/***********************************************************************************************************************
 * Schnittstelle fuer die Prozeduren zur Zeitmessung in der Regelmengenverwaltung 
 ***********************************************************************************************************************
 * Autor:     Juergen Sellentin. Implementiert am 13.02.94
 * Angepasst: Arnim am 04.03.94.  Makros und Funtionen eingesetzt.
 *            Arnim am 04.04.94.  Stack zur Verwaltung von geschachtelten Aufrufen eingebaut.
 *            Arnim bis 24.05.95. Komplett neues Konzept, gettimeofday und Zaehler eingebaut.  
 ***********************************************************************************************************************
 *
 * Bemerkungen zur Zeitmessung:
 * ----------------------------
 *
 * Die Zeitmessungen werden mit der Systemfunktion "getrusage" realisiert. Diese protokolliert die seit dem Beginn des
 * Prozesses verbrauchte Zeit (als user- und system- Anteil).
 *
 * Die Genauigkeit dieser Messung haengt sehr stark von der update-Frequenz von getrusage ab !!!
 * Bei sehr kurzen Prozeduraufrufen werden die getrusage-Daten evtl. NICHT veraendert, so dass die Laufzeit 0
 * als Ergebnis herauskommt. Andererseits kann eine zu hohe Laufzeit gemessen werden, wenn das Aktualisieren der Daten
 * (von getrusage) gerade in einer sehr kurzen Prozedur durchgefuehrt wird.
 * Im statistischen Mittel gleichen sich diese Fehler hoffentlich aus!
 *
 * ACHTUNG: Schachtelungstiefe der Aufrufe begrenzt (siehe Konstante ZM_StackSize im .c-File). Bei hoeherem Bedarf 
 *          die Konstante anpassen, da nicht auf Ueberlauf geprueft wird (Zeit!)
 *
 ***********************************************************************************************************************/


#ifndef ZEITMESSUNG_H
#define ZEITMESSUNG_H

#include "general.h"
#include "InfoStrings.h"
#include "Parameter.h"
#include <sys/time.h>

#ifdef __cplusplus
extern "C" {
#endif


/* ***************************************************************** */
/* * (0) Granularitaet und anderes                                 * */
/* ***************************************************************** */

/* Granularitaet der Zeitmessung:

   Level 0: Nur die Gesamtzeit wird gemessen (dies geschieht immer mit getrusage)
   Level 1: Zusaetzlich werden die Gesamtzeiten fuer 
            * NFBildung,
            * KPBildung mit neuer Regel, 
            * KPBildung mit neuer Gleichung,
            * Interreduktion,
            * Subsumption,
            * Richten der vom Vervollstaendiger angefassten KPs
            gemessen.
   Level 2: Zusaetzlich wird die KP_Bildung in MGU-Suche, Groesser-Test und ggf. Kapur-Test aufgegliedert,
            jeweils unterschieden nach 'zu neuer Regel bzw. Gleichung'.
   Level 3: Zusaetzlich wird die NFBildung in Match-Suche, Termerzeugung, Groesser-Test etc. aufgespalten
   Level 4: Zusaetzlich wird der Zeitbedarf fuer die Verwaltung von KP-, Regel- und Gleichungsmenge sowie die Anlage und Initialisierung
            neuer Speicherbloecke gemessen.
   
   Ferner kann eingestellt werden, ob die Zeitmessung die Funktion gettimeofday (Realzeit mit Aufloesung 1us (nur user-time)) oder getrusage 
   (Prozesszeit mit Aufloesung 10ms (user- und system-time)) verwendet.
   Bei Verwendung von gettimeofday wird keine system-time gemessen. Stattdessen kann noch eingestellt werden, dass die Anzahl der
   Zeitmessungsaufrufe mitgemessen werden soll. Diese Anzahl erscheint dann in der Statistik als Sekunden-Anteil der system-time.
   Ferner wird die Gesamtzahl der Zeitmessungsaufrufe gezaehlt, wenn ZM_STATISTIK auf 1 gesetzt ist (auch bei getrusage).

   Module sind: ZeitMessung.h, ZeitMessung.c
*/

#define ZM_MESSLEVEL 0
/* Alternativen sind: 0 1 2 3 4 */

#define ZM_RUSAGE
/* Alternativen sind: ZM_TIMEOFDAY ZM_RUSAGE */

#define ZM_STATISTIK 0
/* Alternativen: 0 fuer aus, 1 fuer an */
/* Nur relevant bei ZM_TIMEOFDAY */

/* ***************************************************************** */
/* * (I) Korrektheit testen                                        * */
/* ***************************************************************** */

#define ZM_TEST 0
/* Alternativen: 0, 1 */

#if ZM_TEST

void ZM_ZeitStartTest(struct timeval *user, char *Aufruf);
void ZM_ZeitSwitchTest(struct timeval *Stoppuser, char *SwitchAufruf, struct timeval *Startuser);
void ZM_ZeitStoppTest(struct timeval *user, char *Aufruf);
int ZM_ZeitStartTestF(struct timeval *user, char *Aufruf);
int ZM_ZeitSwitchTestF(struct timeval *Stoppuser, char *SwitchAufruf, struct timeval *Startuser);
int ZM_ZeitStoppTestF(struct timeval *user, char *Aufruf);
extern struct timeval ZeitDummy;       /* noetig fuer Zeit-Starts, wo ueberfl./notw. noch nicht bekannt ist */

#else

#define ZM_ZeitStartTest(a,b)      /* Dummy */
#define ZM_ZeitSwitchTest(a,b,c)    /* Dummy */
#define ZM_ZeitStoppTest(a,b)  /* Dummy */
#define ZM_ZeitStartTestF(a,b)      0/* Dummy */
#define ZM_ZeitSwitchTestF(a,b,c)    0/* Dummy */
#define ZM_ZeitStoppTestF(a,b)  0/* Dummy */

#endif 

/* ***************************************************************** */
/* * (II) Initialisierung und Allgemeine Zeitmessung               * */
/* ***************************************************************** */

void ZM_InitApriori(void);
void ZM_ReInitApriori(void);

void ZM_StartGesamtZeit(void);                       /* Die Gesamtzeitmessung wird immer ausgefuehrt! */
void ZM_StoppGesamtZeit(void);

void ZM_Times(unsigned long int *cpu, unsigned long int *wct);
/* Liefert cputime und wallclocktime in msec zurueck (mindestens 1 msec)
   (jedenfalls die bislang gestoppten Zeiten)
*/

/* ***************************************************************** */
/* * (III) Variablen                                               * */
/* ***************************************************************** */

extern struct timeval       ZM_GesamtZeit_ut;               /* insgesamt verbrauchte Zeit exklusive Parsen und Plattklopfen. */
extern struct timeval       ZM_GesamtZeit_st;

/* ... fuer : Reduktion ......................................................................................................*/
extern struct timeval       ZM_NFBildung_ut;                /* enthaelt alle Unter-Zeiten */
extern struct timeval       ZM_NFBildung_st;

extern struct timeval       ZM_MatchenRm_ut;                /* Matchen mit Regel- bzw. GleichungsMENGE (feiner wird nicht gemessen!) */
extern struct timeval       ZM_MatchenRm_st;
extern struct timeval       ZM_MatchenGm_ut;
extern struct timeval       ZM_MatchenGm_st;
extern struct timeval       ZM_TERedGm_ut;                  /* Ueberfluessige(!) Termerzeugung bei RE-Schritt */
extern struct timeval       ZM_TERedGm_st;
extern struct timeval       ZM_Rewrite_ut;
extern struct timeval       ZM_Rewrite_st;
/* ... fuer : Unifikation ......................................................................................................*/
extern struct timeval       ZM_KPBildungZuRegel_ut;         /* enthaelt alle Unter-Zeiten */
extern struct timeval       ZM_KPBildungZuRegel_st;
extern struct timeval       ZM_KPBildungZuGleichung_ut;
extern struct timeval       ZM_KPBildungZuGleichung_st;
extern struct timeval       ZM_MGUZuR_ut;                   /* Zeit fuer MGU-Suche mit neuer Regel */
extern struct timeval       ZM_MGUZuR_st;
extern struct timeval       ZM_MGUZuG_ut;                   /* Zeit fuer MGU-Suche mit neuer Gleichung */
extern struct timeval       ZM_MGUZuG_st;
extern struct timeval       ZM_Kapurtest_ut;                /* Ueberpruefung des Kapur-Kriteriums bei KP-Bildung */
extern struct timeval       ZM_Kapurtest_st;
/* ... fuer : Subsumption ......................................................................................................*/
extern struct timeval       ZM_Subsumption_ut;              /* Zeit fuer Subsumption */
extern struct timeval       ZM_Subsumption_st;
/* ... fuer : Interreduktion ...................................................................................................*/
extern struct timeval       ZM_Interred_ut;                 /* Zeit fuer Interreduktion der verschiedenen Mengen */
extern struct timeval       ZM_Interred_st;
/* ....fuer : Ordnungen........................................................................................................*/
extern struct timeval       ZM_OrdnungEntscheidung_ut;      /* Zeit, die in der Ordnung (Entsch) verbraucht wird.   */
extern struct timeval       ZM_OrdnungEntscheidung_st;
extern struct timeval       ZM_OrdnungRedTest_ut;           /* Zeit, die in der Ordnung (>-Test bei Red.) verbraucht wird.   */
extern struct timeval       ZM_OrdnungRedTest_st;
extern struct timeval       ZM_OrdnungUnifTest_ut;          /* Zeit, die in der Ordnung (>-Test bei Unif.) verbraucht wird.   */
extern struct timeval       ZM_OrdnungUnifTest_st;
/* ....fuer : Parsen und Plattklopfen .........................................................................................*/
extern struct timeval       ZM_PP_ut;                       /* user- / system- Zeit, die zum Parsen und Plattklopfen verbraucht wird. */
extern struct timeval       ZM_PP_st;
/* ... fuer : Regelmengenverwaltung ...........................................................................................*/
extern struct timeval       ZM_RegelEinf_ut;                /* user- / system- Zeit, die zum Einfuegen von Regeln verbraucht wird. */
extern struct timeval       ZM_RegelEinf_st;
extern struct timeval       ZM_RegelLoeschen_ut;            /* user- / system- Zeit, die zum Loeschen von Regeln verbraucht wird. */
extern struct timeval       ZM_RegelLoeschen_st;
/* ... fuer : Gleichungsmengenverwaltung ......................................................................................*/
extern struct timeval       ZM_GleichgEinf_ut;              /* user- / system- Zeit, die zum Einfuegen von Gleichungen verbraucht wird. */
extern struct timeval       ZM_GleichgEinf_st;
extern struct timeval       ZM_GleichgLoeschen_ut;          /* user- / system- Zeit, die zum Loeschen von Gleichungen verbraucht wird. */
extern struct timeval       ZM_GleichgLoeschen_st;
/* ... fuer : Kritische-Paar-Mengen-Verwaltung ................................................................................*/
extern struct timeval       ZM_KPEinf_ut;                   /* user- / system- Zeit, die zum Einfuegen von KP's verbraucht wird. */
extern struct timeval       ZM_KPEinf_st;
extern struct timeval       ZM_KPLoeschen_ut;               /* user- / system- Zeit, die zum Loeschen von KP's verbraucht wird. */
extern struct timeval       ZM_KPLoeschen_st;
extern struct timeval       ZM_KPEinpacken_ut;              /* user- / system- Zeit, die zum Einpacken von KP's verbraucht wird. */
extern struct timeval       ZM_KPEinpacken_st;
extern struct timeval       ZM_KPAuspacken_ut;              /* user- / system- Zeit, die zum Auspacken von gepackten KP's verbraucht wird. */
extern struct timeval       ZM_KPAuspacken_st;
extern struct timeval       ZM_GepackteKPsLoeschen_ut;      /* user- / system- Zeit, die zum Loeschen von gepackten KP's verbraucht wird. */
extern struct timeval       ZM_GepackteKPsLoeschen_st;
/* ... fuer : Speicherverwaltung  ................................................................................*/
extern struct timeval       ZM_NeuerSpeicherBlock_ut;       /* user- / system- Zeit zum Anlegen und Initialisierung neuer Speicherbloecke. */
extern struct timeval       ZM_NeuerSpeicherBlock_st;


/* ***************************************************************** */
/* * (IV) Statistik                                                * */
/* ***************************************************************** */

void ZM_Statistik(InfoStringT IS);

/* ***************************************************************** */
/* * (V) Grundlegende Funktionen                                   * */
/* ***************************************************************** */

/* Diese Prozeduren werden durch die unten realisierten Makros aufgerufen.
   
   Aufrufe bitte nur ueber diese Makros!
*/

void ZM_ZeitStart(void);
/* Speichert den aktuellen Stand von user- und system- Zeit auf dem Zeit-Stack*/

void ZM_ZeitStopp(struct timeval *user,struct timeval *system);
/* Holt die zuletzt auf dem Stack gespeicherte Zeit und addiert die seitdem vergangene Zeit
   zu den uebergebenen Variablen dazu. */

void ZM_ZeitSwitch(struct timeval *alteUT, struct timeval *alteST);
/* Diese Prozedur wechselt vom aktuellen Zeitmessungs-Kontext in einen anderen. Dadurch werden - verglichen mit einer
   Stopp-Start-Kombination - ein Aufruf der Zeitmessung sowie die entsprechenden Stack-Operationen eingespart.
   Bsp.: Waehrend der Unifikation wird die Normalformbildung aufgerufen => Switch vom Unifizieren zum Reduzieren.
   Die Zeit, in deren Kontext gewechselt wurde, kann wieder mit einem ZeitSwitch oder aber mit ZM_Stoppxxx gestoppt
   werden.
*/

int ZM_ZeitStartF(void);

int ZM_ZeitStoppF(struct timeval *user,struct timeval *system);

int ZM_ZeitSwitchF(struct timeval *alteUT, struct timeval *alteST);

void ZM_AlleZeitenAnhalten(void);
/* Haelt alle auf dem Stack liegenden Zeiten 'an'. Weitere Zeitmessungen koennen gestartet werden.
   Beachte: Aufrufe dieser Funktion sind _NICHT_ geschachtelt moeglich.*/
   
void ZM_AlleZeitenWeiter(void);
/* Laesst die angehaltenen Zeiten weiterlaufen.
   Vor.: Alle in der Zwischenzeit auf den Stack gelegten Zeiten sind wieder entfernt worden.
*/

void ZM_ZeitenAddieren(struct timeval *alteUT, struct timeval *alteST, struct timeval *summandUT, struct timeval *summandST);
/* Addiert die summandxT zu altexT, die SummandxT wird anschliessend auf Null gesetzt.*/

void ZM_ZeitStartOhneStack(void);
/* Speichert den aktuellen Stand von in einer Dateiglobalen Variablen.*/

void ZM_ZeitStoppOhneStack(struct timeval *user,struct timeval *system);
/* Holt die zuletzt mit ZM_ZeitStartOhneStack gespeicherte Zeit und addiert
   die seitdem vergangene Zeit zu den uebergebenen Variablen dazu. */

BOOLEAN ZM_ZeitstackOK(void);
/* Ueberprueft den Zeitstack. Liefert TRUE, wenn leer, sonst FALSE. 
   Gibt ferner statistische Informationen ueber die Zeitmessung aus.
   Sollte am Programmende aufgerufen werden.
*/

#if (ZM_MESSLEVEL > 0)
 #ifdef ZM_RUSAGE
  #define ZM_VerwendeteZeitmessung() "getrusage"
  #define ZM_GettimeofdayEingestellt() FALSE
  #define ZM_GetrusageEingestellt() TRUE
 #else
  #define ZM_GetrusageEingestellt() FALSE
  #define ZM_GettimeofdayEingestellt() TRUE
  #define ZM_VerwendeteZeitmessung() "gettimeofday"
 #endif
#else
 #define ZM_VerwendeteZeitmessung() "getrusage"
 #define ZM_GettimeofdayEingestellt() FALSE
 #define ZM_GetrusageEingestellt() TRUE
#endif

#ifdef ZM_AUFRUFE_ZAEHLEN
#define ZM_AufrufeZaehlen() TRUE
#else
#define ZM_AufrufeZaehlen() FALSE
#endif


/* ***************************************************************** */
/* * (VI) Makros zur Zeitmessung                                   * */
/* ***************************************************************** */

/* Zeitmessungen bitte mit den hier vorgegebenen Makros durchfuehren.
   Messungen koennen geschachtelt durchgefuehrt werden, die Start-Zeiten werden auf einem Stack gespeichert.
   Dabei ist aber zu beachten, dass bei jedem Stopp-Befehl die ZULETZT gespeicherte Zeit als zugehoerige
   Start-Zeit vom Stack geholt wird. Verschraenkungen sind also nicht moeglich (und wohl auch nicht sinnvoll)*/



/* Die uebrigen Makros werden zunaechst auf 'Nix tun' eingestellt und sukzessive in Abhaengigkeit vom eingestellten Zeitmessungslevel
   neu besetzt. So erspart man sich die Duplikation der Definitionen (jedes Makros ist nur genau einmal als 'Nix tun' bzw. als
   'aktiviert' definiert.
*/

/* ... fuer : Reduktion ......................................................................................*/

#define ZM_StartNFBildung()                          /* nix tun */
#define ZM_StoppNFBildung()                          /* nix tun */

#define ZM_StartMatchenRm()                          /* nix tun */
#define ZM_StoppMatchenRm()                          /* nix tun */
#define ZM_StoppMatchenRmF()                       0 /* nix tun */

#define ZM_StartMatchenGm()                          /* nix tun */
#define ZM_StoppMatchenGm()                          /* nix tun */

#define ZM_StartTERedGm()                            /* nix tun */
#define ZM_StoppTERedGm()                            /* nix tun */

#define ZM_StartRewrite()                            /* nix tun */
#define ZM_StoppRewrite()                            /* nix tun */

#define ZM_StartOrdnungRedTest()                     /* nix tun */
#define ZM_StoppOrdnungRedTest()                     /* nix tun */
                                        
/* Switches: */
#define ZM_MatchGm2TERedGm()                         /* nix tun */
#define ZM_TERedGm2RedTest()                         /* nix tun */
#define ZM_TestRed2MatchGm()                         /* nix tun */

/* ... fuer : Unifikation.....................................................................................*/

#define ZM_StartKPBildungZuRegel()                   /* nix tun */
#define ZM_StartKPBildungZuRegelF() 0                /* nix tun */
#define ZM_StoppKPBildungZuRegel()                   /* nix tun */
#define ZM_StartKPBildungZuGleichung()               /* nix tun */
#define ZM_StartKPBildungZuGleichungF() 0            /* nix tun */
#define ZM_StoppKPBildungZuGleichung()               /* nix tun */

#define ZM_KPBildungAnhaltenWgKPBehandlung()         /* nix tun */
#define ZM_KPBildungWeiterNachKPBehandlung()         /* nix tun */

#define ZM_StartMGUZuR()                             /* nix tun */
#define ZM_StoppMGUZuR()                             /* nix tun */
#define ZM_StoppMGUZuRF()                          0 /* nix tun */
#define ZM_StartMGUZuG()                             /* nix tun */
#define ZM_StoppMGUZuG()                             /* nix tun */
#define ZM_StoppMGUZuGF()                          0 /* nix tun */

#define ZM_StartKapurtest()                          /* nix tun */
#define ZM_StoppKapurtest()                          /* nix tun */

/* ... fuer : Subsumption.....................................................................................*/

#define ZM_StartSubsumption()                        /* nix tun */
#define ZM_StoppSubsumption()                        /* nix tun */
                                        
/* ... fuer : Interreduktion .................................................................................*/
                                                                         
#define ZM_StartInterred()                           /* nix tun */       
#define ZM_StoppInterred()                           /* nix tun */       
                                                                         
/* ....fuer : Ordnungen......................................................................................*/
                                                                         
#define ZM_StartOrdnungEntsch()                      /* nix tun */       
#define ZM_StoppOrdnungEntsch()                      /* nix tun */       
                                                                         
#define ZM_StartOrdnungUnifTest()                    /* nix tun */       
#define ZM_StoppOrdnungUnifTest()                    /* nix tun */       
                                                                         
/* ....fuer : Parsen und Plattklopfen .......................................................................*/
                                                                         
#define ZM_StartPP()                                 /* nix tun */       
#define ZM_StoppPP()                                 /* nix tun */       
                                                                         
/* ... fuer : Regelmengenverwaltung .........................................................................*/
                                                                         
#define ZM_StartRegelEinf()                          /* nix tun */       
#define ZM_StoppRegelEinf()                          /* nix tun */       
                                                                         
#define ZM_StartRegelLoeschen()                      /* nix tun */       
#define ZM_StoppRegelLoeschen()                      /* nix tun */       
                                                                         
/* ... fuer : Gleichungsmengenverwaltung ....................................................................*/
                                                                         
#define ZM_StartGleichgEinf()                        /* nix tun */       
#define ZM_StoppGleichgEinf()                        /* nix tun */       
                                                                         
#define ZM_StartGleichgLoeschen()                    /* nix tun */       
#define ZM_StoppGleichgLoeschen()                    /* nix tun */       
                                                                         
/* ... fuer : Kritische-Paar-Mengen-Verwaltung ..............................................................*/

#define ZM_StartKPEinf()                             /* nix tun */
#define ZM_StoppKPEinf()                             /* nix tun */

#define ZM_StartKPLoeschen()                         /* nix tun */
#define ZM_StoppKPLoeschen()                         /* nix tun */

#define ZM_StartKPEinpacken()                        /* nix tun */
#define ZM_StoppKPEinpacken()                        /* nix tun */

#define ZM_StartKPAuspacken()                        /* nix tun */
#define ZM_StoppKPAuspacken()                        /* nix tun */

#define ZM_StartGepackteKPsLoeschen()                /* nix tun */
#define ZM_StoppGepackteKPsLoeschen()                /* nix tun */

/* ... fuer : Speicherverwaltung  ............................................................................*/
 
#define ZM_StartSpeicherVerwaltung()                 /* nix tun */
#define ZM_StoppSpeicherVerwaltung()                 /* nix tun */

/* ... fuer : MultipleNormalformen  ..........................................................................*/
#define ZM_StartMatchenMNF()                         /* nix tun */
#define ZM_StoppMatchenMNF()                         /* nix tun */



#if (ZM_MESSLEVEL > 0)

#undef ZM_StartNFBildung
#undef ZM_StoppNFBildung
#define ZM_StartNFBildung()                    {ZM_ZeitStartTest(&ZM_NFBildung_ut,"ZM_StartNFBildung"); ZM_ZeitStart();}
#define ZM_StoppNFBildung()                    {ZM_ZeitStoppTest(&ZM_NFBildung_ut,"ZM_StoppNFBildung"); ZM_ZeitStopp(&ZM_NFBildung_ut, &ZM_NFBildung_st);}

#undef ZM_StartKPBildungZuRegel
#undef ZM_StartKPBildungZuRegelF
#undef ZM_StoppKPBildungZuRegel
#define ZM_StartKPBildungZuRegel()          {ZM_ZeitStartTest(&ZM_KPBildungZuRegel_ut,"ZM_StartKPBildungZuRegel"); ZM_ZeitStart();}
#define ZM_StartKPBildungZuRegelF()         (ZM_ZeitStartTestF(&ZM_KPBildungZuRegel_ut,"ZM_StartKPBildungZuRegelF"), ZM_ZeitStart(), 0)
#define ZM_StoppKPBildungZuRegel()          {ZM_ZeitStoppTest(&ZM_KPBildungZuRegel_ut,"ZM_StoppKPBildungZuRegel"); ZM_ZeitStopp(&ZM_KPBildungZuRegel_ut, &ZM_KPBildungZuRegel_st);}

#undef ZM_KPBildungAnhaltenWgKPBehandlung
#undef ZM_KPBildungWeiterNachKPBehandlung
#define ZM_KPBildungAnhaltenWgKPBehandlung()   ZM_AlleZeitenAnhalten()
#define ZM_KPBildungWeiterNachKPBehandlung()   ZM_AlleZeitenWeiter()

#undef ZM_StartKPBildungZuGleichung
#undef ZM_StartKPBildungZuGleichungF
#undef ZM_StoppKPBildungZuGleichung
#define ZM_StartKPBildungZuGleichung()     {ZM_ZeitStartTest(&ZM_KPBildungZuGleichung_ut,"ZM_StartKPBildungZuGleichung"); ZM_ZeitStart();}
#define ZM_StartKPBildungZuGleichungF()    (ZM_ZeitStartTestF(&ZM_KPBildungZuGleichung_ut,"ZM_StartKPBildungZuGleichungF"), ZM_ZeitStart(),0)
#define ZM_StoppKPBildungZuGleichung()     {ZM_ZeitStoppTest(&ZM_KPBildungZuGleichung_ut,"ZM_StoppKPBildungZuGleichung"); ZM_ZeitStopp(&ZM_KPBildungZuGleichung_ut, &ZM_KPBildungZuGleichung_st);}

#undef ZM_StartSubsumption
#undef ZM_StoppSubsumption
#define ZM_StartSubsumption()                  {ZM_ZeitStartTest(&ZM_Subsumption_ut,"ZM_StartSubsumption"); ZM_ZeitStart();}
#define ZM_StoppSubsumption()                  {ZM_ZeitStoppTest(&ZM_Subsumption_ut,"ZM_StoppSubsumption"); ZM_ZeitStopp(&ZM_Subsumption_ut,&ZM_Subsumption_st);}

#undef ZM_StartInterred
#undef ZM_StoppInterred
#define ZM_StartInterred()                     {ZM_ZeitStartTest(&ZM_Interred_ut,"ZM_StartInterred"); ZM_ZeitStart();}
#define ZM_StoppInterred()                     {ZM_ZeitStoppTest(&ZM_Interred_ut,"ZM_StoppInterred"); ZM_ZeitStopp(&ZM_Interred_ut,&ZM_Interred_st);}

#undef ZM_StartOrdnungEntsch
#undef ZM_StoppOrdnungEntsch
#define ZM_StartOrdnungEntsch()                {ZM_ZeitStartTest(&ZM_OrdnungEntscheidung_ut,"ZM_StartOrdnungEntsch"); ZM_ZeitStart();}
#define ZM_StoppOrdnungEntsch()                {ZM_ZeitStoppTest(&ZM_OrdnungEntscheidung_ut,"ZM_StoppOrdnungEntsch"); ZM_ZeitStopp(&ZM_OrdnungEntscheidung_ut,&ZM_OrdnungEntscheidung_st);}

#endif

#if (ZM_MESSLEVEL > 1)

/* Unifikation dazu */

#undef ZM_StartOrdnungUnifTest
#undef ZM_StoppOrdnungUnifTest
#define ZM_StartOrdnungUnifTest()       {ZM_ZeitStartTest(&ZM_OrdnungUnifTest_ut,"ZM_StartOrdnungUnifTest"); ZM_ZeitStart();}
#define ZM_StoppOrdnungUnifTest()       {ZM_ZeitStoppTest(&ZM_OrdnungUnifTest_ut,"ZM_StoppOrdnungUnifTest"); ZM_ZeitStopp(&ZM_OrdnungUnifTest_ut,&ZM_OrdnungUnifTest_st);}

#undef  ZM_StartMGUZuR
#undef  ZM_StoppMGUZuR
#undef  ZM_StoppMGUZuRF
#undef  ZM_StartMGUZuG
#undef  ZM_StoppMGUZuG
#undef  ZM_StoppMGUZuGF
#define ZM_StartMGUZuR()                   {ZM_ZeitStartTest(&ZM_MGUZuR_ut, "ZM_StartMGUZuR"); ZM_ZeitStart();}
#define ZM_StoppMGUZuR()                   {ZM_ZeitStoppTest(&ZM_MGUZuR_ut, "ZM_StoppMGUZuR"); ZM_ZeitStopp(&ZM_MGUZuR_ut,&ZM_MGUZuR_st);}
#define ZM_StoppMGUZuRF()                  (ZM_ZeitStoppTestF(&ZM_MGUZuR_ut, "ZM_StoppMGUZuRF"), ZM_ZeitStoppF(&ZM_MGUZuR_ut,&ZM_MGUZuR_st))
#define ZM_StartMGUZuG()               {ZM_ZeitStartTest(&ZM_MGUZuG_ut, "ZM_StartMGUZuG"); ZM_ZeitStart();}
#define ZM_StoppMGUZuG()               {ZM_ZeitStoppTest(&ZM_MGUZuG_ut, "ZM_StoppMGUZuG"); ZM_ZeitStopp(&ZM_MGUZuG_ut,&ZM_MGUZuG_st);}
#define ZM_StoppMGUZuGF()              (ZM_ZeitStoppTestF(&ZM_MGUZuG_ut, "ZM_StoppMGUZuGF"), ZM_ZeitStoppF(&ZM_MGUZuG_ut,&ZM_MGUZuG_st))

#undef  ZM_StartKapurtest
#undef  ZM_StoppKapurtest
#define ZM_StartKapurtest()                    {ZM_ZeitStartTest(&ZM_Kapurtest_ut, "ZM_StartKapurtest"); ZM_ZeitStart();}
#define ZM_StoppKapurtest()                    {ZM_ZeitStoppTest(&ZM_Kapurtest_ut, "ZM_StoppKapurtest"); ZM_ZeitStopp(&ZM_Kapurtest_ut,&ZM_Kapurtest_st);}

#endif

#if (ZM_MESSLEVEL > 2)

/* Reduktion dazu */

#undef ZM_StartOrdnungRedTest
#undef ZM_StoppOrdnungRedTest
#define ZM_StartOrdnungRedTest()        {ZM_ZeitStartTest(&ZM_OrdnungRedTest_ut,"ZM_StartOrdnungRedTest"); ZM_ZeitStart();}
#define ZM_StoppOrdnungRedTest()        {ZM_ZeitStoppTest(&ZM_OrdnungRedTest_ut,"ZM_StoppOrdnungRedTest"); ZM_ZeitStopp(&ZM_OrdnungRedTest_ut,&ZM_OrdnungRedTest_st);}

#undef ZM_StartMatchenRm
#undef ZM_StoppMatchenRm
#undef ZM_StoppMatchenRmF
#define ZM_StartMatchenRm()             {ZM_ZeitStartTest(&ZM_MatchenRm_ut,"ZM_StartMatchenRm"); ZM_ZeitStart();}
#define ZM_StoppMatchenRm()             {ZM_ZeitStoppTest(&ZM_MatchenRm_ut,"ZM_StoppMatchenRm"); ZM_ZeitStopp(&ZM_MatchenRm_ut, &ZM_MatchenRm_st);}
#define ZM_StoppMatchenRmF()            (ZM_ZeitStoppTestF(&ZM_MatchenRm_ut,"ZM_StoppMatchenRmF"), ZM_ZeitStopp(&ZM_MatchenRm_ut, &ZM_MatchenRm_st), 0)

#undef ZM_StartMatchenGm
#undef ZM_StoppMatchenGm
#define ZM_StartMatchenGm()             {ZM_ZeitStartTest(&ZM_MatchenGm_ut,"ZM_StartMatchenGm"); ZM_ZeitStart();}
#define ZM_StoppMatchenGm()             {ZM_ZeitStoppTest(&ZM_MatchenGm_ut,"ZM_StoppMatchenGm"); ZM_ZeitStopp(&ZM_MatchenGm_ut, &ZM_MatchenGm_st);}

#undef ZM_StartTERedGm
#undef ZM_StoppTERedGm
#define ZM_StartTERedGm()               {ZM_ZeitStartTest(&ZM_TERedGm_ut,"ZM_StartTERedGm"); ZM_ZeitStart();}
#define ZM_StoppTERedGm()               {ZM_ZeitStoppTest(&ZM_TERedGm_ut,"ZM_StoppTERedGm"); ZM_ZeitStopp(&ZM_TERedGm_ut, &ZM_TERedGm_st);}

#undef ZM_StartRewrite
#undef ZM_StoppRewrite
#define ZM_StartRewrite()               {ZM_ZeitStartTest(&ZM_Rewrite_ut,"ZM_StartRewrite"); ZM_ZeitStart();}
#define ZM_StoppRewrite()               {ZM_ZeitStoppTest(&ZM_Rewrite_ut,"ZM_StoppRewrite"); ZM_ZeitStopp(&ZM_Rewrite_ut, &ZM_Rewrite_st);}

/* Switches: */
#undef ZM_MatchGm2TERedGm
#undef ZM_TERedGm2RedTest
#undef ZM_TestRed2MatchGm

#define ZM_MatchGm2TERedGm()            {ZM_ZeitSwitchTest(&ZM_MatchenGm_ut,"ZM_MatchGm2TERedGm",&ZM_TERedGm_ut); ZM_ZeitSwitchF(&ZM_MatchenGm_ut, &ZM_MatchenGm_st);}
#define ZM_TERedGm2RedTest()            {ZM_ZeitSwitchTest(&ZM_TERedGm_ut,"ZM_TERedGm2RedTest",&ZM_OrdnungRedTest_ut); ZM_ZeitSwitchF(&ZM_TERedGm_ut, &ZM_TERedGm_st);}
#define ZM_TestRed2MatchGm()            {ZM_ZeitSwitchTest(&ZM_OrdnungRedTest_ut,"ZM_TestRed2MatchGm",&ZM_MatchenGm_ut); ZM_ZeitSwitchF(&ZM_OrdnungRedTest_ut, &ZM_OrdnungRedTest_st);}
#endif

#if (ZM_MESSLEVEL > 3)

/* Verwaltung dazu */

#undef ZM_StartPP
#undef ZM_StoppPP
#define ZM_StartPP()                    {ZM_ZeitStartTest(&ZM_PP_ut,"ZM_StartPP"); ZM_ZeitStart();}
#define ZM_StoppPP()                    {ZM_ZeitStoppTest(&ZM_PP_ut,"ZM_StoppPP"); ZM_ZeitStopp(&ZM_PP_ut,&ZM_PP_st);}

#undef ZM_StartRegelEinf
#undef ZM_StoppRegelEinf
#define ZM_StartRegelEinf()             {ZM_ZeitStartTest(&ZM_RegelEinf_ut,"ZM_StartRegelEinf"); ZM_ZeitStart();}
#define ZM_StoppRegelEinf()             {ZM_ZeitStoppTest(&ZM_RegelEinf_ut,"ZM_StoppRegelEinf"); ZM_ZeitStopp(&ZM_RegelEinf_ut,&ZM_RegelEinf_st);}

#undef ZM_StartRegelLoeschen
#undef ZM_StoppRegelLoeschen
#define ZM_StartRegelLoeschen()         {ZM_ZeitStartTest(&ZM_RegelLoeschen_ut,"ZM_StartRegelLoeschen"); ZM_ZeitStart();}
#define ZM_StoppRegelLoeschen()         {ZM_ZeitStoppTest(&ZM_RegelLoeschen_ut,"ZM_StoppRegelLoeschen"); ZM_ZeitStopp(&ZM_RegelLoeschen_ut,&ZM_RegelLoeschen_st);}

#undef ZM_StartGleichgEinf
#undef ZM_StoppGleichgEinf
#define ZM_StartGleichgEinf()           {ZM_ZeitStartTest(&ZM_GleichgEinf_ut,"ZM_StartGleichgEinf"); ZM_ZeitStart();}
#define ZM_StoppGleichgEinf()           {ZM_ZeitStoppTest(&ZM_GleichgEinf_ut,"ZM_StoppGleichgEinf"); ZM_ZeitStopp(&ZM_GleichgEinf_ut,&ZM_GleichgEinf_st);}

#undef ZM_StartGleichgLoeschen
#undef ZM_StoppGleichgLoeschen
#define ZM_StartGleichgLoeschen()       {ZM_ZeitStartTest(&ZM_GleichgLoeschen_ut,"ZM_StartGleichgLoeschen"); ZM_ZeitStart();}
#define ZM_StoppGleichgLoeschen()       {ZM_ZeitStoppTest(&ZM_GleichgLoeschen_ut,"ZM_StoppGleichgLoeschen"); ZM_ZeitStopp(&ZM_GleichgLoeschen_ut,&ZM_GleichgLoeschen_st);}

#undef ZM_StartKPEinf
#undef ZM_StoppKPEinf
#define ZM_StartKPEinf()                {ZM_ZeitStartTest(&ZM_KPEinf_ut,"ZM_StartKPEinf"); ZM_ZeitStart();}
#define ZM_StoppKPEinf()                {ZM_ZeitStoppTest(&ZM_KPEinf_ut,"ZM_StoppKPEinf"); ZM_ZeitStopp(&ZM_KPEinf_ut,&ZM_KPEinf_st);}

#undef ZM_StartKPLoeschen
#undef ZM_StoppKPLoeschen
#define ZM_StartKPLoeschen()            {ZM_ZeitStartTest(&ZM_KPLoeschen_ut,"ZM_StartKPLoeschen"); ZM_ZeitStart();}
#define ZM_StoppKPLoeschen()            {ZM_ZeitStoppTest(&ZM_KPLoeschen_ut,"ZM_StoppKPLoeschen"); ZM_ZeitStopp(&ZM_KPLoeschen_ut,&ZM_KPLoeschen_st);}

#undef ZM_StartKPEinpacken
#undef ZM_StoppKPEinpacken
#define ZM_StartKPEinpacken()           {ZM_ZeitStartTest(&ZM_KPEinpacken_ut,"ZM_StartKPEinpacken"); ZM_ZeitStart();}
#define ZM_StoppKPEinpacken()           {ZM_ZeitStoppTest(&ZM_KPEinpacken_ut,"ZM_StoppKPEinpacken"); ZM_ZeitStopp(&ZM_KPEinpacken_ut,&ZM_KPEinpacken_st);}

#undef ZM_StartKPAuspacken
#undef ZM_StoppKPAuspacken
#define ZM_StartKPAuspacken()           {ZM_ZeitStartTest(&ZM_KPAuspacken_ut,"ZM_StartKPAuspacken"); ZM_ZeitStart();}
#define ZM_StoppKPAuspacken()           {ZM_ZeitStoppTest(&ZM_KPAuspacken_ut,"ZM_StoppKPAuspacken"); ZM_ZeitStopp(&ZM_KPAuspacken_ut,&ZM_KPAuspacken_st);}

#undef ZM_StartGepackteKPsLoeschen
#undef ZM_StoppGepackteKPsLoeschen
#define ZM_StartGepackteKPsLoeschen()   {ZM_ZeitStartTest(&ZM_GepackteKPsLoeschen_ut,"ZM_StartGepackteKPsLoeschen"); ZM_ZeitStart();}
#define ZM_StoppGepackteKPsLoeschen()   {ZM_ZeitStoppTest(&ZM_GepackteKPsLoeschen_ut,"ZM_StoppGepackteKPsLoeschen"); ZM_ZeitStopp(&ZM_GepackteKPsLoeschen_ut,&ZM_GepackteKPsLoeschen_st);}

#undef ZM_StartSpeicherVerwaltung
#undef ZM_StoppSpeicherVerwaltung
#define ZM_StartSpeicherVerwaltung()    {ZM_ZeitStartTest(&ZM_NeuerSpeicherBlock_ut,"ZM_StartSpeicherVerwaltung"); ZM_ZeitStart();}
#define ZM_StoppSpeicherVerwaltung()    {ZM_ZeitStoppTest(&ZM_NeuerSpeicherBlock_ut,"ZM_StoppSpeicherVerwaltung"); ZM_ZeitStopp(&ZM_NeuerSpeicherBlock_ut,&ZM_NeuerSpeicherBlock_st);}

#endif


/* Konzept zur Korrektheits-Pruefung der Zeitmessungs-AUFRUFE aus den verschiedenen Modulen:
   -----------------------------------------------------------------------------------------

   Geprueft werden soll hauptsaechlich, ob die mit ZM_Stopp??? beendeten Zeiten dieselben sind,
   die zuvor mit ZM_Start??? angestossen wurden. Bislang wird durch ZM_Start??? lediglich die
   aktuelle Prozess-Zeit auf den Zeit-Stack gelegt. Beim naechsten Stopp oder Switch wird die
   seither verstrichene Zeit der hierbei uebergebenen Zeit-Variablen aufaddiert.

   Zur Ueberpruefung der Start/Stopp-Paare auf Konsistenz sind also folgende Massnahmen erforderlich:
   1) Bei jedem Start bzw. Switch ist die angestossene Zeit-Variable mitzuspeichern
   2) Bei jedem Stopp bzw. Switch ist die angegebene Zeit-Variable mit der angestossenen zu vergleichen.

   Zur Realisierung sind folgende Massnahmen erforderlich:
   - zusaetzlicher Stack, auf den bei Start und Switch die Adressen der angestossenen User-Time-Variablen
     abgelegt werden koennen.
   - dazu sollte der Name des aufgerufenen Makros als String (zur Fehler-Kontrolle) abgelegt werden.
   - Bei jedem Stopp und Switch muss die abgeholte Zeit mit der angestossenen verglichen werden.

   Hierzu gibt es zwei Moeglichkeiten:
   1) Alles wird in den bisher existierenden Prozeduren verborgen.
      Vorteile:  - keine Neu-Uebersetzung des ganzen Systems, wenn von/auf Testbetrieb umgeschaltet wird
      Nachteile: - die unterliegenden Prozeduren erhalten im Nicht-Testfall zuviele Parameter
                   ==> Warnungen waehrend der Uebersetzung
   2) Die Test-Prozeduren werden in den Makros zusaetzlich zu den uebrigen Prozeduren aufgerufen
      Vorteile:  - Im Nicht-Testbetrieb gleiche Effizienz wie jetzt
      Nachteile: - Neu-Uebersetzung des kompletten Systems bei Betriebs-Umschaltung.

   Wegen Effizienz wird Variante 2 vorgezogen.

   Test-Prozeduren:

   ZM_ZeitStartTest(struct timeval *user, char *Aufruf);
   (* Die Adresse der angestossenen User-Zeit-Variablen sowie der Name des aufgerufenen Makros
      werden auf einem Stack gespeichert.
    *)

   ZM_ZeitSwitchTest(struct timeval *Stoppuser, char *StoppAufruf, struct timeval *Startuser, char *StartAufruf);
   (* Die Adresse der gestoppten User-Time wird mit derjenigen auf dem Test-Stack verglichen.
      Bei Unterschied werden der gespeicherte Start-Aufruf sowie der Stopp-Aufruf ausgegeben.
      Ferner wird ein Fehler-Zaehler erhoeht, der spaeter bei ZM_ZeitstackOK ueberprueft wird.

      Anschliessend wird der alte, oberste Stackeintrag durch Startuser und StartAufruf ersetzt.
    *)

   ZM_ZeitStoppTest(struct timeval *user, char *Aufruf);
   (* Die Adresse der gestoppten User-Time wird mit derjenigen auf dem Test-Stack verglichen.
      Bei Unterschied werden der gespeicherte Start-Aufruf sowie der Stopp-Aufruf ausgegeben.
      Ferner wird ein Fehler-Zaehler erhoeht, der spaeter bei ZM_ZeitstackOK ueberprueft wird.
    *)

*/


#ifdef __cplusplus
}
#endif

#endif
