/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Hauptkomponenten.c                                                                                         */
/* - *******************                                                                                        */
/* -                                                                                                            */
/* - Enthaelt die Operationen, die zum Programm-Start mit den in der KP-Menge enthaltenen Gleichungen der Spezi-*/
/* - fikation ausgefuehrt werden (I) sowie eine Prozedur, die alle KP's der Reihe nach behandelt und sie zu-    */
/* - sammenzufuehren sucht (II).                                                                                */
/* - 29.06.94 Arnim. AlsRegel/GleichungAufnehmen fertiggemacht.                                                 */
/* - 10.94/1.11.94 Arnim. Prozeduren fuer die failing completion eingebaut.                                     */
/* -------------------------------------------------------------------------------------------------------------*/
#include "compiler-extra.h"
#include "general.h"

#include "Ausgaben.h"
#include "DSBaumOperationen.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "Interreduktion.h"
#include "KBO.h"
#if COMP_WMTW
#include "KPBildung.h"
#endif
#include "KPMengenInterreduktion.h"
#include "LPO.h"
#include "MatchOperationen.h"
#include "MNF.h"
#include "NFBildung.h"
#include "Ordnungsgenerator.h"
#include "Parameter.h"
#if COMP_WMTW
#include "PriorityQueue.h"
#endif
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "SelectClassify.h"
#include "SpezNormierung.h"
#include "Statistiken.h"
#include "SUEM.h"
#include "Subgleichungen.h"
#include "Subsumption.h"
#include "Termpaare.h"
#include "Unifikation1.h"
#include "ZeitMessung.h"
#include "Zielverwaltung.h"
#include "general.h"
#include "pcl.h"
#include "Ordnungen.h"
#include <string.h>
#include "math.h"
/* -------------------------------------------------------------------------------*/
/* --------------- KPMenge Behandeln (I und II) ----------------------------------*/
/* -------------------------------------------------------------------------------*/

static void AlsGleichungAufnehmen(TermT links, TermT rechts, BOOLEAN ProofModus);
static void AlsRegelAufnehmenOhneGM(BOOLEAN Kreuzen);
static void AlsRegelAufnehmen_FAILING(TermT lhs, TermT rhs, BOOLEAN Kreuzen, BOOLEAN ProofModus);

/* (I) */
/*-----*/

BOOLEAN HK_GewichtsloseKPsBehandeln(void)
/* Behandelt der Reihe nach die kritischen Paare mit Gewicht Null (also die Ausgangsgleichungen) in
   Abhaengigkeit von den verschiedenen (einfachen!) Modi wie folgt:
   
   reduction:        Richten links->rechts,  -> R.
   termination:      Richten nach Ordnung, richtbare -> R, unrichtbare -> TP-Puffer.
   confluence :      Richten links->rechts, -> R. KP-Bildung.
   convergence:      Richten nach Ordnung, richtbare -> R, unrichtbare -> TP-Puffer. KPBildung.
   
   Rueckgabe                  == TRUE   : Alle behandelten KP's konnten gerichtet werden.
                                 FALSE  : nicht alle behandelten KP's konnten gerichtet werden.

   Seiteneffekte                        : Nicht richtbare Ausgangsgleichungen werden ausgegeben und in TP-Puffer aufgenommen.

   Vor.: Alle KPs mit Gewicht 0 sind noch gueltig.
   In allen Modi bleibt die Gleichungsmenge leer.
*/
{
   BOOLEAN AlleRichtbar = TRUE;
   TermT links, rechts;
   RegelOderGleichungsT rog;

#ifndef _ACIDWM_AN
   IO_DruckeUeberschrift("Gleichungen aus der Spezifikation behandeln");
   IO_DruckeFlex("... Richten %s und \n    %s Interreduktion und KP-Bildung in Regel- bzw. Gleichungsmenge aufnehmen.\n\n",
                 ((PA_ExtModus == termination)||(PA_ExtModus==convergence))? "nach Ordnung":"von links nach rechts",
                 ((PA_ExtModus == confluence)||(PA_ExtModus==convergence))? "mit":"ohne");
#endif


   if ((PA_ExtModus == confluence) || (PA_ExtModus == convergence) || (PA_ExtModus == termination))
      PU_TPPufferNeu();

   while(SUE_SelectInitial()){
      links = SUE_selectedCPlhs();
      rechts = SUE_selectedCPrhs();
      
      switch (PA_ExtModus){
        
      case confluence:                                                      /* Konfluenz */
        SUE_selectedCPRE();
        AlsRegelAufnehmenOhneGM(FALSE);
        break;
        
      case convergence:                                                     /* Konvergenz */
        switch(ORD_VergleicheTerme(links, rechts)){
        case Groesser:
           SUE_selectedCPRE();
           AlsRegelAufnehmenOhneGM(FALSE);
           break;
        case Kleiner:
           SUE_selectedCPRE();
           AlsRegelAufnehmenOhneGM(TRUE);
           break;
        case Unvergleichbar:
           /* In diesem Falle werden die KPs zu den nachfolgenden Regeln mit der Regelmenge weiterhin
              gebildet. Da sie aber nicht behandelt werden, wird das Zeitverhalten davon nicht merklich 
              beeinflu\3t .*/
           SUE_selectedCPRE();
           PU_TPEinfuegen(links, rechts, SUE_selectedCPNr());
           AlleRichtbar = FALSE;
           break;
        default:
           SUE_selectedCPjoined();
           break;
        }
        break;
        
      case reduction:                                                       /* Reduktion */
        RE_IncAnzRegelnGesamt();
        rog = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
#ifndef _ACIDWM_AN
        IO_DruckeMeldungNeueRegelAufgenommen(RE_AnzRegelnGesamt(),links, rechts);
#endif
        RE_ZuletztErzeugteRegelEinfuegen();
        SUE_selectedCPRE();
        break; 
        
      case termination:                                                      /* Termination */
        switch(ORD_VergleicheTerme(links, rechts)){
        case Groesser:
          RE_IncAnzRegelnGesamt();
          rog = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
          IO_DruckeMeldungNeueRegelAufgenommen(RE_AnzRegelnGesamt(), links, rechts);
          RE_ZuletztErzeugteRegelEinfuegen();
          SUE_selectedCPRE();
          break;
        case Kleiner:
          RE_IncAnzRegelnGesamt();
          rog = RE_ErzeugteRegel(rechts, links, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
          IO_DruckeMeldungNeueRegelAufgenommen(RE_AnzRegelnGesamt(), rechts, links);
          RE_ZuletztErzeugteRegelEinfuegen();
          SUE_selectedCPRE();
          break;
        case Unvergleichbar:
          RE_IncAnzGleichungenGesamt();
          PU_TPEinfuegen(links, rechts, SUE_NummerAktKP());
          SUE_selectedCPRE();
          AlleRichtbar = FALSE;
          break;
        default:
          SUE_selectedCPjoined();
          break;
        }
        break;
    default: break;                    /* in den Modi proof und completion wird diese Routine sowieso nicht aufgerufen */
    }
  }
#ifndef _ACIDWM_AN
  IO_LF();
#endif
  return AlleRichtbar;
}

/* (II) */
/*------*/

BOOLEAN HK_KPsZusammenfuehrbar(void)
/* Behandelt alle kritischen Paare der Reihe nach (wie sie in der KP-Menge zu finden sind) wie folgt:
   Es wird versucht, linke und rechte Seite zusammenzufuehren. War dies moeglich,
   so wird fortgefahren, andernfalls wird das kritische Paar (nach der Reduktion!) ausgegeben
   Der Rueckgabewert gibt an, ob alle KP's zusammenfuehrbar waren. 
   Seiteneffekt : nicht richtbare KP's werden ausgegeben und im TP-Puffer gespeichert.
*/
{
   BOOLEAN AlleZusammenfuehrbar = TRUE;
   TermT links, rechts;
   
   IO_DruckeUeberschrift("Kritische Paare auf Zusammenfuehrbarkeit testen");
   PU_TPPufferNeu();
   while(SUE_Select()){
      links = SUE_selectedCPlhs();
      rechts = SUE_selectedCPrhs();
      PCL_Reduktionsbeginn(Gleichung, SUE_selectedCPPclID(), linkeSeite, rechts);
      NF_NFBildungZuRegeln(links, KPsSel);
      PCL_Reduktionsbeginn(Gleichung, SUE_selectedCPPclID(), rechteSeite, links);
      NF_NFBildungZuRegeln(rechts, KPsSel);
      if (!TO_TermeGleich(links, rechts)){
         RE_IncAnzGleichungenGesamt();
         IO_DruckeMeldungMeldungKPnichtZusammengefuehrt(links,rechts);
         PU_TPEinfuegen(links, rechts, SUE_NummerAktKP());
         SUE_selectedCPRE();
         AlleZusammenfuehrbar = FALSE;
      }
      else
         SUE_selectedCPjoined();
   }
   IO_DruckeFlex("\n\n\nEs waren %salle kritischen Paare zusammenfuehrbar. :-%s\n\n",
		 (AlleZusammenfuehrbar)? "":"nicht ",
		 (AlleZusammenfuehrbar)? ")":"(");
   return AlleZusammenfuehrbar;
}


/* -------------------------------------------------------------------------------*/
/* --------------- Ziel-Behandlung (III) -----------------------------------------*/
/* -------------------------------------------------------------------------------*/

void HK_GleichungenUndZieleHolen(void)
/* Die Gleichungen aus der Spezifikation plattklopfen und mit Gewicht 0 in die KP-Menge stecken.
   Fuer jedes Symbol testen, ob es AC-Eigenschaft hat.
   Die Ziele aus der Spezifikation plattklopfen, dabei skolemisieren und in die Ziele-Liste stecken.*/
{
   int Index;
   SymbolT SymbolIndex;
   BOOLEAN 
     *IstASymbol = array_alloc(SO_AnzahlFunktionen(),BOOLEAN),
     *IstCSymbol = array_alloc(SO_AnzahlFunktionen(),BOOLEAN),
     ExistenzVonACSymbolen = FALSE;

#if !COMP_ACIDWM
   IO_DruckeUeberschrift("Fetching goals and equations");
#endif

   if (!(PA_ExtModus == completion) && !(PA_ExtModus == proof))
      IO_AusgabenAus();

   SO_forFktSymbole(SymbolIndex)
     IstASymbol[SymbolIndex] = IstCSymbol[SymbolIndex] = FALSE;

   Z_ZieleAusSpezifikationHolen();
     
   for (Index=0; Index<SN_AnzahlGleichungen; Index++){
     term  Term1 = SN_LHS[Index], 
           Term2 = SN_RHS[Index];
     TermT Flunder1, Flunder2;
     TO_PlattklopfenLR(Term1,Term2,&Flunder1,&Flunder2);
     BO_TermpaarNormierenAlsKP(Flunder1, Flunder2);
     if (TO_IstAssoziativitaet(Flunder1,Flunder2))
       IstASymbol[TO_TopSymbol(Flunder1)] = TRUE;
     if (TO_IstKommutativitaet(Flunder1,Flunder2))
       IstCSymbol[TO_TopSymbol(Flunder1)] = TRUE;
     SUE_CPinitial(Flunder1, Flunder2);
     OG_GleichungMelden(Flunder1, Flunder2);
     SUE_recentCPinsert();
   }

   if (!(PA_ExtModus == completion) && !(PA_ExtModus == proof)) IO_AusgabenEin();

   SO_forFktSymbole(SymbolIndex)
     if ((SO_IstACSymbol[SymbolIndex] = IstASymbol[SymbolIndex] && IstCSymbol[SymbolIndex])) 
       ExistenzVonACSymbolen = TRUE;
   our_dealloc(IstASymbol);
   our_dealloc(IstCSymbol);
}


/* (III) */
/*-------*/


BOOLEAN HK_ZieleAufZusammenfuehrbarkeitTesten(void)
/* Alle Ziele aus der Zielmenge bzgl. des aktuellen Regel- und Gleichungs-Systems abschliessend auf NF bringen. */
{
   BOOLEAN alleZusammengefuehrt=TRUE;

   IO_DruckeUeberschrift("Treatment of Goals");
   if (Z_ZielmengeLeer()){
     fprintf(IO_Strom, "\n... there are no goals!\n\n");
     return TRUE;
   }

   alleZusammengefuehrt = Z_FinalerZielDurchlauf();

   IO_DruckeFlex("\n\n\n%s goals were joinable.%s\n\n\n",
		 (alleZusammengefuehrt)? "All":"Not all",
		 (alleZusammengefuehrt)? ":-)":":-(");
   return alleZusammengefuehrt;
}


/* -------------------------------------------------------------------------------*/
/* --------------- Aufnahme neuer Regeln/Gleichungen in RM/GM---------------------*/
/* -------------------------------------------------------------------------------*/

static RegelT zuletztEingefuegteRegel;
static GleichungsT zuletztEingefuegteGleichung;

static void AlsRegelAufnehmenOhneGM(BOOLEAN Kreuzen)
/* Nimmt kp.links=kp.rechts als Regel auf, und zwar von links nach rechts, falls Kreuzen == FALSE,
   ist Kreuzen == TRUE, dann von rechts nach links, fuehrt die Interreduktion der rechten Seiten der Regelmenge 
   sowie die KP-Bildung durch.
   Gleichungen werden nicht beachtet!!!!!!!
   Vor.: linke und rechte Seite des KPs sind auf NF bzgl. R!
   Diese Prozedur wird in HK_GewichtsloseKPsbehandlen in den Modi confluence und convergence verwendet.
   Es erfolgt keine Interreduktion auf den linken Regel-Seiten!
   Der Aufrufparameter Kreuzen wurde eingefuehrt, um die Orientierungsrichtung an das PCL-Protokoll weiter-
   geben zu koennen.*/
{
   TermT links,rechts;

   
   if (Kreuzen){                                                     /* Richtige Richtung herausfinden: */
     links = SUE_selectedCPrhs();
     rechts = SUE_selectedCPlhs();
   }
   else{
     links = SUE_selectedCPlhs();
     rechts = SUE_selectedCPrhs();
   }

   RE_IncAnzRegelnGesamt();

   zuletztEingefuegteRegel = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());

   IO_DruckeMeldungNeueRegelAufgenommen(RE_AnzRegelnGesamt(),links,rechts);

   PCL_KPZuRegel(links, rechts, Kreuzen ? UeberKreuz:VonLinksNachRechts, 
                 TP_PclID(zuletztEingefuegteRegel));

   PCL_NeueRegel(TP_PclID(zuletztEingefuegteRegel), TP_ElternNr(zuletztEingefuegteRegel));

   RE_ZuletztErzeugteRegelEinfuegen();
   
   U1_KPsBildenZuRegelFailing(zuletztEingefuegteRegel);
   /* ausser auf Toplevel KP-Bildung zwischen der neuen Regel und sich selbst.
      Ferner alle KP's zwischen l->r und RM bilden und in den KPPuffer einfuegen.
      Dabei dient das KP als Eltern-Verweis fuer das PCL-Protokoll.*/

   /* =====> KPMenge ist jetzt komplett; */
}

static void AlsGleichungAufnehmen(TermT links, TermT rechts, BOOLEAN ProofModus)
/* Nimmt selectedCP als Gleichung auf, fuehrt die Interreduktion sowie die KP-Bildung durch.
   links und rechts sind auf NF bzgl. R und R(E)!
   Rueckgabe == TRUE, wenn die neue Gleichung nicht wegsubsummiert wurde!*/
{
   RE_IncAnzGleichungenGesamt();
   
   zuletztEingefuegteGleichung = RE_ErzeugteGleichung(links, rechts, RE_AnzGleichungenGesamt(), SUE_selectedCPPclID());
   
   IO_DruckeMeldungGleichungInGMAufgenommen(links, rechts, TP_IstMonogleichung(zuletztEingefuegteGleichung));
   
   IR_NeueGleichungIst(zuletztEingefuegteGleichung);
   
   IR_InterreduktionStarten();						    /* Meldung; Puffer vorbereiten */
   
   IR_GMSubsummierenMitGleichung();
   /* Suche alle mit l=r oder r=l subsummierbaren Gleichungen
      und entferne sie aus der GM (---> wegwerfen, nicht puffern!!!)*/
   
   IR_GMInterredMitGleichung();
   /* Suche alle mit l=r oder r=l reduzierbaren Gleichungen, fuehre den Reduktionsschritt
      aus und entferne sie aus der GM (--->PU_REPuffer)*/
   
   IR_RMLinksInterredMitGleichung();
   /* Suche alle links mit l=r oder r=l reduzierbaren Regeln, fuehre den Reduktionsschritt aus und entferne sie aus RM
      (---> PU_REPuffer) */
   
   /* ====> GM und RM sind links interreduziert */
   
   RE_ZuletztErzeugteHinrichtungEinfuegen(SGM_SubgleichungenZuObjektAufnehmen((SGM_BetroffenesObjektT) zuletztEingefuegteGleichung));
   /* Fuege l=r als Gleichung in GM ein. Spaete Aufnahme erspart Ausschluss bei
      Subsumption und Interred. der GM. */
   
   RE_ZuletztErzeugteHerrichtungEinfuegen();
   
   PCL_NeueGleichung(TP_PclID(zuletztEingefuegteGleichung), TP_ElternNr(zuletztEingefuegteGleichung));

   /* ====> NF-Bildung wieder moeglich (dh. minimal) */
   
   IR_RMRechtsInterredMitGleichung();
   /* Suche alle rechts mit l=r oder r=l reduzierbaren Regeln, fuehre den Reduktionsschritt aus 
      und fahre mit der NF-Bildung fort*/
   
   IR_InterreduktionBeendet();
   
   /* ====> RM ist auch rechts interreduziert */
   
   /* ====> System ist interreduziert */
   
   if (ProofModus){
      (void)Z_ZieleErledigtMitNeuerGleichung(zuletztEingefuegteGleichung);
      if (Z_ZieleBewiesen()){  /* Falls Ziele erledigt, keine KP-Bildung mehr! */
         IO_AusgabenAus();     /* Dafuer aber noch den IR-Puffer abarbeiten!!! */
         IR_PufferAuslesen();
         SUE_NewParent(zuletztEingefuegteGleichung, NULL, NULL);
         SUE_RedirectParent();
         SUE_FinishParent();
         IO_AusgabenEin();
         return;
      }
      /* ====> Ziele sind bearbeitet */
   }
   
   KPIR_KPMengeInterreduzieren(); 
   /* Interreduktion der KP-Menge erst hier moeglich, da vorher NF-Bildung nicht vollst.*/
   /* insb. auch vor KP-Bildung (sonst doppelte Arbeit auf den neuen KPs) */
   
   U1_KPsBildenZuGleichung(zuletztEingefuegteGleichung);
   
   IR_PufferAuslesen();
   
   /* =====> KPMenge ist jetzt komplett; */
   
}

static void AlsRegelAufnehmen_UNFAILING(TermT lhs, TermT rhs, BOOLEAN Kreuzen, BOOLEAN ProofModus)
/* Nimmt links=rechts als Regel auf, fuehrt die Interreduktion sowie die KP-Bildung durch.
   links und rechts sind auf NF bzgl. R und R(E)!
   Der Aufrufparameter Kreuzen wurde eingefuehrt, um die Orientierungsrichtung an das PCL-Protokoll weiter-
   geben zu koennen.*/
{
   TermT links,rechts;

   if (Kreuzen){
     links = rhs;
     rechts = lhs;
   }
   else{
     links = lhs;
     rechts = rhs;
   }

   RE_IncAnzRegelnGesamt();

   zuletztEingefuegteRegel = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());

   IO_DruckeMeldungNeueRegelAufgenommen(RE_AnzRegelnGesamt(),links,rechts);

   IR_NeueRegelIst(zuletztEingefuegteRegel);

   IR_InterreduktionStarten();

   PCL_KPZuRegel(links, rechts, Kreuzen ? UeberKreuz:VonLinksNachRechts, 
                 TP_PclID(zuletztEingefuegteRegel));

   PCL_NeueRegel(TP_PclID(zuletztEingefuegteRegel), TP_ElternNr(zuletztEingefuegteRegel));

   IR_RMLinksInterredMitRegel();
   /* Suche alle Regeln, deren linke Seite von l gematcht werden, fuehre diesen einen Reduktionsschritt aus und
      entferne sie aus der Regelmenge; (wohin? ---> Puffer).
      Vorgezogen, um Ausschluss ueberfluessig zu machen */

   IR_GMInterredMitRegel();
   /* Suche alle mit l->r reduzierbaren Gleichungen, fuehre diesen einen Reduktionsschritt aus und entferne sie aus GM; 
      (wohin? ---> Puffer) */

   /* =====> GM ist wieder interreduziert */

   /* =====> NF-Bildung ist wieder konsistent.*/

   RE_ZuletztErzeugteRegelEinfuegen();                 /* Fuege l->r als neue Regel in RM ein; */

   IR_RMRechtsInterredMitRegel();
   /* Suche alle rechts mit l->r reduzierbaren Regeln, fuehre den Reduktionsschritt aus und
      rufe die Normalformbildung dazu auf */

   /* =====> RM ist wieder interreduziert */

   IR_InterreduktionBeendet();

   /* =====> System ist interreduziert */

   if (ProofModus){
      (void)Z_ZieleErledigtMitNeuerRegel(zuletztEingefuegteRegel);
      if (Z_ZieleBewiesen()){  /* Falls Ziele erledigt, keine KP-Bildung mehr! */
         IO_AusgabenAus();      /* Dafuer aber noch den IR-Puffer abarbeiten!!! */
         IR_PufferAuslesen();
         SUE_NewParent(zuletztEingefuegteRegel, NULL, NULL);
         SUE_FinishParent();
         IO_AusgabenEin();
         return;
      }
         /* ====> Ziele sind bearbeitet */
   }

   KPIR_KPMengeInterreduzieren(); /* Interreduktion der KP-Menge vor KP-Bildung (sonst doppelte Arbeit auf den neuen KPs) */

   U1_KPsBildenZuRegelUnfailing(zuletztEingefuegteRegel);
   /* ausser auf Toplevel KP-Bildung zwischen der neuen Regel und sich selbst.
      Ferner alle KP's zwischen l->r und RM bilden und in den KPPuffer einfuegen.*/

   IR_PufferAuslesen();

   /* =====> KPMenge ist jetzt komplett; */
}


static void AlsRegelAufnehmen_FAILING(TermT lhs, TermT rhs, BOOLEAN Kreuzen, BOOLEAN ProofModus)
/* Nimmt links=rechts als Regel auf, fuehrt die Interreduktion sowie die KP-Bildung durch.
   links und rechts sind auf NF bzgl. R !
   Failing-Version ==> keine Beruehrung der Gleichungsmenge.
   Der Aufrufparameter Kreuzen wurde eingefuehrt, um die Orientierungsrichtung an das PCL-Protokoll weiter-
   geben zu koennen.*/
{
   TermT links,rechts;


   if (Kreuzen){
     links = rhs;
     rechts = lhs;
   }
   else{
     links = lhs;
     rechts = rhs;
   }

   RE_IncAnzRegelnGesamt();

   zuletztEingefuegteRegel = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());

   IO_DruckeMeldungNeueRegelAufgenommen(RE_AnzRegelnGesamt(),links,rechts);

   IR_NeueRegelIst(zuletztEingefuegteRegel);

   IR_InterreduktionStarten();  /* RE-Puffer initialisieren */

   PCL_KPZuRegel(links, rechts, Kreuzen ? UeberKreuz:VonLinksNachRechts, 
                 TP_PclID(zuletztEingefuegteRegel));

   PCL_NeueRegel(TP_PclID(zuletztEingefuegteRegel), TP_ElternNr(zuletztEingefuegteRegel));

   IR_RMLinksInterredMitRegel();
   /* Suche alle Regeln, deren linke Seite von l gematcht werden, fuehre diesen einen Reduktionsschritt aus und
      entferne sie aus der Regelmenge; (wohin? ---> Puffer).
      Vorgezogen, um Ausschluss ueberfluessig zu machen */

   /* =====> NF-Bildung ist wieder konsistent.*/

   RE_ZuletztErzeugteRegelEinfuegen();

   IR_RMRechtsInterredMitRegel();
   /* Suche alle rechts mit l->r reduzierbaren Regeln, fuehre den Reduktionsschritt aus und
      rufe die Normalformbildung dazu auf */

   /* =====> RM ist wieder interreduziert */

   /* =====> System ist interreduziert, entfernte Regeln in KPMenge, ggf. NF-Bildung beenden */

   IR_InterreduktionBeendet();

   if (ProofModus){
      (void)Z_ZieleErledigtMitNeuerRegel(zuletztEingefuegteRegel);
      if (Z_ZieleBewiesen()){  /* Falls Ziele erledigt, keine KP-Bildung mehr! */
         IO_AusgabenAus();      /* Dafuer aber noch den IR-Puffer abarbeiten!!! */
         IR_PufferAuslesen();
         SUE_NewParent(zuletztEingefuegteRegel, NULL, NULL); /* So tun als ob! */
         SUE_FinishParent();
         IO_AusgabenEin();
         return;
      }
         /* ====> Ziele sind bearbeitet */
   }

   KPIR_KPMengeInterreduzieren(); /* KP-Mengen-IR vor der Aufnahme der aktuellen KPs */

   U1_KPsBildenZuRegelFailing(zuletztEingefuegteRegel);
   /* ausser auf Toplevel KP-Bildung zwischen der neuen Regel und sich selbst.
      Ferner alle KP's zwischen l->r und RM bilden und in den KPPuffer einfuegen.*/

   IR_PufferAuslesen();

   /* =====> KPMenge ist jetzt komplett; */
}

/* -------------------------------------------------------------------------------*/
/* --------------- Vervollstaendigung (IV) ---------------------------------------*/
/* -------------------------------------------------------------------------------*/
/* Diejenigen Prozeduren mit einem grossen Eff ('F') als letztem Buchstaben sind  */
/* die jeweiligen abbrechenden Versionen der Vervollstaendigung                   */
/* (failing completion).                                                          */
/* -------------------------------------------------------------------------------*/
static BOOLEAN TWInitialeHolen= FALSE;

#define MeldungBerechnungBeendet()                                              \
  if (!TWInitialeHolen)                                                        \
    fprintf(IO_Strom, "\nsystem completed%s.\n", PA_unfair()?" (unfair)":"")

static unsigned short PermGrad_E, PermGrad_R;
static BOOLEAN VervollstaendigungAbgebrochen, Normalform_R=FALSE, Normalform_E=FALSE, Subsumption=FALSE;

static BOOLEAN Proofmodus;

#define Unvergleichbar_UNFAILING__                                                                                          \
if (!SUE_selectedCPpostponed()){                                                                                            \
   /* kp als neue Gleichung aufnehmen, falls nicht zurueckstellbar */                                                       \
   IO_DruckeMeldungNeueGleichungGebildet(RE_AnzGleichungenGesamt() + 1, links, rechts); /* vorgezogen wg. Interreduktion */ \
   SUE_selectedCPRE();                                                                                                      \
   AlsGleichungAufnehmen(SUE_selectedCPlhs(), SUE_selectedCPrhs(), Proofmodus);                                                                                       \
}

#define Unvergleichbar_UNFAILING_S                                                                                             \
if (!SUE_selectedCPpostponed()){                                                                                               \
   /* kp als neue Gleichung aufnehmen, falls nicht zurueckstellbar */                                                          \
   IO_DruckeMeldungNeueGleichungGebildet(RE_AnzGleichungenGesamt() + 1, links, rechts); /* vorgezogen wg. Interreduktion */    \
   if (SS_GleichungWirdSubsummiertMitSGPuffernGM(SUE_selectedCPlhs(), SUE_selectedCPrhs()))/* ggf. Meldung: ..wurde subsummiert */ \
      SUE_selectedCPsubsumed();  /*Gleichung wurde subsummiert => wegwerfen. */                                                \
   else{                                                                                                                       \
      SUE_selectedCPRE();                                                                                                      \
      AlsGleichungAufnehmen(SUE_selectedCPlhs(), SUE_selectedCPrhs(), Proofmodus);                                                                                       \
   }                                                                                                                           \
}

#define Unvergleichbar_FAILING_KERN                                                                             \
if (!SUE_selectedCPpostponed()){                                                                                \
   IO_MeldeAbbruchWegenUnrichtbar(SUE_selectedCPlhs(), SUE_selectedCPrhs());                                    \
   VervollstaendigungAbgebrochen = TRUE;                                                                  \
   SUE_selectedCPsubsumed();                                                                                    \
   return;                                                                                                      \
}

#define Unvergleichbar_FAILING__  Unvergleichbar_FAILING_KERN
#define Unvergleichbar_FAILING_S  Unvergleichbar_FAILING_KERN



#define KPRed(NFBWomit)                                                                                                           \
   PCL_Reduktionsbeginn(Gleichung, SUE_selectedCPPclID(), linkeSeite, SUE_selectedCPrhs()); /* letztes Arg. ist GEGENRICHTUNG! */ \
   NF_## NFBWomit(SUE_selectedCPlhs(), KPsSel);                                                                                      \
   PCL_Reduktionsbeginn(Gleichung, SUE_selectedCPPclID(), rechteSeite, SUE_selectedCPlhs());                                      \
   NF_## NFBWomit(SUE_selectedCPrhs(), KPsSel);                                                                                \

#define KPRed_R_ KPRed(NFBildungZuRegeln)

#define KPRed__E KPRed(NFBildungZuGleichungen)

#define KPRed_RE KPRed(NFBildung)

#define KPRed___


#define /*BOOLEAN*/ Permsub_P(/*unsigned int*/ Grad, /*TermT*/ links, /*TermT*/ rechts)                         \
  (TO_ACnGleich(Grad,links,rechts))


#define /*BOOLEAN*/ Permsub__(/*unsigned int*/ Grad, /*TermT*/ links, /*TermT*/ rechts)                         \
  (FALSE)

#define VVBODY(NFBildungWomit, Failing, Subsummieren, GgfPermR, GgfPermE)                                       \
{                                                                                                               \
   KPRed_ ## NFBildungWomit                                                                                     \
                                                                                                                \
   switch(ORD_VergleicheTerme(SUE_selectedCPlhs(), SUE_selectedCPrhs())){                                       \
   case Gleich:                                                                                                 \
      SUE_selectedCPjoined();                                                                                   \
      break;                                                                                                    \
   case Unvergleichbar:                                                                                         \
      if (Permsub_##GgfPermE  (PermGrad_E,SUE_selectedCPlhs(), SUE_selectedCPrhs())) {                          \
                          SUE_selectedCPpermsubsumed();           /*Gleichung wurde subsummiert => wegwerfen. */\
      }                                                                                                         \
      else {                                                                                                    \
        Unvergleichbar_ ## Failing ## _ ## Subsummieren                                                         \
      }                                                                                                         \
      break;                                                                                                    \
   case Kleiner :                                                                                               \
      /* kp als neue Regel rechts->links aufnehmen */                                                           \
      if (Permsub_ ## GgfPermR  (PermGrad_R,SUE_selectedCPlhs(), SUE_selectedCPrhs())) {                        \
       SUE_selectedCPpermsubsumed();  /*Gleichung wurde subsummiert => wegwerfen. */                            \
      }                                                                                                         \
      else {                                                                                                    \
        SUE_selectedCPRE();                                                                                     \
        AlsRegelAufnehmen_ ## Failing(SUE_selectedCPlhs(), SUE_selectedCPrhs(), TRUE, Proofmodus);              \
      }                                                                                                         \
      break;                                                                                                    \
   case Groesser:                                                                                               \
      /* kp als neue Regel links->rechts aufnehmen */                                                           \
      if (Permsub_ ## GgfPermR (PermGrad_R,SUE_selectedCPlhs(), SUE_selectedCPrhs())) {                         \
        SUE_selectedCPpermsubsumed();  /*Gleichung wurde subsummiert => wegwerfen. */                           \
      }                                                                                                         \
      else {                                                                                                    \
        SUE_selectedCPRE();                                                                                     \
        AlsRegelAufnehmen_ ## Failing(SUE_selectedCPlhs(), SUE_selectedCPrhs(), FALSE, Proofmodus);             \
      }                                                                                                         \
      break;                                                                                                    \
   default:                                                                                                     \
      our_error("\n\nundef'd value from term-comparison!\n\n");                                                 \
      break;                                                                                                    \
   }                                                                                                            \
   if (HK_Feierabend())                                                                                         \
     return;                                                                                                    \
   if (FB_InterruptflagGesetzt())                                                                               \
     ST_IRMenue("main loop");                                                                                   \
}


#define MNFVVBODY(/* ohne Parameter */)                                                                 \
/* Die Vervollstaendigung mit MNF wird vorlaeufig nur fuer unfailing unterstuetzt!                      \
   failing erfordert bei Abbruch wg. unrichtbar das Loeschen der gesamten Kreuz-Menge!                  \
*/                                                                                                      \
{                                                                                                       \
   SC_REPairsT *REPairs=NULL;                                                                           \
   MNF_NFSetT  *NFLeft;                                                                                 \
   MNF_NFSetT  *NFRight;                                                                                \
   TermT       joinedTerm;                                                                              \
   TermT       lhs = SUE_selectedCPlhs();                                                               \
   TermT       rhs = SUE_selectedCPrhs();                                                               \
                                                                                                        \
   if (MNF_KPSBearbeite(lhs, rhs, &joinedTerm, &NFLeft, &NFRight))     /* MNF-joined */                 \
      SUE_selectedCPMNFjoined(joinedTerm);                                                              \
   else {                                                                                               \
      SC_REPairsT *runny;                                                                               \
                                                                                                        \
      /*In SC_Classify wird subsummmiert und somit kann es passieren,                                   \
        dass der Rueckgabeparameter REPairs NULL ist.                                                   \
        SC_Classify ruft SUE_selectedCPMNF* Funktionen auf                                              \
        */                                                                                              \
      SC_Classify(NFLeft, NFRight, &REPairs);                                                           \
      MNF_DeleteNFSet (&NFLeft, &NFRight);                                                              \
      runny = REPairs;                                                                                  \
      while (runny){                                                                                    \
         switch (runny->vergleich){                                                                     \
         case Kleiner :                                                                                 \
            /* kp als neue Regel rechts->links aufnehmen */                                             \
            AlsRegelAufnehmen_UNFAILING(runny->left, runny->right, TRUE, Proofmodus);                   \
            break;                                                                                      \
         case Groesser :                                                                                \
            /* kp als neue Regel links->rechts aufnehmen */                                             \
            AlsRegelAufnehmen_UNFAILING(runny->left, runny->right, FALSE, Proofmodus);                  \
            break;                                                                                      \
         case Unvergleichbar:                                                                           \
            /* vorgezogen wg. Interreduktion: */                                                        \
            IO_DruckeMeldungNeueGleichungGebildet(RE_AnzGleichungenGesamt() + 1, links, rechts);        \
            /* Provisorium: SG-Info auch in SelectClassify berechenbar, aber etwas kompliziert */       \
            SS_OnlyInitSGPuffer (runny->left, runny->right);                                            \
            AlsGleichungAufnehmen(runny->left, runny->right, Proofmodus);                               \
            break;                                                                                      \
         default:                                                                                       \
            printf("Fehler: SelectClassify liefert Termpaar mit %s zurueck!\n",                         \
                   (runny->vergleich)==Gleich?"=":"undef'd");                                           \
         }                                                                                              \
         runny = runny->next;                                                                           \
      }                                                                                                 \
      SC_DeletePairs(&REPairs);                                                                         \
   }                                                                                                    \
   if (HK_Feierabend())                                                                                 \
     return;                                                                                            \
   if (FB_InterruptflagGesetzt())                                                                       \
     ST_IRMenue("main loop");                                                                           \
}


static void MNF_Proof_U (void)
{
   while (!Z_ZieleBewiesen() && SUE_Select())
      MNFVVBODY();
   
   if (Z_ZieleBewiesen())
      fprintf(IO_Strom, "\ngoal(s) proved.\n");
   else {
      /* falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, */
      /* muessen sie am Ende der Vervollstaendigung nochmals durchlaufen werden */
      IR_FinalerRMrechtsDurchlauf();
      MeldungBerechnungBeendet();
   }
}


static void MNF_Completion_U (void)
{
   while (SUE_Select())
      MNFVVBODY();

     MeldungBerechnungBeendet();
}

static void MNF_Completion_F (void)
{
   our_error ("using MNF for CP-Selection is only implemented for Unfailing Completion");
}

static void MNF_Proof_F (void)
{
   our_error ("using MNF for CP-Selection is only implemented for Unfailing Completion");
}


#define NV_C_U_BODY(nf, sub, Rperm, Eperm)      \
{                                               \
   while (SUE_Select())                         \
      VVBODY(nf, UNFAILING, sub, Rperm, Eperm)  \
   MeldungBerechnungBeendet();                  \
}


#define NV_C_F_BODY(nf, sub, Rperm, Eperm)      \
{                                               \
   while (SUE_Select())                         \
      VVBODY(nf, FAILING, sub, Rperm, Eperm)    \
   MeldungBerechnungBeendet();                  \
}


static unsigned int SelectZaehler;
static void ChangeStrategy(void);
   
#define NV_P_U_BODY(nf, sub, Rperm, Eperm)                                              \
{                                                                                       \
   while (!Z_ZieleBewiesen()) {                                                         \
     if (++SelectZaehler==OG_Schranke1)                                                 \
       ChangeStrategy();                                                                \
     if (!SUE_Select())                                                                 \
       break;                                                                           \
      VVBODY(nf, UNFAILING, sub, Rperm, Eperm)                                          \
   }                                                                                    \
   if (Z_ZieleBewiesen())                                                               \
     fprintf(IO_Strom, "\ngoal(s) proved.\n");                                          \
   else{                                                                                \
     /* falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, */  \
     /* muessen sie am Ende der Vervollstaendigung nochmals durchlaufen werden */       \
     IR_FinalerRMrechtsDurchlauf();                                                     \
     MeldungBerechnungBeendet();                                                        \
   }                                                                                    \
}



#define NV_P_F_BODY(nf, sub, Rperm, Eperm)                                              \
{                                                                                       \
   while (!Z_ZieleBewiesen() && SUE_Select())                                           \
      VVBODY(nf, FAILING, sub, Rperm, Eperm)                                            \
                                                                                        \
   if (Z_ZieleBewiesen())                                                               \
     fprintf(IO_Strom, "\ngoal(s) proved.\n");                                          \
   else{                                                                                \
     /* falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, */  \
     /* muessen sie am Ende der Vervollstaendigung nochmals durchlaufen werden */       \
     IR_FinalerRMrechtsDurchlauf();                                                     \
     MeldungBerechnungBeendet();                                                        \
   }                                                                                    \
}

/***************************************************************************
  Auffaecherung der Parametrisierung:

+-----------+------------+------+------+-----+-----+-----+
|  PrCo     |   FunF     | NF-R | NF-E | Sub | psR | psE |
+-----------+------------+------+------+-----+-----+-----+
|Proof      |  Unfailing |  _   |  _   | _   |  _  |  _  |
|           |            |      |      |     |     |     |
|Completion |  Failing   |  R   |  E   | S   |  P  |  E  |
+-----------+------------+------+------+-----+-----+-----+

Es gibt also 2^7 = 128 verschiedene Moeglichkeiten.

Diese werden durch generierende Makros erzeugt.
Die einzelnen Funktionen tragen ihr Verhalten in ihrem Namen kodiert:
NormaleVervollstaendigung_PrCo.FunF.NF.Sub.psR.PsE

Die Funktionen werden dann als Funktionsvariablen im Array VerteilerFeld[0..127]
gespeichert. Den Index der gewuenschten Funktion in diesem Feld erhaelt man wie folgt:

index = 0;

if proof     : index += 64
if unfailing : index += 32
if not R     : index += 16
if not E     : index += 8
if not S     : index += 4
if not psR   : index += 2
if not psE   : index += 1
***************************************************************************/


#define genERPERM(PrCo, FunF, NF, Sub)                                  \
static void NormaleVervollstaendigung_##PrCo##FunF##NF##Sub##PP(void)   \
NV_##PrCo##_##FunF##_BODY(NF,Sub,P,P)                                   \
static void NormaleVervollstaendigung_##PrCo##FunF##NF##Sub##_P(void)   \
NV_##PrCo##_##FunF##_BODY(NF,Sub,_,P)                                   \
static void NormaleVervollstaendigung_##PrCo##FunF##NF##Sub##P_(void)   \
NV_##PrCo##_##FunF##_BODY(NF,Sub,P,_)                                   \
static void NormaleVervollstaendigung_##PrCo##FunF##NF##Sub##__(void)   \
NV_##PrCo##_##FunF##_BODY(NF,Sub,_,_)

#define genSUB(PrCo, FunF, NF)                  \
genERPERM(PrCo,FunF,NF,S)                       \
genERPERM(PrCo,FunF,NF,_)

#define genNF(PrCo, FunF)                       \
genSUB(PrCo, FunF, __)                          \
genSUB(PrCo, FunF, R_)                          \
genSUB(PrCo, FunF, _E)                          \
genSUB(PrCo, FunF, RE)

#define genAll                                  \
genNF(P,F)                                      \
genNF(P,U)                                      \
genNF(C,F)                                      \
genNF(C,U)

genAll

/* (IV)  */
/*-------*/
/* Verteiler: Je nach Variante (failing/unfailing) und Modus (Normal/MultipleNF/KPSharing) wird der 
   entsprechende Alg. aufgerufen. Dies geschieht auf dieser hohen Ebene, um Abfragen innerhalb der
   Algorithmen einzusparen.*/

static BOOLEAN MNF_VV = FALSE;

typedef void (* VV_Func_T)(void);
 
static VV_Func_T VerteilerFeld[128]=
{
  NormaleVervollstaendigung_CFRESPP,
  NormaleVervollstaendigung_CFRESP_,
  NormaleVervollstaendigung_CFRES_P,
  NormaleVervollstaendigung_CFRES__,
  NormaleVervollstaendigung_CFRE_PP,
  NormaleVervollstaendigung_CFRE_P_,
  NormaleVervollstaendigung_CFRE__P,
  NormaleVervollstaendigung_CFRE___,
  NormaleVervollstaendigung_CFR_SPP,
  NormaleVervollstaendigung_CFR_SP_,
  NormaleVervollstaendigung_CFR_S_P,
  NormaleVervollstaendigung_CFR_S__,
  NormaleVervollstaendigung_CFR__PP,
  NormaleVervollstaendigung_CFR__P_,
  NormaleVervollstaendigung_CFR___P,
  NormaleVervollstaendigung_CFR____,
  NormaleVervollstaendigung_CF_ESPP,
  NormaleVervollstaendigung_CF_ESP_,
  NormaleVervollstaendigung_CF_ES_P,
  NormaleVervollstaendigung_CF_ES__,
  NormaleVervollstaendigung_CF_E_PP,
  NormaleVervollstaendigung_CF_E_P_,
  NormaleVervollstaendigung_CF_E__P,
  NormaleVervollstaendigung_CF_E___,
  NormaleVervollstaendigung_CF__SPP,
  NormaleVervollstaendigung_CF__SP_,
  NormaleVervollstaendigung_CF__S_P,
  NormaleVervollstaendigung_CF__S__,
  NormaleVervollstaendigung_CF___PP,
  NormaleVervollstaendigung_CF___P_,
  NormaleVervollstaendigung_CF____P,
  NormaleVervollstaendigung_CF_____,
  NormaleVervollstaendigung_CURESPP,
  NormaleVervollstaendigung_CURESP_,
  NormaleVervollstaendigung_CURES_P,
  NormaleVervollstaendigung_CURES__,
  NormaleVervollstaendigung_CURE_PP,
  NormaleVervollstaendigung_CURE_P_,
  NormaleVervollstaendigung_CURE__P,
  NormaleVervollstaendigung_CURE___,
  NormaleVervollstaendigung_CUR_SPP,
  NormaleVervollstaendigung_CUR_SP_,
  NormaleVervollstaendigung_CUR_S_P,
  NormaleVervollstaendigung_CUR_S__,
  NormaleVervollstaendigung_CUR__PP,
  NormaleVervollstaendigung_CUR__P_,
  NormaleVervollstaendigung_CUR___P,
  NormaleVervollstaendigung_CUR____,
  NormaleVervollstaendigung_CU_ESPP,
  NormaleVervollstaendigung_CU_ESP_,
  NormaleVervollstaendigung_CU_ES_P,
  NormaleVervollstaendigung_CU_ES__,
  NormaleVervollstaendigung_CU_E_PP,
  NormaleVervollstaendigung_CU_E_P_,
  NormaleVervollstaendigung_CU_E__P,
  NormaleVervollstaendigung_CU_E___,
  NormaleVervollstaendigung_CU__SPP,
  NormaleVervollstaendigung_CU__SP_,
  NormaleVervollstaendigung_CU__S_P,
  NormaleVervollstaendigung_CU__S__,
  NormaleVervollstaendigung_CU___PP,
  NormaleVervollstaendigung_CU___P_,
  NormaleVervollstaendigung_CU____P,
  NormaleVervollstaendigung_CU_____,
  NormaleVervollstaendigung_PFRESPP,
  NormaleVervollstaendigung_PFRESP_,
  NormaleVervollstaendigung_PFRES_P,
  NormaleVervollstaendigung_PFRES__,
  NormaleVervollstaendigung_PFRE_PP,
  NormaleVervollstaendigung_PFRE_P_,
  NormaleVervollstaendigung_PFRE__P,
  NormaleVervollstaendigung_PFRE___,
  NormaleVervollstaendigung_PFR_SPP,
  NormaleVervollstaendigung_PFR_SP_,
  NormaleVervollstaendigung_PFR_S_P,
  NormaleVervollstaendigung_PFR_S__,
  NormaleVervollstaendigung_PFR__PP,
  NormaleVervollstaendigung_PFR__P_,
  NormaleVervollstaendigung_PFR___P,
  NormaleVervollstaendigung_PFR____,
  NormaleVervollstaendigung_PF_ESPP,
  NormaleVervollstaendigung_PF_ESP_,
  NormaleVervollstaendigung_PF_ES_P,
  NormaleVervollstaendigung_PF_ES__,
  NormaleVervollstaendigung_PF_E_PP,
  NormaleVervollstaendigung_PF_E_P_,
  NormaleVervollstaendigung_PF_E__P,
  NormaleVervollstaendigung_PF_E___,
  NormaleVervollstaendigung_PF__SPP,
  NormaleVervollstaendigung_PF__SP_,
  NormaleVervollstaendigung_PF__S_P,
  NormaleVervollstaendigung_PF__S__,
  NormaleVervollstaendigung_PF___PP,
  NormaleVervollstaendigung_PF___P_,
  NormaleVervollstaendigung_PF____P,
  NormaleVervollstaendigung_PF_____,
  NormaleVervollstaendigung_PURESPP,
  NormaleVervollstaendigung_PURESP_,
  NormaleVervollstaendigung_PURES_P,
  NormaleVervollstaendigung_PURES__,
  NormaleVervollstaendigung_PURE_PP,
  NormaleVervollstaendigung_PURE_P_,
  NormaleVervollstaendigung_PURE__P,
  NormaleVervollstaendigung_PURE___,
  NormaleVervollstaendigung_PUR_SPP,
  NormaleVervollstaendigung_PUR_SP_,
  NormaleVervollstaendigung_PUR_S_P,
  NormaleVervollstaendigung_PUR_S__,
  NormaleVervollstaendigung_PUR__PP,
  NormaleVervollstaendigung_PUR__P_,
  NormaleVervollstaendigung_PUR___P,
  NormaleVervollstaendigung_PUR____,
  NormaleVervollstaendigung_PU_ESPP,
  NormaleVervollstaendigung_PU_ESP_,
  NormaleVervollstaendigung_PU_ES_P,
  NormaleVervollstaendigung_PU_ES__,
  NormaleVervollstaendigung_PU_E_PP,
  NormaleVervollstaendigung_PU_E_P_,
  NormaleVervollstaendigung_PU_E__P,
  NormaleVervollstaendigung_PU_E___,
  NormaleVervollstaendigung_PU__SPP,
  NormaleVervollstaendigung_PU__SP_,
  NormaleVervollstaendigung_PU__S_P,
  NormaleVervollstaendigung_PU__S__,
  NormaleVervollstaendigung_PU___PP,
  NormaleVervollstaendigung_PU___P_,
  NormaleVervollstaendigung_PU____P,
  NormaleVervollstaendigung_PU_____
};


BOOLEAN HK_Vervollstaendigung(void)
/* Vervollstaendigung, Ziele werden erst am Ende behandelt.
   Rueckgabe ist FALSE, wenn bei failing-Variante Vervollstaendigung abgebrochen wurde. */
{
  int WhichOne = 0;
  VervollstaendigungAbgebrochen = FALSE;

  if ((Proofmodus = (PA_ExtModus == proof)))
    IO_DruckeUeberschrift("COMPLETION - PROOF");

  if (MNF_VV)
    if (PA_FailingCompletion)
      if (PA_ExtModus == proof)
        MNF_Proof_F();
      else
        MNF_Completion_F();
    else
      if (PA_ExtModus == proof)
        MNF_Proof_U();
      else
        MNF_Completion_U();

  if (!PA_FailingCompletion)   WhichOne += 32;
  if (!Normalform_R)           WhichOne += 16;
  if (!Normalform_E)           WhichOne += 8;
  if (!Subsumption)            WhichOne += 4;
  if (PermGrad_R == 0)         WhichOne += 2;
  if (PermGrad_E == 0)         WhichOne += 1;
  
  if (Proofmodus){
    WhichOne += 64;
    
    VerteilerFeld[WhichOne]();
    
    if (!Z_ZieleBewiesen())
      /* falls nicht nach jeder Einfuegung die Ziele behandelt wurden, muessen sie am */
      Z_FinalerZielDurchlauf(); /* Ende der Vervollstaendigung nochmals durchlaufen werden. */
  }
  else{
    IO_DruckeUeberschrift("COMPLETION");
    
    VerteilerFeld[WhichOne]();
    
    /* falls die rechten Regelseiten nicht nach jeder Einfuegung behandelt wurden, */
    /* muessen sie am Ende der Vervollstaendigung nochmals durchlaufen werden */
    
    IR_FinalerRMrechtsDurchlauf();                       
  }
  return !VervollstaendigungAbgebrochen;
}



#if COMP_WMTW
#define NegMask6(b6,b5,b4,b3,b2,b1,b0) \
  (!b6<<6 | !b5<<5 | !b4<<4 | !b3<<3 | !b2<<2 | !b1<<1 | !b0<<0)

void HK_InitialeKPsHolen(void)                               /* mittels Copy und Paste aus HK_Vervollstaendigung */
{
  PQ_SELECT2INITIAL(); /* Select-Funktion umstellen: nur initiale KPs und IR-Opfer */
  KPB_KeineBehandlung(); /* KP-Behandlungsfunktion leeren */
  TWInitialeHolen = TRUE;
  if (MNF_VV)
    if (PA_FailingCompletion)
      if (PA_ExtModus == proof)
        MNF_Proof_F();
      else
        MNF_Completion_F();
    else
      if (PA_ExtModus == proof)
        MNF_Proof_U();
      else
        MNF_Completion_U();
  else {
    VerteilerFeld[NegMask6(!Proofmodus,PA_FailingCompletion,Normalform_R,Normalform_E,Subsumption,PermGrad_R,PermGrad_E)]();
    if (Proofmodus){
      if (!Z_ZieleBewiesen())
        Z_FinalerZielDurchlauf();
    }
    else
      IR_FinalerRMrechtsDurchlauf();                       
  }
  PQ_SELECT2NORMAL();           /* Select-Funktion wieder zurueckstellen */
  KPB_WiederBehandeln();        /* KP-Behandlung nach 'alt' setzen */
  TWInitialeHolen = FALSE;
}
#endif

void HK_Test(void)
{
   TermT links, rechts;
   
   return;
   while(SUE_Select()){
      links = SUE_selectedCPlhs();
      rechts = SUE_selectedCPrhs();
      switch(ORD_VergleicheTerme(links, rechts)){
      case Gleich:
         printf("Komisch!!!\nGleich\n\n");
         SUE_selectedCPjoined();
         break;
      case Unvergleichbar: 
         printf("Ok. Unvergleichbar.\n");
         zuletztEingefuegteGleichung = RE_ErzeugteGleichung(links, rechts, RE_AnzGleichungenGesamt(), SUE_selectedCPPclID());
         RE_ZuletztErzeugteHinrichtungEinfuegen(SGM_SubgleichungenZuObjektAufnehmen((SGM_BetroffenesObjektT) zuletztEingefuegteGleichung));
         RE_ZuletztErzeugteHerrichtungEinfuegen();
         break;
      case Kleiner :
         /* kp als neue Regel rechts->links aufnehmen */
         printf("Komisch!!!\nKleiner\n\n");
         SUE_selectedCPjoined();
         break;
      case Groesser:
         printf("Ok. Groesser.\n");
         zuletztEingefuegteRegel = RE_ErzeugteRegel(links, rechts, RE_AnzRegelnGesamt(), SUE_selectedCPPclID());
         RE_ZuletztErzeugteRegelEinfuegen();
         break;
      }
   }
   Z_FinalerZielDurchlauf();
}

/* -------------------------------------------------------------------------------*/
/* --------------- Intialisierung ------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Dieses Makro soll die Auswertung der Parameter uebernehmen. Die auszufuehrenden
   Operationen werden als Bloecke uebergeben.*/

#define Auswertung(R_GRAD, E_GRAD, NF_R, NF_E, SUB, MNF_ACTION) \
{                                                                                       \
  R_GRAD=0;E_GRAD=0;NF_R=NF_E=SUB=FALSE;                                                \
                                                                                        \
  fprintf(IO_Strom, "\n* TREATMENT OF CPS AFTER SELECTION:  ");                         \
                                                                                        \
  if (IS_FindValuedOption(DEF, USR, 1, "permsubE") == 0){                               \
    char *str = IS_ValueOfOption(DEF, USR, "permsubE");                                 \
    if (!(str && ((E_GRAD = atoi(str)) > 0))) E_GRAD = 0;                               \
  }                                                                                     \
  if (IS_FindValuedOption(DEF, USR, 1, "permsubR") == 0){                               \
    char *str = IS_ValueOfOption(DEF, USR, "permsubR");                                 \
    if (!(str && ((R_GRAD = atoi(str)) > 0))) R_GRAD = 0;                               \
  }                                                                                     \
  if (IS_FindOption(DEF, USR, 1, "s") == 0) SUB  = TRUE;                                \
  if (IS_FindValuedOption(DEF, USR, 1, "mnf") == 0){                                    \
    fprintf(IO_Strom, " %s", SUB?"S":"-");                                              \
    fprintf(IO_Strom, " (E-Perm = %u; R-Perm = %u)", E_GRAD, R_GRAD);                   \
    MNF_ACTION;                                                                         \
  }                                                                                     \
  else{                                                                                 \
    if (IS_FindOption(DEF, USR, 1, "r") == 0) NF_R = TRUE;                              \
    if (IS_FindOption(DEF, USR, 1, "e") == 0) NF_E = TRUE;                              \
    fprintf(IO_Strom, " %s%s", NF_R?"R":"-", NF_E?"E":"-");                             \
    fprintf(IO_Strom, " %s", SUB?"S":"-");                                              \
    fprintf(IO_Strom, " (E-Perm = %u; R-Perm = %u)\n", E_GRAD, R_GRAD);                 \
  }                                                                                     \
}

void HK_ParamInfo (ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  if (what == ParamEvaluate){
    int RGRADInfo = 0, EGRADInfo=0;
    BOOLEAN Rinfo, Einfo, Sinfo;
    Auswertung(RGRADInfo,EGRADInfo, Rinfo, Einfo, Sinfo,
               {fprintf(IO_Strom, "\n");MNF_ParamInfo(what, DEF, USR);SC_ParamInfo(what,DEF,USR);});
  }
  else{
    IO_DruckeKurzUeberschrift("Treatment of CPs after Selection", "-");
    fprintf(IO_Strom, "1. [r]\n");
    fprintf(IO_Strom, "2. [e]\n");
    fprintf(IO_Strom, "3. [s]\n");
    fprintf(IO_Strom, "4. [permsubE=<n>]\n");
    fprintf(IO_Strom, "5. [permsubR=<n>]\n");
    fprintf(IO_Strom, "6. [mnf=[r][e]...]\n");
    fprintf(IO_Strom, "\n");
    fprintf(IO_Strom, "Explanation:\n");
    fprintf(IO_Strom, "\n");
    fprintf(IO_Strom, "1-3.  These letters cause Waldmeister to normalize newly generated\n");
    fprintf(IO_Strom, "      critical pairs with _r_ules and/or _e_quations, with or without\n");
    fprintf(IO_Strom, "      additional _s_ubsumption\n");
    fprintf(IO_Strom, "4.,5. These options tell Waldmeister to suppress AC-equations from the\n");
    fprintf(IO_Strom, "      degree <n> on.\n");
    fprintf(IO_Strom, "6.    (see below)\n\n");
    MNF_ParamInfo(what, DEF, USR);
    SC_ParamInfo(what, DEF, USR);
  }
}
   

void HK_InitAposteriori(InfoStringT DEF, InfoStringT USR)
/* Einrichtung der Funktionspointer fuer die Ordnungsaufrufe.*/
{
   Auswertung(PermGrad_R, PermGrad_E,Normalform_R,Normalform_E,Subsumption,
              {
                fprintf(IO_Strom, "\n");
                if ((MNF_InitKPSAposteriori(DEF, USR) && 
                     SC_InitAposteriori (DEF, USR, PermGrad_R, PermGrad_E, Subsumption)))
                  MNF_VV = TRUE;
              });
   if (PermGrad_R+PermGrad_E > 0) PA_unfairness = TRUE;
   SelectZaehler = 0;
}

static unsigned int Laenge(ListenT Liste)
{
  if (Liste)
    return 1+Laenge(((ListenT)Liste)->cdr);
  else
    return NULL;
}

static ListenT ElementNr(unsigned int N, ListenT Liste)
{
  if (N==1)
    return Liste->car;
  else
    return ElementNr(N-1, Liste->cdr);
}

static void ChangeStrategy(void)
{
  unsigned int i = Laenge(OG_Gleichungen);
  SUE_SOFTCLEANUP();
  KPH_INITIALIZE ("heap:e1:domain=30:statistics=on", "", 1);
  RE_LoescheGleichungsmenge();
  RE_LoescheRegelmenge();
  while (i) {
    ListenT Gleichung = ElementNr(i,OG_Gleichungen);
    TermT l = ((ListenT)Gleichung)->car,
          r = ((ListenT)Gleichung)->cdr;
    SUE_CPinitial(l,r);
    SUE_recentCPinsert();
    i--;
  }
  SO_PraezedenzErneuert(OG_Strategie2);
}
