/* --------------------------------------------------------------------------------------------------------*/
/* -                                                                                                       */
/* - SUEM.c                                                                                                */
/* - ******                                                                                                */
/* -                                                                                                       */
/* - Schnittstelle fuer das Set of unselected equations.                                                   */
/* -                                                                                                       */
/* - 14.03.96 Arnim. Angelegt.                                                                             */
/* -                                                                                                       */
/* -                                                                                                       */
/* --------------------------------------------------------------------------------------------------------*/

#include "InfoStrings.h"
#include "TermOperationen.h"
#include "general.h"

#include "PriorityQueue.h"
#include "Classification.h"
#include "CritGoals.h"
#include "TermpairRepresentation.h"

#include "CStatistics.h"

#include "RUndEVerwaltung.h"

#include "Ausgaben.h"

#include "pcl.h"

#include "Unifikation1.h"

#include "SUEM.h"



/* --------------------------------------------------------------------------------------------------------*/
/* ------------ STATISTICS --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static CounterT        AnzKPGesamt;                       /* Insgesamt berechnete kritischer Paare */

static CounterT        AnzKPAusSpezifikation;             /* Kritische Paare aus der Spezifikation */
static CounterT        AnzKPfromTW;                       /* Kritische Paare von anderen Prozessen (TW) */
static CounterT        AnzKPIROpfer;                      /* KPs aus geloeschten Regeln und/oder Gleichungen */

static CounterT        AnzKPdirektZusammengefuehrt;       /* Sofort nach der Bildung zusammenfuehrbare kritische Paare */
static CounterT        AnzKPdirektSubsummiert;            /* Sofort nach der Bildung subsummierte kritische Paare */
static CounterT        AnzKPdirektMNFSubsummiert;         /* Sofort nach der Bildung MNF-subsummierte kritische Paare */
static CounterT        AnzKPdirektPermSubsummiert;        /* Sofort nach der Bildung PERM-subsummierte kritische Paare */
static CounterT        AnzKPdirektMNFPermSubsummiert;     /* Sofort nach der Bildung MNF-PERM-subsummierte kritische Paare */
static CounterT        AnzKPInserted;                     /* In die Menge eingefuegte KPs */

static CounterT        AnzKPBehandeltInVervollst;         /* Zum Vervollstaendigungsalgorithmus hochgereichte KPs */
static CounterT        AnzKPZurueckgestellt;              /* Kritische Paare, die wegen Unrichtbarkeit zurueckgestellt wurden. */
static CounterT        AnzKPspaeterZusammengefuehrt;      /* Kritische Paare, die bei ihrer Behandlung zusammengefuehrt werden konnten. */
static CounterT        AnzKPspaeterSubsummiert;           /* Kritische Paare, die bei ihrer Behandlung subsummiert werden konnten. */
static CounterT        AnzKPspaeterMNFSubsummiert;        /* Kritische Paare, die bei ihrer Behandlung MNF-subsummiert werden konnten. */
static CounterT        AnzKPspaeterPermSubsummiert;       /* Kritische Paare, die bei ihrer Behandlung PERM-subsummiert werden konnten. */
static CounterT        AnzKPspaeterMNFPermSubsummiert;    /* Kritische Paare, die bei ihrer Behandlung MNF-PERM-subsummiert werden konnten. */
static CounterT        AnzKPZuRE;                         /* Zu Regeln/Gleichungen gemachte KPs */
static CounterT        AnzKPWeggeworfen;                  /* Weggeworfen, weil Gewicht zu gross */


#define IncAnzKPGesamt() (AnzKPGesamt++)

#define IncAnzBehandelterKPsVervollst() (AnzKPBehandeltInVervollst++)

#define IncAnzKPAusSpezifikation() (AnzKPAusSpezifikation++)

#define IncAnzKPfromTW() (AnzKPfromTW++)

#define SUE_STATISTIK 0
/* Alternativen: 0 fuer aus, 1 fuer an */

#if SUE_STATISTIK
#define NrNaechstesKP() (AnzKPGesamt + 1)

#define IncAnzKPIROpfer() (AnzKPIROpfer++)

#define IncAnzKPdirektZusammengefuehrt() (AnzKPdirektZusammengefuehrt++) 
#define IncAnzKPdirektSubsummiert() (AnzKPdirektSubsummiert++) 
#define IncAnzKPdirektMNFSubsummiert() (AnzKPdirektMNFSubsummiert++) 
#define IncAnzKPdirektPermSubsummiert() (AnzKPdirektPermSubsummiert++) 
#define IncAnzKPdirektMNFPermSubsummiert() (AnzKPdirektMNFPermSubsummiert++) 
#define IncAnzKPInserted() (AnzKPInserted++)

#define IncAnzKPZurueckgestellt() (AnzKPZurueckgestellt++)
#define IncAnzKPspaeterZusammengefuehrt() (AnzKPspaeterZusammengefuehrt++)
#define IncAnzKPspaeterSubsummiert() (AnzKPspaeterSubsummiert++)
#define IncAnzKPspaeterMNFSubsummiert() (AnzKPspaeterMNFSubsummiert++)
#define IncAnzKPspaeterPermSubsummiert() (AnzKPspaeterPermSubsummiert++)
#define IncAnzKPspaeterMNFPermSubsummiert() (AnzKPspaeterMNFPermSubsummiert++)
#define IncAnzKPZuRE() (AnzKPZuRE++)
#define IncAnzKPWeggeworfen() (AnzKPWeggeworfen++)

#define MaximiereAnzKPBlaeh(neuwert){if (neuwert>AnzKPBlaeh)AnzKPBlaeh = neuwert;}

#else

#define NrNaechstesKP()                                  /* Nix Tun */

#define IncAnzKPIROpfer()                                /* Nix Tun */

#define IncAnzKPdirektZusammengefuehrt()                 /* Nix Tun */
#define IncAnzKPdirektSubsummiert()                      /* Nix Tun */
#define IncAnzKPdirektMNFSubsummiert()                   /* Nix Tun */
#define IncAnzKPdirektPermSubsummiert()                  /* Nix Tun */
#define IncAnzKPdirektMNFPermSubsummiert()               /* Nix Tun */
#define IncAnzKPInserted()                               /* Nix Tun */

#define IncAnzKPZurueckgestellt()                        /* Nix Tun */
#define IncAnzKPspaeterZusammengefuehrt()                /* Nix Tun */
#define IncAnzKPspaeterSubsummiert()                     /* Nix Tun */
#define IncAnzKPspaeterMNFSubsummiert()                  /* Nix Tun */
#define IncAnzKPspaeterMNFPermSubsummiert()              /* gar nichts */
#define IncAnzKPspaeterPermSubsummiert()                 /* Nix Tun */
#define IncAnzKPZuRE()                                   /* Nix Tun */
#define IncAnzKPWeggeworfen()                            /* dito    */

#endif

static long int lowerBound, upperBound;


static void UEOutput_TW(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   if ((wt.w1 <= upperBound) && (wt.w1>= lowerBound)){
      fprintf(IO_Strom, "CP(%6ld, %6ld), Weight: ", TP_ElternNr(actualP), TP_ElternNr(otherP));
      C_PrintWeight(&wt, actualP, otherP);
      fprintf(IO_Strom, " : ");
      TPR_PrintTermpair(tp, actualP, otherP, " ?= ");
      fprintf(IO_Strom, "\n");
   }
}

static void UEOutput_W(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   if ((wt.w1 <= upperBound) && (wt.w1>= lowerBound)){
      fprintf(IO_Strom, "CP(%6ld, %6ld), Weight: ", TP_ElternNr(actualP), TP_ElternNr(otherP));
      C_PrintWeight(&wt, actualP, otherP);
      fprintf(IO_Strom, "\n");
   }
}

static void UEOutput_T(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   if ((wt.w1 <= upperBound) && (wt.w1>= lowerBound)){
      fprintf(IO_Strom, "CP(%6ld, %6ld) :", TP_ElternNr(actualP), TP_ElternNr(otherP));
      TPR_PrintTermpair(tp, actualP, otherP, " ?= ");
      fprintf(IO_Strom, "\n");
   }
}


static void UEOutput__(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   if ((wt.w1 <= upperBound) && (wt.w1>= lowerBound)){
      fprintf(IO_Strom, "CP(%6ld, %6ld) \n", TP_ElternNr(actualP), TP_ElternNr(otherP));
   }
}

void SUE_Statistics(InfoStringT StatisticsIS)
/* InfoString:
   output -- Mengen-Ausgabe
   weight -- Gewichtsverteilung
   lowerBound=..., upperBound=...,
   noweights,
   noterms,
   time   -- Zeitliche Verteilung der Gewichte
*/
{
   int tiefe1, tiefe2;

   /* output:limit=value:noterms:noweights. Dabei gilt limit fuer das 1. int-Gewicht
      Dabei ist der Default: kein limit, Terme und Gewichte ausgeben */

   
   if (!(IS_IsIn("all", StatisticsIS) || (IS_IsIn("SUEM", StatisticsIS)))){
      if (IS_IsIn("PQ", StatisticsIS))
         PQ_Statistics(StatisticsIS);
      if (IS_IsIn("C", StatisticsIS))
         C_Statistics(StatisticsIS);
      if (IS_IsIn("TPR", StatisticsIS))
         TPR_Statistics(StatisticsIS);
      return;
   }
   if (IS_IsIn("output", StatisticsIS)){

      if (IS_ValueOf("lowerBound", StatisticsIS))
         lowerBound = atoi(IS_ValueOf("lowerBound", StatisticsIS));
      else
         lowerBound = LONG_MIN;

      if (IS_ValueOf("upperBound", StatisticsIS))
         upperBound = atoi(IS_ValueOf("upperBound", StatisticsIS));
      else
         upperBound = LONG_MAX;
   
      
      printf("SUE_Statistics: output (%ld - %ld): %s\n", lowerBound, upperBound, StatisticsIS);

      if (IS_IsIn("noterms", StatisticsIS))
         if (IS_IsIn("noweights", StatisticsIS))
            PQ_WatchingWalkthrough(UEOutput__);
         else
            PQ_WatchingWalkthrough(UEOutput_W);
      else
         if (IS_IsIn("noweights", StatisticsIS))
            PQ_WatchingWalkthrough(UEOutput_T);
         else
            PQ_WatchingWalkthrough(UEOutput_TW);
      return;
   }
   else if (IS_IsIn("weight", StatisticsIS)){
      CS_WeightStatisticsStart(StatisticsIS);
      PQ_WatchingWalkthrough(CS_WeightStatProc);
      /* weight:upperBond=..:lowerBound=.. Es werden nur lebende KPs angemeldet */
      CS_WeightStatisticsEnd();
      return;
   }
   else if (IS_IsIn("time", StatisticsIS)){
      CS_TimeDistribution(StatisticsIS);
      return;
   }
   else if (IS_IsIn("standard",StatisticsIS)){
      IO_DruckeKurzUeberschrift("Unselected Equations","+");
      
      IO_DruckeStatistikZeile("Total number of ues",AnzKPGesamt);
      tiefe1 = IO_Einruecktiefe(0, Maximum(8, AnzKPAusSpezifikation, AnzKPIROpfer,
                                           AnzKPdirektZusammengefuehrt, AnzKPdirektSubsummiert, AnzKPInserted,
					   AnzKPWeggeworfen, AnzKPdirektMNFSubsummiert, 
                                           AnzKPdirektPermSubsummiert));
      IO_DruckeStatistikZeileEingerueckt("of them from Specification",AnzKPAusSpezifikation, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them from Killed RE", AnzKPIROpfer, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them from Generation", AnzKPGesamt-AnzKPAusSpezifikation-AnzKPIROpfer, tiefe1);
      IO_DruckeStatistikLeerZeile();
      
      IO_DruckeStatistikZeileEingerueckt("of them joined immediately",AnzKPdirektZusammengefuehrt, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them subsumed immediately", AnzKPdirektSubsummiert, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them MNF-subsumed immediately", AnzKPdirektMNFSubsummiert, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them perm-subsumed immediately", AnzKPdirektPermSubsummiert, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them thrown away", AnzKPWeggeworfen, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them inserted into PQ", AnzKPInserted, tiefe1);
      IO_DruckeStatistikLeerZeile();
      IO_DruckeStatistikZeileEingerueckt("ues selected",AnzKPBehandeltInVervollst, tiefe1);
      
      tiefe2 = IO_Einruecktiefe(tiefe1, Maximum(7, AnzKPspaeterZusammengefuehrt, AnzKPZuRE,
                                                AnzKPZurueckgestellt, AnzKPspaeterSubsummiert, 
                                                AnzKPspaeterMNFSubsummiert, AnzKPspaeterPermSubsummiert,
                                                AnzKPspaeterMNFPermSubsummiert));
      
      IO_DruckeStatistikZeileEingerueckt("of them not orientable, postponed", AnzKPZurueckgestellt, tiefe1);
      IO_DruckeStatistikZeileEingerueckt("of them joined", AnzKPspaeterZusammengefuehrt, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them subsumed (forward subsumption)", AnzKPspaeterSubsummiert, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them MNF-subsumed (forward subsumption)", AnzKPspaeterMNFSubsummiert, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them perm-subsumed (forward subsumption)", AnzKPspaeterPermSubsummiert, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them MNF-perm-subsumed (forward subsumption)", AnzKPspaeterMNFPermSubsummiert, tiefe2);
      IO_DruckeStatistikZeileEingerueckt("of them made RE", AnzKPZuRE, tiefe2);
      
      PQ_Statistics (StatisticsIS); /* complete, damit bei 'SUEM' die Submodule auch ausgeben */
      C_Statistics  (StatisticsIS);
      TPR_Statistics(StatisticsIS);
      CG_Statistics (StatisticsIS);
   }
   else if (IS_IsIn("short", StatisticsIS)){
      IO_DruckeKurzUeberschrift("Unselected Equations","+");
      IO_DruckeStatistikZeile("Generated", AnzKPGesamt);
      IO_DruckeStatistikZeile("Selected", AnzKPBehandeltInVervollst);
   }
   else if (IS_IsIn("config", StatisticsIS)){
      PQ_Statistics(StatisticsIS);
      C_Statistics(StatisticsIS);
      TPR_Statistics(StatisticsIS);
   }
   else if (IS_IsIn("intern", StatisticsIS)){
      PQ_Statistics(StatisticsIS);
      C_Statistics(StatisticsIS);
      TPR_Statistics(StatisticsIS);
   }
   else if (IS_IsIn("latex-head", StatisticsIS)){
      fprintf(IO_Strom, "SUE: CPs & spec & IR-vict & GEN: join & subs & cut & inserted & SEL: & postp & join & subs & MNF subs & perm-subs & MNF perm-subs & RE & ");
      PQ_Statistics(StatisticsIS);
      C_Statistics(StatisticsIS);
      TPR_Statistics(StatisticsIS);
      CG_Statistics (StatisticsIS);
   }
   else if (IS_IsIn("slatex-head", StatisticsIS)){
      fprintf(IO_Strom, "CPs gen & sel & ");
   }
}

void SUE_AktuelleGroesse(unsigned long int *GesamtKPZahl, unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl)
{
  *GesamtKPZahl = AnzKPGesamt;
  PQ_AktuelleMengenGroesse(AktKPZahl,AktGueltigeKPZahl);
}

#define Auswertung(CUT, PARAINT)                                                                     \
{                                                                                                    \
  CUT=PARAINT=0;                                                                                     \
  if (IS_ValueOfOption(DEFclas, USRclas, "cut1"))                                                    \
    CUT = atol (IS_ValueOfOption (DEFclas, USRclas, "cut1"));                                        \
  if (IS_ValueOfOption(DEFsue, USRsue, "paraint"))                                                   \
    PARAINT = atoi(IS_ValueOfOption(DEFsue, USRsue, "paraint")); /*$gna*/                            \
                                                                                                     \
  if (PARAINT == 0)                                                                                  \
    PARAINT = 2;                                                                                     \
  fprintf(IO_Strom, "\n* SET OF UNSELECTED EQUATIONS:  Paramodulation interval = %ld\n", PARAINT);   \
  if (CUT != 0)                                                                                      \
    fprintf (IO_Strom,"                                throw away CPs with weight w1 > %ld\n", CUT); \
}

void SUE_ParamInfo(ParamInfoArtT what, int whichModule,
                   InfoStringT DEFsue,    InfoStringT USRsue,
                   InfoStringT DEFclas,   InfoStringT USRclas,
                   InfoStringT DEFreclas, InfoStringT USRreclas,
                   InfoStringT DEFcgclas, InfoStringT USRcgclas,
                   InfoStringT DEFpq,     InfoStringT USRpq,
                   InfoStringT DEFtpr,    InfoStringT USRtpr)
{
   if ((whichModule==0) || (whichModule==99)){                /* SUE */
     if (what == ParamEvaluate){
       unsigned long int infoCUT, infoPARA;
       Auswertung(infoCUT, infoPARA);
     }
     else{
       IO_DruckeKurzUeberschrift("Set of unselected equations", "-");
       fprintf(IO_Strom, "1. [cut1=<n>]\n");
       fprintf(IO_Strom, "2. [paraint=<n>]\n");
       fprintf(IO_Strom, "\n");
       fprintf(IO_Strom, "1. Throw away CPs with primary weight > n\n");
       fprintf(IO_Strom, "2. Intervall for paramodulation with existentially quantified goals.\n");
     }
   }

   if ((whichModule == 99) || (whichModule==1) || (whichModule==2) ||(whichModule==3))
     C_ParamInfo(what, whichModule, DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas);

   if ((whichModule==4) || (whichModule==99)) PQ_ParamInfo(what, DEFpq, USRpq);

   if ((whichModule==5) || (whichModule==99)) TPR_ParamInfo(what, DEFtpr, USRtpr);
}
      

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ VARIABLES ---------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

BOOLEAN SUE_FatherLHS, SUE_MotherLHS, SUE_InFather;
UTermT SUE_Superposition;
void *SUE_addref;               /* additional weight-reference */
long int ParamodIntervall = 0, ParamodCounter = 0;
long int cut1 = 0;

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ INITIALIZE --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
void SUE_InitApriori(void)
{
  C_InitApriori();
}

BOOLEAN SUE_INITIALIZE(InfoStringT DEFsue,    InfoStringT USRsue,
                       InfoStringT DEFclas,   InfoStringT USRclas,
                       InfoStringT DEFreclas, InfoStringT USRreclas,
                       InfoStringT DEFcgclas, InfoStringT USRcgclas,
                       InfoStringT DEFpq,     InfoStringT USRpq,
                       InfoStringT DEFtpr,    InfoStringT USRtpr)
{
   int weightType;


   if (!TPR_INITIALIZE(DEFtpr, USRtpr))
      return FALSE;

   if (!C_INITIALIZE(DEFclas, USRclas, DEFreclas, USRreclas, DEFcgclas, USRcgclas, &weightType))
      return FALSE;

   if (!PQ_INITIALIZE(DEFpq, USRpq, weightType))
      return FALSE;

   if (!CG_InitAposteriori())
      return FALSE;

   Auswertung(cut1, ParamodIntervall);

   if (cut1 > 0) PA_unfairness = TRUE;

   SUE_Demigod.linkeSeite = NULL;
   SUE_Demigod.rechteSeite = NULL;
   SUE_Demigod.Nummer = 0;
   SUE_Demigod.SUEInfo = NULL;
   PQ_NEWPARENT(&SUE_Demigod);

   /* Statistik initialisieren. */

   AnzKPGesamt = 0;                  
   AnzKPAusSpezifikation = 0;        
   AnzKPIROpfer = 0;                 
   AnzKPdirektZusammengefuehrt = 0;  
   AnzKPdirektSubsummiert = 0;       
   AnzKPdirektMNFSubsummiert = 0;       
   AnzKPdirektPermSubsummiert = 0;       
   AnzKPdirektMNFPermSubsummiert = 0;       
   AnzKPInserted = 0;                
   AnzKPBehandeltInVervollst = 0;    
   AnzKPZurueckgestellt = 0;
   AnzKPspaeterZusammengefuehrt = 0; 
   AnzKPspaeterSubsummiert = 0;      
   AnzKPspaeterMNFSubsummiert = 0;      
   AnzKPspaeterPermSubsummiert = 0;      
   AnzKPspaeterMNFPermSubsummiert = 0;      
   AnzKPZuRE = 0;                    
   return TRUE;
}
/* --------------------------------------------------------------------------------------------------------*/
/* ------------ PARENTS -----------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


void SUE_NewParent(RegelOderGleichungsT re, TermT lhsReplica, TermT rhsReplica)
{
   SUE_ActualParent = re;
   PQ_NEWPARENT(re);   /* In PQ: SUE_ActualParent = re;*/
   SUE_ReplicationOfLHS = lhsReplica;
   SUE_ReplicationOfRHS = rhsReplica;
}

void SUE_KillParent(RegelOderGleichungsT re)
/* Vorausgesetzt wird, dass alle Regeln/Gleichungen einen gueltigen Info-Block erhalten haben. */
{
  CG_KillParent(re);
  PQ_KILLPARENT(re);
}

void SUE_RedirectParent(void)
{
   TermT switty = SUE_ReplicationOfLHS;

   SUE_ReplicationOfLHS = SUE_ReplicationOfRHS;
   SUE_ReplicationOfRHS = switty;
   PQ_RedirectParent();         /* SUE_ActualParent ist jetzt noch auf der ausgezeichneten Richtung.*/
   SUE_ActualParent = TP_Gegenrichtung(SUE_ActualParent);
}

void SUE_FinishParent(void)
{
   CG_AddCollected();
   PQ_FinishParent();
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ GENERATION --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


void SUE_CPfromTW(TermT lhs, TermT rhs)
{
   IncAnzKPGesamt(); IncAnzKPfromTW();
   IO_DruckeMeldungTWGleichungZuKPGemacht(AnzKPGesamt, lhs, rhs);
   /*   PCL_GleichungAusDerSpezifikation(lhs, rhs, &SUE_recCPpclID);*/
   PCL_StartChain();
   SUE_recCPrhs = rhs;
   SUE_recCPlhs = lhs;
   SUE_ActualParent = SUE_recCPOtherParent = &SUE_Demigod;
   SUE_recCPactualPosition = NULL;
   SUE_recCPGenerationHistory = Initial;
}

void SUE_CPinitial(TermT lhs, TermT rhs)
{
   IncAnzKPGesamt(); IncAnzKPAusSpezifikation();
   IO_DruckeMeldungSpezGleichungZuKPGemacht(AnzKPAusSpezifikation, AnzKPGesamt, lhs, rhs);
   PCL_GleichungAusDerSpezifikation(lhs, rhs, &SUE_recCPpclID);
   PCL_StartChain();
   SUE_recCPrhs = rhs;
   SUE_recCPlhs = lhs;
   SUE_ActualParent = SUE_recCPOtherParent = &SUE_Demigod;
   SUE_recCPactualPosition = NULL;
   SUE_recCPGenerationHistory = Initial;
}

void SUE_CPkilledRE(TermT lhs, TermT rhs, PCL_ObjektNummerT *pclIDPtr)
{
   IncAnzKPGesamt(); IncAnzKPIROpfer();
   IO_DruckeMeldungZuKPgemacht(AnzKPGesamt, lhs, rhs);
   PCL_StartChain();
   SUE_recCPrhs = rhs;
   SUE_recCPlhs = lhs;
   SUE_ActualParent = SUE_recCPOtherParent = &SUE_Demigod;
   SUE_recCPactualPosition = NULL;
   SUE_recCPGenerationHistory = KilledRE;
   PCL_IDZuweisung(&SUE_recCPpclID, *pclIDPtr);
}

void SUE_CPgeneratedWithFather(RegelOderGleichungsT father, TermT lhs, TermT rhs, UTermT actualPosition)
{
   IncAnzKPGesamt(); 
   IO_DruckeMeldungKPgebildet(AnzKPGesamt, TP_ElternNr(SUE_ActualParent), TP_ElternNr(father), lhs, rhs);
   TPR_NewFlatsGenerated();
   PCL_Unifikation(&SUE_recCPpclID, lhs, rhs,
                   father, TP_LinkeSeite(father), actualPosition, SUE_ActualParent);
   SUE_recCPlhs = lhs;
   SUE_recCPrhs = rhs;
   SUE_recCPOtherParent = father;
   SUE_recCPactualPosition = actualPosition;
   SUE_recCPGenerationHistory = CPWithFather;
}


void SUE_CPgeneratedWithMother(RegelOderGleichungsT mother, TermT lhs, TermT rhs, UTermT actualPosition)
{
   IncAnzKPGesamt(); 
   IO_DruckeMeldungKPgebildet(AnzKPGesamt, TP_ElternNr(SUE_ActualParent), TP_ElternNr(mother), lhs, rhs);
   TPR_NewFlatsGenerated();
   PCL_Unifikation(&SUE_recCPpclID, lhs, rhs,
                   SUE_ActualParent, SUE_ReplicationOfLHS, actualPosition, mother);
   SUE_recCPlhs = lhs;
   SUE_recCPrhs = rhs;
   SUE_recCPOtherParent = mother;
   SUE_recCPactualPosition = actualPosition;
   SUE_recCPGenerationHistory = CPWithMother;
}


void SUE_recentCPjoined(void)
{
   IO_DruckeMeldungKPZusammengefuehrt(SUE_recCPlhs);
   PCL_DeleteChain();
   IncAnzKPdirektZusammengefuehrt();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}

void SUE_recentCPMNFjoined(TermT joinedTerm)
{
   IO_DruckeMeldungKPMNFZusammengefuehrt(joinedTerm);
   TO_TermLoeschen(joinedTerm);
   PCL_DeleteChain();
   IncAnzKPdirektZusammengefuehrt();/* neuer Zaehler? */
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs); 
}

void SUE_recentCPsubsumed(void)
{
   IO_DruckeMeldungKPSubsummiert();
   PCL_DeleteChain();
   IncAnzKPdirektSubsummiert();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}


void SUE_recentCPMNFsubsumed(void)
{
   IO_DruckeMeldungKPMNFSubsummiert();
   PCL_DeleteChain();
   IncAnzKPdirektMNFSubsummiert();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}

void SUE_recentCPpermsubsumed(void)
{
   IO_DruckeMeldungKPPermSubsummiert();
   PCL_DeleteChain();
   IncAnzKPdirektPermSubsummiert();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}


void SUE_recentCPMNFpermsubsumed(void)
{
   IO_DruckeMeldungKPMNFPermSubsummiert();
   PCL_DeleteChain();
   IncAnzKPdirektMNFPermSubsummiert();
   TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
}


void SUE_recentCPinsert(void)
/*$gna*/
{
   TermpairT tp;
   WeightEntryT wt;
   BOOLEAN IsCritGoal = FALSE;


   if ((IsCritGoal = (SO_Paramodulation() &&
                      (SO_SymbGleich(TO_TopSymbol(SUE_recCPlhs), SO_EQSymbol()) || 
                       SO_SymbGleich(TO_TopSymbol(SUE_recCPrhs), SO_EQSymbol()))))){
     
     TermT lhs, rhs;
     
     if (SO_SymbGleich(TO_TopSymbol(SUE_recCPlhs), SO_EQSymbol())) /* fuer CG-Heuristik */
       lhs = TO_Schwanz(SUE_recCPlhs);
     else
       lhs = TO_Schwanz(SUE_recCPrhs);

     rhs = TO_Schwanz(TO_TermEnde(lhs));

     wt.w2=0;                  /* Fuer lex. Vergleich auf Liste */

     C_CGCLASSIFY(&wt, lhs, rhs, SUE_ActualParent, SUE_recCPOtherParent, NULL);
   }
   else

     C_CLASSIFY(&wt, SUE_recCPlhs, SUE_recCPrhs, SUE_ActualParent, SUE_recCPOtherParent, 
                SUE_recCPactualPosition, SUE_recCPGenerationHistory);

   if ((cut1 != 0) && (!IsCritGoal) && (wt.w1 > cut1)) {
      /* throw CP away */
      IO_DruckeMeldungKPWeggeworfen ();
      PCL_DeleteChain();
      IncAnzKPWeggeworfen();
      TO_TermeLoeschen(SUE_recCPlhs, SUE_recCPrhs);
      return;
   }

   IO_DruckeMeldungAlsKPaufgenommen();

   tp = TPR_FLATTERMStoTERMPAIR(SUE_recCPlhs, SUE_recCPrhs, SUE_ActualParent, SUE_recCPOtherParent, SUE_recCPactualPosition, SUE_recCPpclID);
   
   if (IsCritGoal)
     CG_Collect(tp, wt, SUE_ActualParent, SUE_recCPOtherParent);
   else if (IsDemigod(SUE_recCPOtherParent))
     PQ_INSERTImmortal(tp, wt, SUE_ActualParent, SUE_recCPOtherParent);
   else
     PQ_INSERT(tp, wt, SUE_ActualParent, SUE_recCPOtherParent);
   
   IncAnzKPInserted();
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------------ SELECTION ---------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

BOOLEAN SUE_SelectInitial(void)
/* Liefert alle unsterblichen KPs zurueck, die zusaetzlich ein (absolut) minimales Gewicht haben. */
{
   if (PQ_SELECTINITIAL(&SUE_selCPTermpair, &SUE_selCPactualParent, &SUE_selCPotherParent, &SUE_selCPaddref)){

     TPR_TERMPAIRtoPARALLEL_FINAL(SUE_selCPTermpair, &SUE_selCPlhs, &SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent, &SUE_selCPpclID);
      
     IO_DruckeMeldungKPwirdBehandelt(TP_ElternNr(SUE_selCPactualParent), TP_ElternNr(SUE_selCPotherParent), SUE_selCPlhs, SUE_selCPrhs);
     
     IncAnzBehandelterKPsVervollst();
     
     return TRUE;
   }
   else
     return FALSE;
}

static void (* IRProc) (TermT, TermT);

static void TouchOnlyImmortals(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
  if (IsDemigod(otherP)){
    TermT lhs, rhs;
    unsigned long int dummyPCL_ID;
    TPR_TERMPAIRtoPARALLEL_INTERMED(tp, &lhs, &rhs, actualP, otherP, &dummyPCL_ID);
    IRProc(lhs, rhs);
    TPR_PARALLELtoTERMPAIR(tp, lhs, rhs, actualP, otherP, FALSE, dummyPCL_ID); /* nicht veraendert */
  }
}

void SUE_TouchImmortals(void (* Proc) (TermT, TermT))
/* Liefert alle unsterblichen KPs zurueck, unabhaengig vom Gewicht. */
{
  IRProc = Proc;
  PQ_WatchingWalkthrough(TouchOnlyImmortals);
}




BOOLEAN SUE_Select(void)
/*$gna*/
{
  if ((SO_Paramodulation()
        && (((ParamodCounter%ParamodIntervall)
               &&  
               PQ_SELECT(&SUE_selCPTermpair, &SUE_selCPactualParent, &SUE_selCPotherParent, &SUE_selCPaddref))
            ||
           (CG_Select(&SUE_selCPTermpair, &SUE_selCPactualParent, &SUE_selCPotherParent, &SUE_selCPaddref))))
      ||(PQ_SELECT(&SUE_selCPTermpair, &SUE_selCPactualParent, &SUE_selCPotherParent, &SUE_selCPaddref))){

     TPR_TERMPAIRtoPARALLEL_FINAL(SUE_selCPTermpair, &SUE_selCPlhs, &SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent, &SUE_selCPpclID);
    
     IO_DruckeMeldungKPwirdBehandelt(TP_ElternNr(SUE_selCPactualParent), TP_ElternNr(SUE_selCPotherParent), SUE_selCPlhs, SUE_selCPrhs);
    
     IncAnzBehandelterKPsVervollst();
     
     return TRUE;
   }
   else
      return FALSE;
      
}


void SUE_selectedCPjoined(void)
{
   IO_DruckeMeldungKPZusammengefuehrt(SUE_selCPlhs);
   C_KillAdditionalReference(SUE_selCPaddref);
   TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
   IncAnzKPspaeterZusammengefuehrt();
}

void SUE_selectedCPMNFjoined(TermT joinedTerm)
{
   IO_DruckeMeldungKPMNFZusammengefuehrt(joinedTerm);
   TO_TermLoeschen(joinedTerm);
   C_KillAdditionalReference(SUE_selCPaddref);
   TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
   IncAnzKPspaeterZusammengefuehrt();
}


void SUE_selectedCPsubsumed(void)
{
/*    if (SO_Paramodulation()) ParamodCounter++;*/
    IO_DruckeMeldungKPSubsummiert();
    C_KillAdditionalReference(SUE_selCPaddref);
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPspaeterSubsummiert();
}


void SUE_selectedCPMNFsubsumed(void)
{
/*    if (SO_Paramodulation()) ParamodCounter++;*/
    IO_DruckeMeldungKPMNFSubsummiert();
    C_KillAdditionalReference(SUE_selCPaddref);
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPspaeterMNFSubsummiert();
}

void SUE_selectedCPpermsubsumed(void)
{
/*    if (SO_Paramodulation()) ParamodCounter++;*/
    IO_DruckeMeldungKPPermSubsummiert();
    C_KillAdditionalReference(SUE_selCPaddref);
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPspaeterPermSubsummiert();
}


void SUE_selectedCPMNFpermsubsumed(void)
{
/*    if (SO_Paramodulation()) ParamodCounter++;*/
    IO_DruckeMeldungKPMNFPermSubsummiert();
    C_KillAdditionalReference(SUE_selCPaddref);
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPspaeterMNFPermSubsummiert();
}


void SUE_selectedCPRE(void)
{
    if (SO_Paramodulation()) ParamodCounter++;
    C_KillAdditionalReference(SUE_addref);
    TPR_PARALLELtoFLATTERMS(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPZuRE();
}

void SUE_selectedCPMNFRE(void)
{
    if (SO_Paramodulation()) ParamodCounter++;
    C_KillAdditionalReference(SUE_addref);
    TPR_DESTROYPARALLEL(SUE_selCPTermpair, SUE_selCPlhs, SUE_selCPrhs, SUE_selCPactualParent, SUE_selCPotherParent);
    IncAnzKPZuRE();
}


BOOLEAN SUE_selectedCPpostponed(void)
{
    return FALSE;
}




/* --------------------------------------------------------------------------------------------------------*/
/* ------------ WALKTHROUGH -------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

unsigned long int SUE_AnzKPsToGo(void)
{
   return PQ_AnzKPsToGo();
}

void SUE_WALKTHROUGH(PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
{
   PQ_WALKTHROUGH(AP, NeedsTerms);
}

void SUE_TermeVerfolgen(void (*OperationAufTermen) (TermT, TermT))
/* Alle Terme. Achtung: Geht nicht mit CGs!!!  */
{
   PQ_TermeVerfolgen(OperationAufTermen);
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ----------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

void SUE_CLEANUP(void)
{
   PQ_Cleanup();
   PQ_KILLPARENT(&SUE_Demigod);
   TPR_CleanUp();
   CG_CleanUp();
   C_CleanUp();
}


void SUE_SOFTCLEANUP(void)
{
  PQ_SoftCleanup();             /* Strahlen bei Heap und HeapOfHeaps loeschen */
  /* KillParent leert nur den Info-Eintrag in SUE_Demigod; dieser wird aber ohnehin geklaert. */
  /* TPR_Cleanup tut eh' nix */
  CG_SoftCleanUp();
  C_CleanUp();                  /* bleibt noetig, weil CStatistik ueber our_alloc laeuft! */
}
