/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFAdhoc.c                                                                                                  */
/* - *********                                                                                                  */
/* -                                                                                                            */
/* - Enthaelt die Adhoc-Version der Normalform-Bildung, dh. Termdurchlauf ohne jede Pfadverwaltung.             */
/* - Es wird auch nur zwischen zwei Varianten unterschieden, naemlich zwischen 'nochmal' und 'direkt'.          */
/* - Bei 'nochmal' wird eine reduzierte Stelle zunaechst fertigreduziert, bevor weitergelaufen wird;            */
/* - Bei 'direkt' wird auch nach einer Reduktion weitergelaufen.                                                */
/* -                                                                                                            */
/* - Strategie: Der zu reduzierende Term wird von vorne nach hinten durchlaufen. Dabei wird jede Stelle auf     */
/* -            Reduzierbarkeit geprueft und gegebenenfalls reduziert. War waehrend eines solchen Durchlaufes   */
/* -            eine Reduktion moeglich, so wird der Durchlauf wiederholt, andernfalls ist der Term jetzt       */
/* -            auf Normalform.                                                                                 */
/* -                                                                                                            */
/* - 02.03.95 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/



#include "general.h"
#include "RUndEVerwaltung.h"
#include "ZeitMessung.h"
#include "NFBasis.h"
#include "NFAdhoc.h"

void NFA_Init(void)
{
   NFB_Init();
}

   
BOOLEAN NFA_NormalformZuRegelnDirekt(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach erfolgter Reduktion wird direkt weitergegangen.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   BOOLEAN ImLetztenDurchlaufReduziert;

   ZM_StartNFBildung();
   NFB_Reduziert = FALSE;
   do{
      AdHocPfadNeu(Term, objekt);
      ImLetztenDurchlaufReduziert = FALSE;
      do{
         ImLetztenDurchlaufReduziert |= NFB_RegelAngewendetAH();
      }while(AdHocPfadWeitersetzen());
   }while(ImLetztenDurchlaufReduziert);
   ZM_StoppNFBildung();
   return NFB_Reduziert;
}

BOOLEAN NFA_NormalformZuRegelnNochmal(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach erfolgter Reduktion wird zunaechst die betroffene Stelle fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   BOOLEAN ImLetztenDurchlaufReduziert;

   ZM_StartNFBildung();
   NFB_Reduziert = FALSE;
   do{
      AdHocPfadNeu(Term, objekt);
      ImLetztenDurchlaufReduziert = FALSE;
      do{
         if (NFB_RegelAngewendetAH()){
            ImLetztenDurchlaufReduziert = TRUE;
            while(NFB_RegelAngewendetAH());
         }
      }while(AdHocPfadWeitersetzen());
   }while(ImLetztenDurchlaufReduziert);
   ZM_StoppNFBildung();
   return NFB_Reduziert;
}

BOOLEAN NFA_NormalformZuGleichungenDirekt(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach erfolgter Reduktion wird direkt weitergegangen.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   BOOLEAN ImLetztenDurchlaufReduziert;

   ZM_StartNFBildung();
   NFB_Reduziert = FALSE;
   do{
      AdHocPfadNeu(Term, objekt);
      ImLetztenDurchlaufReduziert = FALSE;
      do{
         ImLetztenDurchlaufReduziert |= NFB_GleichungAngewendetAH();
      }while(AdHocPfadWeitersetzen());
   }while(ImLetztenDurchlaufReduziert);
   ZM_StoppNFBildung();
   return NFB_Reduziert;
}


BOOLEAN NFA_NormalformZuGleichungenNochmal(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach erfolgter Reduktion wird zunaechst die betroffene Stelle fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
   BOOLEAN ImLetztenDurchlaufReduziert;

   ZM_StartNFBildung();
   NFB_Reduziert = FALSE;
   do{
      AdHocPfadNeu(Term, objekt);
      do{
         if ((ImLetztenDurchlaufReduziert = NFB_GleichungAngewendetAH())){
            while(NFB_GleichungAngewendetAH());
         }
      }while(AdHocPfadWeitersetzen());
   }while(ImLetztenDurchlaufReduziert);
   ZM_StoppNFBildung();
   return NFB_Reduziert;
}

BOOLEAN NFA_NormalformBildenNochmal(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regel- und Gleichungssystems.
   Nach erfolgter Reduktion wird zunaechst die betroffene Stelle fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
  BOOLEAN reduziertG;

  NFB_Reduziert = FALSE;
  do{
    NFA_NormalformZuRegelnNochmal(Term, objekt);      /* zuerst NF zu Regeln */
    ZM_StartNFBildung();
    AdHocPfadNeu(Term, objekt);                                      /* ... dann zu Gleichungen */
    do{
      reduziertG = NFB_GleichungAngewendetAH();
    }while(!reduziertG && AdHocPfadWeitersetzen());
    ZM_StoppNFBildung();
  }while(reduziertG);
  return NFB_Reduziert;
}

BOOLEAN NFA_NormalformBildenDirekt(TermT Term, ObjektT objekt)
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regel- und Gleichungssystems.
   Nach erfolgter Reduktion wird direkt weitergegangen.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/
{
  BOOLEAN reduziertG;
  
  NFB_Reduziert = FALSE;
  do{
    NFA_NormalformZuRegelnDirekt(Term, objekt);
    ZM_StartNFBildung();
    AdHocPfadNeu(Term, objekt);
    do{
      reduziertG = NFB_GleichungAngewendetAH();
    }while(!reduziertG && AdHocPfadWeitersetzen());
    ZM_StoppNFBildung();
  }while(reduziertG);
  return NFB_Reduziert;
}
