/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - pcl.c                                                                                                      */
/* - *****                                                                                                      */
/* -                                                                                                            */
/* - Enthaelt alle fuer den PCL-Modus notwendigen Routinen, Zaehler etc..                                       */
/* -                                                                                                            */
/* - ACHTUNG: Das Abklemmen des PCL-Modus bewirkt, dass - aehnlich wie im Modul                                 */
/* -          Ausgaben.h die Makros - die Funktionen auf 'Nix tun'-Makros  umgestellt werden. Es ist also       */
/* -          keine weitere Aenderung erforderlich. Ferner finden auch alle Abfragen innerhalb dieser Funktionen*/
/* -          statt, so dass an den betreffenden Stellen im Code nur der jeweilige Makro-Aufruf erscheint.      */
/* -                                                                                                            */
/* - 30.11.94 Arnim. Angelegt.                                                                                  */
/* - Mai 1996 Arnim. Chaining, Mikro-PCL eingebaut                                                              */
/* - Mai 1997 Bernd. iPCL                                                                                       */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/


/* -------------------------------------------------------------------------------------------------------------
   Aufteilung in nPCL, uPCLp, uPCLr und iPCL
   - Es gibt fuenf Modi: 
     * den normalen, ausfuehrlichen PCL-Modus (nPCL),
     * den produzierenden Mikro-PCL Modus (uPCLp),
     * den reproduzierenden Mikro-PCL Modus (uPCLr),
     * den integrierten PCL-Modus (iPCL) und
     * den abgeschalteten PCL-Modus.
   - Im uPCLp-Modus werden nur Abhaengigkeiten von Fakten in die Protokoll-Datei geschrieben,
     die Nummern der Fakten selbst werden nicht ausgegeben.
     * Die Behandlung der verschiedenen Schritte zeigt folgende Tabelle:
         initial       0     1
         hypothesis    0     2
         orient       Nr     0
         cp          Vater Mutter
         tes-red      Nr1   Nr2
         instance     Nr1   Nr2
         tes-final     0     0    Nr1   0
              oder     0     0    Nr1  Nr2
     * Damit haben alle Ausgaben eine gerade Laenge
     * Werden die Nummern der Fakten byte-weise geschrieben, so ist positionieren in der Datei
       leicht moeglich.
   - Extraktion erzeugt eine Liste der noetigen Fakten.
   - Im uPCLr-Modus erhaelt das System diese Liste der noetigen Fakten. Es werden nur noch diese
     notwendigen Fakten ausfuehrlich ausgegeben.

   - im iPCL-Modus werden die Rekonstruktionsdaten in ein dynamisches Array geschriebe. Am Ende wird
     ein extract auf dem Array gemacht und die benoetigten Inferenzen werden nachvollzogen.

-------------------------------------------------------------------------------------------------------------*/

#include "general.h"
#include <errno.h>
#include <stdio.h>
#include "FehlerBehandlung.h"
#include "TermOperationen.h"
#include "Ausgaben.h"
#include "pcl.h"
#include "Parameter.h"
#include "Unifikation1.h"
#include <unistd.h>
#include "StrahlMemorium.h"
#include "MatchOperationen.h"
#include "DSBaumOperationen.h"
#include <string.h>

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe pcl.h) -------------------------------*/
/* ------------------------------------------------------------------------*/

#if COMP_PCL >= 1
void (* PCLProc_GleichungAusDerSpezifikation) (TermT, TermT, PCL_ObjektNummerT *);
void (* PCLProc_ZielAusDerSpezifikation) (TermT, TermT, PCL_ObjektNummerT *);

void (* PCLProc_Reduktionsbeginn) (PCL_ObjektT, PCL_ObjektNummerT *, PCL_SeitenT, TermT);
void (* PCLProc_Reduktion) (TermT, UTermT, PCL_ObjektNummerT *, PCL_SeitenT);

void (* PCLProc_Unifikation) (PCL_ObjektNummerT *, TermT, TermT, RegelOderGleichungsT,
                     TermT/* ggf umbenannt!!! */, UTermT, RegelOderGleichungsT);
void (* PCLProc_ReUnifikation) (PCL_ObjektNummerT *, TermT, TermT, RegelOderGleichungsT, int,
                                RegelOderGleichungsT);

void (* PCLProc_KPZuRegel) (TermT, TermT, PCL_RichtungsRichtungT, PCL_ObjektNummerT *);

void (* PCLProc_ZielSubsummiert) (PCL_ObjektNummerT *,
                         TermT, TermT,
                         PCL_ObjektNummerT *);

void (* PCLProc_ZielBewiesen) (PCL_ObjektNummerT *, 
                               TermT, TermT);
void (* PCLProc_StartChain) (void);
void (* PCLProc_StartFeyChain) (void);
void (* PCLProc_SaveChain) (void);
void (* PCLProc_DeleteChain) (void);

void (* PCLProc_VVAbschlussFaktDrucken) (PCL_ObjektNummerT *, TermT, TermT, PCL_ObjektT);

void (* PCLProcNeueRegel) (PCL_ObjektNummerT *, long int);
void (* PCLProcNeueGleichung) (PCL_ObjektNummerT *, long int);
void (* PCLProcRegelRechtsReduziert) (PCL_ObjektNummerT *, long int);

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Modul-lokale Variablen und Typen ------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static PCL_ObjektNummerT PCL_Objektzaehler;

#define PCL_KOMMENTARE 0                           /* Gibt in jeder Pcl-Zeile einen Kommentar mit aus, falls auf 1 */

#if PCL_KOMMENTARE
#define Kommentar(/*String*/ kommentar) fprintf(PCL_Protokolldatei, "     # %s\n", kommentar)
#else
#define Kommentar(/*String*/ kommentar) /*Nix tun*/
#endif

static PCL_ObjektNummerT *AktuelleID;
static PCL_ObjektT AktuelleObjektArt;
static PCL_SeitenT AktuelleSeite;
static TermT AktuelleAndereSeite;

static PCL_ObjektNummerT NaechsteWichtigeID=0, NaechsteWichtigeIDSAVED;
static unsigned long int NaechsteLesePositionSAVED;

static char *QuelldateiName, *ZieldateiName, *CorrdateiName;
static FILE *Quelldatei, *Zieldatei, *Corrdatei;

static StellenT *UnifStelle;
static unsigned int UnifStelleDimension;                                            /* letztes beschreibbares Feldelement */
static int IndexInUnifStelle;                                        /* -1 plus genaue Listenlaenge (leere Liste -> null) */

/* ----------------------------------------------------------------------------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------------------------- */
/* -- iPCL --------------------------------------------------------------------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------------------------- */

#if COMP_PCL == 2

static DomaenenT ipcl_Domaene;
iPCL_Rec *iPCL_Memo;
long iPCL_lastId, iPCL_AktPclID;
PCL_ObjektNummerT *iPCL_AktuelleID;
PCL_SeitenT iPCL_AktuelleSeite;
unsigned long iPCL_Performed[iPCL_EntryTypeSize], iPCL_Recorded[iPCL_EntryTypeSize], iPCL_Used[iPCL_EntryTypeSize];

int iPCL_doIt = 0, iPCL_doItReally = 0;

#else

static DomaenenT ipcl_Domaene;
static iPCL_Rec *iPCL_Memo;
static long iPCL_lastId, iPCL_AktPclID;
static PCL_ObjektNummerT *iPCL_AktuelleID;
static PCL_SeitenT iPCL_AktuelleSeite;
static unsigned long iPCL_Performed[iPCL_EntryTypeSize], iPCL_Recorded[iPCL_EntryTypeSize], iPCL_Used[iPCL_EntryTypeSize];

#endif

static const char *restpstr[] = {"eqn", "rule", "goal", "final"};

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Pfad berechnen ------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



static void GgfUnifStelleVerlaengern(TermT Term)
{
  unsigned int Tiefe = TO_Termtiefe(Term);
  if (Tiefe>UnifStelleDimension)
    UnifStelle = (StellenT *) our_realloc(UnifStelle,
     (( UnifStelleDimension = NF_PfadArrayGroesse+Tiefe)+1)*sizeof(StellenT),"Erweitern des Feldes UnifStelle");
}


static BOOLEAN StelleBerechnenRek(TermT Term, UTermT Teilterm)
/* Berechnet die Stelle, die Teilterm in Term einnimmt. Wird in das Array UnifStelle geschrieben und kann
   anschliessend im PCL-Protokoll zur Ausgabe verwendet werden.*/
{
  if (TO_PhysischGleich(Term,Teilterm))
    return TRUE;
  else if (TO_Nullstellig(Term))
    return FALSE;
  else {
    UTermT NULLTerm = TO_NULLTerm(Term);
    UnifStelle[++IndexInUnifStelle] = 1;
    Term = TO_ErsterTeilterm(Term);
    do {
      if (StelleBerechnenRek(Term,Teilterm))
        return TRUE;
      else
        UnifStelle[IndexInUnifStelle]++;
    } while (TO_PhysischUngleich(Term = TO_NaechsterTeilterm(Term),NULLTerm));
    IndexInUnifStelle--;
    return FALSE;
  }
}


static unsigned int TiefeBerechneterStelle(TermT Term, UTermT Teilterm)
{
  GgfUnifStelleVerlaengern(Term);
  IndexInUnifStelle = -1;
  StelleBerechnenRek(Term,Teilterm);
/*  IO_DruckeFlex("%t in %t hat Stelle:  ", Teilterm, Term);
{
   int i;
   for (i=0; i<= IndexInUnifStelle; i++)
      printf("%d.", UnifStelle[i]);
   printf("\n");
}*/
  return 1+IndexInUnifStelle;
}

static int AktuellePosition;

static BOOLEAN StelleBerechnenAusNummerRek(TermT Term)
/* Berechnet die Stelle, die Teilterm in Term einnimmt. Wird in das Array UnifStelle geschrieben und kann
   anschliessend im PCL-Protokoll zur Ausgabe verwendet werden.*/
{
  if (AktuellePosition==0)
    return TRUE;
  else if (TO_Nullstellig(Term))
    return FALSE;
  else {
    UTermT NULLTerm = TO_NULLTerm(Term);
    UnifStelle[++IndexInUnifStelle] = 1;
    Term = TO_ErsterTeilterm(Term);
    do {
       AktuellePosition--;
       if (StelleBerechnenAusNummerRek(Term))
          return TRUE;
       else
          UnifStelle[IndexInUnifStelle]++;
    } while (TO_PhysischUngleich(Term = TO_NaechsterTeilterm(Term),NULLTerm));
    IndexInUnifStelle--;
    return FALSE;
  }
}

static unsigned int TiefeBerechneterStelleAusNummer(TermT Term, int Nummer)
{
  GgfUnifStelleVerlaengern(Term);
  IndexInUnifStelle = -1;
  AktuellePosition = Nummer;
  StelleBerechnenAusNummerRek(Term);
  return 1+IndexInUnifStelle;
}

void pcltest(TermT term)
{
   TermT Index;
   int Tiefe, i;
   TO_doStellen(term, Index, 
                {
                   Tiefe = TiefeBerechneterStelle(term, Index);
                   IO_DruckeFlex("%t in %t hat Tiefe %d, die Stelle ist : ", Index, term, Tiefe);
                   for(i=0; i<=IndexInUnifStelle; i++){
                      fprintf(IO_Strom, "%d ", UnifStelle[i]);
                   }
                   fprintf(IO_Strom, "\n\n");
                });
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Alle Funktionen -----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void nPCL_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void nPCL_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void nPCL_Reduktionsbeginn(PCL_ObjektT objektArt,PCL_ObjektNummerT *nr,PCL_SeitenT seite,TermT andereSeite);
void nPCL_Reduktion(TermT ergebnis,UTermT betroffenerTeilterm,PCL_ObjektNummerT *AngewandtesFakt, 
                    PCL_SeitenT angewandteSeite);
void nPCL_Unifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
                      TermT betroffenerVaterTerm,UTermT betroffenerVaterTeilterm,RegelOderGleichungsT Mutter);
void nPCL_ReUnifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
                        int NummerBetroffeneVaterStelle,RegelOderGleichungsT Mutter);
void nPCL_KPZuRegel(TermT links, TermT rechts,PCL_RichtungsRichtungT RichtungsRichtung,
                    PCL_ObjektNummerT *RegelID);
void nPCL_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,TermT links, TermT rechts,
                          PCL_ObjektNummerT *subsummierendeGleichung);
void nPCL_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,TermT links, TermT rechts);
void nPCL_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art);

void uPCLp_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void uPCLp_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void uPCLp_Reduktionsbeginn(PCL_ObjektT objektArt,PCL_ObjektNummerT *nr,PCL_SeitenT seite,TermT andereSeite);
void uPCLp_Reduktion(TermT ergebnis,UTermT betroffenerTeilterm,PCL_ObjektNummerT *AngewandtesFakt, 
                     PCL_SeitenT angewandteSeite);
void uPCLp_Unifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
                       TermT betroffenerVaterTerm,UTermT betroffenerVaterTeilterm,RegelOderGleichungsT Mutter);
void uPCLp_ReUnifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
                         int NummerBetroffeneVaterStelle,RegelOderGleichungsT Mutter);
void uPCLp_KPZuRegel(TermT links, TermT rechts,PCL_RichtungsRichtungT RichtungsRichtung,
                     PCL_ObjektNummerT *RegelID);
void uPCLp_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,TermT links, TermT rechts,
                           PCL_ObjektNummerT *subsummierendeGleichung);
void uPCLp_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,TermT links, TermT rechts);
void uPCLp_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art);

void uPCLr_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void uPCLr_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void uPCLr_Reduktionsbeginn(PCL_ObjektT objektArt,PCL_ObjektNummerT *nr,PCL_SeitenT seite,TermT andereSeite);
void uPCLr_Reduktion(TermT ergebnis,UTermT betroffenerTeilterm,PCL_ObjektNummerT *AngewandtesFakt, 
                     PCL_SeitenT angewandteSeite);
void uPCLr_Unifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
                       TermT betroffenerVaterTerm,UTermT betroffenerVaterTeilterm,RegelOderGleichungsT Mutter);
void uPCLr_ReUnifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
                         int NummerBetroffeneVaterStelle,RegelOderGleichungsT Mutter);
void uPCLr_KPZuRegel(TermT links, TermT rechts,PCL_RichtungsRichtungT RichtungsRichtung,
                     PCL_ObjektNummerT *RegelID);
void uPCLr_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,TermT links, TermT rechts,
                           PCL_ObjektNummerT *subsummierendeGleichung);
void uPCLr_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,TermT links, TermT rechts);
void uPCLr_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art);

void iPCL_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void iPCL_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void iPCL_Reduktionsbeginn(PCL_ObjektT objektArt,PCL_ObjektNummerT *nr,PCL_SeitenT seite,TermT andereSeite);
void iPCL_Reduktion(TermT ergebnis,UTermT betroffenerTeilterm,PCL_ObjektNummerT *AngewandtesFakt, 
            	    PCL_SeitenT angewandteSeite);
void iPCL_Unifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
          	      TermT betroffenerVaterTerm,UTermT betroffenerVaterTeilterm,RegelOderGleichungsT Mutter);
void iPCL_ReUnifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
            		int NummerBetroffeneVaterStelle,RegelOderGleichungsT Mutter);
void iPCL_KPZuRegel(TermT links, TermT rechts,PCL_RichtungsRichtungT RichtungsRichtung,
           	    PCL_ObjektNummerT *RegelID);
void iPCL_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,TermT links, TermT rechts,
          		  PCL_ObjektNummerT *subsummierendeGleichung);
void iPCL_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,TermT links, TermT rechts);
void iPCL_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art);

void Dummy_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void Dummy_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr);
void Dummy_Reduktionsbeginn(PCL_ObjektT objektArt,PCL_ObjektNummerT *nr,PCL_SeitenT seite,TermT andereSeite);
void Dummy_Reduktion(TermT ergebnis,UTermT betroffenerTeilterm,PCL_ObjektNummerT *AngewandtesFakt, 
                     PCL_SeitenT angewandteSeite);
void Dummy_Unifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
                       TermT betroffenerVaterTerm,UTermT betroffenerVaterTeilterm,RegelOderGleichungsT Mutter);
void Dummy_ReUnifikation(PCL_ObjektNummerT *nr,TermT links, TermT rechts,RegelOderGleichungsT Vater,
                         int NummerBetroffeneVaterStelle,RegelOderGleichungsT Mutter);
void Dummy_KPZuRegel(TermT links, TermT rechts,PCL_RichtungsRichtungT RichtungsRichtung,
                     PCL_ObjektNummerT *RegelID);
void Dummy_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,TermT links, TermT rechts,
                           PCL_ObjektNummerT *subsummierendeGleichung);
void Dummy_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,TermT links, TermT rechts);
void Dummy_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art);

void PCL_SaveFeyChain(void);
void PCL_DeleteFeyChain(void);

void FULL_StartFeyChain(void);
void FULL_StartChain(void);
void FULL_SaveChain(void);
void FULL_DeleteChain(void);

void uPCLr_StartChain(void);
void uPCLr_SaveChain(void);
void uPCLr_DeleteChain(void);

void iPCL_StartFeyChain(void);
void iPCL_StartChain(void);
void iPCL_SaveChain(void);
void iPCL_DeleteChain(void);

void Dummy_StartFeyChain(void);
void Dummy_StartChain(void);
void Dummy_SaveChain(void);
void Dummy_DeleteChain(void);

void FULL_NeueRegel(PCL_ObjektNummerT *pclIDPtr, long int RegelNummer);
void FULL_NeueGleichung(PCL_ObjektNummerT *pclIDPtr, long int GleichungsNummer);
void FULL_RegelRechtsReduziert(PCL_ObjektNummerT *pclIDPtr, long int RegelNummer);

void Dummy_NeueRegel(PCL_ObjektNummerT *pclIDPtr, long int RegelNummer);
void Dummy_NeueGleichung(PCL_ObjektNummerT *pclIDPtr, long int GleichungsNummer);
void Dummy_RegelRechtsReduziert(PCL_ObjektNummerT *pclIDPtr, long int RegelNummer);

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Funktionspointer besetzen -------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define DummyMode 0 
#define nPCLMode  1
#define uPCLpMode 2
#define uPCLrMode 3
#define iPCLMode  4

#define DummyChainMode 0
#define FULLChainMode  1
#define uPCLrChainMode 2
#define iPCLChainMode  3

#define DummyCorrMode 0
#define FullCorrMode  1

static int PCL_Mode, Chain_Mode, Corr_Mode;


#define FunktionspointerBesetzenBODY(Praefix)                                                                       \
PCLProc_GleichungAusDerSpezifikation   = Praefix ## _GleichungAusDerSpezifikation;                              \
PCLProc_ZielAusDerSpezifikation        = Praefix ## _ZielAusDerSpezifikation;                                   \
PCLProc_Reduktionsbeginn               = Praefix ## _Reduktionsbeginn;                                          \
PCLProc_Reduktion                      = Praefix ## _Reduktion;                                                 \
PCLProc_Unifikation                    = Praefix ## _Unifikation;                                               \
PCLProc_ReUnifikation                  = Praefix ## _ReUnifikation;                                             \
PCLProc_KPZuRegel                      = Praefix ## _KPZuRegel;                                                 \
PCLProc_ZielSubsummiert                = Praefix ## _ZielSubsummiert;                                           \
PCLProc_ZielBewiesen                   = Praefix ## _ZielBewiesen;                                              \
PCLProc_VVAbschlussFaktDrucken         = Praefix ## _VVAbschlussFaktDrucken;

#define ChainingBesetzenBODY(Praefix)                                                                               \
PCLProc_StartChain   = Praefix ## _StartChain;                                                                  \
PCLProc_SaveChain    = Praefix ## _SaveChain;                                                                   \
PCLProc_DeleteChain  = Praefix ## _DeleteChain;

#define CorrBesetzenBODY(Praefix)                                                                               \
PCLProcNeueRegel             = Praefix ## _NeueRegel;                                                           \
PCLProcNeueGleichung         = Praefix ## _NeueGleichung;                                                       \
PCLProcRegelRechtsReduziert  = Praefix ## _RegelRechtsReduziert;


static void FunktionspointerBesetzen(int Mode)
{
   switch(Mode){
   case nPCLMode:  FunktionspointerBesetzenBODY(nPCL);break;
   case uPCLpMode: FunktionspointerBesetzenBODY(uPCLp);break;
   case uPCLrMode: FunktionspointerBesetzenBODY(uPCLr);break;
   case iPCLMode:  FunktionspointerBesetzenBODY(iPCL);break;
   case DummyMode:
   default:        FunktionspointerBesetzenBODY(Dummy);break;
   }
}

static void ChainingBesetzen(int Mode)
{
   switch (Mode){
   case FULLChainMode :  ChainingBesetzenBODY(FULL);  PCLProc_StartFeyChain = FULL_StartFeyChain; break;
   case uPCLrChainMode:  ChainingBesetzenBODY(uPCLr); PCLProc_StartFeyChain = FULL_StartFeyChain; break;
   case iPCLChainMode:   ChainingBesetzenBODY(iPCL);  PCLProc_StartFeyChain = FULL_StartFeyChain; break;
   case DummyChainMode: 
   default:              ChainingBesetzenBODY(Dummy); PCLProc_StartFeyChain = Dummy_StartFeyChain;break;
   }
}

static void CorrBesetzen(int Mode)
{
   switch(Mode){
   case FullCorrMode:  CorrBesetzenBODY(FULL); break;
   case DummyCorrMode:
   default:            CorrBesetzenBODY(Dummy); break;
   }
}


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CHAINING ------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


static BOOLEAN ChainStarted, FeyChainStarted;

static long AktPos;

static PCL_ObjektNummerT AktPclID;

/* interne Konsistenztest: Alternativen 0,1 */
#define PCL_TEST 0
#if PCL_TEST
#define PCLPrint(bla) fprintf(Zieldatei, bla)
#else
#define PCLPrint(bla) 
#endif

void FULL_StartFeyChain(void)
/* Alle folgenden Schritte gar nicht protokollieren. */
{
   /* Funktions-Pointer auf Dummies */
   FunktionspointerBesetzen(DummyMode);
   PCLProc_SaveChain = PCL_SaveFeyChain;
   PCLProc_DeleteChain = PCL_DeleteFeyChain;

#if PCL_TEST
   /* Aktuelle Pcl-ID und Filepos merken */
   AktPclID = PCL_Objektzaehler;
   AktPos = ftell(Zieldatei);
#endif
   PCLPrint("vvvvv START FEY CHAIN vvvvv\n");
   if (FeyChainStarted){
      printf(">>>>> ERROR: 2x PCL_StartFeyChain\n");
      PCLPrint(">>>>> ERROR: 2x PCL_StartFeyChain\n");
   }
   else
      FeyChainStarted = TRUE;
}

void PCL_SaveFeyChain(void)
{
   printf(">>>>> ERROR: SaveFeyChain\n");
   PCLPrint(">>>>> ERROR: SaveFeyChain\n");
}

void PCL_DeleteFeyChain(void)
/* Vorlaeufig protokollierte Schritte loeschen. */
{
   /* Pointer ruecksetzen */
   FunktionspointerBesetzen(PCL_Mode);
   ChainingBesetzen(Chain_Mode);
   PCLPrint("^^^^^ DELETE FEY CHAIN ^^^^^\n");
   if (!FeyChainStarted){
      printf(">>>>> ERROR: PCL_DeleteFeyChain ohne PCL_StartFeyChain\n");
      PCLPrint(">>>>> ERROR: PCL_DeleteFeyChain ohne PCL_StartFeyChain\n");
   }
   else
      FeyChainStarted = FALSE;
   
#if PCL_TEST
   if (PCL_Objektzaehler != AktPclID){
      printf(">>>>> ERROR: In Fey-Chain Pcl-Schritte geschrieben (IDs)\n");
      PCLPrint(">>>>> ERROR: In Fey-Chain Pcl-Schritte geschrieben (IDs)\n");
   }
   if (AktPos != ftell(Zieldatei)){
      printf(">>>>> ERROR: In Fey-Chain Pcl-Schritte geschrieben (FilePos)\n");
      PCLPrint(">>>>> ERROR: In Fey-Chain Pcl-Schritte geschrieben (FilePos)\n");
   }
#endif
}


void FULL_StartChain(void)
/* Alle folgenden Schritte nur vorlaeufig protokollieren. */
{
   /* Aktuelle Pcl-ID merken */
   AktPclID = PCL_Objektzaehler;
   AktPos = ftell(Zieldatei);
   PCLPrint("vvvvv START CHAIN vvvvv\n");
   if (ChainStarted){
      printf(">>>>> ERROR: 2x PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: 2x PCL_StartChain\n");
   }
   else
      ChainStarted = TRUE;
}


void FULL_SaveChain(void)
/* Vorlaeufig protokollierte Schritte speichern. */
{
   /* temporaere Datei loeschen/auf Anfang setzen */
   PCLPrint("^^^^^ SAVE CHAIN ^^^^^\n");
   if (!ChainStarted){
      printf(">>>>> ERROR: PCL_SaveChain ohne PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: PCL_SaveChain ohne PCL_StartChain\n");
   }
   else
      ChainStarted = FALSE;
}

void FULL_DeleteChain(void)
/* Vorlaeufig protokollierte Schritte loeschen. */
{
   PCLPrint("^^^^^ DELETE CHAIN ^^^^^\n");
   if (!ChainStarted){
      printf(">>>>> ERROR: PCL_DeleteChain ohne PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: PCL_DeleteChain ohne PCL_StartChain\n");
   }
   else
      ChainStarted = FALSE;
   /* In StartChain gemerkte ID als aktuelle rueck-uebernehmen */
   PCL_Objektzaehler = AktPclID;
   /* Pcl-Datei zuruecksetzen setzen */
   fseek(Zieldatei, AktPos, SEEK_SET);
}


void uPCLr_StartChain(void)
/* Alle folgenden Schritte nur vorlaeufig protokollieren. */
{
   /* Aktuelle Pcl-ID merken */
   AktPclID = PCL_Objektzaehler;
   NaechsteWichtigeIDSAVED = NaechsteWichtigeID;

   AktPos = ftell(Zieldatei);
   NaechsteLesePositionSAVED = ftell(Quelldatei);

   PCLPrint("vvvvv START CHAIN vvvvv\n");
   if (ChainStarted){
      printf(">>>>> ERROR: 2x PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: 2x PCL_StartChain\n");
   }
   else
      ChainStarted = TRUE;
}


void uPCLr_SaveChain(void)
/* Vorlaeufig protokollierte Schritte speichern. */
{
   /* temporaere Datei loeschen/auf Anfang setzen */
   PCLPrint("^^^^^ SAVE CHAIN ^^^^^\n");
   if (!ChainStarted){
      printf(">>>>> ERROR: PCL_SaveChain ohne PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: PCL_SaveChain ohne PCL_StartChain\n");
   }
   else
      ChainStarted = FALSE;
}

void uPCLr_DeleteChain(void)
/* Vorlaeufig protokollierte Schritte loeschen. */
{
   PCLPrint("^^^^^ DELETE CHAIN ^^^^^\n");
   if (!ChainStarted){
      printf(">>>>> ERROR: PCL_DeleteChain ohne PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: PCL_DeleteChain ohne PCL_StartChain\n");
   }
   else
      ChainStarted = FALSE;

   /* In StartChain gemerkte ID als aktuelle rueck-uebernehmen */
   PCL_Objektzaehler = AktPclID;
   NaechsteWichtigeID = NaechsteWichtigeIDSAVED;

   /* Pcl-Datei zuruecksetzen */
   fseek(Zieldatei, AktPos, SEEK_SET);
   fseek(Quelldatei, NaechsteLesePositionSAVED, SEEK_SET);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* -- iPCL -----------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void iPCL_StartChain(void)
/* Alle folgenden Schritte nur vorlaeufig protokollieren. */
{
   /* Aktuelle Pcl-ID merken */
   iPCL_AktPclID = iPCL_lastId;
   PCLPrint("vvvvv START CHAIN vvvvv\n");
   if (ChainStarted){
      printf(">>>>> ERROR: 2x PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: 2x PCL_StartChain\n");
   }
   else
      ChainStarted = TRUE;
}


void iPCL_SaveChain(void)
/* Vorlaeufig protokollierte Schritte speichern. */
{
   PCLPrint("^^^^^ SAVE CHAIN ^^^^^\n");
   if (!ChainStarted){
      printf(">>>>> ERROR: PCL_SaveChain ohne PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: PCL_SaveChain ohne PCL_StartChain\n");
   }
   else
      ChainStarted = FALSE;
}

void iPCL_DeleteChain(void)
/* Vorlaeufig protokollierte Schritte loeschen. */
{
   PCLPrint("^^^^^ DELETE CHAIN ^^^^^\n");
   if (!ChainStarted){
      printf(">>>>> ERROR: PCL_DeleteChain ohne PCL_StartChain\n");
      PCLPrint(">>>>> ERROR: PCL_DeleteChain ohne PCL_StartChain\n");
   }
   else
      ChainStarted = FALSE;
   /* In StartChain gemerkte ID als aktuelle rueck-uebernehmen */
   iPCL_lastId = iPCL_AktPclID;
}



/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- NORMALES PCL --------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Gleichung/Ziel aus der Spezifikation --------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void nPCL_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
/* Die Gleichungen aus der Spezifiaktion werden beim Einfuegen in die KP-Menge mit Gewicht 0 als 'initial' ins PCL-Protokoll (PP)
   geschrieben. Die Nummer wird mit einer neuen PCL-ID besetzt.
<pcl-id> : tes-eqn : links = rechts : initial
*/
{
   *nr = ++PCL_Objektzaehler;
   fprintf(Zieldatei,"%lu : tes-eqn : ",*nr);
   IO_DruckeTermStrom(Zieldatei, links);
   fprintf(Zieldatei, " = ");
   IO_DruckeTermStrom(Zieldatei, rechts);
   fprintf(Zieldatei, " : initial\n");
   Kommentar("equation from specification");
}



void nPCL_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
/* Einlesen der Ziele aus der Spezifikation:
   Die Ziele werden bei der Aufnahme in die interne Datenstruktur als tes-goal gemeldet.
   Die Nummer wird mit einer neuen PCL-ID besetzt.
<pcl-id> : tes-goal : links = rechts : hypothesis
*/
{
   *nr = ++PCL_Objektzaehler;
   fprintf(Zieldatei,"%lu : tes-goal : ",*nr);
   IO_DruckeTermStrom(Zieldatei, links);
   fprintf(Zieldatei, " = ");
   IO_DruckeTermStrom(Zieldatei, rechts);
   fprintf(Zieldatei, " : hypothesis\n");
   Kommentar("goal from spezification");
}


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------------- REDUKTION -----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


/* --------------------------------------------------------------------------------------------------------------------------------
   Reduktion:
   Jeder Reduktionsschritt wird mit betroffenem <fact>, angewandter <rule>/<eqn>, betroffener Seite+Stelle sowie angewandter Seite
   ins PPP geschrieben.
<pcl-id> : <step-type> : links = rechts : tes-red( <pcl-expr>, <place>, <pcl-expr>, <side>)
  ^Nummer   ^Art des entst.  ^Ergebnis              ^Nr. des red. ^Pos.     ^Nr der  ^Richtung der Anwendung(l->r bzw. r->l)
             Faktums          der Red.               Faktums       der Red.  angewandten Regel/Gleichung

   Was wird reduziert?

   1. KPs
   2. Regeln rechts
   3. Regeln links
   4. Gleichungen
   5. Ziele
   
   Welche Meldungen sind hierbei erforderlich?

   1. <pcl-id> : tes-eqn : links = rechts : tes-red( <pcl-expr>, <place>, <pcl-expr>, <side>)
   2. <pcl-id> : tes-rule : links -> rechts : tes-red( <pcl-expr>, <place>, <pcl-expr>, <side>)
   3. <pcl-id> : tes-eqn : links = rechts : tes-red( <pcl-expr>, <place>, <pcl-expr>, <side>)
   4. <pcl-id> : tes-eqn : links = rechts : tes-red( <pcl-expr>, <place>, <pcl-expr>, <side>)
   5. <pcl-id> : tes-goal : links = rechts : tes-red( <pcl-expr>, <place>, <pcl-expr>, <side>)

   Damit kann man folgendermassen vorgehen:
   - Vor/bei Eintritt in eine der Reduktionsroutinen aus NFBildung wird das zu reduzierende Objekt
     mit pcl-Nr. und Objekt-Art (RegelLinks, RegelRechts, KPLinks, KPRechts, GleichungLinks,
     GleichungRechts, ZielLinks, ZielRechts) sowie der jeweils anderen Seite angemeldet.
   - Jeder erfolgte Reduktionsschritt wird mit PCL_Reduktion gemeldet. In dieser Routine wird dann
     die neue pcl-Nr. erzeugt und (an den VAR-Parameter) vergeben sowie die entsprechende Meldung ausgeloest
     und - falls es sich um die linke Seite einer Regel handelte - der Objekt-Typ auf KP umgestellt.

   Es faellt auf, dass es im wesentlichen nur vier - hier - relevante Objektklassen gibt:

   1. Eine Seite einer Gleichung (tes-eqn) (umfasst also KPs und Gleichungen)
   2. Rechte Seite einer Regel
   3. Linke Seite einer Regel
   4. Eine Seite eines Zieles

   Wenn man nun also beim Aufruf die Parameter "PCL-ObjektT, SeitenT, AndereSeite" mitgibt, so ist die Anmelde-Prozedur
   ausreichend ueber die Art des Objektes informiert.
   Dabei ist PCL-ObjektT = {Regel, Gleichung, Ziel}, SeitenT = {linkeSeite, rechteSeite}, AndereSeite = TermT.

   Die Reduktionsprozedur muss dann nur noch das Ergebnis der Reduktion sowie angewandtes Fakt und die betroffene Stelle melden.
   Ueber den VAR-Parameter wird dann sogar die pcl-nr automatisch angepasst (??? Vorsicht bei Interreduktion!!!!).

   ENTSCHEIDUNG: Die PCL-Nr. wird nicht ueber einen VAR-Parameter angepasst!
   Begruendung:  Bei der Interreduktion der Regelmenge (linke Seiten) ist nicht gewaehrleistet, dass die Nummer des reduzierten
                 Objektes auch physikalisch die gleiche bleibt. Ferner ist im PCL-Modus das Zeitverhalten ohnehin irrelevant.
   Entscheidung geaendert, da interreduzierte Objekte ja zunaechst im Puffer gespeichert werden, die PCL-ID also erhalten
   bleibt, und deshalb bei Weiterbehandlung des Objekts benutzt werden kann.
*/

void nPCL_Reduktionsbeginn(PCL_ObjektT objektArt, 
                          PCL_ObjektNummerT *nr,
                          PCL_SeitenT seite,
                          TermT andereSeite)
/* Zu Beginn der Normalformbildung wird das zu behandelnde Objekt angemeldet. Am Ende der NF-Bildung
   kann mit der Funktion PCL_NummerNachLetzterReduktion die Nummer des Objektes nach der letzten Reduktion
   abgeholt werden.
   Die nr wird bei jeder Reduktion auf die neue PCL-ID gesetzt.*/
{
   AktuelleID = nr;
   AktuelleObjektArt = objektArt;
   AktuelleSeite = seite;
   AktuelleAndereSeite = andereSeite;
}

void nPCL_Reduktion(TermT ergebnis,
                   UTermT betroffenerTeilterm,
                   PCL_ObjektNummerT *AngewandtesFakt, 
                   PCL_SeitenT angewandteSeite)
{
   StellenT i=0;
   StellenT TBS;
   
   if ((AktuelleObjektArt == Regel) && (AktuelleSeite == linkeSeite)){   /* Regel links reduziert -> ist jetzt Gleichung */
      AktuelleObjektArt = Gleichung;
   }
   fprintf(Zieldatei,"%lu : tes-", ++PCL_Objektzaehler);
   switch (AktuelleObjektArt){
   case Regel: fprintf(Zieldatei,"rule : ");
      break;
   case Ziel: fprintf(Zieldatei,"goal : ");
      break;
   default:  fprintf(Zieldatei,"eqn : ");
      break;
   }
   IO_DruckeTermStrom(Zieldatei, (AktuelleSeite == linkeSeite)?ergebnis:AktuelleAndereSeite);
   if (AktuelleObjektArt == Regel){
      fprintf(Zieldatei," -> ");
   }
   else{
      fprintf(Zieldatei," = ");
   }
   IO_DruckeTermStrom(Zieldatei, (AktuelleSeite == linkeSeite)?AktuelleAndereSeite:ergebnis);
   fprintf(Zieldatei," : tes-red( %lu, ",*AktuelleID);
   fprintf(Zieldatei,"%s", (AktuelleSeite == linkeSeite)?"L":"R");
   TBS = TiefeBerechneterStelle(ergebnis, betroffenerTeilterm);
   while (i<TBS){
      fprintf(Zieldatei,".%d", UnifStelle[i++]);
   }
   fprintf(Zieldatei,", %lu, %s)\n", *AngewandtesFakt, (angewandteSeite == linkeSeite)? "L":"R");
   *AktuelleID = PCL_Objektzaehler;
   Kommentar((AktuelleObjektArt == Regel)?
             ((AktuelleSeite == linkeSeite)?
              "lhs of rule rewritten":"rhs of rule rewritten"):
             ((AktuelleObjektArt == Ziel)?
              ((AktuelleSeite == linkeSeite)?
               "lhs of goal rewritten":"rhs of goal rewritten"):
              ((AktuelleSeite == linkeSeite)?
               "lhs of equation or critical pair rewritten":"rhs of equation or critical pair rewritten")));
}



/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Unifikation ---------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



void nPCL_Unifikation(PCL_ObjektNummerT *nr,
                     TermT links, TermT rechts,
                     RegelOderGleichungsT Vater,
                     TermT betroffenerVaterTerm, /* ggf umbenannt!!! */
                     UTermT betroffenerVaterTeilterm,
                     RegelOderGleichungsT Mutter)
/* Unifikation:
   Jede gelungene Unifikation wird als <cp-expr> ins PP geschrieben: beide Elternteile (<rule> und/oder <eqn>) 
   sowie genaue Beschreibung der Position. Die nr wird mit einer neuen PCL-ID besetzt. 
   Betroffene Stelle gibt die Position innerhalb des Vaters an.
   <pcl-id> : tes-eqn : links=rechts : cp(Vater, <place>, Mutter, <side>)
*/
{
   StellenT i=0,
     maxTiefe = TiefeBerechneterStelle(betroffenerVaterTerm, betroffenerVaterTeilterm);
   *nr = ++PCL_Objektzaehler;
   fprintf(Zieldatei,"%lu : tes-eqn : ", *nr);
   IO_DruckeTermStrom(Zieldatei, links);
   fprintf(Zieldatei," = ");
   IO_DruckeTermStrom(Zieldatei, rechts);
   fprintf(Zieldatei," : cp(%lu, %s", *(TP_PclID(Vater)), 
           RE_IstRegel(Vater)? "L":
           TP_RichtungAusgezeichnet(Vater)?"L":"R");
   while (i<maxTiefe) fprintf(Zieldatei,".%d", UnifStelle[i++]);
   fprintf(Zieldatei,", %lu, %s)\n", *(TP_PclID(Mutter)),
           RE_IstRegel(Mutter)? "L":
           TP_RichtungAusgezeichnet(Mutter)?"L":"R");
   Kommentar("critical pair generated");
}


void nPCL_ReUnifikation(PCL_ObjektNummerT *nr,
                     TermT links, TermT rechts,
                     RegelOderGleichungsT Vater,
                     int NummerBetroffeneVaterStelle,
                     RegelOderGleichungsT Mutter)
/* NummerBetroffeneVaterStelle ist die Nummer desjenigen Symbols in der linken Vater-Seite,
   das an der Stelle steht, wo die Ueberlappung stattfand.*/
{
   StellenT i=0,
   maxTiefe = TiefeBerechneterStelleAusNummer(TP_LinkeSeite(Vater), NummerBetroffeneVaterStelle);
   *nr = ++PCL_Objektzaehler;
   fprintf(Zieldatei,"%lu : tes-eqn : ", *nr);
   IO_DruckeTermStrom(Zieldatei, links);
   fprintf(Zieldatei," = ");
   IO_DruckeTermStrom(Zieldatei, rechts);
   fprintf(Zieldatei," : cp(%lu, %s", *(TP_PclID(Vater)), 
           RE_IstRegel(Vater)? "L":
           TP_RichtungAusgezeichnet(Vater)?"L":"R");
   while (i<maxTiefe) fprintf(Zieldatei,".%d", UnifStelle[i++]);
   fprintf(Zieldatei,", %lu, %s)\n", *(TP_PclID(Mutter)),
           RE_IstRegel(Mutter)? "L":
           TP_RichtungAusgezeichnet(Mutter)?"L":"R");
   Kommentar("critical pair rebuild");
}


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Aufnahme von KPs als Regeln oder Gleichungen ------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



void nPCL_KPZuRegel(TermT links, TermT rechts,
                   PCL_RichtungsRichtungT RichtungsRichtung,
                   PCL_ObjektNummerT *RegelID)
/* Orientierung:
   Die Orientierung von kritischen Paaren, also <eqn>'s, wird nur im Erfolgsfall ins PP gemeldet:
   <pcl-id> : tes-rule: links -> rechts : orient(<pcl-id>,x|u)
   *gerichtetesObjekt enthaelt beim Aufruf die PCL-ID des orientierten Objektes und wird dann mit der neuen PCL-ID besetzt.
*/
{
   fprintf(Zieldatei,"%lu : tes-rule : ", ++PCL_Objektzaehler);
   IO_DruckeTermStrom(Zieldatei, links);
   fprintf(Zieldatei," -> ");
   IO_DruckeTermStrom(Zieldatei, rechts);
   fprintf(Zieldatei," : orient(%lu, %s)\n", *RegelID, (RichtungsRichtung==VonLinksNachRechts)?"u":"x");
   *RegelID = PCL_Objektzaehler;
   Kommentar ("critical pair oriented as rule");
}



/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Ziele subsummieren/ zusammenfuehren ---------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



void nPCL_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,
                         TermT links, TermT rechts,   /* linke und rechte Seite des subsummierten Zieles */
                         PCL_ObjektNummerT *subsummierendeGleichung)
/* Subsumption:
   Wurde ein Ziel erfolgreich subsummiert, so ist dies als 
   <pcl-id> : tes-final : links = rechts : instance(<pcl-expr>, <pcl-expr>)
                                           das subs. Ziel^           ^die subsummierende Gleichung
   *subsummiertesZiel enthaelt beim Aufruf die PCL-ID des subsummierten Zieles und wird dann mit der neuen PCL-ID besetzt.
*/
{
   fprintf(Zieldatei,"%lu : tes-final : ", ++PCL_Objektzaehler);
   IO_DruckeTermStrom(Zieldatei, links);
   fprintf(Zieldatei," = ");
   IO_DruckeTermStrom(Zieldatei, rechts);
   fprintf(Zieldatei," : instance(%lu, %lu)\n", *subsummiertesZiel, *subsummierendeGleichung);
   *subsummiertesZiel = PCL_Objektzaehler;
   Kommentar ("subsumed goal");
}

void nPCL_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,
                      TermT links, TermT rechts)   /* linke und rechte Seite des bewiesenen Zieles */
/* Beweis von Zielen:
   Konnte ein Ziel bewiesen, i.e. zusammengefuehrt, werden, so wird dies mit <pcl-id> : tes-final: links=rechts:<quote-expr> gemeldet,
   wobei die <quote-expr> gerade die vorhergehende PCL-ID des Ziels bezeichnet.
   <pcl-id> : tes-final: links=rechts: <quote-expr>
   *bewiesenesZiel enthaelt die PCL-ID des bewiesenen Zieles und wird dann auf die neue ID gesetzt.
*/
{
   fprintf(Zieldatei,"%lu : tes-final : ", ++PCL_Objektzaehler);
   IO_DruckeTermStrom(Zieldatei, links);
   fprintf(Zieldatei," = ");
   IO_DruckeTermStrom(Zieldatei, rechts);
   fprintf(Zieldatei," : %lu\n", *bewiesenesZiel);
   *bewiesenesZiel = PCL_Objektzaehler;
   Kommentar("joined goal");
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Abschluss der Vervollstaendigung: alle Fakten ausgeben --------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void nPCL_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art)
/* Abschluss der Vervollstaendigung: Wurde ein (grund?, siehe PCL, S. 20) konvergentes Regelsystem erzielt, so
   sollen alle erzeugten Regeln und Gleichungen nochmal mit tes-final gemeldet werden.
   Dies soll hier geschehen. Mit jeder Regel und Gleichung ist diese Prozedur aufzurufen.
   Die neue PCL-ID wird an nr weitergegeben.*/
{
   PCL_ObjektNummerT alteID = *nr;
   
   *nr = ++PCL_Objektzaehler;
   fprintf(Zieldatei,"%lu : tes-final : ", *nr);
   IO_DruckeTermStrom(Zieldatei, links);
   fprintf(Zieldatei, " %s ",(art==Regel)?"->":"=");
   IO_DruckeTermStrom(Zieldatei, rechts);
   fprintf(Zieldatei," : %lu\n", alteID);
   Kommentar("part of final system");
}

/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- MIKRO-PCL -----------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- MIKRO-PCL Reproduzieren ---------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


BOOLEAN IstWichtig(PCL_ObjektNummerT ID)
{
   if (ID==NaechsteWichtigeID){
      fread(&NaechsteWichtigeID, sizeof(unsigned long int), 1, Quelldatei);
      if (feof(Quelldatei))
         NaechsteWichtigeID=0;
      return TRUE;
   }
   else{
      return FALSE;
   }
}
void uPCLr_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
{
   *nr = ++PCL_Objektzaehler;
   if (IstWichtig(*nr)){
      fprintf(Zieldatei,"%lu : tes-eqn : ",*nr);
      IO_DruckeTermStrom(Zieldatei, links);
      fprintf(Zieldatei, " = ");
      IO_DruckeTermStrom(Zieldatei, rechts);
      fprintf(Zieldatei, " : initial\n");
      Kommentar("Gleichung aus der Spezifikation");
   }
}



void uPCLr_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
{
   *nr = ++PCL_Objektzaehler;
   if (IstWichtig(*nr)){
      fprintf(Zieldatei,"%lu : tes-goal : ",*nr);
      IO_DruckeTermStrom(Zieldatei, links);
      fprintf(Zieldatei, " = ");
      IO_DruckeTermStrom(Zieldatei, rechts);
      fprintf(Zieldatei, " : hypothesis\n");
      Kommentar("goal from specification");
   }
}


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------------- REDUKTION -----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void uPCLr_Reduktionsbeginn(PCL_ObjektT objektArt, 
                          PCL_ObjektNummerT *nr,
                          PCL_SeitenT seite,
                          TermT andereSeite)
{
   AktuelleID = nr;
   AktuelleObjektArt = objektArt;
   AktuelleSeite = seite;
   AktuelleAndereSeite = andereSeite;
}

void uPCLr_Reduktion(TermT ergebnis,
                   UTermT betroffenerTeilterm,
                   PCL_ObjektNummerT *AngewandtesFakt, 
                   PCL_SeitenT angewandteSeite)
{
   ++PCL_Objektzaehler;
   if (IstWichtig(PCL_Objektzaehler)){
      StellenT i=0;
      StellenT TBS;

      if ((AktuelleObjektArt == Regel) && (AktuelleSeite == linkeSeite)){   /* Regel links reduziert -> ist jetzt Gleichung */
         AktuelleObjektArt = Gleichung;
      }
      fprintf(Zieldatei,"%lu : tes-", PCL_Objektzaehler);
      switch (AktuelleObjektArt){
      case Regel: fprintf(Zieldatei,"rule : ");
         break;
      case Ziel: fprintf(Zieldatei,"goal : ");
         break;
      default:  fprintf(Zieldatei,"eqn : ");
         break;
      }
      IO_DruckeTermStrom(Zieldatei, (AktuelleSeite == linkeSeite)?ergebnis:AktuelleAndereSeite);
      if (AktuelleObjektArt == Regel){
         fprintf(Zieldatei," -> ");
      }
      else{
         fprintf(Zieldatei," = ");
      }
      IO_DruckeTermStrom(Zieldatei, (AktuelleSeite == linkeSeite)?AktuelleAndereSeite:ergebnis);
      fprintf(Zieldatei," : tes-red( %lu, ",*AktuelleID);
      fprintf(Zieldatei,"%s", (AktuelleSeite == linkeSeite)?"L":"R");
      TBS = TiefeBerechneterStelle(ergebnis, betroffenerTeilterm);
      while (i<TBS){
         fprintf(Zieldatei,".%d", UnifStelle[i++]);
      }
      fprintf(Zieldatei,", %lu, %s)\n", *AngewandtesFakt, (angewandteSeite == linkeSeite)? "L":"R");
      Kommentar((AktuelleObjektArt == Regel)?
                ((AktuelleSeite == linkeSeite)?
                 "lhs of rule rewritten":"rhs of rule rewritten"):
                ((AktuelleObjektArt == Ziel)?
              ((AktuelleSeite == linkeSeite)?
               "lhs of goal rewritten":"rhs of goal rewritten"):
                 ((AktuelleSeite == linkeSeite)?
                  "lhs of equation or critical pair rewritten":"rhs of equation or critical pair rewritten")));
   }
   *AktuelleID = PCL_Objektzaehler;
}



/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Unifikation ---------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



void uPCLr_Unifikation(PCL_ObjektNummerT *nr,
                     TermT links, TermT rechts,
                     RegelOderGleichungsT Vater,
                     TermT betroffenerVaterTerm, /* ggf umbenannt!!! */
                     UTermT betroffenerVaterTeilterm,
                     RegelOderGleichungsT Mutter)
{
   *nr = ++PCL_Objektzaehler;
   if (IstWichtig(PCL_Objektzaehler)){
      StellenT i=0,
      maxTiefe = TiefeBerechneterStelle(betroffenerVaterTerm, betroffenerVaterTeilterm);
      fprintf(Zieldatei,"%lu : tes-eqn : ", *nr);
      IO_DruckeTermStrom(Zieldatei, links);
      fprintf(Zieldatei," = ");
      IO_DruckeTermStrom(Zieldatei, rechts);
      fprintf(Zieldatei," : cp(%lu, %s", *(TP_PclID(Vater)), 
              RE_IstRegel(Vater)? "L":
              TP_RichtungAusgezeichnet(Vater)?"L":"R");
      while (i<maxTiefe) fprintf(Zieldatei,".%d", UnifStelle[i++]);
      fprintf(Zieldatei,", %lu, %s)\n", *(TP_PclID(Mutter)),
              RE_IstRegel(Mutter)? "L":
              TP_RichtungAusgezeichnet(Mutter)?"L":"R");
      Kommentar("critical pair generated");
   }
}


void uPCLr_ReUnifikation(PCL_ObjektNummerT *nr,
                     TermT links, TermT rechts,
                     RegelOderGleichungsT Vater,
                     int NummerBetroffeneVaterStelle,
                     RegelOderGleichungsT Mutter)
{
   *nr = ++PCL_Objektzaehler;
   if (IstWichtig(PCL_Objektzaehler)){
      StellenT i=0,
      maxTiefe = TiefeBerechneterStelleAusNummer(TP_LinkeSeite(Vater), NummerBetroffeneVaterStelle);
      fprintf(Zieldatei,"%lu : tes-eqn : ", *nr);
      IO_DruckeTermStrom(Zieldatei, links);
      fprintf(Zieldatei," = ");
      IO_DruckeTermStrom(Zieldatei, rechts);
      fprintf(Zieldatei," : cp(%lu, %s", *(TP_PclID(Vater)), 
              RE_IstRegel(Vater)? "L":
              TP_RichtungAusgezeichnet(Vater)?"L":"R");
      while (i<maxTiefe) fprintf(Zieldatei,".%d", UnifStelle[i++]);
      fprintf(Zieldatei,", %lu, %s)\n", *(TP_PclID(Mutter)),
              RE_IstRegel(Mutter)? "L":
              TP_RichtungAusgezeichnet(Mutter)?"L":"R");
      Kommentar("critical pair rebuild");
   }
}


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Aufnahme von KPs als Regeln oder Gleichungen ------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



void uPCLr_KPZuRegel(TermT links, TermT rechts,
                   PCL_RichtungsRichtungT RichtungsRichtung,
                   PCL_ObjektNummerT *RegelID)
{
   ++PCL_Objektzaehler;
   if (IstWichtig(PCL_Objektzaehler)){
      fprintf(Zieldatei,"%lu : tes-rule : ", PCL_Objektzaehler);
      IO_DruckeTermStrom(Zieldatei, links);
      fprintf(Zieldatei," -> ");
      IO_DruckeTermStrom(Zieldatei, rechts);
      fprintf(Zieldatei," : orient(%lu, %s)\n", *RegelID, (RichtungsRichtung==VonLinksNachRechts)?"u":"x");
      Kommentar("critical pair oriented as rule");
   }
   *RegelID = PCL_Objektzaehler;
}



/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Ziele subsummieren/ zusammenfuehren ---------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



void uPCLr_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,
                         TermT links, TermT rechts,   /* linke und rechte Seite des subsummierten Zieles */
                         PCL_ObjektNummerT *subsummierendeGleichung)
{
   ++PCL_Objektzaehler;
   if (IstWichtig(PCL_Objektzaehler)){
      fprintf(Zieldatei,"%lu : tes-final : ", PCL_Objektzaehler);
      IO_DruckeTermStrom(Zieldatei, links);
      fprintf(Zieldatei," = ");
      IO_DruckeTermStrom(Zieldatei, rechts);
      fprintf(Zieldatei," : instance(%lu, %lu)\n", *subsummiertesZiel, *subsummierendeGleichung);
      Kommentar("goal subsumed");
   }
   *subsummiertesZiel = PCL_Objektzaehler;
}

void uPCLr_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,
                      TermT links, TermT rechts)   /* linke und rechte Seite des bewiesenen Zieles */
{
   ++PCL_Objektzaehler;
   if (IstWichtig(PCL_Objektzaehler)){
      fprintf(Zieldatei,"%lu : tes-final : ", PCL_Objektzaehler);
      IO_DruckeTermStrom(Zieldatei, links);
      fprintf(Zieldatei," = ");
      IO_DruckeTermStrom(Zieldatei, rechts);
      fprintf(Zieldatei," : %lu\n", *bewiesenesZiel);
      Kommentar("goal joined");
   }
   *bewiesenesZiel = PCL_Objektzaehler;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Abschluss der Vervollstaendigung: alle Fakten ausgeben --------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void uPCLr_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art)
{
   PCL_ObjektNummerT alteID = *nr;
   
   *nr = ++PCL_Objektzaehler;
   if (IstWichtig(PCL_Objektzaehler)){
      fprintf(Zieldatei,"%lu : tes-final : ", *nr);
      IO_DruckeTermStrom(Zieldatei, links);
      fprintf(Zieldatei, " %s ",(art==Regel)?"->":"=");
      IO_DruckeTermStrom(Zieldatei, rechts);
      fprintf(Zieldatei," : %lu\n", alteID);
      Kommentar("part of final system");
   }
}

/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- MIKRO-PCL Produzieren -----------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void DruckeMikro(PCL_ObjektNummerT ID1, PCL_ObjektNummerT ID2)
{
   fwrite(&ID1, sizeof(PCL_ObjektNummerT), 1, Zieldatei);
   fwrite(&ID2, sizeof(PCL_ObjektNummerT), 1, Zieldatei);
}

void uPCLp_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
{
   *nr = ++PCL_Objektzaehler;
   DruckeMikro(0,1);
}

void uPCLp_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
{
   *nr = ++PCL_Objektzaehler;
   DruckeMikro(0,2);
}

void uPCLp_Reduktionsbeginn(PCL_ObjektT objektArt, 
                          PCL_ObjektNummerT *nr,
                          PCL_SeitenT seite,
                          TermT andereSeite)
{
   AktuelleID = nr;
}

void uPCLp_Reduktion(TermT ergebnis,
                   UTermT betroffenerTeilterm,
                   PCL_ObjektNummerT *AngewandtesFakt, 
                   PCL_SeitenT angewandteSeite)
{
   DruckeMikro(*AktuelleID, *AngewandtesFakt);
   *AktuelleID = ++PCL_Objektzaehler;
}


void uPCLp_Unifikation(PCL_ObjektNummerT *nr,
                     TermT links, TermT rechts,
                     RegelOderGleichungsT Vater,
                     TermT betroffenerVaterTerm, /* ggf umbenannt!!! */
                     UTermT betroffenerVaterTeilterm,
                     RegelOderGleichungsT Mutter)
{
   *nr = ++PCL_Objektzaehler;
   DruckeMikro(*(TP_PclID(Vater)), *(TP_PclID(Mutter)));
}


void uPCLp_ReUnifikation(PCL_ObjektNummerT *nr,
                     TermT links, TermT rechts,
                     RegelOderGleichungsT Vater,
                     int NummerBetroffeneVaterStelle,
                     RegelOderGleichungsT Mutter)
{
   *nr = ++PCL_Objektzaehler;
   DruckeMikro(*(TP_PclID(Vater)), *(TP_PclID(Mutter)));
}

void uPCLp_KPZuRegel(TermT links, TermT rechts,
                   PCL_RichtungsRichtungT RichtungsRichtung,
                   PCL_ObjektNummerT *RegelID)
{
   DruckeMikro(*RegelID, 0);
   *RegelID = ++PCL_Objektzaehler;
}

void uPCLp_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,
                         TermT links, TermT rechts,   /* linke und rechte Seite des subsummierten Zieles */
                         PCL_ObjektNummerT *subsummierendeGleichung)
{
   DruckeMikro(0,0);
   DruckeMikro(*subsummiertesZiel, *subsummierendeGleichung);
   *subsummiertesZiel = ++PCL_Objektzaehler;
}

void uPCLp_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,
                      TermT links, TermT rechts)   /* linke und rechte Seite des bewiesenen Zieles */
{
   DruckeMikro(0,0);
   DruckeMikro(*bewiesenesZiel, 0);
   *bewiesenesZiel = ++PCL_Objektzaehler;
}

void uPCLp_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art)
{
   DruckeMikro(0,0);
   DruckeMikro(*nr, 0);
   *nr = ++PCL_Objektzaehler;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- INTEGRATED PCL-------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Gleichung/Ziel aus der Spezifikation --------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void iPCL_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
{
    iPCL_Memo[iPCL_lastId].p.p1 = TO_Termkopie(links);
    iPCL_Memo[iPCL_lastId].p.p2 = TO_Termkopie(rechts);
    iPCL_Memo[iPCL_lastId].p.misc = iPCL_ComposeMisc(iPCL_init,0,0,0);
    *(nr) = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_init);
}

void iPCL_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
{
    iPCL_Memo[iPCL_lastId].p.p1 = TO_Termkopie(links);
    iPCL_Memo[iPCL_lastId].p.p2 = TO_Termkopie(rechts);
    iPCL_Memo[iPCL_lastId].p.misc = iPCL_ComposeMisc(iPCL_hypo,0,0,0);
    *(nr) = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_hypo);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------------- REDUKTION -----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void iPCL_Reduktionsbeginn(PCL_ObjektT objektArt, 
			   PCL_ObjektNummerT *nr,
			   PCL_SeitenT seite,
			   TermT andereSeite)
{
    iPCL_AktuelleID = nr;
    iPCL_AktuelleSeite = seite;
}

void iPCL_Reduktion(TermT ergebnis,
		    UTermT betroffenerTeilterm,
		    PCL_ObjektNummerT *AngewandtesFakt, 
		    PCL_SeitenT angewandteSeite)
{
    iPCL_Memo[iPCL_lastId].id.id1 = *iPCL_AktuelleID;
    iPCL_Memo[iPCL_lastId].id.id2 = *AngewandtesFakt;
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_red,
						      TO_StelleInTerm(betroffenerTeilterm,ergebnis),
						      iPCL_AktuelleSeite,
						      angewandteSeite);
    *iPCL_AktuelleID = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_red);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Unifikation ---------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void iPCL_Unifikation(PCL_ObjektNummerT *nr,
		      TermT links, TermT rechts,
		      RegelOderGleichungsT Vater,
		      TermT betroffenerVaterTerm, /* ggf umbenannt!!! */
		      UTermT betroffenerVaterTeilterm,
		      RegelOderGleichungsT Mutter)
{
    iPCL_Memo[iPCL_lastId].id.id1 = *(TP_PclID(Vater));
    iPCL_Memo[iPCL_lastId].id.id2 = *(TP_PclID(Mutter));
    iPCL_Memo[iPCL_lastId].id.misc = 
      iPCL_ComposeMisc(iPCL_cp,
		       TO_StelleInTerm(betroffenerVaterTeilterm,betroffenerVaterTerm),
		       RE_IstRegel(Vater) ? linkeSeite : TP_RichtungAusgezeichnet(Vater) ? linkeSeite : rechteSeite,
		       RE_IstRegel(Mutter)? linkeSeite : TP_RichtungAusgezeichnet(Mutter)? linkeSeite : rechteSeite);
    *nr = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_cp);
}

void iPCL_ReUnifikation(PCL_ObjektNummerT *nr,
			TermT links, TermT rechts,
			RegelOderGleichungsT Vater,
			int NummerBetroffeneVaterStelle,
			RegelOderGleichungsT Mutter)
{
    iPCL_Memo[iPCL_lastId].id.id1 = *(TP_PclID(Vater));
    iPCL_Memo[iPCL_lastId].id.id2 = *(TP_PclID(Mutter));
    iPCL_Memo[iPCL_lastId].id.misc = 
      iPCL_ComposeMisc(iPCL_cp,
		       NummerBetroffeneVaterStelle,
		       RE_IstRegel(Vater) ? linkeSeite : TP_RichtungAusgezeichnet(Vater) ? linkeSeite : rechteSeite,
		       RE_IstRegel(Mutter)? linkeSeite : TP_RichtungAusgezeichnet(Mutter)? linkeSeite : rechteSeite);
    *nr = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_cp);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Aufnahme von KPs als Regeln oder Gleichungen ------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void iPCL_KPZuRegel(TermT links, TermT rechts,
		    PCL_RichtungsRichtungT RichtungsRichtung,
		    PCL_ObjektNummerT *RegelID)
{
    iPCL_Memo[iPCL_lastId].id.id1 = *RegelID;
    iPCL_Memo[iPCL_lastId].id.id2 = 0;
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_orient,0,0,RichtungsRichtung);
    *RegelID = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_orient);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Ziele subsummieren/ zusammenfuehren ---------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



void iPCL_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,
			  TermT links, TermT rechts,   /* linke und rechte Seite des subsummierten Zieles */
			  PCL_ObjektNummerT *subsummierendeGleichung)
{
    iPCL_Memo[iPCL_lastId].id.id1 = *subsummiertesZiel;
    iPCL_Memo[iPCL_lastId].id.id2 = *subsummierendeGleichung;
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_inst,0,0,0);
    iPCL_SetUsed(iPCL_Memo[iPCL_lastId]);
    *subsummiertesZiel = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_inst);
}

void iPCL_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,
		       TermT links, TermT rechts)   /* linke und rechte Seite des bewiesenen Zieles */
{
    iPCL_Memo[iPCL_lastId].id.id1 = *bewiesenesZiel;
    iPCL_Memo[iPCL_lastId].id.id2 = 0;
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_proved,0,0,0);
    iPCL_SetUsed(iPCL_Memo[iPCL_lastId]);
    *bewiesenesZiel = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_proved);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Abschluss der Vervollstaendigung: alle Fakten ausgeben --------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void iPCL_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art)
{
    iPCL_Memo[iPCL_lastId].id.id1 = *nr;
    iPCL_Memo[iPCL_lastId].id.id2 = 0;
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_final,0,0,0);
    iPCL_SetUsed(iPCL_Memo[iPCL_lastId]);
    *nr = iPCL_lastId;
    iPCL_lastId++;
    iPCL_RecordPerformed(iPCL_final);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CORRESPONDANCE produzieren ------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void FULL_NeueRegel(PCL_ObjektNummerT *pclIDPtr, long int RegelNummer)
{
   fprintf(Corrdatei, "%10ld R %ld\n", *pclIDPtr, RegelNummer);
}

void FULL_NeueGleichung(PCL_ObjektNummerT *pclIDPtr, long int GleichungsNummer)
{
   fprintf(Corrdatei, "%10ld E %ld\n", *pclIDPtr, GleichungsNummer);
}

void FULL_RegelRechtsReduziert(PCL_ObjektNummerT *pclIDPtr, long int RegelNummer)
{
   fprintf(Corrdatei, "%10ld R %ldrr\n", *pclIDPtr, RegelNummer);
}


/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- DUMMIES -------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void Dummy_GleichungAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
{;}

void Dummy_ZielAusDerSpezifikation(TermT links, TermT rechts, PCL_ObjektNummerT *nr)
{;}

void Dummy_Reduktionsbeginn(PCL_ObjektT objektArt, 
                          PCL_ObjektNummerT *nr,
                          PCL_SeitenT seite,
                          TermT andereSeite)
{;}

void Dummy_Reduktion(TermT ergebnis,
                   UTermT betroffenerTeilterm,
                   PCL_ObjektNummerT *AngewandtesFakt, 
                   PCL_SeitenT angewandteSeite)
{;}

void Dummy_Unifikation(PCL_ObjektNummerT *nr,
                     TermT links, TermT rechts,
                     RegelOderGleichungsT Vater,
                     TermT betroffenerVaterTerm, /* ggf umbenannt!!! */
                     UTermT betroffenerVaterTeilterm,
                     RegelOderGleichungsT Mutter)
{;}


void Dummy_ReUnifikation(PCL_ObjektNummerT *nr,
                     TermT links, TermT rechts,
                     RegelOderGleichungsT Vater,
                     int NummerBetroffeneVaterStelle,
                     RegelOderGleichungsT Mutter)
{;}

void Dummy_KPZuRegel(TermT links, TermT rechts,
                   PCL_RichtungsRichtungT RichtungsRichtung,
                   PCL_ObjektNummerT *RegelID)
{;}


void Dummy_ZielSubsummiert(PCL_ObjektNummerT *subsummiertesZiel,
                         TermT links, TermT rechts,   /* linke und rechte Seite des subsummierten Zieles */
                         PCL_ObjektNummerT *subsummierendeGleichung)
{;}

void Dummy_ZielBewiesen(PCL_ObjektNummerT *bewiesenesZiel,
                      TermT links, TermT rechts)   /* linke und rechte Seite des bewiesenen Zieles */
{;}

void Dummy_VVAbschlussFaktDrucken(PCL_ObjektNummerT *nr, TermT links, TermT rechts, PCL_ObjektT art)
{;}

void Dummy_StartChain(void)
{;}

void Dummy_StartFeyChain(void)
{;}

void Dummy_SaveChain(void)
{;}

void Dummy_DeleteChain(void)
{;}

void Dummy_NeueRegel(PCL_ObjektNummerT *pclIDPtr, long int RegelNummer)
{;}

void Dummy_NeueGleichung(PCL_ObjektNummerT *pclIDPtr, long int GleichungsNummer)
{;}

void Dummy_RegelRechtsReduziert(PCL_ObjektNummerT *pclIDPtr, long int RegelNummer)
{;}



/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Initialisierung/Abschluss des Protokolls ----------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define Ausgaben()                                                                                      \
{                                                                                                       \
  BOOLEAN dummyMode = FALSE;                                                                            \
   fprintf(IO_Strom, "\n* PCL:   ");                                                                    \
   /***************************************** nPCL *****************************/                       \
   if ((ZieldateiName = IS_ValueOfOption(DEF, USR, "PCL"))){                                            \
      fprintf(IO_Strom, "write PCL -> %s", ZieldateiName);                                              \
   }                                                                                                    \
   /***************************************** uPCLp *****************************/                      \
   else if ((ZieldateiName = IS_ValueOfOption(DEF, USR, "uPCL"))){                                      \
      fprintf(IO_Strom, "write uPCL -> %s", ZieldateiName);                                             \
   }                                                                                                    \
   /***************************************** uPCLr *****************************/                      \
   else if ((QuelldateiName = IS_ValueOfOption(DEF, USR, "uPCLQuelle")) &&                              \
         (ZieldateiName = IS_ValueOfOption(DEF, USR, "uPCLZiel"))){                                     \
      fprintf(IO_Strom, "reproduce PCL (-> %s) from uPCL in %s ", ZieldateiName, QuelldateiName);       \
   }                                                                                                    \
   /***************************************** iPCL *****************************/                       \
   else if (IS_FindOption(DEF, USR, 1, "iPCL") == 0){                                                   \
      fprintf(IO_Strom, "integrated PCL ");                                                             \
   }                                                                                                    \
   /***************************************** Dummies *****************************/                    \
   else{                                                                                                \
      fprintf(IO_Strom, " off");                                                                        \
      dummyMode = TRUE;                                                                                 \
   }                                                                                                    \
                                                                                                        \
   /***************************************** Correspondance WM--PCL **************/                    \
   if (!dummyMode){                                                                                     \
      if (IS_ValueOfOption(DEF, USR, "corr")){                                                          \
         fprintf(IO_Strom, ", corrdatei = %s", IS_ValueOfOption(DEF, USR, "corr"));                     \
      }                                                                                                 \
   }                                                                                                    \
   fprintf(IO_Strom, "\n");                                                                             \
}

void PCL_InitAposteriori(InfoStringT DEF, InfoStringT USR)
/* PCL-Protokoll-Datei oeffnen, Funktionspointer besetzen.
   Parametrisierung:
   PCL=Dateiname oder uPCL=Dateiname oder uPCLQuelle=Datei1:uPCLZiel=Datei1
*/
{

#if COMP_PCL == 2
  Ausgaben();
  if ((ZieldateiName = IS_ValueOfOption(DEF, USR, "PCL")) ||
      (ZieldateiName = IS_ValueOfOption(DEF, USR, "uPCL")) ||
      (QuelldateiName = IS_ValueOfOption(DEF, USR, "uPCLQuelle")) ||
      (ZieldateiName = IS_ValueOfOption(DEF, USR, "uPCLZiel"))){
    our_error("compiled with COMP_PCL == 2: only iPCL available!\n");
  }

  else if (IS_FindOption(DEF, USR, 1, "iPCL") == 0){                              
     PCL_Mode        = iPCLMode;
     Chain_Mode      = iPCLChainMode;
     ipcl_Domaene    = SM_NeueStrahldomaene(5,5,10,"iPCL_Memo");
     iPCL_Memo       = ipcl_Domaene->Start;
     iPCL_lastId     = 1; /* Important: id 0 is bogus! */
     iPCL_doIt       = 1;
     iPCL_doItReally = 1;
  }
  else {
     PCL_Mode        = DummyMode;
     iPCL_doIt       = 0;
     iPCL_doItReally = 0;

  }

  if (PCL_Mode != DummyMode){
    if((CorrdateiName = IS_ValueOfOption(DEF, USR, "corr"))){
      our_error("compiled with COMP_PCL == 2: no correspondance available!\n");
    }
  }
  
  Corr_Mode = DummyCorrMode; /* ueberfluessig, soll compiler ruhigstellen, correspondance ist nicht vorges. fuer COMP_PCL==2 */

#else

  Ausgaben();
   /***************************************** nPCL *****************************/
   if ((ZieldateiName = IS_ValueOfOption(DEF, USR, "PCL"))){
      PCL_Objektzaehler = 0;
      Zieldatei = fopen(ZieldateiName, "w");
      if (!Zieldatei){
         fprintf(IO_Strom, " (unable to open file, PCL turned off)");
      }
      UnifStelle = (StellenT *) our_alloc(((UnifStelleDimension = NF_PfadArrayGroesse)+1)*sizeof(StellenT));
      PCL_Mode = nPCLMode;
      Chain_Mode = FULLChainMode;
   }
   /***************************************** uPCLp *****************************/
   else if ((ZieldateiName = IS_ValueOfOption(DEF, USR, "uPCL"))){
      PCL_Objektzaehler = 0;
      Zieldatei = fopen(ZieldateiName, "w");
      if (!Zieldatei){
         fprintf(IO_Strom, " (unable to open file, PCL turned off)");
      }
      UnifStelle = (StellenT *) our_alloc(((UnifStelleDimension = NF_PfadArrayGroesse)+1)*sizeof(StellenT));
      PCL_Mode = uPCLpMode;
      Chain_Mode = FULLChainMode;
   }
   /***************************************** uPCLr *****************************/
   else if ((QuelldateiName = IS_ValueOfOption(DEF, USR, "uPCLQuelle")) &&
         (ZieldateiName = IS_ValueOfOption(DEF, USR, "uPCLZiel"))){
      PCL_Objektzaehler = 0;
      Zieldatei = fopen(ZieldateiName, "w");
      Quelldatei = fopen(QuelldateiName, "r");
      if (! (Zieldatei && Quelldatei)){
         PCL_Mode = DummyMode;
         Chain_Mode = DummyChainMode;
         fprintf(IO_Strom, " (unable to open file(s), PCL turned off)");
      }
      else{
         int ern;
         UnifStelle = (StellenT *) our_alloc(((UnifStelleDimension = NF_PfadArrayGroesse)+1)*sizeof(StellenT));
         PCL_Mode = uPCLrMode;
         Chain_Mode = uPCLrChainMode;

         /* erste wichtige ID besetzen */
         fread(&NaechsteWichtigeID, sizeof(unsigned long int), 1, Quelldatei);
         if ((ern = ferror(Quelldatei))){
            perror(sys_errlist[ern]);
            fprintf(IO_Strom, " (error in uPCL-source-file, PCL turned off)");
            PCL_Mode = DummyMode;
            Chain_Mode = DummyChainMode;
         }
      }
   }
   /***************************************** iPCL *****************************/  
   else if (IS_FindOption(DEF, USR, 1, "iPCL") == 0){                              
     PCL_Mode        = iPCLMode;
     Chain_Mode      = iPCLChainMode;
     ipcl_Domaene    = SM_NeueStrahldomaene(5,5,10,"iPCL_Memo");
     iPCL_Memo       = ipcl_Domaene->Start;
     iPCL_lastId     = 1; /* Important: id 0 is bogus! */
   }
   /***************************************** Dummies *****************************/
   else{
      fprintf(IO_Strom, " off");
      PCL_Mode = DummyMode;
      Chain_Mode = DummyChainMode;
   }

   FunktionspointerBesetzen(PCL_Mode);
   ChainingBesetzen(Chain_Mode);


   /***************************************** Correspondance WM--PCL **************/
   Corr_Mode = DummyCorrMode;
   
   if (PCL_Mode != DummyMode){
      CorrdateiName = IS_ValueOfOption(DEF, USR, "corr");
      if (CorrdateiName){
         if (!(Corrdatei = fopen(CorrdateiName, "w"))){
            fprintf(IO_Strom, " (unable to open)");
         }
         else
            Corr_Mode = FullCorrMode;
      }
   }

   CorrBesetzen(Corr_Mode);
   ChainStarted = FeyChainStarted = FALSE;
#endif
}

void PCL_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  if (what == ParamEvaluate){
    Ausgaben();

#if COMP_PCL == 2
    fprintf (IO_Strom, "WARNING: compiled with COMP_PCL == 2: only iPCL available!\n");
#endif

  }
  else{
    fprintf(IO_Strom, "PCL_ParamInfo not yet implemented\n\n");
  }
}

void iPCL_Beenden(void){

  long i,j;
  int  pos, side, pside ;
  int  restp; 
  TermT lhs, rhs; 
  TermT l1,r1,l2,r2;
  StellenT maxTiefe;
  UTermT actTermBefore;

  /* Phase 1: downwards scanning & marking */

  for (i = iPCL_lastId - 1; i > 0; i--){
    if (iPCL_IsUsed(iPCL_Memo[i]) /* is used is set by inst, proved & final */
        && iPCL_GetEntryType(iPCL_Memo[i]) > iPCL_hypo){ /* not if pointer in id1 and id2 */
      iPCL_SetUsed(iPCL_Memo[iPCL_Memo[i].id.id1]);
      iPCL_SetUsed(iPCL_Memo[iPCL_Memo[i].id.id2]);
    }
  }
  
  /* Pase 2: upwards reconstructing and renumbering */

  for (i = 1; i < iPCL_lastId; i++){
    iPCL_RecordRecorded(iPCL_GetEntryType(iPCL_Memo[i]));
    if (iPCL_IsUsed(iPCL_Memo[i])){
      iPCL_RecordUsed(iPCL_GetEntryType(iPCL_Memo[i]));
      switch (iPCL_GetEntryType(iPCL_Memo[i])){
      case iPCL_init: 
        IO_DruckeFlexStrom(IO_Strom, "%u : tes-eqn : %t = %t : initial\n", i, iPCL_Memo[i].p.p1, iPCL_Memo[i].p.p2);
        restp = iPCL_eqn;
        break;
      case iPCL_hypo: 
        IO_DruckeFlexStrom(IO_Strom, "%u : tes-goal : %t = %t : hypothesis\n", i, iPCL_Memo[i].p.p1, iPCL_Memo[i].p.p2);
        restp = iPCL_goal;
        break;
      case iPCL_red: 
        lhs   = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].id.id1].p.p1);
        rhs   = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].id.id1].p.p2);
        pos   = iPCL_GetPos(iPCL_Memo[i]);
        pside = iPCL_GetPSide(iPCL_Memo[i]);
        side  = iPCL_GetSide(iPCL_Memo[i]);
	l1    = (side == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].id.id2].p.p1 : iPCL_Memo[iPCL_Memo[i].id.id2].p.p2;
	r1    = (side == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].id.id2].p.p2 : iPCL_Memo[iPCL_Memo[i].id.id2].p.p1;
        BO_TermpaarNormierenAlsKP(l1,r1);
        
	if (pos > 0){
	  actTermBefore = TO_TermAnStelle((pside == linkeSeite) ? lhs : rhs, pos - 1);
	  if (MO_TermpaarPasst(l1, r1, TO_Schwanz(actTermBefore)))
            TO_NachfBesetzen(actTermBefore, MO_SigmaRInLZ());
	  else 
	    our_error("#1...");
	}
	else
	  if (MO_TermpaarPasst(l1, r1, (pside == linkeSeite) ? lhs : rhs))
	    MO_SigmaRInEZ();
	  else
	    our_error("#2...");

	restp = iPCL_GetResultType(iPCL_Memo[iPCL_Memo[i].id.id1]);
        if ((restp == iPCL_rule) && (pside == linkeSeite)){
	  restp = iPCL_eqn;
	}

	BO_TermpaarNormierenAlsKP(lhs, rhs);

        IO_DruckeFlexStrom(IO_Strom, "%u : tes-%s : %t %s %t : tes-red(%u,%s", i, restpstr[restp], 
                      lhs, (restp == iPCL_rule) ? "->" : "=", rhs, iPCL_Memo[i].id.id1, (pside == linkeSeite) ? "L" : "R");

        maxTiefe = TiefeBerechneterStelleAusNummer((pside == linkeSeite) ? lhs : rhs, pos);
        for (j = 0; j < maxTiefe; j++){
          fprintf(IO_Strom,".%d", UnifStelle[j]);
        }

        IO_DruckeFlexStrom(IO_Strom, ",%u,%s)\n", iPCL_Memo[i].id.id2, (side == linkeSeite) ? "L" : "R");
        
        iPCL_Memo[i].p.p1 = lhs;
        iPCL_Memo[i].p.p2 = rhs;
        break;
      case iPCL_cp: 
	pos   = iPCL_GetPos(iPCL_Memo[i]);
        pside = iPCL_GetPSide(iPCL_Memo[i]);
        side  = iPCL_GetSide(iPCL_Memo[i]);

	l1 = (pside == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].id.id1].p.p1 : iPCL_Memo[iPCL_Memo[i].id.id1].p.p2;
	r1 = (pside == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].id.id1].p.p2 : iPCL_Memo[iPCL_Memo[i].id.id1].p.p1;
	BO_TermpaarNormierenAlsKP(l1, r1);

	l2 = (side == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].id.id2].p.p1 : iPCL_Memo[iPCL_Memo[i].id.id2].p.p2;
	r2 = (side == linkeSeite) ? iPCL_Memo[iPCL_Memo[i].id.id2].p.p2 : iPCL_Memo[iPCL_Memo[i].id.id2].p.p1;
	  
	U1_KPAusTermenRekonstruieren(l1, r1, l2, r2, pos, &lhs, &rhs); 
	BO_TermpaarNormierenAlsKP(lhs, rhs);

        IO_DruckeFlexStrom(IO_Strom, "%u : tes-eqn : %t = %t : cp(%u,%s", 
			   i, lhs, rhs, iPCL_Memo[i].id.id1, (pside == linkeSeite) ? "L" : "R");

        maxTiefe = TiefeBerechneterStelleAusNummer(l1, pos);
        for (j = 0; j < maxTiefe; j++){
          fprintf(IO_Strom,".%d", UnifStelle[j]);
        }

        IO_DruckeFlexStrom(IO_Strom, ",%u,%s)\n", iPCL_Memo[i].id.id2, (side == linkeSeite) ? "L" : "R");
        
        restp  = iPCL_eqn;
        iPCL_Memo[i].p.p1 = lhs;
        iPCL_Memo[i].p.p2 = rhs;

        break;
      case iPCL_orient: 
        side = iPCL_GetSide(iPCL_Memo[i]);
	lhs = TO_Termkopie((side == UeberKreuz) ? iPCL_Memo[iPCL_Memo[i].id.id1].p.p2 : iPCL_Memo[iPCL_Memo[i].id.id1].p.p1);
	rhs = TO_Termkopie((side == UeberKreuz) ? iPCL_Memo[iPCL_Memo[i].id.id1].p.p1 : iPCL_Memo[iPCL_Memo[i].id.id1].p.p2);

        BO_TermpaarNormierenAlsKP(lhs, rhs);

        IO_DruckeFlexStrom(IO_Strom, "%u : tes-rule : %t -> %t : orient(%u,%s)\n", 
		      i, lhs, rhs, iPCL_Memo[i].id.id1, (side == UeberKreuz) ? "x" : "u");

        restp  = iPCL_rule;
        iPCL_Memo[i].p.p1 = lhs;
        iPCL_Memo[i].p.p2 = rhs;
        break;
      case iPCL_inst:
        lhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].id.id1].p.p1);
        rhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].id.id1].p.p2);
        BO_TermpaarNormierenAlsKP(lhs, rhs);
        IO_DruckeFlexStrom(IO_Strom, "%u : tes-final : %t = %t : instance(%u,%u)\n", 
                      i, lhs, rhs, iPCL_Memo[i].id.id1, iPCL_Memo[i].id.id2);

        restp  = iPCL_finalr;
        iPCL_Memo[i].p.p1 = lhs;
        iPCL_Memo[i].p.p2 = rhs;
        break;
      case iPCL_proved: 
        lhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].id.id1].p.p1);
        rhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].id.id1].p.p2);
        BO_TermpaarNormierenAlsKP(lhs, rhs);
        IO_DruckeFlexStrom(IO_Strom, "%u : tes-final : %t = %t : %u\n", i, lhs, rhs, iPCL_Memo[i].id.id1);

        restp  = iPCL_finalr;
        iPCL_Memo[i].p.p1 = lhs;
        iPCL_Memo[i].p.p2 = rhs;
        break;
      case iPCL_final:
        lhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].id.id1].p.p1);
        rhs = TO_Termkopie(iPCL_Memo[iPCL_Memo[i].id.id1].p.p2);
        BO_TermpaarNormierenAlsKP(lhs, rhs);
        IO_DruckeFlexStrom(IO_Strom, "%u : tes-final : %t %s %t : %u\n", i, lhs, 
		      (iPCL_GetResultType(iPCL_Memo[iPCL_Memo[i].id.id1]) == iPCL_rule) ? "->" : "=", rhs, iPCL_Memo[i].id.id1);

        restp  = iPCL_finalr;
        iPCL_Memo[i].p.p1 = lhs;
        iPCL_Memo[i].p.p2 = rhs;
        break;
      default:
	restp = iPCL_finalr; /* shut off warning */
        our_error("#3...");
      }
      iPCL_SetResultType(iPCL_Memo[i], restp);
    }
  }

  /* Phase 3: dispose everything */
  for (i = iPCL_lastId; i > 0; i--){
    if (iPCL_IsUsed(iPCL_Memo[i])){
      /* dispose id1, id2 */
    }
  }
  /* dispose domaene */
  /* Some output: */
return;
  fprintf(IO_Strom, "\n\n");
  fprintf(IO_Strom, "     PCL-inferences | performed |  recorded |      used\n");
  fprintf(IO_Strom, "     ---------------+-----------+-----------+-----------\n");
  fprintf(IO_Strom, "        initial     | %9lu | %9lu | %9lu\n", iPCL_Performed[iPCL_init], iPCL_Recorded[iPCL_init], iPCL_Used[iPCL_init]);
  fprintf(IO_Strom, "        hypothesis  | %9lu | %9lu | %9lu\n", iPCL_Performed[iPCL_hypo], iPCL_Recorded[iPCL_hypo], iPCL_Used[iPCL_hypo]);
  fprintf(IO_Strom, "        reduction   | %9lu | %9lu | %9lu\n", iPCL_Performed[iPCL_red], iPCL_Recorded[iPCL_red], iPCL_Used[iPCL_red]);
  fprintf(IO_Strom, "        crit. pair  | %9lu | %9lu | %9lu\n", iPCL_Performed[iPCL_cp], iPCL_Recorded[iPCL_cp], iPCL_Used[iPCL_cp]);
  fprintf(IO_Strom, "        orient      | %9lu | %9lu | %9lu\n", iPCL_Performed[iPCL_orient], iPCL_Recorded[iPCL_orient], iPCL_Used[iPCL_orient]);
  fprintf(IO_Strom, "        subsumed    | %9lu | %9lu | %9lu\n", iPCL_Performed[iPCL_inst], iPCL_Recorded[iPCL_inst], iPCL_Used[iPCL_inst]);
  fprintf(IO_Strom, "        proved      | %9lu | %9lu | %9lu\n", iPCL_Performed[iPCL_proved], iPCL_Recorded[iPCL_proved], iPCL_Used[iPCL_proved]);
  fprintf(IO_Strom, "        final       | %9lu | %9lu | %9lu\n", iPCL_Performed[iPCL_final], iPCL_Recorded[iPCL_final], iPCL_Used[iPCL_final]);
  fprintf(IO_Strom, "     ---------------+-----------+-----------+-----------\n");
  fprintf(IO_Strom, "        summary     | %9lu | %9lu | %9lu\n",
iPCL_Performed[iPCL_init] + iPCL_Performed[iPCL_hypo] + iPCL_Performed[iPCL_red] + iPCL_Performed[iPCL_cp] + iPCL_Performed[iPCL_orient] + iPCL_Performed[iPCL_inst] + iPCL_Performed[iPCL_proved] + iPCL_Performed[iPCL_final],
iPCL_Recorded[iPCL_init] + iPCL_Recorded[iPCL_hypo] + iPCL_Recorded[iPCL_red] + iPCL_Recorded[iPCL_cp] + iPCL_Recorded[iPCL_orient] + iPCL_Recorded[iPCL_inst] + iPCL_Recorded[iPCL_proved] + iPCL_Recorded[iPCL_final], 
iPCL_Used[iPCL_init] + iPCL_Used[iPCL_hypo] + iPCL_Used[iPCL_red] + iPCL_Used[iPCL_cp] + iPCL_Used[iPCL_orient] + iPCL_Used[iPCL_inst] + iPCL_Used[iPCL_proved] + iPCL_Used[iPCL_final]);
  fprintf(IO_Strom, "\n\n");
}


void PCL_Beenden(void)
/* PCL-Protokoll-Datei wieder schliessen.*/
{
  if (PCL_Mode == iPCLMode){
    iPCL_Beenden();
  }
  else {
   if (Zieldatei) {
      long Pos;
      Pos = ftell(Zieldatei);
      /*
        printf("Schreibposition beim Schliessen ist: %ld\n", Pos);
        */
      fclose(Zieldatei);
      truncate(ZieldateiName, Pos);
   }
   if (Quelldatei)
      fclose(Quelldatei);
  }
   if (Corrdatei)
      fclose(Corrdatei);
}


#elif COMP_PCL == 0

void PCL_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  fprintf(IO_Strom, "\n* PCL: not available. Try pclwm\n\n");
}
void PCL_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
  PCL_ParamInfo(ParamEvaluate, "", "");
}
#endif

