/*----------------------------------------------------------------------------------------------------
  HEAP.
  
  KP-Mengen-Verwaltung mit einem normalen Heap (ARRAY).
  Keine Sonderfunktionen wie Waisenmord etc., dafuer aber kein zusaetzlicher Verwaltungsaufwand und
  Platzbedarf ausser dem Array.

  11.03.94. Angelegt. Arnim.
  14.03.95. Heap und KPHeap zusammengelegt.

Vorgehensweise:

 Sortierreihenfolge : kleinstes Element am Anfang
 Entnehmen          : Falls Heap-Flag ok, erstes Element entnehmen, Flag auf nicht-ok setzen
                        sonst Heap reorganisieren, erstes Element entnehmen, Flag bleibt nicht-ok
 Einfuegen          : Falls Flag ok, am Ende anhaengen, reorganisieren
                        sonst am Anfang einsetzen, einsinken lassen.

 Zu jedem Heap werden zwei Verwaltungsinformationen gespeichert, die seltener eine Reorganisation
 noetig machen. Zum einen gibt es einen Eintrag 'letztesBesetztesElement', zum anderen ein Flag
 mit Namen 'HeapIstHeap'. Dieses Flag gibt an, ob der Heap in reorganisierter Form vorliegt, oder
 ob nach der letzten Entnahme eines Elementes die oberste Stelle verwaist geblieben ist.
 Dies hat den Vorteil, dass bei einer auf die Entnahme folgenden Einfuege-Operation eine Reorgani-
 sation gespart werden kann (naemlich die Reorganisation nach der Entnahme). Je nach Wert dieses
 Flags kommt dem Verweis 'letztesBesetztesElement' aber eine unterschiedliche Bedeutung zu:

 Ist das Flag TRUE, so wird tatsaechlich auf das letzte besetzte Element verwiesen.
 Ist das Flag aber FALSE, so zeigt der Verweis zwar immer noch auf das letzte besetzte Element,
     an der ersten Stelle im Heap klafft aber eine Luecke. Deshalb wird in dieser Situation nach
     dem naechsten Einfuegen der Verweis nicht erhoeht!


 Ein Sonderfall tritt in diesem Zusammenhang auf:
   Ist der Heap leer, so ist das Flag FALSE, und der Verweis zeigt auf das nullte Element.

+--------------------------------------------+
|             IstHeap       LetztesBesetztes |
+--------------------------------------------+
|leer             F                0         |
|1 Elt auf 0      T                0         |
|1 Elt auf 1      F                1         |
|n ELt ab 0       T                n         |
|n ELt ab 1       F               n+1        |
+--------------------------------------------+

 Daraus ergibt sich folgender Grob-Algorithmus:

 Einfuegen:
 - Flag = TRUE: + Verweis++
                + Heap[Verweis] = neues Element;
                + Position (Verweis div 2) rekursiv einsinken lassen

        = FALSE:+ Heap[0] = neues Element
                + Position 0 einsinken lassen
                + Flag = TRUE

 Entnehmen:
 - Flag = TRUE: + return Heap[0]
                + Flag = FALSE

        = FALSE:+ if Verweis==0 => return NULL (Heap ist leer)
                + Heap[0] = Heap[Verweis]
                + Verweis--
                + Position 0 Einsinken lassen
                + return Heap[0]
*/

/*-----------------------------------------------------------------------------*/
/*------------------------------ Includes -------------------------------------*/
/*-----------------------------------------------------------------------------*/

#include "antiwarnings.h"
#include "general.h"
#include "SpeicherVerwaltung.h"
#include "FehlerBehandlung.h"
#include "Parameter.h"
#include "ZeitMessung.h"
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>
#include "KPHeap.h"
#include "Ausgaben.h"

#include <string.h>

#include "StrahlMemorium.h"



#include "SUEBasis.h"
#include "Classification.h"
#include "TermpairRepresentation.h"

#include "PQStatistik.h"

#define _NOT_CAREFULLY

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe KPHeap.h) ----------------------------*/
/* ------------------------------------------------------------------------*/

void (* KPH_TermVerfolgung) (void (* OperationAufTermen) (TermT, TermT));
void (* KPH_WALKTHROUGH) (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms);
BOOLEAN (* KPH_SELECT) (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **);
BOOLEAN (* KPH_SELECTINITIAL) (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **);
void (* KPH_INSERT)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);
void (* KPH_INSERTImmortal)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);

BOOLEAN (* KPH_POSTPONE) (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);

/*-----------------------------------------------------------------------------*/
/*------------------------------ Typen ----------------------------------------*/
/*-----------------------------------------------------------------------------*/

typedef struct InfoTS{          /* Info-Block */
   long int ElternNummer;
   unsigned long AnzNochNichtBehandelterKPs;
   RegelOderGleichungsT re;
   BOOLEAN LebtNoch;
   unsigned long AnzInitialerKPs;
} InfoT;


/* ---------- zum Allokieren ---------------------- */

typedef struct InfoTS_MS{	/* Info-Block mit Statistik */
   long int ElternNummer;
   unsigned long AnzNochNichtBehandelterKPs;
   RegelOderGleichungsT re;
   BOOLEAN LebtNoch;
   unsigned long AnzInitialerKPs;
} InfoT_MitStatistik;

typedef struct InfoTS_OS{	/* Info-Block ohne Statistik */
   long int ElternNummer;
   unsigned long AnzNochNichtBehandelterKPs;
   RegelOderGleichungsT re;
   BOOLEAN LebtNoch;
} InfoT_OhneStatistik;

typedef struct{			/* Eintrag im Heap: 1 long als Gewicht */
   InfoT *Vater, *Mutter;
   TermpairT tp;
   long int w1;
} HeapZellenT_0;

static HeapZellenT_0   *Elemente_0;    /* Das Array mit den eigentlichen Eintraegen.*/


typedef struct{			/* Eintrag im Heap: 2 long als Gewicht */
   InfoT *Vater, *Mutter;
   TermpairT tp;
   long int w1, w2;
} HeapZellenT_1;

static HeapZellenT_1   *Elemente_1;    /* Das Array mit den eigentlichen Eintraegen.*/


typedef struct{			/* Eintrag im Heap: 2 long + 1 pointer als Gewicht */
   InfoT *Vater, *Mutter;
   TermpairT tp;
   long int w1, w2;
   void *addref;
} HeapZellenT_2;

static HeapZellenT_2   *Elemente_2;    /* Das Array mit den eigentlichen Eintraegen.*/




static BOOLEAN IstHeap;                 /* Zeigt an, ob die Struktur z.Zt. als Heap organisiert ist
                                           oder beim naechsten Zugriff reorganisiert werden muss:
                                           T : besetzt an Positionen 0..LetztesBesetztes.
                                           F : besetzt an Positionen 1..(LetztesBesetztes+1).
                                           (LetztesBesetztes == 0 bedeutet Heap ist leer)
                                         */

static unsigned long  LetztesBesetztes;

static long VerwendeteSaat;   

static int WeightEntryStrukturTyp; 
/* 0 -> 1 long 
   1 -> 2 long
   2 -> 2 long + 1 pointer
*/

static unsigned long DomainGran;

static DomaenenT Domaene;

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Testing ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
#define KPH_TEST 0
/* Alternativen: 0 fuer aus, 1 fuer an */
#if KPH_TEST
static unsigned long int testNr = 0;
static BOOLEAN ok = TRUE;

#define top(StrukturTyp) Elemente_ ## StrukturTyp [index]
#define HeapFehler(str, nr1, nr2) (ok = FALSE, printf("\n%ld: %s (%ld, %ld)\n", index, str, nr1, nr2))

#define ElterOk(el)                                                                                                                     \
{                                                                                                                                       \
   if (!el)                                                                                                                             \
      HeapFehler("Elternteil nicht besetzt",0L,0L);                                                                                     \
                                                                                                                                        \
   if ((el->LebtNoch) && (el->ElternNummer != 0) && !(el->re))                                                                          \
      HeapFehler("lebendem Elternteil anhaengende re nicht besetzt",0L,0L);                                                             \
                                                                                                                                        \
   if ((el->LebtNoch) && (el->re)){                                                                                                     \
      if (el->re->Nummer != el->ElternNummer)                                                                                           \
         HeapFehler("Elternteil mit falscher Nummer",el->re->Nummer , el->ElternNummer);                                                \
                                                                                                                                        \
      if (((InfoT *) TP_SUEInfo(el->re)) != el)                                                                                         \
         HeapFehler("re zeigt auf anderen Info-Block",0L,0L);                                                                           \
                                                                                                                                        \
      if (RE_IstRegel(el->re)){                                                                                                         \
         if (el->ElternNummer > RE_AnzRegelnGesamt())                                                                                   \
            HeapFehler("Regel mit falscher Nummer", el->ElternNummer, RE_AnzRegelnGesamt());                                            \
                                                                                                                                        \
         if (el->AnzNochNichtBehandelterKPs > 2*(AnzKPEingefuegt-AnzKPGeloescht))                                                       \
            HeapFehler("Regel mit zuvielen noch unbeh. KPs",el->AnzNochNichtBehandelterKPs,  2*(AnzKPEingefuegt-AnzKPGeloescht));       \
                                                                                                                                        \
      }                                                                                                                                 \
      else{                                                                                                                             \
         if (el->ElternNummer < -RE_AnzGleichungenGesamt())                                                                             \
            HeapFehler("Gleichung mit falscher Nummer",el->ElternNummer,  -RE_AnzGleichungenGesamt());                                  \
                                                                                                                                        \
         if (el->AnzNochNichtBehandelterKPs > 2*(AnzKPEingefuegt-AnzKPGeloescht))                                                       \
            HeapFehler("Gleichung mit zuvielen noch unbeh. KPs",el->AnzNochNichtBehandelterKPs, 2*(AnzKPEingefuegt-AnzKPGeloescht));    \
                                                                                                                                        \
      }                                                                                                                                 \
   }                                                                                                                                    \
}

#define ElementKleiner_0(elt1, elt2) (elt1.w1<elt2.w1)
#define ElementKleiner_1(elt1, elt2) ((elt1.w1<elt2.w1) || ((elt1.w1==elt2.w1) && (elt1.w2<elt2.w2)))
#define ElementKleiner_2(elt1, elt2) ((elt1.w1<elt2.w1) || ((elt1.w1==elt2.w1) && (elt1.w2<elt2.w2)))

#define HeapTest_BODY(/* {0, 1, 2} */ StrukturTyp)                                                                                      \
{                                                                                                                                       \
   unsigned long int index = IstHeap?0:1;                                                                                               \
                                                                                                                                        \
   while (index <= LetztesBesetztes){                                                                                                   \
                                                                                                                                        \
      ElterOk(top(StrukturTyp).Vater);                                                                                                  \
      ElterOk(top(StrukturTyp).Mutter);                                                                                                 \
                                                                                                                                        \
      if ((2*index+1<= LetztesBesetztes) &&                                                                                             \
           ElementKleiner_ ##  StrukturTyp (Elemente_ ## StrukturTyp [2*index+1], Elemente_ ## StrukturTyp [index]))                    \
         HeapFehler("Sortierung falsch-1",Elemente_ ## StrukturTyp [index].w1, Elemente_ ## StrukturTyp [2*index+1].w1);                  \
                                                                                                                                        \
      if ((2*index+2<= LetztesBesetztes) &&                                                                                             \
           ElementKleiner_ ##  StrukturTyp (Elemente_ ## StrukturTyp [2*index+2], Elemente_ ## StrukturTyp [index]))                    \
         HeapFehler("Sortierung falsch-2", Elemente_ ## StrukturTyp [index].w1, Elemente_ ## StrukturTyp [2*index+2].w1);                 \
                                                                                                                                        \
      if (!top(StrukturTyp).tp)                                                                                                         \
         HeapFehler("Termpaar fehlt!",0L,0L);                                                                                           \
                                                                                                                                        \
      index++;                                                                                                                          \
   }                                                                                                                                    \
}

static void HeapTest(char *str)
{
  printf("%s HEAP1-TEST %ld %s        \r", ok?"++":"--", ++testNr, str);
/*   printf("+");*/
   
   if (testNr == 50)
      printf("bla\n");
   switch (WeightEntryStrukturTyp){
   case 0:
      HeapTest_BODY(0);
      break;
   case 1:
      HeapTest_BODY(1);
      break;
   case 2:
      HeapTest_BODY(2);
      break;
   default:
      break;
   }

}
#else
#define HeapTest(str)
#endif

/*-----------------------------------------------------------------------------*/
/*------------------------------ Potenzierung ---------------------------------*/
/*-----------------------------------------------------------------------------*/

#define /*int*/ ZweitePotenzVon(/*int*/ n) (1<<(n))

/*-----------------------------------------------------------------------------*/
/*------------------------------ Einsinken Lassen -----------------------------*/
/*-----------------------------------------------------------------------------*/
/* Diese Funktion laesst das Element an der Stelle 'position' in den Heap      */
/* einsinken. Dabei wird das Gewicht als Vergleichs-Mittel herangezogen.       */
/* Der Rueckgabewert zeigt an, ob noch Veraenderungen im Heap noetig waren.    */
/* Die folgenden Funktionen unterscheiden sich in der Art und Weise, wie KPs   */
/* mit gleichem Gewicht behandelt werden. Eingestellt wird die entsprechende   */
/* Funktion in InitAposteriori.                                                */
/*-----------------------------------------------------------------------------*/



/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Element-Vergleiche ----------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/



#define C_Groesser_0(e1, e2) (e1.w1 > e2.w1)

#define C_Groesser_1(e1,e2) ((e1.w1 > e2.w1) || ((e1.w1 == e2.w1) && (e1.w2 > e2.w2)))

#define C_Groesser_2(e1, e2)                                                                                    \
  ( (e1.w1 > e2.w1) || ((e1.w1 == e2.w1) && (e1.w2 > e1.w2)) ||                                                 \
   (  (e1.Vater->LebtNoch && e1.Mutter->LebtNoch && e2.Vater->LebtNoch && e2.Mutter->LebtNoch)?                 \
    ( (e1.w1 == e2.w1) && (e1.w2 == e2.w2) &&                                                                   \
     (C_AdditionalComparison (e1.w1, e1.w2, e1.addref, e1.Vater->re, e1.Mutter->re, e1.tp,                      \
                              e2.w1, e2.w2, e2.addref, e2.Vater->re, e2.Mutter->re, e2.tp) == 1)) : TRUE))


/* Die additional comparison darf nur mit lebenden KPs aufgerufen werden!
   Die Gleichheits-Pruefung mit C_Gleich_2 wird nur aufgerufen, wenn die Groesser-Pruefung
   fehlgeschlagen ist. Bei toten Eltern liefert diese aber TRUE, daher sind die Eltern
   bei Aufruf von C_Gleich_2 und DS_Groesser_? auf jeden Fall am Leben.*/



#define C_Gleich_0(e1, e2) (e1.w1 == e2.w1)

#define C_Gleich_1(e1, e2) ( (e1.w1 == e2.w1) && (e1.w2 == e2.w2))

#define C_Gleich_2(e1, e2)                                                                              \
  ( (e1.w1 == e2.w1) && (e1.w2 == e2.w2) &&                                                             \
    (C_AdditionalComparison(e1.w1,e1.w2,e1.addref, e1.Vater->re, e1.Mutter->re, e1.tp,                  \
                            e2.w1,e2.w2,e2.addref, e2.Vater->re, e2.Mutter->re, e2.tp) == 0))


#define DS_Groesser_0(eintrag1, eintrag2) TRUE

#define DS_Groesser_1(eintrag1, eintrag2) FALSE

#define DS_Groesser_2(eintrag1, eintrag2) (32768&rand())

#define DS_Groesser_3(eintrag1, eintrag2)                                                                       \
((eintrag1.Vater->LebtNoch) && (eintrag1.Mutter->LebtNoch) &&                                                   \
 (eintrag2.Vater->LebtNoch) && (eintrag2.Mutter->LebtNoch)                                                      \
 && TPR_UnambiguousGreaterThan(eintrag1.tp, eintrag1.Vater->re, eintrag1.Mutter->re,                            \
                               eintrag2.tp, eintrag2.Vater->re, eintrag2.Mutter->re))

/* ----------------------------------------------------------------------------*/
/* ------- Lasse Einsinken ----------------------------------------------------*/
/* ----------------------------------------------------------------------------*/
/* Diese Funktion laesst das Element an der Stelle 'position' in den Heap      */
/* einsinken. Dabei wird das Gewicht als Vergleichs-Mittel herangezogen.       */
/* Der Rueckgabewert zeigt an, ob noch Veraenderungen im Heap noetig waren.    */
/* Die folgenden Funktionen unterscheiden sich in der Art und Weise, wie KPs   */
/* mit gleichem Gewicht behandelt werden. Eingestellt wird die entsprechende   */
/* Funktion in InitAposteriori.                                                */
/*-----------------------------------------------------------------------------*/



#define LasseEinsinkenBody(position, /*{0,1,2}*/ CVgl,                                                                                  \
                           /*{0,1,2}*/ DSVgl_horizontal, /*{0,1,2}*/ DSVgl_vertikal)                                                    \
{                                                                                                                                       \
   BOOLEAN changed = FALSE,isHeap = FALSE;                                                                                              \
   unsigned long j=position*2+1;                                                                                                        \
   HeapZellenT_ ## CVgl merker=Elemente_ ## CVgl [position];                                                                            \
                                                                                                                                        \
   while ((!isHeap) && (j<=LetztesBesetztes)){                                                                                          \
      if ((j<LetztesBesetztes) &&                                      /* Position j+1 noch gueltig? */                                 \
          (C_Groesser_ ## CVgl (Elemente_ ## CVgl [j], Elemente_ ## CVgl [j+1]) ||                                                      \
           ( DS_Groesser_ ## DSVgl_horizontal(Elemente_ ## CVgl [j], Elemente_ ## CVgl [j+1]) &&                                        \
             C_Gleich_ ## CVgl (Elemente_ ## CVgl [j], Elemente_ ## CVgl [j+1]))))                                                      \
         j++;                                                           /* => j hat jetzt den Wert des leichteren Nachfolgers */        \
      if ((C_Groesser_ ## CVgl (merker, Elemente_ ## CVgl [j]) ||                                                                       \
           ( DS_Groesser_ ## DSVgl_vertikal(merker, Elemente_ ## CVgl [j]) &&                                                           \
             C_Gleich_ ## CVgl (merker, Elemente_ ## CVgl [j])))){                                                                      \
         Elemente_ ## CVgl [position]=Elemente_ ## CVgl [j];                                                                            \
         changed = TRUE;                                                                                                                \
         position = j;                                                                                                                  \
         j *=2;                                                                                                                         \
         j++;                                                                                                                           \
      }                                                                                                                                 \
      else                                                                                                                              \
         isHeap = TRUE;                                                                                                                 \
   }                                                                                                                                    \
   if (changed)                                                                                                                         \
      Elemente_ ## CVgl [position] = merker;                                                                                            \
   return changed;                                                                                                                      \
}

#define LasseEinsinkenBauen(n1)                                                 \
static BOOLEAN LasseEinsinken_##n1##_egal1(unsigned long position)              \
LasseEinsinkenBody(position, n1, 0, 0)                                          \
static BOOLEAN LasseEinsinken_##n1##_egal2(unsigned long position)              \
LasseEinsinkenBody(position, n1, 0, 1)                                          \
static BOOLEAN LasseEinsinken_##n1##_egal3(unsigned long position)              \
LasseEinsinkenBody(position, n1, 1, 0)                                          \
static BOOLEAN LasseEinsinken_##n1##_egal4(unsigned long position)              \
LasseEinsinkenBody(position, n1, 1, 1)                                          \
static BOOLEAN LasseEinsinken_##n1##_random(unsigned long position)             \
LasseEinsinkenBody(position, n1, 2, 2)                                          \
static BOOLEAN LasseEinsinken_##n1##_unambiguous(unsigned long position)        \
LasseEinsinkenBody(position, n1, 3, 3)

LasseEinsinkenBauen(0)
LasseEinsinkenBauen(1)
LasseEinsinkenBauen(2)


static BOOLEAN (*LasseEinsinkenProz) (unsigned long);

#define /*BOOLEAN*/ LasseEinsinken(/*unsigned long*/ i) ((*LasseEinsinkenProz)(i))

/*-----------------------------------------------------------------------------*/
/*------------------------------ Zugriffe auf den Heap ------------------------*/
/*-----------------------------------------------------------------------------*/

#define /*unsigned long*/ HP_AnzahlEintraege(/*void*/) ( (IstHeap)? LetztesBesetztes:(LetztesBesetztes-1))
/* Gibt die Anzahl der momentan im Heap befindlichen Eintraege zurueck.*/

#define /*BOOLEAN*/ HP_HeapIstLeer(/*void*/) (!(IstHeap) && (LetztesBesetztes==0))
/* Gibt zurueck, ob der Heap leer ist oder nicht. */

#define /*BOOLEAN*/ HP_HeapIstNichtLeer(/*void*/) ((IstHeap) || (LetztesBesetztes>0))
/* Gibt zurueck, ob der Heap Elemente enthaelt oder nicht. */

/* ---------------------------------------------------------------*/
/* --------------- Typen & Prozeduren fuer normalen Heap ---------*/
/* ---------------------------------------------------------------*/

static SV_SpeicherManagerT InfoManager;

static void KPVonElternGeloescht(InfoT *Vater, InfoT *Mutter)
/* Dekrementiert den Zaehler der noch nicht behandelten KPs in Vater und Mutter und loescht
   die Info-Eintraege, falls die zug. Regel o. Gleichung als tot markiert ist und es keine
   unbehandelten KPs zu dieser Regel oder Gleichung im Heap mehr gibt.*/
{
   (Vater)->AnzNochNichtBehandelterKPs--;
   (Mutter)->AnzNochNichtBehandelterKPs--; 
   if (Vater!=Mutter){
      if (((Vater)->AnzNochNichtBehandelterKPs == 0)
          && !((Vater)->LebtNoch)){
         SV_Dealloc(InfoManager, Vater);
      }
      
      if (((Mutter)->AnzNochNichtBehandelterKPs == 0)
          && !((Mutter)->LebtNoch)){
         SV_Dealloc(InfoManager, Mutter);
      }
   }
   else if (((Vater)->AnzNochNichtBehandelterKPs == 0)
            && !((Vater)->LebtNoch)){
      SV_Dealloc(InfoManager, Vater);
   }
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Insert ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define GewichteEintragen_0(elt, W1, W2, ADDREF) (elt.w1 = W1)
#define GewichteEintragen_1(elt, W1, W2, ADDREF) (elt.w1 = W1, elt.w2 = W2)
#define GewichteEintragen_2(elt, W1, W2, ADDREF) (elt.w1 = W1, elt.w2 = W2, elt.addref = ADDREF)

#define INSERT_BODY(tp, w1, w2, addref, father,  mother, /*{0,1,2}*/ WeightEntries, /*{Stat, NoStat}*/ statisticsMode)                   \
{                                                                                                                                        \
   HeapTest("Anfang INSERT");                                                                                                            \
   IncAnzKPEingefuegt_ ## statisticsMode();                                                                                              \
   IncAnzInitialerKPs_ ## statisticsMode(!IsDemigod(mother), (((InfoT *) TP_SUEInfo(father))->AnzInitialerKPs));                        \
                                                                                                                                         \
   if (IstHeap){                                                                                                                         \
      unsigned long position;                                                                                                            \
                                                                                                                                         \
      LetztesBesetztes++;                                                                                                                \
                                                                                                                                         \
      Elemente_ ## WeightEntries [LetztesBesetztes].Vater = (InfoT *) TP_SUEInfo(father);                                                \
      Elemente_ ## WeightEntries [LetztesBesetztes].Mutter = (InfoT *) TP_SUEInfo(mother);                                               \
      Elemente_ ## WeightEntries [LetztesBesetztes].tp = tp;                                                                             \
      GewichteEintragen_ ## WeightEntries (Elemente_ ## WeightEntries [LetztesBesetztes], w1, w2, addref);                               \
                                                                                                                                         \
      position = LetztesBesetztes;                                    /* Da der Heap nicht leer war, sind jetzt min. 2 Elemente drin. */ \
      do{                                                                                                                                \
         position = (position-1)/2;                                                                                                      \
      }while( (LasseEinsinken(position)) && (position > 0) ); /* Reihenfolge wichtig: Einsinkenlassen muss durchgefuehrt werden,*/       \
                                                                    /* auch wenn position schon null ist. */                             \
   }                                                                                                                                     \
   else{                                                            /* Heap hat leeres Top-Element */                                    \
      Elemente_ ## WeightEntries [0].Vater = (InfoT *) TP_SUEInfo(father);                                                               \
      Elemente_ ## WeightEntries [0].Mutter = (InfoT *) TP_SUEInfo(mother);                                                              \
      Elemente_ ## WeightEntries [0].tp = tp;                                                                                            \
      GewichteEintragen_ ## WeightEntries (Elemente_ ## WeightEntries [0], w1, w2, addref);                                              \
      LasseEinsinken(0);                                                                                                                 \
      IstHeap = TRUE;                                                                                                                    \
   }                                                                                                                                     \
   ((InfoT *) TP_SUEInfo(father))->AnzNochNichtBehandelterKPs++;                                                                         \
   ((InfoT *) TP_SUEInfo(mother))->AnzNochNichtBehandelterKPs++;                                                                         \
   HeapTest("Ende INSERT");                                                                                                              \
}

static void KPH_INSERT_0_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, 0, NULL, actualParent, otherParent, 0, NoStat)                                                      
                                                                                                                           
static void KPH_INSERT_1_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, wt.w2, NULL, actualParent, otherParent, 1, NoStat)                                                  
                                                                                                                           
static void KPH_INSERT_2_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, wt.w2, wt.addRef, actualParent, otherParent, 2, NoStat)


static void KPH_INSERT_0_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, 0, NULL, actualParent, otherParent, 0, Stat)                                                            
                                                                                                                               
static void KPH_INSERT_1_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, wt.w2, NULL, actualParent, otherParent, 1, Stat)                                                        
                                                                                                                               
static void KPH_INSERT_2_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, wt.w2, wt.addRef, actualParent, otherParent, 2, Stat)





BOOLEAN KPH_POSTPONE_012(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
{return FALSE;}
/* evtl. auch ohne Parameter, denn es wird ja einfach nur das zuletzt selected'e KP postponed */


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Select ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


#define DESTROY_WEIGHT_0(addref)
#define DESTROY_WEIGHT_1(addref)
#define DESTROY_WEIGHT_2(addref) (C_KillAdditionalReference(addref))

#define ADDREF_BESETZEN_0(addrefPtr, addref) 
#define ADDREF_BESETZEN_1(addrefPtr, addref) 
#define ADDREF_BESETZEN_2(addrefPtr, addref) (*addrefPtr = addref) 


#define SELECT_BODY(tpPtr, actualParentPtr,  otherParentPtr, addrefPtr, /* {0,1,2}*/ DestWeight, /*{Stat, NoStat}*/ statisticsMode)   \
{                                                                                                                                               \
   HeapTest("Anfang SELECT");                                                                                                                   \
   while (HP_HeapIstNichtLeer()){                                                                                                               \
                                                                                                                                                \
      IncAnzKPEntfernt_ ## statisticsMode();                                                                                                    \
                                                                                                                                                \
      if (!IstHeap){                                                                                                                            \
         Elemente_ ## DestWeight [0] = Elemente_ ## DestWeight [LetztesBesetztes];                                                              \
         LetztesBesetztes--;                                                                                                                    \
         LasseEinsinken(0);                                                                                                                     \
      }                                                                                                                                         \
      else                                                                                                                                      \
         IstHeap = FALSE;                                                                                                                       \
                                                                                                                                                \
      /* Jetzt ist Elemente_ ## DestWeight [0] gueltig und der Heap restauriert. */                                                             \
      /* Nach der Entnahme eines Elementes ist IstHeap auf jeden Falle FALSE! */                                                                \
                                                                                                                                                \
      if (Elemente_ ## DestWeight [0].Vater->LebtNoch && Elemente_ ## DestWeight [0].Mutter->LebtNoch){                                         \
         *tp = Elemente_ ## DestWeight [0].tp;                                                                                                  \
         *actualParentPtr = Elemente_ ## DestWeight [0].Vater->re;                                                                              \
         *otherParentPtr = Elemente_ ## DestWeight [0].Mutter->re;                                                                              \
         ADDREF_BESETZEN_ ## DestWeight (addrefPtr, Elemente_ ## DestWeight [0].addref);                                                        \
         KPVonElternGeloescht(Elemente_ ## DestWeight [0].Vater, Elemente_ ## DestWeight [0].Mutter);                                           \
                                                                                                                                                \
         HeapTest("Ende SELECT" );                                                                                                              \
         return TRUE;                                                                                                                           \
      }                                                                                                                                         \
      else{                                                                                                                                     \
         /* Meldung, Zaehler */                                                                                                                 \
         TPR_DESTROYTERMPAIR(Elemente_ ## DestWeight [0].tp, Elemente_ ## DestWeight [0].Vater->re, Elemente_ ## DestWeight [0].Mutter->re);    \
         DESTROY_WEIGHT_ ## DestWeight(Elemente_ ## DestWeight [0].addref);                                                                     \
         KPVonElternGeloescht(Elemente_ ## DestWeight [0].Vater, Elemente_ ## DestWeight [0].Mutter);                                           \
         IncAnzKPSpaeterWaisenmord_ ## statisticsMode();                                                                                        \
      }                                                                                                                                         \
   }                                                                                                                                            \
   return FALSE;                                                                                                                                \
}

#define SELECTINITIAL_BODY(tp, actualParent, otherParent, addref, /* {0,1,2}*/ DestWeight, /*{Stat, NoStat}*/ statisticsMode) \
{                                                                                                                             \
  int Index = 0;                                                                                                              \
  HeapTest("Anfang SELECT-INITIAL");                                                                                          \
                                                                                                                              \
  if (LetztesBesetztes == 0) /* Heap leer -> alle initialen weg */                                                            \
    return FALSE;                                                                                                             \
                                                                                                                              \
  if (!IstHeap){                                                                                                              \
    Elemente_##DestWeight[0] = Elemente_##DestWeight[LetztesBesetztes];                                                       \
    LetztesBesetztes--;                                                                                                       \
    LasseEinsinken(0);                                                                                                        \
    IstHeap = TRUE;                                                                                                           \
  }                                                                                                                           \
                                                                                                                              \
  while(!IsDemigod(Elemente_##DestWeight[Index].Vater->re) && (Elemente_##DestWeight[Index].w1!=LONG_MIN)                     \
        && (Index<LetztesBesetztes))                                                                                          \
    Index++;                                                                                                                  \
                                                                                                                              \
  if (Index==LetztesBesetztes) return FALSE;                                                                                  \
  else{                                                                                                                       \
    IncAnzKPEntfernt_ ## statisticsMode();                                                                                    \
                                                                                                                              \
    *tp = Elemente_##DestWeight [Index].tp;                                                                                   \
    *actualParent = Elemente_##DestWeight [Index].Vater->re;                                                                  \
    *otherParent = Elemente_##DestWeight [Index].Mutter->re;                                                                  \
                                                                                                                              \
    KPVonElternGeloescht(Elemente_##DestWeight[Index].Vater, Elemente_##DestWeight[Index].Mutter);                            \
                                                                                                                              \
    if (Index<LetztesBesetztes) Elemente_##DestWeight[Index]=Elemente_##DestWeight[LetztesBesetztes];                         \
                                                                                                                              \
    if (LetztesBesetztes>0){                                                                                                  \
      int run=Index;                                                                                                          \
      LetztesBesetztes--;                                                                                                     \
      while (run>0){                                                                                                          \
        LasseEinsinken(run);                                                                                                  \
        run = (run-1)/2;                                                                                                      \
      }                                                                                                                       \
      LasseEinsinken(0);                                                                                                      \
    }                                                                                                                         \
                                                                                                                              \
    HeapTest("Ende SELECT-INITIAL" );                                                                                         \
                                                                                                                              \
    return TRUE;                                                                                                              \
  }                                                                                                                           \
}
      
      
#define SelectBauen(w,s)                                                                                                              \
static BOOLEAN KPH_SELECT_##w##_##s(TermpairT *tp, RegelOderGleichungsT *actualParent, RegelOderGleichungsT *otherParent, void **addref)        \
SELECT_BODY(tp, actualParent, otherParent, addref, w, s)                                                                                 \
static BOOLEAN KPH_SELECTINITIAL_##w##_##s(TermpairT *tp, RegelOderGleichungsT *actualParent, RegelOderGleichungsT *otherParent, void **addref) \
SELECTINITIAL_BODY(tp, actualParent, otherParent, addref, w, s)

SelectBauen(0,Stat)
SelectBauen(1,Stat)
SelectBauen(2,Stat)
SelectBauen(0,NoStat)
SelectBauen(1,NoStat)
SelectBauen(2,NoStat)


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Parents ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void KPH_NEWPARENT(RegelOderGleichungsT re)
{
   InfoT *infoHP;

   SV_Alloc(infoHP,InfoManager,InfoT *);
   infoHP -> LebtNoch = TRUE;
   infoHP -> ElternNummer = TP_ElternNr(re);
   infoHP -> AnzNochNichtBehandelterKPs = 0;
   infoHP -> re = re;

   if (PQStatisticsAn) infoHP -> AnzInitialerKPs = 0;
      
   TP_SUEInfo(re) = (void *) infoHP;

   if (!TP_TermpaarIstRegel(re)){ /* dh. Gleichung */
      SV_Alloc(infoHP,InfoManager,InfoT *);
      infoHP -> LebtNoch = TRUE;
      infoHP -> ElternNummer = TP_ElternNr(re);
      infoHP -> AnzNochNichtBehandelterKPs = 0;
      infoHP -> re = TP_Gegenrichtung(re);
      if (PQStatisticsAn) infoHP -> AnzInitialerKPs = 0;
      TP_SUEInfo(TP_Gegenrichtung(re)) = (void *) infoHP;
   }
   HeapTest("NEWPARENT");
}


void KPH_KILLPARENT(RegelOderGleichungsT re)
{
   InfoT *info = (InfoT *)TP_SUEInfo(re);

   info-> LebtNoch = FALSE;
   if (info->AnzNochNichtBehandelterKPs == 0){
      SV_Dealloc(InfoManager, info);
   }
   else
      info ->re = NULL;
      
   TP_SUEInfo(re) = (void *) NULL;

   if (!TP_TermpaarIstRegel(re)){ /* dh. Gleichung */
      info = (InfoT *)TP_SUEInfo(TP_Gegenrichtung(re));
      info-> LebtNoch = FALSE;
      if ( info->AnzNochNichtBehandelterKPs == 0){
         SV_Dealloc(InfoManager, info);
      }
      else
         info -> re =NULL;
      TP_SUEInfo(TP_Gegenrichtung(re)) = (void *) NULL;
   }
   HeapTest("KILLPARENT");
}

void KPH_REDIRECTPARENT(void)
{
   ;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Walkthrough -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

unsigned long int KPH_AnzKPsToGo(void)
{
   return IstHeap?LetztesBesetztes:(LetztesBesetztes+1);
}

#define TermeAuspacken_terms(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr) \
TPR_TERMPAIRtoPARALLEL_INTERMED(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr)
#define TermeAuspacken_noterms(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr) /* nix tun */

#define TermeWiederEinpacken_terms(tp, lhs, rhs, actualParent, otherParent, changed, pclID)\
TPR_PARALLELtoTERMPAIR(tp, lhs, rhs, actualParent, otherParent, changed, pclID)
#define TermeWiederEinpacken_noterms(tp, lhs, rhs, actualParent, otherParent, changed, pclID)  tp

#define TermeZerstoeren_terms(tp, lhs, rhs, actualParent, otherParent)\
TPR_DESTROYPARALLEL(tp, lhs, rhs, actualParent, otherParent)
#define TermeZerstoeren_noterms(tp, lhs, rhs, actualParent, otherParent) /* nix tun */

#define GewichteExtrahieren_0(wt, elt) (wt.w1 = elt.w1)
#define GewichteExtrahieren_1(wt, elt) (wt.w1 = elt.w1, wt.w2 = elt.w2)
#define GewichteExtrahieren_2(wt, elt) (wt.w1 = elt.w1, wt.w2 = elt.w2, wt.addRef = elt.addref)



#define WALKTHROUGH_BODY(/* {terms, noterms} */ TermeBenoetigt, /* {0, 1, 2}*/ WeightType, /*{Stat, NoStat}*/ statisticsMode) \
{                                                                                                                             \
   if (!IstHeap && (LetztesBesetztes>0)){                                                                                     \
     Elemente_ ## WeightType [0]= Elemente_ ## WeightType [LetztesBesetztes--];                                               \
     LasseEinsinken(0); /* Heap reorganisieren */                                                                             \
     IstHeap = TRUE;                                                                                                          \
   }                                                                                                                          \
                                                                                                                              \
   HeapTest("Anfang WALKTHROUGH");                                                                                            \
   if (HP_HeapIstNichtLeer()){                                                                                                \
      unsigned long int index = 0;                                                                                            \
      BOOLEAN resort = FALSE;                                                                                                 \
                                                                                                                              \
      while((index <= LetztesBesetztes) && IstHeap){                                                                          \
                                                                                                                              \
         IncAnzKPIRBetrachtet_ ## statisticsMode();                                                                           \
         if (!(Elemente_ ## WeightType [index].Vater->LebtNoch && Elemente_ ## WeightType [index].Mutter->LebtNoch)){         \
            TPR_DESTROYTERMPAIR(Elemente_ ## WeightType [index].tp, Elemente_ ## WeightType [index].Vater->re,                \
                                Elemente_ ## WeightType [index].Mutter->re);                                                  \
            DESTROY_WEIGHT_ ## WeightType (Elemente_ ## WeightType [index].addref);                                           \
            KPVonElternGeloescht(Elemente_ ## WeightType [index].Vater, Elemente_ ## WeightType [index].Mutter);              \
            IncAnzKPIRWaisenmord_ ## statisticsMode();                                                                        \
            IncAnzKPEntfernt_ ## statisticsMode();                                                                            \
            if (index<LetztesBesetztes)                                                                                       \
               Elemente_ ## WeightType [index] = Elemente_ ## WeightType [LetztesBesetztes];                                  \
            if (LetztesBesetztes){                                                                                            \
               LetztesBesetztes--;                                                                                            \
            }                                                                                                                 \
            else IstHeap = FALSE;                                                                                             \
                                                                                                                              \
            resort = TRUE;                                                                                                    \
         }                                                                                                                    \
         else{                                                                                                                \
            int APresult;                                                                                                     \
            PCL_ObjektNummerT pclID;                                                                                          \
            TermT lhs = NULL, rhs = NULL;                                                                                     \
                                                                                                                              \
            TermeAuspacken_ ## TermeBenoetigt(Elemente_ ## WeightType [index].tp, &lhs, &rhs,                                 \
                                              Elemente_ ## WeightType [index].Vater->re,                                      \
                                              Elemente_ ## WeightType [index].Mutter->re, &pclID);                            \
                                                                                                                              \
            if ((APresult = (* AP) (lhs, rhs, Elemente_ ## WeightType [index].Vater->re,                                      \
                                    Elemente_ ## WeightType [index].Mutter->re, &pclID)) == 1){                               \
               TermeZerstoeren_ ## TermeBenoetigt(Elemente_ ## WeightType [index].tp, lhs, rhs,                               \
                                                  Elemente_ ## WeightType [index].Vater->re,                                  \
                                                  Elemente_ ## WeightType [index].Mutter->re);                                \
               DESTROY_WEIGHT_ ## WeightType (Elemente_ ## WeightType [index].addref);                                        \
               KPVonElternGeloescht(Elemente_ ## WeightType [index].Vater,                                                    \
                                    Elemente_ ## WeightType [index].Mutter);                                                  \
               if (index<LetztesBesetztes)                                                                                    \
                  Elemente_ ## WeightType [index] = Elemente_ ## WeightType [LetztesBesetztes];                               \
               if (LetztesBesetztes){                                                                                         \
                  LetztesBesetztes--;                                                                                         \
               }                                                                                                              \
               else IstHeap = FALSE;                                                                                          \
               resort = TRUE;                                                                                                 \
               IncAnzKPEntfernt_ ## statisticsMode();IncAnzKPIRGeloescht_ ## statisticsMode();                                \
            }                                                                                                                 \
            else if (APresult == 2){                                                                                          \
               WeightEntryT wt;                                                                                               \
                                                                                                                              \
               IncAnzKPIRVeraendert_ ## statisticsMode();                                                                     \
               GewichteExtrahieren_ ## WeightType(wt, Elemente_ ## WeightType [index]);                                       \
               C_RECLASSIFY(&wt, lhs, rhs, Elemente_ ## WeightType [index].Vater->re,                                         \
                            Elemente_ ## WeightType [index].Mutter->re);                                                      \
               Elemente_ ## WeightType [index].tp =                                                                           \
                  TermeWiederEinpacken_ ## TermeBenoetigt(Elemente_ ## WeightType [index].tp, lhs, rhs,                       \
                                                          Elemente_ ## WeightType [index].Vater->re,                          \
                                                          Elemente_ ## WeightType [index].Mutter->re, TRUE, pclID);           \
               GewichteEintragen_ ## WeightType(Elemente_ ## WeightType [index], wt.w1, wt.w2, wt.addRef);                    \
               resort = TRUE;                                                                                                 \
               index++;                                                                                                       \
            }                                                                                                                 \
            else{               /* noop */                                                                                    \
               Elemente_ ## WeightType [index].tp =                                                                           \
                  TermeWiederEinpacken_ ## TermeBenoetigt(Elemente_ ## WeightType [index].tp, lhs, rhs,                       \
                                                          Elemente_ ## WeightType [index].Vater->re,                          \
                                                          Elemente_ ## WeightType [index].Mutter->re, TRUE, pclID);                  \
               index++;                                                                                                       \
            }                                                                                                                 \
         }                                                                                                                    \
      }                                                                                                                       \
                                                                                                                              \
      if ((LetztesBesetztes>0) && resort){                                                                                    \
         index = (LetztesBesetztes-1) / 2;                                                                                    \
         while (index>0){                                                                                                     \
            index--;                                                                                                          \
            LasseEinsinken(index);                                                                                            \
         }                                                                                                                    \
      }                                                                                                                       \
   }                                                                                                                          \
   HeapTest("Ende WALKTHROUGH");                                                                                              \
}

#define WALKTHROUGH_WEIGHT_BODY(/* {0, 1, 2}*/ WeightType)                                                                      \
{                                                                                                                               \
   if (PQStatisticsAn)                                                                                                          \
      if (NeedsTerms)  WALKTHROUGH_BODY(terms, WeightType, Stat)                                                                \
      else             WALKTHROUGH_BODY(noterms, WeightType, Stat)                                                              \
   else                                                                                                                         \
      if (NeedsTerms)  WALKTHROUGH_BODY(terms, WeightType, NoStat)                                                              \
      else             WALKTHROUGH_BODY(noterms, WeightType, NoStat)                                                            \
}

static void KPH_WALKTHROUGH_0(PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGH_WEIGHT_BODY(0)

static void KPH_WALKTHROUGH_1(PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGH_WEIGHT_BODY(1)

static void KPH_WALKTHROUGH_2(PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGH_WEIGHT_BODY(2)


#define WW_CONDITION(/* {0, 1, 2}*/ WeightType)                                                                       \
(Elemente_ ## WeightType [index].Vater->LebtNoch && Elemente_ ## WeightType [index].Mutter->LebtNoch)


#define WW_BODY(/* {0, 1, 2}*/ WeightType)                                                                                   \
{                                                                                                                            \
   unsigned long int index;                                                                                                  \
                                                                                                                             \
   if (!IstHeap && (LetztesBesetztes==0)) return; /* Heap ist leer */                                                        \
                                                                                                                             \
   index = IstHeap?0:1;                                                                                                      \
                                                                                                                             \
   while(index <= LetztesBesetztes){                                                                                         \
                                                                                                                             \
      if (WW_CONDITION(WeightType)){                                                                                         \
        WeightEntryT wt;                                                                                                     \
        GewichteExtrahieren_ ## WeightType(wt, Elemente_ ## WeightType [index]);                                             \
        (* AP)(wt, Elemente_ ## WeightType [index].tp, Elemente_ ## WeightType [index].Vater->re,                            \
                   Elemente_ ## WeightType [index].Mutter->re);                                                              \
      }                                                                                                                      \
      index++;                                                                                                               \
   }                                                                                                                         \
}

void KPH_WATCHINGWALKTHROUGH(PQ_WatchingWalkthroughProcedureT AP)
{
   switch(WeightEntryStrukturTyp){
   case 1: WW_BODY(1); break;
   case 2: WW_BODY(2); break;
   case 0:
   default: WW_BODY(0); break;
   }
}

#define TermVerfolgungsBody(/* {0, 1, 2}*/ WeightType)                                                                      \
{                                                                                                                            \
   unsigned long int index;                                                                                                  \
                                                                                                                             \
   if (!IstHeap && (LetztesBesetztes==0)) return; /* Heap ist leer */                                                        \
                                                                                                                             \
   index = IstHeap?0:1;                                                                                                      \
                                                                                                                             \
   while(index <= LetztesBesetztes){                                                                                         \
      TPR_OperationWaehrendTermverfolgung(OperationAufTermen, Elemente_ ## WeightType [index].tp,                            \
                                          Elemente_ ## WeightType [index].Vater->re,                                         \
                                          Elemente_ ## WeightType [index].Mutter->re);                                       \
      index++;                                                                                                               \
   }                                                                                                                         \
}


static void KPH_TermVerfolgung_0(void (* OperationAufTermen) (TermT, TermT))
TermVerfolgungsBody(0)

static void KPH_TermVerfolgung_1(void (* OperationAufTermen) (TermT, TermT))
TermVerfolgungsBody(1)

static void KPH_TermVerfolgung_2(void (* OperationAufTermen) (TermT, TermT))
TermVerfolgungsBody(2)
         
/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

unsigned long int Lebende;

static void LebendsKPsZaehlen(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   Lebende++;
}


void KPH_STATISTICS(InfoStringT is)
{
   if (!(IS_IsIn("all", is) || (IS_IsIn("PQ", is))))
      return;                   /* Modul nicht angesprochen */
   if (IS_IsIn("standard", is)){
      IO_DruckeKurzUeberschrift("Priority-Queue: Regular heap", "+");
      IO_DruckeStatistikZeile("actual number of elements in heap", HP_HeapIstLeer()?0:(LetztesBesetztes - IstHeap?0:1));
      if (PQStatisticsAn){
         Lebende = 0;
         KPH_WATCHINGWALKTHROUGH(LebendsKPsZaehlen);
         PQStatistik(is, Lebende);
      }
      if (VerwendeteSaat != 0){
         fprintf(IO_Strom, "\n");
         IO_DruckeStatistikZeile("seed for random-comparion", VerwendeteSaat);
      }
   }
   else if(IS_IsIn("latex-head", is)){
      fprintf(IO_Strom, "PQ: total &");
      if (PQStatisticsAn)
         PQStatistik(is, 0);
   }
   else if(IS_IsIn("intern", is))
      IO_DruckeStatistikZeile("KPH: domain granularity", DomainGran);
   else if(IS_IsIn("config", is)){
      fprintf(IO_Strom, "KPH: Statistik %s, ", PQStatisticsAn?" AN":"AUS");
#if KPH_TEST
      fprintf(IO_Strom, "Heap-Test-Modus  ANn");
#else
      fprintf(IO_Strom, "Heap-Test-Modus AUS\n");
#endif
   }      
}


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Clean up --------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define AbraeumenBody(/* {0,1,2} */ WeightType)                                                                                 \
{                                                                                                                               \
   unsigned long int i;                                                                                                         \
                                                                                                                                \
   for (i=IstHeap?0:1; i<=LetztesBesetztes; i++){                                                                               \
                                                                                                                                \
      TPR_DESTROYTERMPAIR(Elemente_ ## WeightType[i].tp, Elemente_ ## WeightType[i].Vater->re,                                  \
                          Elemente_ ## WeightType[i].Mutter->re);                                                               \
                                                                                                                                \
      DESTROY_WEIGHT_ ## WeightType(Elemente_ ## WeightType[i].addref);                                                         \
                                                                                                                                \
      KPVonElternGeloescht(Elemente_ ## WeightType[i].Vater, Elemente_ ## WeightType[i].Mutter);                                \
   }                                                                                                                            \
                                                                                                                                \
   SM_NichtGefrorenenStrahlLoeschen(Elemente_ ## WeightType);                                                                   \
}

void KPH_SoftAbraeumen(void)
/* Nur Nicht-SV-Manager-abhaengiges leeren, aber Wiederaufsatz-kompatibel */
{
  SM_StrahldomaeneLoeschen(Domaene);
}

void KPH_Abraeumen(void)
/* Alles loeschen*/
{
   switch (WeightEntryStrukturTyp){
   case 0:
      AbraeumenBody(0);
      break;
   case 1:
      AbraeumenBody(1);
      break;
   case 2:
      AbraeumenBody(2);
      break;
   default:
      break;
   }
   SM_StrahldomaeneLoeschen(Domaene);
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
static BOOLEAN KPH_SchonInitialisiert = FALSE;


#define Initialisierung(/* {0, 1, 2} */ StrukturTyp)                                                                       \
{                                                                                                                          \
   Elemente_ ## StrukturTyp = SM_NeuerStrahl(Domaene);                                                                     \
   KPH_WALKTHROUGH = KPH_WALKTHROUGH_ ## StrukturTyp;                                                                      \
   KPH_TermVerfolgung = KPH_TermVerfolgung_ ## StrukturTyp;                                                                \
   switch (IS_FindOption(DEF, USR, 6, "e1", "e2", "e3", "e4", "random", "unambiguous")){                                   \
   case 0:                                                                                                                 \
     {                                                                                                                     \
       if (!KPH_SchonInitialisiert) fprintf(IO_Strom, "e1\n");                                                             \
       LasseEinsinkenProz = LasseEinsinken_ ## StrukturTyp ## _egal1;                                                      \
       break;                                                                                                              \
     }                                                                                                                     \
   case 1:                                                                                                                 \
     {                                                                                                                     \
       if (!KPH_SchonInitialisiert) fprintf(IO_Strom, "e2\n");                                                             \
       LasseEinsinkenProz = LasseEinsinken_ ## StrukturTyp ## _egal2;                                                      \
       break;                                                                                                              \
     }                                                                                                                     \
   case 2:                                                                                                                 \
     {                                                                                                                     \
       if (!KPH_SchonInitialisiert) fprintf(IO_Strom, "e3\n");                                                             \
       LasseEinsinkenProz = LasseEinsinken_ ## StrukturTyp ## _egal3;                                                      \
       break;                                                                                                              \
     }                                                                                                                     \
   case 3:                                                                                                                 \
     {                                                                                                                     \
       if (!KPH_SchonInitialisiert) fprintf(IO_Strom, "e4\n");                                                             \
       LasseEinsinkenProz = LasseEinsinken_ ## StrukturTyp ## _egal4;                                                      \
       break;                                                                                                              \
     }                                                                                                                     \
   case 4:                                                                                                                 \
   {                                                                                                                       \
      struct timezone tzp;                                                                                                 \
      struct timeval tp;                                                                                                   \
                                                                                                                           \
      LasseEinsinkenProz = LasseEinsinken_ ## StrukturTyp ## _random;                                                      \
      tp.tv_sec = 0;                                                                                                       \
      tp.tv_usec = 0;                                                                                                      \
      tzp.tz_minuteswest = 0;                                                                                              \
      tzp.tz_dsttime = 0;                                                                                                  \
                                                                                                                           \
      if (gettimeofday(&tp, &tzp) == -1)                                                                                   \
         our_error("gettimeofday in KPHeap.c fehlgeschlagen.\n\n");                                                        \
      srand(VerwendeteSaat=tp.tv_usec);                                                                                    \
      fprintf(IO_Strom, "random (computed seed = %ld)\n", VerwendeteSaat);                                                 \
      break;                                                                                                               \
   }                                                                                                                       \
   case 5: fprintf(IO_Strom, "unambiguous\n"); LasseEinsinkenProz = LasseEinsinken_ ## StrukturTyp ## _unambiguous; break; \
   default:                                                                                                                \
      if (IS_FindValuedOption(DEF, USR, 1, "random") == 0){                                                                \
         LasseEinsinkenProz = LasseEinsinken_ ## StrukturTyp ## _random;                                                   \
         VerwendeteSaat = atoi(IS_ValueOfOption(DEF, USR, "random"));                                                      \
         fprintf(IO_Strom, "random (given seed = %ld)\n", VerwendeteSaat);                                                 \
         srand(VerwendeteSaat);                                                                                            \
      }                                                                                                                    \
      else{                                                                                                                \
         LasseEinsinkenProz = LasseEinsinken_ ## StrukturTyp ## _egal1;                                                    \
         fprintf(IO_Strom, "e1\n");                                                                                        \
      }                                                                                                                    \
      break;                                                                                                               \
  }                                                                                                                        \
}


#define Zuweisen(w)                                             \
{                                                               \
  if (PQStatisticsAn){                                          \
    KPH_INSERT = KPH_INSERT_##w##_STATISTICS;                   \
    KPH_POSTPONE = KPH_POSTPONE_012;                            \
    KPH_SELECT = KPH_SELECT_##w##_Stat;                         \
    KPH_SELECTINITIAL = KPH_SELECTINITIAL_##w##_Stat;           \
  }                                                             \
  else{                                                         \
    KPH_INSERT = KPH_INSERT_##w##_NOSTAT;                       \
    KPH_POSTPONE = KPH_POSTPONE_012;                            \
    KPH_SELECT = KPH_SELECT_##w##_NoStat;                       \
    KPH_SELECTINITIAL = KPH_SELECTINITIAL_##w##_NoStat;         \
  }                                                             \
}

BOOLEAN KPH_INITIALIZE(InfoStringT DEF, InfoStringT USR, int WeightEntryType)
{
   char *statOption;

   WeightEntryStrukturTyp = WeightEntryType;

   if ((statOption = IS_ValueOfOption(DEF, USR, "statistics")))
      PQStatisticsAn = (strcmp(statOption, "on") == 0);
   else
      PQStatisticsAn = FALSE;
   
   if (PQStatisticsAn)
      PQStatistikInitialisieren();
   
   if (IS_ValueOfOption(DEF, USR, "domgran"))
      DomainGran = atoi(IS_ValueOfOption(DEF, USR, "domgran"));
   else
      DomainGran = 10;
   
   Domaene = SM_NeueStrahldomaene(KPH_Domaenenstart, KPH_Domaenenende, DomainGran, "Domain for regular heap");

   if (!KPH_SchonInitialisiert)
     fprintf(IO_Strom, "Regular Heap. Weight-EntryType = %d. statistics = %s. "
             "domgran = %ld. ", 
             WeightEntryType, PQStatisticsAn ? "on":"off", DomainGran);
   
   IstHeap = FALSE;        /* Dh. die Wurzel ist unbesetzt ! */

   LetztesBesetztes = 0;

   if (!KPH_SchonInitialisiert)
     if (PQStatisticsAn){
       SV_ManagerInitialisieren(&InfoManager, sizeof(InfoT_MitStatistik), PQ_AnzahlInfoBloeckeJeBlock, "KP-Infobloecke (Heap)2");
     }
     else{
       SV_ManagerInitialisieren(&InfoManager, sizeof(InfoT_OhneStatistik), PQ_AnzahlInfoBloeckeJeBlock, "KP-Infobloecke (Heap)3");
     }

   switch (WeightEntryType){
   case 0:
      Initialisierung(0);
      Zuweisen(0);
      break;
   case 1:
      Initialisierung(1);
      Zuweisen(1);

      break;
   case 2:
      Initialisierung(2);
      Zuweisen(2);

      break;
   default:
      return FALSE;
      break;
   }
   KPH_INSERTImmortal = KPH_INSERT;
   KPH_SchonInitialisiert = TRUE;
   return TRUE;
}

void KPH_ActSize(unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl)
{
  Lebende = 0;
  KPH_WATCHINGWALKTHROUGH(LebendsKPsZaehlen);
  
  *AktGueltigeKPZahl = Lebende;
  *AktKPZahl = AnzKPEingefuegt - AnzKPGeloescht;
}
