/* ------------------------------------------------------------------------------------------------------------*/
/*                                                                                                             */
/*  KPPairing.c                                                                                                */
/*  ***********                                                                                                */
/*                                                                                                             */
/*  Dieses Modul wurde am 14.03.95 aus den beiden Modulen PaiirngHeap.c und KPPairing.c zusammengelegt, um     */
/*  das laestige Casten sowie unnoetiges Data-Hiding zu vermeiden.                                             */
/*  In diesem Modul wird zunaechst ein Pairing-Heap realisiert, der dynamisches Neubewerten sowie das Entfernen*/
/*  beliebiger Elemente des Heaps ermoeglicht.                                                                 */
/*  Darauf aufbauend wird dann die Funktionalitaet der KP-Menge realisiert.                                    */
/*                                                                                                             */
/*                                                                                                             */
/*                                                                                                             */
/*  I) Der Pairing-Heap                                                                                        */
/*                                                                                                             */
/*  Realisierung eines Pairing-Heaps zur KP-Verwaltung.                                                        */
/*  Moeglich sind Funktionen wie Einfuegen, KleinstesEntnehmen, BeliebigesEntnehmen und BeliebigesNeubewerten. */
/*                                                                                                             */
/*                                                                                                             */
/*  DEFINITION Pairing-Heap:                                                                                   */
/*  ========================                                                                                   */
/*                                                                                                             */
/*  - Die leere Menge ist ein Pairing-Heap.                                                                    */
/*  - Die aus einem Element bestehende Menge ist ein Pairing-Heap.                                             */
/*  - Sind h1, h2 .. hm Pairing-Heaps, so ist auch                                                             */
/*                                                                                                             */
/*                    <n>                                                                                      */
/*                   / | \ ..                                                                                  */
/*                  V  V  V                                                                                    */
/*                 h1  h2 .... hm                                                                              */
/*                                                                                                             */
/*    ein Pairing-Heap, wenn <n> ein Knoten ist, dessen Eintrag n kleiner oder gleich allen Eintraegen der     */
/*    Heaps h1 .. hm ist.                                                                                      */
/*                                                                                                             */
/*                                                                                                             */
/*  REALISIERUNG/DATENSTRUKTUR                                                                                 */
/*  ==========================                                                                                 */
/*                                                                                                             */
/*  Ein Knoten im Pairing-Heap beinhaltet neben dem zugehoerigen Eintrag und dem entsprechenden Gewicht einen  */
/*  Verweis auf die 'Menge' seiner Nachfolger sowie einen Verweis auf seinen Nachbarn und Vorgaenger-Knoten.   */
/*  Obige Skizze ist also wie folgt zu interpretieren:                                                         */
/*                                                                                                             */
/*                                                                                                             */
/*                                                                                                             */
/*                  <n> <______                                                                                */
/*                  /^^        \                                                                               */
/*                 V/  \        \                                                                              */
/*                 h1->h2->...->hn-|                                                                           */
/*                                                                                                             */
/*  Jeder Knoten beinhaltet also einen Zeiger auf seinen juengsten Nachfolger sowie auf seinen rechten Nach-   */
/*  barn und seinen Vater.                                                                                     */
/*  Der Wurzelknoten eines Pairing-Heaps hat per Konvention die Eintraege Vater == NULL und Bruder ==  NULL.   */
/*                                                                                                             */
/*  13.03.95. Einbau des Heap-Durchlaufs.                                                                      */
/*                                                                                                             */
/*  Erweiterung der Datenstruktur um einen Zeiger, der die Verkettung der zu loeschenden bzw. zu verschiebenden*/
/*  Elemente waehrend des Durchlaufs ermoeglicht. Dies ist noetig, da eine Verschiebung/Entfernung waehrend    */
/*  des Durchlaufens den Heap so veraendern wuerde, dass keine Kontrolle ueber schon behandelte oder noch zu   */
/*  behandelnde Elemente moeglich waere.                                                                       */
/*                                                                                                             */
/* ------------------------------------------------------------------------------------------------------------*/

/*-------------------------------------------------------------------------------------------------------------*/
/*-- 1) Includes ----------------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------*/

#include "antiwarnings.h"
#include <stdlib.h>
#include <string.h>

/* Fuer die Zufallszahlen */
#include <sys/time.h>
#include <sys/resource.h>

#include "KPPairing.h"
#include "Parameter.h"
#include "ZeitMessung.h"
#include "NFBildung.h"
#include "general.h"
#include "SpeicherVerwaltung.h"
#include "FehlerBehandlung.h"
#include "Ausgaben.h"
#include "RUndEVerwaltung.h"
#include "PQStatistik.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe KPPairing.h) -------------------------*/
/* ------------------------------------------------------------------------*/                                                                             
void (* KPP_INSERT)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);
void (* KPP_INSERTImmortal)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);

BOOLEAN (* KPP_POSTPONE) (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);
BOOLEAN (* KPP_SELECT) (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **);
BOOLEAN (* KPP_SELECTINITIAL) (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **);
void (* KPP_WALKTHROUGH) (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms);

/*-------------------------------------------------------------------------------------------------------------*/
/*-- 2) Allgemeine Typen --------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------*/

typedef struct KPZellenTS *ZellenZeigerT;

typedef struct{                 /* Rest der KP-Zelle, falls es sich um einen Info-Block handelt */
   long int ElternNummer;
   RegelOderGleichungsT re;
   unsigned long int AnzInitialerKPs;
} InfoRestT;



typedef struct{                 /* Rest der KP-Zelle, falls es sich um ein KP handelt */
   /* Eintraege fuer Heap-Verwaltung */
   ZellenZeigerT 
     PHBruder,           /* Geschwister*/
     PHNachfolger,       /* Nachfolger */
     PHnextOne,          /* fuer zu loeschende oder verschiebende Eintraege waehrend des Heap-Durchlauf */
     PHVater;            /* Vorgaenger */
   /* Eintraege fuer verkettete Struktur */
   ZellenZeigerT MutterVorg, MutterNachf, Mutter;  /* Der Verweis auf den Muttereintrag hat wegen der union denselben Typ */
   TermpairT tp;
   int w1, w2;
   void *addref;
} KpzRestT;

typedef struct KPZellenTS {
   ZellenZeigerT VaterVorg, VaterNachf, Vater;  /* Der Verweis auf den Vatereintrag hat wegen der union denselben Typ */
   union{
      InfoRestT InfoBlock;
      KpzRestT KPZelle;
   } Variante;
}KPZellenT;



/* ----------------Fuers Allokieren: ----------------*/
typedef struct{                 /* Rest der KP-Zelle, falls es sich um einen Info-Block handelt */
   long int ElternNummer;
   RegelOderGleichungsT re;
   unsigned long int AnzInitialerKPs;
} InfoRestT_Stat;

typedef struct{                 /* Rest der KP-Zelle, falls es sich um einen Info-Block handelt */
   long int ElternNummer;
   RegelOderGleichungsT re;
} InfoRestT_NoStat;



typedef struct{
   ZellenZeigerT VaterVorg, VaterNachf, Vater;
   InfoRestT_Stat InfoBlock;
} _IBT_Stat;

typedef struct{
   ZellenZeigerT VaterVorg, VaterNachf, Vater;
   InfoRestT_NoStat InfoBlock;
} _IBT_NoStat;

typedef struct{                 /* Rest der KP-Zelle, falls es sich um ein KP handelt */
   ZellenZeigerT 
     PHBruder,           /* Geschwister*/
     PHNachfolger,       /* Nachfolger */
     PHnextOne,          /* fuer zu loeschende oder verschiebende Eintraege waehrend des Heap-Durchlauf */
     PHVater;            /* Vorgaenger */
   ZellenZeigerT MutterVorg, MutterNachf, Mutter;  /* Der Verweis auf den Muttereintrag hat wegen der union denselben Typ */
   TermpairT tp;
   long int w1;
} KpzRestT_0;

typedef struct{                 /* Rest der KP-Zelle, falls es sich um ein KP handelt */
   ZellenZeigerT 
     PHBruder,           /* Geschwister*/
     PHNachfolger,       /* Nachfolger */
     PHnextOne,          /* fuer zu loeschende oder verschiebende Eintraege waehrend des Heap-Durchlauf */
     PHVater;            /* Vorgaenger */
   ZellenZeigerT MutterVorg, MutterNachf, Mutter;  /* Der Verweis auf den Muttereintrag hat wegen der union denselben Typ */
   TermpairT tp;
   long int w1, w2;
} KpzRestT_1;

typedef struct{                 /* Rest der KP-Zelle, falls es sich um ein KP handelt */
   ZellenZeigerT 
     PHBruder,           /* Geschwister*/
     PHNachfolger,       /* Nachfolger */
     PHnextOne,          /* fuer zu loeschende oder verschiebende Eintraege waehrend des Heap-Durchlauf */
     PHVater;            /* Vorgaenger */
   ZellenZeigerT MutterVorg, MutterNachf, Mutter;  /* Der Verweis auf den Muttereintrag hat wegen der union denselben Typ */
   TermpairT tp;
   long int w1, w2;
   void *addref;
} KpzRestT_2;

typedef struct{
   ZellenZeigerT VaterVorg, VaterNachf, Vater;
   KpzRestT_0 KPZelle;
} _KPZT_0;

typedef struct{
   ZellenZeigerT VaterVorg, VaterNachf, Vater;
   KpzRestT_1 KPZelle;
} _KPZT_1;

typedef struct{
   ZellenZeigerT VaterVorg, VaterNachf, Vater;
   KpzRestT_2 KPZelle;
} _KPZT_2;


#define VaterREVon(KPZ) (KPZ->Vater->Variante.InfoBlock.re)
#define MutterREVon(KPZ) ((KPZ->Variante.KPZelle.Mutter == KPZ) ?                                               \
                            (KPZ->Vater->Variante.InfoBlock.re):                                                \
                            (KPZ->Variante.KPZelle.Mutter->Variante.InfoBlock.re))


/*-------------------------------------------------------------------------------------------------------------*/
/*-- 3) Globale Variablen -------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------*/

static ZellenZeigerT PHeap;

static SV_SpeicherManagerT KPZellenManager, InfoManager;

static long VerwendeteSaat;

static int WeightEntryStrukturTyp;

/*-------------------------------------------------------------------------------------------------------------*/
/*-- 5) Der Pairing-Heap --------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------*/

/*------------------------------ Makros fuer Lesbarkeit -------------------------------------------------------*/

#define PHNachf(/*ZellenZeigerT*/ heap) (heap->Variante.KPZelle.PHNachfolger)

#define PHBruder(/*ZellenZeigerT*/ heap) (heap->Variante.KPZelle.PHBruder)

#define PHVater(/*ZellenZeigerT*/ heap) (heap->Variante.KPZelle.PHVater)

#define PHnextOne(/*ZellenZeigerT*/ heap) (heap->Variante.KPZelle.PHnextOne)

#define Gewicht(/*ZellenZeigerT*/ heap) (heap->Variante.KPZelle.w1)

/*------------------------------ Verschmelzen -----------------------------------------------------------------*/

/* Eigentliches Verschmelzen. Beide Heaps sind nicht leer, und das Gewicht der Wurzel von Heap1 ist kleiner als
   das der Wurzel von Heap2. 
   Ferner gilt: Vater(Heap1) == Vater(Heap2) == Bruder(Heap1) == Bruder(Heap2) == NULL.
   Rueckgabe ist ein gueltiger Pairing-Heap, falls Heap1 und Heap2 gueltig waren!
   Ansonsten bleiben der Vater- und der Bruder-Zeiger von Heap1 ungueltig, falls sie dies vorher schon waren*/


#define C_Groesser_0(e1, e2) (e1->Variante.KPZelle.w1 > e2->Variante.KPZelle.w1)

#define C_Groesser_1(e1,e2)                                                                                             \
  ((e1->Variante.KPZelle.w1 > e2->Variante.KPZelle.w1) ||                                                               \
   ((e1->Variante.KPZelle.w1 == e2->Variante.KPZelle.w1) && (e1->Variante.KPZelle.w2 > e2->Variante.KPZelle.w2)))

#define C_Groesser_2(e1, e2)                                                                                        \
  ((e1->Variante.KPZelle.w1 > e2->Variante.KPZelle.w1) ||                                                           \
   ((e1->Variante.KPZelle.w1 == e2->Variante.KPZelle.w1) && (e1->Variante.KPZelle.w2 > e2->Variante.KPZelle.w2)) || \
   ((e1->Variante.KPZelle.w1 == e2->Variante.KPZelle.w1) && (e1->Variante.KPZelle.w2 == e2->Variante.KPZelle.w2) && \
    (C_AdditionalComparison (e1->Variante.KPZelle.w1, e1->Variante.KPZelle.w2, e1->Variante.KPZelle.addref,         \
                             VaterREVon(e1), MutterREVon(e1),                                                       \
                             e1->Variante.KPZelle.tp,                                                               \
                             e2->Variante.KPZelle.w1, e2->Variante.KPZelle.w2, e2->Variante.KPZelle.addref,         \
                             VaterREVon(e2), MutterREVon(e2),                                                       \
                             e2->Variante.KPZelle.tp) == 1)))



#define C_Gleich_0(e1, e2) (e1->Variante.KPZelle.w1 == e2->Variante.KPZelle.w1)

#define C_Gleich_1(e1, e2) ( (e1->Variante.KPZelle.w1 == e2->Variante.KPZelle.w1) &&                                            \
                            (e1->Variante.KPZelle.w2 == e2->Variante.KPZelle.w2))

#define C_Gleich_2(e1, e2)                                                                                         \
  ((e1->Variante.KPZelle.w1 == e2->Variante.KPZelle.w1) && (e1->Variante.KPZelle.w2 == e1->Variante.KPZelle.w2) && \
   (C_AdditionalComparison(e1->Variante.KPZelle.w1,e1->Variante.KPZelle.w2,e1->Variante.KPZelle.addref,            \
                           VaterREVon(e1), MutterREVon(e1),                                                        \
                           e1->Variante.KPZelle.tp,                                                                \
                           e2->Variante.KPZelle.w1,e2->Variante.KPZelle.w2,e2->Variante.KPZelle.addref,            \
                           VaterREVon(e2), MutterREVon(e2),                                                        \
                           e2->Variante.KPZelle.tp) == 0))




#define DS_Groesser_0(e1, e2) TRUE

#define DS_Groesser_1(e1, e2) FALSE

#define DS_Groesser_2(e1, e2) (32768&rand())

#define DS_Groesser_3(e1, e2)                                                                                   \
(TPR_UnambiguousGreaterThan(e1->Variante.KPZelle.tp, VaterREVon(e1), MutterREVon(e1),                           \
                            e2->Variante.KPZelle.tp, VaterREVon(e2), MutterREVon(e2)))


#define /*void*/ PHMeld(/*ZellenZeigerT*/ Heap1, /*ZellenZeigerT*/ Heap2)                                       \
{                                                                                                               \
   PHBruder(Heap2) = PHNachf(Heap1);              /* Heap2 vor Nachfolgerliste von Heap1 haengen */             \
   PHNachf(Heap1) = Heap2;                                                                                      \
   PHVater(Heap2) = Heap1;                        /* und Vater von Heap2 besetzen */                            \
}
      

#define Verschmelze_BODY(/* {0,1,2} */ WeightEntryType, /* {0,1,2,3} */ DSVgl)                                  \
{                                                                                                               \
   if (C_Groesser_ ## WeightEntryType (Heap1, Heap2) ||                                                         \
       (DS_Groesser_ ## DSVgl (Heap1, Heap2) && C_Gleich_ ## WeightEntryType (Heap1, Heap2))){                  \
      PHMeld(Heap2, Heap1);                                                                                     \
      return Heap2;                                                                                             \
   }                                                                                                            \
   else{                                                                                                        \
     PHMeld(Heap1, Heap2);                                                                                      \
     return Heap1;                                                                                              \
   }                                                                                                            \
}      



/* Verschmilzt die beiden uebergebenen Heaps (beide nicht leer!) und gibt den Zeiger auf die neue Wurzel
   zuruck. Die Ausgangs-Heaps werden beim Melding zerstoert!
   Bei gleichem Gewicht: moeglichst wenige Operationen.*/

static ZellenZeigerT PHVerschmelze_0_egal1(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(0,0)

static ZellenZeigerT PHVerschmelze_0_egal2(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(0,1)

static ZellenZeigerT PHVerschmelze_0_random(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(0,2)

static ZellenZeigerT PHVerschmelze_0_unambiguous(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(0,3)


/* Verschmilzt die beiden uebergebenen Heaps (beide nicht leer!) und gibt den Zeiger auf die neue Wurzel
   zuruck. Die Ausgangs-Heaps werden beim Melding zerstoert!
   Bei gleichem Gewicht: moeglichst wenige Operationen.*/

static ZellenZeigerT PHVerschmelze_1_egal1(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(1,0)              
                                   
static ZellenZeigerT PHVerschmelze_1_egal2(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(1,1)             
                                   
static ZellenZeigerT PHVerschmelze_1_random(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(1,2)

static ZellenZeigerT PHVerschmelze_1_unambiguous(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(1,3)


/* Verschmilzt die beiden uebergebenen Heaps (beide nicht leer!) und gibt den Zeiger auf die neue Wurzel
   zuruck. Die Ausgangs-Heaps werden beim Melding zerstoert!
   Bei gleichem Gewicht: moeglichst wenige Operationen.*/

static ZellenZeigerT PHVerschmelze_2_egal1(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(2,0)              
                                   
static ZellenZeigerT PHVerschmelze_2_egal2(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(2,1)              
                                   
static ZellenZeigerT PHVerschmelze_2_random(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(2,2)

static ZellenZeigerT PHVerschmelze_2_unambiguous(ZellenZeigerT Heap1, ZellenZeigerT Heap2)
Verschmelze_BODY(2,3)




static ZellenZeigerT (*PHVerschmelzeProz) (ZellenZeigerT, ZellenZeigerT);

#define /*ZellenZeigerT*/ PHVerschmelze(/*ZellenZeigerT*/ Heap1, /*ZellenZeigerT*/ Heap2) ( (*PHVerschmelzeProz) (Heap1, Heap2))



static ZellenZeigerT PHeapOhneWurzelBinaer(ZellenZeigerT Heap)
/* Schmilzt die Nachfolgerliste eines Heaps zusammen und gibt als Resultat den entstehenden Heap zurueck.
   Im Nachfolger-Zeiger der urspruenglichen Wurzel steht ein Verweis auf den neu entstandenen Heap, ansonsten wird die urspr. Wurzel
   nicht weiter beeintraechtigt.
   Dabei werden die Eintraege der Nachfolgerliste paarweise verschmolzen (schichtweise).*/
{
  ZellenZeigerT haken = NULL, rest, *vorg;
  
  if (PHNachf(Heap)){                                                       /* >0 Nachfolger */
    while (PHBruder(PHNachf(Heap))){
      haken = PHNachf(Heap);
      vorg = &PHNachf(Heap);
      while(haken && PHBruder(haken)){
        rest = PHBruder(PHBruder(haken));
        haken = *vorg = PHVerschmelze(haken, PHBruder(haken));
        vorg = &PHBruder(haken);
        haken = PHBruder(haken) = rest;
      }
    }
    PHVater(PHNachf(Heap)) = NULL;
    PHBruder(PHNachf(Heap)) = NULL; /* Wurzel des 'verschmolzenen' Nachfolgerheaps gueltig machen. Vaterflag(haken) ist FALSE. */
  }
  return PHNachf(Heap);
}

static ZellenZeigerT PHeapOhneWurzelLinks(ZellenZeigerT Heap)
/* Schmilzt die Nachfolgerliste eines Heaps zusammen und gibt als Resultat den entstehenden Heap zurueck.
   Im Nachfolger-Zeiger der urspruenglichen Wurzel steht ein Verweis auf den neu entstandenen Heap, ansonsten wird die urspr. Wurzel
   nicht weiter beeintraechtigt.
   Dabei werde sukzessive alle Elemente der Nachfolgerliste mit dem am weitesten links stehenden verschmolzen.*/
{
  ZellenZeigerT haken=NULL, rest;
  
  if (PHNachf(Heap)){                    /* >0 Nachfolger */
    haken = PHNachf(Heap);
    while(PHBruder(haken)){
      rest = PHBruder(PHBruder(haken));
      haken = PHVerschmelze(haken, PHBruder(haken));
      PHBruder(haken) = rest;
    }
    PHNachf(Heap) = haken;
    PHVater(haken) = NULL;
    PHBruder(haken) = NULL; /* Wurzel des 'verschmolzenen' Nachfolgerheaps gueltig machen. Vaterflag(haken) ist FALSE. */
  }
  return haken;
}

static ZellenZeigerT RechtsRekVerschmelzen(ZellenZeigerT Knoten)
/* Verschmilzt rekursiv Knoten mit seinen Bruedern und gibt den Zeiger auf den entstandenen Heap zurueck. */
{
  if (!PHBruder(Knoten))
     return Knoten;              /* Knoten hat keine Brueder */
  else
     return PHVerschmelze(Knoten, RechtsRekVerschmelzen(PHBruder(Knoten))); /* Knoten hat noch genau einen Bruder */
}

static ZellenZeigerT PHeapOhneWurzelRechts(ZellenZeigerT Heap)
/* Schmilzt die Nachfolgerliste eines Heaps zusammen und gibt als Resultat den entstehenden Heap zurueck.
   Im Nachfolger-Zeiger der urspruenglichen Wurzel steht ein Verweis auf den neu entstandenen Heap, ansonsten 
   wird die urspr. Wurzel nicht weiter beeintraechtigt.
   Dabei werde sukzessive alle Elemente der Nachfolgerliste mit dem am weitesten links stehenden verschmolzen.*/
{
  if (PHNachf(Heap)){
    PHNachf(Heap) = RechtsRekVerschmelzen(PHNachf(Heap));
    PHVater(PHNachf(Heap)) = NULL;
    PHBruder(PHNachf(Heap)) = NULL;
    return PHNachf(Heap);
  }
  else
    return NULL;
}



static ZellenZeigerT (*PHNachlisteZusammenschmelzenProz) (ZellenZeigerT);

#define /*ZellenZeigerT*/ PHeapOhneWurzel(/*ZellenZeigerT*/ Heap) ((*PHNachlisteZusammenschmelzenProz) (Heap))


/*------------------------------ Loesen aus Struktur ----------------------------------------------------------*/

static void PHLoeseKnotenAusPHeap(ZellenZeigerT K)
/* Loest den angegebenen Knoten aus dem globalen Heap. Dabei wird PHeap destruktiv veraendert, Knoten bleibt
   erhalten. Die Verweise in der Struktur Knoten sind jedoch nicht mehr gueltig.
   Es wird davon ausgegangen, dass Knoten ein Element von PHeap ist.

   Zum Loesen eines Knotens aus der Struktur wird die Tatsache ausgenutzt, dass die jeweils linken
   Nachfolger eines Knotens (d.h. die Elemente der mit dem Verweis Nachfolger erreichbaren Knotenliste)
   groessere (oder gleiche) Schluessel als der Knoten selbst haben, also ohne weiteres in die Nachfolgerliste
   des Vaterknotens des zu loeschenden Knotens aufzunehmen sind. Beim Loesen wird dann im wesentlichen der Platz des
   Knotens in der Struktur durch seine Nachfolger aufgefuellt. Dadurch entfaellt ein moegliches Verschmelzen an dieser
   Stelle, da diese Operation (beim Waisenmord oder Neubewerten von KP's) wenn ueberhaupt, dann haeufig ausgefuehrt wird,
   und entsprechend haeufiges Verschmelzen noetig werden wuerde.

   Beim Ersetzen des Knotens sind folgende Sonderfaelle zu beachten:
   1. Der Knoten hat keine Nachfolger. Hat der Knoten noch Geschwister, so wird er einfach aus der Nachfolgerliste
      geloest, andernfalls wird die Nachfolgerliste des Vorgaengers ganz geloescht.
   2. Der Knoten ist der erste in der Nachfolgerliste seines Vorgaengers. Der entsprechende Verweis beim Vorgaenger
      muss geaendert werden.
   3. Der Knoten ist der letzte in der Nachfolgerliste seines Vorgaengers. Dann muss der Vorgaenger von K den Vater-Verweis
      aufnehmen.

   Fall 3 wird nicht explizit behandelt, sondern durch das 'Einpassen' des letzten Nachfolgers von K bzw. des Vorgaengers
   von K entschaerft.
*/
{
   if (PHVater(K))                                           /* K ist nicht die Wurzel von PHeap */
    
    if (PHNachf(PHVater(K)) == K)                           /* K ist der erste Eintrag in der Nachfolgerliste des Vaters */
      if (PHNachf(K)){                                      /* ... und hat Nachfolger */
        ZellenZeigerT temp = PHNachf(K), altTemp;
        
        PHNachf(PHVater(K)) = temp;                         /*     Nachfolger als ersten Nachfolger des Vaters einhaengen */
        
        do{                                                 /*     Nachfolger von K durchgehen und Vaterzeiger aendern */
          PHVater(temp) = PHVater(K);
          altTemp = temp;
          temp = PHBruder(temp);
        }while(temp);
        
        PHBruder(altTemp) = PHBruder(K);                    /*     Letzten Nachfolger von K (altTemp) */
                                                            /*     anstelle von K in die Nachfolgerliste von Vater einpassen */
      }
      else                                                  /* ... und hat keine Nachfolger */
        PHNachf(PHVater(K)) = PHBruder(K);                  /*     Nachfolger des Vaters = Bruder von K (ggf. NULL) */
  
    else{                                                   /* K ist nicht erster Nachfolger von Vater */
      ZellenZeigerT temp = PHNachf(PHVater(K));
      
      while(PHBruder(temp) != K)                            /*     Vorgaenger von K suchen */
        temp = PHBruder(temp);                              /*     -> temp */

      if (PHNachf(K)){                                      /* ... K hat Nachfolger */
        PHBruder(temp) = PHNachf(K);                        /*     an Vorgaenger von K anhaengen, */
        
        while(PHBruder(temp)){                              /*     bei allen Nachfolgern Vater-Zeiger aendern */
          temp = PHBruder(temp);
          PHVater(temp) = PHVater(K);
        }                                                   /*     temp steht jetzt auf dem letzten Nachfolger von K */
      }
      PHBruder(temp) = PHBruder(K);                         /* letzten Nachfolger von K oder Vorgaenger von K einpassen anstelle von K */
    }
  else                                                      /* K ist Wurzel von Heap */
    PHeap = PHeapOhneWurzel(PHeap);
}


/*-------------------------------------------------------------------------------------------------------------*/
/*------------------------------ Einfuegen --------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------*/

static void PHElementEinfuegen(ZellenZeigerT Element)
/* eine bereits angelegte und mit allen ausser den fuer die Pairing-Heap-Struktur noetigen Werten besetzte Zelle
   in den Pairing-Heap PHeap aufnehmen.*/
{
   PHBruder(Element)= PHNachf(Element)= PHVater(Element)=NULL;
   if (PHeap)
      PHeap = PHVerschmelze(Element, PHeap);                /* Mit Ausgangs-Heap verschmelzen, falls dieser nicht leer ist */
   else
      PHeap = Element;                                      /* Falls Ausgangs-Heap leer ist, neuen Element als Wurzel einsetzen. */
}



/* ------------------------------------------------------*/
/* -------------- Testen --------------------------------*/
/* ------------------------------------------------------*/

static void spaces(int n)
{   int i;
    for(i=0; i<n; i++)
      fprintf(IO_Strom, " ");
}

static void DruckeKnoten(ZellenZeigerT Knoten, int einruecktiefe)
{
  spaces(einruecktiefe);
  TPR_PrintTermpair(Knoten->Variante.KPZelle.tp, Knoten->Vater->Variante.InfoBlock.re, 
                    Knoten->Variante.KPZelle.Mutter->Variante.InfoBlock.re, " = ");
  IO_DruckeFlex("Gewicht %l, ",Knoten->Variante.KPZelle.w1 );
  IO_DruckeFlex("Vater %l, ", TP_ElternNr(Knoten->Vater->Variante.InfoBlock.re));
  IO_DruckeFlex("Mutter %l, ", TP_ElternNr(Knoten->Vater->Variante.InfoBlock.re));
  IO_DruckeFlex("Bruder %l, \n", TP_ElternNr(Knoten->Variante.KPZelle.Mutter->Variante.InfoBlock.re));
}

static void DruckePairingHeap(ZellenZeigerT wurzel, int einruecktiefe)
{
   ZellenZeigerT runny;

   if (wurzel){
/*      spaces(einruecktiefe);*/
      IO_DruckeAdresse(wurzel);
      DruckeKnoten(wurzel, einruecktiefe);
      runny = PHNachf(wurzel);
      if (runny){
         while(PHBruder(runny)){
            DruckePairingHeap(runny, einruecktiefe+5);
            runny=PHBruder(runny);
         }
         DruckePairingHeap(runny, einruecktiefe+5);
      }
/*      else{
         spaces(einruecktiefe+5);
         fprintf(IO_Strom, "---Keine Nachfolger---\n");
      }*/
   }
}

#define KPP_TEST 0
/* Alternativen: 0 fuer aus, 1 fuer an */
#if KPP_TEST
unsigned long int testNr = 0;

#define HeapFehler(str, nr1, nr2) printf("%s (%ld, %ld)\n", str, nr1, nr2)

#define ElterOk(el)                                                                                                                     \
{                                                                                                                                       \
   if (!(el))                                                                                                                           \
      HeapFehler("lebendem Elternteil anhaengende re nicht besetzt",0L,0L);                                                             \
                                                                                                                                        \
   if (((el)->Variante.InfoBlock.re)){                                                                                                  \
      if ((el)->Variante.InfoBlock.re->Nummer != (el)->Variante.InfoBlock.ElternNummer)                                                 \
         HeapFehler("Elternteil mit falscher Nummer",(el)->Variante.InfoBlock.re->Nummer , (el)->Variante.InfoBlock.ElternNummer);      \
                                                                                                                                        \
      if (RE_IstRegel((el)->Variante.InfoBlock.re)){                                                                                    \
         if ((el)->Variante.InfoBlock.ElternNummer > RE_AnzRegelnGesamt())                                                              \
            HeapFehler("Regel mit falscher Nummer", (el)->Variante.InfoBlock.ElternNummer, RE_AnzRegelnGesamt());                       \
                                                                                                                                        \
      }                                                                                                                                 \
      else{                                                                                                                             \
         if ((el)->Variante.InfoBlock.ElternNummer < -RE_AnzGleichungenGesamt())                                                        \
            HeapFehler("Gleichung mit falscher Nummer",(el)->Variante.InfoBlock.ElternNummer,  -RE_AnzGleichungenGesamt());             \
                                                                                                                                        \
      }                                                                                                                                 \
   }                                                                                                                                    \
}

static void TestDurchlauf(char *str)
/* Diese Prozedur testet den P-Heap auf Korrektheit.*/
{
   ZellenZeigerT runny, runny2;
   printf("********************** Heap-Test %ld  (%s)\n", ++testNr, str);
   if (PHeap){
      runny = PHeap;
      if (PHVater(PHeap) || PHBruder(PHeap)){
         fprintf(IO_Strom, "\nFEHLER im Pairing-Heap: Wurzel.\n");
         DruckePairingHeap(PHeap, 0);
      }
      while(runny){
         runny2 = PHNachf(runny);
         while(runny2){
            if (runny2 != runny2->Variante.KPZelle.Mutter) ElterOk(runny2->Variante.KPZelle.Mutter);
            ElterOk(runny2->Vater);
            if (PHVater(runny2) != runny){
               fprintf(IO_Strom, "\nFEHLER im Pairing-Heap: Vater-Zeiger stimmt nicht mit Vater ueberein.\n");
               fprintf(IO_Strom, "\nrunny: ");
               DruckeKnoten(runny, 0);
               fprintf(IO_Strom, "\nrunny2:");
               DruckeKnoten(runny2, 0);
               fprintf(IO_Strom, "\n");
               DruckePairingHeap(PHeap, 0);
            }
            if (Gewicht(runny2) < Gewicht(runny)){
               fprintf(IO_Strom, "\nFEHLER im Pairing-Heap: Gewichte.\n");
               DruckePairingHeap(PHeap, 0);
            }
            runny2 = PHBruder(runny2);
         }
         if (runny!=runny->Variante.KPZelle.Mutter) ElterOk(runny->Variante.KPZelle.Mutter);
         ElterOk(runny->Vater);
         if (PHNachf(runny))
            runny = PHNachf(runny);
         else if (PHBruder(runny)) /* keine Nachfolger -> Bruderliste durchgehen*/
            runny = PHBruder(runny);
         else{                  /* auch keine Brueder -> solange aufsteigen, bis Wurzel oder Knoten mit Bruedern gefunden*/
            while(runny && !PHBruder(runny))
               runny = PHVater(runny);
            if (runny)
               runny = PHBruder(runny); /* Im Wurzel-Casus dann runny==NULL */
         }
      }
   }
}

#else
#define TestDurchlauf(strPtr)
#endif


/* ------------------------------------------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------------*/
/*  II) Realisierung der KP-Menge                                                                              */
/*                                                                                                             */
/*  Die KP-Menge selbst wird mit Hilfe des Pairing-Heaps so realisiert, dass jederzeit auch von den Eltern     */
/*  eines kritischen Paares aus auf dieses Zugriff moeglich ist. Damit laesst sich einfach der vorzeitige      */
/*  Waisenmord realisieren.                                                                                    */
/*                                                                                                             */
/* ------------------------------------------------------------------------------------------------------------*/
/* ------------------------------------------------------------------------------------------------------------*/

/* Fuer den Modus Waisenmord, dh. vorzeitigen Kindermord, ist es erforderlich, alle
   kritischen Paare, an denen eine Regel oder Gleichung beteiligt ist, jederzeit
   erreichen zu koennen. Insbesondere muessen diese kritischen Paare ohne groessere
   Probleme aus der gewaehlten Sortierung geloest werden koennen. Um diesen Anforderungen
   zu genuegen, werden kritische Paare in zwei verschiedene Strukturen eingebettet:
   1. Einen Pairing-Heap, der die Sortierung gemaess der eingestellten Heuristik vornimmt
      und ferner ein effizientes Entfernen eines beliebigen - also nicht nur des kleinsten -
      Elementes ermoeglicht (genaueres siehe oben). 
   2. Eine Struktur, die zu jeder Regel und Gleichung einen Ring vorsieht, der alle KPs
      enthaelt, die zu dieser Regel oder Gleichung gebildet wurden und noch in der KP-Menge
      enthalten sind. Jedes KP ist deshalb in zwei solche doppelt verketteten Ringe eingebettet,
      die die KPs zu jeweils einem der beiden Elternteile enthalten. Um die Verkettungen unterscheiden
      zu koennen, wird zu den jeweiligen Zeigern noch ein Verweis auf die entsprechende Elternregel
      oder -gleichung gespeichert. Ein Sonderfall ist die Behandlung von KPs, die nur einen Elternteil
      haben. Bei solchen KPs wird die Vater-Verkettung 'ganz normal' eingezogen, die Mutter-Verkettung
      jedoch zeigt in beiden Richtungen auf das KP selbst. Ebenso ist der Mutter-Verweis auf das KP selbst
      gerichtet, um Verwechslungen auszuschliessen.
*/

/* --------------- Loeschen aus Ketten -------------*/

static void LoeseKPZelleAusVaterKette(ZellenZeigerT kpz)
/* Loest die angegebene KP-Zelle aus der Vater-bezogenen Verkettung. */
{
   if (kpz->VaterNachf->Vater == kpz->Vater)
      kpz->VaterNachf->VaterVorg = kpz->VaterVorg;
   else
      kpz->VaterNachf->Variante.KPZelle.MutterVorg = kpz->VaterVorg;
   if (kpz->VaterVorg->Vater == kpz->Vater)
      kpz->VaterVorg->VaterNachf = kpz->VaterNachf;
   else
      kpz->VaterVorg->Variante.KPZelle.MutterNachf = kpz->VaterNachf;
}

static void LoeseKPZelleAusMutterKette(ZellenZeigerT kpz)
/* Loest die angegebene KP-Zelle aus der Mutter-bezogenen Verkettung. */
{
   if (kpz->Variante.KPZelle.MutterNachf->Vater == kpz->Variante.KPZelle.Mutter)
      kpz->Variante.KPZelle.MutterNachf->VaterVorg = kpz->Variante.KPZelle.MutterVorg;
   else
      kpz->Variante.KPZelle.MutterNachf->Variante.KPZelle.MutterVorg = kpz->Variante.KPZelle.MutterVorg;
   if (kpz->Variante.KPZelle.MutterVorg->Vater == kpz->Variante.KPZelle.Mutter)
      kpz->Variante.KPZelle.MutterVorg->VaterNachf = kpz->Variante.KPZelle.MutterNachf;
   else
      kpz->Variante.KPZelle.MutterVorg->Variante.KPZelle.MutterNachf = kpz->Variante.KPZelle.MutterNachf;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Insert ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define GewichteEintragen_0(elt, W1, W2, ADDREF) elt->Variante.KPZelle.w1 = W1
#define GewichteEintragen_1(elt, W1, W2, ADDREF) elt->Variante.KPZelle.w1 = W1, elt->Variante.KPZelle.w2 = W2
#define GewichteEintragen_2(elt, W1, W2, ADDREF) elt->Variante.KPZelle.w1 = W1; elt->Variante.KPZelle.w2 = W2;                \
                                                  elt->Variante.KPZelle.addref = ADDREF

#define INSERT_BODY(tp, w1, w2, addref, father,  mother, /*{0,1,2}*/ WeightEntryType, /*{Stat, NoStat}*/ statisticsMode) \
/* Fuegt das kritische Paar kp in die KP-Menge ein, mit Verweisen auf die angegebenen Eltern und gemaess dem                    \
   angegebenen Gewicht.                                                                                                         \
*/                                                                                                                              \
{                                                                                                                               \
   KPZellenT *kpz;                                                                                                              \
   KPZellenT *Vater = TP_SUEInfo(actualParent);                                                                                 \
   KPZellenT *Mutter = TP_SUEInfo(otherParent);                                                                                 \
                                                                                                                                \
   IncAnzKPEingefuegt_ ## statisticsMode();                                                                                     \
   IncAnzInitialerKPs_ ## statisticsMode( !IsDemigod(otherParent), (Vater->Variante.InfoBlock.AnzInitialerKPs));           \
                                                                                                                                \
   TestDurchlauf("Anfang INSERT");                                                                                              \
   ZM_StartKPEinf();                                                                                                            \
   SV_Alloc(kpz,KPZellenManager,KPZellenT *);                           /* Anlegen und KP einsetzen */                          \
   kpz->Variante.KPZelle.tp = tp;                                                                                               \
   kpz->Vater = Vater;                                                                                                          \
                                                                                                                                \
   kpz->VaterNachf = Vater->VaterNachf;       /* Einsetzen in Vater-Verkettung */                                               \
   kpz->VaterVorg = Vater;                                                                                                      \
                                                                                                                                \
   if (kpz->VaterNachf->Vater == Vater)                                                                                         \
      kpz->VaterNachf->VaterVorg = kpz;                                                                                         \
   else                                                                                                                         \
      kpz->VaterNachf->Variante.KPZelle.MutterVorg = kpz;                                                                       \
                                                                                                                                \
   Vater->VaterNachf = kpz;                                                                                                     \
                                                                                                                                \
   if (Vater == Mutter){                                                                                                        \
      kpz->Variante.KPZelle.MutterNachf             /* Falls die Eltern gleich sind, muss die Konsistenz beim Loeschen */       \
         =  kpz->Variante.KPZelle.MutterVorg           /* gewaehrleistet werden (siehe dort und oben). */                       \
            = kpz->Variante.KPZelle.Mutter                                                                                      \
               = kpz;                                                                                                           \
   }                                                                                                                            \
   else{                                                                                                                        \
      kpz->Variante.KPZelle.Mutter = Mutter;                                                                                    \
      kpz->Variante.KPZelle.MutterNachf = Mutter->VaterNachf;         /* Einsetzen in Mutter-Verkettung */                      \
      kpz->Variante.KPZelle.MutterVorg = Mutter;                                                                                \
      if (kpz->Variante.KPZelle.MutterNachf->Vater == Mutter)                                                                   \
         kpz->Variante.KPZelle.MutterNachf->VaterVorg = kpz;                                                                    \
      else                                                                                                                      \
         kpz->Variante.KPZelle.MutterNachf->Variante.KPZelle.MutterVorg = kpz;                                                  \
      Mutter->VaterNachf = kpz;                                                                                                 \
   }                                                                                                                            \
   GewichteEintragen_ ## WeightEntryType (kpz, w1, w2, addref);                                                                 \
   PHElementEinfuegen(kpz);                                                                                                     \
   TestDurchlauf("Ende INSERT");                                                                                                \
   ZM_StoppKPEinf();                                                                                                            \
}


static void INSERT_0_Stat   (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, 0,0, actualParent, otherParent, 0, Stat)

static void INSERT_0_NoStat   (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, 0,0, actualParent, otherParent, 0, NoStat)


static void INSERT_1_Stat   (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, wt.w2, 0, actualParent, otherParent, 1, Stat)

static void INSERT_1_NoStat   (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, wt.w2, 0, actualParent, otherParent, 1, NoStat)


static void INSERT_2_Stat   (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, wt.w2, wt.addRef, actualParent, otherParent, 2, Stat)

static void INSERT_2_NoStat   (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
INSERT_BODY(tp, wt.w1, wt.w2, wt.addRef, actualParent, otherParent, 2, NoStat)


static BOOLEAN POSTPONE (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT actualParent, RegelOderGleichungsT otherParent)
{
   return FALSE;
}

/* KILLPARENT mit PP: 
       Im Postponing-Fall ist hier auf nichts weiter zu achten, da das zuletzt entfernte KP bereits beim
       Einfuegen der interreduzierenden Regel bzw. Gleichung physikalisch geloescht wird. Ferner koennen
       beim Waisenmord geloeschte KPs auch nicht zurueckgestellt werden, es muss also nur der KP-Mengen-
       Zaehler dekrementiert werden.
       Da aber beim finalen Loeschen der Regelmenge auch Waisenmord betrieben wird, kann dies ziemlich in die
       Hose gehen, deshalb muss das gemerkte KP auch an dieser Stelle geloescht werden!
*/


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Select ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define DESTROY_WEIGHT_0(addref)
#define DESTROY_WEIGHT_1(addref)
#define DESTROY_WEIGHT_2(addref) (C_KillAdditionalReference(addref))

#define ADDREF_BESETZEN_0(addrefPtr, addref) 
#define ADDREF_BESETZEN_1(addrefPtr, addref) 
#define ADDREF_BESETZEN_2(addrefPtr, addref) (*addrefPtr = addref) 

#define SELECT_BODY(tpPtr, actualParentPtr,  otherParentPtr, addrefPtrPtr, /* {0,1,2}*/ DestWeight,                             \
                    /*{Stat, NoStat}*/ statisticsMode)                                                                          \
/* Liefert das gemaess eingestellter Heuristik naechste zu behandelnde KP, das gleichzeitig                                     \
   aus der KP-Menge entfernt wird. Dabei wird die Statistik angepasst, und die entsprechenden Ausgaben durchgefuehrt.           \
   ACHTUNG: Nur aufrufen, wenn KPMenge noch gueltige KPs enthaelt.                                                              \
*/                                                                                                                              \
{                                                                                                                               \
   if (PHeap){                                                                                                                  \
      KPZellenT *alte = PHeap;                                                                                                  \
      TestDurchlauf("Anfang SELECT");                                                                                           \
      IncAnzKPEntfernt_ ## statisticsMode();                                                                                    \
      ZM_StartKPLoeschen();                                                                                                     \
      *tpPtr = PHeap->Variante.KPZelle.tp;                                                                                      \
      ADDREF_BESETZEN_ ## DestWeight (addrefPtr, PHeap->Variante.KPZelle.addref);                                               \
      *actualParentPtr = VaterREVon(PHeap);                                                                                     \
      *otherParentPtr = MutterREVon(PHeap);                                                                                     \
      LoeseKPZelleAusVaterKette(PHeap);                                                                                         \
      LoeseKPZelleAusMutterKette(PHeap);                                                                                        \
      PHeap = PHeapOhneWurzel(PHeap);                                                                                           \
      SV_Dealloc(KPZellenManager, alte);                                                                                        \
      ZM_StoppKPLoeschen();                                                                                                     \
      TestDurchlauf("Ende SELECT");                                                                                             \
      return TRUE;                                                                                                              \
   }                                                                                                                            \
   else                                                                                                                         \
      return FALSE;                                                                                                             \
}


#define SELECTINITIAL_BODY(tpPtr, actualParentPtr,  otherParentPtr, addrefPtrPtr, /* {0,1,2}*/ DestWeight,      \
                    /*{Stat, NoStat}*/ statisticsMode)                                                          \
{                                                                                                               \
  KPZellenT *ElternTeil = TP_SUEInfo(&SUE_Demigod);                                                             \
  KPZellenT *haken=ElternTeil->VaterNachf;                                                                      \
                                                                                                                \
  TestDurchlauf("Anfang SELECT-Initial");                                                                       \
  while(haken != ElternTeil){                                                                                   \
    if (haken->Variante.KPZelle.w1==LONG_MIN){                                                                  \
      ZM_StartKPLoeschen();                                                                                     \
      LoeseKPZelleAusMutterKette(haken);                                                                        \
      LoeseKPZelleAusVaterKette(haken);                                                                         \
      PHLoeseKnotenAusPHeap(haken);                                                                             \
      IncAnzKPEntfernt_##statisticsMode();                                                                      \
      *tpPtr = haken->Variante.KPZelle.tp;                                                                      \
      ADDREF_BESETZEN_##DestWeight (addrefPtr, haken->Variante.KPZelle.addref);                                 \
      *actualParentPtr = VaterREVon(haken);                                                                     \
      *otherParentPtr = MutterREVon(haken);                                                                     \
      SV_Dealloc(KPZellenManager, haken);                                                                       \
      ZM_StoppKPLoeschen();                                                                                     \
      return TRUE;                                                                                              \
    }                                                                                                           \
    else                                                                                                        \
      haken = haken->VaterNachf;                                                                                \
  }                                                                                                             \
  return FALSE;                                                                                                 \
}
  
#define SelectBauen(w)                                                                                  \
static BOOLEAN SELECT_##w##_Stat (TermpairT *tpPtr, RegelOderGleichungsT *actualParentPtr,              \
                                  RegelOderGleichungsT *otherParentPtr, void **addrefPtr)               \
SELECT_BODY(tpPtr, actualParentPtr,  otherParentPtr, addrefPtr, w, Stat)                                \
static BOOLEAN SELECT_##w##_NoStat (TermpairT *tpPtr, RegelOderGleichungsT *actualParentPtr,            \
                                    RegelOderGleichungsT *otherParentPtr, void **addrefPtr)             \
SELECT_BODY(tpPtr, actualParentPtr,  otherParentPtr, addrefPtr, w, NoStat)                              \
static BOOLEAN SELECTINITIAL_##w##_Stat (TermpairT *tpPtr, RegelOderGleichungsT *actualParentPtr,       \
                                         RegelOderGleichungsT *otherParentPtr, void **addrefPtr)        \
SELECTINITIAL_BODY(tpPtr, actualParentPtr,  otherParentPtr, addrefPtr, w, Stat)                         \
static BOOLEAN SELECTINITIAL_##w##_NoStat (TermpairT *tpPtr, RegelOderGleichungsT *actualParentPtr,     \
                                           RegelOderGleichungsT *otherParentPtr, void **addrefPtr)      \
SELECTINITIAL_BODY(tpPtr, actualParentPtr,  otherParentPtr, addrefPtr, w, NoStat)


SelectBauen(0)
SelectBauen(1)
SelectBauen(2)

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Parents ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void KPP_NEWPARENT(RegelOderGleichungsT re)
/* Legt einen neuen Info-Block an und gibt den Zeiger darauf zurueck. Dieser Info-Block
   enthaelt dann die angegebene Nummer des Elternteils, die auch wieder abgerufen werden kann (s.u.)*/
{
   KPZellenT *info;
   
   
   TestDurchlauf("Anfang NEWPARENT");
   SV_Alloc(info,InfoManager,KPZellenT *);
   info->VaterNachf = info->VaterVorg = info;
   info->Variante.InfoBlock.ElternNummer = TP_ElternNr(re);
   info->Vater = info;
   if (PQStatisticsAn) info -> Variante.InfoBlock.AnzInitialerKPs = 0;
   TP_SUEInfo(re) = (void *) info;
   info->Variante.InfoBlock.re = re;

   if (!TP_TermpaarIstRegel(re)){ /* dh. Gleichung */
      SV_Alloc(info,InfoManager,KPZellenT *);
      info->VaterNachf = info->VaterVorg = info;
      info->Variante.InfoBlock.ElternNummer = TP_ElternNr(re);
      info->Vater = info;
      if (PQStatisticsAn) info -> Variante.InfoBlock.AnzInitialerKPs = 0;
      TP_SUEInfo(TP_Gegenrichtung(re)) = (void *) info;
      info->Variante.InfoBlock.re = TP_Gegenrichtung(re);
   }
   TestDurchlauf("Ende NEWPARENT");
}

#define DESTROY_WEIGHT_0(addref)
#define DESTROY_WEIGHT_1(addref)
#define DESTROY_WEIGHT_2(addref) (C_KillAdditionalReference(addref))


#define Kill_BODY(/* {Stat, NoStat} */ statisticsMode, /* {0,1,2} */ WeightEntryType)   \
{                                                                                       \
   haken = ElternTeil->VaterNachf;                                                      \
   while(haken != ElternTeil){                                                          \
      if (haken->Vater == ElternTeil){                                                  \
         LoeseKPZelleAusMutterKette(haken);                                             \
         nachf = haken->VaterNachf;                                                     \
      }                                                                                 \
      else{                                                                             \
         LoeseKPZelleAusVaterKette(haken);                                              \
         nachf = haken->Variante.KPZelle.MutterNachf;                                   \
      }                                                                                 \
      IncAnzKPFrueherWaisenmord_ ## statisticsMode();                                   \
      IncAnzKPEntfernt_ ## statisticsMode();                                            \
      PHLoeseKnotenAusPHeap(haken);                                                     \
      TPR_DESTROYTERMPAIR(haken->Variante.KPZelle.tp,                                   \
                          haken->Vater->Variante.InfoBlock.re,                          \
                          (haken->Variante.KPZelle.Mutter == haken)?                    \
                          haken->Variante.KPZelle.Mutter->Variante.InfoBlock.re:        \
                          haken->Vater->Variante.InfoBlock.re);                         \
      DESTROY_WEIGHT_ ## WeightEntryType (haken->Variante.KPZelle.addref);              \
      SV_Dealloc(KPZellenManager, haken);                                               \
      haken = nachf;                                                                    \
   }                                                                                    \
}

#define KILL_Verteiler()                        \
{                                               \
   if (PQStatisticsAn){                         \
      switch (WeightEntryStrukturTyp){          \
      case 0:                                   \
         Kill_BODY(Stat,0)                      \
         break;                                 \
      case 1:                                   \
         Kill_BODY(Stat,1)                      \
         break;                                 \
      case 2:                                   \
      default:                                  \
         Kill_BODY(Stat,2)                      \
         break;                                 \
      }                                         \
   }                                            \
   else{                                        \
      switch (WeightEntryStrukturTyp){          \
      case 0:                                   \
         Kill_BODY(NoStat,0)                    \
         break;                                 \
      case 1:                                   \
         Kill_BODY(NoStat,1)                    \
         break;                                 \
      case 2:                                   \
      default:                                  \
         Kill_BODY(NoStat,2)                    \
         break;                                 \
      }                                         \
   }                                            \
}

void KPP_KILLPARENT(RegelOderGleichungsT re)
/* Meldet die Loeschung einer Regel oder Gleichung an die KP-Verwaltung weiter. Es werden
   alle zugehoerigen KP's geloescht.*/
{
   KPZellenT *haken, *nachf;
   KPZellenT *ElternTeil = TP_SUEInfo(re);
   
   TestDurchlauf("Anfang KILLPARENT");
   ZM_StartKPLoeschen();
   KILL_Verteiler();
   SV_Dealloc(InfoManager, ElternTeil);
   TP_SUEInfo(re) = (void *) NULL;

   if (!TP_TermpaarIstRegel(re)){ /* dh. Gleichung */
      ElternTeil = TP_SUEInfo(TP_Gegenrichtung(re));
      KILL_Verteiler();
      SV_Dealloc(InfoManager, ElternTeil);
      TP_SUEInfo(TP_Gegenrichtung(re)) = (void *) NULL;
   }

   ZM_StoppKPLoeschen();
   TestDurchlauf("Ende KILLPARENT");
}

void KPP_REDIRECTPARENT(void)
{
   ;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Walkthrough -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

unsigned long int KPP_AnzKPsToGo(void)
{
   return PQStatisticsAn?(AnzKPEingefuegt - AnzKPGeloescht) : 0;
}

#define TermeAuspacken_terms(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr) \
TPR_TERMPAIRtoPARALLEL_INTERMED(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr)
#define TermeAuspacken_noterms(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr) /* nix tun */

#define TermeWiederEinpacken_terms(tp, lhs, rhs, actualParent, otherParent, changed, pclID)\
TPR_PARALLELtoTERMPAIR(tp, lhs, rhs, actualParent, otherParent, changed, pclID)
#define TermeWiederEinpacken_noterms(tp, lhs, rhs, actualParent, otherParent, changed, pclID)  tp

#define TermeZerstoeren_terms(tp, lhs, rhs, actualParent, otherParent)                                                          \
TPR_DESTROYPARALLEL(tp, lhs, rhs, actualParent, otherParent)
#define TermeZerstoeren_noterms(tp, lhs, rhs, actualParent, otherParent) /* nix tun */

#define GewichteExtrahieren_0(wt, elt) (wt.w1 = elt.w1)
#define GewichteExtrahieren_1(wt, elt) (wt.w1 = elt.w1, wt.w2 = elt.w2)
#define GewichteExtrahieren_2(wt, elt) (wt.w1 = elt.w1, wt.w2 = elt.w2, wt.addRef = elt.addref)


#define WALKTHROUGH_BODY(/* {terms, noterms} */ TermeBenoetigt, /* {0, 1, 2}*/ WeightType, /*{Stat, NoStat}*/ statisticsMode)   \
/* Diese Prozedur wendet AP auf jedes Element des Heaps an. Liefert AP als Ergebnis WiegenUndVerschieben,                       \
   so wird das entsprechende Element mit WP gewogen und entsprechend des neuen Gewichts im Heap verschoben.                     \
   Liefert AP Entfernen, so wird das Element aus dem Heap geloescht.*/                                                          \
{                                                                                                                               \
   ZellenZeigerT runny, zuLoeschende = NULL, zuVerschiebende = NULL;                                                            \
                                                                                                                                \
   TestDurchlauf("Anfang WALKTHROUGH");                                                                                         \
   runny = PHeap;                                                                                                               \
   while(runny){                                                                                                                \
      int APresult;                                                                                                             \
      RegelOderGleichungsT otherParent;                                                                                         \
      TermT lhs=NULL, rhs=NULL;                                                                                                 \
      PCL_ObjektNummerT pclID;                                                                                                  \
                                                                                                                                \
      IncAnzKPIRBetrachtet_ ## statisticsMode();                                                                                \
      otherParent = (runny->Variante.KPZelle.Mutter == runny)?                                                                  \
             runny->Vater->Variante.InfoBlock.re: runny->Variante.KPZelle.Mutter->Variante.InfoBlock.re;                        \
      TermeAuspacken_ ## TermeBenoetigt(runny->Variante.KPZelle.tp, &lhs, &rhs,                                                 \
                                        runny->Vater->Variante.InfoBlock.re, otherParent, &pclID);                              \
                                                                                                                                \
      if ((APresult = (* AP) (lhs, rhs, runny->Vater->Variante.InfoBlock.re, otherParent,                                       \
                              &pclID)) == 1){                                                                                   \
         TermeZerstoeren_ ## TermeBenoetigt(runny->Variante.KPZelle.tp, lhs, rhs,                                               \
                                            runny->Vater->Variante.InfoBlock.re, otherParent);                                  \
         DESTROY_WEIGHT_ ## WeightType(runny->Variante.KPZelle.addref);                                                         \
         IncAnzKPEntfernt_ ## statisticsMode();IncAnzKPIRGeloescht_ ## statisticsMode();                                        \
         PHnextOne(runny) = zuLoeschende;                                                                                       \
         zuLoeschende = runny;                                                                                                  \
      }                                                                                                                         \
      else if (APresult == 2){                                                                                                  \
               WeightEntryT wt;                                                                                                 \
                                                                                                                                \
               PHnextOne(runny) = zuVerschiebende;                                                                              \
               zuVerschiebende = runny;                                                                                         \
               IncAnzKPIRVeraendert_ ## statisticsMode();                                                                       \
               GewichteExtrahieren_ ## WeightType(wt, runny->Variante.KPZelle);                                                 \
               C_RECLASSIFY(&wt, lhs, rhs, runny->Vater->Variante.InfoBlock.re, otherParent);                                   \
               runny->Variante.KPZelle.tp =                                                                                     \
                  TermeWiederEinpacken_ ## TermeBenoetigt(runny->Variante.KPZelle.tp, lhs, rhs,                                 \
                                                          runny->Vater->Variante.InfoBlock.re, otherParent, TRUE, pclID);       \
               GewichteEintragen_ ## WeightType(runny, wt.w1, wt.w2, wt.addRef);                                                \
      }                                                                                                                         \
      else{                                                                                                                     \
         runny->Variante.KPZelle.tp =                                                                                           \
                  TermeWiederEinpacken_ ## TermeBenoetigt(runny->Variante.KPZelle.tp, lhs, rhs,                                 \
                                                          runny->Vater->Variante.InfoBlock.re, otherParent, FALSE, pclID);      \
      }                                                                                                                         \
      if (PHNachf(runny))                                                                                                       \
         runny = PHNachf(runny);                                                                                                \
      else{                                                                                                                     \
         while(runny && !PHBruder(runny))     /* so lange aufsteigen, bis Knoten mit Bruder oder die Wurzel erreicht ist*/      \
            runny = PHVater(runny);                                                                                             \
         if (runny)                                                                                                             \
            runny = PHBruder(runny);           /* Im Wurzelfall ist dann runny == NULL (nach Konvention) */                     \
      }                                                                                                                         \
   }                                                                                                                            \
                                                                                                                                \
   while((runny = zuLoeschende)){           /* zu loeschende Eintraege killen */                                                \
      zuLoeschende = PHnextOne(runny);      /* naechsten Eintrag retten */                                                      \
      LoeseKPZelleAusVaterKette(runny);                                                                                         \
      LoeseKPZelleAusMutterKette(runny);                                                                                        \
      PHLoeseKnotenAusPHeap(runny);         /* Knoten aus Struktur loesen */                                                    \
      SV_Dealloc(KPZellenManager, runny);                                                                                       \
   }                                                                                                                            \
                                                                                                                                \
   runny = zuVerschiebende;                 /* zu verschiebende Eintraege zuerst aus Struktur entfernen */                      \
   while(runny){                                                                                                                \
      PHLoeseKnotenAusPHeap(runny);                                                                                             \
      PHNachf(runny) = NULL;                                                                                                    \
      PHBruder(runny)=NULL;                                                                                                     \
      PHVater(runny)=NULL;                                                                                                      \
      runny = PHnextOne(runny);                                                                                                 \
   }                                                                                                                            \
   runny = zuVerschiebende;                 /* ... und mit dem neuen Gewicht erneut in die Struktur einfuegen. */               \
   while(runny){                                                                                                                \
      if (PHeap)                                                                                                                \
         PHeap = PHVerschmelze(runny,PHeap);                                                                                    \
      else                                                                                                                      \
         PHeap = runny;                                                                                                         \
      runny = PHnextOne(runny);                                                                                                 \
   }                                                                                                                            \
   TestDurchlauf("Ende WALKTHROUGH");                                                                                           \
}

#define WALKTHROUGH_WEIGHT_BODY(/* {0, 1, 2}*/ WeightType)              \
{                                                                       \
   if (PQStatisticsAn)                                                  \
      if (NeedsTerms)  WALKTHROUGH_BODY(terms, WeightType, Stat)        \
      else             WALKTHROUGH_BODY(noterms, WeightType, Stat)      \
   else                                                                 \
      if (NeedsTerms)  WALKTHROUGH_BODY(terms, WeightType, NoStat)      \
      else             WALKTHROUGH_BODY(noterms, WeightType, NoStat)    \
}

static void WALKTHROUGH_0 (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGH_WEIGHT_BODY(0)

static void WALKTHROUGH_1 (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGH_WEIGHT_BODY(1)

static void WALKTHROUGH_2 (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGH_WEIGHT_BODY(2)

#define WW_BODY(/* {0, 1, 2}*/ WeightType)                                                                                      \
/* Diese Prozedur wendet AP auf jedes Element des Heaps an. Rueckgabe ist TRUE, wenn der Heap nicht leer ist.*/                 \
{                                                                                                                               \
   ZellenZeigerT runny;                                                                                                         \
                                                                                                                                \
   runny = PHeap;                                                                                                               \
   while(runny){                                                                                                                \
      WeightEntryT wt;                                                                                                          \
      GewichteExtrahieren_ ## WeightType(wt, runny->Variante.KPZelle);                                                          \
      (*AP)(wt,runny->Variante.KPZelle.tp, runny->Vater->Variante.InfoBlock.re,                                                 \
             ((runny->Variante.KPZelle.Mutter == runny)?                                                                        \
              (runny->Vater->Variante.InfoBlock.re) : (runny->Variante.KPZelle.Mutter->Variante.InfoBlock.re)));                \
      if (PHNachf(runny))                                                                                                       \
         runny = PHNachf(runny);                                                                                                \
      else if (PHBruder(runny))         /* keine Nachfolger -> Bruderliste durchgehen*/                                         \
         runny = PHBruder(runny);                                                                                               \
      else{                             /* auch keine Brueder -> solange aufsteigen, bis Wurzel oder Knoten mit Bruedern gef.*/ \
         while(!PHBruder(runny) && PHVater(runny))                                                                              \
            runny = PHVater(runny);                                                                                             \
         runny = PHBruder(runny);       /* Im Wurzel-Casus ist der Bruder nicht besetzt -> runny==NULL */                       \
      }                                                                                                                         \
   }                                                                                                                            \
}

void KPP_WATCHINGWALKTHROUGH(PQ_WatchingWalkthroughProcedureT AP)
{
   switch(WeightEntryStrukturTyp){
   case 1:
      WW_BODY(1);
      break;   
   case 2:     
      WW_BODY(2);
      break;   
   case 0:     
   default:    
      WW_BODY(0);
      break;
   }
}

#define TV_BODY(/* {0, 1, 2}*/ WeightType)                                                                                      \
/* Diese Prozedur wendet AP auf jedes Element des Heaps an. Rueckgabe ist TRUE, wenn der Heap nicht leer ist.*/                 \
{                                                                                                                               \
   ZellenZeigerT runny;                                                                                                         \
                                                                                                                                \
   runny = PHeap;                                                                                                               \
   while(runny){                                                                                                                \
      TPR_OperationWaehrendTermverfolgung(OperationAufTermen, runny->Variante.KPZelle.tp,                                       \
                                          runny->Vater->Variante.InfoBlock.re,                                                  \
                                          ((runny->Variante.KPZelle.Mutter == runny)?                                           \
                                           (runny->Vater->Variante.InfoBlock.re) :                                              \
                                           (runny->Variante.KPZelle.Mutter->Variante.InfoBlock.re)));                           \
      if (PHNachf(runny))                                                                                                       \
         runny = PHNachf(runny);                                                                                                \
      else if (PHBruder(runny))         /* keine Nachfolger -> Bruderliste durchgehen*/                                         \
         runny = PHBruder(runny);                                                                                               \
      else{                             /* auch keine Brueder -> solange aufsteigen, bis Wurzel oder Knoten mit Bruedern gef.*/ \
         while(!PHBruder(runny) && PHVater(runny))                                                                              \
            runny = PHVater(runny);                                                                                             \
         runny = PHBruder(runny);       /* Im Wurzel-Casus ist der Bruder nicht besetzt -> runny==NULL */                       \
      }                                                                                                                         \
   }                                                                                                                            \
}

void KPP_TermVerfolgung(void (* OperationAufTermen) (TermT, TermT))
{
   switch(WeightEntryStrukturTyp){
   case 1:
      TV_BODY(1);
      break;   
   case 2:     
      TV_BODY(2);
      break;   
   case 0:     
   default:    
      TV_BODY(0);
      break;
   }
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void KPP_STATISTICS(InfoStringT is)
{
   if (!(IS_IsIn("all", is) || (IS_IsIn("PQ", is))))
      return;                   /* Modul nicht angesprochen */
   if (IS_IsIn("standard", is)){
      IO_DruckeKurzUeberschrift("Priority-Queue: Pairing heap", "+");
      if (PQStatisticsAn)
         PQStatistik(is, AnzKPEingefuegt - AnzKPGeloescht);
      if (VerwendeteSaat != 0){
         fprintf(IO_Strom, "\n");
         IO_DruckeStatistikZeile("seed for random-comparison", VerwendeteSaat);
      }
   }
   else if(IS_IsIn("latex-head", is)){
      if (PQStatisticsAn)
         PQStatistik(is, 0);
   }
   else if (IS_IsIn("intern", is))
      ;
   else if(IS_IsIn("config", is)){
      fprintf(IO_Strom, "KPP: Statistik %s, ", PQStatisticsAn?" AN":"AUS");
#if KPP_TEST
      fprintf(IO_Strom, "Heap-Test-Modus  AN\n");
#else
      fprintf(IO_Strom, "Heap-Test-Modus AUS\n");
#endif
   }
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

static BOOLEAN KPP_SchonInitialisiert = FALSE;

#define INIT_BODY(/* {0,1,2} */ StrukturTyp)                                                                                  \
{                                                                                                                             \
   if (!KPP_SchonInitialisiert)\
     SV_ManagerInitialisieren(&KPZellenManager, sizeof(_KPZT_ ## StrukturTyp), PQ_AnzahlKPZellenJeBlock, "KP-Zellen (P-Heap)"); \
   PHeap = NULL;                                                                                                              \
   VerwendeteSaat = 0;                                                                                                        \
                                                                                                                              \
   switch (IS_FindOption(DEF, USR, 3, "bin", "left", "right")){                                                               \
   case 1:                                                                                                                    \
      fprintf(IO_Strom, "meld left, ");                                                                                       \
      PHNachlisteZusammenschmelzenProz = PHeapOhneWurzelLinks;                                                                \
      break;                                                                                                                  \
   case 2:                                                                                                                    \
      fprintf(IO_Strom, "meld right, ");                                                                                      \
      PHNachlisteZusammenschmelzenProz = PHeapOhneWurzelRechts;                                                               \
      break;                                                                                                                  \
   case 0:                                                                                                                    \
   default:                                                                                                                   \
      fprintf(IO_Strom, "meld binary, ");                                                                                     \
      PHNachlisteZusammenschmelzenProz = PHeapOhneWurzelBinaer;                                                               \
      break;                                                                                                                  \
   }                                                                                                                          \
                                                                                                                              \
   switch (IS_FindOption(DEF, USR, 4, "e1", "e2", "random", "unambiguous")){                                                  \
   case 0: fprintf(IO_Strom, "e1\n"); PHVerschmelzeProz = PHVerschmelze_ ## StrukturTyp ## _egal1; break;                     \
   case 1: fprintf(IO_Strom, "e2\n"); PHVerschmelzeProz = PHVerschmelze_ ## StrukturTyp ## _egal2; break;                     \
   case 2:                                                                                                                    \
   {                                                                                                                          \
      struct timezone tzp;                                                                                                    \
      struct timeval tp;                                                                                                      \
                                                                                                                              \
      PHVerschmelzeProz = PHVerschmelze_ ## StrukturTyp ## _random;                                                           \
      tp.tv_sec = 0;                                                                                                          \
      tp.tv_usec = 0;                                                                                                         \
      tzp.tz_minuteswest = 0;                                                                                                 \
      tzp.tz_dsttime = 0;                                                                                                     \
                                                                                                                              \
      if (gettimeofday(&tp, &tzp) == -1)                                                                                      \
         our_error("gettimeofday in KPHeap.c fehlgeschlagen.\n\n");                                                           \
      srand(VerwendeteSaat=tp.tv_usec);                                                                                       \
      fprintf(IO_Strom, "random (computed seed = %ld)\n", VerwendeteSaat); break;                                             \
   }                                                                                                                          \
   case 3: fprintf(IO_Strom, "unambiguous\n"); PHVerschmelzeProz = PHVerschmelze_ ## StrukturTyp ## _unambiguous; break;      \
   default:                                                                                                                   \
      if (IS_FindValuedOption(DEF, USR, 1, "random") == 0){                                                                   \
         PHVerschmelzeProz = PHVerschmelze_ ## StrukturTyp ## _random;                                                        \
         VerwendeteSaat = atoi(IS_ValueOfOption(DEF, USR, "random"));                                                         \
         fprintf(IO_Strom, "random (given seed = %ld)\n", VerwendeteSaat);                                                    \
         srand(VerwendeteSaat);                                                                                               \
      }                                                                                                                       \
      else{                                                                                                                   \
         PHVerschmelzeProz = PHVerschmelze_ ## StrukturTyp ## _egal1;                                                         \
         fprintf(IO_Strom, "e1\n");                                                                                           \
      }                                                                                                                       \
      break;                                                                                                                  \
   }                                                                                                                          \
}
#define Zuweisen(destweight)                                    \
{                                                               \
  KPP_WALKTHROUGH = WALKTHROUGH_##destweight;                   \
  if (PQStatisticsAn){                                          \
    KPP_INSERT = INSERT_##destweight##_Stat;                    \
    KPP_SELECT = SELECT_##destweight##_Stat;                    \
    KPP_SELECTINITIAL = SELECTINITIAL_##destweight##_Stat;      \
    KPP_POSTPONE = POSTPONE;                                    \
  }                                                             \
  else{                                                         \
    KPP_INSERT = INSERT_##destweight##_NoStat;                  \
    KPP_SELECT = SELECT_##destweight##_NoStat;                  \
    KPP_SELECTINITIAL = SELECTINITIAL_##destweight##_NoStat;    \
    KPP_POSTPONE = POSTPONE;                                    \
  }                                                             \
}

BOOLEAN KPP_INITIALIZE(InfoStringT DEF, InfoStringT USR, int WeightEntryType)
{
   char *statOption;

   WeightEntryStrukturTyp = WeightEntryType;
   
   if ((statOption = IS_ValueOfOption(DEF, USR, "statistics")))
      PQStatisticsAn = (strcmp(statOption, "on") == 0);
   else
      PQStatisticsAn = FALSE;
   
   if (PQStatisticsAn){
     if (!KPP_SchonInitialisiert)
       SV_ManagerInitialisieren(&InfoManager, sizeof(_IBT_Stat), PQ_AnzahlInfoBloeckeJeBlock, "KP-Infobloecke (Pairing-Heap Stat)");
     PQStatistikInitialisieren();
   }
   else
     if (!KPP_SchonInitialisiert)
       SV_ManagerInitialisieren(&InfoManager, sizeof(_IBT_NoStat), PQ_AnzahlInfoBloeckeJeBlock, "KP-Infobloecke (Pairing-Heap NoStat)");


   fprintf(IO_Strom, "Pairing Heap.\n  Weight-EntryType = %d. statistics = %s.\n  ",
           WeightEntryType, PQStatisticsAn ? "on":"off");

   switch (WeightEntryStrukturTyp){
   case 1:
      INIT_BODY(1);
      Zuweisen(1);
      break;
   case 2:
      INIT_BODY(2);
      Zuweisen(2);
      break;
   case 0:
   default:
      INIT_BODY(0);
      Zuweisen(0);
      break;
   }  
   KPP_INSERTImmortal = KPP_INSERT;
   KPP_SchonInitialisiert = TRUE;
   return TRUE;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Clean up --------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void KPP_SoftAbraeumen(void){;}

void KPP_Abraeumen(void)
/* Alles loeschen*/
{
  /* Diese Routine ist arbeitslos, weil saemtliche KPs beim Leeren der Regel- und
     Gleichungsmenge wie Waisen behandelt und somit entfernt werden.
   */
  ;
}  

void KPP_ActSize(unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl)
{
  *AktKPZahl = *AktGueltigeKPZahl = AnzKPEingefuegt - AnzKPGeloescht;
}
