/*--------------------------------------------------------------*/
/*      infer.c
/*
/*      Copyright Newborn Software 1996
/*
/*      functions:
/*         long binary_resolvents()    ................... #1
/*         long binary_factors()       ................... #2
/*         standardize()               ................... #3
/*         clear_var_substitution_tble()  ................ #4
/*--------------------------------------------------------------*/
 
#include "const.h"
#include "clause.h"

/*--------------------------------------------------------------*/
/* externally defined functions                                 */
/*--------------------------------------------------------------*/
 
extern unify();                     /* unify two literals       */
extern install_clause();            /* installation of clauses  */
extern clause *generate_clause();   /* generate a new clause    */
 
/*--------------------------------------------------------------*/
/* externally defined constants                                 */
/*--------------------------------------------------------------*/

extern long *var_substitution_tbl[];   /* variable subs table   */
 
extern long ply, iter_depth;
extern int o_nc;
 
/*--------------------------------------------------------------*/
/* internally defined functions                                 */
/*--------------------------------------------------------------*/

int binary_resolvents();
int binary_factors();
void standardize();
void clear_var_substitution_tble();

/*------------------------------------------------------------------*/
/*      binary_resolvents(c1,c2,m)         ................... #1
/*
/*      binary res is an inference rule that produces a new clause
/*      from two parent clauses.
/*
/*      Let C = {Li}n and D = {Mj}m (n,m = number of lit'ls in clause).
/*      If Li and Mj are unifiable with mgu
/*      s, then the binary resolvent is (C - {Li})s + (D - {Mj})s.
/*
/*      The first step in the resolution process is standardization of
/*      the variables in the two clauses. MAXVARS is maximum number of
/*      distinct vars in a clause. By convention, the variables of a
/*      clause start at either 0 or MAXVARS. Thus, we need only st'dize
/*      50% of the time (with luck, the two clauses will already be
/*      standardized).
/*
/*      Since variables of a clause are linked together, we need only
/*      follow the chain to st'dize the variables apart. Naturally, we
/*      use the nvar field so that we do the min work possible; i.e.
/*      we standardize the clause with the least number of variables.
/*
/*      parameters:     c1,c2 : clauses to resolve
/*                      m : [UNIT]|[MERGE]
/*
/*      called by:  calc_inferences()                search.c
/*
/*      calls:      standardize()                    bres.c
/*                  clear_var_substitution_tble()    unify.c
/*                  unify()                          unify.c
/*                  generate_clause()                gen.c
/*                  install_clause()                 search.c
/*
/*      returns:    # of clauses produced
/*------------------------------------------------------------------*/
 
int binary_resolvents(c1,c2,m)
clause *c1,*c2;
long m;
{
register literal *l1, *l2;
register long id1, id2;
clause *new_cl;
register int ix1, ix2;	
int numb_resolvents;
int one_standardize;
 
/*----------------------------------------------*/
/* check resolution restrictions                */
/*----------------------------------------------*/
 
id1 = c1->c_bits; id2 = c2->c_bits;
 
/*--------------------------------------------------------------*/
/* if extended-search opt. is on, and c2 is not a unit          */
/* clause, return                                               */
/*--------------------------------------------------------------*/
 
if(m & UNIT) {
   if((id2 & UNIT) == 0L) {
      return(0);}}			
		
/*--------------------------------------------------------------*/
/* if MERGE opt. is on, and neither c1 or c2 is a merge, return	*/
/*--------------------------------------------------------------*/
 
if(m & MERGE) {
   if((id2 & MERGE) == 0L) {
      if((id1 & MERGE) == 0L) {
         return(0);}}}
					
/*----------------------------------------------*/
/* if one of the clauses is not ACTIVE, return  */
/*----------------------------------------------*/
 
if((id1 & id2 & ACTIVE) == 0L) {
   return(0);}
 
/*----------------------------------------------*/
/* if at ply 1 and both clauses are equality    */
/* substitution axioms, don't form resolvent.   */
/* Also, if NC is on and neither clause         */
/* is in the NC, return                         */
/*----------------------------------------------*/
 
if(ply == 1L) {
   if((id1 & EQCLS) != 0L && (id2 & EQCLS) != 0L) {
      return(0);}
   if(o_nc) {
      if(((id1 | id2) & NC) == 0L) {
         return(0);}}}
 
standardize(c1,c2);                   /* standardize vars apart   */
 
numb_resolvents = 0;       /* count of the # of binary resolvents */
one_standardize = 0;
 
ix1 = 0;
for(l1 = c1->c_lit ; l1 != 0L ; l1 = l1->l_next) {
  if((l1->l_id & HHITMASK) == 0L) {
    ix1++;    
    ix2 = 0;
    for(l2 = c2->c_lit ; l2 != 0L ; l2 = l2->l_next) {
      if((l2->l_id & HHITMASK) == 0L) {
        ix2++;
        id1 = l1->l_id & IDMASK;          /* predicate IDs  */
        id2 = l2->l_id & IDMASK;
 
/*----------------------------------------------------------*/
/* if IDs are the same then we are in a predicate group.    */
/* we attempt unification if literals are complementary.    */
/*----------------------------------------------------------*/
 
        if(id1 == id2) {
          if((l1->l_id ^ l2->l_id) == NEGMASK) {
            if(one_standardize++ == 0) {
	       standardize(c1,c2);}  /* standardize vars apart   */
            clear_var_substitution_tble(c1->c_bits,c1->c_nv);
            clear_var_substitution_tble(c2->c_bits,c2->c_nv);
            if(unify(&l1->l_id,&l2->l_id)) {
              if((new_cl=generate_clause(c1,c2,l1,l2,ix1,ix2))!=0L) {
                if(install_clause(new_cl)) {
                  numb_resolvents++;}}}}}}}}}
										
return(numb_resolvents);
}
 
/*-------------------------------------------------------------*/
/*      binary_factors()      ................... #2
/*
/*      factoring is an inference rule that produces a new clause
/*      from a given non-unit cl. A typical example of factoring
/*      is the following:
/*
/*                     p(x) | p(y)
/*
/*      can be factored into p(x). What we do is attempt to find an
/*      mgu such that we collapse two literals of a clause into a
/*      singleton. A factored cl is a merge if parent is a merge
/*
/*      binary_factors will generate all factors of a given cause.
/*      Since the new clause will have only one parent, we use the
/*      left parent (and NULL the right parent).
/*
/*      parameters: c : non-unit clause to factor
/*
/*      called by:  calc_inferences()               search.c
/*
/*      calls:      clear_var_substitution_tble()   unify.c
/*                  unify()                         unify.c
/*                  generate_clause()               gen.c
/*                  install_clause()                search.c
/*
/*      returns:    the # of factors produced
/*-------------------------------------------------------------*/
 
int binary_factors(c)
clause *c;
{
register literal *l1, *l2;
clause *new_cl;
register int ix1, ix2;
int numb_factors;
	
/*---------------------------------------------------*/
/* if at ply 1 and nc strategy is ON and clause c    */
/* is not in the negated conclusion, return          */
/*---------------------------------------------------*/
 
if(o_nc) {  
   if(ply == 1L) {  
      if((c->c_bits & NC) == 0L)  {
         return(0L);}}}
 
if((c->c_bits & NC) == 0L && ply <= iter_depth)  {
   return(0);}
 
ix1 = ix2 = 0;
numb_factors = 0;                    /* # of factors produced  */
 
for(l1 = c->c_lit; l1->l_next != 0L; l1 = l1->l_next) {
   if((l1->l_id & HHITMASK) == 0L) {
      ix1++;
      ix2 = ix1;
      for(l2 = l1->l_next; l2 != 0L; l2=l2->l_next) {
         if((l2->l_id & HHITMASK) == 0L) {
            ix2++;
            if(l1->l_id != l2->l_id) continue;
            clear_var_substitution_tble(c->c_bits,c->c_nv);
            if(unify(&l1->l_id,&l2->l_id)) {    
               new_cl = generate_clause(c,0L,l1,0L,ix1,ix2);
               if(new_cl != 0L) {
                  if(install_clause(new_cl)) {
                     numb_factors++;}}}}}}}
return(numb_factors);
}
 
/*--------------------------------------------------------------*/
/*      standardize(c1,c2)            ................... #3
/*
/*      renames variables
/*
/*      called by:  bres()          bres.c
/*
/*      calls:      nobody
/*
/*      returns:    nothing
/*--------------------------------------------------------------*/
 
void standardize(c1,c2)
clause *c1,*c2;
{
register long i, a, *p;
literal *l;
 
/*---------------------------------------------*/
/* if variable in one clauses are numbered     */
/* beginning at 0 in one clauses and 200 in    */
/* the other, var are already standardized,    */
/* and the program should return               */
/*---------------------------------------------*/
 
if((c1->c_bits & VID) != (c2->c_bits & VID)) {
   return;}
 
/*---------------------------------------------*/
/* renumber the variables in clause c2         */
/*---------------------------------------------*/
 
if((c2->c_bits & VID) == 0L) {
   c2->c_bits |= VID;
   i = MAX_VARS;}        /* if now 0, add 200  */
else {
   c2->c_bits &= ~VID;
   i = -MAX_VARS;}      /* otherwise, sub 200  */
   
/*----------------------------------------------*/
/* now pass through variable chain while        */
/* renumbering variables                        */
/*----------------------------------------------*/
 
for(l = c2->c_lit ; l != 0L ; l = l->l_next) {
   p = &l->l_id;
   for(a = *p++ & ARGMASK; a > 0L ; a--) {
      if((*p & FCNMASK) != 0L) {
         a += *p++ & ARGMASK;}
      else {
         *p += i;
         p++;}}}
 
}
 
 
/*-------------------------------------------------------------*/
/*      clear_var_substitution_tble(i,n) .............. #4
/*
/*      vstclr is used to zero a section of the vst table (i.e.
/*      set each entry to NULL). bres, fact fxns call this fxn
/*      when they wants to perform a unification
/*
/*      calling arguments:
/*                  i = c_base (ptr to first variable in cl
/*                  n = c_nv (number of distinct var in cl
/*
/*      called by:  binary_resolutions()    infer.c
/*                  binary_factors()        infer.c
/*                  bres_ply_1()            elim.c
/*                  bres_replace()          elim.c
/*                  bres_link()             elim.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*--------------------------------------------------------------*/
 
void clear_var_substitution_tble(i,n)
long i,n;
{
register long **p, j;
 
if((i & VID) == 0L) {
   j = 0L;}
else {
   j = MAX_VARS;}
 
for(p = &var_substitution_tbl[j]; n > 0L; n--) {
   *p++ = 0L;}
}
 
