%###########################################################################
%# mod_replnum.pl
%# - replace numbers by atoms in clauses
%# - modification mode is line by line
%###########################################################################
%# predicates to be used from outside:
%#	- replnum/4
%#	- process_switch_replnum/1
%# dependent on
%#	- mylib.pl
%#	- clause info facts generated by siginfo.pl
%###########################################################################
%#        Max Moser                         Automated Reasoning Research Group
%# Email: moser@informatik.tu-muenchen.de     Lehrstuhl VIII f"ur Informatik
%# Tel.:  +49-89/521096			     Technische Universit"at M"unchen
%# Fax.:  +49-89/526502			         D-80290 M"unchen, Germany
%# http://wwwjessen.informatik.tu-muenchen.de/~moser
%###########################################################################

:- dynamic new_number_symbol/1.


%*****************************************************************************
%* replnum( clausewise, +Clause, +SwitchList, -NewClauseList )
%*****************************************************************************
%* - replace numbers in clauses by respective atoms
%* - every clause is mapped to a new clause, i.e. | NewClauseList | = 1
%* - up to now, no switches are necessary
%*****************************************************************************

replnum( clausewise, comment(Comment), _, [comment(Comment)] ).
	% well, comments don't need to be modified

replnum( clausewise, Clause, _, [NewClause] ) :-
	Clause = clause(Name,Type,LitList,Cons,Dict),
	replnum_expr( LitList, NewLitList ),
	NewClause = clause(Name,Type,NewLitList,Cons,Dict).

replnum( clausewise, Clause, _, [NewClause] ) :-
	Clause = rule(Name,Type,LitList,Cons,Dict),
	replnum_expr( LitList, NewLitList ),
	NewClause = rule(Name,Type,NewLitList,Cons,Dict).

replnum( clausewise, Clause, _, [NewClause] ) :-
	Clause = nomodclause(Name,Type,LitList,Cons,Dict),
	replnum_expr( LitList, NewLitList ),
	NewClause = nomodclause(Name,Type,NewLitList,Cons,Dict).

replnum( clausewise, Clause, _, [NewClause] ) :-
	Clause = nomodrule(Name,Type,LitList,Cons,Dict),
	replnum_expr( LitList, NewLitList ),
	NewClause = nomodrule(Name,Type,NewLitList,Cons,Dict).

replnum( clausewise, Clause, _, [Clause] ) :-
	writeln( 'Error: replnum: This case should not appear.' ).

%*****************************************************************************


%*****************************************************************************
%* replnum_expr( +Expr, -ReplExpr )
%*****************************************************************************
%* - replace numbers in clauses by respective atoms
%*****************************************************************************

replnum_expr( [], [] ).
		% empty expressions don't contain numbers

replnum_expr( [Expr|Rest], [Expr|ReplRest] ) :-
		% variables are not changed
	var( Expr ),
	!,
		% continue with rest of list
	replnum_expr( Rest, ReplRest ).

replnum_expr( [Expr|Rest], [NewExpr|ReplRest] ) :-
		% number found
	number( Expr ),
	!,
		% replace it by corresponding atom
	replace_number( Expr, NewExpr ),
		% continue with rest of list
	replnum_expr( Rest, ReplRest ).

replnum_expr( [Expr|Rest], [NewExpr|ReplRest] ) :-
		% complex expression found
	Expr =.. [Symb|Args],
		% consider arguments of complex expression
	replnum_expr( Args, ReplArgs ),
		% finally, reconstruct new expression
	NewExpr =.. [Symb|ReplArgs],
		% continue with rest of list
	replnum_expr( Rest, ReplRest ).

%*****************************************************************************


%*****************************************************************************
%* replace_number( +Number, -NumberAtom )
%*****************************************************************************
%* - return new atom for number
%*****************************************************************************

replace_number( Number, UniqueNumberAtom ) :-
		% add index Number to atom 'number'
	add_index_to_atom( 'number', Number, NumberAtom ),
		% get list of known function symbols
	functaritylist( FunctArityList ),
		% get list of all new symbols introduced by replnum
	findall( NumSym/0, new_number_symbol( NumSym ), NumSymArList ),
		% throw away newly introduced symbols 
	subtract( FunctArityList, NumSymArList, RemFunctArityList ),
		% add 'new_' to new symbol until unique in RemFunctArityList
	add_prefix_until_unique_in( NumberAtom, 'new_', RemFunctArityList,
							UniqueNumberAtom ),
		% delete old Number from FunctArityList if still there
	siginfo_del( functaritylist, Number/0 ),
		% remember new symbol in signature
	siginfo_add( functaritylist, UniqueNumberAtom/0 ),
		% remember another locally introduced symbol
	assert( new_number_symbol( UniqueNumberAtom ) ).

%*****************************************************************************


%*****************************************************************************
%* process_switch_replnum( +ArrangedSwitch )
%*****************************************************************************
%* - what is to be done if -replnum is found
%*****************************************************************************

process_switch_replnum( ["-replnum"] ) :-
	!,
	assert( cmd_arg(replnum,[]) ),
	writeln( '# *** option: replnum' ).

%*****************************************************************************

