/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                ANALYSIS OF CLAUSE SETS                 * */
/* *                                                        * */
/* *  $Module:   ANALYZE                                    * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.49 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/06/20 09:07:25 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: analyze.c,v $ */

#include "analyze.h"

/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

BOOL ana_PEQUATIONS;     /* Problem contains positive equations */
BOOL ana_NEQUATIONS;     /* Problem contains negative equations */
BOOL ana_FUNCTIONS;      /* Problem contains non-constant function symbols */
BOOL ana_PROP;           /* Problem contains propositional variables */
BOOL ana_GROUND;         /* Problem contains non-propositional, non-equational ground atoms */
BOOL ana_NONUNIT;        /* Problem contains non-unit clauses */
BOOL ana_MONADIC;        /* Problem contains non-ground monadic predicates */
BOOL ana_NONMONADIC;     /* Problem contains non-ground n-place predicates, n>1, not equality */
BOOL ana_SORTRES;        /* Problem contains literal not(S(x)) for some S */
BOOL ana_USORTRES;       /* Problem contains literal not(S(t)) for some S */
BOOL ana_FINDOMAIN;      /* Problem contains clause implying a finite domain */
BOOL ana_CONGROUND;      /* Conjecture is ground */

BOOL ana_PUREEQUATIONAL;    /* Problem is purely equational */
BOOL ana_PUREPROPOSITIONAL; /* Problem is purely propositional */

NAT  ana_AXIOMCLAUSES;   /* Number of axiom clauses */
NAT  ana_CONCLAUSES;     /* Number of conjecture clauses */
NAT  ana_NONHORNCLAUSES; /* Number of Non-Horn clauses */


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void ana_AnalyzeProblem(LIST Clauses)
/**************************************************************
  INPUT:   A list of clauses.
  RETURNS: Void.
  EFFECT:  Analyzes the clauses and sets the anlayze variables.
           Recomputes the weight for the clauses.
***************************************************************/
{
  CLAUSE Clause;

  ana_PEQUATIONS  = FALSE;             /* Defaults for properties */
  ana_NEQUATIONS  = FALSE; 
  ana_FUNCTIONS   = FALSE;
  ana_FINDOMAIN   = FALSE;
  ana_MONADIC     = FALSE;
  ana_NONMONADIC  = FALSE;
  ana_PROP        = FALSE;
  ana_GROUND      = FALSE;
  ana_SORTRES     = FALSE;
  ana_USORTRES    = FALSE;
  ana_NONUNIT     = FALSE;
  ana_CONGROUND   = TRUE;

  ana_AXIOMCLAUSES    = 0; 
  ana_CONCLAUSES      = 0;
  ana_NONHORNCLAUSES  = 0;

  if (list_Empty(Clauses))
    return;

  while(!list_Empty(Clauses)) {
    Clause = (CLAUSE)list_Car(Clauses);
    clause_SetWeight(Clause, clause_ComputeWeight(Clause));

    if (clause_GetFlag(Clause,CONCLAUSE))
      ana_CONCLAUSES++;
    else
      ana_AXIOMCLAUSES++;

    if (clause_NumOfSuccLits(Clause) > 1)
      ana_NONHORNCLAUSES++;

    if (ana_CONGROUND && clause_GetFlag(Clause,CONCLAUSE) &&
	clause_MaxVar(Clause) != symbol_GetInitialStandardVarCounter())
      ana_CONGROUND = FALSE;
    if (!ana_PEQUATIONS && clause_ContainsPositiveEquations(Clause)) {
      ana_PEQUATIONS = TRUE;
    }
    if (!ana_NEQUATIONS && clause_ContainsNegativeEquations(Clause)) {
      ana_NEQUATIONS = TRUE;
    }
    if (!ana_MONADIC || !ana_NONMONADIC || !ana_PROP || !ana_GROUND)
      clause_ContainsFolAtom(Clause,&ana_PROP,&ana_GROUND,&ana_MONADIC,&ana_NONMONADIC);

    if (!ana_FUNCTIONS && clause_ContainsFunctions(Clause)) {
      ana_FUNCTIONS = TRUE;
    }
    if (!ana_FINDOMAIN && clause_ImpliesFiniteDomain(Clause)) {
      ana_FINDOMAIN = TRUE;
    }
    if (!ana_NONUNIT && clause_Length(Clause) > 1) {
      ana_NONUNIT = TRUE;
    }
    if (!ana_SORTRES || !ana_USORTRES) 
      clause_ContainsSortRestriction(Clause,&ana_SORTRES,&ana_USORTRES);
    
    Clauses = list_Cdr(Clauses);
  }

  ana_PUREEQUATIONAL    = ((ana_PEQUATIONS || ana_NEQUATIONS) && !ana_MONADIC && !ana_NONMONADIC 
			   && !ana_PROP && !ana_GROUND);
  ana_PUREPROPOSITIONAL = (!ana_PEQUATIONS && !ana_NEQUATIONS &&!ana_MONADIC && !ana_NONMONADIC && ana_PROP);

  return;
}


void ana_Print(void)
{
  char* Horn;

  if (ana_NONHORNCLAUSES == 0)
    Horn = "Horn";
  else
    Horn = "Non-Horn";
  
  if (ana_MONADIC && !ana_NONMONADIC) {
    printf("\n This is a monadic %s problem",Horn);
    if (ana_PEQUATIONS || ana_NEQUATIONS)
      printf(" with equality.");
    else
      printf(" without equality.");
  }
  else
    if (ana_PEQUATIONS || ana_NEQUATIONS)
      if (ana_NONMONADIC || ana_MONADIC || ana_PROP)
	printf("\n This is a first-order %s problem containing equality.",Horn);
      else
	if (ana_NONUNIT)
	  printf("\n This is a pure equality %s problem.",Horn);
	else
	  printf("\n This is a unit equality problem.");
    else
      if (ana_NONMONADIC || ana_MONADIC)
	printf("\n This is a first-order %s problem without equality.",Horn);

  if (ana_PUREPROPOSITIONAL)
    printf("\n This is a propositional %s problem.",Horn);
  else
    if (ana_FINDOMAIN || !ana_FUNCTIONS) {
      printf("\n This is a problem that has, if any, a finite domain model.");
      if (ana_FINDOMAIN)
	printf("\n There is a finite domain clause.");
      if (!ana_FUNCTIONS)
	printf("\n There are no function symbols.");
    }
      
  
  if (ana_SORTRES)
    printf("\n This is a problem that contains sort information.");

  if (ana_CONGROUND && !ana_PUREPROPOSITIONAL)
    printf("\n The conjecture is ground.");

  printf("\n Axiom clauses: %d Conjecture clauses: %d",ana_AXIOMCLAUSES,ana_CONCLAUSES);

  flag_PrintInferenceRules();
  flag_PrintReductionRules();
  printf("\n Extras    : ");
  if (flag_GetFlagValue(flag_SATINPUT))
    printf("Input Saturation, ");
  else
    printf("No Input Saturation, ");
  if (flag_GetFlagValue(flag_SELECT) == 0)
    printf("No Selection, ");
  else
    if (flag_GetFlagValue(flag_SELECT) == 1)
      printf("Dynamic Selection, ");
    else
      printf("Always Selection, ");
  if (flag_GetFlagValue(flag_SPLITS) == -1)
    printf("Full Splitting, ");
  else
    if (flag_GetFlagValue(flag_SPLITS) == 0)
      printf("No Splitting, ");
    else
      printf("Maximum of %d Splits, ",flag_GetFlagValue(flag_SPLITS));
  if (flag_GetFlagValue(flag_FULLRED))
    printf("Full Reduction, ");
  else
    printf("Lazy Reduction, ");
  printf(" Ratio: %d, FuncWeight: %d, VarWeight: %d",flag_GetFlagValue(flag_WDRATIO),
	 flag_GetFlagValue(flag_FUNCWEIGHT), flag_GetFlagValue(flag_VARWEIGHT));
  printf("\n Precedence: ");
  fol_PrintPrecedence();
}

void ana_AutoConfiguration(LIST Clauses)
{
  LIST       Functions,Predicates,Scan;
  SYMBOL     Symbol;
  SIGNATURE  Entry;
  int        i;

  Functions  = list_Nil();            /* Set precedence */
  Predicates = list_Nil();

  for (i=1;i<symbol_ACTINDEX;i++) {
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)i)) {
      Entry  = symbol_Signature(i);
      Symbol = Entry->info;
      if (symbol_IsPredicate(Symbol))
	Predicates = list_Cons((POINTER)Symbol,Predicates);
      else
	if (symbol_IsFunction(Symbol))
	  Functions = list_Cons((POINTER)Symbol,Functions);
    }
  }
  Functions  = list_NumberSort(Functions,symbol_PositiveArity);
  Predicates = list_NumberSort(Predicates,symbol_PositiveArity); 
  
  for (Scan=Functions;!list_Empty(Scan);Scan=list_Cdr(Scan))
    symbol_SetOrdering((SYMBOL)list_Car(Scan),symbol_GetIncreasedOrderingCounter());

  for (Scan=Predicates;!list_Empty(Scan);Scan=list_Cdr(Scan))  /* Predicates < Functions */
    symbol_SetOrdering((SYMBOL)list_Car(Scan),symbol_GetIncreasedOrderingCounter());

  list_Delete(Functions);
  list_Delete(Predicates);
  for(Scan=Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
    clause_OrientAndReInit(list_Car(Scan));


  flag_ClearInferenceRules();
  flag_ClearReductionRules();

  flag_SetFlagValue(flag_ROBV,1);
  flag_SetFlagValue(flag_RTAUT,1);
  flag_SetFlagValue(flag_RFSUB,1);
  flag_SetFlagValue(flag_RBSUB,1);
  flag_SetFlagValue(flag_RFCLR,1);
  flag_SetFlagValue(flag_RBCLR,1);
  flag_SetFlagValue(flag_FULLRED,1);
  /*flag_SetFlagValue(flag_FUNCWEIGHT,1);
  flag_SetFlagValue(flag_VARWEIGHT,1);*/
  flag_SetFlagValue(flag_WDRATIO,5);

  if (ana_NEQUATIONS)
    flag_SetFlagValue(flag_IEQR,1);

  if (ana_PEQUATIONS) {
    flag_SetFlagValue(flag_ISPR,1);
    flag_SetFlagValue(flag_ISPL,1);
    if (ana_NONHORNCLAUSES >0)
      flag_SetFlagValue(flag_IEQF,1);
    if (ana_NONUNIT) 
      flag_SetFlagValue(flag_RCON,1);
    flag_SetFlagValue(flag_RFREW,1);
    flag_SetFlagValue(flag_RBREW,1);
    /*if (ana_PUREEQUATIONAL) {
      CLAUSE Clause;
      flag_SetFlagValue(flag_FUNCWEIGHT,2);
      while (Clauses) {
	Clause = list_Car(Clauses);
	clause_SetWeight(Clause, clause_ComputeWeight(Clause));
	Clauses = list_Cdr(Clauses);
      }
    }*/
  }
  
  if (ana_SORTRES) {
    flag_SetFlagValue(flag_IEMS,1);
    flag_SetFlagValue(flag_ISOR,1);
    flag_SetFlagValue(flag_RSSI,1);
    if (!ana_PEQUATIONS)
      flag_SetFlagValue(flag_RSST,1);
  }
  else
    flag_SetFlagValue(flag_SORTS,0);

  if (ana_MONADIC || ana_NONMONADIC) { /* Problem contains real predicates */
    flag_SetFlagValue(flag_IGER,1);
    if (ana_NONHORNCLAUSES > 0)
      flag_SetFlagValue(flag_IGEF,1);
    if (ana_NONUNIT) 
      flag_SetFlagValue(flag_RCON,1);
  }


  if (!ana_FUNCTIONS)
    flag_SetFlagValue(flag_SELECT,2);
  else
    if (ana_NONUNIT)
      flag_SetFlagValue(flag_SELECT,1);
    else
      flag_SetFlagValue(flag_SELECT,0);

  if (ana_CONCLAUSES < ana_AXIOMCLAUSES || (ana_CONGROUND && !ana_PUREPROPOSITIONAL))
    flag_SetFlagValue(flag_SATINPUT,1);
  else
    flag_SetFlagValue(flag_SATINPUT,0);

  if (ana_NONHORNCLAUSES > 0)
    flag_SetFlagValue(flag_SPLITS,-1);
  else
    flag_SetFlagValue(flag_SPLITS,0);
}



