% I-THOP  (A first-order theorem prover based on PTTP technology)
% 
%      version 1.01
%      July 2, 1997
%      Copyright (C) 1997, Koji Iwanuma (iwanuma@esi.yamanashi.ac.jp)
% 
% This program is free software; you can redistribute it or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either versions 1, or (at your option)
% any later version.
%  
% This program is distributed in the hope that it will be useful but
% WITHOUT ANY WARRANTY.  See the GNU General Public License for more
% details.
% ---------------
%
%   comPRE.pl
%
% ---------------

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% 
%%%      preprocessing tools
%%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

preprocessing(X) :-
	nl,
	write('Input Formulas:'),nl,
	make_tmp_data(X),
	%
	call(axiom_att(horn,XX)),
	nl,
	(XX == horn ->
	   write('[Input Formulas constitute a Horn theory !!]'),nl ;
	 XX == non_horn_existed ->
	   write('[Input Formulas constitute a Non-Horn theory !!]'),nl ;
	%true ->
	   nl,write('Error in axiom_att(horn,_) !!'),nl
	),
	call(axiom_att(max_term_depth,DD)),
	write('[Max term depth is '),write(DD),write(']'),nl,
	%
	set_default_mode(XX),
	set_user_mode_and_data.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_tmp_data((CarCl,CdrCl)) :-
	!,
	assert_tmp_clauses(CarCl),
	make_tmp_data(CdrCl).

make_tmp_data((Cl)):-
	assert_tmp_clauses(Cl).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

assert_tmp_clauses((query :- Goals)) :-	 % query
	!,
	conjunction_to_list_with_dummy(Goals, GList),
	get_max_term_depth_in_literal_list(GList,D),
	compare_and_max_set(D),
	%
        complemented_list(GList,CompGList),
	check_Non_Horn(CompGList,Flag),
	(Flag \== '@Negative_clause' ->
	      retract(axiom_att(horn,_)),
	      assertz(axiom_att(horn,non_horn_existed)) ;
        %true ->
	      true
	),
	%
	list_to_conjunction(GList, NewGoals),
	call_id_number(clause,ID),inc_id_number(clause),
	assertz(contrapositive(query,query,NewGoals,ID)),
	%
	write_indent_for_number(ID),
	write(ID),write('   '),
	write_lettervars((query :- Goals)),write('.'),nl.

assert_tmp_clauses((Head :- Body)) :-   % Horn Clause 
	!,
	add_dummy_argument(Head, NewHead),
	conjunction_to_list_with_dummy(Body, BList),
	get_max_term_depth_in_literal_list([NewHead|BList],D),
	compare_and_max_set(D),
	%
	list_to_conjunction(BList, NewBody),
	call_id_number(clause,ID),inc_id_number(clause),
	assertz(contrapositive(horn,NewHead,NewBody,ID)),
	%
	write_indent_for_number(ID),
	write(ID),write('   '),
	write_lettervars((Head :- Body)),write('.'),nl.

assert_tmp_clauses((CarL;CdrL)) :-      % Non-horn and Non-unit
	!,
	clause_to_list_with_dummy((CarL;CdrL),CList),
	get_max_term_depth_in_literal_list(CList,D),
	compare_and_max_set(D),
	%
	check_Non_Horn(CList,Flag),
	(Flag == '@Non_Horn_clause' ->
	      retract(axiom_att(horn,_)),
	      assertz(axiom_att(horn,non_horn_existed)) ;
        %true ->
	      true
	),
	%
	call_id_number(clause,ID),inc_id_number(clause),
	(Flag == '@Negative_clause' ->
	      assertz(contrapositive(clause,negative,CList,ID));
	%true ->
	      assertz(contrapositive(clause,non_nega,CList,ID))
	),
	%
	write_indent_for_number(ID),
	write(ID),write('   '),
	write_lettervars((CarL;CdrL)),write('.'),nl.

assert_tmp_clauses(Unit) :-             % Unit
	add_dummy_argument(Unit, NewUnit),
	get_max_term_depth_in_literal_list([NewUnit],D),
	compare_and_max_set(D),
	%
	call_id_number(clause,ID),inc_id_number(clause),
	assertz(contrapositive(unit,NewUnit,true,ID)),
	%
	write_indent_for_number(ID),
	write(ID),write('   '),
	write_lettervars(Unit),write('.'),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Converting a proposition to a predicate with a dummy argument
%
%  The dummy argument added in this routine is 'dummy_arg.'
%  Notice: In order to change dummy argument to '@dummy_arg'etc,
%  A lot of other part of this program code must be rewritten,
%  for symbols '@..' cannot pass any file write and read operation. 
%  I notice more, perhaps write_canonical/1 is not compatible with
%  comCOM.pl
%  

add_dummy_argument(Predicate, NewPred) :-
      ( atom(Predicate) ->
              NewPred =.. [Predicate, 'dummy_arg']
      ;
              NewPred = Predicate
      ).

clause_to_list_with_dummy((Lite;RemClause),[Pred|RemList]) :-
      !,
      add_dummy_argument(Lite, Pred),
      clause_to_list_with_dummy(RemClause,RemList).
clause_to_list_with_dummy(Lite,[Pred]) :-
      add_dummy_argument(Lite, Pred).

conjunction_to_list_with_dummy((Lite,RemCon),[Pred|RemList]) :-
      !,
      add_dummy_argument(Lite, Pred),
      conjunction_to_list_with_dummy(RemCon,RemList).
conjunction_to_list_with_dummy(Lite,[Pred]) :-
      add_dummy_argument(Lite, Pred).
  
  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

get_max_term_depth_in_literal_list([],0) :-!.
get_max_term_depth_in_literal_list([Car|Cdr],D) :-
	get_term_depth(Car,X1),D1 is X1-1,
        get_max_term_depth_in_literal_list(Cdr,D2),
	(D1 >= D2 ->  D=D1 ; D=D2).

get_term_depth(X,1):-
	(var(X);atom(X)),!.

get_term_depth(Term,N) :-
           functor(Term,_,S),
	   get_term_depth_args(S,Term,N1),
	   N is N1 + 1.

get_term_depth_args(0,_,0):- !.
get_term_depth_args(S,Term,N) :-
	arg(S,Term,Arg),
        get_term_depth(Arg,N1),
	S1 is S-1, get_term_depth_args(S1,Term,N2),
	(N1 >= N2 -> N = N1 ; N = N2 ).

%%%%%%%%%%%%%%%%%

check_Non_Horn(List,Flag) :-
	chk_Horn_cl_to_list(List,0,PN2),
	(PN2 == 1 ->
	     Flag = '@Horn_clause' ;
	 PN2 >= 2 ->
	     Flag = '@Non_Horn_clause' ;
	 PN2 == 0 ->
	     Flag = '@Negative_clause' ;
	%true ->
	     nl,write('Error in check_Horn_and_clause_to_list !!!'),nl
	).
 
chk_Horn_cl_to_list([],PN1,PN1) :-!.
chk_Horn_cl_to_list([Car|Cdr],PN1,PN2) :-
	!,
	functor(Car,F,_),name(F,FList),
	(append("not_",_,FList) -> 
              PN3 is PN1 ; PN3 is PN1 + 1 ),
	chk_Horn_cl_to_list(Cdr,PN3,PN2).

%%%%%%%%%%%%%%%%%

compare_and_max_set(D) :-
	call(axiom_att(max_term_depth,DD)),
	(DD >= D ->
	     true ;
	%true ->
	     retract(axiom_att(max_term_depth,DD)),
	     assertz(axiom_att(max_term_depth,D))
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_lettervars(X) :-
	\+ \+ (lettervars(X),write(X)).

lettervars(X) :-
	tt_varlist(X,V1),
	unify_vars(V1,
	 ['X','Y','Z','U','V','W','A','B','C','D','E','F',
	  'X1','Y1','Z1','U1','V1','W1','A1','B1','C1','D1','E1','F1',
	  'X2','Y2','Z2','U2','V2','W2','A2','B2','C2','D2','E2','F2',
	  'X3','Y3','Z3','U3','V3','W3','A3','B3','C3','D3','E3','F3',
	  'X4','Y4','Z4','U4','V4','W4','A4','B4','C4','D4','E4','F4',
	  'X5','Y5','Z5','U5','V5','W5','A5','B5','C5','D5','E5','F5',
	  'X6','Y6','Z6','U6','V6','W6','A6','B6','C6','D6','E6','F6',
	  'X7','Y7','Z7','U7','V7','W7','A7','B7','C7','D7','E7','F7',
	  'X8','Y8','Z8','U8','V8','W8','A8','B8','C8','D8','E8','F8',
	  'X9','Y9','Z9','U9','V9','W9','A9','B9','C9','D9','E9','F9',
	  'X10','Y10','Z10','U10','V10','W10',
	  'X11','Y11','Z11','U11','V11','W11',
	  'X12','Y12','Z12','U12','V12','W12',
	  'X13','Y13','Z13','U13','V13','W13',
	  'X14','Y14','Z14','U14','V14','W14',
	  'X15','Y15','Z15','U15','V15','W15',
	  'X16','Y16','Z16','U16','V16','W16',
	  'X17','Y17','Z17','U17','V17','W17',
	  'X18','Y18','Z18','U18','V18','W18',
	  'X19','Y19','Z19','U19','V19','W19',
	  'X20','Y20','Z20','U20','V20','W20',
	  'X21','Y21','Z21','U21','V21','W21',
	  'X22','Y22','Z22','U22','V22','W22',
	  'X23','Y23','Z23','U23','V23','W23',
	  'X24','Y24','Z24','U24','V24','W24'
	 ]),!.

unify_vars([X|L1],[X|L2]) :- !,
	unify_vars(L1,L2).
unify_vars([_|L1],[_|L2]) :- !,
	unify_vars(L1,L2).
unify_vars(_,_).

%% tt_varlist(T,L,[]) <- L is all occurances of distinct variables in term T
tt_varlist(X,L) :- tt_varlist(X,L,[]),!.

tt_varlist(X,[X|L],L) :- var(X),!.
tt_varlist(T,L0,L) :- T =.. [_|A],!,tt_varlist1(A,L0,L).

tt_varlist1([T|A],L0,L) :- tt_varlist(T,L0,L1),!,tt_varlist1(A,L1,L).
tt_varlist1([],L,L).

%%%%%%%%%%%%%%%%%%%%%%

is_Negative(Lite) :-
	functor(Lite,F,_),name(F,Flist),
	append("not_",_,Flist),
	!.

%%%%%%%%%%%%%%%%%%%%%%
