/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *               INPUT OF SPASS CLAUSES                   * */
/* *                                                        * */
/* *  $Module:   DFGPARSER                                  * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.72 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/01 14:09:27 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: dfgparser.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dfgparser.h"


/**************************************************************/
/* Static Variables                                           */
/**************************************************************/

/* Used for Tokens                                            */
static int    dfg_LOOKAHEAD = '\0';

/* Axiom formulae are appended at the end of 'dfg_AXIOMLIST', */
/* so two list pointers are needed.                           */
static LIST   dfg_AXIOMLIST;
static LIST   dfg_LASTAXIOM;

/* Conjecture formulae are appended at the end of             */
/* 'dfg_CONJECLIST', so two list pointers are needed.         */
static LIST   dfg_CONJECLIST;
static LIST   dfg_LASTCONJEC;

/* Clauses are appended at the end of 'dfg_CLAUSELIST',       */
/* so two list pointers are needed.                           */
static LIST   dfg_CLAUSELIST;
static LIST   dfg_LASTCLAUSE;

/* Input clauses and proofs are appended to the proof list    */
static LIST   dfg_PROOFLIST;
static LIST   dfg_LASTPROOF;

/* The symbol list holds all symbols whose arity wasn't       */
/* specified in the symbol declaration section.               */
/* If the arity of a symbol was not specified in this section */
/* it is first set to 0. If the symbol occurs with always the */
/* same arity 'A' the arity of this symbol is set to 'A'.     */
static LIST   dfg_SYMBOLLIST;

/* List of quantified variables in the current input formula. */
/* This list is used to find symbols that by mistake weren't  */
/* declared in the symbol declaration section                 */
/* --> free variables                                         */
static LIST   dfg_VARLIST;

/* Used for the texts from description section.               */
typedef struct {
  char*     name;
  char*     author;
  char*     version;
  char*     logic;
  DFG_STATE status;
  char*     description;
  char*     date;
} DFG_DESCRIPTIONTYPE;

static DFG_DESCRIPTIONTYPE dfg_DESC;

/**************************************************************/
/* Prototypes for static functions used only in this module   */
/**************************************************************/

static TERM   dfg_Atom(BOOL);
static TERM   dfg_Term(BOOL);
static LIST   dfg_TermList(BOOL);
static LIST   dfg_VarList(void);
static LIST   dfg_QuantifierTermList(void);
static void   dfg_Description(void);
static void   dfg_SymbolList(void);
static void   dfg_SymbolDeclWithArity(int);
static void   dfg_SortSymbolDecl(void);
static void   dfg_DeclList(void);
static void   dfg_Settings(void);
static void   dfg_PrecedenceList(void);
static void   dfg_Flag(void);

static void   dfg_Match(int);
static void   dfg_Warning(const char*);
static CLAUSE dfg_CreateClauseFromTerm(TERM, BOOL);

static SYMBOL dfg_VarLookup(const char*);
static SYMBOL dfg_VarDefine(const char*);
static void   dfg_VarListDelete(void);

static void   dfg_SymAdd(SYMBOL Symbol);
static void   dfg_SymCheck(SYMBOL Symbol, NAT Arity);
static void   dfg_SymCleanUp(void);

/**************************************************************/
/* Function Definitions                                       */
/**************************************************************/

static void dfg_InitParser(FILE* Input)
/**************************************************************
  INPUT:   The input file stream for the parser.
  RETURNS: Nothing.
  EFFECT:  The parser and scanner are initialized.
           Memory is allocated, e.g. for the internal scanner buffers.
***************************************************************/
{
  dfg_InitScanner(Input);

  dfg_LOOKAHEAD           = dfg_Scanner();
  dfg_AXIOMLIST           = dfg_LASTAXIOM  = list_List(list_Nil());
  dfg_CONJECLIST          = dfg_LASTCONJEC = list_List(list_Nil());
  dfg_CLAUSELIST          = dfg_LASTCLAUSE = list_List(list_Nil());
  dfg_PROOFLIST           = dfg_LASTPROOF  = list_List(list_Nil());
  dfg_SYMBOLLIST          = list_Nil();
  dfg_VARLIST             = list_Nil();
  dfg_DESC.name           = (char*) NULL;
  dfg_DESC.author         = (char*) NULL;
  dfg_DESC.version        = (char*) NULL;
  dfg_DESC.logic          = (char*) NULL;
  dfg_DESC.status         = UNKNOWN;
  dfg_DESC.description    = (char*) NULL;
  dfg_DESC.date           = (char*) NULL;
}

static void dfg_FreeParser(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Memory used by the parser and scanner is freed.
***************************************************************/
{
  dfg_FreeScanner();

  dfg_AXIOMLIST  = list_Pop(dfg_AXIOMLIST);
  dfg_CONJECLIST = list_Pop(dfg_CONJECLIST);
  dfg_CLAUSELIST = list_Pop(dfg_CLAUSELIST);
  dfg_PROOFLIST  = list_Pop(dfg_PROOFLIST);

  dfg_SymCleanUp();   /* dfg_SYMBOLLIST is deleted */

  /* dfg_VARLIST is already deleted */

  /* The description entries are freed with function dfg_Free */
}

void dfg_Free(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Frees memory used by the problem description.
***************************************************************/
{
  if (dfg_DESC.name != NULL)
    memory_Free(dfg_DESC.name, strlen(dfg_DESC.name) + 1);
  if (dfg_DESC.author != NULL)
    memory_Free(dfg_DESC.author, strlen(dfg_DESC.author) + 1);
  if (dfg_DESC.version != NULL)
    memory_Free(dfg_DESC.version, strlen(dfg_DESC.version) + 1);
  if (dfg_DESC.logic != NULL)
    memory_Free(dfg_DESC.logic, strlen(dfg_DESC.logic) + 1);
  if (dfg_DESC.description != NULL)
    memory_Free(dfg_DESC.description, strlen(dfg_DESC.description) + 1);
  if(dfg_DESC.date != NULL)
    memory_Free(dfg_DESC.date, strlen(dfg_DESC.date) + 1);
}

const char* dfg_ProblemName(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's name from the description section.
***************************************************************/
{
  return dfg_DESC.name;
}

const char* dfg_ProblemAuthor(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's author from the description section.
***************************************************************/
{
  return dfg_DESC.author;
}

const char* dfg_ProblemVersion(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's version from the description section.
***************************************************************/
{
  return dfg_DESC.version;
}

const char* dfg_ProblemLogic(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's logic from the description section.
***************************************************************/
{
  return dfg_DESC.logic;
}

DFG_STATE dfg_ProblemStatus(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's status from the description section.
***************************************************************/
{
  return dfg_DESC.status;
}

const char* dfg_ProblemStatusString(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The string representation of the problem's status.
***************************************************************/
{
  char* result = "";

  switch (dfg_DESC.status) {
  case SATISFIABLE:
    result = "satisfiable"; break;
  case UNSATISFIABLE:
    result = "unsatisfiable"; break;
  case UNKNOWN:
    result = "unknown"; break;
  default:
    fprintf(stderr, "Invalid status in function dfg_StatusString\n\n");
    misc_Error();
  }
  return result;
}

const char* dfg_ProblemDescription(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's description from the decsription section.
***************************************************************/
{
  return dfg_DESC.description;
}

const char* dfg_ProblemDate(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The problem's date from the description section.
***************************************************************/
{
  return dfg_DESC.date;
}

NAT dfg_DescriptionLength(void)
/**************************************************************
  INPUT:   None.
  RETURNS: An upper bound for the length of the description
           section if printed as string with dfg_SPrintDescription.
***************************************************************/
{
  NAT result = 200;
  /* Estimation for all syntactic strings like "end_of_list" */

  if (dfg_DESC.name != NULL)
    result += strlen(dfg_DESC.name);
  if (dfg_DESC.author != NULL)
    result += strlen(dfg_DESC.author);
  if (dfg_DESC.version != NULL)
    result += strlen(dfg_DESC.version);
  if (dfg_DESC.logic != NULL)
    result += strlen(dfg_DESC.logic);
  result += strlen(dfg_ProblemStatusString());
  if (dfg_DESC.description != NULL)
    result += strlen(dfg_DESC.description);
  if (dfg_DESC.date != NULL)
    result += strlen(dfg_DESC.date);

  return result;
}

void dfg_FPrintDescription(FILE* File)
/**************************************************************
  INPUT:   A file stream.
  RETURNS: Nothing.
  EFFECT:  The description section from the input file
           is printed to 'File'. You must call the parser first
	   before calling this function.
***************************************************************/
{
  fputs("list_of_descriptions.\n  name({*", File);
  if (dfg_DESC.name != NULL)
    fputs(dfg_DESC.name, File);
  fputs("*}).\n  author({*", File);
  if (dfg_DESC.author != NULL)
    fputs(dfg_DESC.author, File);
  fputs("*}).\n", File);
  if (dfg_DESC.version != NULL) {
    fputs("  version({*", File);
    fputs(dfg_DESC.version, File);
    fputs("*}).\n", File);
  }
  if (dfg_DESC.logic != NULL) {
    fputs("  logic({*", File);
    fputs(dfg_DESC.logic, File);
    fputs("*}).\n", File);
  }
  fputs("  status(", File);
  fputs(dfg_ProblemStatusString(), File);
  fputs(").\n  description({*", File);
  if (dfg_DESC.description != NULL)
    fputs(dfg_DESC.description, File);
  fputs("*}).\n", File);
  if (dfg_DESC.date != NULL) {
    fputs("  date({*", File);
    fputs(dfg_DESC.date, File);
    fputs("*}).\n", File);
  }
  fputs("end_of_list.", File);
}

char* dfg_SPrintDescription(char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: Nothing.
  EFFECT:  The description section from the input file
           is printed to 'String'. You must call the parser first
	   before calling this function.
***************************************************************/
{
  String += sprintf(String, "list_of_descriptions.\n  name({*%s*}).\n"
		    "  author({*%s*}).\n",
		    dfg_DESC.name   == NULL ? "" : dfg_DESC.name,
		    dfg_DESC.author == NULL ? "" : dfg_DESC.author);
  if (dfg_DESC.version != NULL)
    String += sprintf(String, "  version({*%s*}).\n", dfg_DESC.version);
  if (dfg_DESC.logic != NULL)
    String += sprintf(String, "  logic({*%s*}).\n", dfg_DESC.logic);
  String += sprintf(String, "  status(%s).\n  description({*%s*}).\n",
		    dfg_ProblemStatusString(),
		    dfg_DESC.description == NULL ? "" : dfg_DESC.description);
  if (dfg_DESC.date != NULL)
    String += sprintf(String, "  date({*%s*}).\n", dfg_DESC.date);
  String += sprintf(String, "end_of_list.");
  return String;
}


/**************************************************************/
/* Functions for clause input                                 */
/**************************************************************/

static void dfg_CNFClausesOnly(void);
static void dfg_ClauseList(BOOL IsAxiom);
static TERM dfg_Clause(void);
static LIST dfg_LitList(void);

LIST dfg_ClauseParser(char* Filename)
/**************************************************************
  INPUT:   The name of the input file containing clauses in DFG
           syntax.
  RETURNS: The list of clauses read from the input file.
           This list includes both the axiom and conjecture clauses
	   from the DFG file.
***************************************************************/
{
  FILE* input;

  input = misc_OpenFile(Filename,"r");  

  dfg_InitParser(input);      /* Initialize the parser */

  dfg_Match(DFG_BEGPROBLEM);
  dfg_Match('(');
  dfg_Match(DFG_ID);
  dfg_Match(')');
  dfg_Match('.');
  dfg_Description();
  dfg_SymbolList();
  dfg_CNFClausesOnly();
  dfg_Settings();
  dfg_Match(DFG_ENDPROBLEM);

  if (dfg_LOOKAHEAD != EOF)
    dfg_Warning("Text after \"endproblem.\" ignored");

  misc_CloseFile(input,Filename);

  dfg_FreeParser();
  
  if (list_Empty(dfg_CLAUSELIST)) {
    /* Found no clauses in CNF */
    fputs("No CNF clauses found in input file!\n\n", stderr);
    misc_Error();
  }

  list_Apply(clause_OrientAndReInit, dfg_CLAUSELIST);

  return dfg_CLAUSELIST;
}

static void dfg_CNFClausesOnly(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses several clause and formula lists
           each from 'list_of_clauses' or 'list_of_formulae'
           to 'end_of_list'.
	   Only clauses in CNF are stored, clauses in DNF and
	   formulae are ignored.
***************************************************************/
{
  /* Ignore formulae */
  while (dfg_LOOKAHEAD == DFG_FORMLIST) {
    do
      dfg_Match(dfg_LOOKAHEAD);
    while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF);
    dfg_Match(DFG_ENDLIST);
  }

  while (dfg_LOOKAHEAD == DFG_CLAUSELIST) {
    BOOL isAxiom = TRUE;

    dfg_Match(DFG_CLAUSELIST);
    dfg_Match('(');
    if (dfg_LOOKAHEAD == DFG_AXIOMS)
      dfg_Match(DFG_AXIOMS);
    else {
      isAxiom = FALSE;
      dfg_Match(DFG_CONJECS);
    }
    dfg_Match(',');
    if (dfg_LOOKAHEAD == DFG_CNF) {
      dfg_Match(DFG_CNF);
      dfg_Match(')');
      dfg_Match('.');
      dfg_ClauseList(isAxiom);   /* Use dfg_CLAUSELIST for the clauses */
    } else
      while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF)
	dfg_Match(dfg_LOOKAHEAD);        /* Ignore clauses in DNF */
    
    dfg_Match(DFG_ENDLIST);
  }   /* while */

  /* Ignore proofs */
  while (dfg_LOOKAHEAD == DFG_PROOF) {
    do
      dfg_Match(dfg_LOOKAHEAD);
    while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF);
    dfg_Match(DFG_ENDLIST);
  }
}


static void dfg_ClauseList(BOOL IsAxiom)
/**************************************************************
  INPUT:   A boolean value.
  RETURNS: Nothing.
  EFFECT:  This function parses several clauses each starting
           with 'clause(' and ending with ').'.
	   If 'IsAxiom' is TRUE all clauses are stored as axioms
	   in dfg_CLAUSELIST, else they're stored as conjectures.
***************************************************************/
{
  CLAUSE clause;
  TERM   clauseterm;

  while (dfg_LOOKAHEAD == DFG_CLAUSE) {
    dfg_Match(DFG_CLAUSE);

    dfg_LOOKUPSYMBOLS = TRUE; /* Tell the scanner to search for symbol names */
    dfg_Match('(');

    clauseterm = dfg_Clause();
    clause     = dfg_CreateClauseFromTerm(clauseterm, IsAxiom);

    dfg_LOOKUPSYMBOLS = FALSE;

    if (clause != NULL) {
      list_Rplacd(dfg_LASTCLAUSE, list_List(clause));
      dfg_LASTCLAUSE = list_Cdr(dfg_LASTCLAUSE);
    }

    if (dfg_LOOKAHEAD == ',') {
      dfg_Match(',');
      if (dfg_LOOKAHEAD == DFG_NUM)
	dfg_Match(DFG_NUM);
      else
	dfg_Match(DFG_ID);
    }
    dfg_Match(')');
    dfg_Match('.');

    dfg_VarListDelete(); /* Delete list of quantified variables */
  }
}


static TERM dfg_Clause(void)
/**************************************************************
  INPUT:   None.
  RETURNS: A term with top symbol 'forall' or 'or'.
  EFFECT:  This function parses a single clause with forall
           or 'or' as top symbol. Only the literals are returned.
***************************************************************/
{
  TERM result;
  LIST args;

  if (dfg_LOOKAHEAD == DFG_FORALL) {
    LIST variables;

    dfg_Match(DFG_FORALL);
    dfg_Match('(');
    variables = dfg_VarList();
    dfg_Match(',');
    dfg_Match(DFG_OR);
    dfg_Match('(');
    args = dfg_LitList();
    dfg_Match(')');
    dfg_Match(')');
    result = fol_CreateQuantifier(fol_All(), variables,
				  list_List(term_Create(fol_Or(), args)));
  } else {
    dfg_Match(DFG_OR);
    dfg_Match('(');
    args = dfg_LitList();
    dfg_Match(')');
    result = term_Create(fol_Or(), args);
  }
  return result; 
}


static LIST dfg_LitList(void)
/**************************************************************
  INPUT:   None.
  RETURNS: A list of literal terms with a predicate or 'not'
           as top symbol.
  EFFECT:  This function parses a comma-separated list of literals.
***************************************************************/
{
  LIST result = list_List(list_Nil()), last = result;
  BOOL loop = TRUE;
  TERM literalterm;

  do {
    if (dfg_LOOKAHEAD == DFG_NOT) {
      TERM atom;

      dfg_Match(DFG_NOT);
      dfg_Match('(');
      atom = dfg_Atom(FALSE);  /* Not in sort */
      dfg_Match(')');
      literalterm = term_Create(fol_Not(), list_List(atom));
    } else
      literalterm = dfg_Atom(FALSE);  /* Not in sort */

    if (dfg_LOOKAHEAD == ',')
      dfg_Match(',');
    else
      loop = FALSE;
    
    /* Insert at end of result list */
    list_Rplacd(last, list_List(literalterm));
    last = list_Cdr(last);
  } while (loop);
      
  result = list_Pop(result); /* Remove the empty item from the beginning */
  return result;
}


/**************************************************************/
/* Functions for formula input                                */
/**************************************************************/

static void dfg_FormulaeOnly(void);
static void dfg_FormulaList(BOOL Axioms);
static TERM dfg_Formula(void);
static LIST dfg_FormList(void);

void dfg_Parser(char* Filename, LIST* AxiomList, LIST* ConjectureList)
/**************************************************************
  INPUT:   The name of the input file containing clauses in DFG
           syntax. AxiomList and ConjectureList are pointers to
	   lists used as return values.
  RETURNS: Nothing.
  EFFECT:  Reads formulae from the input file.
           The axioms and conjectures from the input file are
	   appended to the respective lists, the lists are not
	   deleted !
***************************************************************/
{
  FILE* input;

  input = misc_OpenFile(Filename,"r");  

  dfg_InitParser(input);

  dfg_Match(DFG_BEGPROBLEM);
  dfg_Match('(');
  dfg_Match(DFG_ID);
  dfg_Match(')');
  dfg_Match('.');
  dfg_Description();
  dfg_SymbolList();
  dfg_FormulaeOnly();
  dfg_Settings();
  dfg_Match(DFG_ENDPROBLEM);

  if (dfg_LOOKAHEAD != EOF)
    dfg_Warning("Text after \"end_problem.\" ignored");
  misc_CloseFile(input,Filename);

  dfg_FreeParser();

  if (list_Empty(dfg_AXIOMLIST) && list_Empty(dfg_CONJECLIST)) {
    /* Input file contains no formulae */
    fputs("No formulae found in input file!\n\n", stderr);
    misc_Error();
  }

  *AxiomList = list_Nconc(*AxiomList, dfg_AXIOMLIST);
  *ConjectureList = list_Nconc(*ConjectureList, dfg_CONJECLIST);

  /* Not necessary, but safer */
  dfg_AXIOMLIST  = dfg_LASTAXIOM  = list_Nil();
  dfg_CONJECLIST = dfg_LASTCONJEC = list_Nil();
}


static void dfg_FormulaeOnly(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses several clause and formula lists
           each from 'list_of_clauses' or 'list_of_formulae'
           to 'end_of_list'.
	   Only formulae are stored, clauses are ignored.
***************************************************************/
{
  while (dfg_LOOKAHEAD == DFG_FORMLIST) {
    BOOL isAxiom = TRUE;

    dfg_Match(DFG_FORMLIST);
    dfg_Match('(');
    if (dfg_LOOKAHEAD == DFG_AXIOMS)
      dfg_Match(DFG_AXIOMS);
    else {
      isAxiom = FALSE;
      dfg_Match(DFG_CONJECS);
    }
    dfg_Match(')');
    dfg_Match('.');
    /* Use dfg_AXIOMLIST or dfg_ConjectureList for the formulae */
    dfg_FormulaList(isAxiom);
    dfg_Match(DFG_ENDLIST);
  }   /* while */
  
  /* Ignore clause lists */
  while (dfg_LOOKAHEAD == DFG_CLAUSELIST) {
    do
      dfg_Match(dfg_LOOKAHEAD);
    while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF);
    dfg_Match(DFG_ENDLIST);
  }

  /* Ignore proof lists */
  while (dfg_LOOKAHEAD == DFG_PROOF) {
    do
      dfg_Match(dfg_LOOKAHEAD);
    while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF);
    dfg_Match(DFG_ENDLIST);
  }
}


static void dfg_FormulaList(BOOL Axioms)
/**************************************************************
  INPUT:   A boolean value.
  RETURNS: Nothing.
  EFFECT:  This function parses several formulae each starting
           with 'formula(' and ending with ').'.
	   If 'Axioms' is TRUE all formulae are appended to
	   the axiom list else they're appended to the conjecture
	   list.
***************************************************************/
{
  TERM formula;

  while (dfg_LOOKAHEAD == DFG_FORMULA) {
    dfg_Match(DFG_FORMULA);

    dfg_LOOKUPSYMBOLS = TRUE;
    dfg_Match('(');
    formula = dfg_Formula();

    dfg_LOOKUPSYMBOLS = FALSE;

    if (dfg_LOOKAHEAD == ',') {
      dfg_Match(',');
      if (dfg_LOOKAHEAD == DFG_NUM)
	dfg_Match(DFG_NUM);
      else
	dfg_Match(DFG_ID);
    }
    dfg_Match(')');
    dfg_Match('.');

    if (Axioms) {
      list_Rplacd(dfg_LASTAXIOM, list_List(formula));
      dfg_LASTAXIOM = list_Cdr(dfg_LASTAXIOM);
    } else {
      list_Rplacd(dfg_LASTCONJEC, list_List(formula));
      dfg_LASTCONJEC = list_Cdr(dfg_LASTCONJEC);
    }

    dfg_VarListDelete(); /* Delete list of quantified variables */
  }    /* while */

  dfg_LOOKUPSYMBOLS = FALSE;
}


static TERM dfg_Formula(void)
/**************************************************************
  INPUT:   None.
  RETURNS: A formula with a predicate or a junctor as top
           symbol.
  EFFECT:  This function parses a single formula.
***************************************************************/
{
  SYMBOL topsymbol;
  LIST   arglist, termlist, sortlist, lastsort, varlist, lastvar, scan;
  TERM   result;

  switch (dfg_LOOKAHEAD) {
  case DFG_FORALL:
  case DFG_EXISTS:
    /* Only these two quantifiers are allowed in FOL */
    if (dfg_LOOKAHEAD == DFG_EXISTS)
      topsymbol = fol_Exist();
    else
      topsymbol = fol_All();
    dfg_Match(dfg_LOOKAHEAD); /* Match the quantifier itself */
    dfg_Match('(');
    termlist = dfg_QuantifierTermList();  /* May contain variables or atoms */
    dfg_Match(',');
    arglist = list_List(dfg_Formula());
    dfg_Match(')');
    
    varlist  = lastvar  = list_List(list_Nil());
    sortlist = lastsort = list_List(list_Nil());

    for ( ; !list_Empty(termlist); termlist = list_Pop(termlist)) {
      TERM term = list_Car(termlist);

      if (term_IsVariable(term)) {
	list_Rplacd(lastvar, list_List((POINTER) term_TopSymbol(term)));
	lastvar = list_Cdr(lastvar);
	term_Delete(term);
      } else {
	list_Rplacd(lastsort, list_List(term));
	lastsort = list_Cdr(lastsort);
	list_Rplacd(lastvar, term_VariableSymbols(term));
	while (!list_Empty(list_Cdr(lastvar)))
	  lastvar = list_Cdr(lastvar); /* Set lastvar to the new end of list */
      }
    }

    varlist  = list_Pop(varlist);
    sortlist = list_Pop(sortlist);

    varlist = list_PointerDeleteDuplicates(varlist);
    /* Transform variable symbols in varlist to terms */
    for (scan = varlist; !list_Empty(scan); scan = list_Cdr(scan))
      list_Rplaca(scan, term_Create((SYMBOL)list_Car(scan), list_Nil()));

    if (!list_Empty(sortlist)) {
      if (topsymbol == fol_All()) {
	/* Add the conjunction of all sort terms to the front of arglist */
	if (list_Empty(list_Cdr(sortlist))) {
	  list_Rplacd(sortlist, arglist);
	  arglist = sortlist;
	} else
	  arglist = list_Cons(term_Create(fol_And(), sortlist), arglist);
	/* The conjunction of sort terms implies the quantified term */
	arglist = list_List(term_Create(fol_Implies(), arglist));
      } else {
	/* Append quantified term to sort list */
	list_Rplacd(lastsort, arglist);
	/* The argument is the conjunction of sort terms and quantified term */
	arglist = list_List(term_Create(fol_And(), sortlist));
      }
    }   
     
    result = fol_CreateQuantifier(topsymbol, varlist, arglist);
    break;
  case DFG_NOT:
    /* Unary operator */
    dfg_Match(DFG_NOT);
    dfg_Match('(');
    arglist = list_List(dfg_Formula());
    dfg_Match(')');
    result = term_Create(fol_Not(), arglist);
    break;
  case DFG_IMPLIES:
  case DFG_IMPLIED:
  case DFG_EQUIV:
    /* Binary operators */
    if (dfg_LOOKAHEAD == DFG_IMPLIES)
      topsymbol = fol_Implies();
    else if (dfg_LOOKAHEAD == DFG_IMPLIED)
      topsymbol = fol_Implied();
    else
      topsymbol = fol_Equiv();
    dfg_Match(dfg_LOOKAHEAD); /* Match the operator itself */
    dfg_Match('(');
    arglist = list_List(dfg_Formula());             /* Set 1st argument */
    dfg_Match(',');
    list_Rplacd(arglist, list_List(dfg_Formula())); /* Set 2nd argument */
    dfg_Match(')');
    result = term_Create(topsymbol, arglist);
    break;
  case DFG_AND:
  case DFG_OR:
    /* Operators with arbitrary arity */
    if (dfg_LOOKAHEAD == DFG_AND)
      topsymbol = fol_And();
    else
      topsymbol = fol_Or();
    dfg_Match(dfg_LOOKAHEAD);
    dfg_Match('(');
    arglist = dfg_FormList();
    dfg_Match(')');
    result = term_Create(topsymbol, arglist);
    break;
  default:
    result = dfg_Atom(FALSE);       /* Not in sort */
  }

  return result;
}


static LIST dfg_FormList(void)
/**************************************************************
  INPUT:   None.
  RETURNS: A list of formulae with a predicate or junctor as
           top symbol.
  EFFECT:  This function parses a comma separated list of
           formulae.
***************************************************************/
{
  LIST result = list_List(list_Nil()), last = result;
  BOOL loop = TRUE;

  do {
    TERM formula = dfg_Formula();
    if (dfg_LOOKAHEAD == ',')
      dfg_Match(',');
    else
      loop = FALSE;

    /* Insert at end of result list */
    list_Rplacd(last, list_List(formula));
    last = list_Cdr(last);
  } while (loop);

  result = list_Pop(result); /* Remove the empty item from the beginning */
  return result;
}


/**************************************************************/
/* Functions for proof input                                  */
/**************************************************************/

static void dfg_InputClausesAndProof(void);
static void dfg_InputClauses(BOOL IsAxiom);

LIST dfg_ProofParser(char *Filename)
/**************************************************************
  INPUT:   The name of the input file containing clauses in DFG
           syntax.
  RETURNS: The proof list read from the input file.
           The elements of the list are lists with three items.
	   The first item is the label (a string) of a clause,
	   the second is the clause in TERM format and the third
	   is the list of justification labels.
	   Note that the justification list is empty for input
	   clauses.
***************************************************************/
{
  FILE* input;

  input = misc_OpenFile(Filename,"r");  

  dfg_InitParser(input);

  dfg_Match(DFG_BEGPROBLEM);
  dfg_Match('(');
  dfg_Match(DFG_ID);
  dfg_Match(')');
  dfg_Match('.');
  dfg_Description();
  dfg_SymbolList();
  dfg_InputClausesAndProof();
  dfg_Settings();
  dfg_Match(DFG_ENDPROBLEM);

  if (dfg_LOOKAHEAD != EOF)
    dfg_Warning("Text after \"endproblem.\" ignored");
  misc_CloseFile(input,Filename);

  dfg_FreeParser();

  if (list_Empty(dfg_PROOFLIST)) {
    /* Found no proof in file */
    fputs("No proof found in input file!\n\n", stderr);
    misc_Error();
  }

  return dfg_PROOFLIST;
}

static void dfg_InputClausesAndProof(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing
  EFFECT:  This function parses several clause, formula and
           proof lists, each from 'list_of_clauses', 'list_of_formulae'
           or 'list_of_proof' to 'end_of_list'.
           These "proof items" are appended to the global list
           dfg_PROOFLIST.
***************************************************************/
{
  /* Ignore formulae */
  while (dfg_LOOKAHEAD == DFG_FORMLIST) {
    do
      dfg_Match(dfg_LOOKAHEAD);
    while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF);
    dfg_Match(DFG_ENDLIST);
  }

  while (dfg_LOOKAHEAD == DFG_CLAUSELIST) {
    BOOL isAxiom = TRUE;

    dfg_Match(DFG_CLAUSELIST);
    dfg_Match('(');
    if (dfg_LOOKAHEAD == DFG_AXIOMS)
      dfg_Match(DFG_AXIOMS);
    else {
      isAxiom = FALSE;
      dfg_Match(DFG_CONJECS);
    }
    dfg_Match(',');
    if (dfg_LOOKAHEAD == DFG_CNF) {
      dfg_Match(DFG_CNF);
      dfg_Match(')');
      dfg_Match('.');
      dfg_InputClauses(isAxiom);
    } else
      while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF)
	dfg_Match(dfg_LOOKAHEAD);        /* Ignore clauses in DNF */
    
    dfg_Match(DFG_ENDLIST);
  }

  while (dfg_LOOKAHEAD == DFG_PROOF) {
    dfg_Match(DFG_PROOF);
    dfg_Match('(');
    dfg_Match(DFG_RESOLUTION);
    dfg_Match(')');
    dfg_Match('.');
    while (dfg_LOOKAHEAD == DFG_CLAUSE) {
      LIST justification, justiflast, tripel;
      TERM  clause;
      char* label;
      
      dfg_Match(DFG_CLAUSE);

      dfg_LOOKUPSYMBOLS = TRUE;
      dfg_Match('(');
      clause = dfg_Clause();
      dfg_LOOKUPSYMBOLS = FALSE;

      dfg_Match(',');
      label = dfg_LVAL.string;
      dfg_LVAL.string = NULL;
      if (dfg_LOOKAHEAD == DFG_NUM)
	dfg_Match(DFG_NUM);
      else
	dfg_Match(DFG_ID);

      dfg_Match(',');
      /* ignore justification label */
      if (dfg_LOOKAHEAD == DFG_NUM)
	dfg_Match(DFG_NUM);
      else
	dfg_Match(DFG_ID);

      dfg_Match(',');
      dfg_Match('[');
      justification = justiflast = list_List(list_Nil());
      while (dfg_LOOKAHEAD == DFG_ID || dfg_LOOKAHEAD == DFG_NUM) {
	list_Rplacd(justiflast, list_List(dfg_LVAL.string));
	justiflast = list_Cdr(justiflast);
	dfg_LVAL.string = NULL;
	dfg_Match(dfg_LOOKAHEAD);
	if (dfg_LOOKAHEAD == ',')
	  dfg_Match(',');
      }
      justification = list_Pop(justification);
      dfg_Match(']');
      dfg_Match(')');
      dfg_Match('.');

      tripel = list_Cons(label, list_Cons(clause, list_List(justification)));
      list_Rplacd(dfg_LASTPROOF, list_List(tripel));
      dfg_LASTPROOF = list_Cdr(dfg_LASTPROOF);
    }
    dfg_Match(DFG_ENDLIST);
  }
}


static void dfg_InputClauses(BOOL IsAxiom)
/**************************************************************
  INPUT:   A boolean value.
  RETURNS: Nothing
  EFFECT:  This function parses several clauses each starting
           with 'clause(' and ending with ').'.
	   Only clauses with labels are returned.
	   If 'IsAxiom' is TRUE all clauses are stored as axioms
	   else as conjectures.
***************************************************************/
{
  LIST  tripel;
  TERM  clause;
  char* label;

  while (dfg_LOOKAHEAD == DFG_CLAUSE) {
    dfg_Match(DFG_CLAUSE);

    dfg_LOOKUPSYMBOLS = TRUE; /* Tell the scanner to search for symbol names */
    dfg_Match('(');
    clause = dfg_Clause();
    dfg_LOOKUPSYMBOLS = FALSE;

    label = NULL;
    if (dfg_LOOKAHEAD == ',') {
      dfg_Match(',');
      label = dfg_LVAL.string;
      dfg_LVAL.string = NULL;
      if (dfg_LOOKAHEAD == DFG_NUM)
	dfg_Match(DFG_NUM);
      else
	dfg_Match(DFG_ID);
    }
    dfg_Match(')');
    dfg_Match('.');

    dfg_VarListDelete(); /* Delete list of quantified variables */

    if (label != NULL) {
      /* The justification list is empty for input clauses */
      tripel = list_Cons(label, list_Cons(clause, list_List(list_Nil())));
      list_Rplacd(dfg_LASTPROOF, list_List(tripel));
      dfg_LASTPROOF = list_Cdr(dfg_LASTPROOF);
    } else
      term_Delete(clause);
  }
}

/**************************************************************/
/* Functions for clause and formula input (at the same time)  */
/**************************************************************/

static void dfg_FormulaeAndClauses(void);

LIST dfg_DFGParser(char* Filename, LIST* Axioms, LIST* Conjectures)
/**************************************************************
  INPUT:   The name of the input file containing clauses in DFG
           syntax. Axioms and Conjectures are pointers to
	   lists used as return values.
  RETURNS: A list of clauses.
  EFFECT:  Reads formulae and clauses from the input file.
           The axioms and conjectures are appended to the
	   respective lists, the lists are not deleted !
***************************************************************/
{
  FILE* input;

  input = misc_OpenFile(Filename,"r");  

  dfg_InitParser(input);      /* Initialize the parser */

  dfg_Match(DFG_BEGPROBLEM);
  dfg_Match('(');
  dfg_Match(DFG_ID);
  dfg_Match(')');
  dfg_Match('.');
  dfg_Description();
  dfg_SymbolList();
  dfg_DeclList();
  dfg_FormulaeAndClauses();
  dfg_Settings();
  dfg_Match(DFG_ENDPROBLEM);

  if (dfg_LOOKAHEAD != EOF)
    dfg_Warning("Text after \"endproblem.\" ignored");
  misc_CloseFile(input,Filename);

  dfg_FreeParser();

  if (list_Empty(dfg_AXIOMLIST) && list_Empty(dfg_CONJECLIST) &&
      list_Empty(dfg_CLAUSELIST)) {
    fputs("No formulae and clauses found in input file!\n\n", stderr);
    misc_Error();
  }

  *Axioms      = list_Nconc(*Axioms, dfg_AXIOMLIST);
  *Conjectures = list_Nconc(*Conjectures, dfg_CONJECLIST);

  /* Not necessary, but safer */
  dfg_AXIOMLIST  = dfg_LASTAXIOM  = list_Nil();
  dfg_CONJECLIST = dfg_LASTCONJEC = list_Nil();
  
  list_Apply(clause_OrientAndReInit, dfg_CLAUSELIST);

  return dfg_CLAUSELIST;
}


static void dfg_FormulaeAndClauses(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses several clause and formula lists
           each from 'list_of_clauses' or 'list_of_formulae'
           to 'end_of_list'.
	   Both formulae and CNF clauses are stored.
***************************************************************/
{
  BOOL isAxiom;

  /* First read formulae */
  while (dfg_LOOKAHEAD == DFG_FORMLIST) {
    isAxiom = TRUE;

    dfg_Match(DFG_FORMLIST);
    dfg_Match('(');
    if (dfg_LOOKAHEAD == DFG_AXIOMS)
      dfg_Match(DFG_AXIOMS);
    else {
      isAxiom = FALSE;
      dfg_Match(DFG_CONJECS);
    }
    dfg_Match(')');
    dfg_Match('.');
    /* Use dfg_AXIOMLIST or dfg_ConjectureList for the formulae */
    dfg_FormulaList(isAxiom);
    dfg_Match(DFG_ENDLIST);
  }   /* while */

  /* Now read CNF clauses */
  while (dfg_LOOKAHEAD == DFG_CLAUSELIST) {
    isAxiom = TRUE;

    dfg_Match(DFG_CLAUSELIST);
    dfg_Match('(');
    if (dfg_LOOKAHEAD == DFG_AXIOMS)
      dfg_Match(DFG_AXIOMS);
    else {
      isAxiom = FALSE;
      dfg_Match(DFG_CONJECS);
    }
    dfg_Match(',');
    if (dfg_LOOKAHEAD == DFG_CNF) {
      dfg_Match(DFG_CNF);
      dfg_Match(')');
      dfg_Match('.');
      dfg_ClauseList(isAxiom);   /* Use dfg_CLAUSELIST for the clauses */
    } else
      while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF)
	dfg_Match(dfg_LOOKAHEAD);        /* Ignore clauses in DNF */
    
    dfg_Match(DFG_ENDLIST);
  }   /* while */

  /* Ignore proofs */
  while (dfg_LOOKAHEAD == DFG_PROOF) {
    do
      dfg_Match(dfg_LOOKAHEAD);
    while (dfg_LOOKAHEAD != DFG_ENDLIST && dfg_LOOKAHEAD != EOF);
    dfg_Match(DFG_ENDLIST);
  }
}

/**************************************************************/
/* Common functions for clause or formula input               */
/**************************************************************/

static TERM dfg_Atom(BOOL IsInSort)
/**************************************************************
  INPUT:   A boolean value
  RETURNS: An atom term with a predicate, 'equal', 'true'
           or 'false' as top symbol.
  EFFECT:  This function parses a single atom.
           If <IsInSort> is TRUE, all occuring variables
	   are added to the list of quantified variables.
***************************************************************/
{
  LIST   argList   = list_Nil();
  SYMBOL topsymbol;

  switch (dfg_LOOKAHEAD) {
  case DFG_PREDICATE:
    topsymbol = dfg_LVAL.symbol;
    dfg_Match(DFG_PREDICATE);
    if (dfg_LOOKAHEAD == '(') {
      dfg_Match('(');
      argList = dfg_TermList(IsInSort);
      dfg_Match(')');
    }
    dfg_SymCheck(topsymbol, list_Length(argList));
    break;
  case DFG_EQUAL:
    topsymbol = fol_Equality();
    dfg_Match(DFG_EQUAL);
    dfg_Match('(');
    argList = list_List(dfg_Term(IsInSort));            /* 1st argument */
    dfg_Match(',');
    list_Rplacd(argList, list_List(dfg_Term(FALSE)));  /* 2nd argument */
    dfg_Match(')');
    break;
  case DFG_TRUE:
    topsymbol = fol_True();
    dfg_Match(DFG_TRUE);
    break;
  case DFG_FALSE:
    topsymbol = fol_False();
    dfg_Match(DFG_FALSE);
    break;
  default:
    dfg_Match(DFG_PREDICATE);  /* Create error "predicate expected" */
    /* Unreachable. Just added to please the ignorant compiler ;-) */
    return (TERM) NULL;
  }
  
  return term_Create(topsymbol, argList);
}


static TERM dfg_Term(BOOL IsInSort)
/**************************************************************
  INPUT:   None.
  RETURNS: A term with a variable or function as top symbol.
  EFFECT:  This function parses a single term.
***************************************************************/
{
  SYMBOL topsymbol;
  LIST   argList = list_Nil();
  if (dfg_LOOKAHEAD == DFG_ID) {
    if (IsInSort)
      topsymbol = dfg_VarDefine(dfg_LVAL.string);
    else
      topsymbol = dfg_VarLookup(dfg_LVAL.string);
    dfg_Match(DFG_ID);
  }
  else {
    topsymbol = dfg_LVAL.symbol;
    dfg_Match(DFG_FUNCTION);
    if (dfg_LOOKAHEAD == '(') {
      dfg_Match('(');
      argList = dfg_TermList(IsInSort);
      dfg_Match(')');
    }
    dfg_SymCheck(topsymbol, list_Length(argList));
  }
  return term_Create(topsymbol, argList);
}


static LIST dfg_TermList(BOOL IsInSort)
/**************************************************************
  INPUT:   None.
  RETURNS: A list of terms with a variable or function as top
           symbol.
  EFFECT:  This function parses a comma separated list of terms.
***************************************************************/
{
  LIST result = list_List(list_Nil()), last = result;
  BOOL loop = TRUE;

  do {
    TERM term = dfg_Term(IsInSort);
    if (dfg_LOOKAHEAD == ',')
      dfg_Match(',');
    else
      loop = FALSE;

    /* Insert at end of result list */
    list_Rplacd(last, list_List(term));
    last = list_Cdr(last);    
  } while (loop);

  result = list_Pop(result); /* Remove the empty item from the beginning */
  return result;
}

static LIST dfg_VarList(void)
/**************************************************************
  INPUT:   None.
  RETURNS: A list of variable symbols.
  EFFECT:  This function parses a comma separated list of variables.
***************************************************************/
{
  LIST result = list_List(list_Nil()), last = result;
  BOOL loop = TRUE;
 
  dfg_Match('[');
 
  do {
    SYMBOL variable;
    TERM   vterm = NULL;
    if (dfg_LOOKAHEAD == DFG_ID) {
      variable = dfg_VarDefine(dfg_LVAL.string);
      vterm = term_Create(variable, list_Nil());
    }
    dfg_Match(DFG_ID);
    /* Insert at end of result list */
    list_Rplacd(last, list_List(vterm));
    last = list_Cdr(last);

    if (dfg_LOOKAHEAD == ',')
      dfg_Match(',');
    else
      loop = FALSE;
  } while (loop);
 
  dfg_Match(']');
 
  result = list_Pop(result); /* Remove the empty item from the beginning */
  return result;
}

static LIST dfg_QuantifierTermList(void)
/**************************************************************
  INPUT:   None.
  RETURNS: A list of variable symbols.
  EFFECT:  This function parses a comma separated list of variables
           and sorts..
***************************************************************/
{
  LIST   result = list_List(list_Nil()), last = result;
  BOOL   loop = TRUE;
  SYMBOL topsymbol;
  TERM   term;

  dfg_Match('[');

  do {
    if (dfg_LOOKAHEAD == DFG_ID) {
      topsymbol = dfg_VarDefine(dfg_LVAL.string);
      term = term_Create(topsymbol, list_Nil());
      dfg_Match(DFG_ID);
    } else 
      term = dfg_Atom(TRUE);  /* We are in a sort */

    /* Insert at end of result list */
    list_Rplacd(last, list_List(term));
    last = list_Cdr(last);

    if (dfg_LOOKAHEAD == ',')
      dfg_Match(',');
    else
      loop = FALSE;
  } while (loop);

  dfg_Match(']');

  result = list_Pop(result); /* Remove the empty item from the beginning */
  return result;
}

static void dfg_Description(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses the description section
           starting with 'list_of_descriptions' and ending with
	   'end_of_list'.
***************************************************************/
{
  dfg_Match(DFG_DESCLIST);
  dfg_Match(DFG_NAME);
  dfg_Match('(');
  dfg_Match('{');
  if (dfg_LOOKAHEAD != '*')
    dfg_Match('*');    /* Create an error */
  dfg_DESC.name = dfg_ReadText(TRUE);  /* Returns text within {* and *} */
  dfg_LOOKAHEAD = dfg_Scanner();
  dfg_Match('*');
  dfg_Match('}');
  dfg_Match(')');
  dfg_Match('.');
  dfg_Match(DFG_AUTHOR);
  dfg_Match('(');
  dfg_Match('{');
  if (dfg_LOOKAHEAD != '*')
    dfg_Match('*');    /* Create an error */
  dfg_DESC.author = dfg_ReadText(TRUE);
  dfg_LOOKAHEAD = dfg_Scanner();
  dfg_Match('*');
  dfg_Match('}');
  dfg_Match(')');
  dfg_Match('.');
  if (dfg_LOOKAHEAD == DFG_VERSION) {
    /* optional version information */
    dfg_Match(DFG_VERSION);
    dfg_Match('(');
    dfg_Match('{');
    if (dfg_LOOKAHEAD != '*')
      dfg_Match('*');    /* Create an error */
    dfg_DESC.version = dfg_ReadText(TRUE);
    dfg_LOOKAHEAD = dfg_Scanner();
    dfg_Match('*');
    dfg_Match('}');
    dfg_Match(')');
    dfg_Match('.');
  }
  if (dfg_LOOKAHEAD == DFG_LOGIC) {
    /* optional logic information */
    dfg_Match(DFG_LOGIC);
    dfg_Match('(');
    dfg_Match('{');
    if (dfg_LOOKAHEAD != '*')
      dfg_Match('*');    /* Create an error */
    dfg_DESC.logic = dfg_ReadText(TRUE);
    dfg_LOOKAHEAD = dfg_Scanner();
    dfg_Match('*');
    dfg_Match('}');
    dfg_Match(')');
    dfg_Match('.');
  }
  dfg_Match(DFG_STATUS);
  dfg_Match('(');
  if (dfg_LOOKAHEAD == DFG_SATISFIABLE) {
    dfg_DESC.status = SATISFIABLE;
    dfg_Match(DFG_SATISFIABLE);
  } else if (dfg_LOOKAHEAD == DFG_UNSATISFIABLE) {
    dfg_DESC.status = UNSATISFIABLE;
    dfg_Match(DFG_UNSATISFIABLE);
  } else {
    dfg_DESC.status = UNKNOWN;
    dfg_Match(DFG_UNKNOWN);
  }
  dfg_Match(')');
  dfg_Match('.');
  dfg_Match(DFG_DESCRIPTION);
  dfg_Match('(');
  dfg_Match('{');
  if (dfg_LOOKAHEAD != '*')
    dfg_Match('*');    /* Create an error */
  dfg_DESC.description = dfg_ReadText(TRUE);
  dfg_LOOKAHEAD = dfg_Scanner();
  dfg_Match('*');
  dfg_Match('}');
  dfg_Match(')');
  dfg_Match('.');
  if (dfg_LOOKAHEAD == DFG_DATE) {
    dfg_Match(DFG_DATE);
    dfg_Match('(');
    dfg_Match('{');
    if (dfg_LOOKAHEAD != '*')
      dfg_Match('*');    /* Create an error */
    dfg_DESC.date = dfg_ReadText(TRUE);
    dfg_LOOKAHEAD = dfg_Scanner();
    dfg_Match('*');
    dfg_Match('}');
    dfg_Match(')');
    dfg_Match('.');
  }
  dfg_Match(DFG_ENDLIST);
}


static void dfg_SymbolList(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses the symbol declaration section
           starting with 'list_of_symbols' and ending with
	   'end_of_list'.
***************************************************************/
{
  dfg_Match(DFG_SYMBOLLIST);  

  if (dfg_LOOKAHEAD == DFG_FUNCTIONS)
    dfg_SymbolDeclWithArity(DFG_FUNCTIONS);
  if (dfg_LOOKAHEAD == DFG_PREDICATS)
    dfg_SymbolDeclWithArity(DFG_PREDICATS);
  if (dfg_LOOKAHEAD == DFG_SORTS)
    dfg_SortSymbolDecl();
  if (dfg_LOOKAHEAD == DFG_OPERATORS)
    dfg_SymbolDeclWithArity(DFG_OPERATORS);
  if (dfg_LOOKAHEAD == DFG_QUANTIFS)
    dfg_SymbolDeclWithArity(DFG_QUANTIFS);
  
  dfg_Match(DFG_ENDLIST);
}


static void dfg_SymbolDeclWithArity(int symboltype)
/**************************************************************
  INPUT:   A token defining a symbol type. 
  RETURNS: Nothing.
  EFFECT:  This function parses a single list of symbol
           declarations. 'symboltype' defines the type of all
	   symbols declared in this list. The list starts with
	   the symbol type and ends with "]."
***************************************************************/
{
  long   arity;
  BOOL   loop = TRUE, specified;
  char*  name = NULL;
  SYMBOL actSymbol;

  dfg_Match(symboltype);     /* Matches the symbol type token */
  dfg_Match('[');
  
  do {
    arity = -1;          /* arbitrary */
    specified = FALSE;   /* Assume arity is not specified */

    if (dfg_LOOKAHEAD == DFG_NUM || dfg_LOOKAHEAD == DFG_ID) {
      name = dfg_LVAL.string;
      dfg_LVAL.string = NULL;
      dfg_Match(dfg_LOOKAHEAD);   /* Match DFG_NUM or DFG_ID */
    } else {
      /* Symbol is specified with arity */
      specified = TRUE;
      dfg_Match('(');
      if (dfg_LOOKAHEAD == DFG_NUM || dfg_LOOKAHEAD == DFG_ID) {
	name = dfg_LVAL.string;
	dfg_LVAL.string = NULL;
	dfg_Match(dfg_LOOKAHEAD);
      } else
	/* neither NUM nor ID -> create an error */
	dfg_Match(DFG_ID);

      dfg_Match(',');
      arity = dfg_LVAL.number;
      dfg_Match(DFG_NUM);
      dfg_Match(')');
    }

    /* Check if this symbol was declared earlier */
    actSymbol = symbol_Lookup(name);
    if (actSymbol != 0) {
      /* Symbol was declared before */
      /* Check if the old and new symbol type are equal */
      if ((symboltype == DFG_FUNCTIONS && !symbol_IsFunction(actSymbol)) ||
	  (symboltype == DFG_PREDICATS && !symbol_IsPredicate(actSymbol)) ||
	  (symboltype == DFG_JUNCTOR   && !symbol_IsJunctor(actSymbol))) {
	fprintf(stderr, "Line %u: symbol %s was already declared as ",
		dfg_ActLineNumber(), name);
	switch (symbol_Type(actSymbol)) {
	case symbol_CONSTANT:
	case symbol_FUNCTION:
	  fputs("function\n", stderr); break;
	case symbol_PREDICATE:
	  fputs("predicate\n", stderr); break;
	case symbol_JUNCTOR:
	  fputs("junctor", stderr); break;
	default:
	  fputs("unknown type\n", stderr);
	}
	misc_Error();
      }
      /* Now check the old and new arity if specified */
      if (specified && arity != symbol_Arity(actSymbol)) {
	fprintf(stderr, "Line %u: symbol %s was already declared with arity %d\n",
		dfg_ActLineNumber(), name, symbol_Arity(actSymbol));
	misc_Error();
      }
    } else {
      /* Symbol was not declared before */
      switch (symboltype) {
      case DFG_FUNCTIONS:
	actSymbol = symbol_CreateFunction(name, arity, symbol_STATLEX);
	break;
      case DFG_PREDICATS:
	actSymbol = symbol_CreatePredicate(name, arity, symbol_STATLEX);
	break;
      default:
	actSymbol = symbol_CreateJunctor(name, arity, symbol_STATLEX);
      }
      if (!specified)
	dfg_SymAdd(actSymbol);
    }

    dfg_StringFree(name);  /* name was copied */

    if (dfg_LOOKAHEAD == ',')
      dfg_Match(',');
    else
      loop = FALSE;
  } while (loop);

  dfg_Match(']');
  dfg_Match('.');
}

static void dfg_SortSymbolDecl(void)
/**************************************************************
  INPUT:   None. 
  RETURNS: Nothing.
  EFFECT:  This function parses a single list of sort symbol
           declarations. The list starts with "sorts[" and
	   ends with "].". All symbols in this list are declared
	   as monadic predicates.
***************************************************************/
{
  BOOL   loop = TRUE;
  char*  name = NULL;
  SYMBOL actSymbol;

  dfg_Match(DFG_SORTS);
  dfg_Match('[');
  
  do {
    if (dfg_LOOKAHEAD == DFG_NUM || dfg_LOOKAHEAD == DFG_ID) {
      name = dfg_LVAL.string;
      dfg_LVAL.string = NULL;
      dfg_Match(dfg_LOOKAHEAD);   /* Match DFG_NUM or DFG_ID */
    } else
      dfg_Match(DFG_ID);          /* Creates an error */

    /* Check if this symbol was declared earlier */
    actSymbol = symbol_Lookup(name);
    if (actSymbol != 0) {
      /* Symbol was declared before */
      fprintf(stderr, "Line %u: symbol %s was already declared as",
	      dfg_ActLineNumber(), name);
      switch (symbol_Type(actSymbol)) {
      case symbol_CONSTANT:
      case symbol_FUNCTION:
	fputs("function", stderr); break;
      case symbol_PREDICATE:
	fputs("predicate", stderr); break;
      case symbol_JUNCTOR:
	fputs("junctor", stderr); break;
      default:
	fputs("unknown type", stderr);
      }
      fprintf(stderr, " with arity %d\n", symbol_Arity(actSymbol));
      misc_Error();
    } else
      /* Symbol was not declared before */
      symbol_CreatePredicate(name, 1, symbol_STATLEX);

    dfg_StringFree(name);  /* name was copied */

    if (dfg_LOOKAHEAD == ',')
      dfg_Match(',');
    else
      loop = FALSE;
  } while (loop);

  dfg_Match(']');
  dfg_Match('.');
}

static void dfg_DeclList(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses the declaration section from
           'list_of_declarations' to 'end_of_list'.
***************************************************************/
{
  TERM term;

  if (dfg_LOOKAHEAD == DFG_DECLLIST) {
    dfg_LOOKUPSYMBOLS = TRUE;
    dfg_Match(DFG_DECLLIST);

    while (dfg_LOOKAHEAD != DFG_ENDLIST) {
      if (dfg_LOOKAHEAD == DFG_SUBSORT) {
	SYMBOL s1, s2;
	TERM   t1, t2, varterm;

	dfg_Match(DFG_SUBSORT);
	/* Tell the scanner to search for symbol names */
	dfg_Match('(');
	s1 = dfg_LVAL.symbol;
	dfg_Match(DFG_PREDICATE);
	dfg_SymCheck(s1, 1); /* Must be a monadic predicate */
	dfg_Match(',');
	s2 = dfg_LVAL.symbol;
	dfg_Match(DFG_PREDICATE);
	dfg_SymCheck(s2, 1); /* Must be a monadic predicate */
	dfg_Match(')');
	dfg_Match('.');

	varterm = term_Create(symbol_CreateStandardVariable(), list_Nil());
	symbol_ResetStandardVarCounter();

	t1   = term_Create(s1, list_List(varterm));
	t2   = term_Create(s2, list_List(term_Copy(varterm)));
	term = term_Create(fol_Implies(), list_Cons(t1, list_List(t2)));
	term = fol_CreateQuantifier(fol_All(), list_List(term_Copy(varterm)),
				    list_List(term));
	list_Rplacd(dfg_LASTAXIOM, list_List(term));
	dfg_LASTAXIOM = list_Cdr(dfg_LASTAXIOM);
      } else if (dfg_LOOKAHEAD == DFG_SORTPREDICATE) {
	/* ignore "predicate(...)." */
	dfg_Match(DFG_SORTPREDICATE);
	dfg_LOOKUPSYMBOLS = FALSE;
	dfg_Match('(');
	dfg_Match(DFG_ID);
	while (dfg_LOOKAHEAD == ',') {
	  dfg_Match(',');
	  dfg_Match(DFG_ID);
	}
	dfg_Match(')');
	dfg_LOOKUPSYMBOLS = TRUE;
	dfg_Match('.');
      } else if (dfg_LOOKAHEAD == DFG_FORALL) {
	LIST arglist, termlist, varlist, lastvar, sortlist, lastsort, scan;

	dfg_Match(DFG_FORALL); /* Match the quantifier itself */
	dfg_LOOKUPSYMBOLS = TRUE;
	dfg_Match('(');
	termlist = dfg_QuantifierTermList();
	/* May contain variables or atoms */
	dfg_Match(',');
	arglist = list_List(dfg_Atom(FALSE));   /* instead of dfg_Formula */
	dfg_Match(')');

	dfg_VarListDelete(); /* Delete list of quantified variables */
    
	varlist  = lastvar  = list_List(list_Nil());
	sortlist = lastsort = list_List(list_Nil());

	for ( ; !list_Empty(termlist); termlist = list_Pop(termlist)) {
	  term = list_Car(termlist);
	  
	  if (term_IsVariable(term)) {
	    list_Rplacd(lastvar, list_List((POINTER) term_TopSymbol(term)));
	    lastvar = list_Cdr(lastvar);
	    term_Delete(term);
	  } else {
	    list_Rplacd(lastsort, list_List(term));
	    lastsort = list_Cdr(lastsort);
	    list_Rplacd(lastvar, term_VariableSymbols(term));
	    while (!list_Empty(list_Cdr(lastvar)))
	      lastvar = list_Cdr(lastvar);
	    /* Set lastvar to the new end of list */
	  }
	}

	varlist  = list_Pop(varlist);
	sortlist = list_Pop(sortlist);

	varlist = list_PointerDeleteDuplicates(varlist);
	/* Transform variable symbols in varlist to terms */
	for (scan = varlist; !list_Empty(scan); scan = list_Cdr(scan))
	  list_Rplaca(scan, term_Create((SYMBOL)list_Car(scan), list_Nil()));

	if (!list_Empty(sortlist)) {
	  /* Add the conjunction of all sort terms to the front of arglist */
	  if (list_Empty(list_Cdr(sortlist))) {
	    list_Rplacd(sortlist, arglist);
	    arglist = sortlist;
	  } else
	    arglist = list_Cons(term_Create(fol_And(), sortlist), arglist);
	  /* The conjunction of sort terms implies the quantified term */
	  arglist = list_List(term_Create(fol_Implies(), arglist));
	}
     
	term = fol_CreateQuantifier(fol_All(), varlist, arglist);

	list_Rplacd(dfg_LASTAXIOM, list_List(term));
	dfg_LASTAXIOM = list_Cdr(dfg_LASTAXIOM);
      } else {
	/* Must be an atom without variables */
	term = dfg_Atom(FALSE);
	list_Rplacd(dfg_LASTAXIOM, list_List(term));
	dfg_LASTAXIOM = list_Cdr(dfg_LASTAXIOM);
      }
    }
    
    dfg_LOOKUPSYMBOLS = FALSE;
    dfg_Match(DFG_ENDLIST);
  }
}

static void dfg_Settings(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses the settings section from
           'list_of_settings(SPASS)' to 'end_of_list'.
***************************************************************/
{
  while (dfg_LOOKAHEAD == DFG_SETTINGS) {
    dfg_Match(DFG_SETTINGS);
    dfg_Match('(');
    if (dfg_LOOKAHEAD == DFG_SPASS) {
      dfg_Match(DFG_SPASS);
      dfg_Match(')');
      dfg_Match('.');
      dfg_Match('{');
      dfg_Match('*');
      while (dfg_LOOKAHEAD != '*') {
	if (dfg_LOOKAHEAD == DFG_ID) {
	  if (strcmp(dfg_LVAL.string, "set_precedence") == 0) {
	    dfg_Match(DFG_ID);
	    dfg_PrecedenceList();
	    dfg_Match('.');
	  } else if (strcmp(dfg_LVAL.string , "set_flag") == 0) {
	    dfg_Match(DFG_ID);
	    dfg_Match('(');
	    dfg_Flag();
	    dfg_Match(')');
	    dfg_Match('.');
	  } else {
	    fprintf(stderr, "Line %u: Unknown command %s in settings"
		    " section\n", dfg_ActLineNumber(), dfg_LVAL.string);
	    misc_Error();
	  }
	} else
	  /* Not DFG_ID */
	  dfg_Match('*');  /* Create an error */
      }   /* while */
    } else {
      /* Settings for another Prover */
      dfg_Match(DFG_ID);  /* ?????? */
      dfg_Match(')');
      dfg_Match('.');
      dfg_Match('{');
      if (dfg_LOOKAHEAD != '*')
	dfg_Match('*');   /* Creates an error */
      dfg_ReadText(FALSE);  /* Read text within {* and *} and ignore it */
    }
    dfg_Match('*');
    dfg_Match('}');
    dfg_Match(DFG_ENDLIST);
  } /* while  Lookahead == DFG_SETTINGS */
}


static void dfg_PrecedenceList(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses a comma separated list of symbols
           whose precedence will be set according to their order
	   of appearance.
	   The weight of a symbol may also be declared.
***************************************************************/
{
  BOOL loop = TRUE;

  dfg_LOOKUPSYMBOLS = TRUE;
  dfg_Match('(');

  while (loop) {
    SYMBOL    symbol;

    symbol = (SYMBOL)NULL;
    if (dfg_LOOKAHEAD == '(') {
      /* Symbol specified with weight */
      dfg_Match('(');
      if (dfg_LOOKAHEAD >= DFG_FUNCTION && dfg_LOOKAHEAD <= DFG_JUNCTOR) {
	symbol = dfg_LVAL.symbol;
	dfg_LOOKUPSYMBOLS = FALSE;
	dfg_Match(dfg_LOOKAHEAD);   /* Match the symbol */
	dfg_Match(',');
	symbol_SetWeight(symbol, (int) dfg_LVAL.number);
	dfg_Match(DFG_NUM);
	if (dfg_LOOKAHEAD == ',') {    /* Found a tripel with symbol status */
	  dfg_LOOKUPSYMBOLS = FALSE;
	  dfg_Match(',');
	  if (dfg_LOOKAHEAD == DFG_ID) {
	    if (strcmp(dfg_LVAL.string, "r") == 0)
	      /* Symbol status ORDRIGHT */
	      symbol_AddProperty(symbol, ORDRIGHT);
	    else if (strcmp(dfg_LVAL.string, "m") == 0)
	      /* Symbol status ORDMUL */
	      symbol_AddProperty(symbol, ORDMUL);
	    else if (strcmp(dfg_LVAL.string, "l") != 0) {
	      fprintf(stderr, "Line %u: Unknown symbol status \"%s\"\n\n",
		      dfg_ActLineNumber(), dfg_LVAL.string);
	      dfg_FPrintErrorLine(stderr);
	      misc_Error();
	    }
	  }
	  
	  dfg_Match(DFG_ID); /* Match the symbol status */
  	  dfg_LOOKUPSYMBOLS = TRUE;
	}
	dfg_Match(')');
      } else
	dfg_Match(DFG_FUNCTION);  /* Creates an error */
    } else if (dfg_LOOKAHEAD >= DFG_FUNCTION && dfg_LOOKAHEAD <= DFG_JUNCTOR) {
      symbol = dfg_LVAL.symbol;
      dfg_LOOKUPSYMBOLS = FALSE;
      dfg_Match(dfg_LOOKAHEAD);   /* Match the symbol */
    } else
      dfg_Match(DFG_FUNCTION); /* Creates an error */

    symbol_SetOrdering(symbol, symbol_GetIncreasedOrderingCounter());
    
    if (dfg_LOOKAHEAD == ',') {
      dfg_LOOKUPSYMBOLS = TRUE;
      dfg_Match(',');
    } else
      loop = FALSE;
  }

  dfg_Match(')');
}


static void dfg_Flag(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function parses a single flag definition line
           within the SPASS settings section.
***************************************************************/
{
  int flag;

  if (dfg_LOOKAHEAD == DFG_ID) {
    char* flagname = dfg_LVAL.string;

    flag = flag_Id(flagname);
    if (flag == -1) {
      fprintf(stderr, "Line %u: Found unkown flag %s\n",
	     dfg_ActLineNumber(), flagname);
      misc_Error();
    }
    dfg_Match(DFG_ID);
    dfg_Match(',');
    if (dfg_LOOKAHEAD == DFG_NUM)
      flag_SetFlagValue(flag, dfg_LVAL.number);
    dfg_Match(DFG_NUM);   /* Creates an error if Lookahead isn't NUM */
  } else
    dfg_Match(DFG_ID);    /* Creates an error because Lookahead isn't ID */
}


/**************************************************************/
/* Basic Functions                                            */
/**************************************************************/

static void dfg_Match(int Token)
/**************************************************************
  INPUT:   A token.
  RETURNS: Nothing.
  EFFECT:  This function compares 'Token' and the actually read
           token. If they match the next token is scanned
	   else an error message is printed to stderr.
***************************************************************/
{
  if (dfg_LOOKAHEAD == Token)
    dfg_LOOKAHEAD = dfg_Scanner();
  else {
    fprintf(stderr, "Line %u: Parse error, \"", dfg_ActLineNumber());
    dfg_FPrintToken(stderr, dfg_LOOKAHEAD); fputs("\" found, \"", stderr);
    dfg_FPrintToken(stderr, Token); fputs("\" expected\n\n", stderr);
    dfg_FPrintErrorLine(stderr);
    misc_Error();
  }
}


static void dfg_Warning(const char* Warning)
/**************************************************************
  INPUT:   A warning message.
  RETURNS: Nothing.
  EFFECT:  The actual line number and the message are printed
           to stderr.
***************************************************************/
{
  fprintf(stderr, "Line %u: %s\n\n", dfg_ActLineNumber(), Warning);
}


static CLAUSE dfg_CreateClauseFromTerm(TERM Clause, BOOL IsAxiom)
/**************************************************************
  INPUT:   A clause term and a boolean value.
  RETURNS: The clause term converted to a CLAUSE.
  EFFECT:  This function converts a clause stored as term
           into an EARL clause structure.
	   If 'IsAxiom' is TRUE the clause is treated as axiom
	   clause else as conjecture clause.
  MEMORY:  The clause term is deleted.
***************************************************************/
{
  LIST   literals, conLits, antLits, sucLits;
  TERM   literal;
  CLAUSE result;
  
  if (term_TopSymbol(Clause) == fol_All()) {
    /* Remove and free the quantifier and the OR term */
    literals = term_ArgumentList(term_SecondArgument(Clause));
    term_RplacArgumentList(term_SecondArgument(Clause), list_Nil());
  } else {
    /* Remove and free the OR term */
    literals = term_ArgumentList(Clause);
    term_RplacArgumentList(Clause, list_Nil());
  }
  term_Delete(Clause);

  conLits = antLits = sucLits = list_Nil();

  for (; !list_Empty(literals); literals = list_Pop(literals)) {
    literal = (TERM) list_Car(literals);
    if (symbol_IsPredicate(term_TopSymbol(literal))) {  /* Positive literal */
      if (fol_IsTrue(literal)) {
	/* Clause is a tautology */
	list_Delete(conLits);
	list_Delete(antLits);
	list_Delete(sucLits);
	list_DeleteWithElement(literals,term_Delete);
	return (CLAUSE) NULL;
      } else if (fol_IsFalse(literal))
	/* Ignore this literal */
	term_Delete(literal);
      else {
	/* Found a succedent literal */
	sucLits = list_Cons(literal, sucLits);
      }
    } else {  /* Found an antecedent literal */
      TERM atom = term_FirstArgument(literal);
      if (fol_IsFalse(atom)) {
	/* Clause is a tautology */
	list_Delete(conLits);
	list_Delete(antLits);
	list_Delete(sucLits);
	list_DeleteWithElement(literals,term_Delete);
	return (CLAUSE) NULL;
      } else if (fol_IsTrue(atom))
	/* Ignore this literal */
	term_Delete(literal);
      else {
	antLits = list_Cons(literal, antLits);
      }
    }
  }

  result = clause_CreateCrude(conLits,antLits,sucLits,!IsAxiom);

  list_Delete(conLits);
  list_Delete(antLits);
  list_Delete(sucLits);

  return result;
}


/**************************************************************/
/* Functions for the Variable Table                           */
/**************************************************************/

typedef struct {
  char*  name;
  SYMBOL symbol;
} DFG_VARENTRY, *DFG_VAR;

static __inline__ char* dfg_VarName(DFG_VAR Entry)
{
  return Entry->name;
}

static __inline__ SYMBOL dfg_VarSymbol(DFG_VAR Entry)
{
  return Entry->symbol;
}

static __inline__ DFG_VAR dfg_VarCreate(void)
{
  return (DFG_VAR) memory_Malloc(sizeof(DFG_VARENTRY));
}

static __inline__ void dfg_VarFree(DFG_VAR Entry)
{
  dfg_StringFree(dfg_VarName(Entry));
  memory_Free(Entry, sizeof(DFG_VARENTRY));
}

static SYMBOL dfg_VarDefine(const char* Name)
/**************************************************************
  INPUT:   A variable name.
  RETURNS: The corresponding variable symbol.
  EFFECT:  Marks the variable with name <Name> as quantified.
           If the variable name wasn't quantified before,
           a new variable symbol is created.
***************************************************************/
{
  LIST scan = dfg_VARLIST;

  while (!list_Empty(scan) &&
	 (strcmp(dfg_VarName(list_Car(scan)), Name) != 0))
    scan = list_Cdr(scan);

  if (list_Empty(scan)) {
    /* Variable not found */
    DFG_VAR newEntry = dfg_VarCreate();
    newEntry->name   = dfg_StringCopy(Name);
    newEntry->symbol = symbol_CreateStandardVariable();
    dfg_VARLIST = list_Cons(newEntry,dfg_VARLIST);
    return dfg_VarSymbol(newEntry);
  } else
    return dfg_VarSymbol(list_Car(scan));
}

static SYMBOL dfg_VarLookup(const char* Name)
/**************************************************************
  INPUT:   A variable name.
  RETURNS: The corresponding variable symbol.
  EFFECT:  If the variable name wasn't quantified before,
           an error message is printed and the program exits.
***************************************************************/
{
  LIST scan = dfg_VARLIST;

  while (!list_Empty(scan) &&
	 (strcmp(dfg_VarName(list_Car(scan)), Name) != 0))
    scan = list_Cdr(scan);

  if (list_Empty(scan)) {
    /* Variable not found */
    fprintf(stderr, "Line %u: Free Variable %s\n\n",
	    dfg_ActLineNumber(), Name);
    dfg_FPrintErrorLine(stderr);
    misc_Error();
  }

  return dfg_VarSymbol(list_Car(scan));
}


static void dfg_VarListDelete(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Frees the memory used by the variable list.
***************************************************************/
{
  symbol_ResetStandardVarCounter();
  while (!list_Empty(dfg_VARLIST)) {
    dfg_VarFree(list_Car(dfg_VARLIST));
    dfg_VARLIST = list_Pop(dfg_VARLIST);
  }
}

/**************************************************************/
/* Functions for the Symbol Table                             */
/**************************************************************/

typedef struct {
    SYMBOL symbol;
    BOOL   valid;
    int    arity;
} DFG_SYMENTRY, *DFG_SYM;

static __inline__ SYMBOL dfg_SymSymbol(DFG_SYM Entry)
{
  return Entry->symbol;
}

static __inline__ int dfg_SymArity(DFG_SYM Entry)
{
  return Entry->arity;
}

static __inline__ BOOL dfg_SymValid(DFG_SYM Entry)
{
  return Entry->valid;
}

static __inline__ DFG_SYM dfg_SymCreate(void)
{
  return (DFG_SYM) memory_Malloc(sizeof(DFG_SYMENTRY));
}

static __inline__ void dfg_SymFree(DFG_SYM Entry)
{
  memory_Free(Entry, sizeof(DFG_SYMENTRY));
}


static void dfg_SymAdd(SYMBOL Symbol)
/**************************************************************
  INPUT:   A symbol.
  RETURNS: Nothing.
  EFFECT:  This function adds 'Symbol' to the symbol list.
           The arity of these symbols will be checked every time
	   the symbol occurs.
***************************************************************/
{
  DFG_SYM newEntry = dfg_SymCreate();
  newEntry->symbol = Symbol;
  newEntry->valid  = FALSE;
  newEntry->arity  = 0;
  dfg_SYMBOLLIST = list_Cons(newEntry, dfg_SYMBOLLIST);
}


static void dfg_SymCheck(SYMBOL Symbol, NAT Arity)
/**************************************************************
  INPUT:   A symbol and the current arity of this symbol.
  RETURNS: Nothing.
  EFFECT:  This function compares the previous arity of 'Symbol'
           with the actual 'Arity'. If these values differ
	   the symbol's arity is set to arbitrary.
	   The arity of symbols whose arity was specified in
	   the symbol declaration section is checked and a warning
	   is printed to stderr in case of differences.
***************************************************************/
{
  LIST scan = dfg_SYMBOLLIST;
  while (!list_Empty(scan)) {
    DFG_SYM actEntry = (DFG_SYM) list_Car(scan);
    if (actEntry->symbol == Symbol) {
      if (actEntry->valid) {
	if (actEntry->arity != Arity)
	  actEntry->arity = -1;    /* Arbitrary Arity */
      } else {
	/* Not valid => first time */
	actEntry->arity = Arity;
	actEntry->valid = TRUE;
      }
      return;
    }
    scan = list_Cdr(scan);
  }

  /* Symbol isn't in SymbolList, so its arity was specified.        */
  /* Check if the specified arity corresponds with the actual arity */
  if (symbol_Arity(Symbol) != symbol_ArbitraryArity() &&
      symbol_Arity(Symbol) != Arity) {
    fprintf(stderr, "Line %u: Symbol ", dfg_ActLineNumber());
    symbol_FPrint(stderr, Symbol);
    fprintf(stderr, " was declared with arity %u\n", symbol_Arity(Symbol));
    misc_Error();
  }
}


static void dfg_SymCleanUp(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function corrects all symbols whose arity wasn't
           specified in the symbol declaration section but seem
	   to occur with always the same arity.
	   The memory for the symbol list is freed.
***************************************************************/
{
  while (!list_Empty(dfg_SYMBOLLIST)) {
    DFG_SYM actEntry  = (DFG_SYM) list_Car(dfg_SYMBOLLIST);
    SYMBOL  actSymbol = actEntry->symbol;

    if (actEntry->arity != symbol_Arity(actSymbol))
      symbol_SetArity(actSymbol, actEntry->arity);

    dfg_SymFree(actEntry);
    dfg_SYMBOLLIST = list_Pop(dfg_SYMBOLLIST);
  }  /* while */
}
