/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *               INPUT OF SPASS CLAUSES                   * */
/* *                                                        * */
/* *  $Module:   DFGSCANNER                                 * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.64 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/05/26 12:32:51 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: dfgscanner.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include <ctype.h>
#include "dfgscanner.h"


/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

DFG_LVALTYPE dfg_LVAL;
BOOL         dfg_LOOKUPSYMBOLS;

/**************************************************************/
/* Local Variables                                            */
/**************************************************************/

/* The local input file buffer.                      */
/* The whole input file is read into this buffer.    */
static char* dfg_FileBuffer;
static NAT   dfg_FileBufferLength;

/* The actual scanner position.                      */
static char* dfg_ActPtr;

/* The scanner input file.                           */
static FILE* dfg_Input;

/**************************************************************/
/* Prototypes for static Functions                            */
/**************************************************************/

static void dfg_EatComment(void);
static int  dfg_Identifier(char* Name);


/**************************************************************/
/* Inline Functions                                           */
/**************************************************************/

static __inline__ void dfg_NextChar(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function reads the next character.
***************************************************************/
{
  dfg_ActPtr++;
}

static __inline__ void dfg_Skip(int N)
/**************************************************************
  INPUT:   The number of characters to be skipped.
  RETURNS: Nothing.
  EFFECT:  This function ignores N characters from the
           scanner's input.
  CAUTION: You must be sure not to exceed the file buffer.
***************************************************************/
{
  dfg_ActPtr += N;
}

static __inline__ void dfg_SkipLine(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function ignores the rest of the actual line.
***************************************************************/
{
  do 
    dfg_NextChar();
  while (*dfg_ActPtr != '\0' && *dfg_ActPtr != '\n');
}

static __inline__ void dfg_EatSpace(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function skips 'white space' characters.
***************************************************************/
{
  while (isspace(*dfg_ActPtr))
    dfg_NextChar();
}

    
/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void dfg_InitScanner(FILE* Input)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Initializes all internal scanner variables.
           Only used by the parser.
***************************************************************/
{
  dfg_LVAL.number      = 0;
  dfg_LVAL.string      = NULL;
  dfg_LVAL.symbol      = symbol__NULL;

  dfg_LOOKUPSYMBOLS    = FALSE;
  dfg_FileBuffer       = (char*) NULL;
  dfg_FileBufferLength = 0;
  dfg_ActPtr           = (char*) NULL;
  dfg_Input            = Input;

  /* Now read the entire file into dfg_FileBuffer */
  fseek(dfg_Input, 0, SEEK_END);
  dfg_FileBufferLength = ftell(dfg_Input) + 1;   /* File length + '\0' */
  if (dfg_FileBufferLength == 1) {
    /* File is empty! */
    fprintf(stderr, "Input file is empty!\n\n");
    misc_Error();
  }
  rewind(dfg_Input);
  dfg_FileBuffer = memory_Malloc(sizeof(char) * dfg_FileBufferLength);
  fread(dfg_FileBuffer, sizeof(char), dfg_FileBufferLength, dfg_Input);
  dfg_FileBuffer[dfg_FileBufferLength-1] = '\0';

  /* Set the scanner position to beginning of the file buffer. */
  dfg_ActPtr = dfg_FileBuffer;
}


void dfg_FreeScanner(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  Frees the memory used for the local file buffer.
           Only used by the parser.
***************************************************************/ 
{
  if (dfg_LVAL.string != NULL) {
    dfg_StringFree(dfg_LVAL.string);
    dfg_LVAL.string = NULL;
  }
  if (dfg_FileBufferLength != 0) {
    memory_Free(dfg_FileBuffer, dfg_FileBufferLength);
    dfg_FileBuffer = NULL;
  }
}


NAT dfg_ActLineNumber(void)
/**************************************************************
  INPUT:   None.
  RETURNS: The actual line number.
***************************************************************/
{
  NAT lineNumber = 1;
  char* ptr = dfg_FileBuffer;

#ifdef DEBUG
  if (dfg_FileBuffer == NULL) {
    fprintf(stderr, "Uninitialized file buffer in function dfg_ActLineNumber");
    misc_Error();
  }
#endif

  /* Count the newlines from the beginning to the actual position */
  while (ptr < dfg_ActPtr) {
    if (*ptr == '\n')
      lineNumber++;
    ptr++;
  }
  return lineNumber;
}


void dfg_FPrintErrorLine(FILE* Output)
/**************************************************************
  INPUT:   A file the error message is printed to.
  RETURNS: Nothing.
  EFFECT:  This function prints the contents of the current
           input line into the file. A linefeed is inserted to
           mark the current scanner position. The actual
	   position is the end of the first printed line.
***************************************************************/
{
  char* lineEnd   = strchr(dfg_ActPtr, '\n');
  char* lineStart = dfg_ActPtr;

  if (lineEnd != NULL)
    /* Found newline marking the end of the actual line */
    *lineEnd = '\0';
  /* Now search start of actual line */
  while (lineStart > dfg_FileBuffer && *lineStart != '\n')
    lineStart--;
  if (*lineStart == '\n')
    lineStart++;

  fprintf(Output, "%.*s    <----\n%s\n", dfg_ActPtr - lineStart,
	  lineStart, dfg_ActPtr);
}


int dfg_Scanner(void)
/**************************************************************
  INPUT:   None.
  RETURNS: A token. This might be a character (casted to int)
           or one of the grammar tokens of type 'dfg_token'
	   defined in 'dfgscanner.h.
  EFFECT:  This function reads the next token from the input file.
***************************************************************/
{
  int   result = EOF;
  char  *actToken, *ptr;
  char  nextc, *nextptr;  /* Points to the character after the actual token */
  int   tokenlength;
  BOOL  loop = TRUE;

  while (loop) {
    loop = FALSE;

    /* Read Whitespace and comments */
    dfg_EatComment();
    if (*dfg_ActPtr == '\0')
      return EOF;

    ptr = dfg_ActPtr;
    while (*ptr != '\0' && (isalnum(*ptr) || *ptr == '_'))
      ptr++;
    tokenlength = ptr - dfg_ActPtr;
    if (tokenlength == 0) {
      if (strncmp(dfg_ActPtr, "-1", 2) == 0) {
	dfg_Skip(2);
	dfg_LVAL.number = -1;
	result = DFG_NUM;
      } else {
	result = *dfg_ActPtr;
	dfg_NextChar();
      }
    } else {
      actToken = dfg_ActPtr;
      dfg_Skip(tokenlength);   /* Set ActPtr to end of actToken */
      nextptr = dfg_ActPtr;
      nextc = *nextptr;
      *nextptr = '\0';     /* Mark the end of actToken */
      switch (tokenlength) {
      case 1:
	result = dfg_Identifier(actToken);
	break;
      case 2:
	if (strcmp(actToken, "or") == 0)
	  result = DFG_OR;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 3:
	if (strcmp(actToken, "not") == 0)
	  result = DFG_NOT;
	else if (strcmp(actToken, "and") == 0)
	  result = DFG_AND;
	else if (strcmp(actToken, "cnf") == 0)
	  result = DFG_CNF;
	else if (strcmp(actToken, "dnf") == 0)
	  result = DFG_DNF;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 4:
	if (strcmp(actToken, "true") == 0)
	  result = DFG_TRUE;
	else if (strcmp(actToken, "name") == 0)
	  result = DFG_NAME;
	else if (strcmp(actToken, "date") == 0)
	  result = DFG_DATE;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 5:
	if (strcmp(actToken, "equal") == 0)
	  result = DFG_EQUAL;
	else if (strcmp(actToken, "equiv") == 0)
	  result = DFG_EQUIV;
	else if (strcmp(actToken, "false") == 0)
	  result = DFG_FALSE;
	else if (strcmp(actToken, "SPASS") == 0)
	  result = DFG_SPASS;
	else if (strcmp(actToken, "logic") == 0)
	  result = DFG_LOGIC;
	else if (strcmp(actToken, "sorts") == 0)
	  result = DFG_SORTS;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 6:
	if (strcmp(actToken, "clause") == 0)
	  result = DFG_CLAUSE;
	else if (strcmp(actToken, "forall") == 0)
	  result = DFG_FORALL;
	else if (strcmp(actToken, "exists") == 0)
	  result = DFG_EXISTS;
	else if (strcmp(actToken, "axioms") == 0)
	  result = DFG_AXIOMS;
	else if (strcmp(actToken, "author") == 0)
	  result = DFG_AUTHOR;
	else if (strcmp(actToken, "status") == 0)
	  result = DFG_STATUS;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 7:
	if (strcmp(actToken, "formula") == 0)
	  result = DFG_FORMULA;
	else if (strcmp(actToken, "implies") == 0)
	  result = DFG_IMPLIES;
	else if (strcmp(actToken, "implied") == 0)
	  result = DFG_IMPLIED;
	else if (strcmp(actToken, "version") == 0)
	  result = DFG_VERSION;
	else if (strcmp(actToken, "unknown") == 0)
	  result = DFG_UNKNOWN;
	else if (strcmp(actToken, "subsort") == 0)
	  result = DFG_SUBSORT;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 9:
	if (strcmp(actToken, "functions") == 0)
	  result = DFG_FUNCTIONS;
	else if (strcmp(actToken, "operators") == 0)
	  result = DFG_OPERATORS;
	else if (strcmp(actToken, "predicate") == 0)
	  result = DFG_SORTPREDICATE;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 10:
	if (strcmp(actToken, "predicates") == 0)
	  result = DFG_PREDICATS;
        else if (strcmp(actToken, "resolution") == 0)
	  result = DFG_RESOLUTION;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 11:
	if (strcmp(actToken, "end_of_list") == 0 && nextc == '.') {
	  dfg_NextChar();
	  result = DFG_ENDLIST;
	}
	else if (strcmp(actToken, "end_problem") == 0 && nextc == '.') {
	  dfg_NextChar();
	  result = DFG_ENDPROBLEM;
	}
	else if (strcmp(actToken, "quantifiers") == 0)
	  result = DFG_QUANTIFS;
	else if (strcmp(actToken, "conjectures") == 0)
	  result = DFG_CONJECS;
	else if (strcmp(actToken, "description") == 0)
	  result = DFG_DESCRIPTION;
	else if (strcmp(actToken, "satisfiable") == 0)
	  result = DFG_SATISFIABLE;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 13:
	if (strcmp(actToken, "begin_problem") == 0)
	  result = DFG_BEGPROBLEM;
	else if (strcmp(actToken, "unsatisfiable") == 0)
	  result = DFG_UNSATISFIABLE;
	else if (strcmp(actToken, "list_of_proof") == 0)
	  result = DFG_PROOF;
	else
	  result = dfg_Identifier(actToken);
	break;
      case 15:
      case 16:
      case 20:
      case 24:
	if (strncmp(actToken, "list_of_", 8) == 0) {
	  char* newbuf = actToken + 8;

	  if (tokenlength==15 && strcmp(newbuf, "clauses") == 0)
	    result = DFG_CLAUSELIST;
	  else if (tokenlength==15 && strcmp(newbuf, "symbols") == 0 && nextc == '.') {
	    dfg_NextChar();
	    result = DFG_SYMBOLLIST;
	  }
	  else if (tokenlength==16 && strcmp(newbuf, "formulae") == 0)
	    result = DFG_FORMLIST;
	  else if (tokenlength==16 && strcmp(newbuf, "settings") == 0)
	    result = DFG_SETTINGS;
	  else if (tokenlength==20 && strcmp(newbuf, "descriptions")==0 && nextc=='.') {
	    dfg_NextChar();     /* Skip '.' */
	    result = DFG_DESCLIST;
	  }
	  else if (tokenlength==20 && strcmp(newbuf, "declarations")==0 && nextc=='.') {
	    dfg_NextChar();
	    result = DFG_DECLLIST;
	  }
	  else if (tokenlength==24 && strcmp(newbuf, "general_settings") == 0) {
	    /* Read and ignore everything up to end_of_list. */
	    ptr = strstr(dfg_ActPtr + 1, "end_of_list.");
	    
	    loop = TRUE;
	    if (ptr == NULL) {
	      /* Reached EOF */
	      fprintf(stderr, "Unexpected end of file in list_of_%s\n\n", newbuf);
	      misc_Error();
	    }
	    dfg_ActPtr = ptr;
	    dfg_Skip(12);    /* Skip "end_of_list." */
	  } else
	    /* Identifier starting with "list_of_" */
	    result = dfg_Identifier(actToken);
	}
	else
	  /* Tokenlength 15, 16, 20 or 24 but not starting with "list_of_" */
	  /* and thus no reserved word.                                    */
	  result = dfg_Identifier(actToken);
	break;
      default:
	result = dfg_Identifier(actToken);
      }  /* switch (tokenlength) */

      *nextptr = nextc;         /* Restore old buffer contents */
    }  /* if identifier [a-zA-Z0-9_]* */
  }    /* while (loop) */

  return result;
}

char* dfg_ReadText(BOOL CopyText)
/**************************************************************
  INPUT:   A Boolean value.
  RETURNS: if <CopyText> is TRUE the text from actual scanner
           position up to the next "*}". The text may contain
	   pairs of {* and *}.
	   If <CopyText> is FALSE, NULL is returned.
  EFFECT:  If <CopyText> is TRUE, memory is allocated for the text.
***************************************************************/
{
  long  braces = 0;
  char* ptr    = dfg_ActPtr;
  char* Result = NULL;

  while (braces >= 0 && ptr < dfg_FileBuffer+dfg_FileBufferLength-2) {
    if (*ptr == '{' && *(ptr+1) == '*')
      braces++;
    else if (*ptr == '*' && *(ptr+1) == '}')
      braces--;
    if (braces >= 0)
      ptr++;
  }

  if (braces < 0)
    *ptr = '\0';
  if (CopyText) {
    /* Allocates memory for at least 1 character */
    Result = memory_Malloc(sizeof(char) * (ptr - dfg_ActPtr + 1));
    strcpy(Result, dfg_ActPtr);
  }
  if (braces < 0)
    *ptr = '*';   /* Restore old buffer contents */

  dfg_ActPtr = ptr;
  return Result;
}


char* dfg_StringCopy(const char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: A copy of 'String'.
  EFFECT:  Memory is allocated for the copy.
***************************************************************/
{
  char* result = memory_Malloc(symbol__SYMBOLMAXLEN);

#ifdef DEBUG
  if (strlen(String) >= symbol__SYMBOLMAXLEN) {
    fprintf(stderr, "Line %u: %s is a too long symbol name\n",
	    dfg_ActLineNumber(), String);
    misc_Error();
  }
#endif

  strcpy(result, String);
  return result;
}


void dfg_StringFree(char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: Nothing.
  EFFECT:  This function frees the memory for 'String'.
***************************************************************/
{
  memory_Free(String, symbol__SYMBOLMAXLEN);
}


void dfg_FPrintToken(FILE* Output, int Token)
/**************************************************************
  INPUT:   A token.
  RETURNS: Nothing.
  EFFECT:  This function prints 'Token' to stdout.
***************************************************************/
{
  switch (Token) {
  case EOF:               fputs("EOF", Output); break;
  case DFG_CLAUSE:        fputs("clause", Output); break;
  case DFG_FORMULA:       fputs("formula", Output); break;
  case DFG_ENDLIST:       fputs("end_of_list.", Output); break;
  case DFG_BEGPROBLEM:    fputs("begin_problem", Output); break;
  case DFG_ENDPROBLEM:    fputs("end_problem.", Output); break;
  case DFG_SYMBOLLIST:    fputs("list_of_symbols.", Output); break;
  case DFG_CLAUSELIST:    fputs("list_of_clauses", Output); break;
  case DFG_FORMLIST:      fputs("list_of_formulae", Output); break;
  case DFG_PROOF:         fputs("list_of_proof", Output); break;
  case DFG_FUNCTIONS:     fputs("functions", Output); break;
  case DFG_OPERATORS:     fputs("operators", Output); break;
  case DFG_PREDICATS:     fputs("predicates", Output); break;
  case DFG_QUANTIFS:      fputs("quantifiers", Output); break;
  case DFG_SORTS:         fputs("sorts", Output); break;
  case DFG_NUM:           fputs("NUM", Output); break;
  case DFG_ID:            fputs("ID", Output); break;
  case DFG_FUNCTION:      fputs("FUNCTION", Output); break;
  case DFG_PREDICATE:     fputs("PREDICATE", Output); break;
  case DFG_JUNCTOR:       fputs("JUNCTOR", Output); break;
  case DFG_SETTINGS:      fputs("list_of_settings", Output); break;
  case DFG_SPASS:         fputs("SPASS", Output); break;
  case DFG_AXIOMS:        fputs("axioms", Output); break;
  case DFG_CONJECS:       fputs("conjectures", Output); break;
  case DFG_RESOLUTION:    fputs("resolution", Output); break;
  case DFG_CNF:           fputs("cnf", Output); break;
  case DFG_DNF:           fputs("dnf", Output); break;
  case DFG_FORALL:        fputs("forall", Output); break;
  case DFG_EXISTS:        fputs("exists", Output); break;
  case DFG_AND:           fputs("and", Output); break;
  case DFG_OR:            fputs("or", Output); break;
  case DFG_NOT:           fputs("not", Output); break;
  case DFG_IMPLIES:       fputs("implies", Output); break;
  case DFG_IMPLIED:       fputs("implied", Output); break;
  case DFG_EQUIV:         fputs("equiv", Output); break;
  case DFG_EQUAL:         fputs("equal",Output); break;
  case DFG_TRUE:          fputs("true", Output); break;
  case DFG_FALSE:         fputs("false", Output); break;
  case DFG_DESCLIST:      fputs("list_of_descriptions.", Output); break;
  case DFG_NAME:          fputs("name", Output); break;
  case DFG_AUTHOR:        fputs("author", Output); break;
  case DFG_VERSION:       fputs("version", Output); break;
  case DFG_LOGIC:         fputs("logic", Output); break;
  case DFG_STATUS:        fputs("status", Output); break;
  case DFG_DESCRIPTION:   fputs("description", Output); break;
  case DFG_DATE:          fputs("date", Output); break;
  case DFG_SATISFIABLE:   fputs("satisfiable", Output); break;
  case DFG_UNSATISFIABLE: fputs("unsatisfiable", Output); break;
  case DFG_UNKNOWN:       fputs("unknown", Output); break;
  case DFG_DECLLIST:      fputs("list_of_declarations.", Output); break;
  case DFG_SUBSORT:       fputs("subsort", Output); break;
  case DFG_SORTPREDICATE: fputs("predicate", Output); break;
  default:
    if (isprint(Token))
      fputc(Token, Output);
    else
      fprintf(Output, "Unknown Token %d", Token);
  }
}


/**************************************************************/
/* Static Functions                                           */
/**************************************************************/

static void dfg_EatComment(void)
/**************************************************************
  INPUT:   None.
  RETURNS: Nothing.
  EFFECT:  This function reads (and ignores) several consecutive
           comments.
***************************************************************/
{
  BOOL loop;
  do {
    loop = FALSE;
    dfg_EatSpace();
    if (*dfg_ActPtr == '%') {
      /* A one line comment */
      loop = TRUE;
      dfg_SkipLine();
    }
  } while (*dfg_ActPtr != '\0' && loop);
}


static int dfg_Identifier(char* Name)
/**************************************************************
  INPUT:   Name represents an identifier or a number.
  RETURNS: A token of type 'dfg_token' casted as int.
           The token might be 'DFG_FUNCTION', 'DFG_PREDICATE',
	   'DFG_PREDICATE', 'DFG_JUNCTOR', 'DFG_NUM' or 'DFG_ID'.
  EFFECT:  If a symbol was previously defined (e.g. as predicate)
           this function returns the symbol type (DFG_PREDICATE).
***************************************************************/
{
  long number;
  char* end;

  if (dfg_LOOKUPSYMBOLS) {
    SYMBOL s = symbol_Lookup(Name);
    if (s != 0) {
      dfg_LVAL.symbol = s;
      if (symbol_IsFunction(s))
	return DFG_FUNCTION;
      else if (symbol_IsPredicate(s))
	return DFG_PREDICATE;
      else if (symbol_IsConstant(s))
	return DFG_FUNCTION;
      /* Symbol is junctor */
      return DFG_JUNCTOR;
    }
  }

  if (dfg_LVAL.string != NULL)
    dfg_StringFree(dfg_LVAL.string);
  dfg_LVAL.string = dfg_StringCopy(Name);

  /* Symbol is a variable or a number */
  /* First check if it's a number */
  end = Name;
  number = strtol(Name, &end, 10);
  if (number != LONG_MAX && *end == '\0') {
    /* Number isn't too large and buffer can be converted completely */
    /* Since dfg_LVAL is a structure (and not a union) the number is */
    /* both stored as string and as numerical value                  */
    dfg_LVAL.number = number;
    return DFG_NUM;
  }

  /* Variable */
  return DFG_ID;
}
