/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              FIRST ORDER LOGIC SYMBOLS                 * */
/* *                                                        * */
/* *  $Module:   FOL  DFG                                   * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.58 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/02 14:27:50 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: foldfg.c,v $ */

#include "foldfg.h"


SYMBOL  fol_ALL;
SYMBOL  fol_EXIST;
SYMBOL  fol_AND;
SYMBOL  fol_OR;
SYMBOL  fol_NOT;
SYMBOL  fol_IMPLIES;
SYMBOL  fol_IMPLIED;
SYMBOL  fol_EQUIV;
SYMBOL  fol_VARLIST;
SYMBOL  fol_EQUALITY;
SYMBOL  fol_TRUE;
SYMBOL  fol_FALSE;

LIST    fol_SYMBOLS;


/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  FUNCTIONS                                             * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

void fol_Init(BOOL All)
/**************************************************************
  INPUT:   A boolean value detrmining whether only 'equality' or
           all fol symbols are created.
  RETURNS: None.
  USAGE:   extern public.
  SUMMARY: Initializes the Fol Module.
  EFFECTS: If <All> then all fol-symbols are created,
           only 'equality' otherwise.
  CAUTION: MUST BE CALLED BEFORE ANY OTHER fol-FUNCTION.
***************************************************************/
{ 
  if (All) {

    fol_ALL      = symbol_CreateJunctor("forall", 2, symbol_STATLEX);
    fol_EXIST    = symbol_CreateJunctor("exists", 2, symbol_STATLEX);
    fol_AND      = symbol_CreateJunctor("and", symbol_ArbitraryArity(), symbol_STATLEX);
    fol_OR       = symbol_CreateJunctor("or", symbol_ArbitraryArity(), symbol_STATLEX);
    fol_NOT      = symbol_CreateJunctor("not", 1, symbol_STATLEX);
    fol_IMPLIES  = symbol_CreateJunctor("implies", 2, symbol_STATLEX);
    fol_IMPLIED  = symbol_CreateJunctor("implied", 2, symbol_STATLEX);
    fol_EQUIV    = symbol_CreateJunctor("equiv", 2, symbol_STATLEX);
    fol_VARLIST  = symbol_CreateJunctor("", symbol_ArbitraryArity(), symbol_STATLEX);
    fol_EQUALITY = symbol_CreatePredicate("equal", 2, symbol_STATLEX);
    fol_TRUE     = symbol_CreatePredicate("true", 0, symbol_STATLEX);
    fol_FALSE    = symbol_CreatePredicate("false", 0, symbol_STATLEX);

    fol_SYMBOLS =
      list_Cons((POINTER)fol_ALL, list_Cons((POINTER)fol_EXIST, 
	list_Cons((POINTER)fol_AND, list_Cons((POINTER)fol_OR,
	  list_Cons((POINTER)fol_NOT,
	    list_Cons((POINTER)fol_IMPLIES, list_Cons((POINTER)fol_IMPLIED,
              list_Cons((POINTER)fol_EQUIV, list_Cons((POINTER)fol_VARLIST,
                list_Cons((POINTER)fol_EQUALITY, list_Cons((POINTER)fol_TRUE,
		  list_List((POINTER)fol_FALSE))))))))))));
  }
  else {
    fol_EQUALITY = symbol_CreatePredicate("equal", 2, symbol_STATLEX);
    fol_NOT      = symbol_CreateJunctor("not", 1, symbol_STATLEX);
    fol_SYMBOLS  = list_Cons((POINTER) fol_NOT, list_List((POINTER)fol_EQUALITY));
  }
}


SYMBOL fol_IsStringPredefined(const char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: The symbol iff String is a predefined fol string,
           symbol NULL otherwise
  EFFECT:  None
***************************************************************/
{
  LIST Scan;
  for(Scan=fol_SYMBOLS; !list_Empty(Scan); Scan=list_Cdr(Scan))
    if (strcmp(String, symbol_Name((SYMBOL)list_Car(Scan))) == 0)
      return ((SYMBOL)list_Car(Scan));
  return symbol_Null();
}


TERM fol_CreateQuantifier(SYMBOL Quantifier, LIST VarList, LIST Arguments)
/*******************************************************
  INPUT:   A symbol (which MUST be a fol quantifier),
           a list of variables that will be bound, and
           a list of arguments.
  RETURNS: A quantified term.
********************************************************/
{                                                                  
#if DEBUG                                                          
  if (!(fol_IsQuantifier(Quantifier))) {
    printf("\n\nIllegal input in fol_CreateQuantifier\n");
    misc_Error();
  }
#endif                                                             
                                                                   
  return term_Create(Quantifier,                                             
		     list_Cons(term_Create(fol_Varlist(), VarList),   
			       Arguments));  
}



TERM fol_ComplementaryTerm(TERM Term)
/*********************************************************
  INPUT:   A formula.
  RETURNS: The (copied) complementary (in sign) formula of <Term>
*******************************************************/
{
  if (symbol_Equal(term_TopSymbol(Term), fol_Not()))
    return term_Copy((TERM)list_First(term_ArgumentList(Term)));
  else
    return term_Create(fol_Not(), list_List(term_Copy(Term)));
}

void fol_NormalizeVarsIntern(TERM Formula)
/*********************************************************
  INPUT:    A sentence.
  EFFECT:   The quantifier variables of the formula are
            normalized, i.e., no two different quantifiers
	    bind the same variable.
  CAUTOION: Desctructive.
  RETURNS:  void.
*******************************************************/
{
  SYMBOL   Top;
  LIST     Scan1;
  
#if DEBUG
  if (!term_IsTerm(Formula)) {
    printf("\n\nIllegal input in fol_NormalizeVarsIntern\n");
    misc_DumpCore();
  }
#endif

  Top = term_TopSymbol(Formula);
  
  if (term_IsComplex(Formula))
    if (fol_IsQuantifier(Top)) {
      SYMBOL    Var;
      LIST      OldVars,Scan2;
      OldVars = list_Nil();
      for (Scan1=fol_QuantifierVariables(Formula);!list_Empty(Scan1);Scan1=list_Cdr(Scan1)) {
	Var     = term_TopSymbol(list_Car(Scan1));
	OldVars = list_Nconc(OldVars,list_List((POINTER)term_BindingValue(Var)));
	term_CreateValueBinding(Var, term_OldMark(), symbol_CreateStandardVariable());
      }
      fol_NormalizeVarsIntern(term_SecondArgument(Formula));
      for (Scan1=fol_QuantifierVariables(Formula),Scan2=OldVars;
	   !list_Empty(Scan1);
	   Scan1=list_Cdr(Scan1),Scan2=list_Cdr(Scan2)) {
	Var     = term_TopSymbol(list_Car(Scan1));
	term_RplacTop(list_Car(Scan1),term_BindingValue(Var));
	term_CreateValueBinding(Var, term_OldMark(), (NAT)list_Car(Scan2));
      }
      list_Delete(OldVars);
    }
    else
      for (Scan1=term_ArgumentList(Formula);!list_Empty(Scan1);Scan1=list_Cdr(Scan1))
	fol_NormalizeVarsIntern(list_Car(Scan1));
  else 
    if (symbol_IsVariable(Top))
      term_RplacTop(Formula,term_BindingValue(Top));

  return;
}

void fol_NormalizeVars(TERM Formula)
/*********************************************************
  INPUT:    A sentence.
  EFFECT:   The quantifier variables of the formula are
            normalized, i.e., no two different quantifiers
	    bind the same variable.
  CAUTOION: Desctructive.
  RETURNS:  void.
*******************************************************/
{
  symbol_ResetStandardVarCounter();
  fol_NormalizeVarsIntern(Formula);
}

BOOL fol_VarOccursFreely(TERM Var,TERM Term)
/**************************************************************
  INPUT:   A variable and a term.
  RETURNS: TRUE iff <Var> occurs freely in <Term>
***************************************************************/
{ 
  LIST    Scan;
  int     Stack;
  SYMBOL  Top;
  BOOL    Hit;

#if DEBUG
  if (!term_IsTerm(Term) || !term_IsVariable(Var)) {
    printf("\n\nIllegal input in fol_VarOccursFreely\n");
    misc_DumpCore();
  }
#endif

  Stack = stack_Bottom();

  do {
    Top = term_TopSymbol(Term);
    if (term_IsComplex(Term)) {
      if (fol_IsQuantifier(Top)) {
	Hit = TRUE;
	for (Scan=fol_QuantifierVariables(Term);!list_Empty(Scan)&&Hit;Scan=list_Cdr(Scan))
	  if (symbol_Equal(term_TopSymbol(list_Car(Scan)),term_TopSymbol(Var)))
	    Hit = FALSE;
	if (Hit)
	  stack_Push(list_Cdr(term_ArgumentList(Term)));
      }
      else
	stack_Push(term_ArgumentList(Term));
    }
    else {
      if (symbol_IsVariable(Top) && symbol_Equal(Top,term_TopSymbol(Var))) {
	stack_SetBottom(Stack);
	return TRUE;
      }
    }
    while (!stack_Empty(Stack) && list_Empty(stack_Top()))
      stack_Pop();
    if (!stack_Empty(Stack)) {
      Term = (TERM)list_Car(stack_Top());
      stack_RplacTop(list_Cdr(stack_Top()));
    }
  } while (!stack_Empty(Stack));

  return FALSE;
}

LIST fol_FreeVariables(TERM Term)
/**************************************************************
  INPUT:   A term where we assume that no variable is bound by more than
           one quantifier in <Term> !!!!!
  RETURNS: The list of variables occurring in the term. Variables are
           not (!) copied.
           Note that there may be many terms with same variable symbol.
	   All Variable terms are newly created.
***************************************************************/
{ 
  LIST    Variables,Scan;
  int     Stack;
  SYMBOL  Top;
  NAT     BoundMark,FreeMark;

#if DEBUG
  if (!term_IsTerm(Term) ||  term_InBindingPhase()) {
    printf("\n\nIllegal input or context in fol_FreeVariables\n");
    misc_DumpCore();
  }
  else
    term_StartBinding();
#endif

  Variables  = list_Nil();
  Stack      = stack_Bottom();
  BoundMark  = term_ActMark();
  FreeMark   = term_ActMark();

  do {
    Top = term_TopSymbol(Term);
    if (term_IsComplex(Term)) {
      if (fol_IsQuantifier(Top)) {
	for (Scan=fol_QuantifierVariables(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	  if (!term_VarIsMarked(term_TopSymbol(list_Car(Scan)),FreeMark))
	    term_CreateBinding(term_TopSymbol(list_Car(Scan)),BoundMark);
	stack_Push(term_ArgumentList(Term));                           /* Mark has to be removed ! */
	stack_Push(list_Cdr(term_ArgumentList(Term)));
      }
      else
	if (symbol_Equal(Top,fol_Varlist())) {
	  for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	    if (!term_VarIsMarked(term_TopSymbol(list_Car(Scan)),FreeMark))
	      term_CreateBinding(term_TopSymbol(list_Car(Scan)),0);    /* Mark has to be removed ! */
	  stack_RplacTop(list_Cdr(stack_Top()));                   /* Second Argument is Quantifier Arg */
	}
	else
	  stack_Push(term_ArgumentList(Term));
    }
    else {
      if (symbol_IsVariable(Top) && !term_VarIsMarked(Top,FreeMark) && !term_VarIsMarked(Top,BoundMark)) {
	Variables = list_Cons(Term, Variables);
	term_CreateBinding(Top, FreeMark);
      }
    }
    while (!stack_Empty(Stack) && list_Empty(stack_Top()))
      stack_Pop();
    if (!stack_Empty(Stack)) {
      Term = (TERM)list_Car(stack_Top());
      stack_RplacTop(list_Cdr(stack_Top()));
    }
  } while (!stack_Empty(Stack));

#if DEBUG
  term_StopBinding();
#endif  

  return Variables;
}


void fol_Free(void)
/**************************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  The fol memory used by the fol modul is freed.
***************************************************************/
{
  list_Delete(fol_SYMBOLS);
}


BOOL fol_FormulaIsClause(TERM Formula)
/**************************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  The fol memory used by the fol modul is freed.
***************************************************************/
{
  LIST LitList;

  if (term_TopSymbol(Formula) == fol_ALL)
    Formula = term_SecondArgument(Formula);

  if (term_TopSymbol(Formula) != fol_OR)
    return FALSE;
  
  LitList = term_ArgumentList(Formula);

  while (!list_Empty(LitList)) {
    if (!fol_IsLiteral(list_Car(LitList)))
      return FALSE;
    LitList = list_Cdr(LitList);
  }

  return TRUE;
}

void fol_PrintDFGSignature(FILE* File)
/**************************************************************
  INPUT:   A file stream.
  RETURNS: Nothing.
  SUMMARY: Prints all signature symbols in DFG format to the
           file stream.
***************************************************************/

{
  NAT    i;
  SYMBOL symbol;
  LIST   functions, predicates;

  functions = predicates = list_Nil();

  for (i = 1; i < symbol_ActIndex(); i++)
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)i)) {
      symbol = symbol_Signature(i)->info;
      switch (symbol_Type(symbol)) {

      case symbol_CONSTANT:
      case symbol_FUNCTION:
	functions = list_Push((POINTER)symbol, functions);
	break;
      case symbol_PREDICATE:
	predicates = list_Push((POINTER)symbol, predicates);
	break;
      }
    }

  if (!list_Empty(functions)) {
    fprintf(File, "  functions[");
    i = 0;
    do {
      symbol = (SYMBOL) list_Top(functions);
      fprintf(File, "(%s, %d)", symbol_Name(symbol), symbol_Arity(symbol));
      functions = list_Pop(functions);
      if (!list_Empty(functions))
	fprintf(File, ", ");
      if (i<15)
	i++;
      else {
	i= 0;
	fprintf(File, "\n\t");
      }
      
    } while (!list_Empty(functions));
    fprintf(File, "].\n");
  }

  predicates = list_DeleteElement(predicates, (POINTER)fol_VARLIST, symbol_Equal);
  predicates = list_DeleteElement(predicates, (POINTER)fol_EQUALITY, symbol_Equal);
  predicates = list_DeleteElement(predicates, (POINTER)fol_TRUE, symbol_Equal);
  predicates = list_DeleteElement(predicates, (POINTER)fol_FALSE, symbol_Equal);
  if (!list_Empty(predicates)) {
    i = 0;
    fprintf(File, "  predicates[");
    do {
      symbol = (SYMBOL) list_Top(predicates);
      fprintf(File, "(%s, %d)", symbol_Name(symbol), symbol_Arity(symbol));
      predicates = list_Pop(predicates);
      if (!list_Empty(predicates))
	fprintf(File, ", ");
      if (i<15)
	i++;
      else {
	i= 0;
	fprintf(File, "\n\t");
      }
    } while (!list_Empty(predicates));
    fprintf(File, "].\n");
  }
  list_Delete(predicates);
  list_Delete(functions);
}

char* fol_SPrintDFGSignature(char* File)
/**************************************************************
  INPUT:   A file stream.
  RETURNS: Nothing.
  SUMMARY: Prints all signature symbols in DFG format to the
           file stream.
***************************************************************/

{
  NAT    i;
  SYMBOL symbol;
  LIST   functions, predicates;

  functions = predicates = list_Nil();

  for (i = 1; i < symbol_ActIndex(); i++)
    if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)i)) {
      symbol = symbol_Signature(i)->info;
      switch (symbol_Type(symbol)) {

      case symbol_CONSTANT:
      case symbol_FUNCTION:
	functions = list_Push((POINTER)symbol, functions);
	break;
      case symbol_PREDICATE:
	predicates = list_Push((POINTER)symbol, predicates);
	break;
      }
    }

  if (!list_Empty(functions)) {
    File += sprintf(File, "  functions[");
    i = 0;
    do {
      symbol = (SYMBOL) list_Top(functions);
      File += sprintf(File, "(%s, %d)", symbol_Name(symbol), symbol_Arity(symbol));
      functions = list_Pop(functions);
      if (!list_Empty(functions))
	File +=  sprintf(File, ", ");
      if (i<15)
	i++;
      else {
	i= 0;
	File += sprintf(File, "\n\t");
      }
      
    } while (!list_Empty(functions));
    File += sprintf(File, "].\n");
  }

  predicates = list_DeleteElement(predicates, (POINTER)fol_VARLIST, symbol_Equal);
  predicates = list_DeleteElement(predicates, (POINTER)fol_EQUALITY, symbol_Equal);
  predicates = list_DeleteElement(predicates, (POINTER)fol_TRUE, symbol_Equal);
  predicates = list_DeleteElement(predicates, (POINTER)fol_FALSE, symbol_Equal);
  if (!list_Empty(predicates)) {
    i = 0;
    File += sprintf(File, "  predicates[");
    do {
      symbol = (SYMBOL) list_Top(predicates);
      File += sprintf(File, "(%s, %d)", symbol_Name(symbol), symbol_Arity(symbol));
      predicates = list_Pop(predicates);
      if (!list_Empty(predicates))
	File += sprintf(File, ", ");
      if (i<15)
	i++;
      else {
	i= 0;
	File += sprintf(File, "\n\t");
      }
    } while (!list_Empty(predicates));
    File += sprintf(File, "].\n");
  }
  list_Delete(predicates);
  list_Delete(functions);
  return File;
}


void fol_TermListPrintDFG(LIST List)
/**************************************************************
  INPUT:   A list of terms.
  RETURNS: None.
  USAGE:   intern private.
***************************************************************/
{
  for(; !list_Empty(List); List=list_Cdr(List)) {
    fol_PrintDFG(list_Car(List));
    if (!list_Empty(list_Cdr(List)))
      printf(",");
  }
}


void fol_TermListFPrintDFG(FILE* File, LIST List)
/**************************************************************
  INPUT:   A list of terms.
  RETURNS: None.
  USAGE:   intern private.
***************************************************************/
{
  for(; !list_Empty(List); List=list_Cdr(List)) {
    fol_FPrintDFG(File,list_Car(List));
    if (!list_Empty(list_Cdr(List)))
      fprintf(File,",");
  }
}


void fol_FPrintDFG(FILE* File, TERM Term)
/**************************************************************
  INPUT:   A file and a term.
  RETURNS: none.
  SUMMARY: Prints the term in prefix notation to the file. 
  CAUTION: Uses the other fol_Output functions.
***************************************************************/
{
  if (term_IsComplex(Term))
    if (fol_IsQuantifier(term_TopSymbol(Term))) {
      symbol_FPrint(File,term_TopSymbol(Term));
      fprintf(File,"([");
      fol_TermListFPrintDFG(File,fol_QuantifierVariables(Term));
      fprintf(File,"],");
      fol_FPrintDFG(File, term_SecondArgument(Term));
      fprintf(File,")");
    }
    else {
      symbol_FPrint(File,term_TopSymbol(Term));
      fprintf(File,"(");
      fol_TermListFPrintDFG(File,term_ArgumentList(Term));
      fprintf(File,")");
    }
  else 
    symbol_FPrint(File,term_TopSymbol(Term));
}

char* fol_TermListSPrintDFG(char* File, LIST List)
/**************************************************************
  INPUT:   A list of terms.
  RETURNS: None.
  USAGE:   intern private.
***************************************************************/
{
  for(; !list_Empty(List); List=list_Cdr(List)) {
    File = fol_SPrintDFG(File,list_Car(List));
    if (!list_Empty(list_Cdr(List)))
      File += sprintf(File,",");
  }
  return File;
}


char* fol_SPrintDFG(char* File, TERM Term)
/**************************************************************
  INPUT:   A file and a term.
  RETURNS: none.
  SUMMARY: Prints the term in prefix notation to the file. 
  CAUTION: Uses the other fol_Output functions.
***************************************************************/
{
  if (term_IsComplex(Term))
    if (fol_IsQuantifier(term_TopSymbol(Term))) {
      File  = symbol_SPrint(File,term_TopSymbol(Term));
      File += sprintf(File,"([");
      File  = fol_TermListSPrintDFG(File,fol_QuantifierVariables(Term));
      File += sprintf(File,"],");
      File  = fol_SPrintDFG(File, term_SecondArgument(Term));
      File += sprintf(File,")");
    }
    else {
      File  = symbol_SPrint(File,term_TopSymbol(Term));
      File += sprintf(File,"(");
      File  = fol_TermListSPrintDFG(File,term_ArgumentList(Term));
      File += sprintf(File,")");

    } else 
      File  = symbol_SPrint(File,term_TopSymbol(Term));

  return File;
}


void fol_PrintDFG(TERM Term)
/**************************************************************
  INPUT:   A term.
  RETURNS: none.
  SUMMARY: Prints the term in prefix notation to stdout. 
  CAUTION: Uses the other fol_Output functions.
***************************************************************/
{
  if (term_IsComplex(Term))
    if (fol_IsQuantifier(term_TopSymbol(Term))) {
      symbol_Print(term_TopSymbol(Term));
      printf("([");
      fol_TermListPrintDFG(fol_QuantifierVariables(Term));
      printf("],");
      fol_PrintDFG(term_SecondArgument(Term));
      printf(")");
    }
    else {
      symbol_Print(term_TopSymbol(Term));
      printf("(");
      fol_TermListPrintDFG(term_ArgumentList(Term));
      printf(")");
    } else 
      symbol_Print(term_TopSymbol(Term));
}


void fol_PrintPrecedence(void)
/**************************************************************
  INPUT:   None
  RETURNS: void
  EFFECT:  Prints the current precedence to stdout,
           fol symbols are excluded.
***************************************************************/
{
  SIGNATURE ActSig;

  if (symbol_HASSIGNATURE) {
    LIST   Symbols,Scan;
    SYMBOL Symbol;
    int    Index;
    Symbols = list_Nil();
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	Symbol = symbol_Signature(Index)->info;
	if  (Symbol != fol_TRUE && Symbol != fol_FALSE && Symbol != fol_EQUALITY &&
	     (symbol_IsPredicate(symbol_Signature(Index)->info) ||
	      symbol_IsFunction(symbol_Signature(Index)->info)))
	  Symbols = list_Cons((POINTER)(symbol_Signature(Index)->info),Symbols);
      }
    Symbols = list_Sort(Symbols, symbol_PrecedenceGreater);
    for(Scan = Symbols; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      ActSig = symbol_Signature(symbol_Index((SYMBOL)list_Car(Scan)));
      if (list_Empty(list_Cdr(Scan)))
	printf("%s", ActSig->name);
      else
	printf("%s > ", ActSig->name);
    }
    list_Delete(Symbols);
  }
}

void fol_FPrintPrecedence(FILE *File)
/**************************************************************
  INPUT:   None
  RETURNS: void
  EFFECT:  Prints the current precedence as a setting command to <File>
           fol symbols are excluded.
***************************************************************/
{
  SIGNATURE ActSig;

  if (symbol_HASSIGNATURE) {
    LIST   Symbols,Scan;
    SYMBOL Symbol;
    int    Index;
    Symbols = list_Nil();
    for(Index = 1; Index < symbol_ACTINDEX; Index++)
      if (!list_PointerMember(symbol_FREEDSYMBOLS,(POINTER)Index)) {
	Symbol = symbol_Signature(Index)->info;
	if  (Symbol != fol_TRUE && Symbol != fol_FALSE && Symbol != fol_EQUALITY &&
	     (symbol_IsPredicate(symbol_Signature(Index)->info) ||
	      symbol_IsFunction(symbol_Signature(Index)->info)))
	  Symbols = list_Cons((POINTER)(symbol_Signature(Index)->info),Symbols);
      }
    Symbols = list_Sort(Symbols, symbol_PrecedenceGreater);
    Index   = 0;
    fprintf(File,"set_precedence(");
    for(Scan = Symbols; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      ActSig = symbol_Signature(symbol_Index((SYMBOL)list_Car(Scan)));
      if (list_Empty(list_Cdr(Scan)))
	fprintf(File,"(%s,%d,%c)", ActSig->name, ActSig->weight,
		(symbol_HasProperty((SYMBOL)list_Car(Scan),ORDRIGHT) ? 'r' :
		 (symbol_HasProperty((SYMBOL)list_Car(Scan),ORDMUL) ? 'm' : 'l')));
      else
	fprintf(File,"(%s,%d,%c),", ActSig->name, ActSig->weight,
		(symbol_HasProperty((SYMBOL)list_Car(Scan),ORDRIGHT) ? 'r' :
		 (symbol_HasProperty((SYMBOL)list_Car(Scan),ORDMUL) ? 'm' : 'l')));
      if (Index>15) {
	Index = 0;
	fprintf(File,"\n\t");
      }
      else
	Index++;
    }
    fprintf(File,").");
    list_Delete(Symbols);
  }
}

BOOL fol_AssocEquation(TERM Term, SYMBOL *Result)
/**************************************************************
  INPUT:   A term.
  RETURNS: TRUE if the term is an equation defining associativity
           for some function symbol.
  EFFECT:  If the <Term> is an assoc equation, then <*Result> is
           assigned the assoc symbol.
***************************************************************/
{
  
  if (fol_IsEquality(Term)) {
    SYMBOL Top;
    TERM   Left,Right;
    Left = term_FirstArgument(Term);
    Right= term_SecondArgument(Term);
    Top  = term_TopSymbol(Left);
    if (symbol_IsFunction(Top) && symbol_Arity(Top) == 2 && 
	symbol_Equal(Top,term_TopSymbol(Right))) {
      SYMBOL v1,v2,v3;
      if (term_IsVariable(term_FirstArgument(Left)))
	v1 = term_TopSymbol(term_FirstArgument(Left));
      else
	if (term_IsVariable(term_FirstArgument(Right))) {
	  Term  = Right;
	  Right = Left;
	  Left  = Term;
	  v1 = term_TopSymbol(term_FirstArgument(Left));
	}
	else
	  return FALSE;
      if (symbol_Equal(term_TopSymbol(term_SecondArgument(Left)),Top) &&
	  symbol_IsVariable((v2=term_TopSymbol(term_FirstArgument(term_SecondArgument(Left)))))&&
	  symbol_IsVariable((v3=term_TopSymbol(term_SecondArgument(term_SecondArgument(Left)))))&&
	  symbol_Equal(term_TopSymbol(term_FirstArgument(Right)),Top) &&
	  symbol_Equal(v1,term_TopSymbol(term_FirstArgument(term_FirstArgument(Right)))) &&
	  symbol_Equal(v2,term_TopSymbol(term_SecondArgument(term_FirstArgument(Right)))) &&
	  symbol_Equal(v3,term_TopSymbol(term_SecondArgument(Right)))) {
	*Result = Top;
	return TRUE;
      }
    }
  }

  return FALSE;

}
