/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                     FORM2DFG                           * */
/* *                                                        * */
/* *  $Module:      FORM2DFG                                * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.59 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/06/20 09:12:34 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: form2dfg.c,v $ */

/* This Program reads formlas in DFG notation from a file
   and writes the same formulas in DFG syntax to stdout */

#include "input.h"
#include "vector.h"

static void SaveAsDFG(FILE *, LIST);
static void PrintFormula(FILE *, TERM );
static void PrintArguments(FILE * , LIST);
static void PrintSymbol(FILE *, SYMBOL);

void main(int argc, char* argv[])
{
  LIST     TokenList, InputList, FormulaList;
  TERM     Term;
  FILE     *Input,*Output;
  char     *Filename;
   
  memory_Init(0);
  /*memory_Print();*/
  symbol_Init(TRUE);
  stack_Init();
  term_Init();
  fol_Init(TRUE);
  in_Init();
  vec_Init();

  if (argc<3) {
    printf("\n\t          form2dfg Version 0.15");
    printf("\n\tUsage: form2dfg <input-file> <output-file>\n\n");
    return;
  }
  
  Filename = *++argv;
  Input    = misc_OpenFile(Filename,"r");
  FormulaList = list_Nil();
  
  TokenList = in_Scanner(Input);
  misc_CloseFile(Input,Filename);
  
  InputList  = TokenList;
  Term       = (TERM)NULL;

  while (!list_Empty(InputList)) {
    Term  = in_CreateFormulaTerm(in_ExtractTermList(InputList));
    FormulaList = list_Cons(Term,FormulaList);
    InputList = in_NextArgument(InputList);
    in_Free();
    symbol_ResetStandardVarCounter();
  }
  list_DeleteWithElement(TokenList, in_TokenDelete);

  if (Term == (TERM)NULL) {
    printf("\n\tEmpty input file.\n");
    return;
  }
  
  Filename = *++argv;
  Output   = misc_OpenFile(Filename,"w");
  SaveAsDFG(Output, FormulaList);
  misc_CloseFile(Output,Filename);
  printf("\nResult of translation is written to file: %s\n\n",*argv);
  printf("\n\t WARNING: You have to adjust the result by hand, since");
  printf("\n\t all formulae are put in the conjecture list. FLOTTER will");
  printf("\n\t negate these formulae before CNF translation !!\n");
 
  /* memory_Print(); */
  /* symbol_PrintAll(); */
  /* symbol_Dump(); */
  list_DeleteWithElement(FormulaList, term_Delete);
  in_Free();
  fol_Free();
  symbol_FreeAllSymbols();
#ifdef DEBUG
  memory_Print();
#endif
}


static void SaveAsDFG(FILE *File, LIST FormulaList) {
  LIST scan;

  fprintf(File, "begin_problem(Unknown).\n\n");

  fprintf(File, "list_of_descriptions.\n");
  fprintf(File, "name({*Unknown*}).\n");
  fprintf(File, "author({*form2dfg*}).\n");
  fprintf(File, "status(unknown).\n");
  fprintf(File, "description({*File generated by form2dfg*}).\n");
  fprintf(File, "end_of_list.\n\n");
  
  fprintf(File, "list_of_symbols.\n");
  fol_PrintDFGSignature(File);
  fprintf(File, "end_of_list.\n\n");
  fprintf(File, "list_of_formulae(axioms).\n");
  fprintf(File, "end_of_list.\n\n");
  fprintf(File, "list_of_formulae(conjectures).\n");
  for (scan = FormulaList; !list_Empty(scan); scan= list_Cdr(scan)) {
    fprintf(File, "  formula(");
    PrintFormula(File, list_Car(scan));
    fprintf(File, ").\n");
  }
  fprintf(File, "end_of_list.\n\n");
  fprintf(File, "end_problem.\n");
}


static void PrintFormula(FILE *File, TERM Formula)
{
  if (term_IsComplex(Formula)) {
    PrintSymbol(File,term_TopSymbol(Formula));
    fprintf(File,"(");
    if (fol_IsQuantifier(term_TopSymbol(Formula))) {
      fprintf(File,"[");
      PrintArguments(File, term_ArgumentList(term_FirstArgument(Formula)));
      fprintf(File,"],");
      PrintFormula(File, term_SecondArgument(Formula));
    }
    else
      PrintArguments(File, term_ArgumentList(Formula));
    fprintf(File,")");
  } else 
    PrintSymbol(File,term_TopSymbol(Formula));
}


static void PrintSymbol(FILE* File, SYMBOL Symbol)
{
  if (symbol_IsJunctor(Symbol)) {
    if (symbol_Equal(Symbol,fol_All())) {
      fprintf(File,"forall");
      return;
    }
    if (symbol_Equal(Symbol,fol_Exist())) {
      fprintf(File,"exists");
      return;
    }
    if (symbol_Equal(Symbol,fol_And())) {
      fprintf(File,"and");
      return;
    }
    if (symbol_Equal(Symbol,fol_Or())) {
      fprintf(File,"or");
      return;
    }
    if (symbol_Equal(Symbol,fol_Not())) {
      fprintf(File,"not");
      return;
    }
    if (symbol_Equal(Symbol,fol_Implies())) {
      fprintf(File,"implies");
      return;
    }
    if (symbol_Equal(Symbol,fol_Implied())) {
      fprintf(File,"implied");
      return;
    }
    if (symbol_Equal(Symbol,fol_Equiv())) {
      fprintf(File,"equiv");
      return;
    }
  }
  else
    if (symbol_Equal(Symbol,fol_Equality()))
      fprintf(File,"equal");
    else
      symbol_FPrint(File,Symbol);
}


static void PrintArguments(FILE* File, LIST Arguments)
{
  while(!list_Empty(Arguments)) {
    if (list_Empty(list_Cdr(Arguments))) {
      PrintFormula(File,list_Car(Arguments));
    }
    else {
      PrintFormula(File,list_Car(Arguments));
      fprintf(File,",");
    }
    Arguments = list_Cdr(Arguments);
  }
}
