/*
   Builds a database of contrapositives given tptp format input.
   Three types of terms: pos_ext are positive extensions, neg_ext are
   negative extensions, and numbered_clauses are the input_clause(s) 
   with ordered clause numbering.  Where pos_ext and neg_ext args:
      arg1: Predicate (e.g. Predicate of ++P(f(Y),X) is P, the reason
            for this is that Prolog does database indexing of 1st argument
            principle functor, as long as there are no variables
            (see, for example, Quintus manual e-3)
      arg2: head literal with no repeated variables ... repeated variables
            are "linearized" in the Unify_List of arg4.
      arg3: clause number
      arg4: estimate for iterative deepening bound, currently #non_prop extensions
      arg5: unify list of linearized variables in head literal: [unify(V1,V2),unify(V3,V4),...]
      arg6: Extension literals list, each element is a list with 5 elements: [Lit,_,ClN,_,_],
         where Lit is the extension literal, ClN is the clause number, and "_" are variables.

   Also asserts problem('<problem>') term, and if the problem is first order --predicate 
   calculus -- then it asserts prolog term 'first_order.'

   Top clause is tptp('<problem>'). replace <problem> with a tptp problem name, e.g.
   tptp('NUM015-1').  builds a NUM015-1.ap_db all paths database file.

   Heuristics Used
   ***************
   Uses Heuristic of "Unit Clause Preference," i.e., clauses are sorted in ascending order on 
   number of literals

   Also "weight," where input clauses are preferred in order: conjecture, hypothesis, axiom.
( EBS 'Conjecture' replaces 'theorem' in more recent versions of TPTP)
   weight in the database will cause "weight" to be used

   Also, "Deterministic Literal Preference," i.e., extension literals are selected in ascending order
   on the number of extension literals over all other clauses.
   This will be used if det_lit_pref is in the database

   David Sharpe
*/

/* June 26, 1997 Bruce Spencer 

A quick change to this file to prepare for the CASC 14 competition.

- added tptp_sub(<ProblemFile>).
- removed the use of consult_problem
- added ensure_loaded(library(quintus)) which was brought in indirectly 
  before through consult_problem


The new top-level call tptp_sub(<ProblemFile>) assumes that the tptp
problem file is stored in tptp format (i.e. input_clause/3 -- see the
TPTP tech report)  and that it is available.
In effect this means you can run the tptp2X
facility from the command line just before you start Prolog.  Here is a
"typical" such command line:

TPTP-v2.0.0/TPTP2X/tptp2X -t random+shorten -f tptp -d . NUM/NUM015-1.p

This will put NUM015-1+random+shorten.tptp in a subdirectory called NUM, 
creating the subdirectory if required.  Thus with a call to the prolog goal
tptp_sub('NUM/NUM015-1+random+shorten.tptp'), 
tptp_sub will consult this file and go on to create the all-paths database.
in the file 'ap_db's in the current directory.

Search for 'EBS' with your editor to see the other small changes

*/

:- ensure_loaded(library(lists)).
:- ensure_loaded(library(sorts)).
:- pragma(nodebug).
%:- ensure_loaded('consult_problem.pl'). EBS June'97.  Not needed for tptp_sub
:- ensure_loaded(library(quintus)).
:- dynamic first_order/0,	% set by function to remember a first_order problem
	size/0,			% set by user (":-assert(size).") to sort clauses by # literals
	weight/0,		% set by user to weight clauses by thereom, hypothesis, axiom
	det_lit_pref/0.		% set by user to sort literals inversly to number of extensions

det_lit_pref. % July 3, Scott Buffett

/* logical 'sign' operators */
:- op(100,fx,(++)).
:- op(100,fx,(--)).

/* Clear the database to deal with another file */
clear :-
   abolish(input_clause,3),
   retractall(first_order),
   abolish(numbered_clause,4),
   abolish(pos_ext,7),
   abolish(neg_ext,7),
   abolish(proof_start,1).

/* List the current clauses */
list :-
   listing(input_clause).

/* sort the input clauses increasing on size in a Length-InputClause list */
sort_size(KeyL7):-
   weight,
   compose_cl([],conjecture,KeyL0),
   (size -> keysort(KeyL0,KeyL1) ; KeyL1 = KeyL0),
   compose_cl([],hypothesis,KeyL2),
   (size -> keysort(KeyL2,KeyL3) ; KeyL3 = KeyL2),
   compose_cl([],axiom,KeyL4),
   (size -> keysort(KeyL4,KeyL5) ; KeyL5 = KeyL4),
   append(KeyL1,KeyL3,KeyL6),
   append(KeyL6,KeyL5,KeyL7).
sort_size(KeyL1):-
   compose_cl([],_,KeyL0),
   (size -> keysort(KeyL0,KeyL1) ; KeyL1 = KeyL0).

/* compose clause key list, keyed on length */
compose_cl(IDList,InfoIn,[L-input_clause(ID,Info,Body)|KLs]):-

   % find a unique input_clause(by checking ID) of right type
   input_clause(ID,Info,Body),
   \+ member(ID,IDList),
   (  nonvar(InfoIn)
   -> Info == InfoIn
   ;  true
   ),

   (  first_order
   -> true
   ;  first_order_test(Body)
   ),

   % compose this clause and do more clauses
   length(Body,L),
   compose_cl([ID|IDList],InfoIn,KLs).

% no clauses left, output starts with empty list
compose_cl(_,_,[]):-!.

/* sort literals in each clause on the number of extension literals
   over all other clauses.
*/
sort_lit(KL0,KL1):-
   count_lit(KL0,[],KCnt),
   arrange_lit(KL0,KCnt,KL1).

count_lit([],KCnt,KCnt):-
   !.
count_lit([_-input_clause(_,_,Body)|KLs],KCnt0,KCnt2):-
   lit_in_body(Body,KCnt0,KCnt1),
   count_lit(KLs,KCnt1,KCnt2).

lit_in_body([],KCnt,KCnt):-
   !.
lit_in_body([G|Gs],KCnt0,KCnt2):-
   goal_db(G,Gdb),
   (   delete(Cnt-Gdb,KCnt0,KCnt1)
   ->  Cntplus1 is Cnt + 1
   ;   KCnt1 = KCnt0,
       Cntplus1 is 1
   ),
   lit_in_body(Gs,[Cntplus1-Gdb|KCnt1],KCnt2).

arrange_lit([],_,[]):-
   !.
arrange_lit([L-input_clause(ID,Info,Body0)|KLs0],KCnt,
	[L-input_clause(ID,Info,Body1)|KLs1]):-
   compose_lit(Body0,KCnt,KB0),
   keysort(KB0,KB1),
   arrange_lit2(KB1,Body1),
   arrange_lit(KLs0,KCnt,KLs1).

compose_lit([],_,[]):-
   !.
compose_lit([G|Gs0],KCnt,[Cnt-G|Gs1]):-
   goal_db(G,Gdb),
   member(Cnt-Gdb,KCnt),
   compose_lit(Gs0,KCnt,Gs1).

arrange_lit2([],[]):-
   !.
arrange_lit2([_-G|KGs0],[G|KGs1]):-
   arrange_lit2(KGs0,KGs1).

goal_db(++Atom,Gdb):-
   functor(Atom,Pred,Arity),
   name(Pred,PredList),
   name(Arity,ArityList),
   %name(pos, [112, 111, 115]) % EBS
   append([112, 111, 115],PredList,SignPredList),
   append(SignPredList,ArityList,GdbList),
   name(Gdb,GdbList).
goal_db(--Atom,Gdb):-
   functor(Atom,Pred,Arity),
   name(Pred,PredList),
   name(Arity,ArityList),
   %name(neg, [110, 101, 103]), %EBS
   append([110, 101, 103],PredList,SignPredList),
   append(SignPredList,ArityList,GdbList),
   name(Gdb,GdbList).


/* process builds a database of the axiom set
   arg1: clause number
   arg2: keyed list of Length-InputClause 
*/
process(_,[]):-!.
process(N,[L-input_clause(ID,Info,Body)|KLs]):-

   % put the input clause as a numbered clause 
   nl, write('% numbered input clause of length '), write(L), write(': '),nl,
   write('numbered_clause('),
   write(N),write(','),
   write(ID),write(','),
   write(Info),write(','),
   writeq(Body),write(').'),nl,

   % build the positive and negative extensions
   Estimate is L - 1,
   \+ build_axiom_db(N,Estimate,Body),
   !,
   Nplus1 is N + 1,
   process(Nplus1,KLs).

/* build_axiom_db builds a database of resolvent clauses for each literal
   in the clause:
   arg1: clause number
   arg2: clause body
*/
build_axiom_db(N,Estimate0,Body):-
   db_delete(Goal,Body,NewBody),
   calc_estimate(NewBody,Estimate0,Estimate1),
   linearize(Goal,NewGoal,[],_,[],UnifyList),
   clause_to_g_and_v(N,NewBody,G,V),
   NewGoal =.. [Sign,Atom],
   functor(Atom,Pred,_Arity),

   (   Sign == ++  
    -> 
       write('pos_ext(')
    ;
       write('neg_ext(')
   ),
   writeq(Pred),write(','),
   writeq(NewGoal),write(','),
   writeq(N),write(','),
   writeq(Estimate1),write(','),
   writeq(UnifyList),write(','),
   writeq(G),write(','),
   writeq(V),write(').'),nl,
   fail.

build_query_db([L-input_clause(ID,conjecture,Body)|KLs],KLs):-
   copy_term(Body,Query0),
   calc_estimate(Query0,L,Estimate1),
      
   % distinguish as query
   clause_to_g_and_v(1,Query0,G,V),
   nl, write('% top clause (query)'),nl,
   write('numbered_clause('),
   write(1),write(','),
   write(ID),write(','),
   write(conjecture),write(','),
   writeq(Query0),write(').'),nl,

   db_enum_nodes(G,1,N),
   
   nl,write('query('),
   writeq(G),write(','),
   writeq(V),write(','),
   writeq(Estimate1),write(','),
   writeq(N),write(').'),nl,

   (   true		% queries reused for bbf, non fol queries NOT reused for allpaths
   ->  Est is L - 1, \+ build_axiom_db(1,Est,Body),nl
   ;   true
   ),
   !.
build_query_db([KL|KLs0],[KL|KLs1]):-
   build_query_db(KLs0,KLs1).

   
calc_estimate([],Est,Est):-
   !.
calc_estimate(_,0,0):-!.
calc_estimate([H|T],Est0,Est2):-
   atomize(H,A),
   functor(A,_,N),
   (  N = 0
   -> 
      Est1 is Est0 - 1
   ;  % else
      Est1 = Est0
   ),
   calc_estimate(T,Est1,Est2).

first_order_test([]):-!.
first_order_test([H|_]):-
   atomize(H,A),
   functor(A,_,N),
   N > 0,
   assert(first_order),
   !.
first_order_test([_|T]):-
   first_order_test(T).

/* linearize adapted from M. Stickel PTTP.  Transforms a literal so
   that it does not have any repeated variables.  This allows 
   unsound unification to be done on the goal literal (without occurs
   check).  Unsound unification can be done in constant time in 
   Prolog.  The occurs check is returned as a list of unify(L1,L2)
   members which can be completed after the goal is matched.
*/
linearize(TermIn,TermOut,VarsIn,VarsOut,MatchesIn,MatchesOut):-
   nonvar(TermIn) ->
      functor(TermIn,F,N),
      functor(TermOut,F,N),
      linearize_args(TermIn,TermOut,VarsIn,VarsOut,MatchesIn,MatchesOut,1,N)
   ; % elseif
   identical_member(TermIn,VarsIn) ->
      VarsOut = VarsIn,
      MatchesOut = [unify(TermIn,TermOut)|MatchesIn]
   ; % otherwise
      TermOut = TermIn,
      VarsOut = [TermIn|VarsIn],
      MatchesOut = MatchesIn.
linearize_args(TermIn,TermOut,VarsIn,VarsOut,MatchesIn,MatchesOut,I,N):-
   I > N ->
      VarsOut = VarsIn,
      MatchesOut = MatchesIn
   ; % otherwise
      arg(I,TermIn,ArgI),
      linearize(ArgI,NewArgI,VarsIn,Vars1,MatchesIn,Matches1),
      arg(I,TermOut,NewArgI),
      I1 is I + 1,
      linearize_args(TermIn,TermOut,Vars1,VarsOut,Matches1,MatchesOut,I1,N).

/* clause_to_v_and_g converts a clause body into its goal and vis terms,
   (which share variables):
   arg1: clause number
   arg2: clause body
   arg3: clause as a list of goal terms
   arg3: clause as a list of vis (to the right) terms
*/
clause_to_g_and_v(_,[],[],[]):-
   !.
clause_to_g_and_v(ClN,[C|Cs],[g(C,N,ClN,L,F)|Gs],[C-info(N,L,F)|Vs]):-
   clause_to_g_and_v(ClN,Cs,Gs,Vs).

/* enum_nodes assigns a number to each new extension literal and sets level to 1 */
db_enum_nodes([],N,N):-!.
db_enum_nodes([g(_,N0,_,1,_)|Gs],N0,N2):-
   N1 is N0 + 1,
   db_enum_nodes(Gs,N1,N2).

/********************************/
/* Top Level Procedures General */
/********************************/
tptp_sub(ProblemFile) :- %EBS
	compile(ProblemFile),
	create_db_file('ap_db', ProblemFile).

tptp(Problem):-
   clear,
   consult_problem(Problem),
   name(Problem,ProblemList),
   name('.ap_db',Extension),
   append(ProblemList,Extension,DBFileList),
   name(DBFile,DBFileList),
   create_db_file(DBFile, Problem). %EBS  I inserted a split here

create_db_file(DBFile, Problem) :- %EBS   to bypass the first few predicates
   tell(DBFile),
   sort_size(KL1),
   (   det_lit_pref
   ->  sort_lit(KL1,KL2)
   ;   KL2 = KL1
   ),

   % write 'header' info to file, (specific to eclipse)
   write('% all_paths database, for problem '),writeq(Problem),nl,
   write('% db is contrapositives (i.e. with every literal of every clause as head)'),nl,nl,
   write('% Heuristics: '), nl, 
   (  size 
   -> write('%   unit clause preference (clauses sorted by size)'),nl
   ;  true
   ), 

   (  det_lit_pref 
   -> write('%   deterministic lit preference (literals sorted inverse to possible extensions)'),nl 
   ;  true
   ),

   (  weight
   -> write('%   weighted clauses (clauses sorted by conjecture, hypothesis, axiom)'),nl
   ;  true
   ),

   (  not(size), not(det_lit_pref), not(weight)
   -> write('%   no heuristics (clauses & literals listed in TPTP input order)'),nl
   ;  true
   ),

   nl, write('% eclipse directives'),nl,
   write(':- pragma(nodebug).'),nl,
   write(':- lib(scattered).'),nl,
   write(':- scattered'),nl,
   write('      pos_ext/7, neg_ext/7, numbered_clause/4, query/4.'),nl,nl,
   write('problem('),put(39),write(Problem),put(39),write(').'),nl,!,
   
   % writeq query and axioms into file
   build_query_db(KL2,KL3),
   process(2,KL3),

   nl, 
   (   first_order 
   ->  write('first_order.') 
   ;   write('first_order:-fail.')
   ), nl,
   told.

/* build_demo/1 is currently broken
build_demo(X):-
   name('demo',DemoList),
   name(X,XList),
   append(DemoList,XList,DemoXList),
   name('.ap_db',ExtList),
   append(DemoXList,ExtList,DBFileList),
   name(DBFile,DBFileList),
   tell(DBFile),
   write('problem(demo'),write(X),write(').'),nl,
   \+ process(2,[]),
   build_query_db,
   (first_order -> write(':-assert(first_order).'),nl ; true),
   told.
*/

/********************************/
/* General Prolog Utilities     */
/********************************/
/* db_delete(a1,a2 a3),remove first member of a2 that matches a1 into a3:
   arg1: element to test for
   arg2: input list
   arg3: list with first match of arg1 in arg2 off the list
*/
db_delete(A1,[A1|A3],A3).
db_delete(A1,[E|A2],[E|A3]):-
   db_delete(A1,A2,A3).

/* first argument literal with second argument atom */
atomize(++X,X).
atomize(--X,X).

identical_member(X,[Y|_]):-
   X == Y,
   !.
identical_member(X,[_|Y]):-
   identical_member(X,Y).
