/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *             INPUT OF FORMUALAE                         * */
/* *                                                        * */
/* *  $Module:   INPUT                                      * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.44 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/04/30 15:50:34 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: input.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "input.h"

/**************************************************************/
/* Globals                                                    */
/**************************************************************/

static LIST in_STACK;
static int  in_CURRENTLINE;
static char in_CURCHAR;

static char in_IdentString[symbol__SYMBOLMAXLEN];

static char in_OpenString[symbol__SYMBOLMAXLEN]  = {'('};
static char in_CloseString[symbol__SYMBOLMAXLEN] = {')'};

/**************************************************************/
/* Functions                                                  */
/**************************************************************/

static void in_Push(SYMBOL Var, char* String)
/**************************************************************
  INPUT:   A variable symbol and a string.
  RETURNS: void.
  EFFECT:  The variable is pushed on the in_STACK.
***************************************************************/
{
  in_STACK = list_Push(list_PairCreate((POINTER)Var, String), in_STACK);
}


static void in_Pop(void)
/**************************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  Performs a pop on in_STACK.
***************************************************************/
{
  list_PairFree(list_Top(in_STACK));
  in_STACK = list_Pop(in_STACK);
}


static SYMBOL in_Lookup(const char* String)
/**************************************************************
  INPUT:   A string.
  RETURNS: NULL if <String> is not on in_STACK,
           the corresponding symbol otherwise
  EFFECT:  None.
***************************************************************/
{
  LIST Scan;

  for(Scan = in_STACK; !list_Empty(Scan); Scan = list_Cdr(Scan))
    if (strcmp((char *)(list_PairSecond(list_Car(Scan))), String) == 0)
      return ((SYMBOL) list_PairFirst(list_Car(Scan)));
  
  return ((SYMBOL)NULL);
}


void in_Init(void)
/**************************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  The input module is initialized.
***************************************************************/
{
  in_STACK       = list_StackBottom();
  in_CURRENTLINE = 1;
  in_CURCHAR     = ' ';
}


void in_Free(void)
/**************************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  The input module is initialized.
***************************************************************/
{
  while (!list_StackEmpty(in_STACK))
    in_Pop();
}


static TOKEN in_TokenCreate(char* TokenString, int TokenLine)
/**************************************************************
  INPUT:   A string and a line number.
  RETURNS: The token created from the input.
  EFFECT:  None.
***************************************************************/
{
  TOKEN Token;

  Token = memory_Malloc(sizeof(TOKEN_NODE));

  Token->string = TokenString;
  Token->line   = TokenLine;

  return (Token);
}


void in_TokenDelete(TOKEN Token)
/**************************************************************
  INPUT:   A Token.
  RETURNS: void.
  EFFECT:  The Token is deleted.
***************************************************************/
{  
  memory_Free(Token->string, symbol__SYMBOLMAXLEN); 
  memory_Free(Token, sizeof(TOKEN_NODE));
}


void in_TokenPrint(TOKEN Token)
/**************************************************************
  INPUT:   A Token.
  RETURNS: void.
  EFFECT:  The Token is printed to stdout.
***************************************************************/
{  
  printf("String: %s  Line:%d\n", Token->string, Token->line);
}


static BOOL in_IsWhitespace(char Character)
/**************************************************************
  INPUT:   A Character.
  RETURNS: TRUE iff <Character> is a whitespace character
  EFFECT:  None
***************************************************************/
{
  return(Character == ' ' || Character == '\n' ||
	 Character == '\r' || Character == '\t');
}


static BOOL in_IsNameChar(char Character)
/**************************************************************
  INPUT:   A Character.
  RETURNS: TRUE iff <Character> is a  character of a name
  EFFECT:  None
***************************************************************/
{
  return ((Character >= '0' && Character <= '9') ||
	  (Character >= 'A' && Character <= 'Z') ||
	  (Character >= 'a' && Character <= 'z') ||
	  Character == '_' || Character == '-' || Character == '=');
}


static BOOL in_IsSpecialChar(char Character)
/**************************************************************
  INPUT:   A Character.
  RETURNS: TRUE iff <Character> is a special character
  EFFECT:  None
***************************************************************/
{
  return(Character == '(' || Character == ')' || Character == '.');
}


static BOOL in_IsSpecialString(const char* String)
/**************************************************************
  INPUT:   A String.
  RETURNS: TRUE iff <String> is a special character string.
  EFFECT:  None
***************************************************************/
{
  return (strcmp(String, in_CloseString) == 0 ||
	  strcmp(String, in_OpenString) == 0);
}


static BOOL in_IsCommentChar(char Character)
/**************************************************************
  INPUT:   A Character.
  RETURNS: TRUE iff <Character> is the comment character
  EFFECT:  None
***************************************************************/
{
  return(Character == '%');
}


static char in_ReadUntilEndOfLine(FILE* Stream, char CurChar)
/**************************************************************
  INPUT:   A stream and a character.
  RETURNS: The first read character finishing the curent line.
  EFFECT:  Reads from stdin
***************************************************************/
{
  while (CurChar != '\n' && CurChar != '\r' && CurChar != (char)EOF)
    CurChar = (char)fgetc(Stream);

  return CurChar;
}


static char in_FindToken(FILE* Stream, char CurChar, int* CurLinePointer)
/**************************************************************
  INPUT:   A stream, the current character and a pointer to an integer 
           containing the current line
  RETURNS: The first character of a token.
  EFFECT:  Reads a STOP input from stdin until a token character occurs
***************************************************************/
{
  while (in_IsWhitespace(CurChar)) {
    if (CurChar == '\n' || CurChar == '\r')
      (*CurLinePointer)++;
    CurChar = (char)fgetc(Stream);
  }
  
  if ((int)CurChar == '|'){
    CurChar = (char)fgetc(Stream); 
    while (in_IsWhitespace(CurChar)) {
      if (CurChar == '\n' || CurChar == '\r')
	(*CurLinePointer)++;
      CurChar = (char)fgetc(Stream);
    }
    if ((int)CurChar == '|') {
      printf("\n Line %d: %c%c is not an allowed character\n",
	     *CurLinePointer, CurChar, CurChar);
      misc_Error();
    }
  }

  if (!in_IsNameChar(CurChar) && !in_IsSpecialChar(CurChar) &&
      !in_IsCommentChar(CurChar) && CurChar != (char)EOF) {
    printf("\n Line %d: %c is not an allowed character\n",
	   *CurLinePointer, CurChar);
    misc_Error();
  }

  return CurChar;
}


static char in_ReadToken(FILE* Stream, char CurChar, int *CurLinePointer,
			 LIST TokenList, int* ParenthPointer)
/**************************************************************
  INPUT:   A stream, a Character, a pointer to an integer
           containing the current line, a list of tokens
	   and a pointer to an integer containing
	   the number of open parentheses.
  RETURNS: The first read non token char.
  EFFECT:  Reads from stdin and appends the read token to <TokenList>
***************************************************************/
{
  int Position;
  char *TokenString;

  Position = 0;

  if (in_IsSpecialChar(CurChar)) {
    in_IdentString[0] = CurChar;
    in_IdentString[1] = '\0';
    TokenString              = memory_Malloc(symbol__SYMBOLMAXLEN);
    TokenString              = strcpy(TokenString, in_IdentString);
    list_Rplacd(TokenList,
		list_List(in_TokenCreate(TokenString, *CurLinePointer)));
    if (CurChar == '(')
      (*ParenthPointer)++;
    else
      if (CurChar == ')') {
	(*ParenthPointer)--;
	if (*ParenthPointer < 0) {
	  printf("\n Line %d: Too many right parantheses\n", *CurLinePointer);
	  misc_Error();
	}
      }
    CurChar = fgetc(Stream);
  } else if (in_IsNameChar(CurChar)) {
    
    while (in_IsNameChar(CurChar) && Position < symbol__SYMBOLMAXLEN) {
      in_IdentString[Position] = CurChar;
      Position++;
      CurChar = fgetc(Stream);
    }
    
    if (Position >= symbol__SYMBOLMAXLEN) {
      in_IdentString[Position] = '\0';
      printf("\n Line %d: <%s...> too long\n", *CurLinePointer,
	     in_IdentString);
      misc_Error();
    }
    
    if (!in_IsWhitespace(CurChar) && !in_IsCommentChar(CurChar) &&
	!in_IsSpecialChar(CurChar)) {
      printf("\n Line %d: %c is not an allowed character\n",
	     *CurLinePointer, CurChar);
      misc_Error();
    }
    
    in_IdentString[Position] = '\0';
    TokenString              = memory_Malloc(symbol__SYMBOLMAXLEN);
    TokenString              = strcpy(TokenString, in_IdentString);
    list_Rplacd(TokenList,
		list_List(in_TokenCreate(TokenString, *CurLinePointer)));
  }
  
  return(CurChar);
}


LIST in_Scanner(FILE* Stream)
/**************************************************************
  INPUT:   A stream.
  RETURNS: A list of Tokens.
  EFFECT:  Reads a STOP input from stdin, and converts it into
           a Token list.
***************************************************************/
{
  LIST TokenList, EndOfTokenList;
  int  Parentheses;

  TokenList      = list_List(list_Nil());
  EndOfTokenList = TokenList;
  Parentheses    = 0;

  while (in_CURCHAR != (char)EOF) {
    in_CURCHAR =
      in_ReadToken(Stream, in_FindToken(Stream, in_CURCHAR, &in_CURRENTLINE), 
		   &in_CURRENTLINE, EndOfTokenList, &Parentheses);

    if (list_Cdr(EndOfTokenList) != list_Nil())
      EndOfTokenList = list_Cdr(EndOfTokenList);

    if (in_IsCommentChar(in_CURCHAR))
      in_CURCHAR = in_ReadUntilEndOfLine(Stream, in_CURCHAR);
  }

  EndOfTokenList = list_Cdr(TokenList);
  list_Free(TokenList);
  if (Parentheses != 0) {
    printf("\n Too many Opening Parantheses\n");
    misc_Error();
  }

  return EndOfTokenList;
}


LIST in_ScanList(FILE* Stream)
/**************************************************************
  INPUT:   A stream.
  RETURNS: A list of Tokens.
  EFFECT:  Reads a exactly one list of tokens.
***************************************************************/
{
  LIST TokenList, EndOfTokenList;
  int  Parentheses;

  TokenList      = list_List(list_Nil());
  EndOfTokenList = TokenList;
  Parentheses    = 0;

  in_CURCHAR = in_FindToken(Stream, in_CURCHAR, &in_CURRENTLINE);

  if (in_CURCHAR != '(' && in_CURCHAR != (char)EOF && in_CURCHAR != '.') {
    /*'.' dabei*/ 
    printf("\n Line %d: expected an open paranthesis, not %c ! \n", 
	   in_CURRENTLINE, in_CURCHAR);
    misc_Error();
  }
  
  if (in_CURCHAR != (char)EOF) {
    in_CURCHAR = in_ReadToken(Stream, in_CURCHAR, &in_CURRENTLINE,
			      EndOfTokenList, &Parentheses);
    EndOfTokenList = list_Cdr(EndOfTokenList);
  }
  
  while (in_CURCHAR != (char)EOF && Parentheses != 0) {
    in_CURCHAR = in_ReadToken(Stream,
			  in_FindToken(Stream, in_CURCHAR, &in_CURRENTLINE), 
			  &in_CURRENTLINE, 
		          EndOfTokenList,
		          &Parentheses);

    if (list_Cdr(EndOfTokenList) != list_Nil())
      EndOfTokenList = list_Cdr(EndOfTokenList);

    if (in_IsCommentChar(in_CURCHAR))
      in_CURCHAR = in_ReadUntilEndOfLine(Stream, in_CURCHAR);
  }
  EndOfTokenList = list_Cdr(TokenList);
  list_Free(TokenList);
  if (Parentheses != 0) {
    printf("\n Too many Open Parantheses\n");
    misc_Error();
  }

  return EndOfTokenList;
}


LIST in_NextArgument(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, starting with an Open token.
  RETURNS: The list without the first sublist, represented by the tokens.
  EFFECT:  None
***************************************************************/
{
  if (list_Empty(TokenList))
    return(list_Nil());
  
  if (strcmp(((TOKEN)list_Car(TokenList))->string, in_OpenString) == 0) {
    int Parantheses;

    Parantheses = 1;
    TokenList   = list_Cdr(TokenList);

    while (Parantheses !=0) {
      if (strcmp(((TOKEN)list_Car(TokenList))->string, in_OpenString) == 0) 
	Parantheses++;
      else
	if (strcmp(((TOKEN)list_Car(TokenList))->string, in_CloseString) == 0)
	  Parantheses--;
      TokenList   = list_Cdr(TokenList);
    }
  
    return TokenList;
  }
  else
    return list_Cdr(TokenList);
}


LIST in_ExtractTermList(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, starting with an Open token.
  RETURNS: The list of tokens corresponding to the Open token.
  EFFECT:  None
***************************************************************/
{
  int  Parantheses;
  LIST Result, ResultEnd;

  if (strcmp(((TOKEN)list_Car(TokenList))->string, in_OpenString) != 0) {
    printf("\n Line %d: Expected a term, not <%s>\n",
	   ((TOKEN) list_Car(TokenList))->line,
	   ((TOKEN) list_Car(TokenList))->string);
    misc_Error();
  }
  
  Parantheses = 1;
  Result      = list_List(list_Nil());
  ResultEnd   = Result;
  TokenList   = list_Cdr(TokenList);
  
  while (Parantheses !=0) {
    if (strcmp(((TOKEN) list_Car(TokenList))->string, in_OpenString) == 0) 
      Parantheses++;
    else
      if (strcmp(((TOKEN) list_Car(TokenList))->string, in_CloseString) == 0)
	Parantheses--;
    if (Parantheses !=0) {
      list_Rplacd(ResultEnd, list_List(list_Car(TokenList)));
      ResultEnd = list_Cdr(ResultEnd);
    }
    TokenList   = list_Cdr(TokenList);
  }
  ResultEnd   = list_Cdr(Result);
  list_Free(Result);
  if (list_Empty(ResultEnd)) {
    printf("\n Line %d: Expected a term, not ()\n",
	   ((TOKEN) list_Car(TokenList))->line);
    misc_Error();
  }
  
  return ResultEnd;
}


LIST in_ExtractList(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, starting with an Open token.
  RETURNS: The list of tokens corresponding to the Open token.
  EFFECT:  None
***************************************************************/
{
  int  Parantheses;
  LIST Result, ResultEnd;

  Parantheses = 1;
  Result      = list_List(list_Nil());
  ResultEnd   = Result;
  TokenList   = list_Cdr(TokenList);

  while (Parantheses !=0) {
    if (strcmp(((TOKEN)list_Car(TokenList))->string, in_OpenString) == 0) 
      Parantheses++;
    else if (strcmp(((TOKEN)list_Car(TokenList))->string, in_CloseString) == 0)
      Parantheses--;
    if (Parantheses !=0) {
      list_Rplacd(ResultEnd, list_List(list_Car(TokenList)));
      ResultEnd = list_Cdr(ResultEnd);
    }
    TokenList   = list_Cdr(TokenList);
  }
  ResultEnd   = list_Cdr(Result);
  list_Free(Result);
  if (list_Empty(ResultEnd)) {
    printf("\n Line %d: Expected a non empty list.\n",
	   ((TOKEN) list_Car(TokenList))->line);
    misc_Error();
  }

  return ResultEnd;
}


static int in_NumberOfArguments(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, representing the arguments of a term.
  RETURNS: The number of arguments.
  EFFECT:  None
***************************************************************/   
{
  int Result;

  Result = 0;

  while (!list_Empty(TokenList)) {
    Result++;
    TokenList = in_NextArgument(TokenList);
  }

  return(Result);
}


LIST in_ExtractInputList(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, representing an input for STOP.
  RETURNS: A list of three token lists. One for parameters, 
           premises, conclusions, respectively.
  EFFECT:  None.
***************************************************************/   
{
  LIST Result, Parameters, Premises, Conclusions;
  int  Parantheses;

  Result      = list_Nil();
  Parameters  = list_Nil();
  Premises    = list_Nil();
  Conclusions = list_Nil();
  

  if (strcmp(((TOKEN)list_First(TokenList))->string, in_OpenString) == 0 &&
      strcmp(((TOKEN)list_Second(TokenList))->string, in_OpenString) == 0) {
    TokenList = list_Cdr(list_Cdr(TokenList));
    Parantheses = 1;
    while (Parantheses != 0) {
      if (strcmp(((TOKEN)list_First(TokenList))->string, in_OpenString) == 0)
	Parantheses++;
      else
	if (strcmp(((TOKEN)list_First(TokenList))->string, in_CloseString) == 0)
	  Parantheses--;
      if (Parantheses != 0)
	Parameters = list_Nconc(Parameters, list_List(list_First(TokenList)));
      TokenList = list_Cdr(TokenList);
    }
    if (strcmp(((TOKEN)list_First(TokenList))->string, in_OpenString) == 0) {
      TokenList = list_Cdr(TokenList);
      Parantheses = 1;
      while (Parantheses != 0) {
	if (strcmp(((TOKEN)list_First(TokenList))->string, in_OpenString) == 0)
	  Parantheses++;
	else if (strcmp(((TOKEN)list_First(TokenList))->string, in_CloseString) == 0)
	  Parantheses--;
	if (Parantheses != 0)
	  Premises = list_Nconc(Premises, list_List(list_First(TokenList)));
	TokenList = list_Cdr(TokenList);
      }
      if (strcmp(((TOKEN)list_First(TokenList))->string, in_OpenString) == 0) {
	TokenList = list_Cdr(TokenList);
	Parantheses = 1;
	while (Parantheses != 0) {
	  if (strcmp(((TOKEN)list_First(TokenList))->string, in_OpenString) == 0)
	    Parantheses++;
	  else if (strcmp(((TOKEN)list_First(TokenList))->string, in_CloseString) == 0)
	    Parantheses--;
	  if (Parantheses != 0)
	    Conclusions = list_Nconc(Conclusions, list_List(list_First(TokenList)));
	  TokenList = list_Cdr(TokenList);
	}
	
	if (strcmp(((TOKEN)list_First(TokenList))->string, in_CloseString) != 0) {
	  printf("\nExpected a ')' after the conclusion list, not %s.\n",
		 ((TOKEN) list_First(TokenList))->string);
	  misc_Error();
	}
      }
      else {
	printf("\nExpected a ')' after the premis list, not %s.\n",
	       ((TOKEN) list_First(TokenList))->string);
	misc_Error();
      }
    }
    else {
      printf("\nExpected a ')' after the parameter list, not %s.\n",
	     ((TOKEN) list_First(TokenList))->string);
      misc_Error();
    }
  }
  else {
    printf("\nExpected a list of three list as STOP input, not %s.\n",
	   ((TOKEN) list_First(TokenList))->string);
    misc_Error();
  }

  return list_Cons(Parameters,list_Cons(Premises,list_List(Conclusions)));
}


static TERM in_CreateVariable(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, the first token must be the name of a variable.
  RETURNS: A term corresponding to the variable name.
  EFFECT:  A new variable symbol is created.
***************************************************************/   
{
  TOKEN Token;

  Token      = (TOKEN)list_Car(TokenList);

  if (symbol_Lookup(Token->string) != symbol_Null()) {
    printf("\n Line %d: Expected a a variable, not <%s>\n",
	   Token->line,
	   Token->string);
    misc_Error();
  }

  return term_Create(symbol_CreateStandardVariable(), list_Nil());
}

    
static TERM in_CreateFunctionArgument(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, the first token must be the argument of a term.
  RETURNS: The term corresponding to the first token.
  EFFECT:  Non-existing symbols are created.
***************************************************************/   
{
  TOKEN  Token;
  SYMBOL Symbol;

  Token      = (TOKEN)list_Car(TokenList);
 
  if (strcmp(Token->string, in_OpenString) == 0)
    return(in_CreateFunctionTerm(in_ExtractTermList(TokenList)));
   
  if ((Symbol = symbol_Lookup(Token->string)) != symbol_Null()) { 
    if (!symbol_IsConstant(Symbol)) {
      printf("\n Line %d: Expected a constant or a variable, not <%s>\n",
	     Token->line,
	     Token->string);
      misc_Error();
    }
  } else {
    if (fol_IsStringPredefined(Token->string)) {
      printf("\n Line %d: <%s> is a reserved name, "
	     "must not be used for a variable.\n",
	     Token->line,
	     Token->string);
      misc_Error();
    }
    
    if ((Symbol = in_Lookup(Token->string)) == 0) {
      Symbol = symbol_CreateStandardVariable();
      in_Push(Symbol, Token->string);
    }
  }
  return term_Create(Symbol, list_Nil());
}
  

LIST in_CreateFunctionArguments(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, representing the arguments of a term.
  RETURNS: The list of argument terms.
  EFFECT:  Non-existing symbols are created.
***************************************************************/   
{
  LIST   ArgumentList, ArgumentListEnd;

  ArgumentList    = list_List(list_Nil());
  ArgumentListEnd = ArgumentList;

  while (!list_Empty(TokenList)) {
    list_Rplacd(ArgumentListEnd,
		list_List(in_CreateFunctionArgument(TokenList)));
    ArgumentListEnd = list_Cdr(ArgumentListEnd);
    TokenList       = in_NextArgument(TokenList);
  }
  ArgumentListEnd = list_Cdr(ArgumentList);
  list_Free(ArgumentList);
  return ArgumentListEnd;
}
  
  
TERM in_CreateFunctionTerm(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, representing a term starting with
           a function symbol.
  RETURNS: The term corresponding to the list of tokens.
  EFFECT:  If necessary, symbols are created.
***************************************************************/   
{
  char   *SymbolName;
  int    NumberOfArguments;
  SYMBOL Topsymbol;
  LIST   ArgumentList;

  SymbolName        = ((TOKEN)list_Car(TokenList))->string;
  NumberOfArguments = in_NumberOfArguments(list_Cdr(TokenList));

  if (fol_IsStringPredefined(SymbolName)) {
    printf("\n Line %d: <%s> is a reserved name, "
	   "must not be used for a function.\n",
	   ((TOKEN) list_Car(TokenList))->line,
	   SymbolName);
    misc_Error();
  }

  if ((Topsymbol = symbol_Lookup(SymbolName)) != symbol_Null()) {
    if (!symbol_IsFunction(Topsymbol)) {
      printf("\n Line %d: Expected a function, not <%s>.\n",
	     ((TOKEN) list_Car(TokenList))->line,
	     SymbolName);
      misc_Error();
    }
    
    if (NumberOfArguments != symbol_Arity(Topsymbol)) {
      printf("\n Line %d: Function <%s> used with "
	     "different number of arguments.\n",
	     ((TOKEN) list_Car(TokenList))->line,
	     SymbolName);
      misc_Error();
    }
  }
  else {
    if (in_IsSpecialString(SymbolName)) {
      printf("\n Line %d: <%s> is not allowed as a function name.\n",
	     ((TOKEN) list_Car(TokenList))->line,
	     SymbolName);
      misc_Error();
    }
    
    Topsymbol = symbol_CreateFunction(SymbolName, NumberOfArguments,
				      symbol_STATLEX);
  }

  ArgumentList = in_CreateFunctionArguments(list_Cdr(TokenList));
  list_Delete(TokenList);
  return term_Create(Topsymbol, ArgumentList);
}
 
  
TERM in_CreateAtomTerm(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, representing a term starting with
           a predicate symbol.
  RETURNS: The term corresponding to the list of tokens.
  EFFECT:  If necessary, symbols are created.
***************************************************************/   
{
  char   *SymbolName;
  int    NumberOfArguments;
  SYMBOL Topsymbol;
  LIST   ArgumentList;

  SymbolName        = ((TOKEN)list_Car(TokenList))->string;
  NumberOfArguments = in_NumberOfArguments(list_Cdr(TokenList));

  if ((Topsymbol = symbol_Lookup(SymbolName)) != symbol_Null()) {
    if (!symbol_IsPredicate(Topsymbol))
      if (fol_IsStringPredefined(SymbolName)) {
	printf("\n Line %d: Expected a predicate, "
	       "not <%s> a predefined string.\n",
	       ((TOKEN) list_Car(TokenList))->line,
	       SymbolName);
	misc_Error();
      }
      else {
	printf("\n Line %d: Expected a predicate, not <%s>.\n",
	       ((TOKEN) list_Car(TokenList))->line,
	       SymbolName);
	misc_Error();
      }
    
    if (NumberOfArguments != symbol_Arity(Topsymbol)) {
      printf("\n Line %d: Predicate <%s> used with different "
	     "number of arguments.\n",
	     ((TOKEN) list_Car(TokenList))->line,
	     SymbolName);
      misc_Error();
    }
  }
  else {
    if (in_IsSpecialString(SymbolName)) {
      printf("\n Line %d: <%s> is not allowed as a predicate name.\n",
	     ((TOKEN) list_Car(TokenList))->line,
	     SymbolName);
      misc_Error();
    }

    Topsymbol = symbol_CreatePredicate(SymbolName, NumberOfArguments,
				       symbol_STATLEX);
  }

  ArgumentList = in_CreateFunctionArguments(list_Cdr(TokenList));
  list_Delete(TokenList);
  return term_Create(Topsymbol, ArgumentList);
}
	     
  
LIST in_CreateFormulaArguments(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, representing the arguments of a formula.
  RETURNS: The list of argument terms.
  EFFECT:  Non-existing symbols are created.
***************************************************************/   
{
  LIST   ArgumentList, ArgumentListEnd;
  TOKEN  Token;
  SYMBOL Topsymbol;

  ArgumentList    = list_List(list_Nil());
  ArgumentListEnd = ArgumentList;

  while (!list_Empty(TokenList)) {
    Token = (TOKEN)list_Car(TokenList);

    if (strcmp(Token->string, in_OpenString) == 0) {
      list_Rplacd(ArgumentListEnd, 
		  list_List(in_CreateFormulaTerm(in_ExtractTermList(TokenList))));
      ArgumentListEnd = list_Cdr(ArgumentListEnd);
      TokenList       = in_NextArgument(TokenList);
    }
    else if ((Topsymbol = symbol_Lookup(Token->string)) != symbol_Null() &&
	     symbol_IsPredicate(Topsymbol) &&
	     symbol_Arity(Topsymbol) == 0) {
      list_Rplacd(ArgumentListEnd, list_List(term_Create(Topsymbol, list_Nil())));
      ArgumentListEnd = list_Cdr(ArgumentListEnd);
      TokenList       = in_NextArgument(TokenList);
    }
    else {
      printf("\n Line %d: Expected a formula, not <%s>.\n",
	     Token->line, Token->string);
      misc_Error();
    }
  }
  
  ArgumentListEnd = list_Cdr(ArgumentList);
  list_Free(ArgumentList);
  return ArgumentListEnd;
}


static LIST in_CreateQuantifierVariableList(LIST TokenList, int* Variables)
/**************************************************************
  INPUT:   A list of tokens, representing the variable list of a quantifier.
  RETURNS: The list of variables.
  EFFECT:  If necessary, symbols are created.
***************************************************************/   
{
  LIST   VariableList,VariableListEnd;
  TERM   Variable;

  if (strcmp(((TOKEN)list_Car(TokenList))->string, in_OpenString) != 0) {
    printf("\n Line %d: Expected a list of variables, not <%s>.\n",
	   ((TOKEN) list_Car(TokenList))->line,
	   ((TOKEN) list_Car(TokenList))->string);
    misc_Error();
  }
  
  VariableList    = list_List(list_Nil());
  VariableListEnd = VariableList;
  TokenList       = list_Cdr(TokenList);

  while (strcmp(((TOKEN)list_Car(TokenList))->string, in_CloseString) != 0) {
    Variable = in_CreateVariable(TokenList);
    in_Push(term_TopSymbol(Variable), (((TOKEN)list_Car(TokenList))->string));
    (*Variables)++;
    list_Rplacd(VariableListEnd, list_List(Variable));
    VariableListEnd = list_Cdr(VariableListEnd);
    TokenList = in_NextArgument(TokenList);
  }
  VariableListEnd = list_Cdr(VariableList);
  list_Free(VariableList);
  if (list_Empty(VariableListEnd)) {
    printf("\n Line %d: Expected a list of variables, not ().\n",
	   ((TOKEN) list_Car(TokenList))->line);
    misc_Error();
  }

  return(VariableListEnd);
}


TERM in_CreateFormulaTerm(LIST TokenList)
/**************************************************************
  INPUT:   A list of tokens, representing a term starting with
           a predicate symbol.
  RETURNS: The term corresponding to the list of tokens.
  EFFECT:  If necessary, symbols are created.
***************************************************************/   
{
  char   *SymbolName;
  int    NumberOfArguments;
  SYMBOL Topsymbol;
  LIST   ArgumentList, VariableList;

  SymbolName        = ((TOKEN)list_Car(TokenList))->string;
  NumberOfArguments = in_NumberOfArguments(list_Cdr(TokenList));

  if ((Topsymbol = symbol_Lookup(SymbolName)) != symbol_Null() &&
      fol_IsJunctor(Topsymbol)) {
    if (fol_IsQuantifier(Topsymbol)) {
      int Variables;

      Variables = 0;
      if (NumberOfArguments != 2) {
	printf("\n Line %d: A quantifier must have exactly two arguments.\n",
	       ((TOKEN) list_Car(TokenList))->line);
	misc_Error();
      }

      VariableList = in_CreateQuantifierVariableList(list_Cdr(TokenList),
						     &Variables);
      ArgumentList
	  = in_CreateFormulaArguments(in_NextArgument(list_Cdr(TokenList)));
      list_Delete(TokenList);
      while (Variables-->0)
	in_Pop();
      return(fol_CreateQuantifier(Topsymbol, VariableList, ArgumentList));
    }
    if (((symbol_Equal(Topsymbol, fol_Implies()) ||
	  symbol_Equal(Topsymbol, fol_Equiv())) && NumberOfArguments != 2) ||
	(symbol_Equal(Topsymbol, fol_Not()) && NumberOfArguments != 1)) {
      printf("\n Line %d: <%s>, <%s> must have exactly two arguments, "
	     "<%s> must have exactly one argument.\n",
	     ((TOKEN) list_Car(TokenList))->line,
	     symbol_Name(fol_Implies()),
	     symbol_Name(fol_Equiv()),
	     symbol_Name(fol_Not()));
      misc_Error();
    }

    ArgumentList = in_CreateFormulaArguments(list_Cdr(TokenList));
    list_Delete(TokenList);
    return term_Create(Topsymbol, ArgumentList);  
  }
  else
    return in_CreateAtomTerm(TokenList);
}
