/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *         EXTENDED KNUTH BENDIX ORDERING                 * */
/* *                                                        * */
/* *  $Module:   KBO                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.50 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/01 14:14:38 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: kbo.c,v $ */

#include "kbo.h"

const int kbo_MINWEIGHT = 1;

NAT  kbo_VARCOUNT[symbol__MAXVARIABLES][2];

/**************************************************************/
/*  Functions to implement an order on terms                  */
/**************************************************************/

int kbo_TermWeight(TERM Term)
/**************************************************************
  INPUT:   A term.
  RETURNS: The weight of the term.
***************************************************************/
{
  LIST Stack;
  int  Weight;

#if DEBUG
  if (!term_IsTerm(Term)) {
    printf("\n\nIllegal input in kbo_TermWeight\n");
    misc_DumpCore();
  }
#endif

  Stack  = list_StackBottom();
  Weight = kbo_SymbolWeight(term_TopSymbol(Term));
/*  printf("\n kbo_SymbolWeight(%d)\n",
    kbo_SymbolWeight(term_TopSymbol(Term)); */

  do {
    if (term_IsComplex(Term))
      Stack = list_Push(term_ArgumentList(Term), Stack);
    while (!list_StackEmpty(Stack) && list_Empty(list_Top(Stack)))
      Stack = list_Pop(Stack);
    if (!list_StackEmpty(Stack)) {
      Term     = (TERM)list_Car(list_Top(Stack));
      Weight += kbo_SymbolWeight(term_TopSymbol(Term));
      /* printf("\n kbo_SymbolWeight(%d)\n",
	 kbo_SymbolWeight(term_TopSymbol(Term)); */
      list_RplacTop(Stack, list_Cdr(list_Top(Stack)));
    }
  } while (!list_StackEmpty(Stack));

  return Weight;
}

int kbo_ContCompVarCondAndWeightIntern(CONTEXT Context, TERM Term, int Index)
{
  int Weight;

  Weight = 0;
  Term   = cont_Deref(&Context,Term);

  if (term_IsStandardVariable(Term)) {
    kbo_VARCOUNT[term_TopSymbol(Term)][Index]++;
    Weight += kbo_MINWEIGHT;
  }
  else {
    LIST Scan;
    Weight += symbol_Weight(term_TopSymbol(Term));
    for(Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      Weight += kbo_ContCompVarCondAndWeightIntern(Context, list_Car(Scan), Index);
  }

  return Weight;
}

int kbo_ContCompVarCondAndWeight(CONTEXT Context1, TERM Term1, BOOL *VarCond1, 
				 CONTEXT Context2, TERM Term2, BOOL *VarCond2)
/**************************************************************
  INPUT:   Two contexts, two terms and two pointers to booleans.
  EFFECT:  Sets the booleans with respect to the kbo variable condition.
           Computes the kbo weight difference.
	   The terms are interpreted with respect to the bindings in the respective
	   contexts.
***************************************************************/
{
  SYMBOL MaxVar1,MaxVar2;
  int    i,Weight;

  *VarCond1 = *VarCond2 = TRUE;
  MaxVar1   = cont_TermMaxVar(Context1,Term1);
  MaxVar2   = cont_TermMaxVar(Context2,Term2);

  if (MaxVar1<MaxVar2)
    MaxVar1 = MaxVar2;

  for (i=0;i<=MaxVar1;i++) {
    kbo_VARCOUNT[i][0] = 0;
    kbo_VARCOUNT[i][1] = 0;
  }

  Weight = kbo_ContCompVarCondAndWeightIntern(Context1, Term1, 0);
  Weight = Weight - kbo_ContCompVarCondAndWeightIntern(Context2, Term2, 1);

  for (i=0;i<=MaxVar1;i++) {
    if (kbo_VARCOUNT[i][0] < kbo_VARCOUNT[i][1]) {
      *VarCond1 = FALSE;
      if (!*VarCond2)
	return Weight;
    }
   if (kbo_VARCOUNT[i][0] > kbo_VARCOUNT[i][1]) {
      *VarCond2 = FALSE;
      if (!*VarCond1)
	return Weight;
    }
  }
  return Weight;  
}


int kbo_CompVarCondAndWeight(TERM Term1, BOOL *VarCond1, TERM Term2, BOOL *VarCond2)
/**************************************************************
  INPUT:   Two terms and two pointers to booleans.
  EFFECT:  Sets the booleans with respect to the kbo variable condition.
           Computes the kbo weight difference.
***************************************************************/
{
  SYMBOL MaxVar1,MaxVar2;
  TERM   Term;
  LIST   Scan;
  int    i,Stack,Weight;

  *VarCond1 = *VarCond2 = TRUE;
  MaxVar1   = term_MaxVar(Term1);
  MaxVar2   = term_MaxVar(Term2);
  Stack     = stack_Bottom();
  Weight    = 0;

  if (MaxVar1<MaxVar2)
    MaxVar1 = MaxVar2;

  for (i=0;i<=MaxVar1;i++) {
    kbo_VARCOUNT[i][0] = 0;
    kbo_VARCOUNT[i][1] = 0;
  }

  Term = Term1;
  if (term_IsStandardVariable(Term)) {
    kbo_VARCOUNT[term_TopSymbol(Term)][0]++;
    Weight += kbo_MINWEIGHT;
  }
  else {
    Weight += symbol_Weight(term_TopSymbol(Term));
    if (term_IsComplex(Term))
      stack_Push(term_ArgumentList(Term));
  }
  while (!stack_Empty(Stack)) {
    Scan = stack_Top();
    Term = (TERM)list_Car(Scan);
    stack_RplacTop(list_Cdr(Scan));
    if (term_IsStandardVariable(Term)) {
      Weight += kbo_MINWEIGHT;
      kbo_VARCOUNT[term_TopSymbol(Term)][0]++;
    }
    else {
      Weight += symbol_Weight(term_TopSymbol(Term));
      if (term_IsComplex(Term))
	stack_Push(term_ArgumentList(Term));
    }
    while (!stack_Empty(Stack) && list_Empty(stack_Top()))
      stack_Pop();
  }

  Term = Term2;
  if (term_IsStandardVariable(Term)) {
    Weight -= kbo_MINWEIGHT;
    kbo_VARCOUNT[term_TopSymbol(Term)][1]++;
  }
  else {
    Weight -= symbol_Weight(term_TopSymbol(Term));
    if (term_IsComplex(Term))
      stack_Push(term_ArgumentList(Term));
  }
  while (!stack_Empty(Stack)) {
    Scan = stack_Top();
    Term = (TERM)list_Car(Scan);
    stack_RplacTop(list_Cdr(Scan));
    if (term_IsStandardVariable(Term)) {
      Weight -= kbo_MINWEIGHT;
      kbo_VARCOUNT[term_TopSymbol(Term)][1]++;
    }
    else {
      Weight -= symbol_Weight(term_TopSymbol(Term));
      if (term_IsComplex(Term))
	stack_Push(term_ArgumentList(Term));
    }
    while (!stack_Empty(Stack) && list_Empty(stack_Top()))
      stack_Pop();
  }

  for (i=0;i<=MaxVar1;i++) {
    if (kbo_VARCOUNT[i][0] < kbo_VARCOUNT[i][1]) {
      *VarCond1 = FALSE;
      if (!*VarCond2)
	return Weight;
    }
   if (kbo_VARCOUNT[i][0] > kbo_VARCOUNT[i][1]) {
      *VarCond2 = FALSE;
      if (!*VarCond1)
	return Weight;
    }
  }
  return Weight;  
}

kbo_RESULT kbo_Not(kbo_RESULT Result)
/**************************************************************
  INPUT:   A kbo_RESULT
  RETURNS: The negation with respect to kbo argument switching.
***************************************************************/
{
  if (Result == kbo_UNCOMPARABLE || Result == kbo_EQUAL)
    return Result;

  if (Result == kbo_GREATER_THAN)
    return kbo_SMALLER_THAN;

  return kbo_GREATER_THAN;
}
  

kbo_RESULT kbo_CompareStruc(TERM Term1, TERM Term2, int WeightDiff)
/**************************************************************
  INPUT:   Two terms where the kbo-variable condition for <Term1> and
           <Term2> is satisfied and <WeightDiff> is the kbo weight difference
	   between <Term1> and <Term2> 
  RETURNS: kbo_UNCOMPARABLE, if Term1 and Term2 are uncomparable
	   kbo_EQUAL,        if Term1 and Term2 are equal
	   kbo_GREATER_THAN, if Term1 is greater than Term2 
***************************************************************/
{
  LIST   Scan1,Scan2;
  SYMBOL Top1,Top2;

  Top1       = term_TopSymbol(Term1);
  Top2       = term_TopSymbol(Term2);

  /*printf("\n\tKBOCompareStruc: ");
  fol_PrintDFG(Term1);
  printf(" ");
  fol_PrintDFG(Term2);
  printf(" WeightDiff: %d",WeightDiff);*/

  if (WeightDiff>0)
    return kbo_GREATER_THAN;
  else 
    if (WeightDiff==0)
      if (symbol_IsStandardVariable(Top1))
	if (symbol_IsStandardVariable(Top2))
	  return kbo_EQUAL;
	else
	  return kbo_UNCOMPARABLE;
      else
	if (symbol_IsStandardVariable(Top2) || symbol_PrecedenceGreater(Top1,Top2))
	  return kbo_GREATER_THAN;
	else
	  if (Top1 == Top2) {
	    int    RecWeightDiff;
	    BOOL   T1VarCond, T2VarCond;
	    TERM   RecTerm1,RecTerm2;
	    Scan1 = term_ArgumentList(Term1);
	    Scan2 = term_ArgumentList(Term2);
	    if (symbol_HasProperty(Top1,ORDRIGHT)) {
	      int i;
	      for (i=symbol_Arity(Top1);
		   i>0 && term_Equal(list_NthElement(Scan1,i),list_NthElement(Scan2,i));
		   i--);
	      if (i>0) {
		RecTerm1 = (TERM)list_NthElement(Scan1,i);
		RecTerm2 = (TERM)list_NthElement(Scan2,i);
	      }
	      else
		return kbo_EQUAL;
	    }
	    else {	      
	      while (!list_Empty(Scan1) && term_Equal(list_Car(Scan1),list_Car(Scan2))) {
		Scan1 = list_Cdr(Scan1);
		Scan2 = list_Cdr(Scan2);
	      }
	      if (list_Empty(Scan1)) /* Implies that list_Empty(Scan2)  */
		return kbo_EQUAL;
	      else {
		RecTerm1 = (TERM)list_Car(Scan1);
		RecTerm2 = (TERM)list_Car(Scan2);
	      }
	    }	      
	    RecWeightDiff =  kbo_CompVarCondAndWeight(RecTerm1,&T1VarCond,RecTerm2,&T2VarCond);
	    if (RecWeightDiff >= 0 && T1VarCond)
	      return kbo_CompareStruc(RecTerm1, RecTerm2, RecWeightDiff);
	    else
	      return kbo_UNCOMPARABLE;
	  }
	  else
	    return kbo_UNCOMPARABLE;
    else
      return kbo_UNCOMPARABLE;

  return kbo_UNCOMPARABLE;
}
  

kbo_RESULT kbo_Compare(TERM Term1, TERM Term2)
/**************************************************************
  INPUT:   Two terms.
  RETURNS: kbo_UNCOMPARABLE, if Term1 and Term2 are uncomparable because of
                              different variables,
	   kbo_EQUAL,        if Term1 and Term2 are comparable and have the
	                      same weight,
	   kbo_GREATER_THAN, if Term1 is greater than Term2 wrt the kbo with
	                      the actual precedence kbo_Prec and the given 
			      symbol_Weights,
	   kbo_SMALLER_THAN, else.
***************************************************************/
{
  int        WeightDiff;
  BOOL       T1VarCond, T2VarCond;
  kbo_RESULT Result;

#if DEBUG
  if ((!term_IsTerm(Term1)) || (!term_IsTerm(Term2)) ) {
    printf("\n\nIllegal input in kbo_Compare\n");
    misc_DumpCore();
  }
#endif


  WeightDiff =  kbo_CompVarCondAndWeight(Term1,&T1VarCond,Term2,&T2VarCond);

  /*printf("\n\tKBOCompare: ");
  fol_PrintDFG(Term1);
  printf(" ");
  fol_PrintDFG(Term2);
  printf(" WeightDiff: %d Cond1: %d Cond2: %d ",WeightDiff,T1VarCond,T2VarCond);*/

  if (T1VarCond && !T2VarCond)
    return kbo_CompareStruc(Term1,Term2,WeightDiff);

  if (!T1VarCond && T2VarCond)
    return kbo_Not(kbo_CompareStruc(Term2,Term1,-WeightDiff));

  if (T1VarCond && T2VarCond) {
    Result = kbo_CompareStruc(Term1,Term2,WeightDiff);
    if (Result == kbo_UNCOMPARABLE)
      return kbo_Not(kbo_CompareStruc(Term2,Term1,-WeightDiff));
    else
      return Result;
  }

  return kbo_UNCOMPARABLE;
}

kbo_RESULT kbo_ContCompareStruc(CONTEXT Context1, TERM Term1, CONTEXT Context2, TERM Term2, int WeightDiff)
/**************************************************************
  INPUT:   Two contexts and two terms where the kbo-variable condition for <Term1> and
           <Term2> is satisfied and <WeightDiff> is the kbo weight difference
	   between <Term1> and <Term2> 
  RETURNS: kbo_UNCOMPARABLE, if Term1 and Term2 are uncomparable
	   kbo_EQUAL,        if Term1 and Term2 are equal
	   kbo_GREATER_THAN, if Term1 is greater than Term2 
	   The Terms are interpreted with respect to the contexts.
***************************************************************/
{
  LIST   Scan1,Scan2;
  SYMBOL Top1,Top2;

  Term1      = cont_Deref(&Context1,Term1);
  Term2      = cont_Deref(&Context2,Term2);
  Top1       = term_TopSymbol(Term1);
  Top2       = term_TopSymbol(Term2);

  /*printf("\n\tKBOStrCompare: ");
  term_PrintPrefix(Term1);
  printf(" ");
  term_PrintPrefix(Term2);
  printf(" WeightDiff: %d",WeightDiff);*/

  if (WeightDiff>0)
    return kbo_GREATER_THAN;
  else 
    if (WeightDiff==0)
      if (symbol_IsStandardVariable(Top1))
	if (symbol_IsStandardVariable(Top2))
	  return kbo_EQUAL;
	else
	  return kbo_UNCOMPARABLE;
      else
	if (symbol_IsStandardVariable(Top2) || symbol_PrecedenceGreater(Top1,Top2))
	  return kbo_GREATER_THAN;
	else
	  if (Top1 == Top2) {
	    int    RecWeightDiff;
	    BOOL   T1VarCond, T2VarCond;
	    TERM   RecTerm1,RecTerm2;
	    Scan1 = term_ArgumentList(Term1);
	    Scan2 = term_ArgumentList(Term2);
	    if (symbol_HasProperty(Top1,ORDRIGHT)) {
	      int i;
	      for (i=symbol_Arity(Top1);
		   i>0 && cont_TermEqual(Context1,list_NthElement(Scan1,i),
					 Context2,list_NthElement(Scan2,i));
		   i--);
	      if (i>0) {
		RecTerm1 = cont_Deref(&Context1,list_NthElement(Scan1,i));
		RecTerm2 = cont_Deref(&Context2,list_NthElement(Scan2,i));
	      }
	      else
		return kbo_EQUAL;
	    }
	    else {	      
	      while (!list_Empty(Scan1) && term_Equal(list_Car(Scan1),list_Car(Scan2))) {
		Scan1 = list_Cdr(Scan1);
		Scan2 = list_Cdr(Scan2);
	      }
	      if (list_Empty(Scan1)) /* Implies that list_Empty(Scan2)  */
		return kbo_EQUAL;
	      else {
		RecTerm1 = cont_Deref(&Context1,list_Car(Scan1));
		RecTerm2 = cont_Deref(&Context2,list_Car(Scan2));
	      }
	    }	      
	    RecWeightDiff =  kbo_ContCompVarCondAndWeight(Context1,RecTerm1,&T1VarCond,
							  Context2,RecTerm2,&T2VarCond);
	    if (RecWeightDiff >= 0 && T1VarCond)
	      return kbo_ContCompareStruc(Context1, RecTerm1, Context2, RecTerm2, RecWeightDiff);
	    else
	      return kbo_UNCOMPARABLE;
	  }
	  else
	    return kbo_UNCOMPARABLE;
    else
      return kbo_UNCOMPARABLE;

  return kbo_UNCOMPARABLE;
}
  

kbo_RESULT kbo_ContCompare(CONTEXT Context1, TERM Term1, CONTEXT Context2, TERM Term2)
/**************************************************************
  INPUT:   Two contexts and two terms.
  RETURNS: kbo_UNCOMPARABLE, if Term1 and Term2 are uncomparable because of
                              different variables,
	   kbo_EQUAL,        if Term1 and Term2 are comparable and have the
	                      same weight,
	   kbo_GREATER_THAN, if Term1 is greater than Term2 wrt the kbo with
	                      the actual precedence kbo_Prec and the given 
			      symbol_Weights,
	   kbo_SMALLER_THAN, else.
	   The Terms are interpreted with respect to the contexts.
***************************************************************/
{
  int        WeightDiff;
  BOOL       T1VarCond, T2VarCond;
  kbo_RESULT Result;

#if DEBUG
  if ((!term_IsTerm(Term1)) || (!term_IsTerm(Term2)) ) {
    printf("\n\nIllegal input in kbo_Compare\n");
    misc_DumpCore();
  }
#endif


  WeightDiff =  kbo_ContCompVarCondAndWeight(Context1, Term1, &T1VarCond, Context2, Term2, &T2VarCond);

  /*printf("\n\tKBOCompare: ");
  cont_TermPrintPrefix(Context1,Term1);
  printf(" ");
  cont_TermPrintPrefix(Context2,Term2);
  printf(" WeightDiff: %d Cond1: %d Cond2: %d ",WeightDiff,T1VarCond,T2VarCond);*/

  if (T1VarCond && !T2VarCond)
    return kbo_ContCompareStruc(Context1,Term1,Context2,Term2,WeightDiff);

  if (!T1VarCond && T2VarCond)
    return kbo_Not(kbo_ContCompareStruc(Context2,Term2,Context1,Term1,-WeightDiff));

  if (T1VarCond && T2VarCond) {
    Result = kbo_ContCompareStruc(Context1,Term1,Context2,Term2,WeightDiff);
    if (Result == kbo_UNCOMPARABLE)
      return kbo_Not(kbo_ContCompareStruc(Context2,Term2,Context1,Term1,-WeightDiff));
    else
      return Result;
  }

  return kbo_UNCOMPARABLE;
}

BOOL kbo_ContGreaterCompareStruc(CONTEXT Context1, TERM Term1, CONTEXT Context2, TERM Term2)
/**************************************************************
  INPUT:   Two contexts and two terms where the kbo-variable condition for <Term1> and
           <Term2> is satisfied as well as the weight difference between the terms is zero.
 RETURNS:  TRUE if Term1 is greater than Term2 
	   The Terms are interpreted with respect to the contexts.
***************************************************************/
{
  LIST   Scan1,Scan2;
  SYMBOL Top1,Top2;

  Term1      = cont_Deref(&Context1,Term1);
  Term2      = cont_Deref(&Context2,Term2);
  Top1       = term_TopSymbol(Term1);
  Top2       = term_TopSymbol(Term2);

  /*printf("\n\tKBOStrCompare: ");
    term_PrintPrefix(Term1);
    printf(" ");
    term_PrintPrefix(Term2);*/

 
  if (symbol_IsStandardVariable(Top1))
    if (symbol_IsStandardVariable(Top2))
      return FALSE;
    else
      return FALSE;
  else
    if (symbol_IsStandardVariable(Top2) || symbol_PrecedenceGreater(Top1,Top2))
      return TRUE;
    else
      if (Top1 == Top2) {
	int    RecWeightDiff;
	BOOL   T1VarCond, T2VarCond;
	TERM   RecTerm1,RecTerm2;
	Scan1 = term_ArgumentList(Term1);
	Scan2 = term_ArgumentList(Term2);
	if (symbol_HasProperty(Top1,ORDRIGHT)) {
	  int i;
	  for (i=symbol_Arity(Top1);
	       i>0 && cont_TermEqual(Context1,list_NthElement(Scan1,i),
				     Context2,list_NthElement(Scan2,i));
	       i--);
	  if (i>0) {
	    RecTerm1 = cont_Deref(&Context1,list_NthElement(Scan1,i));
	    RecTerm2 = cont_Deref(&Context2,list_NthElement(Scan2,i));
	  }
	  else
	    return FALSE;
	}
	else {	      
	  while (!list_Empty(Scan1) && term_Equal(list_Car(Scan1),list_Car(Scan2))) {
	    Scan1 = list_Cdr(Scan1);
	    Scan2 = list_Cdr(Scan2);
	  }
	  if (list_Empty(Scan1)) /* Implies that list_Empty(Scan2)  */
	    return FALSE;
	  else {
	    RecTerm1 = cont_Deref(&Context1,list_Car(Scan1));
	    RecTerm2 = cont_Deref(&Context2,list_Car(Scan2));
	  }
	}	      
	RecWeightDiff =  kbo_ContCompVarCondAndWeight(Context1,RecTerm1,&T1VarCond,
						      Context2,RecTerm2,&T2VarCond);

	if (T1VarCond)
	  if (RecWeightDiff > 0)
	    return TRUE;
	  else
	    if (RecWeightDiff == 0)
	      return kbo_ContGreaterCompareStruc(Context1, RecTerm1, Context2, RecTerm2);
      }
  
  return FALSE;
}


BOOL kbo_ContGreater(CONTEXT Context1, TERM Term1, CONTEXT Context2, TERM Term2)
/**************************************************************
  INPUT:   Two contexts and two terms.
  RETURNS: TRUE, if Term1 is greater than Term2 wrt the kbo with
           the actual precedence kbo_Prec and the given symbol_Weights
***************************************************************/
{
  int        WeightDiff;
  BOOL       T1VarCond, T2VarCond;

#if DEBUG
  if ((!term_IsTerm(Term1)) || (!term_IsTerm(Term2)) ) {
    printf("\n\nIllegal input in kbo_Compare\n");
    misc_DumpCore();
  }
#endif


  WeightDiff =  kbo_ContCompVarCondAndWeight(Context1, Term1, &T1VarCond, Context2, Term2, &T2VarCond);

  /*printf("\n\tKBOCompare: ");
  cont_TermPrintPrefix(Context1,Term1);
  printf(" ");
  cont_TermPrintPrefix(Context2,Term2);
  printf(" WeightDiff: %d Cond1: %d Cond2: %d ",WeightDiff,T1VarCond,T2VarCond);*/

  if (T1VarCond)
    if (WeightDiff > 0)
      return TRUE;
    else
      if (WeightDiff == 0) 
	return kbo_ContGreaterCompareStruc(Context1,Term1,Context2,Term2);

  return FALSE;
}


kbo_RESULT kbo_LiteralCompare(TERM Lit1, BOOL Orient1, TERM Lit2, BOOL Orient2, BOOL Check)
/*********************************************************
  INPUT:   Two term literals and two flags indicating of whether these
           two literals are oriented equalities. 
	   Additionally a check flag that indicates whether
	   the orietation flags are sufficient or necessary and sufficient:
	   If Check=TRUE then if flag is FALSE it is interpretated that it is not
	   known whether the (possible) equation can be oriented oriented, however
	   if flag is TRUE we assume that it is oriented.
  RETURNS: The result if these literals are compared with
           respect to XKBO. Dependent on their sign, the literals
	   are compared as multiset of their topsymbol terms,
	   where any literal is considered to be an equation
	   and non equational literals are considered to be
	   equations with the artificial, non-existent, minimal
	   constant tt.
  EFFECT:  The arguments of the literals are possibly, destructively flipped.
**********************************************************/
{
  kbo_RESULT Result,AuxResult,AuxRl1r2,AuxRr1r2,AuxRr1l2,Comp1,Comp2;
  BOOL       Pos1,Pos2;

  Pos1  = Pos2  = TRUE;
  Comp1 = Comp2 = Result = kbo_UNCOMPARABLE;
  if (symbol_Equal(term_TopSymbol(Lit1),fol_Not())) {
    Lit1 = term_FirstArgument(Lit1);
    Pos1 = FALSE;
  }
  if (symbol_Equal(term_TopSymbol(Lit2),fol_Not())) {
    Lit2 = term_FirstArgument(Lit2);
    Pos2 = FALSE;
  }
      
  if (fol_IsEquality(Lit1))           /* Real equation */
    if (fol_IsEquality(Lit2)) {
      TERM       l1,r1,l2,r2,aux;
      l1 = term_FirstArgument(Lit1);
      r1 = term_SecondArgument(Lit1);
      l2 = term_FirstArgument(Lit2);
      r2 = term_SecondArgument(Lit2);
      if (Orient1 || 
	  (Check && 
	   ((Comp1 = kbo_Compare(l1,r1)) == kbo_GREATER_THAN || Comp1 == kbo_SMALLER_THAN))) {
	if (Comp1 == kbo_SMALLER_THAN) {
	  aux = l1; l1 = r1; r1 = aux;
	}
	if (Orient2 ||
	    (Check && 
	     ((Comp2 = kbo_Compare(l2,r2)) == kbo_GREATER_THAN || Comp2 == kbo_SMALLER_THAN))) {
	  if (Comp2 == kbo_SMALLER_THAN) {
	    aux = l2; l2 = r2; r2 = aux;
	  }
	  Result = kbo_Compare(l1,l2);
	  if (Result == kbo_EQUAL)
	    if (Pos1)
	      if (Pos2)
		return kbo_Compare(r1,r2);
	      else
		return kbo_SMALLER_THAN;
	    else
	      if (Pos2)
		return kbo_GREATER_THAN;
	      else
		return kbo_Compare(r1,r2);
	  return Result;
	}
	else { /* Lit2 is not oriented equation */
	  if (term_Equal(l1,l2)) {
	    Result = kbo_Compare(r1,r2);
	    if (Result == kbo_EQUAL)
	      if (Pos1 && !Pos2)
		return kbo_SMALLER_THAN;
	      else
		if (!Pos1 && Pos2)
		  return kbo_GREATER_THAN;
	    return Result;
	  }
	  if (term_Equal(l1,r2)) {
	    Result = kbo_Compare(r1,l2);
	    if (Result == kbo_EQUAL)
	      if (Pos1 && !Pos2)
		return kbo_SMALLER_THAN;
	      else
		if (!Pos1 && Pos2)
		  return kbo_GREATER_THAN;
	    return Result;
	  }
	  Result    = kbo_Compare(l1,l2);
	  AuxRl1r2  = kbo_Compare(l1,r2);
	  if (Result == AuxRl1r2)
	    return Result;
	  AuxRr1r2  = kbo_Compare(r1,r2);
	  if (Result == AuxRr1r2)
	    return Result;
	  if (AuxRl1r2 == AuxRr1r2 && AuxRl1r2 == kbo_SMALLER_THAN)
	    return kbo_SMALLER_THAN;
	  AuxRr1l2  = kbo_Compare(r1,l2);
	  if (Result == AuxRr1l2 && Result == kbo_SMALLER_THAN)
	    return  kbo_SMALLER_THAN;	  
	  return kbo_UNCOMPARABLE;                   
	}
      }
      else /* Lit1 is unorientable equation */
	if (Orient2 ||
	    (Check && 
	     ((Comp2 = kbo_Compare(term_FirstArgument(Lit2),
				   term_SecondArgument(Lit2)) == kbo_GREATER_THAN) ||
	      Comp2 == kbo_SMALLER_THAN))) {   /* Lit2 is orientable equation */
	  if (Comp2 == kbo_SMALLER_THAN) {
	    aux = l2; l2 = r2; r2 = aux;
	  }
	  if (term_Equal(l1,l2)) {
	    Result = kbo_Compare(r1,r2);
	    if (Result == kbo_EQUAL)
	      if (Pos1 && !Pos2)
		return kbo_SMALLER_THAN;
	      else
		if (!Pos1 && Pos2)
		  return kbo_GREATER_THAN;
	    return Result;
	  }
	  if (term_Equal(r1,l2)) {
	    Result = kbo_Compare(l1,r2);
	    if (Result == kbo_EQUAL)
	      if (Pos1 && !Pos2)
		return kbo_SMALLER_THAN;
	      else
		if (!Pos1 && Pos2)
		  return kbo_GREATER_THAN;
	    return Result;
	  }
	  
	  Result    = kbo_Compare(l1,l2);
	  AuxRr1l2  = kbo_Compare(r1,l2);
	  if (Result == AuxRr1l2)
	    return Result;
	  AuxRr1r2  = kbo_Compare(r1,r2);
	  if (Result == AuxRr1r2)
	    return Result;
	  if (AuxRr1l2 == AuxRr1r2 && AuxRr1l2 == kbo_GREATER_THAN)
	    return kbo_GREATER_THAN;
	  AuxRl1r2  = kbo_Compare(l1,r2);
	  if (Result == AuxRl1r2 && Result == kbo_GREATER_THAN)
	    return  kbo_GREATER_THAN;	  
	  return kbo_UNCOMPARABLE;
	}
	else { /* Both literals are unorientable equations */
	  if (term_Equal(l1,l2)) {
	    Result = kbo_Compare(r1,r2);
	    if (Result == kbo_EQUAL)
	      if (Pos1 && !Pos2)
		return kbo_SMALLER_THAN;
	      else
		if (!Pos1 && Pos2)
		  return kbo_GREATER_THAN;
	    return Result;
	  }
	  if (term_Equal(r1,l2)) {
	    Result = kbo_Compare(l1,r2);
	    if (Result == kbo_EQUAL)
	      if (Pos1 && !Pos2)
		return kbo_SMALLER_THAN;
	      else
		if (!Pos1 && Pos2)
		  return kbo_GREATER_THAN;
	    return Result;
	  }
	  if (term_Equal(l1,r2)) {
	    Result = kbo_Compare(r1,l2);
	    if (Result == kbo_EQUAL)
	      if (Pos1 && !Pos2)
		return kbo_SMALLER_THAN;
	      else
		if (!Pos1 && Pos2)
		  return kbo_GREATER_THAN;
	    return Result;
	  }
	  if (term_Equal(r1,r2)) {
	    Result = kbo_Compare(l1,l2);
	    if (Result == kbo_EQUAL)
	      if (Pos1 && !Pos2)
		return kbo_SMALLER_THAN;
	      else
		if (!Pos1 && Pos2)
		  return kbo_GREATER_THAN;
	    return Result;
	  }
	  Result    = kbo_Compare(l1,l2);
	  if (Result == kbo_UNCOMPARABLE) {
	    Result = kbo_Compare(l1,r2);
	    if (Result == kbo_UNCOMPARABLE) {
	     if (kbo_Compare(r1,l2) == kbo_GREATER_THAN &&
		 kbo_Compare(r1,r2) == kbo_GREATER_THAN)
	       return kbo_GREATER_THAN;
	     return kbo_UNCOMPARABLE;
	    }
	    if (Result == kbo_GREATER_THAN) {
	      if (kbo_Compare(r1,l2) == kbo_GREATER_THAN)
		return kbo_GREATER_THAN;
	      return kbo_UNCOMPARABLE;
	    }
	    if (Result == kbo_SMALLER_THAN) {
	      if (kbo_Compare(r1,l2) == kbo_SMALLER_THAN ||
		  kbo_Compare(r1,r2) == kbo_SMALLER_THAN)
		return kbo_SMALLER_THAN;
	      return kbo_UNCOMPARABLE;
	    }
	  }
	  if (Result == kbo_GREATER_THAN) {
	    if ((Result = kbo_Compare(l1,r2)) == kbo_GREATER_THAN ||
		(AuxResult = kbo_Compare(r1,r2)) == kbo_GREATER_THAN)
	      return kbo_GREATER_THAN;
	    if (Result == kbo_UNCOMPARABLE || AuxResult == kbo_UNCOMPARABLE)
		return kbo_UNCOMPARABLE;
	    return kbo_SMALLER_THAN;
	  }
	  if (Result == kbo_SMALLER_THAN) {
	    if ((Result = kbo_Compare(r1,l2)) == kbo_SMALLER_THAN ||
		(AuxResult = kbo_Compare(r1,r2)) == kbo_SMALLER_THAN)
	      return kbo_SMALLER_THAN;
	    if (Result == kbo_UNCOMPARABLE || AuxResult == kbo_UNCOMPARABLE)
	      return kbo_UNCOMPARABLE;
	    return kbo_GREATER_THAN;
	  }
	}
    }
    else /* Second Atom is not an equation */
      if (Orient1)
	return kbo_Compare(term_FirstArgument(Lit1),Lit2);  /* They can't be equal ! */
      else {
	Result    = kbo_Compare(term_FirstArgument(Lit1),Lit2);
	if (Result == kbo_UNCOMPARABLE)
	  return Result;
	AuxResult = kbo_Compare(term_SecondArgument(Lit1),Lit2);
	if (Result != AuxResult)
	  return kbo_UNCOMPARABLE;
	return Result;
      }
  else /* First Atom is not an equation */
    if (fol_IsEquality(Lit2))
      if (Orient2)
	return kbo_Compare(Lit1,term_FirstArgument(Lit2));  /* They can't be equal ! */
      else {
	Result    = kbo_Compare(Lit1,term_FirstArgument(Lit2));
	if (Result == kbo_UNCOMPARABLE)
	  return Result;
	AuxResult = kbo_Compare(Lit1,term_SecondArgument(Lit2));
	if (Result != AuxResult)
	  return kbo_UNCOMPARABLE;
	return Result;
      }
    else { /* None of the atoms is an equation */
      Result = kbo_Compare(Lit1,Lit2);
      if (Result == kbo_EQUAL)
	if (Pos1)
	  if (Pos2)
	    return Result;
	  else
	    return kbo_SMALLER_THAN;
	else
	  if (Pos2)
	    return kbo_GREATER_THAN;
	  else
	    return Result;
      return Result;
    }

  return Result;
}





void kbo_Print(kbo_RESULT Result)
/**************************************************************
  INPUT:   A kbo_Result.
  RETURNS: None, prints the Result as a string to stdout.
***************************************************************/
{
  switch(Result) { 
  case kbo_UNCOMPARABLE:  printf(" uncomparable "); break;
  case kbo_EQUAL:         printf(" equal "); break;
  case kbo_GREATER_THAN:  printf(" greater than "); break;
  case kbo_SMALLER_THAN:  printf(" smaller than "); break;
  default:                printf(" Nonsense! ");
  }
}
