/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                     LISTS                              * */
/* *                                                        * */
/* *  $Module:   LIST                                       * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.49 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/01 14:06:44 $                             * */
/* $Author: weidenb $                                         * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: list.c,v $ */

#include "list.h"

/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  MEMORY MANAGEMENT                                     * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

LIST list_Copy(const LIST List)
/**************************************************************
  INPUT:   A List.
  RETURNS: The copy of the list.
  CAUTION: The entries of the list are NOT copied !
***************************************************************/
{
  LIST Copy;
  LIST Scan1,Scan2;

 
  if (list_Empty(List))
    return(list_Nil());

  Copy = list_List(list_Car(List));
  Scan1 = Copy;
  Scan2 = list_Cdr(List);

  while (!list_Empty(Scan2)) {
    list_Rplacd(Scan1, list_List(list_Car(Scan2)));
    Scan1 = list_Cdr(Scan1);
    Scan2 = list_Cdr(Scan2);
  }
  return Copy;
}


LIST list_CopyWithElement(const LIST List, POINTER (*CopyElement)())
/**************************************************************
  INPUT:   A List and a copy function for the elements.
  RETURNS: The copy of the list.
  CAUTION: The entries of the list are NOT copied !
***************************************************************/
{
  LIST Copy;
  LIST Scan1,Scan2;

  if (list_Empty(List))
    return(list_Nil());

  Copy  = list_List(CopyElement(list_Car(List)));
  Scan1 = Copy;
  Scan2 = list_Cdr(List);

  while (!list_Empty(Scan2)) {
    list_Rplacd(Scan1, list_List(CopyElement(list_Car(Scan2))));
    Scan1 = list_Cdr(Scan1);
    Scan2 = list_Cdr(Scan2);
  }
  return Copy;
}


void list_InsertNext(LIST List, POINTER Pointer)
/**************************************************************
  INPUT:	 A list and a pointer to anything.
  RETURNS: A list with Pointer being added at the position that
           follows List.
  SUMMARY: We enqueue the element at position list_Cdr(List);
***************************************************************/
{
#if DEBUG
  if (list_Empty(List)) {
    printf("\n\n NULL Pointer in list_InsertNext\n");
    misc_Error();
  }
#endif

  list_Rplacd(List, list_Cons(Pointer, list_Cdr(List)));
}


LIST list_NMapCar(LIST List, POINTER (*ElementFunc)())
/**************************************************************
  INPUT:   A List and a function for the elements.
  RETURNS: The List where all elements are replaced by the result of 
           the function calls of <ElementFunc> to the elements
  CAUTION: The List is not copied !
***************************************************************/
{
  LIST Scan;

  for (Scan = List; !list_Empty(Scan); Scan = list_Cdr(Scan))
    list_Rplaca(Scan, ElementFunc(list_Car(Scan)));

  return List;
}


void list_Apply(void (*Function)(),  LIST List)
/**************************************************************
  INPUT:	 A pointer to a non-resulting function.
  SUMMARY: Apply the function to all members of the list.
***************************************************************/
{
  while (!list_Empty(List)) {
    Function(list_Car(List));
    List = list_Cdr(List);
  }
}


LIST list_Reverse(const LIST List)
/**************************************************************
  INPUT:   A list
  RETURNS: A new list where the order of the elements is reversed.
***************************************************************/
{
  LIST ReverseList;
  LIST Scan;
  
  ReverseList = list_Nil();

  for (Scan=List;!list_Empty(Scan);Scan=list_Cdr(Scan)) 
    ReverseList = list_Cons(list_Car(Scan), ReverseList);

  return ReverseList;
}


LIST list_NReverse(LIST List)
/**************************************************************
  INPUT:	 A list
  RETURNS: The same list with reversed order of items.
  CAUTION: Destructive.
***************************************************************/
{
  LIST ReverseList;
  LIST Scan1;
  LIST Scan2;
  
  ReverseList = list_Nil();

  for (Scan1=List;!list_Empty(Scan1);Scan1=list_Cdr(Scan1)) 
    ReverseList = list_Cons(list_Car(Scan1),ReverseList);

  for (Scan1=List, Scan2=ReverseList;
       !list_Empty(Scan1);
       Scan1=list_Cdr(Scan1), Scan2=list_Cdr(Scan2)) 
    list_Rplaca(Scan1, list_Car(Scan2));

  list_Delete(ReverseList);
  return List;
}


LIST list_PointerSort(LIST List)
/**************************************************************
  INPUT:	 A list
  RETURNS: The same list where the elements are sorted as pointers.
  CAUTION: Destructive.
***************************************************************/
{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1=List;!list_Empty(Scan1);Scan1=list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2=list_Cdr(Scan1);!list_Empty(Scan2);Scan2=list_Cdr(Scan2)) {
      if ((NAT)list_Car(Scan2)<(NAT)list_Car(Min)) {
	Exchange = list_Car(Min);
	list_Rplaca(Min,list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
    }
  }

  return(List);
}


LIST list_Sort(LIST List, BOOL (*Test)())
/**************************************************************
  INPUT:	 A list and a 'less' function on the elements.
  RETURNS: The same list where the elements are sorted with respect to Test.
  CAUTION: Destructive.
***************************************************************/
{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1=List; !list_Empty(Scan1); Scan1=list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2 = list_Cdr(Scan1);
	 !list_Empty(Scan2);
	 Scan2 = list_Cdr(Scan2))
      if (Test(list_Car(Scan2), list_Car(Min))) {
	Exchange = list_Car(Min);
	list_Rplaca(Min, list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
  }

  return(List);
}

LIST list_NumberSort(LIST List, NAT (*Number)())
/**************************************************************
  INPUT:   A list and function mapping elements to numbers.
  RETURNS: The same list where the elements are sorted with respect to < and the
           Number function.
  CAUTION: Destructive.
***************************************************************/
{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1=List; !list_Empty(Scan1); Scan1=list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2 = list_Cdr(Scan1);
	 !list_Empty(Scan2);
	 Scan2 = list_Cdr(Scan2))
      if (Number(list_Car(Scan2)) < Number(list_Car(Min))) {
	Exchange = list_Car(Min);
	list_Rplaca(Min, list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
  }

  return(List);
}

LIST list_GreaterNumberSort(LIST List, NAT (*Number)())
/**************************************************************
  INPUT:   A list and function mapping elements to numbers.
  RETURNS: The same list where the elements are sorted with respect to > and the
           Number function.
  CAUTION: Destructive.
***************************************************************/
{
  LIST Scan1,Scan2,Min;
  POINTER Exchange;
  
  for (Scan1=List; !list_Empty(Scan1); Scan1=list_Cdr(Scan1)) {
    Min = Scan1;
    for (Scan2 = list_Cdr(Scan1);
	 !list_Empty(Scan2);
	 Scan2 = list_Cdr(Scan2))
      if (Number(list_Car(Scan2)) > Number(list_Car(Min))) {
	Exchange = list_Car(Min);
	list_Rplaca(Min, list_Car(Scan2));
	list_Rplaca(Scan2, Exchange);
      }
  }

  return(List);
}

LIST list_NNumberMerge(LIST List1, LIST List2, NAT (*Number)())
/**************************************************************
  INPUT:   Two sorted lists and function mapping elements to numbers.
  RETURNS: The merge of the lists where the elements are sorted with respect to < and the
           Number function.
  CAUTION: Destructive.
***************************************************************/
{
  LIST    Scan1,Scan2,Result,ResultStart;

  if (list_Empty(List1))
    return List2;

  if (list_Empty(List2))
    return List1;

  Result      = list_List(NULL);
  ResultStart = Result;
  Scan1       = List1;
  Scan2       = List2;

  while (!list_Empty(Scan1) && !list_Empty(Scan2)) {
    if (Number(list_Car(Scan1)) <= Number(list_Car(Scan2))) {
      list_Rplacd(Result,Scan1);
      Scan1  = list_Cdr(Scan1);
    }
    else {
      list_Rplacd(Result,Scan2);
      Scan2  = list_Cdr(Scan2);
    }
    Result = list_Cdr(Result);
  }
  if (list_Empty(Scan1))
    list_Rplacd(Result,Scan2);
  else
    list_Rplacd(Result,Scan1);
  Result = list_Cdr(ResultStart);
  list_Free(ResultStart);
  return Result;
}
      


POINTER list_DequeueNext(LIST List)
/**************************************************************
  INPUT:   A list
  RETURNS: A pointer to a dequeued element.
  SUMMARY: We dequeue the element pointed to by list_Cdr(List);
***************************************************************/
{
  POINTER Pointer;
  LIST    Memo;

  if (list_Empty(List)) return (POINTER)NULL;

  Memo = list_Cdr(List);

  if (list_Empty(Memo)) return (POINTER)NULL;
      
  Pointer = list_Car(Memo);
  list_Rplacd(List, Memo->cdr);
  memory_Free(Memo, sizeof(LIST_NODE));
  return Pointer;
}


POINTER list_NthElement(LIST List, NAT Number)
/**************************************************************
  INPUT:   A List and a natural number.
  RETURNS: The <Number>th element of the list, NULL otherwise
***************************************************************/
{
  while (!list_Empty(List) && --Number > 0)
    List = list_Cdr(List);
  
  if (list_Empty(List))
    return NULL;
  else
    return (list_Car(List));
}


void list_DeleteWithElement(LIST List, void (*ElementDelete)())
/**************************************************************
  INPUT:   A List and a delete function for the elements.
  RETURNS: Nothing.
  EFFECT:  The List and all its elements are deleted.
***************************************************************/
{
  LIST Scan;

  while (!list_Empty(List)) {
    Scan = list_Cdr(List);
    ElementDelete(list_Car(List));
    list_Free(List);
    List = Scan;
    }
}


NAT list_DeleteWithElementCount(LIST List, void (*ElementDelete)())
/**************************************************************
  INPUT:   A List and a delete function for the elements.
  RETURNS: The number of deleted elements.
  EFFECT:  The List and all its elements are deleted.
***************************************************************/
{
  int  Result;
  LIST Scan;

  Result = 0;

  while (!list_Empty(List)) {
    Scan = list_Cdr(List);
    ElementDelete(list_Car(List));
    list_Free(List);
    List = Scan;
    Result++;
  }

  return Result;
}


LIST list_DeleteElement(LIST List, POINTER Element, BOOL (*Test)())
/**************************************************************
  INPUT:   A list, an element pointer, an equality test for elements
  RETURNS: The list where Element is deleted from List with respect to Test.
  EFFECTS: If List contains Element with respect to EqualityTest,
           Element is deleted from List
  CAUTION: Destructive. Be careful, the first element of a list cannot
           be changed destructively by call by reference.
***************************************************************/
{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Test(Element, list_Car(List))) {
    Scan1 = list_Cdr(List);
    list_Free(List);
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Test(Element, list_Car(Scan1))) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}

LIST list_DeleteElementIf(LIST List, BOOL (*Test)(), void (*Delete)())
/**************************************************************
  INPUT:   A list, a test for elements and a delete function for elements.
  RETURNS: The list where an element is deleted if <Test> on it succeeds.
           The element is deleted with <Delete>.
  CAUTION: Destructive. Be careful, the first element of a list cannot
           be changed destructively by call by reference.
***************************************************************/
{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Test(list_Car(List))) {
    Scan1 = list_Cdr(List);
    Delete(list_Car(List));
    list_Free(List);
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Test(list_Car(Scan1))) {
      Delete(list_Car(Scan1));
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}


LIST list_DeleteElementFree(LIST List, POINTER Element, BOOL (*Test)(),
			    void (*Free)())
/**************************************************************
  INPUT:   A list, an element pointer, an equality test for elements
           and a free function for elements
  RETURNS: The list where Element is deleted from List with respect to Test.
  EFFECTS: If List contains Element with respect to EqualityTest,
           Element is deleted from List and freed
  CAUTION: Destructive. Be careful, the first element of a list cannot
           be changed destructively by call by reference.
***************************************************************/
{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Test(Element, list_Car(List))) {
    Scan1 = list_Cdr(List);
    Free(list_Car(List));
    list_Free(List);
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Test(Element, list_Car(Scan1))) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      Free(list_Car(Scan1));
      list_Free(Scan1);
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}


LIST list_PointerDeleteElement(LIST List, POINTER Element)
/**************************************************************
  INPUT:   A list and an element pointer
  RETURNS: The list where Element is deleted from List with respect to
           pointer equality.
  EFFECTS: If List contains Element with respect to pointer equality,
           Element is deleted from List
  CAUTION: Destructive. Be careful, the first element of a list cannot
           be changed destructively by call by reference.
***************************************************************/
{
  LIST   Scan1,Scan2;

  while(!list_Empty(List) && Element == list_Car(List)) {
    Scan1 = list_Cdr(List);
    memory_Free(List, sizeof(LIST_NODE));
    List = Scan1;
  }

  if (list_Empty(List)) return(list_Nil());
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Element == list_Car(Scan1)) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}


LIST list_PointerDeleteOneElement(LIST List, POINTER Element)
/**************************************************************
  INPUT:   A list and an element pointer
  RETURNS: The list where one occurrence of Element is deleted from List 
           with respect to pointer equality.
  EFFECTS: If List contains Element with respect to pointer equality,
           Element is deleted from List
  CAUTION: Destructive. Be careful, the first element of a list cannot
           be changed destructively by call by reference.
***************************************************************/
{
  LIST   Scan1,Scan2;

  if (list_Empty(List))
    return(List);
  else
    if (Element == list_Car(List)) {
      Scan1 = list_Cdr(List);
      memory_Free(List, sizeof(LIST_NODE));
      return(Scan1);
    }
  
  Scan2 = List;
  Scan1 = list_Cdr(List);

  while(!list_Empty(Scan1)) {
    if (Element == list_Car(Scan1)) {
      list_Rplacd(Scan2, list_Cdr(Scan1));
      memory_Free(Scan1, sizeof(LIST_NODE));
      Scan1 = list_Cdr(Scan2);
      return(List);
    }
    else {
      Scan2 = Scan1;
      Scan1 = list_Cdr(Scan1);
    }
  }
  return(List);     
}


BOOL list_DeleteFromList(LIST* List, POINTER Element)
/**************************************************************
  INPUT:   A list and an element pointer
  RETURNS: TRUE, if Element was deleted; FALSE, otherwise.
  EFFECTS: If List contains Element with respect to pointer equality,
           all occurrences of Element are deleted from List.
  CAUTION: Destructive. Be careful, the first element of a list cannot
           be changed destructively by call by reference.
***************************************************************/
{
  BOOL Found;
  LIST Scan1;

  Found = FALSE;

  while(list_Exist(*List) && Element == list_Car(*List)) {
    Scan1 = list_Cdr(*List);
    memory_Free(*List, sizeof(LIST_NODE));
    *List = Scan1;
    Found = TRUE;
  }

  if (list_Exist(*List)) {
    LIST Scan2;

    Scan2 = *List;
    Scan1 = list_Cdr(*List);

    while(list_Exist(Scan1)) {
      if (Element == list_Car(Scan1)) {
	list_Rplacd(Scan2, list_Cdr(Scan1));
	memory_Free(Scan1, sizeof(LIST_NODE));
	Scan1 = list_Cdr(Scan2);
	Found = TRUE;
      } else {
	Scan2 = Scan1;
	Scan1 = list_Cdr(Scan1);
      }
    }
  }

  return Found;
}


BOOL list_DeleteOneFromList(LIST* List, POINTER Element)
/**************************************************************
  INPUT:   A list and an element pointer
  RETURNS: TRUE, if Element was deleted; FALSE, otherwise.
  EFFECTS: If List contains Element with respect to pointer equality,
           the first occurrence of Element is deleted from List.
  CAUTION: Destructive. Be careful, the first element of a list cannot
           be changed destructively by call by reference.
***************************************************************/
{
  if (list_Exist(*List)) {
    LIST Scan1;

    if (Element == list_Car(*List)) {
      Scan1 = list_Cdr(*List);
      memory_Free(*List, sizeof(LIST_NODE));
      *List = Scan1;
      return TRUE;
    } else {
      LIST Scan2;

      for (Scan2 = *List, Scan1 = list_Cdr(*List);
	   list_Exist(Scan1);) {
	if (Element == list_Car(Scan1)) {
	  list_Rplacd(Scan2, list_Cdr(Scan1));
	  memory_Free(Scan1, sizeof(LIST_NODE));
	  Scan1 = list_Cdr(Scan2);
	  return TRUE;
	} else {
	  Scan2 = Scan1;
	  Scan1 = list_Cdr(Scan1);
	}
      }
    }
  }
  return FALSE;
}


LIST list_DeleteDuplicates(LIST List, BOOL (*Test)())
/**************************************************************
  INPUT:   A list, an equality test for elements
  RETURNS: The list where multiple occurrances are deleted.
  CAUTION: Destructive.
***************************************************************/
{
  LIST Scan;
  
  Scan = List;

  while (!list_Empty(Scan)) {
    list_Rplacd(Scan,
		 list_DeleteElement(list_Cdr(Scan), list_Car(Scan), Test));
    Scan = list_Cdr(Scan);
  }
  return List;
}


LIST list_DeleteDuplicatesFree(LIST List, BOOL (*Test)(), void (*Free)())
/**************************************************************
  INPUT:   A list, an equality test for elements, and a free for elements
  RETURNS: The list where multiple occurrances are deleted.
  CAUTION: Destructive and frees all duplicates.
***************************************************************/
{
  LIST Scan;
  
  Scan = List;

  while (!list_Empty(Scan)) {
    list_Rplacd(Scan, list_DeleteElementFree(list_Cdr(Scan), list_Car(Scan), Test, Free));
    Scan = list_Cdr(Scan);
  }
  return List;
}


LIST list_PointerDeleteDuplicates(LIST List)
/**************************************************************
  INPUT:   A list
  RETURNS: The list where multiple occurrances are deleted.
  CAUTION: Destructive.
***************************************************************/
{
  LIST Scan;
  
  Scan = List;

  while (!list_Empty(Scan)) {
    list_Rplacd(Scan, list_PointerDeleteElement(list_Cdr(Scan),
						list_Car(Scan)));
    Scan = list_Cdr(Scan);
  }
  return List;
}


LIST list_NPointerUnion(LIST List1, LIST List2)
/**************************************************************
  INPUT:   Two lists.
  RETURNS: Regarding both lists as sets, the union of the sets.
  CAUTION: Destructive.
***************************************************************/
{
  return(list_PointerDeleteDuplicates(list_Nconc(List1,List2)));
}


LIST list_NUnion(LIST List1, LIST List2, BOOL (*Test)())
/**************************************************************
  INPUT:   Two lists and an equality test for the elements.
  RETURNS: Regarding both lists as sets, the union of the sets.
  CAUTION: Destructive.
***************************************************************/
{
  return(list_DeleteDuplicates(list_Nconc(List1,List2), Test));
}


LIST list_NListTimes(LIST List1, LIST List2)
/**************************************************************
  INPUT:   Two lists of lists.
  RETURNS: The list of combinations of element lists.
  CAUTION: Destroys List1 and List2.
***************************************************************/
{
  LIST Result, Scan1, Scan2;

  Result = list_Nil();

  for(Scan1=List1; !list_Empty(Scan1); Scan1=list_Cdr(Scan1))
    for(Scan2=List2; !list_Empty(Scan2); Scan2=list_Cdr(Scan2))
      Result = list_Cons(list_Append(((LIST)list_Car(Scan1)),
				     list_Copy((LIST)list_Car(Scan2))),
			  Result);

  list_Apply(list_Delete, List1);
  list_Apply(list_Delete, List2);
  list_Delete(List1);
  list_Delete(List2);

  return(Result);
}


LIST list_NIntersect(LIST List1, LIST List2, BOOL (*Test)())
/**************************************************************
  INPUT:   Two lists and an equality test for the elements.
  RETURNS: Regarding both lists as sets, the intersection of the sets.
  CAUTION: Destructive on List1
***************************************************************/
{
  LIST Scan1, Scan2;
  
  while (!list_Empty(List1) && !list_Member(List2, list_Car(List1), Test)) {
    Scan1 = list_Cdr(List1);
    list_Free(List1);
    List1 = Scan1;
  }

  if (list_Empty(List1))
    return(List1);

  Scan1 = List1;
  Scan2 = list_Cdr(List1);

  while(!list_Empty(Scan2)) {
    if (list_Member(List2, list_Car(Scan2), Test)) {
      Scan2 = list_Cdr(Scan2);
      Scan1 = list_Cdr(Scan1);
    }
    else {
      list_Rplacd(Scan1, list_Cdr(Scan2));
      list_Free(Scan2);
      Scan2 = list_Cdr(Scan1);
    }
  }
  return(List1);
}


LIST list_NPointerIntersect(LIST List1, LIST List2)
/**************************************************************
  INPUT:   Two lists.
  RETURNS: Regarding both lists as sets, the intersection of the sets.
  CAUTION: Destructive on List1
***************************************************************/
{
  LIST Scan1, Scan2;
  
  while (!list_Empty(List1) && !list_PointerMember(List2, list_Car(List1))) {
    Scan1 = list_Cdr(List1);
    list_Free(List1);
    List1 = Scan1;
  }

  if (list_Empty(List1))
    return(List1);

  Scan1 = List1;
  Scan2 = list_Cdr(List1);

  while(!list_Empty(Scan2)) {
    if (list_PointerMember(List2, list_Car(Scan2))) {
      Scan2 = list_Cdr(Scan2);
      Scan1 = list_Cdr(Scan1);
    }
    else {
      list_Rplacd(Scan1, list_Cdr(Scan2));
      list_Free(Scan2);
      Scan2 = list_Cdr(Scan1);
    }
  }
  return(List1);
}

void list_NInsert(LIST List1, LIST List2)
/**************************************************************
  INPUT:   Two lists where <List1> must not be empty.
  EFFECT:  <List2> is destructively concatenated after
           the first element of <List1>
  RETURNS: void
  CAUTION: Destructive on List1 and List2
***************************************************************/
{
  LIST Help;

#if DEBUG
  if (list_Empty(List1)) {
    printf("\n\n Empty List argument in list_NInsert\n");
    misc_Error();
  }
#endif

  Help = list_Cdr(List1);
  list_Rplacd(List1,List2);
  List2 = List1;

  while (!list_Empty(list_Cdr(List2)))
    List2 = list_Cdr(List2);

  list_Rplacd(List2,Help);
}

      
BOOL list_HasIntersection(LIST List1, LIST List2)
/**************************************************************
  INPUT:   Two lists .
  RETURNS: TRUE iff List1 and List2 have a comon element.
***************************************************************/
{ 
  LIST Scan;

  for(Scan=List1; !list_Empty(Scan); Scan=list_Cdr(Scan))
    if (list_PointerMember(List2,list_Car(Scan)))
      return TRUE;

  return FALSE;
}


LIST list_NPointerDifference(LIST List1, LIST List2)
/**************************************************************
  INPUT:   Two lists.
  RETURNS: The list List1-List2.
  CAUTION: Destructive on List1
***************************************************************/
{ 
  LIST Scan;

  for(Scan=List2; !list_Empty(Scan); Scan=list_Cdr(Scan))
    List1 = list_PointerDeleteElement(List1, list_Car(Scan));

  return(List1);
}


LIST list_NDifference(LIST List1, LIST List2, BOOL (*Test)())
/**************************************************************
  INPUT:   Two lists and an equality test for elements.
  RETURNS: The list List1-List2 wrt. <Test>.
  CAUTION: Destructive on List1
***************************************************************/
{ 
  LIST Scan;

  for(Scan=List2; !list_Empty(Scan); Scan=list_Cdr(Scan))
    List1 = list_DeleteElement(List1, list_Car(Scan), Test);

  return(List1);
}


BOOL list_Member(LIST List, POINTER Element, BOOL (*Test)())
/**************************************************************
  INPUT:   A list and an element pointer and an equality test for two elements.
  RETURNS: TRUE iff Element is in List with respect to Test
***************************************************************/
{
  while (!list_Empty(List)) {
    if (Test(Element, list_Car(List)))
      return TRUE;
    List = list_Cdr(List);
  }
  
  return FALSE;
}


BOOL list_PointerMember(LIST List, POINTER Element)
/**************************************************************
  INPUT:   A list and an element pointer.
  RETURNS: TRUE iff Element is in List with respect to pointer equality.
***************************************************************/
{
  while (!list_Empty(List)) {
    if (Element== list_Car(List))
      return TRUE;
    List = list_Cdr(List);
  }
  
  return FALSE;
}
    
BOOL list_PointerReplaceMember(LIST List, POINTER Old, POINTER New)
/**************************************************************
  INPUT:   A list, a pointer to an old element, a pointer to a new element
  EFFECT: The first occurrence of new in the list is replcaed by old
  RETURNS: TRUE iff New was replcaed
***************************************************************/
{
    while (!list_Empty(List)) {
	if (Old == list_Car(List)) {
	    list_Rplaca(List, New);
	    return TRUE;
	}
	List = list_Cdr(List);
    }
    return FALSE;
}   
  
NAT list_Length(LIST List)
/**************************************************************
  INPUT:   A List.
  RETURNS: The number of elements..
  EFFECT:  None.
***************************************************************/
{
  NAT Result;

  Result = 0;

  while (!list_Empty(List)) {
    Result++;
    List = list_Cdr(List);
  }

  return(Result);
}


NAT list_Bytes(LIST List)
/**************************************************************
  INPUT:   A List.
  RETURNS: The number of Bytes occupied by the list structure of <List>
  EFFECT:  None.
***************************************************************/
{
  return(list_Length(List)*sizeof(LIST_NODE));
}
