/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                   RENAMING                             * */
/* *                                                        * */
/* *  $Module:   RENAMING                                   * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.52 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/03/06 10:29:09 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: renaming.c,v $ */

#include "cnf.h"
#include "term.h"
#include "renaming.h"

NAT ren_STAMPID;

static BOOL ren_AFactorOk(TERM,TERM);
static BOOL ren_BFactorOk(TERM,TERM);
static BOOL ren_AExtraFactorOk(TERM,TERM);
static BOOL ren_BExtraFactorOk(TERM,TERM);
static TERM ren_FormulaRename(TERM Term, LIST Termlist, LIST *SkolemSymbols);
static LIST ren_GetRenamings(TERM Term1, int Pol1, TERM Term2, int Pol2);
static BOOL ren_HasBenefit(TERM Term1, int Pol1, TERM Term2, int Pol2);
static BOOL ren_PFactorOk(TERM);
static BOOL ren_PExtraFactorOk(TERM);
static int  ren_Polarity(TERM Term);
static BOOL ren_NotPFactorOk(TERM);
static BOOL ren_NotPExtraFactorOk(TERM);
static void ren_ResetTermStamp(TERM Term);

void ren_Init()
/**********************************************************
  INPUT:   None.
  RETURNS: void.
  EFFECT:  Initializes the renaming module, in particular
           the stamp id used in this module.
***********************************************************/
{
  ren_STAMPID = term_GetStampID();
}

static void ren_ResetTermStamp(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: void.
  EFFECT:  The Term stamp of term as well as the stamps of 
           all its subterms (up to atom level) are reset.
***********************************************************/
{
  SYMBOL Top;
 
  term_ResetTermStamp(Term);
  Top = term_TopSymbol(Term);

  if (!symbol_IsPredicate(Top))
    if (fol_IsQuantifier(Top))
      ren_ResetTermStamp(term_SecondArgument(Term));
    else {
      LIST Scan;
      for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	ren_ResetTermStamp(list_Car(Scan)); 
    }
}

static BOOL ren_PExtraFactorOk(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: TRUE if transforming the term <Term> in positive polarity context
           results in more than two clauses.
***********************************************************/
{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) /* Formula already renamed */
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv()))
    if (!fol_IsLiteral(term_FirstArgument(Term)) && !fol_IsLiteral(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
	  
  if (symbol_Equal(Top,fol_And()))
    if (list_Length(term_ArgumentList(Term))>2 ||
	ren_PFactorOk(term_FirstArgument(Term)) ||
	ren_PFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;

  if (symbol_Equal(Top,fol_Implies()))
    return (ren_NotPExtraFactorOk(term_FirstArgument(Term)) || 
	    ren_PExtraFactorOk(term_SecondArgument(Term)));

  if (symbol_Equal(Top,fol_Not()))
    return ren_NotPExtraFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_PExtraFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_PExtraFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; /* Either <Term> is a trivial disjunction or an atom */
}

static BOOL ren_PFactorOk(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: TRUE if transforming the term <Term> in positive polarity context
           results in more than one clause.
***********************************************************/
{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) /* Formula already renamed */
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv()) || symbol_Equal(Top,fol_And()))
    return TRUE;

  if (symbol_Equal(Top,fol_Implies()))
    return (ren_NotPFactorOk(term_FirstArgument(Term)) || ren_PFactorOk(term_SecondArgument(Term)));

  if (symbol_Equal(Top,fol_Not()))
    return ren_NotPFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_PFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_PFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; /* Either <Term> is a trivial disjunction or an atom */
}

static BOOL ren_NotPExtraFactorOk(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: TRUE if transforming the term <Term> in negative polarity context
           results in more than two clauses.
***********************************************************/
{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) /* Formula already renamed */
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv()))
    if (!fol_IsLiteral(term_FirstArgument(Term)) && !fol_IsLiteral(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;
	  
  if (symbol_Equal(Top,fol_Or()))
    if (list_Length(term_ArgumentList(Term))>2 ||
	ren_NotPFactorOk(term_FirstArgument(Term)) ||
	ren_NotPFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;

  if (symbol_Equal(Top,fol_Implies()))
    if (ren_PFactorOk(term_FirstArgument(Term)) ||
	ren_NotPFactorOk(term_SecondArgument(Term)))
      return TRUE;
    else
      return FALSE;

  if (symbol_Equal(Top,fol_Not()))
    return ren_PExtraFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_NotPExtraFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_NotPExtraFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; /* Either <Term> is a trivial conjunction or an atom */
}

static BOOL ren_NotPFactorOk(TERM Term)
/**********************************************************
  INPUT:   A term.
  RETURNS: TRUE if transforming the term <Term> in negative polarity context
           results in more than one clause.
***********************************************************/
{ 
  SYMBOL Top;

  if (term_HasTermStamp(Term)) /* Formula already renamed */
    return FALSE;

  Top = term_TopSymbol(Term);
  
  if (symbol_Equal(Top,fol_Equiv()) || symbol_Equal(Top,fol_Or()) || symbol_Equal(Top,fol_Implies()))
    return TRUE;

  if (symbol_Equal(Top,fol_Not()))
    return ren_PFactorOk(term_FirstArgument(Term));

  if (fol_IsQuantifier(Top))
    return ren_NotPFactorOk(term_SecondArgument(Term));

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      if (ren_NotPFactorOk(list_Car(Scan)))
	return TRUE;
  }

  return FALSE; /* Either <Term> is a trivial conjunction or an atom */
}


static BOOL ren_AFactorOk(TERM Term1, TERM Term2)
/**********************************************************
  INPUT:   Two terms where <Term1> is a superterm of <Term2>
  RETURNS: TRUE if the A-Factor of Term2 in Term1 is larger than one.
***********************************************************/
{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_And()) || fol_IsQuantifier(Top))
    return ren_AFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_BFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    TERM Sub;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2 && ren_PFactorOk(Sub))
	return TRUE;
    }
    return ren_AFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies()))
    if (Term2 == term_FirstArgument(Super))
      return ren_BFactorOk(Term1, Super);
    else
      return (ren_NotPFactorOk(term_FirstArgument(Super)) || ren_AFactorOk(Term1, Super));

  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Pol == 0)
      return TRUE;
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 1)
      return (ren_NotPFactorOk(Term2) || ren_AFactorOk(Term1,Super));
    else
      return (ren_PFactorOk(Term2) ||  ren_BFactorOk(Term1,Super));
  }
  
  misc_DumpCoreOut("Unexpected case in ren_AFactorOk");
  return FALSE; 
}

static BOOL ren_AExtraFactorOk(TERM Term1, TERM Term2)
/**********************************************************
  INPUT:   Two terms where <Term1> is a superterm of <Term2>
  RETURNS: TRUE if the A-Factor of Term2 in Term1 is larger than two.
***********************************************************/
{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_And()) || fol_IsQuantifier(Top))
    return ren_AExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_BExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Or())) {
    LIST Scan;
    TERM Sub;
    BOOL Extra;
    Extra = FALSE;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2)
	if (ren_PFactorOk(Sub))
	  if (Extra || ren_PExtraFactorOk(Sub))
	    return TRUE;
	  else
	    Extra = TRUE;
    }
    return ren_AExtraFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies()))
    if (Term2 == term_FirstArgument(Super))
      return ren_BExtraFactorOk(Term1, Super);
    else
      return (ren_NotPExtraFactorOk(term_FirstArgument(Super)) || ren_AExtraFactorOk(Term1, Super));

  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 0)
      return (ren_PFactorOk(Term2) ||  ren_NotPFactorOk(Term2) ||
	      ren_AFactorOk(Term1,Super) ||  ren_BFactorOk(Term1,Super));

    if (Pol == 1)
      return (ren_NotPExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Super));
    else
      return (ren_PExtraFactorOk(Term2) ||  ren_BExtraFactorOk(Term1,Super));
  }
  
  
  misc_DumpCoreOut("Unexpected case in ren_AExtraFactorOk");
  return FALSE; 
}

static BOOL ren_BFactorOk(TERM Term1, TERM Term2)
/**********************************************************
  INPUT:   Two terms where <Term1> is a superterm of <Term2>
  RETURNS: TRUE if the B-Factor of Term2 in Term1 is larger than one.
***********************************************************/
{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
    return ren_BFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_AFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    TERM Sub;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2 && ren_NotPFactorOk(Sub))
	return TRUE;
    }
    return ren_BFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies()))
    if (Term2 == term_FirstArgument(Super))
      return (ren_PFactorOk(term_SecondArgument(Super)) || ren_AFactorOk(Term1, Super));
    else
      return ren_BFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Pol == 0)
      return TRUE;
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 1)
      return (ren_PFactorOk(Term2) || ren_AFactorOk(Term1,Super));
    else
      return (ren_NotPFactorOk(Term2) || ren_BFactorOk(Term1,Super));
  }
  
  misc_DumpCoreOut("Unexpected case in ren_BFactorOk");
  return FALSE; 
}

static BOOL ren_BExtraFactorOk(TERM Term1, TERM Term2)
/**********************************************************
  INPUT:   Two terms where <Term1> is a superterm of <Term2>
  RETURNS: TRUE if the B-Factor of Term2 in Term1 is larger than two.
***********************************************************/
{ 
  SYMBOL Top;
  TERM   Super;

  if (Term1 == Term2)
    return FALSE;

  Super = term_Superterm(Term2);
  Top   = term_TopSymbol(Super);    
  
  if (symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
    return ren_BExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Not()))
    return ren_AExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_And())) {
    LIST Scan;
    TERM Sub;
    BOOL Extra;
    Extra = FALSE;
    for (Scan=term_ArgumentList(Super);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Sub = (TERM)list_Car(Scan);
      if (Sub != Term2)
	if (ren_NotPFactorOk(Sub))
	  if (Extra || ren_NotPExtraFactorOk(Sub))
	    return TRUE;
	  else
	    Extra = TRUE;
    }
    return ren_BExtraFactorOk(Term1, Super);
  }

  if (symbol_Equal(Top,fol_Implies()))
    if (Term2 == term_FirstArgument(Super))
      return (ren_PExtraFactorOk(term_SecondArgument(Super)) || ren_AExtraFactorOk(Term1, Super));
    else
      return ren_BExtraFactorOk(Term1, Super);

  if (symbol_Equal(Top,fol_Equiv())) {
    int Pol;
    Pol = ren_Polarity(Super);
    if (Term2 == term_FirstArgument(Super))
      Term2 = term_SecondArgument(Super);
    else
      Term2 = term_FirstArgument(Super);

    if (Pol == 0)
      return (ren_PFactorOk(Term2) ||  ren_NotPFactorOk(Term2) ||
	      ren_AFactorOk(Term1,Super) ||  ren_BFactorOk(Term1,Super));

    if (Pol == 1)
      return (ren_PExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Super));
    else
      return (ren_NotPExtraFactorOk(Term2) ||  ren_BExtraFactorOk(Term1,Super));
    }
  
  misc_DumpCoreOut("Unexpected case in ren_BExtraFactorOk");
  return FALSE; 
}

static BOOL ren_HasBenefit(TERM Term1, int Pol1, TERM Term2, int Pol2)
/**********************************************************
  INPUT:   Two terms and the polarity of the terms in the overall formula.
  RETURNS: TRUE if renaming <Term1> in <Term2> results in a positive benefit.
  CAUTION: It is assumed that all superterms are set !!!!
***********************************************************/
{ 
  BOOL  PFacOk, NotPFacOk, AFacOk, BFacOk;
  
  switch (Pol1*Pol2) {
    
  case 0:
    PFacOk    = ren_PFactorOk(Term2);
    NotPFacOk = ren_NotPFactorOk(Term2);
    AFacOk    = ren_AFactorOk(Term1,Term2);
    BFacOk    = ren_BFactorOk(Term1,Term2);

    return ((AFacOk && BFacOk && PFacOk && NotPFacOk) ||
	    (AFacOk && PFacOk && (ren_PExtraFactorOk(Term2) || ren_AExtraFactorOk(Term1,Term2))) ||
	    (BFacOk && NotPFacOk && (ren_NotPExtraFactorOk(Term2) || ren_BExtraFactorOk(Term1,Term2))));
    break;

  case 1:
    return (ren_PFactorOk(Term2) && ren_AFactorOk(Term1,Term2));
    break;

  case -1:
    return (ren_NotPFactorOk(Term2) && ren_BFactorOk(Term1,Term2));
    break;
  }
  
  misc_DumpCoreOut("Unexpected polarity in ren_HasBenefit");
  return FALSE;  
}


static LIST ren_GetRenamings(TERM Term1, int Pol1, TERM Term2, int Pol2) 
/**********************************************************
  INPUT:   Two terms and the polarity of the terms in the overall formula.
  RETURNS: The list of subterms below <Term2> that have a positive renaming
           benefit.
  EFFECT:  All renamed formulae are stamped.
  CAUTION: None.
***********************************************************/
{ 
  SYMBOL Top;
  LIST   Result,Scan;

  Result = list_Nil();

  /*printf("\n %d: ",Pol1); term_Print(Term1); printf("\n %d: ",Pol2); term_Print(Term2);*/

  if (fol_IsLiteral(Term2))
    return Result;

  Top = term_TopSymbol(Term2);

  if (symbol_Equal(Top,fol_Not()))
    return ren_GetRenamings(Term1, Pol1, term_FirstArgument(Term2), Pol2*-1);
  
  if (term_Superterm(Term2) && !fol_IsQuantifier(term_TopSymbol(term_Superterm(Term2))) &&
      ren_HasBenefit(Term1, Pol1, Term2, Pol2)) {
    Result = list_Cons(Term2,Result);
    term_SetTermStamp(Term2);
    Term1 = Term2;
    Pol1  = Pol2;
  }

  if (fol_IsQuantifier(Top))
    return list_Nconc(Result,ren_GetRenamings(Term1, Pol1, term_SecondArgument(Term2), Pol2));

  if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()))
    for (Scan=term_ArgumentList(Term2);!list_Empty(Scan);Scan=list_Cdr(Scan))
      Result = list_Nconc(Result,ren_GetRenamings(Term1,Pol1,list_Car(Scan),Pol2));
  else
    if (symbol_Equal(Top,fol_Implies()))
      Result = list_Nconc(Result,
        list_Nconc(ren_GetRenamings(Term1,Pol1,term_FirstArgument(Term2),Pol2*-1),
		   ren_GetRenamings(Term1,Pol1,term_SecondArgument(Term2),Pol2)));
    else
      if (symbol_Equal(Top,fol_Equiv()))
	Result = list_Nconc(Result,
			    list_Nconc(ren_GetRenamings(Term1,Pol1,term_FirstArgument(Term2),0),
				       ren_GetRenamings(Term1,Pol1,term_SecondArgument(Term2),0)));
      else
	misc_DumpCoreOut("Unexpected case in ren_GetRenamings");

  return Result;
}

static int ren_Polarity(TERM Term)
/**********************************************************
  INPUT:   A term where the existence of superterms is assumed!.
  RETURNS: The polarity of Term with respect to its superterms.
***********************************************************/
{
  TERM SuperTerm;

  SuperTerm = term_Superterm(Term);

  if (SuperTerm) {
    SYMBOL Top;
    Top = term_TopSymbol(SuperTerm);
    if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()) || fol_IsQuantifier(Top))
      return ren_Polarity(SuperTerm);
    if (symbol_Equal(Top,fol_Not()))
      return (-1*ren_Polarity(SuperTerm));
    if (symbol_Equal(Top,fol_Equiv()))
      return 0;
    if (symbol_Equal(Top,fol_Implies()))
      if (Term == term_FirstArgument(SuperTerm))
	return (-1*ren_Polarity(SuperTerm));
      else
	return ren_Polarity(SuperTerm);
    misc_DumpCoreOut("Unexpected case in ren_Polarity");
  }

  return 1;
}


static TERM ren_FormulaRename(TERM Term, LIST Termlist, LIST *SkolemSymbols) 
/**********************************************************
  INPUT:   A term, a list of subterms of term and a list of symbols.
  RETURNS: The renamed formula with respect to termlist and all newly introduced
           Skolem symbols for renamings are added to <SkolemSymbols>
  CAUTION: The term is destructively changed.
***********************************************************/
{
  TERM   Result,ActTerm,DefTerm,Superterm,NewTerm;
  LIST   Scan,FreeVariables,Arg;
  SYMBOL ActSymbol;
  int Length;
 
  DefTerm = (TERM)NULL;

  if (!list_Empty(Termlist))
    Result = term_Create(fol_AND,list_List(Term));
  else
    return Term;

  for(Scan=Termlist;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    ActTerm        = (TERM)list_Car(Scan);
    FreeVariables  = fol_FreeVariables(ActTerm);
    Length         = list_Length(FreeVariables);
    Superterm      = term_Superterm(ActTerm);
    ActSymbol      = symbol_CreateSkolemPredicate(Length);
    *SkolemSymbols = list_Cons((POINTER)ActSymbol,*SkolemSymbols);
    NewTerm        = term_Create(ActSymbol, list_CopyWithElement(FreeVariables,(POINTER (*)())term_Copy));

    switch (ren_Polarity(ActTerm)) {
      
    case 0:
      DefTerm = term_Create(fol_Equiv(),list_Cons(term_Copy(NewTerm),list_List(ActTerm)));
      term_RplacSuperterm(term_FirstArgument(DefTerm),DefTerm);
      term_RplacSuperterm(term_SecondArgument(DefTerm),DefTerm);
      if (!list_Empty(FreeVariables))
	DefTerm = fol_CreateQuantifier(fol_All(),list_NMapCar(FreeVariables,(POINTER (*)())term_Copy),
				       list_List(DefTerm));
      break;

    case 1:
      DefTerm = term_Create(fol_Implies(),list_Cons(term_Copy(NewTerm),list_List(ActTerm)));
      term_RplacSuperterm(term_FirstArgument(DefTerm),DefTerm);
      term_RplacSuperterm(term_SecondArgument(DefTerm),DefTerm);
      if (!list_Empty(FreeVariables))
	DefTerm = fol_CreateQuantifier(fol_All(),list_NMapCar(FreeVariables,(POINTER (*)())term_Copy),
				       list_List(DefTerm));
      break;

    case -1:
      DefTerm = term_Create(fol_Implies(),list_Cons(ActTerm,list_List(term_Copy(NewTerm))));
      term_RplacSuperterm(term_FirstArgument(DefTerm),DefTerm);
      term_RplacSuperterm(term_SecondArgument(DefTerm),DefTerm);
      if (!list_Empty(FreeVariables))
	DefTerm = fol_CreateQuantifier(fol_All(),list_NMapCar(FreeVariables,(POINTER (*)())term_Copy),
				       list_List(DefTerm));
      break;
    }

    /*printf("\nDef:\t");term_Print(DefTerm);*/

    term_RplacArgumentList(Result,list_Nconc(term_ArgumentList(Result),list_List(DefTerm)));

    for(Arg=term_ArgumentList(Superterm);
	!list_Empty(Arg) && !((TERM)list_Car(Arg) == ActTerm);
	Arg=list_Cdr(Arg));                                      /* Search Position in Argumentlist */

    list_Rplaca(Arg,NewTerm);
  }
  
  /*printf("\nRes:\t");term_Print(Result);*/

  return Result;			   
}


TERM ren_Rename(TERM Term, LIST *SkolemSymbols)
/**********************************************************
  INPUT:   A term and a pointer to a list of Skolem symbols.
  RETURNS: The possibly changed Term where subformulae are renamed
           if this results in a smaller clause normal form, with
	   respect to the number of clauses. The newly introduced
	   Skolem predicates are added to <SkolemSymbols>.
  CAUTION: Destructive.
***********************************************************/
{
  LIST Renamings;

  Renamings = list_Nil();

  if (term_StampOverflow(ren_STAMPID))
    ren_ResetTermStamp(Term);

  term_AddFatherLinks(Term);

  term_StartStamp();

  Renamings = ren_GetRenamings(Term,1,Term,1);

  /*printf("\n\nTerm: ");fol_PrintDFG(Term);
  printf("\n\nRenamings:\t");
  for(Scan=Renamings;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    fol_PrintDFG(list_Car(Scan));
    if (!list_Empty(list_Cdr(Scan)))
      printf("\n\t");
    else
      printf("\n");
  }*/


  Term = ren_FormulaRename(Term,Renamings,SkolemSymbols);
    
  list_Delete(Renamings);

  term_StopStamp();

  return Term;
}
