/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                INFERENCE RULES                         * */
/* *                                                        * */
/* *  $Module:   INFRULES                                   * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.62 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/02 14:25:21 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: rules-inf.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "rules-inf.h"

/**************************************************************/
/* Functions                                                  */
/**************************************************************/        


static BOOL inf_InferenceLitsMax(CLAUSE Clause, int i, SUBST Subst,
				 CLAUSE PartnerClause, int j,
				 SUBST PartnerSubst)
/**************************************************************
  INPUT:   The parents of a resolution inference, the respective
           literal indices and substitutions
  RETURNS: TRUE iff the literals are still maximal in the instantiated clause
***************************************************************/
{
  TERM       Max, Lit;
  kbo_RESULT Compare;
  int        k,l;

  l   = clause_Length(Clause);

  if (!clause_GetFlag(Clause,CLAUSESELECT) && l != 1 ) {
    /* The check doesn't make sense for unit, or selected clauses */
    Max = subst_Apply(Subst,term_Copy(clause_GetLiteralTerm(Clause,i)));
    for(k = clause_FirstAntecedentLitIndex(Clause); k < l; k++)
      if (k != i && clause_LiteralIsMaximal(clause_GetLiteral(Clause, k))) {
	Lit     = subst_Apply(Subst,term_Copy(clause_GetLiteralTerm(Clause,k)));
	Compare = kbo_LiteralCompare(Max,clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,k)),
				     Lit,clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,i)),
				     TRUE);
	if (Compare == kbo_SmallerThan() ||
	    (Compare == kbo_Equal() && i > clause_LastAntecedentLitIndex(Clause))) {
	  /*fputs("\nLitMax: ",stdout); clause_Print(Clause);
	    printf(", %d, ",i);term_Print(Max);putchar(' ');
	    subst_Print(Subst);*/
	  term_Delete(Max);
	  term_Delete(Lit);
	  return FALSE;
	}
	term_Delete(Lit);
      }
    term_Delete(Max);
  }

  l   = clause_Length(PartnerClause);

  if (!clause_GetFlag(PartnerClause,CLAUSESELECT) && l != 1 ) {
    /* The check doesn't make sense for unit, or selected clauses */
    Max = subst_Apply(PartnerSubst,term_Copy(clause_GetLiteralTerm(PartnerClause,j)));
    for(k = clause_FirstAntecedentLitIndex(PartnerClause);k<l;k++)
      if (k != j && clause_LiteralIsMaximal(clause_GetLiteral(PartnerClause, k))) {
	Lit = subst_Apply(PartnerSubst,term_Copy(clause_GetLiteralTerm(PartnerClause,k)));
	Compare = kbo_LiteralCompare(Max,clause_LiteralIsOrientedEquality(clause_GetLiteral(PartnerClause,j)),
				     Lit,clause_LiteralIsOrientedEquality(clause_GetLiteral(PartnerClause,k)),TRUE);
	if (Compare == kbo_SmallerThan() ||
	    (Compare == kbo_Equal() && k > clause_LastAntecedentLitIndex(PartnerClause))) {
	  /*fputs("\nLitMax: ",stdout); clause_Print(PartnerClause);
	    printf(", %d, ",j);term_Print(Max);
	    putchar(' ');subst_Print(PartnerSubst);*/
	  term_Delete(Max);
	  term_Delete(Lit);
	  return FALSE;
	}
	term_Delete(Lit);
      }
    term_Delete(Max);
  }
  return TRUE;
}


static BOOL inf_FactorLitMax(CLAUSE Clause, int i, int j, SUBST Subst)
/**************************************************************
  INPUT:   The clause of a factoring inference, the respective
           literal indices and the substitution
  RETURNS: TRUE iff the literals are still maximal in the instantiated clause
***************************************************************/
{
  TERM       Max, Lit;
  kbo_RESULT Compare;
  int        k,l;

  l   = clause_Length(Clause);
  Max = subst_Apply(Subst,term_Copy(clause_GetLiteralTerm(Clause,i)));

  for(k=clause_FirstAntecedentLitIndex(Clause);k<l;k++)
    if (k != i && k != j && clause_LiteralIsMaximal(clause_GetLiteral(Clause, k))) {
      Lit    = subst_Apply(Subst,term_Copy(clause_GetLiteralTerm(Clause,k)));
      Compare = kbo_LiteralCompare(Max,clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,i)),
				   Lit,clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,k)),TRUE);
      if (Compare == kbo_SmallerThan() ||
	  (Compare == kbo_Equal() && k > clause_LastAntecedentLitIndex(Clause))) {
	/*fputs("\nLitMax: ",stdout); clause_Print(Clause);
	  printf(", %d, ",i);term_Print(Max);putchar(' ');subst_Print(Subst);*/
	term_Delete(Max);
	term_Delete(Lit);
	return FALSE;
      }
      term_Delete(Lit);
    }
  term_Delete(Max);

  return TRUE;
}


LIST inf_EqualityResolution(CLAUSE GivenClause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: A list of clauses inferred from the GivenClause by
           EqualityResolution.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{

  LIST Result;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    fputs("\n\nIllegal input in EqualityResolution\n", stderr);
    misc_Error();
  }
#endif

  Result = list_Nil();

  if (!clause_HasEmptyAntecedent(GivenClause) &&
      clause_HasSolvedConstraint(GivenClause)){
    LITERAL ActLit;
    int i,last;
    SYMBOL MinimumVariable;

    last = clause_LastAntecedentLitIndex(GivenClause);

    MinimumVariable = clause_MaxVar(GivenClause);

    for(i = clause_FirstAntecedentLitIndex(GivenClause);
	(i <= last);
	i++){

      ActLit = clause_GetLiteral(GivenClause, i);

      if ((clause_LiteralIsMaximal(ActLit) &&
	   !clause_GetFlag(GivenClause,CLAUSESELECT)) ||
	  clause_LiteralGetFlag(ActLit,LITSELECT)){
	TERM Atom;

	Atom = clause_GetLiteralAtom(GivenClause, i);

	if (fol_IsEquality(Atom)){
	  cont_Check();

	  if(unify_UnifyCom(cont_LeftContext(),
			    term_FirstArgument(Atom), 
			    term_SecondArgument(Atom))){
	    SUBST mgu;
	    CLAUSE NewClause;
	    int j,c,a,s;

	    NewClause = clause_CreateBody(clause_Length(GivenClause) - 1);

	    subst_ExtractUnifierCom(cont_LeftContext(), &mgu);

	    c = clause_LastConstraintLitIndex(GivenClause);
	    clause_SetNumOfConsLits(NewClause,
				    clause_NumOfConsLits(GivenClause));
	    a = clause_LastAntecedentLitIndex(GivenClause);
	    clause_SetNumOfAnteLits(NewClause,
				    (clause_NumOfAnteLits(GivenClause) - 1));
	    s = clause_LastSuccedentLitIndex(GivenClause);
	    clause_SetNumOfSuccLits(NewClause,
				    clause_NumOfSuccLits(GivenClause));

	    for (j = clause_FirstLitIndex(); j <= c; j++)
	      clause_SetLiteral(NewClause, j, 
		clause_LiteralCreate(subst_Apply(mgu, term_Copy(
                  clause_GetLiteralTerm(GivenClause, j))),NewClause));

	    for( ; j <= a; j++)
	      if (j < i)
		/* Just set the literal appropriately! */
		clause_SetLiteral(NewClause, j, 
		  clause_LiteralCreate(subst_Apply(mgu, term_Copy(
		    clause_GetLiteralTerm(GivenClause, j))),NewClause));

	      else if (j > i)
		/* ignore Lit(i), afterwards use shifted index! */
		clause_SetLiteral(NewClause, (j-1), 
		  clause_LiteralCreate(subst_Apply(mgu, term_Copy(
		    clause_GetLiteralTerm(GivenClause, j))),NewClause));

	    for( ; j <= s; j++)
	      /* still use shifted index! */
	      clause_SetLiteral(NewClause, (j-1),
		clause_LiteralCreate(subst_Apply(mgu, term_Copy(
		  clause_GetLiteralTerm(GivenClause, j))),NewClause));

	    clause_SetDataFromFather(NewClause, GivenClause, i);
	    clause_SetFromEqualityResolution(NewClause);
	    /* printf("\n In EqualityResolution entstand aus "
	       "Literal %d die Klausel:  ",i);clause_Print(NewClause); */
	    Result = list_Cons(NewClause, Result);

	    subst_Delete(mgu);
	  }

	  cont_Reset();
	} /*end of if term_IsEquality(Atom). */
      }	/* end of if 'ActLit is maximal'. */
    } /*end of for 'all literals'. */
  }
  return(Result);
}

static CLAUSE inf_ApplyEqualityFactoring(CLAUSE Clause, TERM Left, TERM Right,
					 int i, int j, SUBST Subst)
/**************************************************************
  INPUT:   A clause, two terms, two indeces in the clause and
           a substitution.
  RETURNS: A new clause, where <= Left Right> is added as antecedent atom,
           the <i>th literal is deleted the <j>th kept and
	   <Subst> is applied to a copy of <clause>
***************************************************************/
{
  CLAUSE NewClause;
  TERM   Atom;
  int    k,c,a,s;

  NewClause = clause_CreateBody(clause_Length(Clause));

  c = clause_LastConstraintLitIndex(Clause);
  clause_SetNumOfConsLits(NewClause, clause_NumOfConsLits(Clause));
  a = clause_LastAntecedentLitIndex(Clause);
  clause_SetNumOfAnteLits(NewClause, clause_NumOfAnteLits(Clause) + 1);
  s = clause_LastSuccedentLitIndex(Clause);
  clause_SetNumOfSuccLits(NewClause, clause_NumOfSuccLits(Clause) - 1);

  for (k = clause_FirstLitIndex(); k <= c; k++) {
    clause_SetLiteral(NewClause, k, 
      clause_LiteralCreate(subst_Apply(Subst,
	term_Copy(clause_GetLiteralTerm(Clause, k))),NewClause));
  }

  for( ; k <= a; k++) {
    clause_SetLiteral(NewClause, k, 
      clause_LiteralCreate(subst_Apply(Subst,
        term_Copy(clause_GetLiteralTerm(Clause, k))),NewClause));
  }

  Atom = term_Create(fol_Equality(),
		     list_Cons(term_Copy(Left),list_List(term_Copy(Right))));

  clause_SetLiteral(NewClause, k, clause_LiteralCreate(
    term_Create(fol_Not(), list_List(subst_Apply(Subst,Atom))), NewClause));

  a = 1; /* Shift */

  for ( ; k <= s; k++) {
    if (k == i)
      a = 0;
    else {
      clause_SetLiteral(NewClause, (k + a),
	clause_LiteralCreate(subst_Apply(Subst,
	  term_Copy(clause_GetLiteralTerm(Clause, k))),NewClause));
    }
  }

  clause_SetDataFromFather(NewClause, Clause, i);

  clause_SetFromEqualityFactoring(NewClause);
  clause_AddParentClause(NewClause, clause_Number(Clause));
  clause_AddParentLiteral(NewClause, j);

  return NewClause;
}


static BOOL inf_EqualityFactoringApplicable(CLAUSE Clause, int i, int j,
					    TERM Left, TERM Right, TERM PLeft,
					    TERM PRight, SUBST Subst,
					    BOOL Check, BOOL PCheck)
/**************************************************************
  INPUT:   A clause <Clause>, two indexes in <Clause> where <i> is
           the maximal equality literal und <j> the other, the
	   left and right side of literal <i>, <Left> and <Right>,
	   left and right side of literal <j>, <PLeft> and <PRight>
	   the unifier <Subst> between <Left> and <PLeft>
	   and two boolean values, whether equality ordering restrictions
	   have to be checked for the respective literals
  RETURNS: TRUE iff equality factoring is applicable with respect to the
           ordering restrictions
***************************************************************/
{
  kbo_RESULT Help;
  
  if (Check) {
    TERM NLeft,NRight;
    NLeft  = subst_Apply(Subst,term_Copy(Left));
    NRight = subst_Apply(Subst,term_Copy(Right));
    if ((Help = kbo_Compare(NLeft,NRight)) == kbo_SmallerThan() ||
	Help == kbo_Equal()) {
      term_Delete(NLeft);
      term_Delete(NRight);
      return FALSE;
    }
    term_Delete(NLeft);
    term_Delete(NRight);
  }

  if (PCheck) {
    TERM NLeft,NRight;
    NLeft  = subst_Apply(Subst,term_Copy(PLeft));
    NRight = subst_Apply(Subst,term_Copy(PRight));
    if ((Help = kbo_Compare(NLeft,NRight)) == kbo_SmallerThan() ||
	Help == kbo_Equal()) {
      term_Delete(NLeft);
      term_Delete(NRight);
      return FALSE;
    }
    term_Delete(NLeft);
    term_Delete(NRight);
  }

  return (inf_FactorLitMax(Clause,i,j,Subst));
}


LIST inf_EqualityFactoring(CLAUSE GivenClause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: A list of clauses derivable from 'GivenClause' by EF.
***************************************************************/
{
  LIST Result;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    fputs("\n\nIllegal input in EqualityFactoring\n", stderr);
    misc_Error();
  }
#endif

  Result = list_Nil();

  if (!clause_HasEmptySuccedent(GivenClause) &&
      !clause_GetFlag(GivenClause,CLAUSESELECT) &&
      clause_HasSolvedConstraint(GivenClause)) {
    LITERAL ActLit;
    int     i,j,last;

    last = clause_LastSuccedentLitIndex(GivenClause);

    for (i = clause_FirstSuccedentLitIndex(GivenClause); i <= last; i++) {

      ActLit = clause_GetLiteral(GivenClause, i);

      if (clause_LiteralIsMaximal(ActLit) &&
	  clause_LiteralIsEquality(ActLit)) {
	TERM    Atom,Left,Right;
	LITERAL PartnerLit;

	Atom  = clause_LiteralAtom(ActLit);
	Left  = term_FirstArgument(Atom);
	Right = term_SecondArgument(Atom);

	for (j = clause_FirstSuccedentLitIndex(GivenClause); j <= last; j++) {
	  PartnerLit = clause_GetLiteral(GivenClause, j);
	  if ((i != j) && clause_LiteralIsEquality(PartnerLit) &&
	      (j > i || !clause_LiteralIsMaximal(PartnerLit))) {
	    TERM PartnerAtom,PartnerLeft,PartnerRight;

	    PartnerAtom  = clause_LiteralAtom(PartnerLit);
	    PartnerLeft  = term_FirstArgument(PartnerAtom);
	    PartnerRight = term_SecondArgument(PartnerAtom);

	    if (clause_LiteralIsOrientedEquality(ActLit)) {
	      /* Four different cases */
	      if (clause_LiteralIsOrientedEquality(PartnerLit)) {
		cont_Check();

		if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		  SUBST mgu;
		  subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		  if (inf_FactorLitMax(GivenClause,i,j,mgu))
		    Result = list_Cons(inf_ApplyEqualityFactoring(GivenClause,
								  Right,
								  PartnerRight,
								  i,j,mgu),
				       Result);
		  subst_Delete(mgu);
		}
		
		cont_Reset();
	      } else {
		cont_Check();

		if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		  SUBST mgu;
		  subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		  
		  if (inf_EqualityFactoringApplicable(GivenClause,i,j,Left,
						      Right,PartnerLeft,
						      PartnerRight,mgu,FALSE,
						      TRUE))
		    Result = list_Cons(inf_ApplyEqualityFactoring(
			       GivenClause,Right,PartnerRight,i,j,mgu),Result);
		  subst_Delete(mgu);
		}

		cont_Reset();

		PartnerLeft  = term_SecondArgument(PartnerAtom);
		PartnerRight = term_FirstArgument(PartnerAtom);

		if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		  SUBST mgu;
		  subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		  
		  if (inf_EqualityFactoringApplicable(GivenClause,i,j,Left,
						      Right, PartnerLeft,
						      PartnerRight,mgu,FALSE,
						      TRUE))
		    Result = list_Cons(inf_ApplyEqualityFactoring(
                              GivenClause,Right,PartnerRight,i,j,mgu),Result);
		  subst_Delete(mgu);
		}

		cont_Reset();
	      }
	    } else if (clause_LiteralIsOrientedEquality(PartnerLit)) {
	      cont_Check();

	      if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		SUBST mgu;
		subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		if (inf_EqualityFactoringApplicable(GivenClause,i,j,Left,Right,
						    PartnerLeft,PartnerRight,
						    mgu,TRUE,FALSE))
		  Result =
		    list_Cons(inf_ApplyEqualityFactoring(GivenClause,Right,
							 PartnerRight,i,j,
							 mgu),
			      Result);
		subst_Delete(mgu);
	      }

	      cont_Reset();

	      Left  = term_SecondArgument(Atom);
	      Right = term_FirstArgument(Atom);

	      if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		SUBST mgu;
		subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		if (inf_EqualityFactoringApplicable(GivenClause,i,j,Left,Right,
						    PartnerLeft,PartnerRight,
						    mgu,TRUE,FALSE))
		  Result =
		    list_Cons(inf_ApplyEqualityFactoring(GivenClause,Right,
							 PartnerRight,i,j,
							 mgu),
			      Result);
		subst_Delete(mgu);
	      }
	      
	      cont_Reset();
	    } else {
	      cont_Check();

	      /* if (!cont_IsContextEmpty(cont_LeftContext())){
		 printf("jaja");
		 misc_Error();
		 } */

	      if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		SUBST mgu;
		subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		
		if (inf_EqualityFactoringApplicable(GivenClause,i,j,Left,Right,
						    PartnerLeft,PartnerRight,
						    mgu,TRUE,TRUE))
		  Result =
		    list_Cons(inf_ApplyEqualityFactoring(GivenClause,Right,
							 PartnerRight,i,j,
							 mgu),
			      Result);
		subst_Delete(mgu);
	      }

	      cont_Reset();

	      PartnerLeft  = term_SecondArgument(PartnerAtom);
	      PartnerRight = term_FirstArgument(PartnerAtom);

	      if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		SUBST mgu;
		subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		
		if (inf_EqualityFactoringApplicable(GivenClause,i,j,Left,Right,
						    PartnerLeft,PartnerRight,
						    mgu,TRUE,TRUE))
		  Result =
		    list_Cons(inf_ApplyEqualityFactoring(GivenClause,Right,
							 PartnerRight,i,j,
							 mgu),
			      Result);
		subst_Delete(mgu);
	      }

	      cont_Reset();

	      Left     = term_SecondArgument(Atom);
	      Right    = term_FirstArgument(Atom);

	      if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		SUBST mgu;
		subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		
		if (inf_EqualityFactoringApplicable(GivenClause,i,j,Left,Right,
						    PartnerLeft,PartnerRight,
						    mgu,TRUE,TRUE))
		  Result =
		    list_Cons(inf_ApplyEqualityFactoring(GivenClause,Right,
							 PartnerRight,i,j,
							 mgu),
			      Result);
		subst_Delete(mgu);
	      }

	      cont_Reset();

	      PartnerLeft  = term_FirstArgument(PartnerAtom);
	      PartnerRight = term_SecondArgument(PartnerAtom);

	      if (unify_UnifyCom(cont_LeftContext(), Left, PartnerLeft)) {
		SUBST mgu;
		subst_ExtractUnifierCom(cont_LeftContext(), &mgu);
		
		if (inf_EqualityFactoringApplicable(GivenClause,i,j,Left,Right,
						    PartnerLeft,PartnerRight,
						    mgu,TRUE,TRUE))
		  Result =
		    list_Cons(inf_ApplyEqualityFactoring(GivenClause,Right,
							 PartnerRight,i,j,
							 mgu),
			      Result);
		subst_Delete(mgu);
	      }

	      cont_Reset();
	    }
	  }
	}
      }
    }
  }
  return(Result);
}


static LIST inf_ApplySuperposition(CLAUSE Clause, int ci, SUBST Subst,
				   CLAUSE PartnerClause, int pci,
				   SUBST PartnerSubst, LIST Atoms, BOOL Right)
/**************************************************************
  INPUT:   Two clauses where a superposition inference can be applied
           using the positive equality literal <i> from <Clause>
	   with subst <Subst> using the literal <j> from <PartnerClause>
	   with subst <PArtnerSubst>  where <Atoms> are all
	   possible derivable atoms
	   Right is TRUE if the inference is a superposition right inference
	   Right is FALSE if the inference is a superposition left inference
  RETURNS: The new clauses.
  MEMORY:  Memory for the new clause is allocated.
***************************************************************/
{
  CLAUSE NewClause;
  LIST   Scan,Result;
  int    j,lc,la,ls,pls,pla,plc,help;

  pls = clause_LastSuccedentLitIndex(PartnerClause);
  pla = clause_LastAntecedentLitIndex(PartnerClause);
  plc = clause_LastConstraintLitIndex(PartnerClause);

  ls  = clause_LastSuccedentLitIndex(Clause);
  la  = clause_LastAntecedentLitIndex(Clause);
  lc  = clause_LastConstraintLitIndex(Clause);

  Result = list_Nil();

  for (Scan = Atoms; !list_Empty(Scan); Scan = list_Cdr(Scan)) {

    NewClause = clause_CreateBody(clause_Length(Clause) - 1
				  + clause_Length(PartnerClause));

    clause_SetNumOfConsLits(NewClause, (clause_NumOfConsLits(Clause) +
					clause_NumOfConsLits(PartnerClause)));
    clause_SetNumOfAnteLits(NewClause, (clause_NumOfAnteLits(Clause) +
					clause_NumOfAnteLits(PartnerClause)));
    clause_SetNumOfSuccLits(NewClause, ((clause_NumOfSuccLits(Clause) -1)+
					clause_NumOfSuccLits(PartnerClause)));

    /* First set the literals from the Clause : */
    
    for (j = clause_FirstLitIndex(); j <= lc; j++) {
      clause_SetLiteral(NewClause, j, 
	clause_LiteralCreate(subst_Apply(Subst, term_Copy(
	  clause_GetLiteralTerm(Clause, j))),NewClause));
    }

    /* help = number of literals to leave empty */
    help = clause_NumOfConsLits(PartnerClause);

    for ( ; j <= la; j++) {
      clause_SetLiteral(NewClause, (j + help), 
	clause_LiteralCreate(subst_Apply(Subst, term_Copy(
	  clause_GetLiteralTerm(Clause, j))),NewClause));
    }

    /* help = number of literals to leave empty */
    help += clause_NumOfAnteLits(PartnerClause);

    for ( ; j <= ls; j++) {
      if (j != ci) {
	/* The literal used in the inference isn't copied */
	clause_SetLiteral(NewClause, (j + help), 
	  clause_LiteralCreate(subst_Apply(Subst,
	    term_Copy(clause_GetLiteralTerm(Clause, j))),NewClause));

      } else {
	/*the index has to be decreased to avoid an empty literal! */
	help--;
      }
    }

    /* Now we consider the PartnerClause : */

    /* help = number of already set constraint (Clause) literals */
    help = clause_NumOfConsLits(Clause);

    for (j = clause_FirstLitIndex(); j <= plc; j++) {
      clause_SetLiteral(NewClause, (j + help), 
	clause_LiteralCreate(subst_Apply(PartnerSubst,
	  term_Copy(clause_GetLiteralTerm(PartnerClause, j))),NewClause));
    }

    /* help = number of already set constraint and antecedent Given-literals */
    help += clause_NumOfAnteLits(Clause);

    for ( ; j <= pla; j++) {
      if (j != pci) {
      clause_SetLiteral(NewClause, (j + help), 
        clause_LiteralCreate(subst_Apply(PartnerSubst,
	  term_Copy(clause_GetLiteralTerm(PartnerClause, j))),NewClause));
      } else {
	/* The PartnerLit is modified apropiately!   */
	clause_SetLiteral(NewClause, (j + help), clause_LiteralCreate(
	  term_Create(fol_Not(),list_List(list_Car(Scan))), NewClause));
      }
    }


    /* help = number of already set Given-literals */
    help = clause_Length(Clause) - 1;

    for ( ; j <= pls; j++) {
      if (j != pci) {
	/* The PartnerLit isn't copied! */
	clause_SetLiteral(NewClause, (j + help), 
	  clause_LiteralCreate(subst_Apply(PartnerSubst,
	    term_Copy(clause_GetLiteralTerm(PartnerClause, j))),NewClause));

      } else {
	/* The PartnerLit is modified apropiately!   */
	clause_SetLiteral(NewClause, (j + help),
	  clause_LiteralCreate(list_Car(Scan), NewClause));
      }
    }
  
    if (Right)
      clause_SetFromSuperpositionRight(NewClause);
    else
      clause_SetFromSuperpositionLeft(NewClause);

    clause_SetDataFromParents(NewClause, PartnerClause, pci, Clause, ci);

    /*fputs("\n Neue Klausel SPRight: ",stdout); clause_Print(NewClause); */

    Result = list_Cons(NewClause,Result);
  }

  return Result;
}


static BOOL inf_NAllTermRplacs(TERM Term, TERM TestTerm, TERM RplacTerm,
			       SUBST Subst, int i)
/**************************************************************
  INPUT:   Three terms, a substition and an integer.
           The i-th occurrence of <TestTerm> in <Term> is replaced
           by <RplacTerm>. The substitution <Subst> is applied to <Term>.
  RETURNS: TRUE, if TestTerm was replaced by RplacTerm,
           FALSE otherwise.
  EFFECT:  <Term> is destructively changed!
***************************************************************/
{
  LIST ArgListNode;
  BOOL Replaced;
  int  Bottom;

  Replaced = FALSE;

  if (term_Equal(Term, TestTerm))
    if (i == 0) {
      term_RplacTop(Term,term_TopSymbol(RplacTerm));
      ArgListNode = term_ArgumentList(Term);
      term_RplacArgumentList(Term, list_CopyWithElement(
	term_ArgumentList(RplacTerm), (POINTER(*)()) term_Copy));
      list_DeleteWithElement(ArgListNode, term_Delete);
      return TRUE;
    }
    else
      return FALSE;

  if (!list_Empty(term_ArgumentList(Term))) {

    Bottom = stack_Bottom();
    stack_Push(term_ArgumentList(Term));

    while (!stack_Empty(Bottom)) {
      ArgListNode = stack_Top();
      Term        = (TERM)list_Car(ArgListNode);
      stack_RplacTop(list_Cdr(ArgListNode));

      if (!Replaced && term_Equal(Term, TestTerm))
	if (i > 0) {
	  i--;
	  subst_Apply(Subst,Term);
	} else {
	  Replaced = TRUE;
	  list_Rplaca(ArgListNode, term_Copy(RplacTerm));
	  term_Delete(Term);
	}
      else {
	if (term_IsComplex(Term))
	  stack_Push(term_ArgumentList(Term));
	else if (term_IsVariable(Term))
	  subst_Apply(Subst,Term);
      }
      
      while (!stack_Empty(Bottom) && list_Empty(stack_Top()))
	stack_Pop();
    }
  }
  return Replaced;
}


static LIST inf_AllTermRplacs(TERM Term, TERM TestTerm, TERM RplacTerm,
			      SUBST Subst)
/**************************************************************
  INPUT:   Three terms and a substitution.
  RETURNS: A list of copies of <Term>. In every copy only one
           occurrence of <TestTerm> is replaced by <RplacTerm>
	   and the substitution is applied to the copy.
	   Note that all terms in the list are different!
  MEMORY:  Memory is allocated for the returned list and the terms
           within.
***************************************************************/
{
  LIST   Result  = list_Nil();
  int    i       = 0;
  BOOL   Hit     = TRUE;
  TERM   ActTerm = term_Copy(Term); 

  while (Hit) {
    Hit = inf_NAllTermRplacs(ActTerm,TestTerm,RplacTerm,Subst,i);
    if (Hit) {
      Result  = list_Cons(ActTerm,Result);
      ActTerm = term_Copy(Term);
      i++;
    }
  }
  term_Delete(ActTerm);
  return Result;
}


static LIST inf_AllTermLeftRplacs(TERM Term, TERM TestTerm,
				  TERM RplacTerm, SUBST Subst)
/**************************************************************
  INPUT:   Three terms and a substitution. <Term> is typically
           an equality term.
  RETURNS: A list of copies of <Term>. In every copy only one
           occurrence of <TestTerm> is replaced by <RplacTerm>
	   and the substitution is applied to the copy.
	   Note that all terms in the list are different!
	   In contrast to the previous function this changes
           are applied only to the first (left) argument of <Term>
  MEMORY:  Memory is allocated for the returned list and the terms
           within.
***************************************************************/
{
  LIST   Result  = list_Nil();
  int    i       = 0;
  BOOL   Hit     = TRUE;
  TERM   ActTerm = term_Copy(Term); 

  while (Hit) {
    Hit = inf_NAllTermRplacs(term_FirstArgument(ActTerm),TestTerm,
			     RplacTerm,Subst,i);
    if (Hit) {
      subst_Apply(Subst,term_SecondArgument(ActTerm));
      Result  = list_Cons(ActTerm,Result);
      ActTerm = term_Copy(Term);
      i++;
    }
  }
  term_Delete(ActTerm);
  return Result;
}


static LIST inf_AllTermRightRplacs(TERM Term, TERM TestTerm,
				   TERM RplacTerm, SUBST Subst)
/**************************************************************
  INPUT:   Three terms and a substitution. <Term> is typically
           an equality term.
  RETURNS: A list of copies of <Term>. In every copy only one
           occurrence of <TestTerm> is replaced by <RplacTerm>
	   and the substitution is applied to the copy.
	   Note that all terms in the list are different!
	   In contrast to the previous function this changes
           are applied only to the second (right) argument of <Term>
  MEMORY:  Memory is allocated for the returned list and the terms
           within.
***************************************************************/
{
  LIST   Result  = list_Nil();
  int    i       = 0;
  BOOL   Hit     = TRUE;
  TERM   ActTerm = term_Copy(Term); 

  while (Hit) {
    Hit = inf_NAllTermRplacs(term_SecondArgument(ActTerm),TestTerm,
			     RplacTerm,Subst,i);
    if (Hit) {
      subst_Apply(Subst,term_FirstArgument(ActTerm));
      Result  = list_Cons(ActTerm,Result);
      ActTerm = term_Copy(Term);
      i++;
    }
  }
  term_Delete(ActTerm);
  return Result;
}


static LIST inf_LitSPRight(CLAUSE Clause, TERM Left, TERM Right,
			   int i, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause (unshared) with a positive, strictly maximal
           equality literal at position <i> where <Left> and <Right>
	   are the arguments of just that literal with <Right>
	   is not greater in the ordering than <Left>
  RETURNS: A list of clauses derivable with the literals owning
           clause by superposition right wrt. the Index.
  MEMORY:  The list of clauses is extended, where memory for the
           list and the clauses is allocated.
***************************************************************/
{
  LIST Result,Terms,Scan;

  Result  = list_Nil();
  Terms   = st_GetUnifier(cont_LeftContext(), sharing_Index(ShIndex), cont_RightContext(), Left);

  /*fputs("\n LitSP: ",stdout);clause_Print(Clause);*/

  for (Scan = Terms; !list_Empty(Scan); Scan = list_Cdr(Scan)) {

    LIST Lits,Iter;
    TERM Term;

    Term = (TERM)list_First(Scan);

    if (!term_IsVariable(Term) && !symbol_IsPredicate(term_TopSymbol(Term))) {

      Lits = sharing_GetDataList(Term, ShIndex);

      for(Iter =  Lits; !list_Empty(Iter); Iter = list_Cdr(Iter)) {

	LITERAL PartnerLit;
	TERM    PartnerAtom;
	CLAUSE  PartnerClause;
	int     pli;

	PartnerLit        = (LITERAL)list_Car(Iter);
	PartnerAtom       = clause_LiteralAtom(PartnerLit);
	pli               = clause_LiteralGetIndex(PartnerLit);
	PartnerClause     = clause_LiteralOwningClause(PartnerLit);

	/*fputs("\n\t Partner: ",stdout);clause_Print(PartnerClause);*/

	if (!clause_GetFlag(PartnerClause,CLAUSESELECT) &&
	    clause_LiteralGetFlag(PartnerLit,STRICTMAXIMAL) &&
	    clause_HasSolvedConstraint(PartnerClause) &&
	    clause_GetFlag(PartnerClause,WORKEDOFF) &&
	    clause_LiteralIsPositive(PartnerLit)) {

	  SUBST  Subst, PartnerSubst;
	  TERM   NewLeft,NewRight;
	  SYMBOL PartnerMaxVar;
	  LIST   SupAtoms;

	  SupAtoms      = list_Nil();
	  PartnerMaxVar = clause_MaxVar(PartnerClause);
	  NewLeft       = Left;
	  clause_RenameVarsBiggerThan(Clause, PartnerMaxVar);
	  cont_Check();
	  unify_UnifyNoOC(cont_LeftContext(), Left, cont_RightContext(), Term);
	  subst_ExtractUnifier(cont_LeftContext(), &Subst, cont_RightContext(), &PartnerSubst);
	  cont_Reset();
	  if (inf_InferenceLitsMax(Clause, i, Subst,
				   PartnerClause, pli, PartnerSubst)) {
	    NewRight = subst_Apply(Subst, term_Copy(Right));
	    if (!clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,i)))
	      NewLeft  = subst_Apply(Subst, term_Copy(Left));
	    if (NewLeft == Left ||       /* TRUE, if oriented */
		kbo_Compare(NewLeft,NewRight) != kbo_SmallerThan()) {
	      if (fol_IsEquality(PartnerAtom))
		if (clause_LiteralIsOrientedEquality(PartnerLit))
		  SupAtoms = inf_AllTermLeftRplacs(PartnerAtom,Term,
						   NewRight,PartnerSubst);
		else {
		  TERM NewPartnerLeft,NewPartnerRight;
		  NewPartnerLeft  =
		    subst_Apply(PartnerSubst, 
				term_Copy(term_FirstArgument(PartnerAtom)));
		  NewPartnerRight =
		    subst_Apply(PartnerSubst, 
				term_Copy(term_SecondArgument(PartnerAtom)));
		  switch (kbo_Compare(NewPartnerLeft,NewPartnerRight)) {
		  case kbo_SMALLER_THAN:
		    SupAtoms = inf_AllTermRightRplacs(PartnerAtom,Term,
						      NewRight,PartnerSubst);
		    break;
		  case kbo_GREATER_THAN:
		    SupAtoms = inf_AllTermLeftRplacs(PartnerAtom,Term,
						     NewRight,PartnerSubst);
		    break;
		  default:
		    SupAtoms = inf_AllTermRplacs(PartnerAtom,Term,
						 NewRight,PartnerSubst);
		  }
		  term_Delete(NewPartnerLeft);
		  term_Delete(NewPartnerRight);
		}
	      else
		SupAtoms = inf_AllTermRplacs(PartnerAtom,Term,
					     NewRight,PartnerSubst);

	      Result =
		list_Nconc(inf_ApplySuperposition(Clause, i, Subst,
						  PartnerClause, pli,
						  PartnerSubst, SupAtoms,
						  TRUE), 
			   Result);
	      list_Delete(SupAtoms);
	    }
	    if (NewLeft != Left)
	      term_Delete(NewLeft);
	    term_Delete(NewRight);
	  }
	  subst_Delete(Subst);
	  subst_Delete(PartnerSubst);
	  clause_Normalize(Clause);
	}
      }
      list_Delete(Lits);
    }
  }
  list_Delete(Terms);

  return Result;
}


static LIST inf_SPRightEqToGiven(CLAUSE Clause, int i, BOOL Left,
				 SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   An unshared clause, the index of a strictly maximal
           succedent literal that is an equality literal and
	   a boolean values which argument to use:
	   If Left==TRUE then the left argument is used
	   otherwise the right argument.
  RETURNS: A list of clauses derivable from superposition right on the
           GivenCopy  wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST    Result,TermList,ParentList,Scan,Iter;
  int     Bottom;
  LITERAL Lit;
  TERM    Atom,Term,PartnerTerm,PartnerEq;

  Result = list_Nil();
  Lit    = clause_GetLiteral(Clause,i);
  Atom   = clause_LiteralAtom(Lit);

  Bottom = stack_Bottom();
  if (Left) /* Top Level considered in inf_LitSPRight */ 
    sharing_PushListOnStack(term_ArgumentList(term_FirstArgument(Atom)));
  else
    sharing_PushListOnStack(term_ArgumentList(term_SecondArgument(Atom)));

  while (!stack_Empty(Bottom)) {
    Term = (TERM)stack_PopResult();
    if (!term_IsVariable(Term)) {
      /* Superposition into variables is not necessary */
      TermList = st_GetUnifier(cont_LeftContext(),
			       sharing_Index(ShIndex),
			       cont_RightContext(),
			       Term);
      for(Scan=TermList;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	PartnerTerm = (TERM)list_Car(Scan);
	ParentList  = term_SupertermList(PartnerTerm);
	for (Iter=ParentList;!list_Empty(Iter);Iter=list_Cdr(Iter)) {
	  PartnerEq = (TERM)list_Car(Iter);
	  if (fol_IsEquality(PartnerEq)) {
	    CLAUSE  PartnerClause;
	    LITERAL PartnerLit;
	    TERM    PartnerAtom;
	    LIST    Scl;
	    int     j;
	    for (Scl = sharing_NAtomDataList(PartnerEq);
		 !list_Empty(Scl);
		 Scl=list_Cdr(Scl)) {
	      PartnerLit    = (LITERAL)list_Car(Scl);
	      j             = clause_LiteralGetIndex(PartnerLit);
	      PartnerClause = clause_LiteralOwningClause(PartnerLit);
	      PartnerAtom   = clause_LiteralAtom(PartnerLit);
	      if (!clause_GetFlag(PartnerClause,CLAUSESELECT) &&
		  clause_LiteralGetFlag(PartnerLit,STRICTMAXIMAL) &&
		  clause_LiteralIsPositive(PartnerLit) &&
		  (PartnerTerm == term_FirstArgument(PartnerAtom) ||
		   !clause_LiteralIsOrientedEquality(PartnerLit)) &&
		  clause_GetFlag(PartnerClause,WORKEDOFF) &&
		  clause_HasSolvedConstraint(PartnerClause) &&
		  clause_Number(PartnerClause) != clause_Number(Clause)) {
		/* inf_LitSPRight */
		SYMBOL MaxVar;
		TERM   PartnerLeft,PartnerRight;
		SUBST  Subst, PartnerSubst;
		LIST   SupAtoms;

		SupAtoms    = list_Nil();
		MaxVar      = clause_MaxVar(PartnerClause);
		PartnerLeft = PartnerTerm;
		clause_RenameVarsBiggerThan(Clause,MaxVar);
		cont_Check();
		unify_UnifyNoOC(cont_LeftContext(),Term,cont_RightContext(),PartnerTerm);
		subst_ExtractUnifier(cont_LeftContext(),&Subst,cont_RightContext(),&PartnerSubst);
		cont_Reset();
		if (inf_InferenceLitsMax(Clause,i,Subst,PartnerClause,
					 j,PartnerSubst)) {
		  if (PartnerTerm == term_FirstArgument(PartnerAtom))
		    PartnerRight =
		      subst_Apply(PartnerSubst,
				  term_Copy(term_SecondArgument(PartnerAtom)));
		  else 
		    PartnerRight =
		      subst_Apply(PartnerSubst,
				  term_Copy(term_FirstArgument(PartnerAtom)));
		  if (!clause_LiteralIsOrientedEquality(PartnerLit))
		    PartnerLeft  =
		      subst_Apply(PartnerSubst,term_Copy(PartnerTerm));

		  if (PartnerLeft == PartnerTerm ||
		      kbo_Compare(PartnerLeft,PartnerRight)
		      != kbo_SmallerThan()) {

		    if (clause_LiteralIsOrientedEquality(Lit))
		      SupAtoms = inf_AllTermLeftRplacs(Atom,Term,
						       PartnerRight,Subst);
		    else {
		      TERM NewLeft,NewRight;
		      NewLeft  =
			subst_Apply(Subst,
				    term_Copy(term_FirstArgument(Atom)));
		      NewRight =
			subst_Apply(Subst,
				    term_Copy(term_SecondArgument(Atom)));

		      if (Left) {
			if (kbo_Compare(NewLeft,NewRight) != kbo_SmallerThan())
			  SupAtoms = inf_AllTermLeftRplacs(Atom,Term,
							   PartnerRight,Subst);
		      }
		      else
			if (kbo_Compare(NewRight,NewLeft) != kbo_SmallerThan())
			  SupAtoms =
			    inf_AllTermRightRplacs(Atom,Term,
						   PartnerRight,Subst);
		      term_Delete(NewLeft);
		      term_Delete(NewRight);
		    }

		    Result =
		      list_Nconc(inf_ApplySuperposition(PartnerClause, j,
							PartnerSubst, Clause,
							i, Subst, SupAtoms,
							TRUE), 
				 Result);
		    list_Delete(SupAtoms);
		  }
		  if (PartnerLeft != PartnerTerm)
		    term_Delete(PartnerLeft);
		  term_Delete(PartnerRight);
		}
		subst_Delete(Subst);
		subst_Delete(PartnerSubst);
		clause_Normalize(Clause);
	      }
	    }
	  }
	}
      }
      list_Delete(TermList);
    }
  }
  return Result;
}


static LIST inf_SPRightLitToGiven(CLAUSE Clause, int i, TERM Atom,
				  SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   An unshared clause, the index of a strictly maximal
           succedent literal that is not an equality literal
	   and its atom
  RETURNS: A list of clauses derivable from superposition right on the
           GivenCopy  wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST    Result,TermList,ParentList,Scan,Iter;
  int     Bottom;
  LITERAL Lit;
  TERM    Term,PartnerTerm,PartnerEq;

  Result = list_Nil();
  Lit    = clause_GetLiteral(Clause,i);

  Bottom = stack_Bottom();
  sharing_PushListOnStack(term_ArgumentList(Atom));

  while (!stack_Empty(Bottom)) {
    Term = (TERM)stack_PopResult();
    if (!term_IsVariable(Term)) {
      /* Superposition into variables is not necessary */
      TermList = st_GetUnifier(cont_LeftContext(),
			       sharing_Index(ShIndex),
			       cont_RightContext(),
			       Term);
      for(Scan=TermList;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	PartnerTerm = (TERM)list_Car(Scan);
	ParentList  = term_SupertermList(PartnerTerm);
	for (Iter=ParentList;!list_Empty(Iter);Iter=list_Cdr(Iter)) {
	  PartnerEq = (TERM)list_Car(Iter);
	  if (fol_IsEquality(PartnerEq)) {
	    CLAUSE  PartnerClause;
	    LITERAL PartnerLit;
	    TERM    PartnerAtom;
	    LIST    Scl;
	    int     j;
	    for (Scl = sharing_NAtomDataList(PartnerEq);
		 !list_Empty(Scl);
		 Scl = list_Cdr(Scl)) {
	      PartnerLit    = (LITERAL)list_Car(Scl);
	      j             = clause_LiteralGetIndex(PartnerLit);
	      PartnerClause = clause_LiteralOwningClause(PartnerLit);
	      PartnerAtom   = clause_LiteralAtom(PartnerLit);
	      if (!clause_GetFlag(PartnerClause,CLAUSESELECT) &&
		  clause_LiteralGetFlag(PartnerLit,STRICTMAXIMAL) &&
		  clause_LiteralIsPositive(PartnerLit) &&
		  (PartnerTerm == term_FirstArgument(PartnerAtom) ||
		   !clause_LiteralIsOrientedEquality(PartnerLit)) &&
		  clause_GetFlag(PartnerClause,WORKEDOFF) &&
		  clause_HasSolvedConstraint(PartnerClause) &&
		  clause_Number(PartnerClause) != clause_Number(Clause)) {
		/* inf_LitSPRight */
		SYMBOL MaxVar;
		TERM   PartnerLeft,PartnerRight;
		SUBST  Subst, PartnerSubst;
		LIST   SupAtoms;

		SupAtoms    = list_Nil();
		MaxVar      = clause_MaxVar(PartnerClause);
		clause_RenameVarsBiggerThan(Clause,MaxVar);
		cont_Check();
		unify_UnifyNoOC(cont_LeftContext(),Term,cont_RightContext(),PartnerTerm);
		subst_ExtractUnifier(cont_LeftContext(),&Subst,cont_RightContext(),&PartnerSubst);
		cont_Reset();
		if (inf_InferenceLitsMax(Clause,i,Subst,PartnerClause,
					 j,PartnerSubst)) {
		  PartnerLeft = subst_Apply(PartnerSubst,
					    term_Copy(PartnerTerm));
		  if (PartnerTerm == term_FirstArgument(PartnerAtom))
		    PartnerRight =
		      subst_Apply(PartnerSubst,
				  term_Copy(term_SecondArgument(PartnerAtom)));
		  else
		    PartnerRight =
		      subst_Apply(PartnerSubst,
				  term_Copy(term_FirstArgument(PartnerAtom)));

		
		  if (clause_LiteralIsOrientedEquality(PartnerLit) ||
		      kbo_Compare(PartnerLeft,PartnerRight)
		      != kbo_SmallerThan()) {
		    SupAtoms = inf_AllTermRplacs(Atom,Term,PartnerRight,Subst);
		    Result =
		      list_Nconc(inf_ApplySuperposition(PartnerClause, j,
							PartnerSubst, Clause,
							i, Subst, SupAtoms,
							TRUE), 
				 Result);
		    list_Delete(SupAtoms);
		  }
		  term_Delete(PartnerLeft);
		  term_Delete(PartnerRight);
		}
		subst_Delete(Subst);
		subst_Delete(PartnerSubst);
		clause_Normalize(Clause);
	      }
	    }
	  }
	}
      }
      list_Delete(TermList);
    }
  }
  return Result;
}


static LIST inf_SPRightToGiven(CLAUSE Clause, int i, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   An unshared clause, the index of a strictly maximal
           succedent literal, an index of shared clauses and
	   a List of derived clauses.
  RETURNS: A list of clauses derivable from superposition right on the
           GivenCopy  wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  TERM Atom;
  LIST Result;

#if DEBUG
  if (!(clause_IsClause(Clause))) {
    fputs("\n\nIllegal input in inf_SP_RightToGiven\n", stderr);
    misc_Error();
  }
#endif

  Result  = list_Nil();
  Atom    = clause_LiteralAtom(clause_GetLiteral(Clause,i));

  if (fol_IsEquality(Atom)) {
    Result = list_Nconc(inf_SPRightEqToGiven(Clause,i, TRUE,ShIndex),
			Result);

    if (!clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,i)))
      Result = list_Nconc(inf_SPRightEqToGiven(Clause, i, FALSE, ShIndex),
			  Result);
  } else
    Result = list_Nconc(inf_SPRightLitToGiven(Clause,i,Atom,ShIndex), Result);
  
  return Result;
}



LIST inf_SuperpositionRight(CLAUSE GivenClause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause and an Index, usually the WorkedOffIndex.
  RETURNS: A list of clauses derivable from the Givenclause by 
           superposition right wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST Result;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    fputs("\n\nIllegal input in SuperpositionRight\n", stderr);
    misc_Error();
  }
#endif

  Result = list_Nil();

  if (!clause_GetFlag(GivenClause,CLAUSESELECT) &&
      !clause_HasEmptySuccedent(GivenClause) && 
      clause_HasSolvedConstraint(GivenClause)) {
    TERM    Atom;
    CLAUSE  Copy;
    int     i,n;
    LITERAL ActLit;

    Copy = clause_Copy(GivenClause);
    n    = clause_LastSuccedentLitIndex(Copy);

    for(i=clause_FirstSuccedentLitIndex(Copy); i <= n; i++) {

      ActLit = clause_GetLiteral(Copy, i);
      Atom   = clause_LiteralSignedAtom(ActLit);

      if (clause_LiteralGetFlag(ActLit,STRICTMAXIMAL)) {

	if (fol_IsEquality(Atom)) {

	  Result = list_Nconc(inf_LitSPRight(Copy, term_FirstArgument(Atom), 
					     term_SecondArgument(Atom),
					     i, ShIndex), 
			       Result);

	  if (!clause_LiteralIsOrientedEquality(ActLit))
	    Result = list_Nconc(inf_LitSPRight(Copy,
					       term_SecondArgument(Atom), 
					       term_FirstArgument(Atom),
					       i, ShIndex), 
				Result);
	}
	Result = list_Nconc(inf_SPRightToGiven(Copy, i, ShIndex), Result);
      }
    }
    clause_Delete(Copy);
  }
  return Result;
}


static LIST inf_LitSPLeft(CLAUSE Clause, TERM Left, TERM Right,
			  int i, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause (unshared) with a positive, strictly maximal
           equality literal at position <i> where <Left> and <Right>
	   are the arguments of just that literal with <Right>
	   is not greater in the ordering than <Left>
  RETURNS: A list of clauses derivable with the literals owning
           clause by superposition Left wrt. the Index.
  MEMORY:  The list of clauses is extended, where memory for the
           list and the clauses is allocated.
***************************************************************/
{
  LIST Result,Terms,Scan;

  Result  = list_Nil();
  Terms   = st_GetUnifier(cont_LeftContext(), sharing_Index(ShIndex), cont_RightContext(), Left);

  /*fputs("\n LitSPLeft: ",stdout); clause_Print(Clause);*/

  for (Scan = Terms; !list_Empty(Scan); Scan = list_Cdr(Scan)) {

    LIST Lits,Iter;
    TERM Term;

    Term = (TERM)list_First(Scan);

    if (!term_IsVariable(Term) && !symbol_IsPredicate(term_TopSymbol(Term))) {

      Lits = sharing_GetDataList(Term, ShIndex);

      for(Iter =  Lits; !list_Empty(Iter); Iter = list_Cdr(Iter)){

	LITERAL PartnerLit;
	TERM    PartnerAtom;
	CLAUSE  PartnerClause;
	int     pli;

	PartnerLit        = (LITERAL)list_Car(Iter); /* Antecedent Literal ! */
	PartnerAtom       = clause_LiteralAtom(PartnerLit);  
	pli               = clause_LiteralGetIndex(PartnerLit);
	PartnerClause     = clause_LiteralOwningClause(PartnerLit);

	/*fputs("\n\t Partner: ",stdout); clause_Print(PartnerClause);*/

	if (((clause_LiteralIsMaximal(PartnerLit) &&
	      !clause_GetFlag(PartnerClause,CLAUSESELECT))||
	     clause_LiteralGetFlag(PartnerLit,LITSELECT)) &&
	    clause_HasSolvedConstraint(PartnerClause) &&
	    clause_GetFlag(PartnerClause,WORKEDOFF) &&
	    clause_LiteralIsNegative(PartnerLit) &&
	    clause_Number(PartnerClause) != clause_Number(Clause)) {

	  SUBST  Subst, PartnerSubst;
	  TERM   NewLeft,NewRight;
	  SYMBOL PartnerMaxVar;
	  LIST   SupAtoms;

	  SupAtoms      = list_Nil();
	  PartnerMaxVar = clause_MaxVar(PartnerClause);
	  NewLeft       = Left;
	  clause_RenameVarsBiggerThan(Clause, PartnerMaxVar);
	  cont_Check();
	  unify_UnifyNoOC(cont_LeftContext(), Left, cont_RightContext(), Term);
	  subst_ExtractUnifier(cont_LeftContext(), &Subst, cont_RightContext(), &PartnerSubst);
	  cont_Reset();
	  if (inf_InferenceLitsMax(Clause, i, Subst, PartnerClause,
				   pli, PartnerSubst)) {
	    NewRight = subst_Apply(Subst, term_Copy(Right));
	    if (!clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,i)))
	      NewLeft  = subst_Apply(Subst, term_Copy(Left));
	    if (NewLeft == Left ||
		kbo_Compare(NewLeft,NewRight) != kbo_SmallerThan()) {
	      if (fol_IsEquality(PartnerAtom))
		if (clause_LiteralIsOrientedEquality(PartnerLit))
		  SupAtoms = inf_AllTermLeftRplacs(PartnerAtom,Term,NewRight,
						   PartnerSubst);
		else {
		  TERM NewPartnerLeft,NewPartnerRight;
		  NewPartnerLeft  = subst_Apply(PartnerSubst, 
		       	    term_Copy(term_FirstArgument(PartnerAtom)));
		  NewPartnerRight = subst_Apply(PartnerSubst, 
			    term_Copy(term_SecondArgument(PartnerAtom)));
		  switch (kbo_Compare(NewPartnerLeft,NewPartnerRight)) {
		  case kbo_SMALLER_THAN:
		    SupAtoms = inf_AllTermRightRplacs(PartnerAtom,Term,
						      NewRight,PartnerSubst);
		    break;
		  case kbo_GREATER_THAN:
		    SupAtoms = inf_AllTermLeftRplacs(PartnerAtom,Term,
						     NewRight,PartnerSubst);
		    break;
		  default:
		    SupAtoms = inf_AllTermRplacs(PartnerAtom,Term,
						 NewRight,PartnerSubst);
		  }
		  term_Delete(NewPartnerLeft);
		  term_Delete(NewPartnerRight);
		}
	      else
		SupAtoms = inf_AllTermRplacs(PartnerAtom,Term,NewRight,
					     PartnerSubst);

	      Result = list_Nconc(inf_ApplySuperposition(Clause, i, Subst, 
							 PartnerClause, pli,
							 PartnerSubst, 
							 SupAtoms, FALSE), 
				  Result);
	      list_Delete(SupAtoms);
	    }
	    if (NewLeft != Left)
	      term_Delete(NewLeft);
	    term_Delete(NewRight);
	  }
	  subst_Delete(Subst);
	  subst_Delete(PartnerSubst);
	  clause_Normalize(Clause);
	}
      }
      list_Delete(Lits);
    }
  }
  list_Delete(Terms);

  return Result;
}


static LIST inf_SPLeftEqToGiven(CLAUSE Clause, int i, BOOL Left,
				SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   An unshared clause, the index of a maximal antecedent literal
           that is an equality literal and a boolean values
	   which argument to use:
           If Left==TRUE then the left argument is used
	   otherwise the right argument.
  RETURNS: A list of clauses derivable from superposition Left on the
           GivenCopy  wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST    Result,TermList,ParentList,Scan,Iter;
  int     Bottom;
  LITERAL Lit;
  TERM    Atom,Term,PartnerTerm,PartnerEq;

  Result = list_Nil();
  Lit    = clause_GetLiteral(Clause,i); /* Is an antecedent Literal ! */
  Atom   = clause_LiteralAtom(Lit);

  Bottom = stack_Bottom();
  if (Left) 
    sharing_PushOnStack(term_FirstArgument(Atom));
  else
    sharing_PushOnStack(term_SecondArgument(Atom));

  while (!stack_Empty(Bottom)) {
    Term = (TERM)stack_PopResult();
    if (!term_IsVariable(Term)) {
      /* Superposition into variables is not necessary */
      TermList = st_GetUnifier(cont_LeftContext(),
			       sharing_Index(ShIndex),
			       cont_RightContext(),
			       Term);
      for(Scan=TermList;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	PartnerTerm = (TERM)list_Car(Scan);
	ParentList  = term_SupertermList(PartnerTerm);
	for (Iter=ParentList;!list_Empty(Iter);Iter=list_Cdr(Iter)) {
	  PartnerEq = (TERM)list_Car(Iter);
	  if (fol_IsEquality(PartnerEq)) {
	    CLAUSE  PartnerClause;
	    LITERAL PartnerLit;
	    TERM    PartnerAtom;
	    LIST    Scl;
	    int     j;
	    for (Scl = sharing_NAtomDataList(PartnerEq);
		 !list_Empty(Scl);
		 Scl = list_Cdr(Scl)) {
	      PartnerLit    = (LITERAL)list_Car(Scl);
	      j             = clause_LiteralGetIndex(PartnerLit);
	      PartnerClause = clause_LiteralOwningClause(PartnerLit);
	      PartnerAtom   = clause_LiteralAtom(PartnerLit);
	      if (!clause_GetFlag(PartnerClause,CLAUSESELECT) &&
		  clause_LiteralGetFlag(PartnerLit,STRICTMAXIMAL) &&
		  clause_LiteralIsPositive(PartnerLit) &&
		  (PartnerTerm == term_FirstArgument(PartnerAtom) ||
		   !clause_LiteralIsOrientedEquality(PartnerLit)) &&
		  clause_GetFlag(PartnerClause,WORKEDOFF) &&
		  clause_HasSolvedConstraint(PartnerClause) &&
		  clause_Number(PartnerClause) != clause_Number(Clause)) {
		SYMBOL MaxVar;
		TERM   PartnerLeft,PartnerRight;
		SUBST  Subst, PartnerSubst;
		LIST   SupAtoms;

		SupAtoms    = list_Nil();
		MaxVar      = clause_MaxVar(PartnerClause);
		PartnerLeft = PartnerTerm;
		clause_RenameVarsBiggerThan(Clause,MaxVar);
		cont_Check();
		unify_UnifyNoOC(cont_LeftContext(),Term,cont_RightContext(),PartnerTerm);
		subst_ExtractUnifier(cont_LeftContext(),&Subst,cont_RightContext(),&PartnerSubst);
		cont_Reset();
		if (inf_InferenceLitsMax(Clause,i,Subst,PartnerClause,
					 j,PartnerSubst)) {
		  if (PartnerTerm == term_FirstArgument(PartnerAtom))
		    PartnerRight = subst_Apply(PartnerSubst,
			     term_Copy(term_SecondArgument(PartnerAtom)));
		  else 
		    PartnerRight = subst_Apply(PartnerSubst,
			     term_Copy(term_FirstArgument(PartnerAtom)));
		  if (!clause_LiteralIsOrientedEquality(PartnerLit))
		    PartnerLeft  = subst_Apply(PartnerSubst,
					       term_Copy(PartnerTerm));

		  if (PartnerLeft == PartnerTerm ||
		      kbo_Compare(PartnerLeft,PartnerRight)
		      != kbo_SmallerThan()) {

		    if (clause_LiteralIsOrientedEquality(Lit))
		      SupAtoms = inf_AllTermLeftRplacs(Atom,Term,PartnerRight,
						       Subst);
		    else {
		      TERM NewLeft,NewRight;
		      NewLeft  = subst_Apply(Subst,
				     term_Copy(term_FirstArgument(Atom)));
		      NewRight = subst_Apply(Subst,
				     term_Copy(term_SecondArgument(Atom)));

		      if(Left) {
			if (kbo_Compare(NewLeft,NewRight) != kbo_SmallerThan())
			  SupAtoms = inf_AllTermLeftRplacs(Atom,Term,
							   PartnerRight,Subst);
		      }
		      else
			if (kbo_Compare(NewRight,NewLeft) != kbo_SmallerThan())
			  SupAtoms = inf_AllTermRightRplacs(Atom,Term,
							    PartnerRight,
							    Subst);
		      term_Delete(NewLeft);
		      term_Delete(NewRight);
		    }

		    Result =
		      list_Nconc(inf_ApplySuperposition(PartnerClause, j,
							PartnerSubst, Clause,
							i, Subst, SupAtoms,
							FALSE), 
				 Result);
		    list_Delete(SupAtoms);
		  }
		  if (PartnerLeft != PartnerTerm)
		    term_Delete(PartnerLeft);
		  term_Delete(PartnerRight);
		}
		subst_Delete(Subst);
		subst_Delete(PartnerSubst);
		clause_Normalize(Clause);
	      }
	    }
	  }
	}
      }
      list_Delete(TermList);
    }
  }
  return Result;
}


static LIST inf_SPLeftLitToGiven(CLAUSE Clause, int i, TERM Atom,
				 SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   An unshared clause, the index of a strictly maximal
           succedent literal that is not an equality literal
	   and its atom
  RETURNS: A list of clauses derivable from superposition Left on the
           GivenCopy  wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST    Result,TermList,ParentList,Scan,Iter;
  int     Bottom;
  LITERAL Lit;
  TERM    Term,PartnerTerm,PartnerEq;

  Result = list_Nil();
  Lit    = clause_GetLiteral(Clause,i);

  Bottom = stack_Bottom();
  sharing_PushListOnStack(term_ArgumentList(Atom));

  while (!stack_Empty(Bottom)) {
    Term = (TERM)stack_PopResult();
    if (!term_IsVariable(Term)) {
      /* Superposition into variables is not necessary */
      TermList = st_GetUnifier(cont_LeftContext(),
			       sharing_Index(ShIndex),
			       cont_RightContext(),
			       Term);
      for(Scan=TermList;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	PartnerTerm = (TERM)list_Car(Scan);
	ParentList  = term_SupertermList(PartnerTerm);
	for (Iter=ParentList;!list_Empty(Iter);Iter=list_Cdr(Iter)) {
	  PartnerEq = (TERM)list_Car(Iter);
	  if (fol_IsEquality(PartnerEq)) {
	    CLAUSE  PartnerClause;
	    LITERAL PartnerLit;
	    TERM    PartnerAtom;
	    LIST    Scl;
	    int     j;
	    for (Scl = sharing_NAtomDataList(PartnerEq);
		 !list_Empty(Scl);
		 Scl = list_Cdr(Scl)) {
	      PartnerLit    = (LITERAL)list_Car(Scl);
	      j             = clause_LiteralGetIndex(PartnerLit);
	      PartnerClause = clause_LiteralOwningClause(PartnerLit);
	      PartnerAtom   = clause_LiteralAtom(PartnerLit);
	      if (!clause_GetFlag(PartnerClause,CLAUSESELECT) &&
		  clause_LiteralGetFlag(PartnerLit,STRICTMAXIMAL) &&
		  clause_LiteralIsPositive(PartnerLit) &&
		  (PartnerTerm == term_FirstArgument(PartnerAtom) ||
		   !clause_LiteralIsOrientedEquality(PartnerLit)) &&
		  clause_GetFlag(PartnerClause,WORKEDOFF) &&
		  clause_HasSolvedConstraint(PartnerClause) &&
		  clause_Number(PartnerClause) != clause_Number(Clause)) {
		SYMBOL MaxVar;
		TERM   PartnerLeft,PartnerRight;
		SUBST  Subst, PartnerSubst;
		LIST   SupAtoms;

		SupAtoms    = list_Nil();
		MaxVar      = clause_MaxVar(PartnerClause);
		clause_RenameVarsBiggerThan(Clause,MaxVar);
		cont_Check();
		unify_UnifyNoOC(cont_LeftContext(),Term,cont_RightContext(),PartnerTerm);
		subst_ExtractUnifier(cont_LeftContext(),&Subst,cont_RightContext(),&PartnerSubst);
		cont_Reset();
		if (inf_InferenceLitsMax(Clause,i,Subst,PartnerClause,
					 j,PartnerSubst)) {
		  PartnerLeft  = subst_Apply(PartnerSubst,
				 term_Copy(PartnerTerm));
		  if (PartnerTerm == term_FirstArgument(PartnerAtom))
		    PartnerRight = subst_Apply(PartnerSubst,
				 term_Copy(term_SecondArgument(PartnerAtom)));
		  else
		    PartnerRight = subst_Apply(PartnerSubst,
				 term_Copy(term_FirstArgument(PartnerAtom)));

		
		  if (clause_LiteralIsOrientedEquality(PartnerLit) ||
		      kbo_Compare(PartnerLeft,PartnerRight)
		      != kbo_SmallerThan()) {
		    SupAtoms = inf_AllTermRplacs(Atom,Term,PartnerRight,Subst);
		    Result =
		      list_Nconc(inf_ApplySuperposition(PartnerClause, j,
							PartnerSubst, Clause,
							i, Subst, SupAtoms,
							FALSE), 
				 Result);
		    list_Delete(SupAtoms);
		  }
		  term_Delete(PartnerLeft);
		  term_Delete(PartnerRight);
		}
		subst_Delete(Subst);
		subst_Delete(PartnerSubst);
		clause_Normalize(Clause);
	      }
	    }
	  }
	}
      }
      list_Delete(TermList);
    }
  }
  return Result;
}


static LIST inf_SPLeftToGiven(CLAUSE Clause, int i, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   An unshared clause, the index of a strictly maximal
           antecedent literal, and an index of shared clauses.
  RETURNS: A list of clauses derivable from superposition Left on the
           GivenCopy  wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  TERM Atom;
  LIST Result;

#if DEBUG
  if (!(clause_IsClause(Clause))) {
    fputs("\n\nIllegal input in inf_SPLeftToGiven\n", stderr);
    misc_Error();
  }
#endif

  Result  = list_Nil();
  Atom    = clause_LiteralAtom(clause_GetLiteral(Clause,i));

  if (fol_IsEquality(Atom)) {
    Result = list_Nconc(inf_SPLeftEqToGiven(Clause,i, TRUE,ShIndex), Result);
    if (!clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,i)))
      /* If not oriented also try other direction */
      Result = list_Nconc(inf_SPLeftEqToGiven(Clause, i, FALSE, ShIndex),
			  Result);
  } else
    Result = list_Nconc(inf_SPLeftLitToGiven(Clause,i,Atom,ShIndex), Result);

  return Result;
}


LIST inf_SuperpositionLeft(CLAUSE GivenClause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause and an Index, usually the WorkedOffIndex.
  RETURNS: A list of clauses derivable from the Givenclause by 
           superposition left wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST Result;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    fputs("\n\nIllegal input in SuperpositionLeft\n", stderr);
    misc_Error();
  }
#endif

  Result = list_Nil();

  if (clause_HasSolvedConstraint(GivenClause)) {
    TERM    Atom;
    CLAUSE  Copy;
    int     i,n;
    LITERAL ActLit;

    Copy = clause_Copy(GivenClause);
    n    = clause_LastSuccedentLitIndex(Copy);
    
    if (!clause_GetFlag(Copy,CLAUSESELECT)) {
      for (i = clause_FirstSuccedentLitIndex(Copy); i <= n; i++) {
	ActLit = clause_GetLiteral(Copy, i);
	Atom   = clause_LiteralSignedAtom(ActLit);

	if (clause_LiteralGetFlag(ActLit,STRICTMAXIMAL)) {

	  if (fol_IsEquality(Atom)) {

	    Result =
	      list_Nconc(inf_LitSPLeft(Copy, term_FirstArgument(Atom), 
				       term_SecondArgument(Atom), i, ShIndex), 
			 Result);

	    if (!clause_LiteralIsOrientedEquality(ActLit))
	      Result =
		list_Nconc(inf_LitSPLeft(Copy, term_SecondArgument(Atom), 
					 term_FirstArgument(Atom),i,ShIndex), 
			   Result);
	  }
	}
      }
    }
    n    = clause_LastAntecedentLitIndex(Copy);

    for (i = clause_FirstAntecedentLitIndex(Copy); i <= n; i++) {

      ActLit = clause_GetLiteral(Copy, i);
      Atom   = clause_LiteralAtom(ActLit);

      if ((clause_LiteralIsMaximal(ActLit) &&
	   !clause_GetFlag(Copy,CLAUSESELECT)) ||
	  clause_LiteralGetFlag(ActLit,LITSELECT))
	Result = list_Nconc(inf_SPLeftToGiven(Copy, i, ShIndex), Result);
    }
    clause_Delete(Copy);
  }
  return(Result);
}


static BOOL inf_MParamodLitsMax(CLAUSE Clause, int i, CLAUSE PartnerClause,
				int j, SUBST tau, SUBST rho)
/**************************************************************
  INPUT:   The parents of merging paramodulation inference, the
           respective literal indices and substitutions.
  RETURNS: TRUE iff the literals are still maximal in the
           instantiated clauses.
***************************************************************/
{
  TERM       max, lit;
  kbo_RESULT compare;
  int        k,l;
  BOOL       stop = FALSE;

  l   = clause_Length(Clause);

  if (!clause_GetFlag(Clause,CLAUSESELECT) && l != 1 ) {
    /* The check doesn't make sense for unit, or selected clauses */
    max = subst_Apply(rho,subst_Apply(tau,term_Copy(clause_GetLiteralTerm(Clause,i))));
    for(k = clause_FirstAntecedentLitIndex(Clause); k < l && !stop; k++)
      if (k != i && clause_LiteralIsMaximal(clause_GetLiteral(Clause, k))) {
	lit     = subst_Apply(rho,subst_Apply(tau,term_Copy(clause_GetLiteralTerm(Clause,k))));
	compare = kbo_LiteralCompare(max, clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,i)),
				     lit, clause_LiteralIsOrientedEquality(clause_GetLiteral(Clause,k)),TRUE);
	if (compare == kbo_SmallerThan() ||
	    (compare == kbo_Equal() && k > clause_LastAntecedentLitIndex(Clause)))
	  stop = TRUE;

	term_Delete(lit);
      }
    term_Delete(max);
  }

  l   = clause_Length(PartnerClause);

  if (!stop && !clause_GetFlag(PartnerClause,CLAUSESELECT) && l != 1) {
    /* The check doesn't make sense for unit, or selected clauses */
    max = subst_Apply(rho, subst_Apply(tau,term_Copy(clause_GetLiteralTerm(PartnerClause,j))));
    for(k = clause_FirstAntecedentLitIndex(PartnerClause); k < l && !stop; k++)
      if (k != j && clause_LiteralIsMaximal(clause_GetLiteral(PartnerClause, k))) {
	lit     = subst_Apply(rho, subst_Apply(tau,term_Copy(clause_GetLiteralTerm(PartnerClause,k))));
	compare = kbo_LiteralCompare(max, clause_LiteralIsOrientedEquality(clause_GetLiteral(PartnerClause,j)),
				     lit, clause_LiteralIsOrientedEquality(clause_GetLiteral(PartnerClause,k)),TRUE);
	if (compare == kbo_SmallerThan() ||
	    (compare == kbo_Equal() && k > clause_LastAntecedentLitIndex(PartnerClause)))
	  stop = TRUE;

	term_Delete(lit);
      }
    term_Delete(max);
  }

  return !stop;   /* TRUE, if both for loops were not aborted */
}


static LIST inf_ApplyMParamod(CLAUSE C1, CLAUSE C2, int i, int j, int k,
			      TERM u_tau, TERM v, TERM s2, TERM t,
			      TERM v2_sigma, SUBST tau, SUBST rho)
/**************************************************************
  INPUT:   Two clauses, <i> is a literal index in <C1>, <j> and <k>
           are literal indices in <C2>, <u_tau> with <rho> applied
	   is used as left side of the two new literals, <v> with
	   subterm <s2> replaced by <t> is used as right side of
	   the first new literal, <v2_sigma> as right side of the
	   second new literal and two substitutions.
	   The substitution <rho> is applied after <tau>.
  RETURNS: A list containing one clause derived from the given
           clauses and literals by merging paramodulation.
  MEMORY:  Memory for the list and the clause is allocated.
***************************************************************/
{
  CLAUSE newClause;
  TERM   u_sigma;
  LIST   result;
  int    m, lc, la, ls, pls, pla, plc, help;

  pls = clause_LastSuccedentLitIndex(C2);
  pla = clause_LastAntecedentLitIndex(C2);
  plc = clause_LastConstraintLitIndex(C2);

  ls  = clause_LastSuccedentLitIndex(C1);
  la  = clause_LastAntecedentLitIndex(C1);
  lc  = clause_LastConstraintLitIndex(C1);

  newClause = clause_CreateBody(clause_Length(C1) + clause_Length(C2) - 1);

  clause_SetNumOfConsLits(newClause, (clause_NumOfConsLits(C1) +
				      clause_NumOfConsLits(C2)));
  clause_SetNumOfAnteLits(newClause, (clause_NumOfAnteLits(C1) +
				      clause_NumOfAnteLits(C2)));
  clause_SetNumOfSuccLits(newClause, (clause_NumOfSuccLits(C1) - 1 +
				      clause_NumOfSuccLits(C2)));

  for (m = clause_FirstLitIndex(); m <= lc; m++)
    clause_SetLiteral(newClause, m,
      clause_LiteralCreate(subst_Apply(rho, subst_Apply(tau,
	term_Copy(clause_GetLiteralTerm(C1, m)))), newClause));

  /* help = number of literals to leave empty */
  help = clause_NumOfConsLits(C2);

  for ( ; m <= la; m++)
    clause_SetLiteral(newClause, (m + help),
      clause_LiteralCreate(subst_Apply(rho, subst_Apply(tau,
	term_Copy(clause_GetLiteralTerm(C1, m)))), newClause));

  /* help = number of literals to leave empty */
  help += clause_NumOfAnteLits(C2);

  for ( ; m <= ls; m++) {
    if (m != i)
      /* The literal used in the inference isn't copied */
      clause_SetLiteral(newClause, (m + help), 
	clause_LiteralCreate(subst_Apply(rho, subst_Apply(tau,
	  term_Copy(clause_GetLiteralTerm(C1, m)))),newClause));
    else
      /*the index has to be decreased to avoid an empty literal! */
      help--;
  }

  /* Now we consider the PartnerClause : */
  
  /* help = number of already set constraint (Clause) literals */
  help = clause_NumOfConsLits(C1);
  
  for (m = clause_FirstLitIndex(); m <= plc; m++)
    clause_SetLiteral(newClause, (m + help), 
      clause_LiteralCreate(subst_Apply(rho, subst_Apply(tau,
	term_Copy(clause_GetLiteralTerm(C2, m)))),newClause));
  
  /* help = number of already set constraint and antecedent Given-literals */
  help += clause_NumOfAnteLits(C1);
  
  for ( ; m <= pla; m++)
    clause_SetLiteral(newClause, (m + help), 
      clause_LiteralCreate(subst_Apply(rho, subst_Apply(tau,
        term_Copy(clause_GetLiteralTerm(C2, m)))),newClause));

  /* help = number of already set literals */
  help += clause_NumOfSuccLits(C1) - 1;
  /*help = clause_Length(Clause) - 1;*/

  u_sigma = subst_Apply(rho, term_Copy(u_tau));

  for ( ; m <= pls; m++) {
    TERM newAtom;

    if (m == j) {
      /* The first partner literal is modified appropiately! */
      TERM right;
      if (v == s2)
	/* Necessary because term_ReplaceSubtermBy doesn't treat top level */
	right = term_Copy(t);
      else {
	right = term_Copy(v);
	term_ReplaceSubtermBy(right, s2, t);
      }		
      newAtom = term_Create(fol_Equality(), list_Cons(term_Copy(u_sigma),
		  list_List(subst_Apply(rho, subst_Apply(tau, right)))));
    } else if (m == k) {
      /* The second partner lit is modified appropiately!    */
      newAtom = term_Create(fol_Equality(), list_Cons(term_Copy(u_sigma),
	          list_List(term_Copy(v2_sigma))));
    } else {
      /* Apply substitutions to all other literals */
      newAtom = subst_Apply(rho, subst_Apply(tau,
                  term_Copy(clause_GetLiteralTerm(C2, m))));
    }

    clause_SetLiteral(newClause, (m + help),
		      clause_LiteralCreate(newAtom, newClause));
   }

  term_Delete(u_sigma);

  clause_SetFromMergingParamodulation(newClause);
  
  clause_OrientAndReInit(newClause);
  clause_SetSplitDataFromParents(newClause, C1, C2);
  clause_SetDepth(newClause, misc_Max(clause_Depth(C1), clause_Depth(C2)));
  clause_SetParentClauses(newClause, list_Cons((POINTER)clause_Number(C1),
    list_Cons((POINTER)clause_Number(C2),
      list_List((POINTER)clause_Number(C2)))));
  clause_SetParentLiterals(newClause, list_Cons((POINTER)i,
    list_Cons((POINTER)j,list_List((POINTER) k))));

  result = list_List(newClause);

#ifdef _INF_DEBUGMPARAMOD_
  clause_Print(newClause); putchar('\n');
#endif

  return result;
}


static LIST inf_Lit2MParamod(CLAUSE C1, CLAUSE C2, int i, int j, TERM s,
			     TERM t, TERM s2, TERM u, TERM v, TERM u_tau,
			     TERM v_tau, SUBST tau)
/**************************************************************
  INPUT:   Two clauses, the index <i> of a strict maximal literal
           in <C1>, the index <j> of a strict maximal index in <C2>,
	   <s> and <t> are from literal <i>, <s2> is a subterm
	   of <v>, <u> and <v> are from literal <j>, <u_tau> and
	   <v_tau> are <u> and <v> with substitution <tau> applied.
  RETURNS: A list of clauses derivable from the given data
           by merging paramodulation.
  MEMORY:  Memory is allocated for the list and the clauses.
***************************************************************/
{
  LIST result;
  int  k, last;

  result = list_Nil();

  /* Now find the 3rd literal u' = v' in C2 */
  last = clause_LastSuccedentLitIndex(C2);  /* Last index */
  for (k = clause_FirstSuccedentLitIndex(C2); k <= last; k++) {
    LITERAL partnerLit2  = clause_GetLiteral(C2, k);
    TERM    partnerAtom2 = clause_LiteralSignedAtom(partnerLit2);

    if (k != j && fol_IsEquality(partnerAtom2) /*&&
	!clause_LiteralIsMaximal(partnerLit2))*/) {
      /* partnerLit2: u' = v'  or  v' = u' */
      SUBST      rho;
      TERM       pLeft2, pRight2, v_sigma, s_sigma, t_sigma, v2_sigma;
      kbo_RESULT kboResult;
      BOOL       checkPassed;

      pLeft2  = subst_Apply(tau, term_Copy(term_FirstArgument(partnerAtom2)));
      pRight2 = subst_Apply(tau, term_Copy(term_SecondArgument(partnerAtom2)));
      
      /* First try unification with left side */
      cont_Check();

      if (unify_UnifyCom(cont_LeftContext(), u_tau, pLeft2)) {
	subst_ExtractUnifierCom(cont_LeftContext(), &rho);

	s_sigma = t_sigma = (TERM) NULL;
	checkPassed = TRUE;

#ifdef _INF_DEBUGMPARAMOD_
	fputs("u_tau = ", stdout);   term_PrintPrefix(u_tau);
	fputs(" u2_tau = ", stdout); term_PrintPrefix(pLeft2);
	fputs(" rho = ", stdout);    subst_Print(rho);
	putchar('\n');
#endif

	if (!clause_LiteralIsOrientedEquality(clause_GetLiteral(C1,i))) {
	  s_sigma = subst_Apply(rho, subst_Apply(tau, term_Copy(s)));
	  t_sigma = subst_Apply(rho, subst_Apply(tau, term_Copy(t)));
	  kboResult = kbo_Compare(s_sigma, t_sigma);
	  if (kboResult == kbo_SmallerThan() || kboResult == kbo_Equal())
	    checkPassed = FALSE;
	}

	if (checkPassed) {
	  v_sigma = subst_Apply(rho, term_Copy(v_tau));
	  v2_sigma = subst_Apply(rho, term_Copy(pRight2));
	  kboResult = kbo_Compare(v2_sigma, v_sigma);

	  if (kboResult != kbo_GreaterThan() && kboResult != kbo_Equal() &&
	      inf_MParamodLitsMax(C1, i, C2, j, tau, rho)) {

	    result = list_Nconc(inf_ApplyMParamod(C1,C2,i,j,k,u_tau,v,s2,
						  t,v2_sigma,tau,rho),
				result);
	  }
	  term_Delete(v_sigma);
	  term_Delete(v2_sigma);
	}
	/* Now cleanup */
	if (s_sigma != NULL) {      /* Also t_sigma != NULL */
	  term_Delete(s_sigma);
	  term_Delete(t_sigma);
	}
	subst_Delete(rho);
      }

      cont_Reset();

      /* Now try unification with right side */
      if (unify_UnifyCom(cont_LeftContext(), u_tau, pRight2)) {
	subst_ExtractUnifierCom(cont_LeftContext(), &rho);

	s_sigma = t_sigma = (TERM) NULL;
	checkPassed = TRUE;

#ifdef _INF_DEBUGMPARAMOD_
	fputs("u_tau = ", stdout);   term_PrintPrefix(u_tau);
	fputs(" u2_tau = ", stdout); term_PrintPrefix(pRight2);
	fputs(" rho = ", stdout);    subst_Print(rho);
	putchar('\n');
#endif

	if (!clause_LiteralIsOrientedEquality(clause_GetLiteral(C1,i))) {
	  s_sigma = subst_Apply(rho, subst_Apply(tau, term_Copy(s)));
	  t_sigma = subst_Apply(rho, subst_Apply(tau, term_Copy(t)));
	  kboResult = kbo_Compare(s_sigma, t_sigma);
	  if (kboResult == kbo_SmallerThan() || kboResult == kbo_Equal())
	    checkPassed = FALSE;
	}

	if (checkPassed) {
	  v_sigma = subst_Apply(rho, term_Copy(v_tau));
	  v2_sigma = subst_Apply(rho, term_Copy(pLeft2));
	  kboResult = kbo_Compare(v2_sigma, v_sigma);

	  if (kboResult != kbo_GreaterThan() && kboResult != kbo_Equal() &&
	      inf_MParamodLitsMax(C1, i, C2, j, tau, rho)) {

	    result = list_Nconc(inf_ApplyMParamod(C1,C2,i,j,k,u_tau,v,s2,
						  t,v2_sigma,tau,rho),
				result);
	  }
	  term_Delete(v_sigma);
	  term_Delete(v2_sigma);
	}
	/* Now cleanup */
	if (s_sigma != NULL) {      /* Also t_sigma != NULL */
	  term_Delete(s_sigma);
	  term_Delete(t_sigma);
	}
	subst_Delete(rho);
      }

      cont_Reset();

      term_Delete(pLeft2);
      term_Delete(pRight2);
    }   /* k != j */
  }     /* for k */
  
  return result;
}  
  

static LIST inf_LitMParamod(CLAUSE Clause, int i, BOOL Turn,
			    SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause with a strict maximal equality literal at
           position <i>, a boolean value and a shared index.
	   If <Turn> is TRUE, the left and right term of the
	   equality are exchanged.
  RETURNS: A list of clauses derivable from the given clause
           by merging paramodulation.
***************************************************************/
{
  LIST    result, unifiers, literals;
  LITERAL actLit;
  TERM    s, t, help;

  actLit = clause_GetLiteral(Clause, i);
  s      = term_FirstArgument(clause_LiteralSignedAtom(actLit));
  t      = term_SecondArgument(clause_LiteralSignedAtom(actLit));
  if (Turn) {
    /* Exchange s and t */
    help = s;
    s    = t;
    t    = help;
  }
  result = list_Nil();

#ifdef _INF_DEBUGMPARAMOD_
  fputs("Given: ", stdout); clause_LiteralPrint(actLit); putchar('\n');
#endif

  for (unifiers = st_GetUnifier(cont_LeftContext(),
				sharing_Index(ShIndex),
				cont_RightContext(),
				s);
       !list_Empty(unifiers);
       unifiers= list_Pop(unifiers)) {
    TERM s2 = (TERM) list_Car(unifiers);     /* Unifiable with s */

    if (!term_IsVariable(s2) && !term_IsAtom(s2)) {
      for (literals = sharing_GetDataList(s2, ShIndex);
	   !list_Empty(literals);
	   literals = list_Pop(literals)) {
	LITERAL partnerLit    = (LITERAL) list_Car(literals);  /* u = v[s'] */
	CLAUSE  partnerClause = clause_LiteralOwningClause(partnerLit);
	TERM    partnerAtom   = clause_LiteralAtom(partnerLit);
	int     pli           = clause_LiteralGetIndex(partnerLit);

	if (!clause_GetFlag(partnerClause, CLAUSESELECT) &&
	    clause_GetFlag(partnerClause, WORKEDOFF) && 
	    clause_LiteralGetFlag(partnerLit, STRICTMAXIMAL) &&
	    clause_LiteralIsPositive(partnerLit) &&   /* succedent literal */
	    clause_LiteralIsEquality(partnerLit) &&
	    clause_HasSolvedConstraint(partnerClause)) {
	  TERM partnerLeft    = term_FirstArgument(partnerAtom);
	  TERM partnerRight   = term_SecondArgument(partnerAtom);
	  BOOL inPartnerRight = term_HasPointerSubterm(partnerRight, s2);

#ifdef _INF_DEBUGMPARAMOD_
	  fputs("s = ", stdout); term_PrintPrefix(s);
	  fputs(", s' = ", stdout); term_PrintPrefix(s2);
	  fputs(" in ", stdout); term_PrintPrefix(partnerAtom);
#endif

	  if (!clause_LiteralIsOrientedEquality(partnerLit) || inPartnerRight){
	    /* Don't do this if u=v is oriented and s2 is not a subterm of v */
	    TERM       newPLeft, newPRight;
	    SUBST      tau;
	    SYMBOL     partnerMaxVar;
	    kbo_RESULT kboResult;

	    partnerMaxVar = clause_MaxVar(partnerClause);
	    clause_RenameVarsBiggerThan(Clause, partnerMaxVar);
	    cont_Check();
	    unify_UnifyNoOC(cont_LeftContext(), s, cont_RightContext(), s2);
	    subst_ExtractUnifierCom(cont_LeftContext(), &tau);
	    cont_Reset();

#ifdef _INF_DEBUGMPARAMOD_
	    fputs(" tau = ",stdout);  subst_Print(tau); putchar('\n');
#endif
	    newPLeft  = subst_Apply(tau, term_Copy(partnerLeft));
	    newPRight = subst_Apply(tau, term_Copy(partnerRight));
	    if (clause_LiteralIsOrientedEquality(partnerLit))
	      kboResult = kbo_GreaterThan();
	    else
	      kboResult = kbo_Compare(newPLeft, newPRight);

	    if (inPartnerRight && kbo_IsGreaterThan(kboResult)) {
	      /* Take a look at right side */
#ifdef _INF_DEBUGMPARAMOD_
	      fputs("u = ", stdout);      term_PrintPrefix(partnerLeft);
	      fputs(" v = ", stdout);     term_PrintPrefix(partnerRight);
	      putchar('\n');
#endif
	      result = list_Nconc(inf_Lit2MParamod(Clause,partnerClause,i,pli,
						   s,t,s2,partnerLeft,
						   partnerRight, newPLeft,
						   newPRight,tau),
				  result);
	    }
	    if (kbo_IsSmallerThan(kboResult) &&
		(!inPartnerRight || term_HasPointerSubterm(partnerLeft, s2))) {
	      /* If s2 is not in partnerRight, it MUST be in partnerLeft,
		 else really do the test */
	      /* Take a look at left side */
#ifdef _INF_DEBUGMPARAMOD_
	      fputs("u = ", stdout);  term_PrintPrefix(partnerRight);
	      fputs(" v = ", stdout); term_PrintPrefix(partnerLeft);
	      putchar('\n');
#endif
	      result = list_Nconc(inf_Lit2MParamod(Clause,partnerClause,i,pli,
						   s,t,s2,partnerRight,
						   partnerLeft,newPRight,
						   newPLeft,tau),
				  result);
	    }

	    term_Delete(newPLeft);
	    term_Delete(newPRight);
	    subst_Delete(tau);
	    clause_Normalize(Clause);
	  }
#ifdef _INF_DEBUGMPARAMOD_
	  else
	    puts(", Unifier only in left term.");
#endif
	}
      }  /* for all Literals containing s2 */
    }    /* if s2 isn't a variable */
  }      /* for all unifiers s2 */

#ifdef _INF_DEBUGMPARAMOD_
  putchar('\n');
#endif

  return result;
}


static LIST inf_MParamodLitToGiven(CLAUSE Clause, int j, BOOL Turn,
				   SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause with a strict maximal equality literal at
           position <j>, a boolean value and a shared index.
	   If <Turn> is TRUE, the left and right term of the
	   equality are exchanged.
  RETURNS: A list of clauses derivable from the given clause
           by merging paramodulation with this literal as second
	   literal of the rule.
***************************************************************/
{
  LIST    result, unifiers, superterms, literals;
  LITERAL actLit;
  TERM    u, v;
  int     bottom;

  actLit = clause_GetLiteral(Clause, j);
  u  = term_FirstArgument(clause_LiteralSignedAtom(actLit));
  v = term_SecondArgument(clause_LiteralSignedAtom(actLit));
  if (Turn) {
    /* Exchange s and t */
    TERM help = u;
    u  = v;
    v = help;
  }
  result = list_Nil();
  bottom = stack_Bottom();

  sharing_PushReverseOnStack(v);  /* Without variables !*/

#ifdef _INF_DEBUGMPARAMOD_
  fputs("ToGiven: ", stdout); clause_LiteralPrint(actLit); putchar('\n');
#endif

  while (!stack_Empty(bottom)) {
    TERM s2 = (TERM) stack_PopResult();

    for (unifiers = st_GetUnifier(cont_LeftContext(),
				  sharing_Index(ShIndex),
				  cont_RightContext(),
				  s2);
	 !list_Empty(unifiers);
	 unifiers = list_Pop(unifiers)) {
      TERM s = (TERM) list_Car(unifiers);

      for (superterms = term_SupertermList(s);
	   !list_Empty(superterms);
	   superterms = list_Cdr(superterms)) {
	TERM partnerAtom = (TERM) list_Car(superterms);

	if (fol_IsEquality(partnerAtom)) {
	  for (literals = sharing_NAtomDataList(partnerAtom);
	       !list_Empty(literals);
	       literals = list_Cdr(literals)) {
	    LITERAL partnerLit    = (LITERAL) list_Car(literals);
	    CLAUSE  partnerClause = clause_LiteralOwningClause(partnerLit);
	    int     i             = clause_LiteralGetIndex(partnerLit);

	    if (!clause_GetFlag(partnerClause,CLAUSESELECT) &&
		clause_LiteralGetFlag(partnerLit,STRICTMAXIMAL) &&
		clause_LiteralIsPositive(partnerLit) &&
		(s == term_FirstArgument(partnerAtom) ||
		 !clause_LiteralIsOrientedEquality(partnerLit)) &&
		clause_GetFlag(partnerClause,WORKEDOFF) &&
		clause_HasSolvedConstraint(partnerClause) /*&&
		clause_Number(partnerClause) != clause_Number(Clause)*/) {
	      SYMBOL partnerMaxVar;
	      SUBST  tau;
	      TERM   u_tau, v_tau;
	      BOOL   checkPassed;

	      partnerMaxVar = clause_MaxVar(partnerClause);
	      clause_RenameVarsBiggerThan(Clause, partnerMaxVar);
	      cont_Check();
	      unify_UnifyNoOC(cont_LeftContext(), s, cont_RightContext(), s2);
	      subst_ExtractUnifierCom(cont_LeftContext(), &tau);
	      cont_Reset();

#ifdef _INF_DEBUGMPARAMOD_
	      fputs("s' = ",stdout);  term_PrintPrefix(s2);
	      fputs(" s = ",stdout);   term_PrintPrefix(s);
	      fputs(" in ", stdout);   term_PrintPrefix(partnerAtom);
	      fputs(" tau = ",stdout); subst_Print(tau);
	      putchar('\n');
#endif

	      u_tau = v_tau = (TERM) NULL;
	      checkPassed = TRUE;

	      /* u_tau must be greater than v_tau */
	      if (!clause_LiteralIsOrientedEquality(actLit)) {
		u_tau = subst_Apply(tau, term_Copy(u));
		v_tau = subst_Apply(tau, term_Copy(v));
		if (kbo_Compare(u_tau, v_tau) != kbo_GreaterThan())
		  checkPassed = FALSE;
	      }

	      if (checkPassed) {
		/* u_tau > v_tau */
		TERM t;

		if (s == term_FirstArgument(partnerAtom))
		  t = term_SecondArgument(partnerAtom);
		else
		  t = term_FirstArgument(partnerAtom);
		if (u_tau == (TERM)NULL) {
		  u_tau = subst_Apply(tau, term_Copy(u));
		  v_tau = subst_Apply(tau, term_Copy(v));
		}

		result = list_Nconc(inf_Lit2MParamod(partnerClause,Clause,i,j,
						     s,t,s2,u,v,u_tau,v_tau,
						     tau),
				    result);
	      }

	      /* Now cleanup */
	      if (u_tau != (TERM)NULL) {
		term_Delete(u_tau);
		term_Delete(v_tau);
	      }
	      subst_Delete(tau);
	      clause_Normalize(Clause);
	    }
	  }
	}   /* partnerAtom is equality */
      }
    }
  }

#ifdef _INF_DEBUGMPARAMOD_
  putchar('\n');
#endif

  return result;
}


LIST inf_MergingParamodulation(CLAUSE GivenClause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause and a shard index used to find unifiers.
  RETURNS: A list of clauses derivable from the given clause
           by merging paramodulation.
  MEMORY:  Memory is allocated for the list and the clauses.
***************************************************************/
{
  LIST result = list_Nil();

#if DEBUG
  if (!clause_IsClause(GivenClause)) {
    fputs("\n\nIllegal Input in Paramodulation\n", stderr);
    misc_Error();
  }
#endif

  if (!clause_GetFlag(GivenClause, CLAUSESELECT) &&
      !clause_HasEmptySuccedent(GivenClause) &&
      clause_HasSolvedConstraint(GivenClause)) {
    CLAUSE copy = clause_Copy(GivenClause);
    int    last = clause_LastSuccedentLitIndex(copy);
    int    i;

    for (i = clause_FirstSuccedentLitIndex(copy); i <= last; i++) {
      LITERAL actLit = clause_GetLiteral(copy, i);
      TERM    atom   = clause_LiteralSignedAtom(actLit);

      if (clause_LiteralGetFlag(actLit, STRICTMAXIMAL) &&
	  fol_IsEquality(atom)) {

	result = list_Nconc(inf_LitMParamod(copy, i, FALSE, ShIndex), result);
	/* Assume GivenClause is the second clause of the rule */
	result = list_Nconc(inf_MParamodLitToGiven(copy, i, FALSE, ShIndex),
			    result);
	
	if (!clause_LiteralIsOrientedEquality(actLit)) {
	    /* First check rule with left and right side exchanged */
	  result = list_Nconc(inf_LitMParamod(copy, i, TRUE, ShIndex),
			      result);
	  /* Now assume GivenClause is the second clause of the rule */
	  /* Check with sides exchanged */
	  result = list_Nconc(inf_MParamodLitToGiven(copy, i, TRUE, ShIndex),
			      result);
	}
#ifdef _INF_DEBUGMPARAMOD_
	putchar('\n');
#endif
      }
    }  /* for */
    clause_Delete(copy);
  }
  return result;
}


static CLAUSE inf_ApplyGenRes(LITERAL PosLit, LITERAL NegLit, SUBST SubstTermS, SUBST SubstPartnerTermS)
/**************************************************************
  INPUT:   A clause to use for Resolution, the index of a 
           positive non-equality literal, a unifiable literal
           and the substitutions for the terms to unify.
  RETURNS: A clause derivable from the literals owning
           clause by Resolution wrt. the Index.
  MEMORY:  Memory for the new clause is allocated.
***************************************************************/
{
  CLAUSE NewClause, GivenClause, PartnerClause;
  int    i,j,lc,la,ls,pi,pls,pla,plc,help,ConNeg,AntNeg; /* p=Partner,l=last */

  /*  fputs("\n GeneralResolution is Applied !!! ", stdout); */

  PartnerClause = clause_LiteralOwningClause(NegLit);
  GivenClause   = clause_LiteralOwningClause(PosLit);

  pls = clause_LastSuccedentLitIndex(PartnerClause);
  pla = clause_LastAntecedentLitIndex(PartnerClause);
  plc = clause_LastConstraintLitIndex(PartnerClause);

  pi  = clause_LiteralGetIndex(NegLit);

  ls  = clause_LastSuccedentLitIndex(GivenClause);
  la  = clause_LastAntecedentLitIndex(GivenClause);
  lc  = clause_LastConstraintLitIndex(GivenClause);

  i   = clause_LiteralGetIndex(PosLit);

  if (pi<=plc) {
    ConNeg = 1;
    AntNeg = 0;
  }
  else {
    ConNeg = 0;
    AntNeg = 1;
  }

  NewClause = clause_CreateBody((clause_Length(GivenClause) -1) +
				clause_Length(PartnerClause) -1);

  clause_SetNumOfConsLits(NewClause,
			    (clause_NumOfConsLits(GivenClause) +
			     (clause_NumOfConsLits(PartnerClause)-ConNeg)));
  
  clause_SetNumOfAnteLits(NewClause,
			    (clause_NumOfAnteLits(GivenClause) +
			     (clause_NumOfAnteLits(PartnerClause)-AntNeg)));
  
  clause_SetNumOfSuccLits(NewClause,
			  ((clause_NumOfSuccLits(GivenClause) -1)+
			   clause_NumOfSuccLits(PartnerClause)));


  /* First set the literals from the GivenClause : */

  for (j = clause_FirstLitIndex(); j <= lc; j++) {
    clause_SetLiteral(NewClause, j, 
      clause_LiteralCreate(subst_Apply(SubstTermS,
	term_Copy(clause_GetLiteralTerm(GivenClause, j))),NewClause));
  }

  /* help = number of literals to leave empty */
  help = clause_NumOfConsLits(PartnerClause)-ConNeg;
    
  for ( ; j <= la; j++) {
    clause_SetLiteral(NewClause, (j + help), 
      clause_LiteralCreate(subst_Apply(SubstTermS,
	term_Copy(clause_GetLiteralTerm(GivenClause, j))),NewClause));
  }

  /* help = number of literals to leave empty */
  help += clause_NumOfAnteLits(PartnerClause)-AntNeg;
    
  

  for ( ; j <= ls; j++) {
    if (j != i) {
      /* The ActLit isn't copied! */
      clause_SetLiteral(NewClause, (j + help), 
	clause_LiteralCreate(subst_Apply(SubstTermS,
	  term_Copy(clause_GetLiteralTerm(GivenClause, j))),NewClause));

    } else {
      /*the index has to be decreased to avoid an empty literal! */
      help--;
    }
  }

  /* Now we consider the PartnerClause : */

  /* help = number of already set constraint (GivenClause-) literals */
  help = clause_NumOfConsLits(GivenClause);

  for (j = clause_FirstLitIndex(); j <= plc; j++) {
    if (j != pi) {
    clause_SetLiteral(NewClause, (j + help), 
      clause_LiteralCreate(subst_Apply(SubstPartnerTermS,
	term_Copy(clause_GetLiteralTerm(PartnerClause, j))),NewClause));
    } else {
      help--;
    }
  }

  /* help = number of already set constraint and antecedent Given-literals */
  help += clause_NumOfAnteLits(GivenClause);

  for ( ; j <= pla; j++) {

    if (j != pi) {
      /* The NegLit isn't copied! */
      clause_SetLiteral(NewClause, (j + help), 
	clause_LiteralCreate(subst_Apply(SubstPartnerTermS,
	  term_Copy(clause_GetLiteralTerm(PartnerClause, j))),NewClause));

    } else {
      /* The index has to be shifted as above.  */
      help--;
    }
  }

  /* help = number of already set (GivenClause-) literals */
  help += clause_NumOfSuccLits(GivenClause) - 1;

  for ( ; j <= pls; j++) {
    clause_SetLiteral(NewClause, (j + help), 
      clause_LiteralCreate(subst_Apply(SubstPartnerTermS,
	term_Copy(clause_GetLiteralTerm(PartnerClause, j))),NewClause));
  } /* end of NewClause creation (last for loop). */


  clause_SetDataFromParents(NewClause, PartnerClause, pi, GivenClause, i);
  clause_SetFromGeneralResolution(NewClause);

  /* fputs("\nNewClause aus ApplyGenRes: ",stdout); clause_Print(NewClause);
     fputs("\naus (Given)  : ",stdout); clause_PrintVerbose(GivenClause);
     fputs("\nund (Partner): ",stdout); clause_PrintVerbose(PartnerClause); DBG */

  return(NewClause);
}


LIST inf_GeneralResolution(CLAUSE GivenClause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause and an Index, usually the WorkedOffIndex.
  RETURNS: A list of clauses derivable from the Givenclause by 
           GeneralResolution wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  CLAUSE GivenCopy;
  LIST Result;
  LITERAL ActLit;
  TERM Atom;
  int i,n;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    fputs("\n\nIllegal input in inf_GeneralResolution\n", stderr);
    misc_Error();
  }
#endif

  cont_Check();

  Result = list_Nil();

  if (clause_HasSolvedConstraint(GivenClause)) {
    /*fputs("\n GeneralResolution is tried !!! ",stdout); DBG */

    GivenCopy = clause_Copy(GivenClause);
    
    if (clause_GetFlag(GivenCopy,CLAUSESELECT))
      n = clause_LastAntecedentLitIndex(GivenCopy) + 1;
    else
      n = clause_Length(GivenCopy);
    
    for(i=clause_FirstAntecedentLitIndex(GivenCopy); i < n; i++) {
      
      ActLit = clause_GetLiteral(GivenCopy, i);
      Atom   = clause_LiteralAtom(ActLit);
      
      if (((clause_LiteralIsMaximal(ActLit) &&
	    !clause_GetFlag(GivenCopy,CLAUSESELECT)) ||
	   clause_LiteralGetFlag(ActLit,LITSELECT)) &&
	  !fol_IsEquality(Atom) &&
	  (clause_LiteralIsFromAntecedent(ActLit) || 
	   clause_LiteralGetFlag(ActLit,STRICTMAXIMAL))) {
	LIST TermList,TermListDel;
	
	/* fputs("\nWir versuchen Resolution mit dem 'maximal literal' : ",
                 stdout);
	   term_Print(Atom); fflush(stdout);
	   DBG */
	
	TermListDel = st_GetUnifier(cont_LeftContext(),
				    sharing_Index(ShIndex),
				    cont_RightContext(),
				    Atom);
	
	for (TermList = TermListDel;
	     !list_Empty(TermList);
	     TermList = list_Cdr(TermList)){
	  
	  LIST LitList;
	  TERM PartnerAtom;
	  
	  PartnerAtom = list_First(TermList);
	  
	  if (!term_IsVariable(PartnerAtom)){
	    
	    LITERAL PartnerLit;
	    int     j;
	    CLAUSE  PartnerClause;
	    
	    for(LitList =  sharing_NAtomDataList(PartnerAtom); 
		!list_Empty(LitList);  LitList = list_Cdr(LitList)) {
	      
	      PartnerLit    = list_Car(LitList);
	      j             = clause_LiteralGetIndex(PartnerLit);
	      PartnerClause = clause_LiteralOwningClause(PartnerLit);
	      
	      
	      if(clause_GetFlag(PartnerClause,WORKEDOFF) &&
		 (!clause_GetFlag(PartnerClause,CLAUSESELECT) || 
		  clause_LiteralIsPositive(ActLit)) &&
		 clause_HasSolvedConstraint(PartnerClause) &&
		 ((clause_LiteralIsMaximal(PartnerLit) && 
		   !clause_GetFlag(PartnerClause,CLAUSESELECT)) ||
		  clause_LiteralGetFlag(PartnerLit,LITSELECT)) &&
		 clause_LiteralsAreComplementary(PartnerLit,ActLit) &&
		 (clause_LiteralIsFromAntecedent(PartnerLit) || 
		  clause_LiteralGetFlag(PartnerLit,STRICTMAXIMAL))) {
		
		SUBST SubstAtom,SubstPartnerAtom;
		SYMBOL MaxVar;
		
		
		/*fputs("\nPartner : ",stdout);
		  clause_PrintVerbose(PartnerClause);*/
		
		
		MaxVar = clause_MaxVar(PartnerClause);
		clause_RenameVarsBiggerThan(GivenCopy, MaxVar);
		Atom = clause_LiteralAtom(ActLit);
		cont_Check();
		unify_UnifyNoOC(cont_LeftContext(), Atom, cont_RightContext(), PartnerAtom);
		subst_ExtractUnifier(cont_LeftContext(),
				     &SubstAtom,
				     cont_RightContext(),
				     &SubstPartnerAtom);
		cont_Reset();

		if (inf_InferenceLitsMax(GivenCopy,i,SubstAtom,PartnerClause,
					 j,SubstPartnerAtom))
		  if (clause_LiteralIsNegative(PartnerLit))
		    Result =
		      list_Cons(inf_ApplyGenRes(ActLit, PartnerLit, 
						SubstAtom, SubstPartnerAtom),
				Result);
		
		  else
		    Result =
		      list_Cons(inf_ApplyGenRes(PartnerLit, ActLit, 
						SubstPartnerAtom, SubstAtom),
				Result);
		
		clause_Normalize(GivenCopy);
		/* fputs("\nAus GeneralResolution mit : ",stdout);
		   clause_Print(GivenClause); fputs(" und ",stdout);
		   clause_Print(PartnerClause); 
		   fputs("\nunter den Substitutionen: ",stdout);
		   subst_Print(SubstAtom);
		   fputs(" und ",stdout);  subst_Print(SubstPartnerAtom);
		   fputs("\nErgibt sich: ",stdout); clause_Print(list_Car(Result));
		   DBG */
		subst_Delete(SubstAtom);
		subst_Delete(SubstPartnerAtom);
	      } /* end of if 'we didn't find the GivenClause'
		   and 'clause reductive */
	      /* for PartnerLit'. */
	    } /* end of for (LitListDel = LitList = sharing_GetDataList ...). */
	    
	  } /* end of rule application constraint:
	       if (!term_IsVariable(PartnerAtom)). */
	  
	} /* end of for (TermListDel = TermList = st_GetUnifier...). */
	list_Delete(TermListDel);
	
      } /* end of if (clause_LiteralIsMaximal(ActLit)). */
    } /* end of for 'all succedent literals'. */
    
    clause_Delete(GivenCopy);
    
    /*if list_Empty(Result)
      fputs("\n GeneralResolution wasn't possible! ",stdout);
      else {
      fputs("\n GeneralResolution yielded : \n",stdout);
      clause_ListPrint(Result);
      }*/
    
    cont_Check();
  }
  
  return Result;
}


LIST inf_UnitResolution(CLAUSE GivenClause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause and an Index, usually the WorkedOffIndex.
  RETURNS: A list of clauses derivable from the Givenclause by 
           Unit Resolution (one parent must be a unit) wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  CLAUSE   GivenCopy;
  LIST     Result;
  LITERAL  ActLit;
  TERM     Atom;
  BOOL     GivenIsUnit;
  int      i,n;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    fputs("\n\nIllegal input in inf_UnitResolution\n", stderr);
    misc_Error();
  }
#endif

  cont_Check();

  Result = list_Nil();

  if (clause_HasSolvedConstraint(GivenClause)) {

    GivenCopy   = clause_Copy(GivenClause);
    GivenIsUnit = (clause_Length(GivenCopy) == 1);
    
    if (clause_GetFlag(GivenCopy,CLAUSESELECT))
      n = clause_LastAntecedentLitIndex(GivenCopy) + 1;
    else
      n = clause_Length(GivenCopy);
    
    for(i=clause_FirstAntecedentLitIndex(GivenCopy); i < n; i++) {
      
      ActLit = clause_GetLiteral(GivenCopy, i);
      Atom   = clause_LiteralAtom(ActLit);
      
      if (((clause_LiteralIsMaximal(ActLit) &&
	    !clause_GetFlag(GivenCopy,CLAUSESELECT)) ||
	   clause_LiteralGetFlag(ActLit,LITSELECT)) &&
	  !fol_IsEquality(Atom) &&
	  (clause_LiteralIsFromAntecedent(ActLit) || 
	   clause_LiteralGetFlag(ActLit,STRICTMAXIMAL))) {
	LIST TermList,TermListDel;
	
	TermListDel = st_GetUnifier(cont_LeftContext(),
				    sharing_Index(ShIndex),
				    cont_RightContext(),
				    Atom);
	
	for (TermList = TermListDel;!list_Empty(TermList);TermList = list_Cdr(TermList)) {
	  
	  LIST LitList;
	  TERM PartnerAtom;
	  
	  PartnerAtom = list_First(TermList);
	  
	  if (!term_IsVariable(PartnerAtom)) {
	    
	    LITERAL PartnerLit;
	    int     j;
	    CLAUSE  PartnerClause;
	    
	    for(LitList =  sharing_NAtomDataList(PartnerAtom); 
		!list_Empty(LitList);  LitList = list_Cdr(LitList)) {
	      
	      PartnerLit    = list_Car(LitList);
	      j             = clause_LiteralGetIndex(PartnerLit);
	      PartnerClause = clause_LiteralOwningClause(PartnerLit);
	      
	      
	      if((GivenIsUnit || clause_Length(PartnerClause) == 1) &&
		 clause_GetFlag(PartnerClause,WORKEDOFF) &&
		 (!clause_GetFlag(PartnerClause,CLAUSESELECT) || 
		  clause_LiteralIsPositive(ActLit)) &&
		 clause_HasSolvedConstraint(PartnerClause) &&
		 ((clause_LiteralIsMaximal(PartnerLit) && 
		   !clause_GetFlag(PartnerClause,CLAUSESELECT)) ||
		  clause_LiteralGetFlag(PartnerLit,LITSELECT)) &&
		 clause_LiteralsAreComplementary(PartnerLit,ActLit) &&
		 (clause_LiteralIsFromAntecedent(PartnerLit) || 
		  clause_LiteralGetFlag(PartnerLit,STRICTMAXIMAL))) {
		
		SUBST SubstAtom,SubstPartnerAtom;
		SYMBOL MaxVar;
			       		
		MaxVar = clause_MaxVar(PartnerClause);
		clause_RenameVarsBiggerThan(GivenCopy, MaxVar);
		Atom = clause_LiteralAtom(ActLit);
		cont_Check();
		unify_UnifyNoOC(cont_LeftContext(), Atom, cont_RightContext(), PartnerAtom);
		subst_ExtractUnifier(cont_LeftContext(),
				     &SubstAtom,
				     cont_RightContext(),
				     &SubstPartnerAtom);
		cont_Reset();

		if (inf_InferenceLitsMax(GivenCopy,i,SubstAtom,PartnerClause,
					 j,SubstPartnerAtom))
		  if (clause_LiteralIsNegative(PartnerLit))
		    Result =
		      list_Cons(inf_ApplyGenRes(ActLit, PartnerLit, 
						SubstAtom, SubstPartnerAtom),
				Result);
		
		  else
		    Result =
		      list_Cons(inf_ApplyGenRes(PartnerLit, ActLit, 
						SubstPartnerAtom, SubstAtom),
				Result);
		
		clause_Normalize(GivenCopy);
		/* fputs("\nAus UnitResolution mit : ",stdout);
		   clause_Print(GivenClause); fputs(" und ",stdout);
		   clause_Print(PartnerClause); 
		   fputs("\nunter den Substitutionen: ",stdout);
		   subst_Print(SubstAtom);
		   fputs(" und ",stdout);  subst_Print(SubstPartnerAtom);
		   fputs("\nErgibt sich: ",stdout); clause_Print(list_Car(Result));
		   DBG */
		subst_Delete(SubstAtom);
		subst_Delete(SubstPartnerAtom);
	      } /* end of if 'we didn't find the GivenClause'
		   and 'clause reductive */
	      /* for PartnerLit'. */
	    } /* end of for (LitListDel = LitList = sharing_GetDataList ...). */
	    
	  } /* end of rule application constraint:
	       if (!term_IsVariable(PartnerAtom)). */
	  
	} /* end of for (TermListDel = TermList = st_GetUnifier...). */
	list_Delete(TermListDel);
	
      } /* end of if (clause_LiteralIsMaximal(ActLit)). */
    } /* end of for 'all succedent literals'. */
    
    clause_Delete(GivenCopy);
    
    /*if list_Empty(Result)
      fputs("\n GeneralResolution wasn't possible! ",stdout);
      else {
      fputs("\n GeneralResolution yielded : \n",stdout);
      clause_ListPrint(Result);
      }*/
    
    cont_Check();
  }
  
  return Result;
}

LIST inf_BoundedDepthUnitResolution(CLAUSE GivenClause, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause and an Index, usually the WorkedOffIndex.
  RETURNS: A list of clauses derivable from the Givenclause by 
           GeneralResolution wrt. the Index.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
     /* GivenClause is always a CONCLAUSE */
{
  CLAUSE  GivenCopy;
  LIST    Result;
  LITERAL ActLit;
  TERM    Atom;
  int     i,n,depth;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    printf("\n\nIllegal input in inf_BoundedDepthUnitResolution\n");
    misc_Error();
  }
  cont_Check();
#endif

  Result    = list_Nil();
  GivenCopy = clause_Copy(GivenClause);
  n         = clause_Length(GivenCopy);
  depth     = clause_ComputeTermDepth(GivenCopy);

  for(i=clause_FirstLitIndex(); i < n; i++) {
    LIST TermList;
    LIST TermListDel;

    ActLit      = clause_GetLiteral(GivenCopy, i);
    Atom        = clause_LiteralAtom(ActLit);
    TermListDel = st_GetUnifier(cont_LeftContext(), sharing_Index(ShIndex), cont_RightContext(), Atom);
    for (TermList = TermListDel; !list_Empty(TermList);TermList = list_Cdr(TermList)){
      LIST LitList;
      TERM PartnerAtom;
      PartnerAtom = list_First(TermList);
      if (!term_IsVariable(PartnerAtom)){
	LITERAL PartnerLit;
	int     j;
	CLAUSE  PartnerClause;
	for(LitList =  sharing_NAtomDataList(PartnerAtom);!list_Empty(LitList);LitList = list_Cdr(LitList)) {
	  PartnerLit    = list_Car(LitList);
	  j             = clause_LiteralGetIndex(PartnerLit);
	  PartnerClause = clause_LiteralOwningClause(PartnerLit);

	  if(clause_LiteralsAreComplementary(PartnerLit,ActLit)
	     && ((clause_Length(GivenCopy) == 1) || (clause_Length(PartnerClause) == 1))
	     && (clause_GetFlag(GivenCopy,CONCLAUSE) || clause_GetFlag(PartnerClause,CONCLAUSE))) {
	    SUBST  SubstAtom,SubstPartnerAtom;
	    SYMBOL MaxVar;
	    int    maxdepth;
	    CLAUSE Resolvent;

	    maxdepth = misc_Max(depth,clause_ComputeTermDepth(PartnerClause));
	    MaxVar = clause_SearchMaxVar(PartnerClause);
	    clause_RenameVarsBiggerThan(GivenCopy, MaxVar);
	    Atom = clause_LiteralAtom(ActLit);
	    cont_Check();
	    unify_UnifyNoOC(cont_LeftContext(), Atom, cont_RightContext(), PartnerAtom);
	    subst_ExtractUnifier(cont_LeftContext(), &SubstAtom, cont_RightContext(), &SubstPartnerAtom);
	    cont_Reset();
	    if (clause_LiteralIsNegative(PartnerLit))
	      Resolvent = inf_ApplyGenRes(ActLit, PartnerLit, SubstAtom, SubstPartnerAtom);
	    else
	      Resolvent = inf_ApplyGenRes(PartnerLit, ActLit, SubstPartnerAtom, SubstAtom);
	    /*printf("\n\n BUR: "); clause_Print(Resolvent);printf("\n");*/
	    if (clause_Depth(Resolvent)>maxdepth)
	      clause_Delete(Resolvent);
	    else {
	      Result = list_Cons(Resolvent,Result);
	      /*printf("\n\t KEPT");*/
	    }
	    clause_Normalize(GivenCopy);
	    subst_Delete(SubstAtom);
	    subst_Delete(SubstPartnerAtom);
	  } 
	}
      }
    }
    list_Delete(TermListDel);
  }

  clause_Delete(GivenCopy);

  /*if list_Empty(Result)
    printf("\n GeneralResolution wasn't possible! ");
    else {
    printf("\n GeneralResolution yielded : \n");
    clause_ListPrint(Result);
    }*/

#if DEBUG
  cont_Check();
#endif
  return(Result);
}

static CLAUSE inf_ApplyGeneralFactoring(CLAUSE Clause, NAT i, NAT j,
					SUBST Subst)
/**************************************************************
  INPUT:   A clause an index in the clause and a substitution.
  RETURNS: A new clause obtained from <Clause> by applying <Subst>
           and deleting literal <j> keeping literal <i>
***************************************************************/
{
  CLAUSE NewClause;
  int    k,c,a,s,shift;

  NewClause = clause_CreateBody(clause_Length(Clause)-1);

  c = clause_LastConstraintLitIndex(Clause);
  a = clause_LastAntecedentLitIndex(Clause);
  s = clause_LastSuccedentLitIndex(Clause);


  clause_SetNumOfConsLits(NewClause, clause_NumOfConsLits(Clause));
  clause_SetNumOfAnteLits(NewClause, clause_NumOfAnteLits(Clause));
  clause_SetNumOfSuccLits(NewClause, (clause_NumOfSuccLits(Clause) - 1));


  for (k = clause_FirstLitIndex(); k <= c; k++) {
    clause_SetLiteral(NewClause, k, 
      clause_LiteralCreate(subst_Apply(Subst,
	term_Copy(clause_GetLiteralTerm(Clause, k))),NewClause));
  }

  for ( ; k <= a; k++) {
    clause_SetLiteral(NewClause, k, 
      clause_LiteralCreate(subst_Apply(Subst,
	term_Copy(clause_GetLiteralTerm(Clause, k))),NewClause));
  }

  shift = 0; /* start without shifting the index. */

  for ( ; k <= s; k++) {
    if (k == j)
      shift = 1;
    else
      clause_SetLiteral(NewClause, (k - shift),
	clause_LiteralCreate(subst_Apply(Subst,
	                       term_Copy(clause_GetLiteralTerm(Clause, k))),
			     NewClause));
  }

  clause_SetDataFromFather(NewClause, Clause, j);
  clause_SetFromGeneralFactoring(NewClause);

  clause_AddParentClause(NewClause, clause_Number(Clause));
  clause_AddParentLiteral(NewClause, i);

  return NewClause;
}


LIST inf_GeneralFactoring(CLAUSE GivenClause)
/**************************************************************
  INPUT:   A clause.
  RETURNS: A list of clauses derivable from 'GivenClause' by GF.
***************************************************************/
{

  LIST Result;
  LITERAL ActLit;
  SYMBOL MinimumVariable;
  int i,j,last;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    fputs("\n\nIllegal input in GeneralFactoring\n", stderr);
    misc_Error();
  }
#endif

  cont_Check();

  Result = list_Nil();
  last = clause_LastSuccedentLitIndex(GivenClause);
  MinimumVariable = clause_MaxVar(GivenClause);

  if (clause_HasSolvedConstraint(GivenClause) &&
      !clause_GetFlag(GivenClause,CLAUSESELECT)) {

    for(i = clause_FirstSuccedentLitIndex(GivenClause); i <= last; i++) {

      ActLit = clause_GetLiteral(GivenClause, i);

      if (clause_LiteralIsMaximal(ActLit) &&
	  (flag_GetFlagValue(flag_IMPM) || !fol_IsEquality(clause_LiteralSignedAtom(ActLit)))) {
	TERM Atom;
	LITERAL PartnerLit;

	Atom = clause_LiteralAtom(ActLit);

	for (j = clause_FirstSuccedentLitIndex(GivenClause);j <= last; j++) {

	  if (i != j) {

	    PartnerLit = clause_GetLiteral(GivenClause, j);

	    if (!clause_LiteralIsMaximal(PartnerLit) || j>i){

	      TERM PartnerAtom;

	      PartnerAtom = clause_LiteralAtom(PartnerLit);

	      cont_Check();

	      if (unify_UnifyCom(cont_LeftContext(), Atom, PartnerAtom)) {
		SUBST mgu;

		subst_ExtractUnifierCom(cont_LeftContext(), &mgu);

		if (inf_FactorLitMax(GivenClause,i,j,mgu))
		  Result = list_Cons(inf_ApplyGeneralFactoring(GivenClause,i,
							       j,mgu),
				     Result);

		subst_Delete(mgu);
	      }

	      cont_Reset();

	      if (fol_IsEquality(Atom) &&
		  symbol_Equal(term_TopSymbol(PartnerAtom),fol_Equality()) &&
		  unify_UnifyCom(cont_LeftContext(),
				 term_SecondArgument(Atom), 
				 term_FirstArgument(PartnerAtom)) &&
		  unify_UnifyCom(cont_LeftContext(),
				 term_FirstArgument(Atom), 
				 term_SecondArgument(PartnerAtom))) {
		SUBST mgu;

		subst_ExtractUnifierCom(cont_LeftContext(), &mgu);

		if (inf_FactorLitMax(GivenClause,i,j,mgu))
		  Result = list_Cons(inf_ApplyGeneralFactoring(GivenClause,i,
							       j,mgu),
				     Result);

		subst_Delete(mgu);
	      }

	      cont_Reset();
	    }
	  } 
	}
      }
    }
  }

  cont_Check();

  return Result;
}


LIST inf_DerivableClauses(PROOFSEARCH Search, CLAUSE GivenClause)
/**************************************************************
  INPUT:   A clause and an Index, usually the WorkedOffIndex.
  RETURNS: A list of clauses derivable from 'GivenClause' wrt index.
  EFFECT:  Allocates memory for the clauselistnodes and new clauses.
***************************************************************/
{
  LIST         ListOfDerivedClauses;
  SHARED_INDEX ShIndex;
  SORTTHEORY   Dynamic;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    fputs("\n\nIllegal input in inf_DerivableClauses\n", stderr);
    misc_Error();
  }
#endif

  ListOfDerivedClauses = list_Nil();
  ShIndex              = prfs_WorkedOffSharingIndex(Search);
  Dynamic              = prfs_DynamicSortTheory(Search);

  if (Dynamic && !clause_HasSolvedConstraint(GivenClause)) {

    if (clause_HasTermSortConstraintLits(GivenClause)) {
      if (flag_GetFlagValue(flag_ISOR))
	ListOfDerivedClauses =
	  list_Nconc(inf_ForwardSortResolution(GivenClause, sort_TheoryIndex(Dynamic), Dynamic),
		     ListOfDerivedClauses);
    }
    else
      if (flag_GetFlagValue(flag_IEMS))	
	ListOfDerivedClauses =
	  list_Nconc(inf_ForwardEmptySort(GivenClause, sort_TheoryIndex(Dynamic), Dynamic), 
		     ListOfDerivedClauses); 
  } else { /* Given with solved Constraint! */

    if (Dynamic && flag_GetFlagValue(flag_IEMS))
      ListOfDerivedClauses =
	list_Nconc(inf_BackwardEmptySort(GivenClause, sharing_Index(ShIndex), Dynamic), 
		   ListOfDerivedClauses); 

    if (Dynamic && flag_GetFlagValue(flag_ISOR))
      ListOfDerivedClauses =
	list_Nconc(inf_BackwardSortResolution(GivenClause, sharing_Index(ShIndex), Dynamic),
		   ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_IEQR))
      ListOfDerivedClauses =
	list_Nconc(inf_EqualityResolution(GivenClause), ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_IMPM))
      ListOfDerivedClauses =
	list_Nconc(inf_MergingParamodulation(GivenClause, ShIndex), ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_IEQF))
      ListOfDerivedClauses =
	list_Nconc(inf_EqualityFactoring(GivenClause), ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_IGEF))
      ListOfDerivedClauses =
	list_Nconc(inf_GeneralFactoring(GivenClause), ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_ISPR))
      ListOfDerivedClauses =
	list_Nconc(inf_SuperpositionRight(GivenClause, ShIndex),
		   ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_ISPL))
      ListOfDerivedClauses =
	list_Nconc(inf_SuperpositionLeft(GivenClause, ShIndex),
		   ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_IGER))
      ListOfDerivedClauses =
	list_Nconc(inf_GeneralResolution(GivenClause, ShIndex),
		   ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_IUNR))
      ListOfDerivedClauses =
	list_Nconc(inf_UnitResolution(GivenClause, ShIndex),
		   ListOfDerivedClauses);

    if (flag_GetFlagValue(flag_IBUR))
      ListOfDerivedClauses =
	list_Nconc(inf_BoundedDepthUnitResolution(GivenClause, ShIndex),
		   ListOfDerivedClauses);
  }

  return ListOfDerivedClauses;
}
