/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                REDUCTION RULES                         * */
/* *                                                        * */
/* *  $Module:   REDRULES                                   * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.68 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/03 12:27:51 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: rules-red.c,v $ */

#include "rules-red.h"

/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  Globals                                               * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


NAT red_STAMPID;
 
const NAT red_USABLE    = 1;
const NAT red_WORKEDOFF = 2;
const NAT red_ALL       = 3;

/* Needed for term stamping in red_RewriteRedUnitClause */


/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *  Functions                                             * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/



static void red_HandleRedundantIndexedClauses(PROOFSEARCH Search, LIST Blocked, 
					      CLAUSE RedClause, BOOL Document)
/*********************************************************
  INPUT:   A proof search object, a list <Blocked> of clauses from the proof search object and
           a clause that causes the already indexed clauses in <Blocked> to be redundant.
	   An additional flag indicating whether a proof documentation should take place.
  RETURNS: None.
  EFFECT:  Initializes the Reduction module, in particular
           its stampid to stamp terms.
**********************************************************/
{
  CLAUSE Clause;
  LIST   Scan;
  for(Scan = Blocked; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (clause_SplitLevel(Clause) < clause_SplitLevel(RedClause) ||
	prfs_LastBacktrackLevel(Search) <  clause_SplitLevel(RedClause))
      split_DeleteClauseAtLevel(Search, Clause, clause_SplitLevel(RedClause));
    else {
      if (clause_GetFlag(Clause, WORKEDOFF)) {
	if (Document)
	  prfs_MoveWorkedOffDocProof(Search, Clause);
	else
	  prfs_DeleteWorkedOff(Search, Clause);
      }
      else
	if (Document)
	  prfs_MoveUsableDocProof(Search, Clause);
	else
	  prfs_DeleteUsable(Search, Clause);
    }
  }
}

static void red_HandleRedundantDerivedClauses(PROOFSEARCH Search, LIST Blocked, 
					      CLAUSE RedClause, BOOL Document)
/*********************************************************
  INPUT:   A proof search object, a list <Blocked> of clauses from the proof search object and
           a clause that causes the derived clauses in <Blocked> to be redundant.
	   An additional flag indicating whether a proof documentation should take place.
  RETURNS: None.
  EFFECT:  Initializes the Reduction module, in particular
           its stampid to stamp terms.
**********************************************************/
{
  CLAUSE Clause;
  LIST   Scan;
  for(Scan = Blocked; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (clause_SplitLevel(Clause) < clause_SplitLevel(RedClause) ||
	prfs_LastBacktrackLevel(Search) <  clause_SplitLevel(RedClause)) {
      split_KeepClauseAtLevel(Search, Clause, clause_SplitLevel(RedClause));
    }
    else {
	if (Document)
	  prfs_InsertDocProofClause(Search, Clause);
	else
	  clause_Delete(Clause);
    }
  }
}

void red_Init()
/*********************************************************
  INPUT:   None.
  RETURNS: None.
  EFFECT:  Initializes the Reduction module, in particular
           its stampid to stamp terms.
**********************************************************/
{
  red_STAMPID = term_GetStampID();
}

void red_DocumentObviousReductions(CLAUSE Clause, LIST Indexes)
/*********************************************************
  INPUT:   A clause and a list of literal indexes removed by obvious reductions.
  RETURNS: None
  MEMORY:  The <Indexes> list is consumed.
**********************************************************/
{

  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause, list_Nil());
  clause_SetParentLiterals(Clause, Indexes);

  clause_AddParentClause(Clause,clause_Number(Clause));  /* Has to be done before increasing it! */

  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromObviousReductions(Clause);
}

static BOOL red_ObviousReductions(CLAUSE Clause, CLAUSE *Changed, BOOL Document)
/**********************************************************
  INPUT:   A clause and a flag indicating whether the reductions should be documented.
  RETURNS: TRUE iff  obvious reductions are possible.
           If <Document> is false the clause is destructively change,
	     else a reduced copy of the clause is returned in <*Changed>.
  EFFECT:  Multiple occurrences of the same literal as well as trivial
           equations are removed.
********************************************************/
{
  int     i,j,end;
  LIST    Indexes;
  TERM    Atom, PartnerAtom;

  /* printf("\nred_ObviousReductions");  DBG */

#ifdef DEBUG
  clause_Check(Clause);
#endif

  Indexes = list_Nil();
  end     = clause_LastAntecedentLitIndex(Clause);
    
  for (i = clause_FirstConstraintLitIndex(Clause); i <= end; i++) {
    Atom = clause_LiteralAtom(clause_GetLiteral(Clause,i));
    if (fol_IsEquality(Atom) &&
	term_Equal(term_FirstArgument(Atom),term_SecondArgument(Atom))) {
      Indexes = list_Cons((POINTER)i,Indexes);
    } else
      for (j = i+1; j <= end; j++) {
	PartnerAtom = clause_LiteralAtom(clause_GetLiteral(Clause,j));
	if (term_Equal(PartnerAtom, Atom)) {
	  Indexes = list_Cons((POINTER)i,Indexes);
	  j       = end;
	}
      }
  }

  end    = clause_LastSuccedentLitIndex(Clause);
    
  for (i = clause_FirstSuccedentLitIndex(Clause); i <= end; i++) {
    Atom = clause_LiteralAtom(clause_GetLiteral(Clause,i));
    for (j = i+1; j <= end; j++) {
      PartnerAtom = clause_LiteralAtom(clause_GetLiteral(Clause,j));
      if (term_Equal(PartnerAtom,Atom)) {
	Indexes = list_Cons((POINTER)i,Indexes);
	j       = end;
      }
    }
  }

  if (!list_Empty(Indexes)) {
    if (flag_GetFlagValue(flag_POBV)) {
      printf("\nObvious: ");
      clause_Print(Clause);
      printf(" ==> ");
    }      
    if (Document) {
      CLAUSE Copy;
      Copy = clause_Copy(Clause);
      clause_DeleteLiterals(Copy,Indexes);
      red_DocumentObviousReductions(Copy,Indexes); /* Indexes is consumed */
      if (flag_GetFlagValue(flag_POBV))
	clause_Print(Copy);
      *Changed = Copy;
    }
    else {
      clause_DeleteLiterals(Clause,Indexes);
      list_Delete(Indexes);
      if (flag_GetFlagValue(flag_POBV))
	clause_Print(Clause);
    }
    return TRUE;
  }

  return FALSE;
}

void red_DocumentCondensing(CLAUSE Clause, LIST Indexes)
/*********************************************************
  INPUT:   A clause and a list of literal indexes removed by condensing.
  RETURNS: None
  MEMORY:  The <Indexes> list is consumed.
**********************************************************/
{

  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause, list_Nil());
  clause_SetParentLiterals(Clause, Indexes);

  clause_AddParentClause(Clause,clause_Number(Clause));  /* Has to be done before increasing it! */

  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromCondensing(Clause);
}

static BOOL red_Condensing(CLAUSE Clause, CLAUSE *Changed, BOOL Document)
/**********************************************************
  INPUT:   A pointer to a non-empty unshared clause and a flag indicating whether
           the condensing step should be documented.
  RETURNS: TRUE iff condensing is applicable to <Clause>.
           If <Document> is false the clause is desctructively changed
	     else a condensed copy of the clause is returned in <*Changed>.
***********************************************************/
{
  LIST Indexes;

#if DEBUG
  if (!clause_IsClause(Clause) || (*Changed != (CLAUSE)NULL)) {
    printf("\n\nIllegal input in red_Condensing\n");
    misc_DumpCore();
  }
  clause_Check(Clause);
#endif

  /* printf("\nred_Condensing"); DBG */

  Indexes = cond_CondFast(Clause);

  if (!list_Empty(Indexes)) {
    if (flag_GetFlagValue(flag_PCON)) {
      printf("\nCondensing: "); 
      clause_Print(Clause);
      printf(" ==> ");
    }
    if (Document) {
      CLAUSE Copy;
      Copy = clause_Copy(Clause);
      clause_DeleteLiterals(Copy, Indexes);
      red_DocumentCondensing(Copy, Indexes);
      if (flag_GetFlagValue(flag_PCON))
	clause_Print(Copy);
      *Changed = Copy;
    }
    else {
      clause_DeleteLiterals(Clause, Indexes);
      list_Delete(Indexes);
      if (flag_GetFlagValue(flag_PCON))
	clause_Print(Clause);
    }
    return TRUE;
  }
  return FALSE;
}

BOOL red_Tautology(CLAUSE Clause)
/**********************************************************
  INPUT:   A pointer to a non-empty clause.
  RETURNS: The boolean value TRUE if 'Clause' is a tautology.
***********************************************************/
{
  TERM  Atom;
  int   i,j, la,n;
  BOOL   Result;

#if DEBUG
  if (!clause_IsClause(Clause)) {
    printf("\n\nIllegal input in red_Tautology\n");
    misc_DumpCore();
  }
  clause_Check(Clause);
#endif
  
  la     = clause_LastAntecedentLitIndex(Clause);
  n      = clause_Length(Clause);
  Result = FALSE;

  for (j = clause_FirstSuccedentLitIndex(Clause); j < n && !Result;  j++) {

    Atom = clause_LiteralAtom(clause_GetLiteral(Clause, j));

    if (fol_IsEquality(Atom) &&
	term_Equal(term_FirstArgument(Atom),term_SecondArgument(Atom)))
      Result = TRUE;
    
    for (i = clause_FirstLitIndex(); i <= la && !Result; i++)
      if (term_Equal(Atom, clause_LiteralAtom(clause_GetLiteral(Clause, i))))
	Result = TRUE;
  }

  if (Result && flag_GetFlagValue(flag_PTAUT)) {
    printf("\nTautology: "); 
    clause_Print(Clause);
  }
  return Result;
}

static LITERAL red_GetClRedLit(LITERAL ActLit, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A literal and an Index.
  RETURNS: The most valid clause with a complementary literal,
           (CLAUSE)NULL, if no such clause exists.
  MEMORY:  None.
***************************************************************/
{
  LITERAL NextLit;
  int     i;
  CLAUSE  ActClause;
  TERM    CandTerm;
  LIST    LitScan;

  NextLit     = (LITERAL)NULL;
  ActClause   = clause_LiteralOwningClause(ActLit);
  i           = clause_LiteralGetIndex(ActLit);
  CandTerm    = st_ExistGen(cont_LeftContext(),
			    sharing_Index(ShIndex),
			    clause_LiteralAtom(ActLit));

  while (CandTerm) { /* First check units */
    if (!term_IsVariable(CandTerm)) { /* Has to be an Atom! */
      LitScan   = sharing_NAtomDataList(CandTerm); /* CAUTION !!! the List is not a copy */
      while (!list_Empty(LitScan)) {
	NextLit = list_Car(LitScan);
	if (clause_LiteralsAreComplementary(ActLit,NextLit))
	  if (clause_Length(clause_LiteralOwningClause(NextLit)) == 1 ||
	      subs_SubsumesBasic(clause_LiteralOwningClause(NextLit),ActClause,
				 clause_LiteralGetIndex(NextLit),i)) {
	    st_CancelExistRetrieval();
	    return NextLit;
	  }
	LitScan = list_Cdr(LitScan);
      }
    }
    CandTerm = st_NextCandidate();
  }
  return (LITERAL)NULL;
}

static void red_DocumentClauseReduction(CLAUSE Clause, LIST LitInds, LIST ClauseNums, 
					LIST PLitInds)
/*********************************************************
  INPUT:   A clause, the involved literals indice and the clauses number.
  RETURNS: None
  MEMORY:  All input lists are consumed.
**********************************************************/
{
  LIST   Scan,Help;

  Help  = list_Nil();

  for (Scan=LitInds; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Help = list_Cons((POINTER)clause_Number(Clause), Help);    /* Has to be done before increasing it! */
  }
  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause, list_Nconc(Help,ClauseNums));
  clause_SetParentLiterals(Clause, list_Nconc(LitInds,PLitInds));
  
  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromClauseReduction(Clause);
}

static BOOL red_ClauseReduceRedClause(CLAUSE Clause, SHARED_INDEX ShIndex, CLAUSE *Changed, int Level,
				      BOOL Document)
/**************************************************************
  INPUT:   A clause, an Index and a flag indicating whether the reduction should be documented.
           In addition a level indication the need of a copy if <Clause> is reduced by
	   a clause of higher level than <Level>
  RETURNS: TRUE if reduction wrt the indexed clauses was possible.
           If <Document> is true or the clauses used for reductions have a higher split level
	      then a changed copy is returned in <*CHanged>.
	   Otherwise <Clause> is destructively changed
***************************************************************/
{
  CLAUSE  PClause,Copy;
  LITERAL ActLit,PLit;
  int     i, j, length;
  LIST    ReducedBy,ReducedLits,PLits,Scan1,Scan2;

#if DEBUG
  if (!clause_IsClause(Clause) || (*Changed != (CLAUSE)NULL)) {
    printf("\n\nIllegal input in red_ClauseReduceRedClause\n");
    misc_DumpCore();
  }
  clause_Check(Clause);
#endif
  
  Copy          = Clause;
  length        = clause_Length(Clause);
  ReducedBy     = list_Nil();
  ReducedLits   = list_Nil();
  PLits         = list_Nil();
  i             = clause_FirstLitIndex();
  j             = 0;
  
  while (i < length) {
    ActLit  = clause_GetLiteral(Copy, i);
    
    if (!fol_IsEquality(clause_LiteralAtom(ActLit)) ||   /* Reduce with negative equations. */
	clause_LiteralIsPositive(ActLit)) {
      PLit    = red_GetClRedLit(ActLit, ShIndex);
      if (clause_LiteralExists(PLit)) {
	if (list_Empty(PLits) && flag_GetFlagValue(flag_PCLR)) {
	  printf("\nFClauseReduction: "); 
	  clause_Print(Copy);
	}
	PClause     = clause_LiteralOwningClause(PLit);
	ReducedBy   = list_Cons((POINTER)clause_Number(PClause), ReducedBy);
	PLits       = list_Cons((POINTER)clause_LiteralGetIndex(PLit),PLits);
	ReducedLits = list_Cons((POINTER)(i+j), ReducedLits);
	if (Copy == Clause &&
	    (Document || (clause_SplitLevel(Copy) < clause_SplitLevel(PClause) ||
			   Level < clause_SplitLevel(PClause))))
	  Copy = clause_Copy(Clause);
	clause_UpdateSplitDataFromPartner(Copy, PClause);	  
	clause_DeleteLiteral(Copy,i);
	length--;
	j++;
      }
      else
	i++;
    }
    else
      i++;
  }
  
  if (!list_Empty(ReducedBy)) {
    if (Document) {
      ReducedBy   = list_NReverse(ReducedBy);
      ReducedLits = list_NReverse(ReducedLits);
      PLits       = list_NReverse(PLits);
      red_DocumentClauseReduction(Copy, ReducedLits, ReducedBy, PLits); /* Lists are consumed */
      if (flag_GetFlagValue(flag_PCLR)) {
	printf(" ==> [ ");
	for(Scan1=ReducedBy,Scan2=ReducedLits;!list_Empty(Scan1);
	    Scan1=list_Cdr(Scan1),Scan2=list_Cdr(Scan2))
	  printf("%d.%d ",(NAT)list_Car(Scan1),(NAT)list_Car(Scan2));
	printf("] ");
	clause_Print(Copy);
      }
    }
    else {
      if (flag_GetFlagValue(flag_PCLR)) {
	printf(" ==> [ ");
	for(Scan1=ReducedBy,Scan2=ReducedLits;!list_Empty(Scan1);
	    Scan1=list_Cdr(Scan1),Scan2=list_Cdr(Scan2))
	  printf("%d.%d ",(NAT)list_Car(Scan1),(NAT)list_Car(Scan2));
	printf("] ");
	clause_Print(Copy);
      }
      list_Delete(ReducedBy);
      list_Delete(ReducedLits);
      list_Delete(PLits);
    }
    if (Copy != Clause)
      *Changed = Copy;
    return TRUE;
  }
  return FALSE;
}


static CLAUSE red_ForwardSubsumer(CLAUSE RedCl, SHARED_INDEX ShIndex)
/**********************************************************
  INPUT:   A pointer to a non-empty clause and an index of clauses.
  RETURNS: A possibly NULL clause with minimal split level
           that subsumes <REdCl>
***********************************************************/
{
  TERM   Atom,AtomGen;
  CLAUSE CandCl;
  LIST   LitScan;
  int    i,lits;

#if DEBUG
  if (!clause_IsClause(RedCl)) {
    printf("\n\nIllegal input in red_ForwardSubsumer\n");
    misc_DumpCore();
  }
  clause_Check(RedCl);
#endif

  /* printf("\nred_ForwardSubsumer with "); clause_Print(RedCl); DBG */

  lits     = clause_Length(RedCl);

  for (i = 0; i < lits; i++) {
    Atom    = clause_GetLiteralAtom(RedCl, i);
    AtomGen = st_ExistGen(cont_LeftContext(), sharing_Index(ShIndex), Atom);

    while (AtomGen) {
      if (!term_IsVariable(AtomGen)) {
	for (LitScan = sharing_NAtomDataList(AtomGen);
	     !list_Empty(LitScan);
	     LitScan = list_Cdr(LitScan)) {
	  CandCl = clause_LiteralOwningClause(list_Car(LitScan));

	  if (CandCl != RedCl &&
	      clause_GetLiteral(CandCl,clause_FirstLitIndex()) == (LITERAL)list_Car(LitScan) &&
	      term_TopSymbol(clause_GetLiteralTerm(RedCl,i)) == 
	      term_TopSymbol(clause_GetLiteralTerm(CandCl,
						   clause_FirstLitIndex())) &&
	      subs_SubsumesBasic(CandCl, RedCl, clause_FirstLitIndex(), i)) {
	    st_CancelExistRetrieval();
	    return (CandCl);
	  }
	}
      }
      AtomGen = st_NextCandidate();
    }

    if (fol_IsEquality(Atom) && 
	clause_LiteralIsNotOrientedEquality(clause_GetLiteral(RedCl,i))) {
      Atom = term_Create(fol_Equality(),list_Reverse(term_ArgumentList(Atom)));
      AtomGen = st_ExistGen(cont_LeftContext(), sharing_Index(ShIndex), Atom);
      while (AtomGen) {
	if (!term_IsVariable(AtomGen)) {
	  for (LitScan = sharing_NAtomDataList(AtomGen);
	       !list_Empty(LitScan);
	       LitScan = list_Cdr(LitScan)) {
	    CandCl = clause_LiteralOwningClause(list_Car(LitScan));
	    if ((CandCl != RedCl) && 
		clause_GetLiteral(CandCl,clause_FirstLitIndex())
		== (LITERAL)list_Car(LitScan) &&
		term_TopSymbol(clause_GetLiteralTerm(RedCl,i)) == 
		term_TopSymbol(clause_GetLiteralTerm(CandCl,
						 clause_FirstLitIndex())) &&
		subs_SubsumesBasic(CandCl, RedCl, clause_FirstLitIndex(), i)) {
	      st_CancelExistRetrieval();
	      list_Delete(term_ArgumentList(Atom));
	      term_Free(Atom);
	      return (CandCl);
	    }
	  }
	}
	AtomGen = st_NextCandidate();
      }
      list_Delete(term_ArgumentList(Atom));
      term_Free(Atom);
    }
  }

  return((CLAUSE)NULL);
}


static CLAUSE red_ForwardSubsumption(CLAUSE RedClause, SHARED_INDEX ShIndex)
/**********************************************************
  INPUT:   A pointer to a clause and an index of clauses.
  RETURNS: TRUE iff <RedClause> is subsumed by a clause in <ShIndex>
***********************************************************/
{ 
  CLAUSE Subsumer;

#if DEBUG
  if (!clause_IsClause(RedClause)) {
    printf("\n\nIllegal input in red_ForwardSubsumption\n");
    misc_DumpCore();
  }
  clause_Check(RedClause);
#endif

  Subsumer = red_ForwardSubsumer(RedClause, ShIndex);

  if (flag_GetFlagValue(flag_PSUB) && Subsumer) {
    printf("\nFSubsumption:"); clause_Print(RedClause);
    printf(" by %d %d ",clause_Number(Subsumer),clause_SplitLevel(Subsumer));
  }

  return Subsumer;
}

static void red_DocumentRewriting(CLAUSE Clause, int i, CLAUSE Rule, int ri)
/*********************************************************
  INPUT:   Two clauses and the literals involved in the rewrite step.
  RETURNS: None
**********************************************************/
{
  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));
  clause_SetParentClauses(Clause,
			  list_List((POINTER)clause_Number(Clause)));
  /* Has to be done before increasing the number! */

  clause_SetParentLiterals(Clause, list_List((POINTER)i));
  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromRewriting(Clause);

  clause_AddParentClause(Clause,clause_Number(Rule));
  clause_AddParentLiteral(Clause,ri);
}

static void red_DocumentFurtherRewriting(CLAUSE Clause, int i, CLAUSE Rule, int ri)
/*********************************************************
  INPUT:   Two clauses and the literals involved in the rewrite step.
  RETURNS: None
**********************************************************/
{
  clause_AddParentClause(Clause,
    (int) list_Car(list_Cdr(clause_ParentClauses(Clause))));
  clause_AddParentLiteral(Clause, i);
  clause_AddParentClause(Clause, clause_Number(Rule));
  clause_AddParentLiteral(Clause, ri);
}

static BOOL red_RewriteRedUnitClause(CLAUSE RedClause, SHARED_INDEX ShIndex, CLAUSE *Changed, int Level,
				     BOOL Document)
/**************************************************************
  INPUT:   A unit (!) clause and an Index and a flag indicating whether the
           rewriting should be documented.
	   In addition a level indication the need of a copy if <Clause> is reduced by
	   a clause of higher level than <Level>.
  RETURNS: TRUE iff rewriting was possible.
           If <Document> is true or the split level of the rewrite rule is higher
	     a copy of RedClause that is rewritten wrt. the indexed clauses is returned in <*Changed>.
           Otherwise the clause is destructively rewritten.
***************************************************************/
{
  TERM   RedAtom, RedTermS;
  int    B_Stack;
  BOOL   Rewritten,Result,Oriented,Renamed;
  TERM   TermS,PartnerEq;
  LIST   EqList,EqScan,LitScan;
  CLAUSE Copy;

#if DEBUG
  if (!clause_IsClause(RedClause) || (*Changed != (CLAUSE)NULL)) {
    printf("\n\nIllegal input in red_RewriteRedUnitClause\n");
    misc_DumpCore();
  }
  clause_Check(RedClause);
#endif

  Result      = FALSE;
  Renamed     = FALSE;
  Copy        = RedClause;
  RedAtom     = clause_GetLiteralAtom(Copy, clause_FirstLitIndex());
  Rewritten   = TRUE;

  if (!list_Empty(term_ArgumentList(RedAtom))) { /* Not propositional */
    if (term_StampOverflow(red_STAMPID))
      term_ResetTermStamp(clause_LiteralSignedAtom(clause_GetLiteral(RedClause, clause_FirstLitIndex())));
    term_StartStamp();
    while (Rewritten) {
      Rewritten = FALSE;
      B_Stack = stack_Bottom();
      sharing_PushListOnStackNoStamps(term_ArgumentList(RedAtom));

      while (!stack_Empty(B_Stack)) {
	RedTermS = (TERM)stack_PopResult();
	TermS    = st_ExistGen(cont_LeftContext(), sharing_Index(ShIndex), RedTermS);
	while (TermS && !Rewritten) {
	  EqList = term_SupertermList(TermS);
	  for (EqScan = EqList; !list_Empty(EqScan) && !Rewritten;
	       EqScan = list_Cdr(EqScan)) {
	    PartnerEq = list_Car(EqScan);
	    if (fol_IsEquality(PartnerEq)) {
	      CLAUSE  RewriteClause;
	      LITERAL RewriteLit;
	      TERM    Right;
	      if (TermS == term_FirstArgument(PartnerEq))
		Right = term_SecondArgument(PartnerEq);
	      else
		Right = term_FirstArgument(PartnerEq);
	      for (LitScan = sharing_NAtomDataList(PartnerEq); 
		   !list_Empty(LitScan) && !Rewritten;
		   LitScan = list_Cdr(LitScan)) {
		RewriteLit    = list_Car(LitScan);
		RewriteClause = clause_LiteralOwningClause(RewriteLit);
		if (clause_LiteralIsPositive(RewriteLit) &&
		    clause_Length(RewriteClause) == 1) {
		  Oriented = (clause_LiteralIsOrientedEquality(RewriteLit) && 
			      TermS == term_FirstArgument(PartnerEq));
		  if (!Oriented && !clause_LiteralIsOrientedEquality(RewriteLit)) { 
		    Renamed = TRUE;                            /* If oriented, no renaming needed! */
		    term_StartMaxRenaming(clause_MaxVar(RewriteClause));
		    term_Rename(RedAtom); /* Renaming destructive, no extra match needed !! */
		    Oriented = kbo_ContGreater(cont_LeftContext(),TermS,cont_LeftContext(),Right);
		      
		    /*if (Oriented) {
		      printf("\n\n\tRedAtom: ");term_PrintPrefix(RedAtom);
		      printf("\n\tSubTerm: ");term_PrintPrefix(RedTermS);
		      printf("\n\tGenTerm: ");term_PrintPrefix(TermS);
		      printf("\n\tGenRight: ");term_PrintPrefix(Right);
		      printf("\n");cont_PrintCurrentTrail();printf("\n");
		    }*/
		  }
		  if (Oriented) {
		    TERM   TermT;

		    if (RedClause == Copy &&
			(Document || 
			 clause_SplitLevel(RewriteClause) > clause_SplitLevel(RedClause) ||
			 Level < clause_SplitLevel(RewriteClause))) {
		      Copy    = clause_Copy(RedClause);
		      RedAtom = clause_GetLiteralAtom(Copy, clause_FirstLitIndex());
		    }

		    if (!Result)
		      if (flag_GetFlagValue(flag_PREW)) {
			printf("\nFRewriting: ");
			clause_Print(Copy);
			printf(" ==>[ ");
		      }

		    if (Document)
		      if (!Result)
			red_DocumentRewriting(Copy, clause_FirstLitIndex(),
					      RewriteClause, clause_FirstLitIndex());
		      else
			red_DocumentFurtherRewriting(Copy, clause_FirstLitIndex(), 
						     RewriteClause, clause_FirstLitIndex());
		    Result = TRUE;
		    TermT  = cont_ApplyBindingsModuloMatching(cont_LeftContext(), term_Copy(Right));
		    if (cont_BindingsAreRenamingModuloMatching(cont_RightContext()))
		      term_SetTermSubtermStamp(TermT);
		    term_ReplaceSubtermBy(RedAtom, RedTermS, TermT);
		    Rewritten = TRUE;
		    clause_UpdateSplitDataFromPartner(Copy, RewriteClause);
		    term_Delete(TermT);
		    stack_SetBottom(B_Stack);
		  
		    if (flag_GetFlagValue(flag_PREW))
		      printf("%d.%d ",clause_Number(RewriteClause), clause_FirstLitIndex());
		    clause_SetWeight(Copy, clause_ComputeWeight(Copy));
		  }
		}
	      }
	    }
	  }
	  if (!Rewritten)
	    TermS = st_NextCandidate();
	}
	st_CancelExistRetrieval();
	if (!Rewritten)
	  term_SetTermStamp(RedTermS);
      }
    }
    term_StopStamp(); 
  }
  if (Result) {
    clause_OrientAndReInit(Copy);
    if (Copy != RedClause)
      clause_OrientAndReInit(RedClause);
    if (flag_GetFlagValue(flag_PREW)) {
      printf("] "); 
      clause_Print(Copy);
    }
    if (Copy != RedClause)
      *Changed = Copy;
  }
  else
    if (Renamed)
      clause_OrientAndReInit(Copy);
	
      
#if DEBUG
  clause_Check(Copy);
  clause_Check(RedClause);
#endif
  
  return Result;
}


static BOOL red_RewriteRedClause(CLAUSE RedClause, SHARED_INDEX ShIndex, CLAUSE *Changed, int Level,
				 BOOL Document)
/**************************************************************
  INPUT:   A clause and an Index.
           In addition a level indication the need of a copy if <Clause> is reduced by
	   a clause of higher level than <Level>.
  RETURNS: NULL, if no rewriting was possible.
           If <Document> is true or the split level of the rewrite rule is higher
	     a copy of RedClause that is rewritten wrt. the indexed clauses.
	   Otherwise the clause is destructively rewritten and returned.
***************************************************************/
{
  TERM   RedAtom, RedTermS;
  int    B_Stack;
  int    ci, length;
  BOOL   Rewritten,Result;
  TERM   TermS,PartnerEq;
  LIST   EqList,EqScan,LitScan;
  CLAUSE Copy;

#if DEBUG
  if (!clause_IsClause(RedClause)) {
    printf("\n\nIllegal input in red_RewriteRedClause\n");
    misc_DumpCore();
  }
  clause_Check(RedClause);
#endif
  
  length      = clause_Length(RedClause);

  if (length == 1)
    return red_RewriteRedUnitClause(RedClause, ShIndex, Changed, Level, Document);

  Result      = FALSE;
  Copy        = RedClause;

  for (ci = clause_FirstLitIndex(); ci < length; ci++) {
    Rewritten  = TRUE;
    if (!list_Empty(term_ArgumentList(clause_GetLiteralAtom(Copy, ci)))) {
      while (Rewritten) {
	Rewritten = FALSE;
	RedAtom   = clause_GetLiteralAtom(Copy, ci);

	B_Stack = stack_Bottom();
	sharing_PushListReverseOnStack(term_ArgumentList(RedAtom));

	while (!stack_Empty(B_Stack))  {
	  RedTermS = (TERM)stack_PopResult();
	  TermS    = st_ExistGen(cont_LeftContext(), sharing_Index(ShIndex), RedTermS);

	  while (TermS && !Rewritten) {
	    EqList = term_SupertermList(TermS);

	    for (EqScan = EqList; !list_Empty(EqScan) && !Rewritten;
		 EqScan = list_Cdr(EqScan)) {
	      PartnerEq = list_Car(EqScan);
	      if (fol_IsEquality(PartnerEq) &&
		  (term_FirstArgument(PartnerEq) == TermS)) {
		CLAUSE  RewriteClause;
		LITERAL RewriteLit;
		int     ri;

		for (LitScan = sharing_NAtomDataList(PartnerEq); 
		     !list_Empty(LitScan) && !Rewritten;
		     LitScan = list_Cdr(LitScan)) {
		  RewriteLit    = list_Car(LitScan);
		  RewriteClause = clause_LiteralOwningClause(RewriteLit);
		  ri            = clause_LiteralGetIndex(RewriteLit);

		  if (clause_LiteralIsPositive(RewriteLit) &&
		      clause_LiteralGetFlag(RewriteLit,STRICTMAXIMAL) &&
		      clause_LiteralIsOrientedEquality(RewriteLit) &&
		      subs_SubsumesBasic(RewriteClause, Copy, ri, ci)) {
		    TERM   TermT;

		    if (RedClause == Copy &&
			(Document || 
			 clause_SplitLevel(RewriteClause) > clause_SplitLevel(RedClause) ||
			 Level < clause_SplitLevel(RewriteClause))) {
		      Copy    = clause_Copy(RedClause);
		      RedAtom = clause_GetLiteralAtom(Copy, ci);
		    }

		    if (!Result) {
		      if (flag_GetFlagValue(flag_PREW)) {
			printf("\nFRewriting: ");
			clause_Print(Copy);
			printf(" ==>[ ");
		      }
		    }

		    if (Document)
		      if (!Result)
			red_DocumentRewriting(Copy,ci, RewriteClause,ri);
		      else
			red_DocumentFurtherRewriting(Copy,ci,RewriteClause,ri);

		    Result = TRUE;
		    TermT  = cont_ApplyBindingsModuloMatching(cont_LeftContext(),
							      term_Copy(term_SecondArgument(PartnerEq)));
		    term_ReplaceSubtermBy(RedAtom, RedTermS, TermT);
		    Rewritten = TRUE;
		    clause_UpdateSplitDataFromPartner(Copy,RewriteClause);
		    term_Delete(TermT);
		    stack_SetBottom(B_Stack);

		    if (flag_GetFlagValue(flag_PREW))
		      printf("%d.%d ",clause_Number(RewriteClause), ri);
		    clause_SetWeight(Copy, clause_ComputeWeight(Copy));
		  }
		}
	      }
	    }
	    if (!Rewritten)
	      TermS = st_NextCandidate();
	  }
	  st_CancelExistRetrieval();
	} 
      }
    }
  }
  if (Result) {
    clause_OrientAndReInit(Copy);
    if (flag_GetFlagValue(flag_PREW)) {
      printf("] "); 
      clause_Print(Copy);
    }
    if (Copy != RedClause) {
      clause_OrientAndReInit(RedClause);
      *Changed = Copy;
    }
  }

#if DEBUG
  clause_Check(Copy);
  clause_Check(RedClause);
#endif

  return Result;
}

LIST red_BackSubsumption(CLAUSE RedCl, SHARED_INDEX ShIndex)
/**********************************************************
  INPUT:   A pointer to a non-empty clause and an index of clauses.
  RETURNS: The list of clauses that are subsumed by the clause RedCl.
***********************************************************/
{
  TERM   Atom,CandTerm;
  CLAUSE SubsumedCl;
  LIST   CandLits, Scan, SubsumedList;
  int    i,j;
  
#if DEBUG
  if (!clause_IsClause(RedCl)) {
    printf("\n\nIllegal input in red_BackSubsumption\n");
    misc_DumpCore();
  }
  clause_Check(RedCl);
#endif
  
  /* printf("\nred_BackSubsumption"); DBG */
  
  SubsumedList = list_Nil();
  
  if (!clause_IsEmptyClause(RedCl)) {
    
    i          = clause_FirstLitIndex();
    Atom       = clause_GetLiteralAtom(RedCl, i);
    CandTerm   = st_ExistInstance(cont_LeftContext(), sharing_Index(ShIndex), Atom);
    
    while (CandTerm) {
      CandLits = sharing_NAtomDataList(CandTerm);
      
      for (Scan = CandLits; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	SubsumedCl = clause_LiteralOwningClause(list_Car(Scan));
	j          = clause_LiteralGetIndex(list_Car(Scan));
	
	if ((RedCl != SubsumedCl) && 
	    term_TopSymbol(clause_GetLiteralTerm(RedCl,i)) == 
	    term_TopSymbol(clause_GetLiteralTerm(SubsumedCl,j)) &&
	    !list_PointerMember(SubsumedList, SubsumedCl) &&
	    subs_SubsumesBasic(RedCl, SubsumedCl, i, j))
	  SubsumedList = list_Cons(SubsumedCl, SubsumedList);
      }
      
      CandTerm = st_NextCandidate();
    }
    
    if (fol_IsEquality(Atom) && 
	clause_LiteralIsNotOrientedEquality(clause_GetLiteral(RedCl, i))) {
      Atom      = term_Create(fol_Equality(),
			      list_Reverse(term_ArgumentList(Atom)));
      CandTerm  = st_ExistInstance(cont_LeftContext(), sharing_Index(ShIndex), Atom);
      
      while (CandTerm) {
	CandLits = sharing_NAtomDataList(CandTerm);
	
	for (Scan = CandLits; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  SubsumedCl = clause_LiteralOwningClause(list_Car(Scan));
	  /* if (!clause_GetFlag(SubsumedCl, BLOCKED)) { */
	  j          = clause_LiteralGetIndex(list_Car(Scan));
	  
	  if ((RedCl != SubsumedCl) && 
	      term_TopSymbol(clause_GetLiteralTerm(RedCl,i)) == 
	      term_TopSymbol(clause_GetLiteralTerm(SubsumedCl,j)) &&
	      !list_PointerMember(SubsumedList, SubsumedCl) &&
	      subs_SubsumesBasic(RedCl, SubsumedCl, i, j))
	    SubsumedList = list_Cons(SubsumedCl, SubsumedList);
	  /* } */
	}
	
	CandTerm = st_NextCandidate();
      }
      
      list_Delete(term_ArgumentList(Atom));
      term_Free(Atom);
    }
      
    if (flag_GetFlagValue(flag_PSUB)) {
	for (Scan = SubsumedList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	    SubsumedCl = list_Car(Scan);
	    printf("\nBSubsumption: "); 
	    clause_Print(SubsumedCl);
	    printf(" by %d ",clause_Number(RedCl));
	}
    }
  }
  return SubsumedList;
}

static LIST red_GetBackClRedLits(CLAUSE Clause, LITERAL ActLit, SHARED_INDEX ShIndex)
/**************************************************************
  INPUT:   A clause, one of its literals and an Index.
  RETURNS: A list of clauses with a complementary literal instance 
           that are subsumed if these literals are ignored.
	   the empty list if no such clause exists.
  MEMORY:  Allocates the needed listnodes.
***************************************************************/
{
  CLAUSE  PClause;
  LITERAL PLit;
  LIST    LitScan, PClLits;
  TERM    CandTerm;
  int     i;

  PClLits   = list_Nil();
  i         = clause_LiteralGetIndex(ActLit);

  CandTerm  = st_ExistInstance(cont_LeftContext(),
			       sharing_Index(ShIndex),
			       clause_LiteralAtom(ActLit));

  while (CandTerm) {

    LitScan = sharing_NAtomDataList(CandTerm); /* CAUTION ! */

    for ( ; !list_Empty(LitScan); LitScan = list_Cdr(LitScan)) {

      PLit    = list_Car(LitScan);
      PClause = clause_LiteralOwningClause(PLit);

      if (PClause != Clause &&
	  clause_LiteralsAreComplementary(ActLit,PLit) &&
	  subs_SubsumesBasic(Clause,PClause,i,clause_LiteralGetIndex(PLit)))
	PClLits = list_Cons(PLit, PClLits);
    }
    
    CandTerm = st_NextCandidate();
  }
  return PClLits;
}


LIST red_BackClauseReduction(CLAUSE RedClause, SHARED_INDEX ShIndex, LIST* Result, BOOL Document)
/**************************************************************
  INPUT:   A clause, a shared index, a pointer to a result list and a flag whether the
           reduction should be documented.
  RETURNS: The return value itself contains a list of clauses from <ShIndex> that is reducable by
           <RedClause> via clause reduction.
	   The return value stored in <*Result> contains the result of this operation.
	   If <Document> is true then the clauses in <*Result> contain information abaout the reduction.
***************************************************************/
{
    LIST Blocked;
    CLAUSE Copy;

#if DEBUG
  if (!clause_IsClause(RedClause)) {
    printf("\n\nIllegal input in red_BackClauseReduction\n");
    misc_DumpCore();
  }
  clause_Check(RedClause);
#endif

  Blocked = list_Nil();

  if (clause_Length(RedClause) == 1) {
    LITERAL ActLit, PLit;
    LIST    LitList, Scan, Iter;
    TERM    CandTerm;
    int     RedClNum;
    
    ActLit      = clause_GetLiteral(RedClause, clause_FirstLitIndex());

    if (!fol_IsEquality(clause_LiteralAtom(ActLit)) ||  /* Reduce with negative equations too */
	clause_LiteralIsNegative(ActLit)) {
      CLAUSE PClause;
      int    PInd;
      LIST   PIndL;

      CandTerm = st_ExistInstance(cont_LeftContext(), sharing_Index(ShIndex), clause_LiteralAtom(ActLit));
      RedClNum = clause_Number(RedClause);
      LitList  = list_Nil();

      while (CandTerm) {
	for (Iter = sharing_NAtomDataList(CandTerm); !list_Empty(Iter); Iter = list_Cdr(Iter))
	  if (clause_LiteralsAreComplementary(ActLit,list_Car(Iter)))
	    LitList = list_Cons(list_Car(Iter),LitList);
	CandTerm = st_NextCandidate();
      }

      /* It is important to get all literals first,
	 because there may be several literals in the same clause which can be reduced by <ActLit> */
      
      while (!list_Empty(LitList)) {
	PLit    = list_Car(LitList);
	PInd    = clause_LiteralGetIndex(PLit);
	PIndL   = list_List(PLit);
	PClause = clause_LiteralOwningClause(PLit);
	Blocked = list_Cons(PClause, Blocked);

	if (flag_GetFlagValue(flag_PCLR)) {
	  printf("\nBClauseReduction: ");
	  clause_Print(PClause);
	  printf(" ==>[ %d.%d ] ",clause_Number(RedClause),clause_FirstLitIndex());
	}

	Iter = LitList;
	for (Scan=list_Cdr(LitList);!list_Empty(Scan);Scan=list_Cdr(Scan)) /* Get brothers of PLit */
	  if (PClause == clause_LiteralOwningClause(list_Car(Scan))) {
	    list_Rplacd(Iter,list_Cdr(Scan));
	    list_Rplacd(Scan,PIndL);
	    PIndL = Scan;
	    Scan  = Iter;
	  }
	  else
	    Iter = Scan;
	Iter    = LitList;
	LitList = list_Cdr(LitList);
	list_Free(Iter);
	Copy    = clause_Copy(PClause);
	clause_RemoveFlag(Copy,WORKEDOFF);
	clause_UpdateSplitDataFromPartner(Copy, RedClause);
	for(Scan=PIndL;!list_Empty(Scan);Scan=list_Cdr(Scan))       /* Change lits to indexes */
	  list_Rplaca(Scan,(POINTER)clause_LiteralGetIndex(list_Car(Scan)));
	clause_DeleteLiterals(Copy, PIndL);
	
	if (Document)
	  red_DocumentClauseReduction(Copy, PIndL, list_List((POINTER)RedClNum),  /* Lists are consumed */
				      list_List((POINTER)clause_FirstLitIndex()));
	else
	  list_Delete(PIndL);	
	
	if (flag_GetFlagValue(flag_PCLR))
	  clause_Print(Copy);
	*Result = list_Cons(Copy, *Result);
      }
    }
    return Blocked;
  } 
  else {
    CLAUSE  PClause;
    LITERAL ActLit, PLit;
    LIST    LitScan,LitList;
    int     i,length,RedClNum,PInd;

    RedClNum    = clause_Number(RedClause);
    length      = clause_Length(RedClause);

    for (i = clause_FirstLitIndex(); i < length; i++) {
      ActLit = clause_GetLiteral(RedClause, i);
      
      if (!fol_IsEquality(clause_LiteralAtom(ActLit))) {
	LitList = red_GetBackClRedLits(RedClause,ActLit, ShIndex);
	
	for (LitScan = LitList;!list_Empty(LitScan);LitScan = list_Cdr(LitScan)) {
	  PLit    = list_Car(LitScan);
	  PClause = clause_LiteralOwningClause(PLit);
	  PInd    = clause_LiteralGetIndex(PLit);
	  if (!list_PointerMember(Blocked,PClause))
	    Blocked = list_Cons(PClause, Blocked);

	  Copy    = clause_Copy(PClause);
	  clause_RemoveFlag(Copy,WORKEDOFF);
	  clause_UpdateSplitDataFromPartner(Copy, RedClause);
	  clause_DeleteLiteral(Copy, PInd);

	  if (Document)
	    red_DocumentClauseReduction(Copy, list_List((POINTER)PInd), 
				      list_List((POINTER)RedClNum), 
				      list_List((POINTER)i));
	  
	  if (flag_GetFlagValue(flag_PCLR)) {
	    printf("\nBClauseReduction: "); 
	    clause_Print(PClause);
	    printf(" ==>[ %d.%d ] ",clause_Number(RedClause),i);
	    clause_Print(Copy);
	  }
	  *Result = list_Cons(Copy, *Result);
	}
	list_Delete(LitList);
      }
    }
    return Blocked;
  }
}

static void red_ApplyRewriting(CLAUSE RuleCl, int ri, CLAUSE PartnerClause, int pli, TERM PartnerTermS,
			       BOOL Document)
/**************************************************************
  INPUT:   A clause to use for rewriting, the index of a 
           positive equality literal where the first equality
	   argument is greater, a clause with the index of a literal
	   with subterm <PArtnerTermS> that can be rewritten and
	   a flag whether this operation should be documented.
  RETURNS: None, 
  CAUTION: The atom of literal pli in PartnerClause is
           destructively changed !!!
***************************************************************/
{
  LITERAL PartnerLit;
  TERM    ReplaceTermT,NewAtom;

#ifdef DEBUG
  clause_Check(PartnerClause);
  clause_Check(RuleCl);
#endif

  if (Document)
    red_DocumentRewriting(PartnerClause,pli, RuleCl,ri);

  if (flag_GetFlagValue(flag_PREW)) {
    printf("\nBRewriting: ");
    clause_Print(PartnerClause);
    printf(" ==>[ %d.%d ] ",clause_Number(RuleCl), ri); 
  }

  PartnerLit   = clause_GetLiteral(PartnerClause, pli);

  ReplaceTermT =
    cont_ApplyBindingsModuloMatchingReverse(cont_LeftContext(),
	    term_Copy(term_SecondArgument(clause_LiteralSignedAtom(clause_GetLiteral(RuleCl, ri)))));

  NewAtom      = clause_LiteralSignedAtom(PartnerLit);
  term_ReplaceSubtermBy(NewAtom, PartnerTermS, ReplaceTermT);
  term_Delete(ReplaceTermT);

  clause_OrientAndReInit(PartnerClause);
  clause_UpdateSplitDataFromPartner(PartnerClause, RuleCl);

  if (flag_GetFlagValue(flag_PREW))
    clause_Print(PartnerClause);
}


static LIST red_LiteralRewriting(CLAUSE RedClause, LITERAL ActLit, int ri,
				 SHARED_INDEX ShIndex, LIST* Result, BOOL Document)
/**************************************************************
  INPUT:   A positive equality literal where the first equality
           argument is greater, its index and an index of clauses and
	   a pointer to a list of clauses that were rewritten and
	   a flag indicating whether the reduction should be documented.
  RETURNS: The list of clauses from the index that can be rewritten by <ActLit> and <RedClause>.
           The rewritten clauses are stored in <*Result>.
***************************************************************/
{
  TERM TermS, CandTerm;
  LIST Blocked;

#if DEBUG
  if (!clause_LiteralIsLiteral(ActLit)) {
    printf("\n\nIllegal input in red_LiteralRewriting\n");
    misc_Error();
  }
  clause_Check(RedClause);
#endif

  Blocked    = list_Nil();
  TermS      = term_FirstArgument(clause_LiteralSignedAtom(ActLit));  /* Vars can't be greater ! */

  CandTerm = st_ExistInstance(cont_LeftContext(), sharing_Index(ShIndex), TermS);

  while (CandTerm) {
    LIST LitList,LitListDel;

    if (!term_IsVariable(CandTerm) &&
	!symbol_IsPredicate(term_TopSymbol(CandTerm))) {
      LitListDel = sharing_GetDataList(CandTerm, ShIndex);

      for(LitList =  LitListDel; !list_Empty(LitList); LitList = list_Cdr(LitList)){
	LITERAL PartnerLit;
	CLAUSE PartnerClause;
	int pli;

	PartnerLit    = list_Car(LitList);
	pli           = clause_LiteralGetIndex(PartnerLit);
	PartnerClause = clause_LiteralOwningClause(PartnerLit);

	if (clause_Number(RedClause) != clause_Number(PartnerClause) &&
	    subs_SubsumesBasic(RedClause, PartnerClause, ri, pli)) {
	  CLAUSE Copy;
	  if (!list_PointerMember(Blocked, PartnerClause))
	    Blocked = list_Cons(PartnerClause,Blocked);
	  Copy = clause_Copy(PartnerClause);
	  clause_RemoveFlag(Copy,WORKEDOFF);
	  red_ApplyRewriting(RedClause, ri, Copy, pli, CandTerm, Document);
	  *Result = list_Cons(Copy, *Result);
	}
      }
      list_Delete(LitListDel);
    }
    CandTerm = st_NextCandidate();
  }
  return Blocked;
}


LIST red_BackRewriting(CLAUSE RedClause, SHARED_INDEX ShIndex, LIST* Result, BOOL Document)
/**************************************************************
  INPUT:   A clause, and Index and a pointer to the list of rewritten clauses and a flag
           indicating whether the reduction should be documented
  RETURNS: A list of clauses that can be rewritten with <RedClause> and the
           result of this operation is stored in <*Result>.
***************************************************************/
{
  int     i,length;
  LITERAL ActLit;
  LIST    Blocked;

#if DEBUG
  if (!(clause_IsClause(RedClause))) {
    printf("\n\nIllegal input in red_BackRewriting\n");
    misc_Error();
  }
  clause_Check(RedClause);
#endif
  
  Blocked    = list_Nil();
  length     = clause_Length(RedClause);
  
  if (!clause_HasEmptySuccedent(RedClause)) {
    
    for (i=clause_FirstSuccedentLitIndex(RedClause); i < length; i++) {
      ActLit = clause_GetLiteral(RedClause, i);
      if (clause_LiteralIsOrientedEquality(ActLit) && 
	  clause_LiteralGetFlag(ActLit,STRICTMAXIMAL)) {
	Blocked = list_Nconc(red_LiteralRewriting(RedClause, ActLit, i, ShIndex, Result, Document),
			     Blocked);
      }
      
#if DEBUG
      if (fol_IsEquality(clause_LiteralSignedAtom(ActLit))) {
	kbo_RESULT HelpRes;
	
	HelpRes =
	  kbo_Compare(term_FirstArgument(clause_LiteralSignedAtom(ActLit)), 
		      term_SecondArgument(clause_LiteralSignedAtom(ActLit)));
	
	if (kbo_IsSmallerThan(HelpRes)){ /* For Debugging*/
	  printf("\nERROR, first argument smaller than second in: ");
	  term_Print(clause_LiteralSignedAtom(ActLit));
	  printf("\nin clause: "); clause_Print(RedClause);
	  misc_DumpCore();
	}
      } /* end of if (fol_IsEquality). */
#endif
    } /* end of for 'all succedent literals'. */
  } /* end of if (!clause_HasEmptySuccedent(RedClause)). */
  Blocked = list_PointerDeleteDuplicates(Blocked);
  return Blocked;
}


void red_DocumentSortSimplification(CLAUSE Clause, LIST Indexes, LIST Clauses)
/*********************************************************
  INPUT:   A clause and the literal indicees and clauses
           involved in sort simplification.
  RETURNS: None
  MEMORY:  Consumes the input lists.
**********************************************************/
{
  LIST Scan,Declarations,Self;

  Declarations = list_Nil();
  Self         = list_Nil();

  list_Delete(clause_ParentClauses(Clause));
  list_Delete(clause_ParentLiterals(Clause));  
  
  for(Scan=Indexes;!list_Empty(Scan);Scan=list_Cdr(Scan))
    Self = list_Cons((POINTER)clause_Number(Clause),Self);
  
  for(Scan=Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
   Declarations = list_Cons((POINTER)clause_FirstSuccedentLitIndex(list_Car(Scan)),Declarations);
   list_Rplaca(Scan,(POINTER)clause_Number(list_Car(Scan)));
  }

  clause_SetParentLiterals(Clause, list_Nconc(Indexes,Declarations));
  clause_SetParentClauses(Clause, list_Nconc(Self,Clauses));

  clause_SetNumber(Clause, clause_IncreaseCounter());
  clause_SetFromSortSimplification(Clause);
}


BOOL red_SortSimplification(SORTTHEORY Theory, CLAUSE Clause, CLAUSE *Changed, BOOL Document)
/**********************************************************
  INPUT:   A sort theory, a clause and a flag whether the simplification
           should be documented.
  RETURNS: TRUE iff sort simplification was possible.
           If <Document> is true or the split level of the used declaration clauses
	       requires copying a simplified copy of the clause is returned in <*Changed>.
	   Otherwise the clause is destructively simplified.
***********************************************************/
{ 
  if (Theory != (SORTTHEORY)NULL) {
    TERM      Atom,Term;
    SOJU      SortPair;
    SORT      TermSort,LitSort;
    LIST      Indexes,NewClauses,Clauses,Scan;
    int       i,lc,j,OldSplitLevel;
    CLAUSE    Copy;
    CONDITION Cond;

#if DEBUG
    if (!clause_IsClause(Clause)) {
      printf("\n\nIllegal input in red_SortSimplification\n");
      misc_DumpCore();
    }
    clause_Check(Clause);
#endif

    lc            = clause_LastConstraintLitIndex(Clause);
    i             = clause_FirstLitIndex();
    j             = 0;
    OldSplitLevel = clause_SplitLevel(Clause);
    Copy          = Clause;
    Indexes       = list_Nil();
    Clauses       = list_Nil();
  
    while (i <= lc) {

      Atom        = clause_LiteralAtom(clause_GetLiteral(Copy, i));
      Term        = term_FirstArgument(Atom);
      SortPair    = sort_ComputeSortNoResidues(Theory, Term, Copy, i);
      TermSort    = sort_PairSort(SortPair);
      NewClauses  = sort_ConditionClauses(sort_PairCondition(SortPair));
      sort_ConditionPutClauses(sort_PairCondition(SortPair),list_Nil());
      LitSort     = sort_TheorySortOfSymbol(Theory,term_TopSymbol(Atom));

      if ((Cond = sort_TheoryIsSubsortOfNoResidues(Theory, TermSort, LitSort)) != (CONDITION)NULL) {

	if (j == 0 && flag_GetFlagValue(flag_PSSI)) {
	  printf("\nSortSimplification: ");
	  clause_Print(Copy);
	  printf(" ==>[ ");
	}

	NewClauses = list_Nconc(NewClauses,sort_ConditionClauses(Cond));
	sort_ConditionPutClauses(Cond,list_Nil());
	sort_ConditionDelete(Cond);

	for (Scan = NewClauses; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  if (Clause == Copy && 
	      (Document || clause_SplitLevel(list_Car(Scan)) > OldSplitLevel))
	    Copy   = clause_Copy(Clause);
	  clause_UpdateSplitDataFromPartner(Copy, list_Car(Scan));
	  if (flag_GetFlagValue(flag_PSSI))
	    printf("%d ",clause_Number(list_Car(Scan)));
	}

	if (Document)
	  Indexes = list_Cons((POINTER)(i+j), Indexes);

	clause_DeleteLiteral(Copy, i);
	Clauses = list_Nconc(NewClauses,Clauses);
	j++;
	lc--;
      }	   
      else {
	list_Delete(NewClauses);
	i++;
      }
      sort_DeleteSortPair(SortPair);
      sort_Delete(LitSort);
    }

#ifdef DEBUG
    clause_Check(Copy);
#endif

    if (j > 0) {
      if (Document)
	red_DocumentSortSimplification(Copy,Indexes,Clauses);
      else
	list_Delete(Clauses);
      clause_ReInit(Copy);
      if (flag_GetFlagValue(flag_PSSI)) {
	printf("] ");
	clause_Print(Copy);
      }
      if (Copy != Clause)
	*Changed = Copy;
      return TRUE;
    }
  }

  return FALSE;
}

void red_ExchangeClauses(CLAUSE *RedClause, CLAUSE *Copy, LIST *Result)
{
  if (*Copy) {
    *Result    = list_Cons(*RedClause,*Result);
    *RedClause = *Copy;
    *Copy      = (CLAUSE)NULL;
  }
}
     


BOOL red_SimpleStaticReductions(CLAUSE *RedClause, LIST* Result, BOOL Document)
/**********************************************************
  INPUT:   A clause (by reference), a flag indicating
           whether reductions should be documented.
  EFFECT:  Used reductions are tautology deletion and obvious reductions.
  RETURNS: TRUE if <*RedClause> is redundant.
	   If <Document> is false and no copying is necessary with respect 
	   to splitting, the clause is destructively changed, otherwise
	   (intermediate) copies are made and returned in <*Result>.
***********************************************************/
{
  CLAUSE Copy;

#if DEBUG
  if (!clause_IsClause(*RedClause)) {
    printf("\n\nIllegal input in red_SimpleStaticReductions\n");
    misc_DumpCore();
  }
  clause_Check(*RedClause);
#endif

  Copy = (CLAUSE)NULL;

  if (flag_GetFlagValue(flag_RTAUT) && red_Tautology(*RedClause))
    return TRUE;

  if (flag_GetFlagValue(flag_ROBV)) {
    red_ObviousReductions(*RedClause,&Copy,Document);
    red_ExchangeClauses(RedClause,&Copy,Result);
  }

  if (flag_GetFlagValue(flag_RCON)) {
    red_Condensing(*RedClause, &Copy, Document);
    red_ExchangeClauses(RedClause,&Copy,Result);
  }

  return FALSE;    
}

  


BOOL red_StaticReductions(PROOFSEARCH Search, CLAUSE *Clause, CLAUSE *Subsumer, LIST* Result, BOOL Document,
			  NAT Mode)
/**********************************************************
  INPUT:   A proof search objetc, a clause (by reference) to be reduced, 
           a shared index of clauses and a flag indicating whether reductions should be documented.
  EFFECT:  Used reductions are tautology deletion, obvious reductions,
           forward subsumption, rewriting and clause reduction.
	   Depending on <Mode>, then clauses are reduced with respect to WorkedOff or  Usable Clauses.
  RETURNS: TRUE iff the clause is redundant.
	   If <Document> is false and no copying is necessary with respect 
	   to splitting, the clause is destructively changed, otherwise
	   (intermediate) copies are made and returned in <*Result>.
           If <Clause> gets redundant with respect to forward subsumption,
	   the subsuming clause is returned in <*Subsumer>.
***********************************************************/
{ 
  CLAUSE       Copy;
  BOOL         Redundant;
  SHARED_INDEX Index;

#if DEBUG
  if (!clause_IsClause(*Clause)) {
    printf("\n\nIllegal input in red_StaticReductions\n");
    misc_DumpCore();
  }
  clause_Check(*Clause);
#endif

  
  Index   = (red_OnlyWorkedOffMode(Mode) ? 
	     prfs_WorkedOffSharingIndex(Search) : prfs_UsableSharingIndex(Search));
  Copy      = (CLAUSE)NULL;
  Redundant = red_SimpleStaticReductions(Clause, Result, Document);

  if (Redundant)
    return Redundant;

  if (flag_GetFlagValue(flag_RFSUB)) {
    *Subsumer = red_ForwardSubsumption(*Clause, Index);
    if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
      return Redundant;
  }

  if (flag_GetFlagValue(flag_RFREW) && 
      red_RewriteRedClause(*Clause, Index, &Copy, prfs_LastBacktrackLevel(Search), Document)) {
    red_ExchangeClauses(Clause,&Copy,Result);
    Redundant = red_SimpleStaticReductions(Clause, Result, Document);
    if (Redundant)
      return Redundant;
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
    if (flag_GetFlagValue(flag_RFSUB)) {
      *Subsumer = red_ForwardSubsumption(*Clause, Index);
      if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	return Redundant;
    }
  }

  if (red_OnlyWorkedOffMode(Mode) && flag_GetFlagValue(flag_RSSI)) { 
    red_SortSimplification(prfs_DynamicSortTheory(Search), *Clause, &Copy, Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;    
  }

  if (flag_GetFlagValue(flag_RFCLR)) {
    red_ClauseReduceRedClause(*Clause, Index, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
  }
  if (red_OnlyWorkedOffMode(Mode) && flag_GetFlagValue(flag_RSST))
    Redundant = red_ClauseDeletion(prfs_StaticSortTheory(Search),*Clause);
  
#ifdef DEBUG
  clause_Check(*Clause);
#endif

  return Redundant;
}

BOOL red_CompleteStaticReductions(PROOFSEARCH Search, CLAUSE *Clause, CLAUSE *Subsumer, 
				  LIST* Result, BOOL Document)
/**********************************************************
  INPUT:   A proof search objetc, a clause (by reference) to be reduced, 
           a shared index of clauses and a flag indicating whether reductions should be documented.
  EFFECT:  Used reductions are tautology deletion, obvious reductions,
           forward subsumption, rewriting and clause reduction.
	   Depending on <Mode>, then clauses are reduced with respect to WorkedOff and/or  Usable Clauses.
  RETURNS: TRUE iff the clause is redundant.
	   If <Document> is false and no copying is necessary with respect 
	   to splitting, the clause is destructively changed, otherwise
	   (intermediate) copies are made and returned in <*Result>.
           If <Clause> gets redundant with respect to forward subsumption,
	   the subsuming clause is returned in <*Subsumer>.
***********************************************************/
{ 
  CLAUSE       Copy;
  BOOL         Redundant,Rewritten,Tried;
  SHARED_INDEX WoIndex,UsIndex;

#if DEBUG
  if (!clause_IsClause(*Clause)) {
    printf("\n\nIllegal input in red_StaticReductions\n");
    misc_DumpCore();
  }
  clause_Check(*Clause);
#endif

  
  WoIndex   = prfs_WorkedOffSharingIndex(Search);
  UsIndex   = prfs_UsableSharingIndex(Search);
  Copy      = (CLAUSE)NULL;
  Redundant = red_SimpleStaticReductions(Clause, Result, Document);

  if (Redundant)
    return Redundant;

  if (flag_GetFlagValue(flag_RFSUB)) {
    *Subsumer = red_ForwardSubsumption(*Clause, WoIndex);
    if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
      return Redundant;
    *Subsumer = red_ForwardSubsumption(*Clause, UsIndex);
    if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
      return Redundant;
  }

  Rewritten = flag_GetFlagValue(flag_RFREW);
  Tried     = FALSE;
  while (Rewritten) {
    if (red_RewriteRedClause(*Clause, WoIndex, &Copy, prfs_LastBacktrackLevel(Search), Document)) {
      Rewritten = TRUE;
      red_ExchangeClauses(Clause,&Copy,Result);
      Redundant = red_SimpleStaticReductions(Clause, Result, Document);
      if (Redundant)
	return Redundant;
      if (clause_IsEmptyClause(*Clause))
	return FALSE;
      if (flag_GetFlagValue(flag_RFSUB)) {
	*Subsumer = red_ForwardSubsumption(*Clause, WoIndex);
	if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	  return Redundant;
	*Subsumer = red_ForwardSubsumption(*Clause, UsIndex);
	if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	  return Redundant;
      }
    }
    else
      Rewritten = FALSE;

    if ((!Tried || Rewritten) && 
	red_RewriteRedClause(*Clause, UsIndex, &Copy, prfs_LastBacktrackLevel(Search), Document)) {
      Rewritten = TRUE;
      red_ExchangeClauses(Clause,&Copy,Result);
      Redundant = red_SimpleStaticReductions(Clause, Result, Document);
      if (Redundant)
	return Redundant;
      if (clause_IsEmptyClause(*Clause))
	return FALSE;
      if (flag_GetFlagValue(flag_RFSUB)) {
	*Subsumer = red_ForwardSubsumption(*Clause, WoIndex);
	if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	  return Redundant;
	*Subsumer = red_ForwardSubsumption(*Clause, UsIndex);
	if ((Redundant = (*Subsumer != (CLAUSE)NULL)))
	  return Redundant;
      }
    }
    else
      Rewritten = FALSE;
    Tried = TRUE;
  }

  if (flag_GetFlagValue(flag_RSSI)) { 
    red_SortSimplification(prfs_DynamicSortTheory(Search), *Clause, &Copy, Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE; 
  }
     
  if (flag_GetFlagValue(flag_RFCLR)) {
    red_ClauseReduceRedClause(*Clause, WoIndex, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
    red_ClauseReduceRedClause(*Clause, UsIndex, &Copy, prfs_LastBacktrackLevel(Search), Document);
    red_ExchangeClauses(Clause,&Copy,Result);
    if (clause_IsEmptyClause(*Clause))
      return FALSE;
  }

  if (flag_GetFlagValue(flag_RSST))
    Redundant = red_ClauseDeletion(prfs_StaticSortTheory(Search),*Clause);
  
#ifdef DEBUG
  clause_Check(*Clause);
#endif

  return Redundant;
}


CLAUSE red_ReductionOnDerivedClause(PROOFSEARCH Search, CLAUSE Clause, BOOL Document, NAT Mode)
/**************************************************************
  INPUT:   A proof search object, a derived clause and a flag indicating
           whether the reductions should be documented and a mode indicating which
	   indexes should be used for reductions.
  RETURNS: The non-redundant clause after reducing <Clause>, NULL if <Clause> is redundant.
  EFFECT:  Clauses probably generated, but redundant are kept according
           to <Document> and the spit level of involved clauses.
	   depending on <Mode>, then clauses are reduced with respect to WorkedOff and/or  Usable Clauses.
***************************************************************/
{
  CLAUSE RedClause;
  LIST   Redundant;

#ifdef DEBUG
  cont_SaveState();
#endif

  Redundant = list_Nil();
  RedClause = (CLAUSE)NULL;

  if (red_StaticReductions(Search,&Clause,&RedClause,&Redundant,Document,Mode)) { /* Redundant */
    red_HandleRedundantDerivedClauses(Search,Redundant,Clause,Document);
    list_Delete(Redundant);
    if (RedClause && 
	(clause_SplitLevel(RedClause) > clause_SplitLevel(Clause) ||
	 prfs_LastBacktrackLevel(Search) < clause_SplitLevel(RedClause))) {
      split_KeepClauseAtLevel(Search,Clause,clause_SplitLevel(RedClause));
    }
    else
      if (Document)
	prfs_InsertDocProofClause(Search,Clause);
      else
	clause_Delete(Clause);
    Clause = (CLAUSE)NULL;
  }
  else {
    red_HandleRedundantDerivedClauses(Search,Redundant,Clause,Document);
    list_Delete(Redundant);
  }

#ifdef DEBUG
  cont_CheckState();
#endif

  return Clause;
}

CLAUSE red_CompleteReductionOnDerivedClause(PROOFSEARCH Search, CLAUSE Clause, BOOL Document)
/**************************************************************
  INPUT:   A proof search object, a derived clause and a flag indicating
           whether the reductions should be documented.
  RETURNS: The non-redundant clause after reducing <Clause>, NULL if <Clause> is redundant.
  EFFECT:  Clauses probably generated, but redundant are kept according
           to <Document> and the spit level of involved clauses.
	   The clause is reduced with respect to all indexes.
***************************************************************/
{
  CLAUSE RedClause;
  LIST   Redundant;

#ifdef DEBUG
  cont_SaveState();
#endif

  Redundant = list_Nil();
  RedClause = (CLAUSE)NULL;

  if (red_CompleteStaticReductions(Search,&Clause,&RedClause,&Redundant,Document)) { /* Redundant */
    red_HandleRedundantDerivedClauses(Search,Redundant,Clause,Document);
    list_Delete(Redundant);
    if (RedClause && 
	(clause_SplitLevel(RedClause) > clause_SplitLevel(Clause) ||
	 prfs_LastBacktrackLevel(Search) < clause_SplitLevel(RedClause))) {
      split_KeepClauseAtLevel(Search,Clause,clause_SplitLevel(RedClause));
    }
    else
      if (Document)
	prfs_InsertDocProofClause(Search,Clause);
      else
	clause_Delete(Clause);
    Clause = (CLAUSE)NULL;
  }
  else {
    red_HandleRedundantDerivedClauses(Search,Redundant,Clause,Document);
    list_Delete(Redundant);
  }

#ifdef DEBUG
  cont_CheckState();
#endif

  return Clause;
}


LIST red_ReductionOnDerivedClauses(PROOFSEARCH Search, LIST DerivedClauses, BOOL Document, NAT Mode)
/**************************************************************
  INPUT:   A proof search object, a list of derived clauses and a flag indicating
           whether the reductions should be documented and a mode indicating which
	   indexes should be used for reductions.
  RETURNS: The empty clauses after reduction of <DerivedClauses>.
  EFFECT:  Clauses probably generated, but redundant are kept according
           to <Document> and the spit level of involved clauses.
	   depending on <Mode>, then clauses are reduced with respect to WorkedOff and/or  Usable Clauses.
	   The clauses are inserted into the Usable Index of <Search>.
***************************************************************/
{
  LIST   EmptyClauses;
  CLAUSE Clause;

#ifdef DEBUG
  cont_SaveState();
#endif

  DerivedClauses = clause_ListSortWeighed(DerivedClauses);
  EmptyClauses   = list_Nil();

  while (!list_Empty(DerivedClauses)) {
    Clause = (CLAUSE)list_NCar(&DerivedClauses);
    Clause = red_ReductionOnDerivedClause(Search, Clause, Document, Mode);
    if (Clause) {
      if (flag_GetFlagValue(flag_PKEPT)) {
	printf("\nKept: "); 
	clause_Print(Clause);
      }
      prfs_IncreaseKeptClauses(Search);	
      if (clause_IsEmptyClause(Clause))
	EmptyClauses = list_Cons(Clause,EmptyClauses);
      else 
	prfs_InsertUsableClause(Search,Clause);
    }
  }


#ifdef DEBUG
  cont_CheckState();
#endif

  return EmptyClauses;
}


LIST red_BackReduction(PROOFSEARCH Search, CLAUSE Clause, BOOL Document, NAT Mode)
/**************************************************************
  INPUT:   A proof search object and a clause and a flag indicating whether
           a proof should be documented and a mode flag.
  RETURNS: A list of reduced clauses in usable and worked-off (depending on <Mode>)
           in <Search> with respect to <Clause>.
	   The original clauses that become redundant are either deleted
	   or kept for proof documentation or splitting.
  EFFECT:  The original clauses that become redundant are either deleted
	   or kept for proof documentation or splitting.
***************************************************************/
{
  LIST Result, Redundant;

#ifdef DEBUG
  cont_SaveState();
#endif

  Result    = list_Nil();

  if (flag_GetFlagValue(flag_RBSUB)) {
    Redundant = list_Nil();
    if (red_WorkedOffMode(Mode))
      Redundant = red_BackSubsumption(Clause,prfs_WorkedOffSharingIndex(Search));
    if (red_UsableMode(Mode))
      Redundant = list_Nconc(Redundant,red_BackSubsumption(Clause,prfs_UsableSharingIndex(Search)));
    red_HandleRedundantIndexedClauses(Search, Redundant, Clause, Document);
    list_Delete(Redundant);
  }

  if (flag_GetFlagValue(flag_RBCLR)) {
    Redundant = list_Nil();
    if (red_WorkedOffMode(Mode))
      Redundant = red_BackClauseReduction(Clause,prfs_WorkedOffSharingIndex(Search),&Result,Document);
    if (red_UsableMode(Mode))
      Redundant = list_Nconc(Redundant,
			     red_BackClauseReduction(Clause,prfs_UsableSharingIndex(Search),
						     &Result,Document));
    red_HandleRedundantIndexedClauses(Search, Redundant, Clause, Document);
    list_Delete(Redundant);
  }

  if (flag_GetFlagValue(flag_RBREW)) {
    Redundant = list_Nil();
    if (red_WorkedOffMode(Mode))
      Redundant = red_BackRewriting(Clause,prfs_WorkedOffSharingIndex(Search),&Result,Document);
    if (red_UsableMode(Mode))
      Redundant = list_Nconc(Redundant,
			     red_BackRewriting(Clause,prfs_UsableSharingIndex(Search),&Result,Document));
    red_HandleRedundantIndexedClauses(Search, Redundant, Clause, Document);
    list_Delete(Redundant);
  }

#ifdef DEBUG
  cont_CheckState();
#endif

  return Result;
}

LIST red_CompleteReductionOnDerivedClauses(PROOFSEARCH Search, LIST DerivedClauses, BOOL Document,
					   NAT *Derived)
/**************************************************************
  INPUT:   
  RETURNS: 
  EFFECT:  
***************************************************************/
{
  LIST   EmptyClauses,NewClauses,Scan;
  CLAUSE Clause;

#ifdef DEBUG
  cont_SaveState();
#endif

  EmptyClauses   = list_Nil();
  DerivedClauses = clause_ListSortWeighed(DerivedClauses);

  while (!list_Empty(DerivedClauses)) {
    Clause     = list_NCar(&DerivedClauses);
    Clause     = red_CompleteReductionOnDerivedClause(Search, Clause, Document);
    if (Clause) {
      prfs_IncreaseKeptClauses(Search);
      if (flag_GetFlagValue(flag_PKEPT)) {
	printf("\nKept: "); 
	clause_Print(Clause);
      }
      if (clause_IsEmptyClause(Clause))
	EmptyClauses = list_Cons(Clause,EmptyClauses);
      else {
	NewClauses = red_BackReduction(Search,Clause,Document,red_ALL);
	*Derived  += list_Length(NewClauses);
	if (flag_GetFlagValue(flag_PDER))
	  for(Scan=NewClauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	    printf("\nDerived: "); 
	    clause_Print(list_Car(Scan));
	  }
	NewClauses = split_ExtractEmptyClauses(NewClauses,&EmptyClauses);
	prfs_InsertUsableClause(Search,Clause);
	DerivedClauses = list_NNumberMerge(DerivedClauses,NewClauses,clause_Weight);
      }
    }
  }

#ifdef DEBUG
  cont_CheckState();
#endif

  return EmptyClauses;
}



static CLAUSE red_CDForwardSubsumer(CLAUSE RedCl, st_INDEX Index)
/**********************************************************
  INPUT:   A pointer to a non-empty clause and an index of clauses.
  RETURNS: The first clause from the Approx Set which subsumes 'RedCl'.
  ********************************************************/
{
  TERM   Atom,AtomGen;
  LIST   LitScan;
  int    i,length;
  CLAUSE CandCl;

#if DEBUG
  if (!clause_IsClause(RedCl)) {
    printf("\n\nIllegal input in red_CDForwardSubsumer\n");
    misc_DumpCore();
  }
  clause_Check(RedCl);
#endif

  length  = clause_Length(RedCl);

  for (i = 0; i < length; i++) {
    Atom     = clause_GetLiteralAtom(RedCl, i);
    AtomGen  = st_ExistGen(cont_LeftContext(), Index, Atom);

    while (AtomGen) {
      for (LitScan = term_SupertermList(AtomGen); 
	   !list_Empty(LitScan); LitScan = list_Cdr(LitScan)) {
	CandCl = clause_LiteralOwningClause(list_Car(LitScan));

	if (clause_GetLiteral(CandCl,clause_FirstLitIndex()) == (LITERAL)list_Car(LitScan) &&
	    subs_Subsumes(CandCl, RedCl, clause_FirstLitIndex(), i)) {
	  st_CancelExistRetrieval();
	  return CandCl;
	}
      }
      AtomGen = st_NextCandidate();  
    }
  }
  return (CLAUSE)NULL;
}


static BOOL red_CDForwardSubsumption(CLAUSE RedClause, st_INDEX Index)
/**********************************************************
  INPUT:   A pointer to a clause and an index of clauses.
  RETURNS: The boolean value TRUE if the clause is subsumed
           by an indexed clause, if so, the clause is deleted,
	   either really or locally.
***********************************************************/
{ 
  BOOL   IsSubsumed;
  CLAUSE Subsumer;

#if DEBUG
  if (!clause_IsClause(RedClause)) {
    printf("\n\nIllegal input in red_CDForwardSubsumption\n");
    misc_DumpCore();
  }
  clause_Check(RedClause);
#endif
  IsSubsumed = FALSE;
  Subsumer   = red_CDForwardSubsumer(RedClause, Index);

  if (clause_Exists(Subsumer)) {
    IsSubsumed = TRUE;

    if (flag_GetFlagValue(flag_DOCSST) &&
	flag_GetFlagValue(flag_PSUB)) {
      printf("\nFSubsumption:"); 
      clause_Print(RedClause);
      printf(" by %d ",clause_Number(Subsumer));
    }
  }
  return IsSubsumed;
}


static void red_CDBackSubsumption(CLAUSE RedCl, LIST* UsListPt, LIST* WOListPt, st_INDEX Index)
/**********************************************************
  INPUT:   A pointer to a non-empty clause and an index of clauses.
  RETURNS: None.
***********************************************************/
{
  TERM   Atom,AtomInst;
  CLAUSE SubsumedCl;
  LIST   Scan, SubsumedList;

#if DEBUG
  if (!clause_IsClause(RedCl)) {
    printf("\n\nIllegal input in red_CDBackSubsumption\n");
    misc_DumpCore();
  }
  clause_Check(RedCl);
#endif

  SubsumedList = list_Nil();

  if (!clause_IsEmptyClause(RedCl)) {
    Atom     = clause_GetLiteralAtom(RedCl, clause_FirstLitIndex());
    AtomInst = st_ExistInstance(cont_LeftContext(), Index, Atom);
    
    while(AtomInst) {
      for (Scan = term_SupertermList(AtomInst); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	SubsumedCl = clause_LiteralOwningClause(list_Car(Scan));
	if ((RedCl != SubsumedCl) &&
	    subs_Subsumes(RedCl, SubsumedCl, clause_FirstLitIndex(), 
			  clause_LiteralGetIndex(list_Car(Scan))) &&
	    !list_PointerMember(SubsumedList, SubsumedCl))
	  SubsumedList = list_Cons(SubsumedCl, SubsumedList);
      }
      AtomInst = st_NextCandidate();
    }
    
    for (Scan = SubsumedList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      SubsumedCl = list_Car(Scan);

      if (flag_GetFlagValue(flag_DOCSST) &&  flag_GetFlagValue(flag_PSUB)) {
	printf("\nBSubsumption: "); 
	clause_Print(SubsumedCl);printf(" by %d ",clause_Number(RedCl));
      }
	     

      if (clause_GetFlag(SubsumedCl,WORKEDOFF)) {
	*WOListPt = list_PointerDeleteOneElement(*WOListPt, SubsumedCl);
      }else {
	*UsListPt = list_PointerDeleteOneElement(*UsListPt, SubsumedCl);
      }
      clause_DeleteFlatFromIndex(SubsumedCl, Index);
    }
    list_Delete(SubsumedList);
  }
}


static LIST red_CDDerivables(SORTTHEORY Theory, CLAUSE GivenClause)
/**************************************************************
  INPUT:   A sort theory and a clause.
  RETURNS: A list of clauses derivable from <GivenClause> and the declaration
           clauses in <Theory>.
***************************************************************/
{
  LIST ListOfDerivedClauses;

#if DEBUG
  if (!(clause_IsClause(GivenClause))) {
    printf("\n\nIllegal input in red_CDDerivables\n");
    misc_Error();
  }
  clause_Check(GivenClause);
#endif
  
  if (clause_HasTermSortConstraintLits(GivenClause))
    ListOfDerivedClauses = inf_ForwardSortResolution(GivenClause, sort_TheoryIndex(Theory), Theory);
  else
    ListOfDerivedClauses = inf_ForwardEmptySort(GivenClause, sort_TheoryIndex(Theory), Theory); 
  
  return ListOfDerivedClauses;
}


static BOOL red_CDReduce(SORTTHEORY Theory, CLAUSE RedClause,  LIST *ApproxUsListPt, LIST *ApproxWOListPt, 
			 st_INDEX Index)
/**************************************************************
  INPUT:   A sort theory, a clause (unshared),  and their index.
  RETURNS: TRUE iff <RedClause> is redundant with respect to clauses in the index or theory.
  EFFECT   <RedClause> is destructively changed.
***************************************************************/
{
  CLAUSE Copy;

#ifdef DEBUG
  clause_Check(RedClause);
#endif

  Copy = (CLAUSE)NULL; /* Only needed for interface */
  
  red_ObviousReductions(RedClause,&Copy,FALSE);
  red_SortSimplification(Theory, RedClause, &Copy, FALSE);
  
  if (clause_IsEmptyClause(RedClause))
    return FALSE;

  red_Condensing(RedClause, &Copy, FALSE);

  if (red_CDForwardSubsumption(RedClause, Index))
    return TRUE;
  else {			/* RedClause isn't subsumed! */
    red_CDBackSubsumption(RedClause, ApproxUsListPt, ApproxWOListPt, Index);
    clause_InsertFlatIntoIndex(RedClause, Index);
    *ApproxUsListPt = list_Cons(RedClause, *ApproxUsListPt);
  }

#if DEBUG
  clause_Check(RedClause);
  if (Copy != (CLAUSE)NULL) {
    printf("\n\nUnexpected copy in red_CDReduce !\n");
    misc_DumpCore();
  }
    
#endif

  return FALSE;
}


BOOL red_ClauseDeletion(SORTTHEORY Theory, CLAUSE RedClause)
/**************************************************************
  INPUT:   A sort theory and a clause (unshared).
  RETURNS: TRUE iff the sort constraint of the clause is unsolvable
           with respect to the sort theory.
***************************************************************/
{
  if (Theory != (SORTTHEORY)NULL) {
    CLAUSE   ConstraintClause, GivenClause;
    LIST     ApproxUsableList, ApproxWOList, EmptyClauses, ApproxDerivables, Scan;
    int      i,nc, Count, OldClauseCounter;
    st_INDEX Index;

#if DEBUG
    if (!(clause_IsClause(RedClause))) {
      printf("\n\nIllegal input in red_ClauseDeletion\n");
      misc_Error();
    }
    clause_Check(RedClause);
#endif

    if (clause_HasEmptyConstraint(RedClause) || !flag_GetFlagValue(flag_RSST))
      return FALSE;
 
    if (flag_GetFlagValue(flag_DOCSST)) {
      printf("\n\nClauseDeletion tried on: "); 
      clause_Print(RedClause);
    }

    Index             = st_IndexCreate();
    Scan              = list_Nil();
    nc                = clause_NumOfConsLits(RedClause);
    OldClauseCounter  = clause_Counter();

    /* Make constraint clause, insert it into the Approx-usable-list: */

    for (i = clause_FirstLitIndex(); i < nc; i++)
      Scan = list_Cons(term_Copy(clause_LiteralAtom(
						    clause_GetLiteral(RedClause, i))), Scan);

    Scan             = list_NReverse(Scan);
    ConstraintClause = clause_Create(Scan, list_Nil(), list_Nil());
    list_Delete(Scan);
    clause_InitSplitData(ConstraintClause);
    clause_AddParentClause(ConstraintClause, clause_Number(RedClause));
    clause_AddParentLiteral(ConstraintClause, clause_FirstLitIndex());
    clause_SetFromClauseDeletion(ConstraintClause);
    clause_InsertFlatIntoIndex(ConstraintClause, Index);
    ApproxUsableList = list_List(ConstraintClause);
    ApproxWOList     = list_Nil();

    /* printf("\nConstraint clause: "); clause_Print(ConstraintClause); */

    /* Now the lists are initialized, the subproof is started: */

    EmptyClauses = list_Nil();
    Count        = 0;

    if (flag_GetFlagValue(flag_DOCSST)) {
      printf("\n*************** ClauseDeletion subproof: ***************");
      printf("\nThe usable list:\n"); clause_ListPrint(ApproxUsableList);
      printf("\nThe worked-off list:\n"); clause_ListPrint(ApproxWOList);
      /* printf("\nAll indexed clauses: ");
	 clause_PrintAllIndexedClauses(ShIndex); */
    }
    while (!list_Empty(ApproxUsableList) && list_Empty(EmptyClauses)) {
      GivenClause      = list_Car(ApproxUsableList); 
      clause_SetFlag(GivenClause,WORKEDOFF);

      if (flag_GetFlagValue(flag_DOCSST)) {
	printf("\n\tSubproof Given clause: "); 
	clause_Print(GivenClause); fflush(stdout);
      }
      ApproxWOList     = list_Cons(GivenClause, ApproxWOList);
      ApproxUsableList = list_PointerDeleteOneElement(ApproxUsableList,GivenClause);
      ApproxDerivables = red_CDDerivables(Theory,GivenClause);
      ApproxDerivables = split_ExtractEmptyClauses(ApproxDerivables, &EmptyClauses);
    
      if (!list_Empty(EmptyClauses)) { /* Exit while loop! */
	if (flag_GetFlagValue(flag_DOCSST)) {
	  printf("\nClauseDeletion not possible: "); 
	  clause_Print(list_Car(EmptyClauses));
	}
	clause_DeleteClauseList(ApproxDerivables); 
	ApproxDerivables = list_Nil();
      }
      else {
	CLAUSE DerClause;
	for (Scan = ApproxDerivables; !list_Empty(Scan) && list_Empty(EmptyClauses);
	     Scan = list_Cdr(Scan)) {
	  DerClause = (CLAUSE)list_Car(Scan);
	  if (red_CDReduce(Theory, DerClause, &ApproxUsableList, &ApproxWOList, Index)) 
	    clause_Delete(DerClause);
	  else{
	    Count++;
	    if (flag_GetFlagValue(flag_DOCSST)) {
	      printf("\n"); clause_Print(DerClause);
	    }
	    if (clause_IsEmptyClause(DerClause))
	      EmptyClauses = list_Cons(DerClause,EmptyClauses);
	  }
	  list_Rplaca(Scan,(CLAUSE)NULL);
	}

	if (!list_Empty(EmptyClauses)) {
	  if (flag_GetFlagValue(flag_DOCSST)) {
	    printf(" ClauseDeletion not possible!");
	    clause_Print(list_Car(EmptyClauses));
	  }
	  clause_DeleteClauseList(ApproxDerivables);    /* There still may be clauses in the list */
	  ApproxDerivables = list_Nil();
	}
	else {
	  list_Delete(ApproxDerivables); 
	  ApproxDerivables = list_Nil();
	}
      }
    }

    if (!list_Empty(EmptyClauses)) {
      if (flag_GetFlagValue(flag_DOCSST)) {
	printf("\nClauseDeletion failed, constraint solvable.");
	printf("\n************ ClauseDeletion subproof finished. ************\n");
      }
    }
    else
      if (flag_GetFlagValue(flag_PSST)) {
	printf("\nClauseDeletion: "); 
	clause_Print(RedClause);
      }

    /* Cleanup */
    clause_DeleteClauseListFlatFromIndex(ApproxUsableList, Index);
    clause_DeleteClauseListFlatFromIndex(ApproxWOList, Index);
    st_IndexDelete(Index);
    clause_SetCounter(OldClauseCounter);

    if (!list_Empty(EmptyClauses)) {
      list_DeleteWithElement(EmptyClauses,clause_Delete);
      return FALSE;
    } 

    return TRUE;

#ifdef DEBUG
    clause_Check(RedClause);
#endif
  }
  return FALSE;
}

LIST red_SatUnit(PROOFSEARCH Search, LIST ClauseList, BOOL Document) 
/*********************************************************
  INPUT:   A list of unshared clauses, proof search object and a flag for 
           setting DocProof plus a mode and <Derived> limiting the number
	   of derived clauses.
  RETURNS: A possibly empty list of empty clauses.
  EFFECT:  Does a shallow saturation of the conclauses depending on the
           flag_SATUNIT flag.
**********************************************************/
{
  CLAUSE Given,Clause;
  LIST   Scan, Derivables, EmptyClauses, BackReduced;
  NAT    n, Derived;

  Derived      = flag_GetFlagValue(flag_CNFPROOFSTEPS);
  EmptyClauses = list_Nil();

  ClauseList = clause_ListSortWeighed(ClauseList);
  
  while (!list_Empty(ClauseList) && list_Empty(EmptyClauses)) {
    Given = (CLAUSE)list_NCar(&ClauseList);
    Given = red_ReductionOnDerivedClause(Search, Given, Document, red_USABLE);
    if (Given)
      if (clause_IsEmptyClause(Given))
	EmptyClauses = list_List(Given);
      else {
	/*printf("\n\nGiven: ");clause_Print(Given);*/
	BackReduced = red_BackReduction(Search, Given, Document, red_USABLE);

	if (Derived != 0) {
	  Derivables = inf_BoundedDepthUnitResolution(Given, prfs_UsableSharingIndex(Search));
	  n          = list_Length(Derivables);
	  if (n > Derived)
	    Derived = 0;
	  else
	    Derived = Derived - n;
	}
	else 
	  Derivables = list_Nil();

	Derivables  = list_Nconc(BackReduced,Derivables);
	/*printf("\nDerived: ");clause_ListPrint(Derivables);*/
	Derivables  = split_ExtractEmptyClauses(Derivables, &EmptyClauses);
    
	prfs_InsertUsableClause(Search, Given);

	for(Scan = Derivables; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  Clause     = (CLAUSE)list_Car(Scan);
	  clause_SetDepth(Clause,0);
	}
	ClauseList = list_Nconc(ClauseList,Derivables);
	Derivables = list_Nil();
      }
  }
  for(Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan))
      prfs_InsertUsableClause(Search, list_Car(Scan));
  list_Delete(ClauseList);
  return EmptyClauses;
}


LIST red_SatInput(PROOFSEARCH Search, LIST ClauseList, BOOL Document, NAT *Derived) 
/*********************************************************
  INPUT:   A list of unshared clauses, proof search object and a flag for 
           setting DocProof
  EFFECT:  Does a one stop saturation from the conjectures into the
           axioms/conjectures
	   Keeps track of derived and kept clauses.
**********************************************************/
{
  CLAUSE Given;
  LIST   Scan, Derivables, EmptyClauses, BackReduced;

  EmptyClauses = list_Nil();

  ClauseList = clause_ListSortWeighed(ClauseList);
  
  while (!list_Empty(ClauseList) && list_Empty(EmptyClauses)) {
    Given = (CLAUSE)list_NCar(&ClauseList);
#if DEBUG
  if (!clause_IsClause(Given)) {
    fputs("\n\nIllegal input in red_SatInput\n", stderr);
    misc_DumpCore();
  }
#endif 
    Given = red_ReductionOnDerivedClause(Search, Given, Document, red_USABLE);
    if (Given) {
      prfs_IncreaseKeptClauses(Search);
      if (clause_IsEmptyClause(Given))
	EmptyClauses = list_List(Given);
      else {
	BackReduced = red_BackReduction(Search, Given, Document, red_USABLE);
	*Derived   += list_Length(BackReduced);
	BackReduced = split_ExtractEmptyClauses(BackReduced, &EmptyClauses);
	prfs_InsertUsableClause(Search, Given);
	BackReduced = clause_ListSortWeighed(BackReduced);
	ClauseList = list_NNumberMerge(ClauseList,BackReduced,clause_Weight);
      }
    }
  }
  for(Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan))
    prfs_InsertUsableClause(Search, list_Car(Scan));
  list_Delete(ClauseList);
  if (list_Empty(EmptyClauses) && flag_GetFlagValue(flag_SATINPUT)) {
    ClauseList = list_Nil();
    for(Scan=prfs_UsableClauses(Search);!list_Empty(Scan); Scan = list_Cdr(Scan)) {
      Given = (CLAUSE)list_Car(Scan);
      if (clause_GetFlag(Given,CONCLAUSE)) {
	Derivables = inf_BoundedDepthUnitResolution(Given, prfs_UsableSharingIndex(Search));
	ClauseList = list_Nconc(Derivables,ClauseList);
      }
    }
    ClauseList = split_ExtractEmptyClauses(ClauseList, &EmptyClauses);
    while (!list_Empty(ClauseList) && list_Empty(EmptyClauses)) {
      Given = (CLAUSE)list_NCar(&ClauseList);
      Given = red_ReductionOnDerivedClause(Search, Given, Document, red_USABLE);
      if (Given) {
	prfs_IncreaseKeptClauses(Search);
	if (clause_IsEmptyClause(Given))
	  EmptyClauses = list_List(Given);
	else {
	  BackReduced = red_BackReduction(Search, Given, Document, red_USABLE);
	  *Derived   += list_Length(BackReduced);
	  BackReduced = split_ExtractEmptyClauses(BackReduced, &EmptyClauses);
	  prfs_InsertUsableClause(Search, Given);
	  BackReduced = clause_ListSortWeighed(BackReduced);
	  ClauseList = list_NNumberMerge(ClauseList,BackReduced,clause_Weight);
	}
      }
    }
    for(Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan))
      prfs_InsertUsableClause(Search, list_Car(Scan));
    list_Delete(ClauseList);
  }
  return EmptyClauses;
}

