/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *           INFERENCE RULES FOR SORTS                    * */
/* *                                                        * */
/* *  $Module:   SORTRULES                                  * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.51 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/01 14:26:05 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: rules-sort.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "rules-sort.h"

     
/**************************************************************/
/* Functions                                                  */
/**************************************************************/ 

static LIST inf_GetForwardPartnerLits(LITERAL, st_INDEX);
static SORT inf_GetSortFromLits(LIST, SORTTHEORY);


static BOOL inf_SubsortPrecheck(CLAUSE Clause, LIST TLits, LITERAL Special,
				st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause, a list of constraint literal indices in
           that clause, a special literal, an index of clauses
	   and the actual sort theory.
  RETURNS: TRUE, if there exists any subsort of the <TLits> sort.
***************************************************************/
{
  SORT tSort, unifierSort;
  LIST unifiers;
  BOOL result;

  unifiers = inf_GetForwardPartnerLits(clause_GetLiteral(Clause,(int)list_Car(TLits)),
				       Index);
  unifierSort = inf_GetSortFromLits(unifiers, SortTheory);
  list_Delete(unifiers);

  tSort = sort_TopSort();
  for (; !list_Empty(TLits); TLits = list_Cdr(TLits)) {
    TERM actAtom = clause_GetLiteralAtom(Clause, (int)list_Car(TLits));
    tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory, term_TopSymbol(actAtom)),
			   tSort);
  }
  tSort = list_PointerDeleteDuplicates(tSort);

  if (Special == NULL)
    result = sort_TheoryIsSubsortOf(SortTheory, unifierSort, tSort);
  else {
    SORT extraSort = sort_TheorySortOfSymbol(SortTheory,
					     clause_LiteralPredicate(Special));

    result = sort_TheoryIsSubsortOfExtra(SortTheory, extraSort, unifierSort, tSort);
    sort_Delete(extraSort);
  }

  sort_Delete(tSort);
  sort_Delete(unifierSort);

  /*if (!result)
    printf("\n\t HAB IHN");*/

  return result;
}
     
static LIST inf_GetSortResolutionPartnerLits(TERM Atom, st_INDEX Index)
/**************************************************************
  INPUT:   A clause and an Index of clauses.
  RETURNS: A list of literals with which sortresolution is possible.
  MEMORY:  Allocates memory for the list.
***************************************************************/
{
  LIST    Result, TermList, LitScan;
  LITERAL NextLit;

#if DEBUG  
  if (!term_IsAtom(Atom)) {
    fputs("\nVariable as atom input in inf_GetSortResolutionPartnerLits.\n", stderr);
    misc_Error();
  }
#endif

  Result   = list_Nil();
  TermList = st_GetUnifier(cont_LeftContext(), Index, cont_RightContext(), Atom);
  
  /*printf("\nFor: "); term_PrintPrefix(Atom);*/
  
  for ( ; !list_Empty(TermList); TermList = list_Pop(TermList)) {
    
    if (term_IsAtom(list_Car(TermList))) {
      
      for (LitScan = sharing_NAtomDataList(list_Car(TermList)); 
	   !list_Empty(LitScan); 
	   LitScan = list_Cdr(LitScan)){
	NextLit = list_Car(LitScan);
	
	/*printf("\nWe found the PartnerLit: ");
	  clause_LiteralPrint(NextLit);
	  clause_Print(clause_LiteralOwningClause(NextLit));*/
	
	if (clause_LiteralIsPositive(NextLit) && 
	    clause_LiteralGetFlag(NextLit,STRICTMAXIMAL) &&
	    clause_GetFlag(clause_LiteralOwningClause(NextLit),WORKEDOFF) &&
	    clause_HasSolvedConstraint(clause_LiteralOwningClause(NextLit)) &&
	    !list_PointerMember(Result, NextLit)) 
	  Result = list_Cons(NextLit, Result);
      }
    }
  }
  
  return Result;
}


static CLAUSE inf_BuildConstraintHyperResolvent(CLAUSE Clause, LIST Lits,
						SUBST Subst, LIST Foundlits)
/**************************************************************
  INPUT:   A <Clause> where the sort constraint is resolved.
           A list <Lits> of constraint indicees in <Clause> where
	   all corresponding constraints have the same term.
	   The overall substitution <Subst>.
	   A list <Restlits> of constraint indecees for which
	   a partner clause is searched.
	   A list <Foundlits> of literals of already found partner clauses.
  RETURNS: The list of possible resolvents.
  EFFECT:  None.
***************************************************************/
{
  CLAUSE NewClause, ClauseCopy;
  LIST   Constraint, Antecedent, Succedent, ParentCls, ParentLits, Scan;
  TERM   Atom;
  SYMBOL MaxVar,MaxCand;
  int    i,bound, depth;
  BOOL   IsFromEmptySort;
  LIST   Partners;
  /* printf("\nBuildHyper mit Clause: "); clause_Print(Clause);
     printf("\nderen Lits: ");
     for (Constraint = Lits;
     !list_Empty(Constraint);
     Constraint = list_Cdr(Constraint))
     printf("%d  ",(int)list_Car(Constraint));
     printf("\nder Subst: "); subst_Print(Subst);
     printf("\nund Foundlits: "); clause_LiteralListPrint(Foundlits);DBG */
  
  ParentCls   = list_Nil();
  ParentLits  = list_Nil();
  Constraint  = list_Nil();
  Antecedent  = list_Nil();
  Succedent   = list_Nil();
  Partners    = list_Nil();
  depth       = clause_Depth(Clause);

  for (Scan=Foundlits; !list_Empty(Scan); Scan=list_Cdr(Scan))
    depth = misc_Max(depth,
		     clause_Depth(clause_LiteralOwningClause(list_Car(Scan))));

  ClauseCopy  = clause_Copy(Clause);
  Partners    = list_Cons(ClauseCopy, Partners);
  clause_SubstApply(Subst, ClauseCopy);

  IsFromEmptySort = term_IsVariable(term_FirstArgument(
    clause_GetLiteralAtom(Clause, (int)list_Car(Lits))));
  /* printf("\n %s",(IsFromEmptySort? "EmptySort" : "SortResolution"));DBG */

  bound = clause_LastConstraintLitIndex(ClauseCopy);

  for (i = clause_FirstLitIndex(); i <= bound; i++)
    if (!list_PointerMember(Lits, (POINTER)i)) {
      Atom  = term_Copy(clause_GetLiteralAtom(ClauseCopy, i));
      Constraint = list_Cons(Atom, Constraint);
      /*printf("\nnot regarded Constraint literal %d: ",i);
	term_Print(Atom); DBG */
    } else {
      ParentCls  = list_Cons((POINTER)clause_Number(ClauseCopy), ParentCls);
      ParentLits = list_Cons((POINTER)i, ParentLits);
    }

  bound = clause_LastAntecedentLitIndex(ClauseCopy);
  for ( ; i <= bound; i++) {
    Atom = term_Copy(clause_GetLiteralAtom(ClauseCopy, i));
    Antecedent = list_Cons(Atom, Antecedent);
  }
  bound = clause_LastSuccedentLitIndex(ClauseCopy);
  for (; i <= bound; i++) {
    Atom = term_Copy(clause_GetLiteralAtom(ClauseCopy, i));
    Succedent = list_Cons(Atom, Succedent);
  }
  bound = clause_LastConstraintLitIndex(Clause);
  for (i = clause_FirstLitIndex(); i <= bound; i++) {			
    /* Hier sollen die gematchten Constraintliterale dazu fuehren, dass die */
    /* c,a und s- literale der Partnerclauses in die Listen kommen...       */

    if (list_PointerMember(Lits, (POINTER)i)) {
      CLAUSE  PartnerCopy;
      LITERAL PLit;
      TERM    PAtom;
      SUBST   NewSubst,RightSubst;
      int     j,lc,la,n,PLitInd;

      Atom      = clause_GetLiteralAtom(ClauseCopy, i);
      NewClause = clause_CreateUnnormalized(Constraint, Antecedent, Succedent);
      /* printf("\nNewClause at for-loop entry: ");
	 clause_Print(NewClause);DBG */

      list_Delete(Constraint);
      list_Delete(Antecedent);
      list_Delete(Succedent);
      Constraint  = list_Nil();
      Antecedent  = list_Nil();
      Succedent   = list_Nil();

      /*printf("\nIn BuildHyper suchen wir: "); term_Print(Atom); 
	printf("\nin der Foundlits-Liste: ");
	clause_LiteralListPrint(Foundlits);DBG*/
      
      /* Find corresponding Foundlit: */
      for (Scan = Foundlits; 
	   term_TopSymbol(Atom) !=
	     term_TopSymbol(clause_LiteralAtom(list_Car(Scan))); 
	   Scan = list_Cdr(Scan));
      PLit        = list_Car(Scan);
      PLitInd     = clause_LiteralGetIndex(PLit);
      PartnerCopy = clause_Copy(clause_LiteralOwningClause(PLit));
      Partners    = list_Cons(PartnerCopy, Partners);
      ParentCls   = list_Cons((POINTER)clause_Number(PartnerCopy), ParentCls);
      ParentLits  = list_Cons((POINTER)PLitInd, ParentLits);
      MaxVar      = clause_SearchMaxVar(ClauseCopy);
      MaxCand     = clause_SearchMaxVar(NewClause);
      MaxVar      = ((MaxVar > MaxCand) ? MaxVar : MaxCand);   
      /* MaxVar is the maximal variable in the new clause or the ClauseCopy, */
      /* the latter to guarantee the stability of variable names.            */
      /* printf("\nThe Maxvar: "); symbol_Print(MaxVar);DBG */

      clause_RenameVarsBiggerThan(PartnerCopy, MaxVar);
      PLit  = clause_GetLiteral(PartnerCopy, PLitInd);
      PAtom = clause_LiteralAtom(PLit);
      /* printf("\nrenamed PartnerCopy: "); clause_Print(PartnerCopy);
	 printf("\nAtom: "); term_Print(Atom);
	 printf("  PAtom: "); term_Print(PAtom);DBG */
      
      /* if (!unify_Unify(PAtom, Atom, &Bindings)) {
	 printf("\n \t\tHAHAHA \n");
	 printf("\n %s",(IsFromEmptySort? "EmptySort" : "SortResolution"));
	 term_Print(Atom);
	 printf("\n"); term_Print(PAtom);
	 printf("\nfrom renamed PartnerCopy: "); clause_Print(PartnerCopy);
	 printf("\nAlso BuildHyper mit Clause: "); clause_Print(Clause);
	 printf("\nderen Lits: ");
	 for (DBGScan=Lits; !list_Empty(DBGScan); DBGScan = list_Cdr(DBGScan))
	 printf("%d  ",(int)list_Car(DBGScan));
	 printf("\nder Subst: "); subst_Print(Subst);
	 printf("\nund Foundlits: "); 
	 for (DBGScan=Foundlits;!list_Empty(DBGScan);
	 DBGScan =list_Cdr(DBGScan)) {
	 printf("\n"); clause_LiteralPrint(list_Car(DBGScan));
	 printf("\tfrom: ");
	 clause_Print(clause_LiteralOwningClause(list_Car(DBGScan)));
	 
	 }
	 }DBG */
      
      cont_Check();
      unify_UnifyNoOC(cont_LeftContext(), PAtom, cont_RightContext(), Atom);
      subst_ExtractUnifier(cont_LeftContext(), &RightSubst, cont_RightContext(), &NewSubst);
      cont_Reset();

      /* printf("\nUnifier for NewClause    : "); subst_Print(NewSubst);*/
      /*printf("\nUnifier for partnerclause: "); subst_Print(RightSubst);DBG */

      clause_SubstApply(NewSubst, NewClause);
      clause_SubstApply(NewSubst, ClauseCopy);
      subst_Delete(NewSubst);
      
      n  = clause_Length(PartnerCopy);
      lc = clause_LastConstraintLitIndex(PartnerCopy);
      la = clause_LastAntecedentLitIndex(PartnerCopy);
      for (j = clause_FirstLitIndex(); j < n; j++) {
	if (j <= lc) 
	  Constraint  = list_Cons(subst_Apply(RightSubst,
	    term_Copy(clause_GetLiteralAtom(PartnerCopy, j))),
	      Constraint);
	else if (j <= la) 
	  Antecedent  = list_Cons(subst_Apply(RightSubst,
            term_Copy(clause_GetLiteralAtom(PartnerCopy, j))),
               Antecedent);
	else if (j != PLitInd)
	  Succedent  = list_Cons(subst_Apply(RightSubst,
            term_Copy(clause_GetLiteralAtom(PartnerCopy, j))),
               Succedent);
      }

      /* printf("\nNew atoms from C:\n");
      term_TermListPrint(Constraint);
      printf("\n               A:\n");
      term_TermListPrint(Antecedent);
      printf("\n               S:\n");
      term_TermListPrint(Succedent);DBG */
      
      subst_Delete(RightSubst);
      /*      clause_Delete(PartnerCopy); */
      
      n  = clause_Length(NewClause);
      lc = clause_LastConstraintLitIndex(NewClause);
      la = clause_LastAntecedentLitIndex(NewClause);
      
      for (j = clause_FirstLitIndex(); j < n; j++) {
	if (j <= lc) 
	  Constraint  = list_Cons(term_Copy(clause_GetLiteralAtom(NewClause, j)),
				  Constraint);
	else if (j <= la) 
	  Antecedent  = list_Cons(term_Copy(clause_GetLiteralAtom(NewClause, j)),
				  Antecedent);
	else 
	  Succedent  = list_Cons(term_Copy(clause_GetLiteralAtom(NewClause, j)),
				 Succedent);
      }      
      clause_Delete(NewClause);
      clause_DecreaseCounter();
    }
  }
  /*  clause_Delete(ClauseCopy); */

  /* printf("\nNewClause aus C:\n");
  term_TermListPrint(Constraint);
  printf("\n              A:\n");
  term_TermListPrint(Antecedent);
  printf("\n              S:\n");
  term_TermListPrint(Succedent);DBG */

  NewClause = clause_Create(Constraint, Antecedent, Succedent);

  list_Delete(Constraint); list_Delete(Antecedent); list_Delete(Succedent); 

  if (IsFromEmptySort)
    clause_SetFromEmptySort(NewClause);
  else
    clause_SetFromSortResolution(NewClause);

  clause_SetDepth(NewClause, depth + 1);

  clause_SetSplitDataFromList(NewClause, Partners);
  list_DeleteWithElement(Partners, clause_Delete);

  clause_SetParentClauses(NewClause, list_NReverse(ParentCls));
  clause_SetParentLiterals(NewClause, list_NReverse(ParentLits));

  /* printf("\nAus BuildHyper kommt: "); clause_Print(NewClause);DBG */
  return NewClause;
}


static LIST inf_ConstraintHyperResolvents(CLAUSE Clause, LIST Lits,
					  SUBST Subst, LIST Restlits,
					  LIST Foundlits, st_INDEX Index)
/**************************************************************
  INPUT:   A <Clause> where the sort constraint is resolved.
           A list <Lits> of constraint indicees in <Clause> where
	   all corresponding constraints have the same term.
	   The overall substitution <Subst>.
	   A list <Restlits> of constraint indecees for which
	   a partner clause is searched.
	   A list <Foundlits> of literals of already found partner clauses.
  RETURNS: The list of possible resolvents.
  EFFECT:  None.
***************************************************************/
{
  if (list_Empty(Restlits))
    return list_List(inf_BuildConstraintHyperResolvent(Clause,Lits,Subst,
						       Foundlits));
  else {
    CLAUSE  PartnerCopy;
    LITERAL Lit, PLit;
    LIST    Result, NextLits;
    TERM    AtomCopy;
    SUBST   NewSubst, RightSubst, HelpSubst;
    SYMBOL  MaxVar, MaxCand;
    int     PLitInd;

    Result   = list_Nil();
    Lit      = clause_GetLiteral(Clause, (int) list_Car(Restlits));
    AtomCopy = subst_Apply(Subst, term_Copy(clause_LiteralAtom(Lit)));
    NextLits = inf_GetSortResolutionPartnerLits(AtomCopy, Index);
    MaxVar   = clause_MaxVar(Clause);
    MaxCand  = clause_AtomMaxVar(AtomCopy);
    MaxVar   = (symbol_GreaterVariable(MaxVar, MaxCand) ? MaxVar : MaxCand);
      
    for ( ; !list_Empty(NextLits); NextLits = list_Pop(NextLits)) {
      PLit        = list_Car(NextLits);
      PLitInd     = clause_LiteralGetIndex(PLit);
      Foundlits   = list_Cons(PLit, Foundlits);
      PartnerCopy = clause_Copy(clause_LiteralOwningClause(PLit));

      clause_RenameVarsBiggerThan(PartnerCopy, MaxVar);
      PLit        = clause_GetLiteral(PartnerCopy, PLitInd);

      cont_Check();
      unify_UnifyNoOC(cont_LeftContext(), AtomCopy,
		      cont_RightContext(), clause_LiteralAtom(PLit));
      subst_ExtractUnifier(cont_LeftContext(), &NewSubst, cont_RightContext(), &RightSubst);
      cont_Reset();

      subst_Delete(RightSubst);
      HelpSubst = NewSubst;
      NewSubst  = subst_Merge(NewSubst, subst_Copy(Subst));

      Result = list_Nconc(inf_ConstraintHyperResolvents(Clause, Lits, NewSubst,
							list_Cdr(Restlits), Foundlits,
							Index),
			  Result);
      subst_Delete(NewSubst);
      subst_Delete(HelpSubst);
      clause_Delete(PartnerCopy);

      Foundlits = list_Pop(Foundlits);
    }
    term_Delete(AtomCopy);

    return Result;
  }
}


LIST inf_BackwardSortResolution(CLAUSE GivenClause, st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause with a solved sort constraint and an 'Index' of clauses.
  RETURNS: A list of clauses inferred from the GivenClause by
           SortResolution with the given clause.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST     result;
  int      i, ls;

#if DEBUG
  if (!clause_IsClause(GivenClause) || !clause_HasSolvedConstraint(GivenClause)) {
    fputs("\n\nIllegal input in inf_BackwardSortResolution\n", stderr);
    misc_Error();
  }
#endif

  result = list_Nil();
  ls     = clause_LastSuccedentLitIndex(GivenClause);

  for (i = clause_FirstSuccedentLitIndex(GivenClause); i <= ls; i++) {
    LITERAL pLit  = clause_GetLiteral(GivenClause, i);
    TERM    pAtom = clause_LiteralAtom(pLit);

    if (clause_LiteralGetFlag(pLit,STRICTMAXIMAL) && 
	clause_LiteralIsSort(pLit)) {
      LIST termList = st_GetUnifier(cont_LeftContext(),Index,cont_RightContext(),pAtom);

      for ( ; !list_Empty(termList); termList = list_Pop(termList)){
	if (term_IsAtom(list_Car(termList)) &&
	    !term_IsVariable(term_FirstArgument(list_Car(termList)))) {

	  LIST litScan = sharing_NAtomDataList(list_Car(termList));
	  for ( ; !list_Empty(litScan); litScan = list_Cdr(litScan)) {

	    LITERAL gLit   = list_Car(litScan);
	    CLAUSE gClause = clause_LiteralOwningClause(gLit);

	    if (clause_LiteralGetIndex(gLit) < clause_FirstAntecedentLitIndex(gClause) && 
		clause_GetFlag(gClause,WORKEDOFF)) {
	      TERM   gAtom     = clause_LiteralAtom(gLit);
	      int    lc        = clause_LastConstraintLitIndex(gClause);
	      int    gi        = clause_LiteralGetIndex(gLit);
	      LIST   tLits     = list_List((POINTER)gi);
	      LIST   restLits  = list_Nil();
	      int    j;

	      for (j = clause_FirstLitIndex(); j <= lc; j++) {
		LITERAL tCand = clause_GetLiteral(gClause, j);

		if (j != gi &&
		    term_FirstArgument(clause_LiteralAtom(tCand))
		    == term_FirstArgument(gAtom)) {
		  tLits     = list_Cons((POINTER)j, tLits);
		  restLits  = list_Cons((POINTER)j, restLits);
		}
	      }

	      if (inf_SubsortPrecheck(gClause, tLits, pLit , Index, SortTheory)) {
		CLAUSE pClauseCopy = clause_Copy(GivenClause);
		SYMBOL minVar      = clause_MaxVar(gClause);
		LIST   foundLits   = list_List(pLit);
		SUBST  leftSubst, rightSubst;


		clause_RenameVarsBiggerThan(pClauseCopy, minVar);
		pAtom = clause_GetLiteralAtom(pClauseCopy, i);
		/* set, to unify correctly! */

		cont_Check();
		unify_UnifyNoOC(cont_LeftContext(), gAtom, cont_RightContext(), pAtom);
		subst_ExtractUnifier(cont_LeftContext(), &leftSubst,
				     cont_RightContext(), &rightSubst);
		cont_Reset();

		subst_Delete(rightSubst);

		result =
		  list_Nconc(inf_ConstraintHyperResolvents(gClause, tLits,
							   leftSubst, restLits,
							   foundLits, Index),
			     result);

		pAtom = clause_LiteralAtom(pLit);

		subst_Delete(leftSubst);
		list_Delete(foundLits);
		clause_Delete(pClauseCopy);
	      } /* if Precheck */
	      list_Delete(tLits);
	      list_Delete(restLits);
	    }
	  }
	}
      }
    }
  }
  return result;
}


LIST inf_ForwardSortResolution(CLAUSE GivenClause, st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause with an unsolved sort constraint and an 'Index' of clauses.
  RETURNS: A list of clauses inferred from the GivenClause by
           SortResolution on the given clause.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST    result;
  int     i,lc;
  BOOL    hit;

#if DEBUG
  if (!clause_IsClause(GivenClause) || !clause_HasTermSortConstraintLits(GivenClause)) {
    fputs("\n\nIllegal input in inf_ForwardSortResolution: ", stderr);
    if (clause_IsClause(GivenClause))
      clause_FPrint(stderr, GivenClause);
    putc('\n', stderr);
    misc_Error();
  }
#endif

  result = list_Nil();
  lc     = clause_LastConstraintLitIndex(GivenClause);
  hit    = FALSE;

  for (i = clause_FirstLitIndex(); i <= lc && !hit; i++) {

    if (!term_IsVariable(term_FirstArgument(clause_GetLiteralAtom(GivenClause, i)))) {
      TERM tAtom    = clause_GetLiteralAtom(GivenClause, i);
      LIST tLits, restLits;
      int  j;

      hit = TRUE;

      /* Search the other T_i from <GivenClause> */
      tLits    = list_List((POINTER)i);
      restLits = list_Nil();
      for (j = i+1; j <= lc; j++) {
	if (term_FirstArgument(clause_GetLiteralAtom(GivenClause, j))
	    == term_FirstArgument(tAtom)) {
	  tLits    = list_Cons((POINTER)j, tLits);
	  restLits = list_Cons((POINTER)j, restLits);
	}
      }

      if (inf_SubsortPrecheck(GivenClause, tLits, NULL, Index, SortTheory)) {
	LIST unifiers = inf_GetSortResolutionPartnerLits(tAtom, Index);

	/*printf("\n Forward Sort Resolution: "); clause_Print(GivenClause);*/

	for ( ; !list_Empty(unifiers); unifiers = list_Pop(unifiers)) {
	  SYMBOL  minVar    = clause_MaxVar(GivenClause);
	  LITERAL pLit      = list_Car(unifiers);
	  LIST    foundLits = list_List(pLit);
	  int     pi        = clause_LiteralGetIndex(pLit);
	  CLAUSE  pCopy     = clause_Copy(clause_LiteralOwningClause(pLit));
	  SUBST   leftSubst, rightSubst;

	  clause_RenameVarsBiggerThan(pCopy, minVar);

	  cont_Check();
	  unify_UnifyNoOC(cont_LeftContext(),  tAtom,
			  cont_RightContext(), clause_GetLiteralAtom(pCopy, pi));
	  subst_ExtractUnifier(cont_LeftContext(), &leftSubst, cont_RightContext(), &rightSubst);
	  cont_Reset();

	  subst_Delete(rightSubst);

	  result =
	    list_Nconc(inf_ConstraintHyperResolvents(GivenClause, tLits, leftSubst, restLits,
						     foundLits, Index),
		       result);
	  list_Delete(foundLits);
	  clause_Delete(pCopy);
	  subst_Delete(leftSubst);
	} /*end of for (Scan = FirstLitPartnerLits ... */
      }  /* if Precheck */
      list_Delete(restLits);
      list_Delete(tLits);
    } /*end of if 'literal is ready' */ 
  }
  return result;
}


LIST inf_BackwardEmptySort(CLAUSE GivenClause, st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause with a solved sort constraint and an 'Index' of clauses.
  RETURNS: A list of clauses inferred from the GivenClause by
           EmptySort with the given clause.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST result;
  int  i, ls;

#if DEBUG
  if (!(clause_IsClause(GivenClause)) || !clause_HasSolvedConstraint(GivenClause)) {
    printf("\n\nIllegal input in inf_BackwardEmptySort\n");
    misc_Error();
  }
#endif

  result = list_Nil();
  
  ls = clause_LastSuccedentLitIndex(GivenClause);

  for (i = clause_FirstSuccedentLitIndex(GivenClause); i <= ls; i++) {
    LITERAL pLit  = clause_GetLiteral(GivenClause, i);
    TERM    pAtom = clause_LiteralAtom(pLit);
      
    if (clause_LiteralGetFlag(pLit,STRICTMAXIMAL) && 
	clause_LiteralIsSort(pLit)) {
      LIST unifiers = st_GetUnifier(cont_LeftContext(),Index,cont_RightContext(),pAtom);
	
      for ( ; !list_Empty(unifiers); unifiers = list_Pop(unifiers)){
	if (term_IsAtom(list_Car(unifiers)) &&
	    term_IsVariable(term_FirstArgument(list_Car(unifiers)))) {
	  LIST litScan = sharing_NAtomDataList(list_Car(unifiers));
	    
	  for ( ; !list_Empty(litScan); litScan = list_Cdr(litScan)){
	    LITERAL gLit   = list_Car(litScan);
	    CLAUSE gClause = clause_LiteralOwningClause(gLit);
	      
	    if (clause_LiteralGetIndex(gLit) < clause_FirstAntecedentLitIndex(gClause) &&
		clause_GetFlag(gClause,WORKEDOFF) &&
		clause_HasOnlyVarsInConstraint(gClause)) {
	      TERM   gAtom     = clause_LiteralAtom(gLit);
	      SYMBOL var       = term_TopSymbol(term_FirstArgument(gAtom));
	      int    lc        = clause_LastConstraintLitIndex(gClause);
	      int    gi        = clause_LiteralGetIndex(gLit);
	      BOOL   varOccursNoMore;
	      int    j, bound;

	      varOccursNoMore = TRUE;
	      bound           = clause_LastSuccedentLitIndex(gClause);
		
	      for (j = clause_FirstAntecedentLitIndex(gClause);
		   (j <= bound) && varOccursNoMore;
		   j++) {
		if (term_ContainsSymbol(clause_GetLiteralAtom(gClause, j), var))
		  varOccursNoMore = FALSE;
	      }
		
	      if (varOccursNoMore) {
		LIST tLits, restLits;

		/* Search the other T_i from <gClause> */
		tLits     = list_List((POINTER)gi);
		restLits  = list_Nil();		
		for (j = clause_FirstLitIndex(); j <= lc; j++) {
		  LITERAL tCand = clause_GetLiteral(gClause, j);

		  if (j != gi &&
		      term_FirstArgument(clause_LiteralAtom(tCand))
		      == term_FirstArgument(gAtom)) {
		    tLits     = list_Cons((POINTER)j, tLits);
		    restLits  = list_Cons((POINTER)j, restLits);
		  }
		}

		if (inf_SubsortPrecheck(gClause, tLits, pLit, Index, SortTheory)) {
		  CLAUSE pCopy     = clause_Copy(GivenClause);
		  SYMBOL minVar    = clause_MaxVar(gClause);
		  LIST   foundLits = list_List(pLit);
		  SUBST  leftSubst, rightSubst;

		  clause_RenameVarsBiggerThan(pCopy, minVar);
		  pAtom = clause_GetLiteralAtom(pCopy, i);
		  /* set, to adress the renamed term! */
		  
		  cont_Check();
		  unify_UnifyNoOC(cont_LeftContext(), gAtom, cont_RightContext(), pAtom);
		  subst_ExtractUnifier(cont_LeftContext(),  &leftSubst,
				       cont_RightContext(), &rightSubst);
		  cont_Reset();

		  subst_Delete(rightSubst);
		  
		  result =
		    list_Nconc(inf_ConstraintHyperResolvents(gClause, tLits, leftSubst,
							     restLits, foundLits, Index),
			       result);

		  list_Delete(foundLits);
		  subst_Delete(leftSubst);
		  clause_Delete(pCopy);
		  
		  pAtom = clause_LiteralAtom(pLit);
		  /* reset to original clauses literal! */
		} /* if Precheck */
		list_Delete(tLits);
		list_Delete(restLits);
	      }
	    }
	  }
	}
      }
    }
  }
  return result;
}


LIST inf_ForwardEmptySort(CLAUSE GivenClause, st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause and an 'Index' of clauses, the constraint of 
           'GivenClause' is necessarily unsolved
  RETURNS: A list of clauses inferred from the GivenClause by
           EmptySort on the given clause.
  MEMORY:  A list of clauses is produced, where memory for the list
           and the clauses is allocated.
***************************************************************/
{
  LIST result;
  int  i, lc;
  BOOL hit;

#if DEBUG
  if (clause_HasTermSortConstraintLits(GivenClause) ||
      clause_HasSolvedConstraint(GivenClause)) {
    fputs("\n\nIllegal input in inf_ForwardEmptySort\n", stderr);
    misc_Error();
  }
#endif
  
  /* printf("\nWe try ForwardEmptySort with: "); clause_Print(GivenClause); */
  
  result = list_Nil();
  lc     = clause_LastConstraintLitIndex(GivenClause);
  hit    = FALSE;
  
  for (i = clause_FirstLitIndex(); i <= lc; i++) {

    if (term_IsVariable(term_FirstArgument(clause_GetLiteralAtom(GivenClause,i)))) {
      TERM   tAtom = clause_GetLiteralAtom(GivenClause, i);
      SYMBOL var   = term_TopSymbol(term_FirstArgument(tAtom));
      int    j, ls;
      BOOL   varOccursNoMore;

      ls = clause_LastSuccedentLitIndex(GivenClause);
      varOccursNoMore = TRUE;
      /* Check if the variable occurs in antecedent or succedent literals */
      for (j = clause_FirstAntecedentLitIndex(GivenClause); 
	   (j <= ls) && varOccursNoMore; j++) {
	if (term_ContainsSymbol(clause_GetLiteralAtom(GivenClause,j), var))
	  varOccursNoMore = FALSE;
      }
		
      if (varOccursNoMore) {
	LIST tLits, restLits;

	/* Consider only the first appropriate literal */
	hit = TRUE;
	
	/* Search the other T_i(t) literals */
	tLits    = list_List((POINTER)i);
	restLits = list_Nil();
	for (j = i+1; j <= lc; j++) {
	  TERM tCand = clause_GetLiteralAtom(GivenClause, j);

	  if (symbol_Equal(term_TopSymbol(term_FirstArgument(tCand)), var)) {
	    tLits    = list_Cons((POINTER)j, tLits);
	    restLits = list_Cons((POINTER)j, restLits);
	  }
	}

	if (inf_SubsortPrecheck(GivenClause, tLits, NULL, Index, SortTheory)) {
	  LIST unifiers;
	  SYMBOL  minVar;

	  unifiers = inf_GetSortResolutionPartnerLits(tAtom, Index);

	  minVar   = clause_MaxVar(GivenClause);

	  for ( ; !list_Empty(unifiers); unifiers = list_Pop(unifiers)) {
	    LITERAL partnerLit = list_Car(unifiers);
	    LIST    foundLits  = list_List(partnerLit);
	    int     pi         = clause_LiteralGetIndex(partnerLit);
	    CLAUSE  pCopy      = clause_Copy(clause_LiteralOwningClause(partnerLit));
	    SUBST   leftSubst, rightSubst;

	    clause_RenameVarsBiggerThan(pCopy, minVar);

	    cont_Check();
	    unify_UnifyNoOC(cont_LeftContext(), tAtom,
			    cont_RightContext(), clause_GetLiteralAtom(pCopy, pi));
	    subst_ExtractUnifier(cont_LeftContext(), &leftSubst, cont_RightContext(), &rightSubst);
	    cont_Reset();

	    subst_Delete(rightSubst);

	    result =
	      list_Nconc(inf_ConstraintHyperResolvents(GivenClause, tLits, leftSubst, restLits,
						       foundLits, Index),
			 result);
	    list_Delete(foundLits);
	    clause_Delete(pCopy);
	    subst_Delete(leftSubst);
	  } /*end of for (unifiers ... */
	}   /* if Precheck */
	list_Delete(restLits);
	list_Delete(tLits);
      } /* varOccursNoMore */
    }
  } /* for all constraint lits */ 
  
  return result;
}

/**************************************************************/
/*NEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEUNEU*/
/**************************************************************/

LIST inf_GetSimpleSubsets(LIST Unifiers)
/**************************************************************
  INPUT:   A list of monadic succedent literals.
  RETURNS: A list of subsets of the input list. Each subset
           contains all literals with a variable and some
	   literals whose subterm has the same topsymbol.
	   The subset are represented as lists, so the returned
	   list is a list of lists.
  MEMORY:  Allocates memory for the lists.
  CAUTION: The input list is destroyed!
***************************************************************/
{
  /* Work to do! */



  return list_Nil();
}

static LIST inf_GetForwardPartnerLits(LITERAL Literal, st_INDEX Index)
/**************************************************************
  INPUT:   A monadic literal and an index of clauses.
  RETURNS: A list of monadic succedent literals whose subterm
           is unifiable with the (one) subterm of <Literal>.
	   The literals are strict maximal in their respective clauses,
	   the clauses are "WORKEDOFF" and either the subterm
	   is not a variable or the clause has an empty constraint.
  MEMORY:  Allocates memory for the list.
***************************************************************/
{
  LIST result, unifiers, atomScan, litScan;

#if DEBUG
  if (!clause_LiteralIsLiteral(Literal) || !clause_LiteralIsSort(Literal)) {
    fputs("\n\nIllegal Input in inf_GetForwardPartnerLits\n", stderr);
    misc_Error();
  }
#endif

  result   = list_Nil();
  /* Search unifiers for the literal's subterm */
  unifiers = st_GetUnifier(cont_LeftContext(), Index, cont_RightContext(),
			   term_FirstArgument(clause_LiteralAtom(Literal)));
  
  for ( ; !list_Empty(unifiers); unifiers = list_Pop(unifiers)) {

    if (!term_IsAtom(list_Car(unifiers))) { /* Can happen if arg is variable */
      
      for (atomScan = term_SupertermList(list_Car(unifiers)); 
	   !list_Empty(atomScan); 
	   atomScan = list_Cdr(atomScan)) {
	TERM atomCand = (TERM) list_Car(atomScan);

	if (term_IsDeclaration(atomCand)) {
	  /* We are looking for an unary atom */

	  for (litScan = sharing_NAtomDataList(atomCand);
	       !list_Empty(litScan);
	       litScan = list_Cdr(litScan)) {
	    LITERAL nextLit    = list_Car(litScan);
	    CLAUSE  nextClause = clause_LiteralOwningClause(nextLit);
	    
	    /* printf("\nWe found the PartnerLit: "); */
	    /* clause_LiteralPrint(nextLit); */
	    /* clause_Print(clause_LiteralOwningClause(nextLit)); */

	    if (clause_LiteralIsPositive(nextLit) && 
		clause_LiteralGetFlag(nextLit,STRICTMAXIMAL) &&
		clause_GetFlag(nextClause, WORKEDOFF) &&
		(!term_IsVariable(list_Car(unifiers)) ||
		 clause_HasEmptyConstraint(nextClause)) &&
		clause_HasSolvedConstraint(nextClause)) {
	      /* Add the literal from the copied clause */
	      result = list_Cons(nextLit, result);
	    }
	  }  /* litScan */
	}    /* if IsAtom */
      }      /* atomScan */
    }        /* ! variable */
  }          /* unifiers */
  return result;
}

static BOOL inf_LiteralsHaveSameSubtermAndAreFromSameClause(LITERAL L1, LITERAL L2)
/**************************************************************
  INPUT:   Two literals.
  RETURNS: TRUE, if both literals have the same term and are
           from the same clause, FALSE otherwise.
	   Since both literals are shared, pointer equality
	   is used to detect this.
	   This function is used by inf_GetBackwardPartnerLits().
***************************************************************/
{
  return (term_FirstArgument(clause_LiteralAtom(L1))
	  == term_FirstArgument(clause_LiteralAtom(L2)) &&
	  clause_LiteralOwningClause(L1) == clause_LiteralOwningClause(L2));
}

static void inf_GetBackwardPartnerLits(LITERAL Literal, st_INDEX Index,
				       LIST* ConstraintLits, LIST* Unifiers,
				       BOOL IsFromEmptySort)
/**************************************************************
  INPUT:   A literal, an index of clauses, two pointers to lists
           and a boolean value.
  RETURNS:
  MEMORY:  Allocates memory for the list.
***************************************************************/
{
  LIST candidates, atomScan, litScan;
  LITERAL nextLit;
  CLAUSE  nextClause;

#if DEBUG
  if (!clause_LiteralIsLiteral(Literal) || !clause_LiteralIsSort(Literal)) {
    fputs("\n\nIllegal input inf inf_GetBackwardPartnerLits\n", stderr);
    misc_Error();
  }
#endif

  candidates = st_GetUnifier(cont_LeftContext(), Index, cont_RightContext(),
			     term_FirstArgument(clause_LiteralAtom(Literal)));

  for ( ; !list_Empty(candidates); candidates = list_Pop(candidates)) {
    if (!term_IsAtom(list_Car(candidates))) { /* May happen if arg is variable */
      /* Consider variable unifiers only if called from BackwardEmptySort */
      for (atomScan = term_SupertermList(list_Car(candidates)); 
	   !list_Empty(atomScan); 
	   atomScan = list_Cdr(atomScan)) {
	TERM atomCand = (TERM) list_Car(atomScan);

	if (term_IsDeclaration(atomCand)) {
	  /* We are looking for unary atoms */

	  for (litScan = sharing_NAtomDataList(atomCand);
	       !list_Empty(litScan);
	       litScan = list_Cdr(litScan)) {
	    nextLit    = list_Car(litScan);
	    nextClause = clause_LiteralOwningClause(nextLit);
	    
	    if (clause_GetFlag(nextClause, WORKEDOFF)) {
	      if (clause_LiteralIsPositive(nextLit)) {
		if (clause_LiteralGetFlag(nextLit,STRICTMAXIMAL) &&
		    (!term_IsVariable(list_Car(candidates)) ||
		     clause_HasEmptyConstraint(nextClause)) &&
		    clause_HasSolvedConstraint(nextClause) &&
		    !symbol_Equal(clause_LiteralPredicate(Literal),
				  clause_LiteralPredicate(nextLit))) {
		  /* Don't consider literals with same top symbol as given literal */
		  /* since the given clause must be part of any inference */
		  *Unifiers = list_Cons(nextLit, *Unifiers);
		}
	      } else if (clause_LiteralGetIndex(nextLit) < clause_FirstAntecedentLitIndex(nextClause) &&
			 ((!term_IsVariable(list_Car(candidates)) && !IsFromEmptySort) ||
			  (term_IsVariable(list_Car(candidates)) && IsFromEmptySort &&
			   clause_HasOnlyVarsInConstraint(nextClause)))) {
		*ConstraintLits = list_Cons(nextLit, *ConstraintLits);
	      }
	    }
	  } /* litScan */
	}
      }     /* atomScan */
    }
  }         /* candidates */

  /* We have to avoid constraint literals from the same clause with the same
     term or variable, since those would create the same result clause. */
  *ConstraintLits =
    list_DeleteDuplicates(*ConstraintLits,
			  inf_LiteralsHaveSameSubtermAndAreFromSameClause);
}

static void inf_MakeClausesDisjoint(CLAUSE GClause, LIST Unifiers)
{
  SYMBOL maxVar, maxCand;
  CLAUSE lastClause;

  maxVar     = clause_MaxVar(GClause);
  lastClause = clause_LiteralOwningClause(list_Car(Unifiers));
  clause_RenameVarsBiggerThan(lastClause, maxVar);
  Unifiers   = list_Cdr(Unifiers);

  for ( ; !list_Empty(Unifiers); Unifiers = list_Cdr(Unifiers)) {
    CLAUSE actClause;

    clause_UpdateMaxVar(lastClause);
    maxCand = clause_MaxVar(lastClause);    
    maxVar = (symbol_GreaterVariable(maxVar,maxCand)? maxVar : maxCand);

    actClause = clause_LiteralOwningClause(list_Car(Unifiers));
    clause_RenameVarsBiggerThan(actClause, maxVar);
  }
}

static void inf_CopyUnifierClauses(LIST Unifiers)
{
  for ( ; !list_Empty(Unifiers); Unifiers = list_Cdr(Unifiers)) {
    CLAUSE actClause;
    int    i;

    actClause = clause_LiteralOwningClause(list_Car(Unifiers));
    i         = clause_LiteralGetIndex(list_Car(Unifiers));
    actClause = clause_Copy(actClause);
    list_Rplaca(Unifiers, clause_GetLiteral(actClause, i)); /* Set to literal from copy */
  }
}

static void inf_DeleteUnifierClauses(LIST Unifiers)
{
  for ( ; !list_Empty(Unifiers); Unifiers = list_Cdr(Unifiers)) {
    clause_Delete(clause_LiteralOwningClause(list_Car(Unifiers)));
    list_Rplaca(Unifiers, NULL);
  }
}

static SORT inf_GetSortFromLits(LIST Literals, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A list of literals.
  RETURNS: The sort created from the literals' predicates.
***************************************************************/
{
  SORT result = sort_TopSort();

  for ( ; !list_Empty(Literals); Literals = list_Cdr(Literals)) {
    SORT newSort = sort_TheorySortOfSymbol(SortTheory,
					   clause_LiteralPredicate(list_Car(Literals)));
    
    result = sort_Intersect(newSort, result);
  }

  list_PointerDeleteDuplicates(result);
  /* Necessary for Christoph's funktion */

  return result;
}

static LIST inf_ApplyWeakening(CLAUSE Clause, LIST TLits, LIST Partners,
			       CONDITION Condition)
/**************************************************************
  INPUT:   A clause, a list of constraint indices in <Clause>,
           a list of maximal, monadic succedent literals,
	   an a subsort condition.
  RETURNS: A one-element list with a clause derived from the
           clause and the partner clauses by the Weakening or
	   Empty Sort rule.
  MEMORY:  Memory is allocated for the returned list and the clause.
***************************************************************/
{
  LIST   scan, parents;
  LIST   constraint, antecedent, succedent;
  LIST   parentClauses, parentLits;   /* clause numbers and literal indices */
  int    i, bound, depth;
  TERM   tSubterm;
  CLAUSE newClause;

#if DEBUG
  if (!clause_IsClause(Clause) || list_Empty(TLits) || list_Empty(Partners) ||
      Condition == NULL) {
    fputs("\nIllegal input in inf_ApplyWeakening\n", stderr);
    misc_Error();
  }
#endif

  /* putchar('\n'); clause_Print(Clause); fputs("\nT_k = ", stdout); */
  /* for (scan = List; scan != NULL; scan = scan->next) { */
  /*   clause_LiteralPrint(clause_GetLiteral(Clause, scan->tIndex)); */
  /*   fputs("  ", stdout); */
  /* } */
  /* puts("\nS_k = "); */
  /* for (scan = List; scan != NULL; scan = scan->next) { */
  /*   LITERAL actLit = scan->sLit; */
  /*   clause_LiteralPrint(actLit); fputs(" in ", stdout); */
  /*   clause_Print(clause_LiteralOwningClause(actLit)); */
  /*   putchar('\n'); */
  /* } */

  constraint    = antecedent = succedent = list_Nil();
  parentClauses = parentLits = list_Nil();
  parents       = list_Nil();                      /* Used to set split data */
  depth         = clause_Depth(Clause);
  tSubterm      = term_FirstArgument(clause_GetLiteralAtom(Clause, (int)list_Car(TLits)));

  /* Now collect the literals of the new clause */
  /* First consider the condition atoms */
  for (scan=sort_ConditionConstraint(Condition); !list_Empty(scan); scan=list_Cdr(scan)) {
    TERM termCopy = term_Copy(list_Car(scan));

    term_ReplaceVariable(termCopy, sort_ConditionVar(Condition), tSubterm);
    constraint = list_Cons(cont_CopyAndApplyBindings(cont_LeftContext(), termCopy),
			   constraint);
    term_Delete(termCopy);   /* constraint contains a copy */
  }
  for (scan=sort_ConditionAntecedent(Condition); !list_Empty(scan); scan=list_Cdr(scan)) {
    TERM termCopy = term_Copy(list_Car(scan));

    term_ReplaceVariable(termCopy, sort_ConditionVar(Condition), tSubterm);
    antecedent = list_Cons(cont_CopyAndApplyBindings(cont_LeftContext(), termCopy),
			   antecedent);
    term_Delete(termCopy);   /* antecedent contains a copy */ 
  }
  for (scan=sort_ConditionSuccedent(Condition); !list_Empty(scan); scan=list_Cdr(scan)) {
    TERM termCopy = term_Copy(list_Car(scan));

    term_ReplaceVariable(termCopy, sort_ConditionVar(Condition), tSubterm);
    succedent = list_Cons(cont_CopyAndApplyBindings(cont_LeftContext(), termCopy),
			  succedent);
    term_Delete(termCopy);   /* succedent contains a copy */
  }
  /* update parents and depth from condition clauses */
  for (scan=sort_ConditionClauses(Condition); !list_Empty(scan); scan=list_Cdr(scan)) {
    CLAUSE condClause = list_Car(scan);

    parents = list_Cons(condClause, parents);

    parentClauses = list_Cons((POINTER)clause_Number(condClause), parentClauses);
    /* !!!! Provisorium !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!! */
    parentLits    = list_Cons((POINTER)clause_FirstSuccedentLitIndex(condClause),
			      parentLits);

    depth = misc_Max(depth, clause_Depth(condClause));
  }

  /* Now we consider the partner clauses */
  for (scan = Partners; !list_Empty(scan); scan = list_Cdr(scan)) {
    LITERAL pLit    = list_Car(scan);
    CLAUSE  pClause = clause_LiteralOwningClause(pLit);
    int     pi      = clause_LiteralGetIndex(pLit);

    bound = clause_LastConstraintLitIndex(pClause);
    for (i = clause_FirstLitIndex(); i <= bound; i++) {
      constraint = list_Cons(cont_CopyAndApplyBindings(cont_RightContext(),
						       clause_GetLiteralAtom(pClause,i)),
			     constraint);
    }
    bound = clause_LastAntecedentLitIndex(pClause);
    for (i = clause_FirstAntecedentLitIndex(pClause); i <= bound; i++) {
      antecedent = list_Cons(cont_CopyAndApplyBindings(cont_RightContext(),
						       clause_GetLiteralAtom(pClause,i)),
			     antecedent);
    }
    bound = clause_LastSuccedentLitIndex(pClause);
    for (i = clause_FirstSuccedentLitIndex(pClause); i <= bound; i++) {
      if (i != pi)
	succedent = list_Cons(cont_CopyAndApplyBindings(cont_RightContext(),
							clause_GetLiteralAtom(pClause,i)),
			      succedent);
    }

    parents = list_Cons(pClause, parents);

    parentClauses = list_Cons((POINTER)clause_Number(pClause), parentClauses);
    parentLits    = list_Cons((POINTER) pi, parentLits);

    depth = misc_Max(depth, clause_Depth(pClause));
  }

  /* Last but not least we consider the <Clause> itself */
  bound = clause_LastConstraintLitIndex(Clause);
  for (i = clause_FirstLitIndex(); i <= bound; i++) {
    if (!list_PointerMember(TLits, (POINTER)i))
	constraint = list_Cons(cont_CopyAndApplyBindings(cont_LeftContext(),
							 clause_GetLiteralAtom(Clause,i)),
			       constraint);
    else {
      parentClauses = list_Cons((POINTER)clause_Number(Clause), parentClauses);
      parentLits    = list_Cons((POINTER)i, parentLits);
    }
  }
  bound = clause_LastAntecedentLitIndex(Clause);
  for (i = clause_FirstAntecedentLitIndex(Clause); i <= bound; i++) {
    antecedent = list_Cons(cont_CopyAndApplyBindings(cont_LeftContext(),
						     clause_GetLiteralAtom(Clause,i)),
			   antecedent);
  }
  bound = clause_LastSuccedentLitIndex(Clause);
  for (i = clause_FirstSuccedentLitIndex(Clause); i <= bound; i++) {
    succedent = list_Cons(cont_CopyAndApplyBindings(cont_LeftContext(),
						    clause_GetLiteralAtom(Clause,i)),
			  succedent);
  }

  parents = list_Cons(Clause, parents);
  
  /* Now we've got all data we need */
  newClause = clause_Create(constraint, antecedent, succedent);
  
  list_Delete(constraint); list_Delete(antecedent); list_Delete(succedent);
  
  if (term_IsVariable(tSubterm))
    clause_SetFromEmptySort(newClause);
    /*clause_SetFromEmptySortPlusPlus(newClause);*/
  else
    clause_SetFromSortResolution(newClause);
    /*clause_SetFromWeakening(newClause);*/
  
  clause_SetDepth(newClause, depth+1);
  clause_SetFlag(newClause, DOCCLAUSE);
  
  clause_SetSplitDataFromList(newClause, parents);
  list_Delete(parents);
  
  clause_SetParentClauses(newClause, parentClauses);
  clause_SetParentLiterals(newClause, parentLits);
 
  printf("\nNEW: "); clause_Print(newClause); fflush(stdout);
  return list_List(newClause);
  /*clause_Delete(newClause);
    return list_Nil(); */
}

static LIST inf_InternWeakening(CLAUSE Clause, LIST TLits, LIST Unifiers,
				LITERAL Special, LIST SojuList)
/**************************************************************
  INPUT:   A <Clause> with unsolved sort constraint.
           A list <TLits> of constraint indices in <Clause> where
	   all corresponding constraints have the same term.
	   A list <Unifiers> of monadic succedent literals whose
	   subterm is unifiable with the subterm of the <TLits>.
	   If called from a backward rule the literal <Special>
	   will be the succedent literal from the respective
	   GivenClause, that must be part of every considered
	   SOJU sort. If called from a forward rule <Special> is NULL.
	   A list <SojuList> of sort pairs.
  RETURNS: The list of possible resolvents.
  EFFECT:  ATTENTION: <SojuList> is deleted.
  MEMORY:  Memory is allocated for the returned list and the clauses.
***************************************************************/
{
  LIST result, myUnifiers;
  TERM searchTerm; 
  int  stack;

  LIST scan;
#if DEBUG
  if (!clause_IsClause(Clause) || list_Empty(TLits) || list_Empty(Unifiers)) {
    fputs("\n\nIllegal Input in inf_InterWeakening\n", stderr);
    misc_Error();
  }
#endif

  putchar('\n'); clause_Print(Clause);
  printf("\nT_k = ");
  for (scan = TLits; !list_Empty(scan); scan = list_Cdr(scan)) {
    clause_LiteralPrint(clause_GetLiteral(Clause, (int)list_Car(scan)));
    putchar(' ');
  }
  printf("\nS_k =");
  for (scan = Unifiers; !list_Empty(scan); scan = list_Cdr(scan)) {
    putchar('\n');
    clause_LiteralPrint(list_Car(scan));
    printf(" in ");
    clause_Print(clause_LiteralOwningClause(list_Car(scan)));
  }
  putchar('\n');

  if (list_Empty(SojuList))
    return list_Nil();

  /*return list_Nil();*/
  /* Und Schluss */

  result = list_Nil();

  myUnifiers = list_Copy(Unifiers);
  inf_CopyUnifierClauses(myUnifiers);
  inf_MakeClausesDisjoint(Clause, myUnifiers);

  searchTerm =
    term_FirstArgument(clause_GetLiteralAtom(Clause, (int)list_Car(TLits)));

  stack  = stack_Bottom();

  for ( ; !list_Empty(SojuList); SojuList = list_Pop(SojuList)) {
    SOJU actSoju = list_Car(SojuList);

    printf("\nSOJU: "); sort_PairPrint(actSoju); fflush(stdout);

    if (Special == NULL ||
	sort_ContainsSymbol(sort_PairSort(actSoju),
			    clause_LiteralPredicate(Special))) { 
      LIST actSortSymbols, symbolScan, unifierScan, subset;

      actSortSymbols = sort_GetSymbolsFromSort(sort_PairSort(actSoju));
      subset         = list_Nil();

      symbolScan  = actSortSymbols;
      unifierScan = myUnifiers;

      do {
	while (!list_Empty(symbolScan) && !list_Empty(unifierScan)) {
	  LITERAL actLit = list_Car(unifierScan);

	  if (symbol_Equal(clause_LiteralPredicate(list_Car(unifierScan)),
			   (SYMBOL)list_Car(symbolScan))) {
	    cont_StartBinding();
	    if (unify_UnifyNoOC(cont_LeftContext(), searchTerm, cont_RightContext(),
				term_FirstArgument(clause_LiteralAtom(actLit)))) {
	      /* Found corresponding literal for sort symbol */
	      stack_Push(symbolScan);
	      stack_Push(list_Cdr(unifierScan));
	      subset = list_Cons(actLit, subset);
	      /* Now search literals for the next sort symbol */
	      symbolScan = list_Cdr(symbolScan);

	      if (!list_Empty(symbolScan))
		/* Start search for literal at the beginning of unifier list */
		unifierScan = myUnifiers;
	      else
		unifierScan = list_Cdr(unifierScan);
	    } else {
	      cont_BackTrack();
	      unifierScan = list_Cdr(unifierScan);
	    }
	  } else
	    unifierScan = list_Cdr(unifierScan);
	}

	if (list_Empty(symbolScan)) {
	  /*putchar('\n');
	  clause_LiteralListPrint(subset);*/
	  /* Found subset */
	  result = list_Nconc(inf_ApplyWeakening(Clause, TLits, subset,
						 sort_PairCondition(actSoju)),
			      result);
	}

	while (!stack_Empty(stack) && list_Empty(stack_Top())) {
	  /* No more literals */
	  stack_NPop(2);
	  cont_BackTrack();
	  subset = list_Pop(subset);
	}

	if (!stack_Empty(stack)) {
	  /* Implies that stack_Top is a non-empty list */
	  unifierScan = stack_PopResult();
	  symbolScan  = stack_PopResult();
	  cont_BackTrack();
	  subset = list_Pop(subset);
	}
      } while (!stack_Empty(stack) || !list_Empty(unifierScan));

      list_Delete(actSortSymbols);
    }
    sort_PairDelete(actSoju);
  } /* For all SOJUs */

  inf_DeleteUnifierClauses(myUnifiers);
  list_Delete(myUnifiers);

  return result;
}

LIST inf_ForwardWeakening(CLAUSE GivenClause, st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause and an 'Index' of clauses. The sort constraint
           of the clause must contain a non-variable term
	   (this implies the sort constraint is unsolved). 
  RETURNS: A list of clauses derived from the GivenClause by
           the Weakening rule.
  MEMORY:  Memory is allocated for the returned list and the clauses.
***************************************************************/
{
  LIST    result;
  int     i, lc;
  BOOL    hit;

#if DEBUG
  if (!clause_HasTermSortConstraintLits(GivenClause)) {
    fputs("\n\nIllegal input in inf_ForwardWeakening\n", stderr);
    misc_Error();
  }
#endif

  result = list_Nil();
  lc     = clause_LastConstraintLitIndex(GivenClause);
  hit    = FALSE;

  for (i = clause_FirstLitIndex(); i <= lc && !hit; i++) {
    
    if (!term_IsVariable(term_FirstArgument(clause_GetLiteralAtom(GivenClause, i)))) {
      /* Condition implies that constraint is unsolved */
      LITERAL tLit = clause_GetLiteral(GivenClause, i);
      LIST    unifiers;
      int     j;

      hit = TRUE; /* Try only the first appropriate constraint literal */

      unifiers = inf_GetForwardPartnerLits(tLit, Index);

      if (!list_Empty(unifiers)) {
	TERM tAtom = clause_GetLiteralAtom(GivenClause, i);
	LIST tLits, sojuList;
	SORT tSort, unifierSort;

	printf("\n Forward Weakening: "); clause_Print(GivenClause);

	/* Search the other T_k(t) in GivenClause */
	tLits = list_Nil();
	tSort = sort_TopSort();
	for (j = lc; j > i; j--) {
	  TERM actAtom = clause_GetLiteralAtom(GivenClause, j);

	  if (term_FirstArgument(actAtom) == term_FirstArgument(tAtom)) {
	    tLits = list_Cons((POINTER)j, tLits);
	    tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory, term_TopSymbol(actAtom)),
				   tSort);
	  }
	}
	tLits = list_Cons((POINTER)i, tLits);
	tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory, term_TopSymbol(tAtom)),
			       tSort);
	list_PointerDeleteDuplicates(tSort);
	/* Necessary for Christoph's function */

	unifierSort = inf_GetSortFromLits(unifiers, SortTheory);
	/*putchar('\n'); sort_TheoryPrint(SortTheory); putchar('\n'); */
	sojuList = sort_TheoryComputeAllSubsortHits(SortTheory, unifierSort, tSort);
	sort_Delete(unifierSort);
	sort_Delete(tSort);

	result =
	  list_Nconc(inf_InternWeakening(GivenClause, tLits, unifiers, NULL, sojuList),
		     result);

	list_Delete(tLits);
	list_Delete(unifiers);
      }
    }
  }
  return result;
}

LIST inf_BackwardWeakening(CLAUSE GivenClause, st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause with solved sort constraint and an 'Index' of clauses.
  RETURNS: A list of clauses inferred from the GivenClause by
           the Weakening rule.
  MEMORY:  Memory is allocated for the list and the clauses.
***************************************************************/
{
  LIST     result;
  int      i, ls;

#if DEBUG
  if (!clause_HasSolvedConstraint(GivenClause)) {
    fputs("\n\nIllegal input in inf_BackwardWeakening\n", stderr);
    misc_Error();
  }
#endif

  result = list_Nil();
  ls     = clause_LastSuccedentLitIndex(GivenClause);

  for (i = clause_FirstSuccedentLitIndex(GivenClause); i <= ls; i++) {
    LITERAL sLit  = clause_GetLiteral(GivenClause, i);
    TERM    sAtom = clause_LiteralAtom(sLit);

    if (clause_LiteralGetFlag(sLit, STRICTMAXIMAL) && 
	clause_LiteralIsSort(sLit) &&
	(!term_IsVariable(term_FirstArgument(sAtom)) ||
	 clause_HasEmptyConstraint(GivenClause))) {
      LIST unifiers, partners;
      SORT unifierSort;

      unifiers = partners = list_Nil();
      inf_GetBackwardPartnerLits(sLit, Index, &partners, &unifiers, FALSE);
      unifiers = list_Cons(sLit, unifiers);
      /* <partners> holds monadic constraint literals */
      /* <unifiers> holds monadic, maximal succedent literals */
      unifierSort = inf_GetSortFromLits(unifiers, SortTheory);
      
      for ( ; !list_Empty(partners); partners = list_Pop(partners)) {
	LITERAL tLit    = list_Car(partners);
	CLAUSE  tClause = clause_LiteralOwningClause(tLit);
	TERM    tAtom   = clause_LiteralAtom(tLit);
	int     ti      = clause_LiteralGetIndex(tLit);
	int     lc      = clause_LastConstraintLitIndex(tClause);
	int     j;
	LIST    tLits, sojuList;
	SORT    tSort;
	    
	printf("\nBackward Weakening: ");clause_Print(GivenClause);

	/* Search the other T_k(t) in GivenClause */
	tLits = list_Nil();
	tSort = sort_TopSort();
	for (j = lc; j >= clause_FirstLitIndex(); j--) {
	  TERM actAtom = clause_GetLiteralAtom(tClause, j);

	  if (j != ti &&
	      term_FirstArgument(actAtom) == term_FirstArgument(tAtom)) {
	    tLits = list_Cons((POINTER)j, tLits);
	    tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory, term_TopSymbol(actAtom)),
				   tSort);
	  }
	}
	tLits = list_Cons((POINTER)ti, tLits);
	tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory, term_TopSymbol(tAtom)),
			       tSort);
	list_PointerDeleteDuplicates(tSort);
	/* Necessary for Christoph's function */

	/*sort_TheoryPrint(SortTheory); fflush(stdout);*/

	sojuList = sort_TheoryComputeAllSubsortHits(SortTheory, unifierSort, tSort);
	sort_Delete(tSort);

	cont_StartBinding();
	unify_UnifyNoOC(cont_LeftContext(), tAtom,
			cont_RightContext(), sAtom);

	result =
	  list_Nconc(inf_InternWeakening(tClause, tLits, unifiers, sLit, sojuList),
		     result);

	cont_BackTrack();

	list_Delete(tLits);
      }
      sort_Delete(unifierSort);
      list_Delete(unifiers);
    }
  }

  return result;
}

LIST inf_ForwardEmptySortPlusPlus(CLAUSE GivenClause, st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause and an 'Index' of clauses. The sort constraint
           of the clause must not contain a non-variable term,
	   but the sort constraint has to be unsolved.
  RETURNS: A list of clauses derived from the GivenClause by
           the Empty Sort rule.
  MEMORY:  Memory is allocated for the returned list and the clauses.
***************************************************************/
{
  LIST     result;
  int      i, lc;
  BOOL     hit;
  
#if DEBUG
  if (clause_HasTermSortConstraintLits(GivenClause) ||
      clause_HasSolvedConstraint(GivenClause)) {
    fputs("\n\nIllegal input in inf_ForwardEmptySort\n", stderr);
    misc_Error();
  }
#endif
  
  result = list_Nil();
  lc     = clause_LastConstraintLitIndex(GivenClause);
  hit    = FALSE;
  
  for (i = clause_FirstLitIndex(); i <= lc && !hit; i++) {

    if (term_IsVariable(term_FirstArgument(clause_GetLiteralAtom(GivenClause,i)))) {
      LITERAL tLit = clause_GetLiteral(GivenClause, i);
      TERM    var  = term_FirstArgument(clause_LiteralAtom(tLit));
      int     j, ls;
      BOOL varOccursNoMore;

      ls = clause_LastSuccedentLitIndex(GivenClause);
      varOccursNoMore = TRUE;
      /* Check if the variable occurs in antecedent or succedent literals */
      for (j = clause_FirstAntecedentLitIndex(GivenClause); 
	   (j <= ls) && varOccursNoMore; j++) {
	if (term_ContainsSymbol(clause_GetLiteralAtom(GivenClause,j),
				term_TopSymbol(var)))
	  varOccursNoMore = FALSE;
      }
		
      if (varOccursNoMore) {
	/* Condition implies that constraint is unsolved */
	LIST unifiers = inf_GetForwardPartnerLits(tLit, Index);

	hit = TRUE;  /* We found the first appropriate constraint literal */

	if (!list_Empty(unifiers)) {
	  TERM tAtom = clause_LiteralAtom(tLit);
	  LIST tLits, sojuList;
	  SORT tSort, unifierSort;

	  printf("\n Forward EmptySort: "); clause_Print(GivenClause);

	  /* Search the other T_k(t) in GivenClause */
	  tLits = list_Nil();
	  tSort = sort_TopSort();
	  for (j = lc; j > i; j--) {
	    TERM actAtom = clause_GetLiteralAtom(GivenClause, j);

	    if (term_FirstArgument(actAtom) == var) {   /* tClause is shared */
	      tLits = list_Cons((POINTER)j, tLits);
	      tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory,
							     term_TopSymbol(actAtom)),
				     tSort);
	    }
	  }
	  tLits = list_Cons((POINTER)i, tLits);
	  tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory,
							 term_TopSymbol(tAtom)),
				 tSort);
	  list_PointerDeleteDuplicates(tSort);
	  /* Necessary for Christoph's function */

	  unifierSort = inf_GetSortFromLits(unifiers, SortTheory);
	  putchar('\n'); sort_TheoryPrint(SortTheory); putchar('\n');
	  printf("\nT_i : "); sort_Print(tSort);
	  printf("\nS_i : "); sort_Print(unifierSort); fflush(stdout);
	  sojuList = sort_TheoryComputeAllSubsortHits(SortTheory, unifierSort, tSort);
	  printf("\nDa bin ich wieder!"); fflush(stdout);
	  sort_Delete(unifierSort);
	  sort_Delete(tSort);

	  result =
	    list_Nconc(inf_InternWeakening(GivenClause, tLits, unifiers, NULL, sojuList),
		       result);

	  list_Delete(tLits);
	  list_Delete(unifiers);
	}
      }
    }
  }
  return result;
}

LIST inf_BackwardEmptySortPlusPlus(CLAUSE GivenClause, st_INDEX Index, SORTTHEORY SortTheory)
/**************************************************************
  INPUT:   A clause with solved sort constraint and an 'Index' of clauses.
  RETURNS: A list of clauses inferred from the GivenClause by
           the Empty Sort rule.
  MEMORY:  Memory is allocated for the list and the clauses.
***************************************************************/
{
  LIST     result;
  int      i, ls;

#if DEBUG
  if (!clause_HasSolvedConstraint(GivenClause)) {
    printf("\n\nIllegal input in inf_BackwardEmptySort\n");
    misc_Error();
  }
#endif

  result = list_Nil();  
  ls     = clause_LastSuccedentLitIndex(GivenClause);

  for (i = clause_FirstSuccedentLitIndex(GivenClause); i <= ls; i++) {
    LITERAL sLit  = clause_GetLiteral(GivenClause, i);
    TERM    sAtom = clause_LiteralAtom(sLit);

    if (clause_LiteralGetFlag(sLit,STRICTMAXIMAL) && 
	clause_LiteralIsSort(sLit) &&
	(!term_IsVariable(term_FirstArgument(sAtom)) ||
	 clause_HasEmptyConstraint(GivenClause))) {
      LIST unifiers, partners;
      SORT unifierSort;

      unifiers = partners = list_Nil();
      inf_GetBackwardPartnerLits(sLit, Index, &partners, &unifiers, TRUE);
      unifiers = list_Cons(sLit, unifiers);
      /* <partners> holds monadic constraint literals */
      /* <unifiers> holds monadic, maximal succedent literals */

      unifierSort = inf_GetSortFromLits(unifiers, SortTheory);

      for ( ; !list_Empty(partners); partners = list_Pop(partners)) {
	LITERAL tLit    = list_Car(partners);
	CLAUSE  tClause = clause_LiteralOwningClause(tLit);
	TERM    tAtom   = clause_LiteralAtom(tLit);
	int     ti      = clause_LiteralGetIndex(tLit);
	int     li      = clause_LastSuccedentLitIndex(tClause);
	TERM    var     = term_FirstArgument(tAtom);
	BOOL    varOccursNoMore;
	int     j;

	varOccursNoMore = TRUE;
	for (j = clause_FirstAntecedentLitIndex(tClause);
	     j <= li && varOccursNoMore;
	     j++) {
	  if (term_ContainsSymbol(clause_GetLiteralAtom(tClause,j),
				  term_TopSymbol(var)))
	    varOccursNoMore = FALSE;
	}
 
	if (varOccursNoMore) {
	  /* Condition implies that constraint is unsolved */
	  int  lc  = clause_LastConstraintLitIndex(tClause);
	  LIST tLits, sojuList;
	  SORT tSort;

	  printf("\nBackward EmptySort: ");clause_Print(GivenClause);

	  /* Search the other T_k(t) in GivenClause */
	  tLits = list_Nil();
	  tSort = sort_TopSort();
	  for (j = lc; j >= clause_FirstLitIndex(); j--) {
	    TERM actAtom = clause_GetLiteralAtom(tClause, j);

	    if (j != ti &&
		term_TopSymbol(term_FirstArgument(actAtom)) == term_TopSymbol(var)) {
	      tLits = list_Cons((POINTER)j, tLits);
	      tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory, term_TopSymbol(actAtom)),
				     tSort);
	    }
	  }
	  tLits = list_Cons((POINTER)ti, tLits);
	  tSort = sort_Intersect(sort_TheorySortOfSymbol(SortTheory, term_TopSymbol(tAtom)),
				 tSort);
	  list_PointerDeleteDuplicates(tSort);
	  /* Necessary for Christoph's function */

	  sojuList = sort_TheoryComputeAllSubsortHits(SortTheory, unifierSort, tSort);
	  sort_Delete(tSort);

	  cont_StartBinding();
	  unify_UnifyNoOC(cont_LeftContext(), tAtom,
			  cont_RightContext(), sAtom);

	  result =
	    list_Nconc(inf_InternWeakening(tClause, tLits, unifiers, sLit, sojuList),
		       result);

	  cont_BackTrack();

	  list_Delete(tLits);
	}
      }
      sort_Delete(unifierSort);
      list_Delete(unifiers);
    }
  }

  return result;
}
