/*				FINDER.c
*
*	This is the top file for FINDER, except for Fmain.c where the
*	function main() lives in solitary splendour. When FINDER is 
*	used as an external utility by other programs, this file will 
*	be linked in but Fmain.c will not.
*
*	This is the main file where FINDER's external variables are 
*	officially declared.
*/

        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/

#define MAINFILE
#include "FINDER.h"


/*
*	Found is a supertester. It tests the given clause in the 
*	current model. If the clause is true in the model then it 
*	is added to the current theory iff add_old is true. If it 
*	is not true in the current model and add_new is true then
*	a model of the current theory plus the given clause is 
*	sought. If one is found then it becomes the current model 
*	and the given clause is added to the current theory. The 
*	value returned is in every case true iff the given clause 
*	is true in the (possibly new) current model.
*
*	Note that the timer is reset on every call. This allows a
*	time limit to be set for each individual search.
*/	

int found( char s[], int add_old, int add_new, int seek_model )
{
    int i, j, oldgood, rc;
    boolean update = false;
    functor f;
    void short_timesup();

    signal( SIGALRM, short_timesup );
    addclause(s);
    tr_parameters.done = 0;
    tr_parameters.reps[NULL_BACKTRACKS].i_report = 0;
    tr_parameters.reps[SURJ_BACKTRACKS].i_report = 0;
    tr_parameters.reps[SOLUTIONS].i_report = 0;
    tr_parameters.reps[BAD_TESTS].i_report = 0;
    for ( i = 0; sort[i].sname; i++ ) sort[i].siz = lastsizes[i];
    for ( f = FN; f; f = f->nextfunct ) f->fcc = f->old_fcc;
    tr_parameters.vlength = lastvlength;
  for (f=FN; f; f = f->nextfunct)
        set_fats(f);
    alarm( setting.maxtime );
    rc = model(lastgood, 0);
    if ( seek_model && !rc ) {
	oldgood = good;
	begin_time = timestamp();
	go(0);
	end_time = timestamp();
	if ( good > oldgood ) {
	    rc = 1;
	    update = true;
	}
    }
    if ( rc && (add_old || add_new) ) add_to_theory();
    if ( update ) update_size(1);
    return rc;
}



/*
*	Is_good simply calls found (above) with flags set to add the 
*	clause (old or new) iff the number of calls so far to this 
*	function is small. The default limit on "small" is 25, though
*	this may be altered by means of the setting specification 
*	test-after.
*/

int is_good(char s[])
{
    static int calls;
    int add_if_good;

    calls++;
    add_if_good = (calls <= setting.test_point);
    return found( s, add_if_good, add_if_good, add_if_good );
}



/*
*	Init_finder must be called before much else happens if FINDER
*	is to be used as a supertester. It reads the input file and 
*	returns a boolean (integer) value recording whether it found 
*	any model of the input theory. If the return value is false 
*	then there will be no "current theory", and adding clauses may 
*	be a bad idea.
*/

int init_finder(char s[])
{
    FILE *if_fil;

    initialize();
    if ( !(if_fil = fopen(s,"r")) ) inerror("Unable to open file",s);
    read_job( if_fil );
    fclose( if_fil );
    tr_parameters.done = 0;
    good = 0;
    single_model = true;
    Integer->piped = true;
    if ( setting.pm.verbosity & pjob ) print_job();
    update_size(true);
    begin_time = timestamp();
    go(0);
    end_time = timestamp();
    return good;
}



/*
*	Is_bad was originally used by Scott for goal 
*	deletion. If the clause in c begins with the 
*	character g and the result of stripping off 
*	that character is false in the current model
*	then return true. Otherwise return false.
*/

int is_bad( char c[], char g, boolean verbose )
{
    while (*c == ' ') c++;
    if ( *c == '-' && c[1] == '>' ) c += 2;
    while (*c == ' ') c++;
    if ( !*c ) skipout("Null clause passed",SKIP);
    if ( *c != g ) return 0;
    if ( found(c+1, 0, 0, 1) ) {
	if ( verbose ) printf("deleted  %s\n",c);
	return 1;
    }
    if ( verbose ) printf("kept     %s\n",c);
    return 0;
}



/*
*	Fairly self-explanatory.
*/

int init_goal_finder( char filename[] )
{
    init_finder(filename);
    refute_current();
}



/*
*	Return the theory size the last time this was called with
*	flag new set.
*/

int update_size( boolean new )
{
    static int rc;

    if ( new ) rc = theory_size();
    return rc;
}




/*
*	Go is the main work loop of FINDER. It calls transref to search
*	every search space, defined by setting the cardinalities of the
*	sorts, within the range stipulated. Go is called from the above 
*	functions, and from itself, and also from main().
*/

void go(int x)
{
    if ( negative && !clauses_fail ) return;
    if ( !sort[x].sname ) {
    	set_up_cc();
	firstmodel = true;
	transref(&tr_parameters);
    }
    else if ( sort[x].piped && !tr_parameters.done ) go(x+1);
    else for ( sort[x].siz = sort[x].sizmin;
	      sort[x].siz <= sort[x].sizmax && !tr_parameters.done;
	      sort[x].siz++ )
	go(x+1);
}



/*
*	FINDER can be used to return models falsifying a badguy.
*/

void refute_current()
{
    cc_badguy = true;
}




/*
*	It is sometimes useful for a calling program to know how many 
*	clauses there are in the current theory.
*/

int theory_size()
{
    int i = 0;
    clause c;

    for ( c = TH; c; c = c->nextclause ) i++;
    return i;
}


/*
*	The final trivial function in this section allows a calling
*	program to add FINDER input one line at a time.
*/

void finder_input( char s[] )
{
    line_input = true;
    input_point = 0;
    strcpy( command_line, s );
    strcat( command_line, " end.");
    read_job(0);
}



void short_timesup()
{
    job_done( &tr_parameters, "time limit" );
}


