/*************************************************************************
*   This file contains the functions called by Otter when compiled with  *
*   SCOTT defined. They set up the data structures specific to SCOTT and *
*   call the requisite FINDER functions.                                 *
*************************************************************************/

#include "header.h"

#define FILE_NAME_LENGTH 256
#define MAX_NAME_STRING  64
/* Note MAX_CLAUSE_STRING must be big enough to hold many names */
#define MAX_CLAUSE_STRING 8192

/* Set aside some space for some statistics*/
typedef enum {
    CL_TESTED,
    CL_TRUE_GIVEN,
    CL_TRUE_TESTED,
    CL_DSR_AVOIDED,
    MAX_FSTATS     /* Marker for end of stats list - is really an integer.*/
} Finderstats;

static int Fstats[MAX_FSTATS]; /* Integer array to hold Statictics..?*/
static char Fstatname[MAX_FSTATS][32];

typedef struct {
    char filename[FILE_NAME_LENGTH];
    int   true_weight;
    float true_scale;
    int   sem_res;
} Finderoptions;

static Finderoptions Fopts; /* global, but just for this file */

/************
*
*    finder_init() - reads the command line and rewrites it as an Otter 
*                    command line if necessary. It then calls init_finder 
*                    to get the first model.
*************/

/*-------------------------------------------------------------------------*
 * Scott takes the following command line arguments:                       *
 * d                perform dynamic semantic resolution.                   *
 * f <fileneme>     takes filename as the FINDER input file                *
 * w {n}            uses n as the weight to add to true clauses.           *
 * s {m}            uses m as a scale to apply to true clauses.            *
 * if w is given then f must also be supplied.                             *
 *-------------------------------------------------------------------------*/
 
void finder_init(int argc, char *argv[])
{

    /* Mysterious pointer for executing the getopt function - points to
     * the head of the argument (a string) of the current option */ 
    extern char *optarg;

    int option;
    int i;
    Finderstats fst;

    /* initialise everything */
    strcpy(Fopts.filename,"Finderfile");
    Fopts.true_weight = 0;
    Fopts.true_scale  = 1;
    Fopts.sem_res = FALSE;

    while ((option = getopt (argc, argv, "df:w:s:")) != -1){
        switch( option ) {

        /* perform 'dynamic semantic resolution' */
        case 'd': Fopts.sem_res = TRUE;
            break;

        /* take the following file as FINDER input */
        case 'f': if (strlen(optarg) < FILE_NAME_LENGTH){
				strcpy(Fopts.filename,optarg);
		   }
		  else {
		    abend("filename longer than FILE_NAME_LENGTH characters.");
		   }
            break;
 
        /* fatten TRUE clause with the following weight */
        case 'w': sscanf(optarg,"%d",&(Fopts.true_weight));
            break;

        /* fatten TRUE clause by the following scale */
        case 's': sscanf(optarg,"%f",&(Fopts.true_scale));
            break;
        }
    }

    if (Fopts.sem_res || Fopts.true_weight || Fopts.true_scale != 1){
        /* Get FINDER ready to search for models
         * Takes an initial theory from the finder file 
         * searches for an initial guiding model
         * - fcn in 'FINDER.c' */
             init_finder(Fopts.filename);
        /* Otherwise default to normal Otter */
    }
    
    /* Initialize each statistic */
    for (fst = CL_TESTED; fst < MAX_FSTATS; fst++)
        Fstats[fst] = 0;

     /* Give the stistics Names for Output     */
    /*(Well thats INITing for ya..)         */
    strcpy(Fstatname[CL_TESTED],"Clauses tested");
    strcpy(Fstatname[CL_TRUE_GIVEN],"True given clauses");
    strcpy(Fstatname[CL_TRUE_TESTED],"Clauses true in model");
    strcpy(Fstatname[CL_DSR_AVOIDED],"DSR deleted clauses");

} /* finder_init() */

/***********
*
*    finder_sprint_t() - Convert OTTER term into FINDER term.
*
***********/

void finder_sprint_t(s, t)
char *s;
struct term *t;
{
    struct rel *r;
    char s1[MAX_NAME_STRING];
    
	/* exit if s is too big to append the next name to */
	if (strlen(s) >= (MAX_CLAUSE_STRING - MAX_NAME_STRING))
		abend("Clause too big - MAX_CLAUSE_STRING exceeded");
    if (t->type == NAME)
    strcat(s, sn_to_str(t->sym_num));
    else if (t->type == VARIABLE) {
    strcat(s, "v");
    sprintf(s1, "%d", t->varnum);
    strcat(s, s1);
    }
    else {
    strcat(s, sn_to_str(t->sym_num));
    strcat(s, "(");
    for (r = t->farg; r; r = r->narg) {
        finder_sprint_t(s, r->argval);
        if (r->narg)
        strcat(s, ",");
        }
    strcat(s, ")");
    }

} /* finder_sprint_t() */


/***************
*
*    finder_sprint_cl() - Convert OTTER clause to FINDER clause.
*
***************/

void finder_sprint_cl(s, c)
char *s;
struct clause *c;
{
    struct literal *lit;
    int c_flag;

    *s = '\0'; /* start with an empty FINDER clause */

    /* convert negative literals */
    for (lit = c->first_lit, c_flag = 0; lit; lit = lit->next_lit)
    if (!lit->sign) {
        if (c_flag)
        strcat(s, ", ");
        finder_sprint_t(s, lit->atom);
        c_flag = 1;
        }

    strcat(s, " -> ");

    /* convert positive literals */
    for (lit = c->first_lit, c_flag = 0; lit; lit = lit->next_lit)
    if (lit->sign) {
        if (c_flag)
        strcat(s, ", ");
        finder_sprint_t(s, lit->atom);
        c_flag = 1;
        }

    strcat(s, ".");

} /* finder_sprint_cl() */

/************
*
*    test_a_clause() - calls is_good which test the clause against 
*                      the current model and searches for a new model
*                      if the old model fails to satisfy the clause
*
************/

BOOLEAN test_a_clause(c)
struct clause *c;
{
    char s[MAX_CLAUSE_STRING];

    Fstats[CL_TESTED]++;
    finder_sprint_cl(s, c); /* Convert OTTER clause to FINDER clause */

    /* test clause: improve model - fnc in 'FINDER.c' 
     * but only if clause hasn't already been modeled*/
    if (!c->true_in_model) c->true_in_model = is_good(s); 

    if (c->true_in_model) Fstats[CL_TRUE_TESTED]++;
    return c->true_in_model;

} /* test_a_clause() */


/************
*
*    fps_weight() - return the user specified weight to add (?) to true
*                   clause. Returns 0 if none given (ie defaults to OTTER).
*
************


/* Increase the weight of the clause depending on whether 
 * weighting or scaling strategies are being employed.
 * Returm TRUE iff clause is modeled */

BOOLEAN fps_weight(c)
struct clause *c;
{
    if ( (Fopts.true_weight || Fopts.true_scale != 1) 
          && c && test_a_clause(c) ){
		/* new weight of TRUE clauses = scale * old weight + weight */
		if (Fopts.true_scale != 1)  c->weight *= Fopts.true_scale;
        if (Fopts.true_weight != 0) c->weight += Fopts.true_weight;
		return TRUE;
    }
	return FALSE; /* otherwise weight is unchanged */

 } /* new fps_weight() */



/************
*
*    test_list() - test each clause in the OTTER list;
*                  fatten those that are modeled.
*
************/

BOOLEAN test_list(lptr)
struct list *lptr;
{
    BOOLEAN rc = TRUE;
    struct clause *c;

    if (Fopts.sem_res || Fopts.true_weight || (Fopts.true_scale !=1))
        for (c = lptr->first_cl; c; c = c->next_cl)
			/* OLD  if ( !test_clause(c) ) rc = FALSE; */
            if ( !fps_weight(c) ) rc = FALSE;
    return rc;

} /* test_list() */


/*************
*
*    test_if_dsr() - test the clause if DSR option is set
*
*************/
 
void test_if_dsr(c)
struct clause *c;
{
    if (!c) return;
    if ( Fopts.sem_res && !Fopts.true_weight )
    test_a_clause(c);
 
    /* all clauses have this set ? */
    if (c->true_in_model) Fstats[CL_TRUE_GIVEN]++;
 
} /* test_if_dsr(c) */


/**********
*
*    dsr_bad_parents() - TRUE iff all of the clause's parents
*                        are true in the current model.
*
**********/
 
BOOLEAN dsr_bad_parents(struct clause *c) 
{
    struct int_ptr *ip;
    struct clause *d;
 
    if ( !Fopts.sem_res || !c ) return FALSE;
    for (ip = c->parents; ip; ip = ip->next) {
        if (ip->i >= 0) {
            d = cl_find(ip->i); 
            if (!d->true_in_model) return FALSE;
        }
    }
    Fstats[CL_DSR_AVOIDED]++; 
    return TRUE;
 
} /* dsr_bad_parents() */


/************
*
*    output_finder_stats() - called whenever OTTER outputs it's stats
*
************/

void output_finder_stats(FILE *fp, int level)
{
    Finderstats fst;

    if ( level ) {
        fprintf(fp, "\nFINDER stats\n\n");
        for (fst = CL_TESTED; fst < MAX_FSTATS; fst++)
        fprintf(fp, "%23s%10d\n", Fstatname[fst], Fstats[fst]);
        fprintf(fp, "\nFINDER time      %10.2f\n", clock_val(FINDER_TIME) / 1000.);
    }
} /* output_finder_stats() */
