/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *          REPRESENTATION OF PROOF SEARCH                * */
/* *                                                        * */
/* *  $Module:   PROOF SEARCH                               * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.47 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/02 14:28:30 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: search.c,v $ */


#include "search.h"


/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

static void prfs_PrintSplit(SPLIT);

/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *		    DEBUGGING FUNCTIONS	         	    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

BOOL prfs_Check(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  None.
  RETURNS: TRUE if all invariants about <Search> are valid.
***************************************************************/
{
  LIST   Scan,Clauses;
  SPLIT  Split;
  CLAUSE Clause;

  for (Scan=prfs_UsableClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (clause_GetFlag(Clause,WORKEDOFF) || 
	clause_SplitLevel(Clause) > prfs_ValidLevel(Search) ||
	!clause_IsClause(Clause))
      return FALSE;
  }

  for (Scan=prfs_WorkedOffClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (!clause_GetFlag(Clause,WORKEDOFF) || 
	clause_SplitLevel(Clause) > prfs_ValidLevel(Search) ||
	!clause_IsClause(Clause))
      return FALSE;
  }

  for (Scan=prfs_SplitStack(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Split = (SPLIT)list_Car(Scan);
    if (!(Split->dirty))
      if (!list_Empty(Split->blockedClauses) ||  !list_Empty(Split->deletedClauses)) {
	/*printf("\n");prfs_PrintSplit(Split); printf("\n");*/
	return FALSE;
      }
      else {
	for (Clauses=prfs_UsableClauses(Search);!list_Empty(Clauses);Clauses=list_Cdr(Clauses))
	  if (clause_SplitLevel(list_Car(Clauses)) == Split->splitlevel) {
	    /*printf("\n\n");prfs_PrintSplit(Split); 
	    printf("\n Clause must not exist: ");clause_Print(list_Car(Clauses)); printf("\n");*/
	    return FALSE;
	  }
	for (Clauses=prfs_WorkedOffClauses(Search);!list_Empty(Clauses);Clauses=list_Cdr(Clauses))
	  if (clause_SplitLevel(list_Car(Clauses)) == Split->splitlevel) {
	    /*printf("\n\n");prfs_PrintSplit(Split);
	    printf("\n Clause must not exist: ");clause_Print(list_Car(Clauses)); printf("\n");*/
	    return FALSE;
	  }
      }
  }

  return TRUE;
}

/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *			HIGH LEVEL FUNCTIONS		    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/

static void prfs_InsertInSortTheories(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  If the clause is a declaration clause it is inserted
           into the dynamic and approximated dynamic sort theory.
  RETURNS: void.
***************************************************************/
{
  if ((prfs_DynamicSortTheory(Search) != (SORTTHEORY)NULL
       || prfs_ApproximatedDynamicSortTheory(Search) != (SORTTHEORY)NULL)
      && clause_IsDeclarationClause(Clause)) {
    int     i,l;
    LITERAL Lit;
    CLAUSE  Copy;
    LIST    Approx,Scan;
    l = clause_Length(Clause);
    for(i=clause_FirstSuccedentLitIndex(Clause);i<l;i++) {
      Lit = clause_GetLiteral(Clause,i);
      if (clause_LiteralIsMaximal(Lit) && 
	  symbol_IsBaseSort(term_TopSymbol(clause_LiteralSignedAtom(Lit)))) {
	if (prfs_DynamicSortTheory(Search) != (SORTTHEORY)NULL) {
	  Copy = clause_Copy(Clause);
	  list_Delete(clause_ParentClauses(Copy));
	  clause_SetParentClauses(Copy,list_Nil());
	  list_Delete(clause_ParentLiterals(Copy));
	  clause_SetParentLiterals(Copy,list_Nil());
	  clause_SetNumber(Copy,clause_Number(Clause));
	  sort_TheoryInsertClause(prfs_DynamicSortTheory(Search),Clause,
				  Copy,clause_GetLiteral(Copy,i));
	}
	if (prfs_ApproximatedDynamicSortTheory(Search) != (SORTTHEORY)NULL) {
	  Approx = sort_ApproxMaxDeclClauses(Clause);
	  for(Scan=Approx;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	    Copy = (CLAUSE)list_Car(Scan);
	    sort_TheoryInsertClause(prfs_ApproximatedDynamicSortTheory(Search),
				    Clause, Copy, 
				    clause_GetLiteral(Copy,clause_FirstSuccedentLitIndex(Copy)));
	  }
	  list_Delete(Approx);
	}
      }
    }
  }
}

static void prfs_DeleteFromSortTheories(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  If the clause is a declaration clause it is deleted
           from the dynamic and approximated dynamic sort theory.
  RETURNS: void.
***************************************************************/
{
  if (clause_IsDeclarationClause(Clause)) {
    if (prfs_DynamicSortTheory(Search) != (SORTTHEORY)NULL)
      sort_TheoryDeleteClause(prfs_DynamicSortTheory(Search), Clause);
    if (prfs_ApproximatedDynamicSortTheory(Search) != (SORTTHEORY)NULL)
      sort_TheoryDeleteClause(prfs_ApproximatedDynamicSortTheory(Search), Clause);
  }
}

void prfs_SplitFree(SPLIT S)
/**************************************************************
  INPUT:   A split 
  RETURNS: None.
  MEMORY:  Deletes blocked and deleted clauses.
           Frees the split.
***************************************************************/
{
  clause_DeleteClauseList(S->blockedClauses);
  clause_DeleteClauseList(S->deletedClauses);
  if (S->father != (CLAUSE)NULL)
    clause_Delete(S->father);
  memory_Free(S,sizeof(SPLIT_NODE));
}


void prfs_Delete(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  RETURNS: void.
  EFFECT:  The whole structure including all its substructures 
           is deleted.
***************************************************************/
{
  LIST Scan;

  sort_TheoryDelete(prfs_StaticSortTheory(Search));
  sort_TheoryDelete(prfs_DynamicSortTheory(Search));
  sort_TheoryDelete(prfs_ApproximatedDynamicSortTheory(Search));
  clause_DeleteSharedClauseList(prfs_WorkedOffClauses(Search),prfs_WorkedOffSharingIndex(Search));
  clause_DeleteSharedClauseList(prfs_UsableClauses(Search),prfs_UsableSharingIndex(Search));
  clause_DeleteSharedClauseList(prfs_DocProofClauses(Search),prfs_DocProofSharingIndex(Search));
  sharing_IndexDelete(prfs_WorkedOffSharingIndex(Search));
  sharing_IndexDelete(prfs_UsableSharingIndex(Search));
  if (prfs_DocProofSharingIndex(Search))
    sharing_IndexDelete(prfs_DocProofSharingIndex(Search));

  for (Scan=prfs_SplitStack(Search);!list_Empty(Scan);Scan=list_Cdr(Scan))
    prfs_SplitFree(list_Car(Scan));
  list_Delete(prfs_SplitStack(Search));
  memory_Free(Search,sizeof(PROOFSEARCH_NODE));
}

PROOFSEARCH prfs_Create(BOOL Document)
/**************************************************************
  INPUT:   A flag indicating whether a proof documentation should take place.
  RETURNS: A new proof search object. The worked off and usable
           indexes are created whilst the docproof index and the
	   sort theories are not created, since they are not
	   needed in general.
***************************************************************/
{
  PROOFSEARCH Result;

  Result = memory_Malloc(sizeof(PROOFSEARCH_NODE));

  Result->woindex  = sharing_IndexCreate();
  Result->wolist   = list_Nil();
  Result->usindex  = sharing_IndexCreate();
  Result->uslist   = list_Nil();

  Result->astatic  = (SORTTHEORY)NULL;
  Result->adynamic = (SORTTHEORY)NULL;
  Result->dynamic  = (SORTTHEORY)NULL;
  
  if (Document)
    Result->dpindex  = sharing_IndexCreate();
  else
    Result->dpindex  = (SHARED_INDEX)NULL;
  Result->dplist   = list_Nil();
  
  Result->stack         = list_StackBottom();
  Result->validlevel    = 0;
  Result->lastbacktrack = 0;
  Result->splitcounter  = 0;
  Result->keptclauses   = 0;
    
  return Result;  
}

void prfs_InsertWorkedOffClause(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  MEMORY:  The clause is assumed to be unshared.
  EFFECT:  The clause is inserted into the worked off sharing index
           and list of <Search>. The unshared literals are deleted.
  RETURNS: void.
***************************************************************/
{
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_InsertWorkedOffClause\n", stderr);
    misc_DumpCore();
  }
#endif 

  clause_SetFlag(Clause,WORKEDOFF);
  prfs_PutWorkedOffClauses(Search,list_Cons(Clause, prfs_WorkedOffClauses(Search)));
  clause_InsertIntoSharing(Clause, prfs_WorkedOffSharingIndex(Search));
  prfs_InsertInSortTheories(Search, Clause);
}


void prfs_InsertUsableClause(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  MEMORY:  The clause is assumed to be unshared.
  EFFECT:  The clause is inserted into the usable sharing index
           and list of <Search> sorted with respect to their weight. 
	   The unshared literals are deleted.
  RETURNS: void.
***************************************************************/
{    
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_InsertUsableClause\n", stderr);
    misc_DumpCore();
  }
#endif 
    
  prfs_PutUsableClauses(Search,clause_InsertWeighed(Clause, prfs_UsableClauses(Search)));
  clause_InsertIntoSharing(Clause, prfs_UsableSharingIndex(Search));
}

void prfs_InsertDocProofClause(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  MEMORY:  The clause is assumed to be unshared.
  EFFECT:  The clause is inserted into the prrof documentation sharing index.
	   The unshared literals are deleted.
  RETURNS: void.
***************************************************************/
{ 
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_InsertDocProofClause\n", stderr);
    misc_DumpCore();
  }
#endif   
  
  if (prfs_DocProofSharingIndex(Search) == (SHARED_INDEX)NULL)
    clause_Delete(Clause);
  else {
    prfs_PutDocProofClauses(Search,list_Cons(Clause, prfs_DocProofClauses(Search)));
    clause_InsertIntoSharing(Clause, prfs_DocProofSharingIndex(Search));
  }
}

void prfs_MoveUsableWorkedOff(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is inserted into the worked off sharing index
           and list and it is deleted from the usable index and list.
	   In particular, the WorkedOff flag is set and if <Clause> is a 
	   declaration clause, it is inserted into the respective sort theories.
  RETURNS: void.
***************************************************************/
{
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_MoveUsableWorkedOff\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_PutUsableClauses(Search,list_PointerDeleteElement(prfs_UsableClauses(Search),Clause));
  clause_SetFlag(Clause,WORKEDOFF);
  clause_MoveSharedClause(Clause,prfs_UsableSharingIndex(Search),prfs_WorkedOffSharingIndex(Search));
  prfs_PutWorkedOffClauses(Search,list_Cons(Clause, prfs_WorkedOffClauses(Search)));
  prfs_InsertInSortTheories(Search, Clause);
}

void prfs_MoveWorkedOffDocProof(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is inserted into the doc proof sharing index
           and list of <Search> and it is deleted from the worked off
	   index and list.
  RETURNS: void.
***************************************************************/
{
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_MoveWorkedOffDocProof\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_DeleteFromSortTheories(Search, Clause);
  prfs_PutWorkedOffClauses(Search,list_PointerDeleteElement(prfs_WorkedOffClauses(Search),Clause));
  clause_RemoveFlag(Clause,WORKEDOFF);

  if (prfs_DocProofSharingIndex(Search) == (SHARED_INDEX)NULL)
    clause_DeleteFromSharing(Clause,prfs_WorkedOffSharingIndex(Search));
  else {
    clause_MoveSharedClause(Clause,prfs_WorkedOffSharingIndex(Search),prfs_DocProofSharingIndex(Search));
    prfs_PutDocProofClauses(Search,list_Cons(Clause, prfs_DocProofClauses(Search)));
  }
}

void prfs_MoveUsableDocProof(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is inserted into the doc proof sharing index
           and list of <Search> and it is deleted from the usable
	   index and list.
  RETURNS: void.
***************************************************************/
{
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in  prfs_MoveUsableDocProof\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_PutUsableClauses(Search,list_PointerDeleteElement(prfs_UsableClauses(Search),Clause));

  if (prfs_DocProofSharingIndex(Search) == (SHARED_INDEX)NULL)
    clause_DeleteFromSharing(Clause,prfs_UsableSharingIndex(Search));
  else {
    clause_MoveSharedClause(Clause,prfs_UsableSharingIndex(Search),prfs_DocProofSharingIndex(Search));
    prfs_PutDocProofClauses(Search,list_Cons(Clause, prfs_DocProofClauses(Search)));
  }
}

void prfs_MoveInvalidClausesDocProof(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  All clauses that have a split level higher than the
           current split level of <Search> are moved to the
	   proof documentation index. If it does not exist, i.e.,
	   no proof documentation required, the clauses are
	   deleted.
  RETURNS: void.
***************************************************************/
{
  LIST   Scan,Invalid;
  CLAUSE Clause;

  Invalid = list_Nil();
  for(Scan=prfs_WorkedOffClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (clause_SplitLevel(Clause) > prfs_ValidLevel(Search))
      Invalid = list_Cons(Clause,Invalid);
  }
  for(Scan=Invalid;!list_Empty(Scan);Scan=list_Cdr(Scan))
    prfs_MoveWorkedOffDocProof(Search,list_Car(Scan));
  list_Delete(Invalid);

  Invalid = list_Nil();
  for(Scan=prfs_UsableClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (clause_SplitLevel(Clause) > prfs_ValidLevel(Search))
      Invalid = list_Cons(Clause,Invalid);
  }
  for(Scan=Invalid;!list_Empty(Scan);Scan=list_Cdr(Scan))
    prfs_MoveUsableDocProof(Search,list_Car(Scan));
  list_Delete(Invalid);
  
}

void prfs_ExtractWorkedOff(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is removed from the worked off index and
           list and returned as an unshared clause.
	   Sort theories are updated accordingly.
  RETURNS: void.
***************************************************************/
{
  prfs_DeleteFromSortTheories(Search, Clause);
  clause_RemoveFlag(Clause,WORKEDOFF);
  prfs_PutWorkedOffClauses(Search,list_PointerDeleteElement(prfs_WorkedOffClauses(Search),Clause));
  clause_MakeUnshared(Clause,prfs_WorkedOffSharingIndex(Search));
}

void prfs_ExtractUsable(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is removed from the usable off index and
           list and returned as an unshared clause.
	   Sort theories are updated accordingly.
  RETURNS: void.
***************************************************************/
{
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_ExtractUsable\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_PutUsableClauses(Search,list_PointerDeleteElement(prfs_UsableClauses(Search),Clause));
  clause_MakeUnshared(Clause,prfs_UsableSharingIndex(Search));
}

void prfs_DeleteWorkedOff(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is deleted from the worked off index and list.
	   Sort theories are updated accordingly.
  RETURNS: void.
***************************************************************/
{
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_DeleteWorkedOff\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_DeleteFromSortTheories(Search, Clause);
  prfs_PutWorkedOffClauses(Search,list_PointerDeleteElement(prfs_WorkedOffClauses(Search),Clause));
  clause_DeleteFromSharing(Clause,prfs_WorkedOffSharingIndex(Search));
}


void prfs_DeleteUsable(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a clause.
  EFFECT:  The clause is deleted from the usable index and list.
  RETURNS: void.
***************************************************************/
{
#if DEBUG
  if (!clause_IsClause(Clause)) {
    fputs("\n\nIllegal input in prfs_DeleteUsable\n", stderr);
    misc_DumpCore();
  }
#endif 

  prfs_PutUsableClauses(Search,list_PointerDeleteElement(prfs_UsableClauses(Search),Clause));
  clause_DeleteFromSharing(Clause,prfs_UsableSharingIndex(Search));
}

static void prfs_PrintSplit(SPLIT Split)
/**************************************************************
  INPUT:   A split.
  RETURNS: None, prints the information kept in the split structure.
  MEMORY:  No effects.
***************************************************************/
{
  LIST Scan;

    printf("\n Split: %d %ld", prfs_SplitSplitLevel(Split), (long)Split);
    printf("\n Father: ");
    if (prfs_GetFatherClause(Split) != (CLAUSE)NULL)
      clause_Print(prfs_GetFatherClause(Split));
    else
      printf("No father, unnecessary split.");
    
    if (prfs_IsSplitDirty(Split))
      printf("\n Split is dirty.");
    else
      printf("\n Split is clean.");
    printf("\n Blocked clauses:");
    for(Scan=prfs_SplitBlockedClauses(Split);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      printf("\n ");
      clause_Print(list_Car(Scan));
    }
    printf("\n Deleted clauses:");
    for(Scan=prfs_SplitDeletedClauses(Split);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      printf("\n ");
      clause_Print(list_Car(Scan));
    }
}


void prfs_PrintSplitStack(PROOFSEARCH PS)
/**************************************************************
  INPUT:   None.
  RETURNS: None, prints almost all the information kept in the
           split structure.
  MEMORY:  No effects.
***************************************************************/
{
  LIST Scan;
  int  Level;

  printf("\n Splitstack:");
  Level = prfs_ValidLevel(PS);

  for (Scan = prfs_SplitStack(PS); !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    printf("\n Level: %d",Level--);
    prfs_PrintSplit(list_Car(Scan));
  }
}

void prfs_Print(PROOFSEARCH Search)
/**************************************************************
  INPUT:   A proof search object.
  EFFECT:  The proof search object is printed to stdout.
  RETURNS: void.
***************************************************************/
{
  LIST Scan;

#if DEBUG
  if (!prfs_Check(Search)) {
    fputs("\n\nIllegal input in prfs_Print\n", stderr);
    misc_DumpCore();
  }
#endif 

  printf("\n\n Proofsearch: Current Level: %d Last Backtrack Level: %d Splits: %d",
	 prfs_ValidLevel(Search),prfs_LastBacktrackLevel(Search),prfs_SplitCounter(Search));
  printf("\n Worked Off Clauses:");
  for(Scan=prfs_WorkedOffClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    printf("\n ");
    clause_Print(list_Car(Scan));
  }
  printf("\n Usable Clauses:");
  for(Scan=prfs_UsableClauses(Search);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    printf("\n ");
    clause_Print(list_Car(Scan));
  }
  prfs_PrintSplitStack(Search);
  printf("\n Static Sort Theory:");
  sort_TheoryPrint(prfs_StaticSortTheory(Search));
  printf("\n Dynamic Sort Theory:");
  sort_TheoryPrint(prfs_DynamicSortTheory(Search));
  printf("\n Approximated Dynamic Sort Theory:");
  sort_TheoryPrint(prfs_ApproximatedDynamicSortTheory(Search));
  printf("\n");
}

SPLIT prfs_SplitCreate(PROOFSEARCH PS)
/**************************************************************
  INPUT:   A proofsearch object
  RETURNS: A pointer to a new split node, increases the valid level
  MEMORY:  Allocates memory for the SPLIT_NODE.
***************************************************************/
{
  SPLIT Result;

  prfs_IncValidLevel(PS);

  Result                 = (SPLIT)memory_Malloc(sizeof(SPLIT_NODE));
  Result->splitlevel     = prfs_ValidLevel(PS);
  Result->dirty          = TRUE;
  Result->blockedClauses = list_Nil();
  Result->deletedClauses = list_Nil();
  Result->father         = (CLAUSE) NULL;
  return Result;
}


CLAUSE prfs_DoSplitting(PROOFSEARCH PS, CLAUSE SplitClause, LIST Literals)
/**************************************************************
  INPUT:   An unshared clause to be splitted where 'Literals' is the list
           of literals to keep (in their order in the SplitClause) and 
	   an index where the splitted clause is inserted,
	   and a proofsearch object
  RETURNS: A pointer to the (stack-, not sharing-) inserted splitted clause.
  MEMORY:  The blocked parts and the actparts literals are created
           unshared, memory for the two (more for HornSplits) new 
	   clausenodes is allocated.
***************************************************************/
{

  SPLIT   NewSplit;
  CLAUSE  NewClause, BlockedClause;
  LITERAL NextLit,NewLit;
  int     i,j,lengthBlocked,lengthNew,lc,la,ls,nc,na,ns;
#if DEBUG_RULES_SPLIT
  if (!(clause_IsClause(SplitClause))) {
    printf("\n\nIllegal input in prfs_DoSplitting\n");
    misc_Error();
  }
#endif

  prfs_DecrementSplitCounter(PS);
  NewSplit = prfs_SplitCreate(PS);

  prfs_SetFatherClause(NewSplit, SplitClause);

  lengthNew = list_Length(Literals);
  lengthBlocked = clause_Length(SplitClause) - lengthNew;

  NewClause = clause_CreateBody(lengthNew);
  BlockedClause = clause_CreateBody(lengthBlocked);
  clause_DecreaseCounter(); /* reset internally increased counter! */
  clause_SetNumber(BlockedClause, 0);
  /* To detect forgotten setting at insertion! */

  lc = clause_LastConstraintLitIndex(SplitClause);
  la = clause_LastAntecedentLitIndex(SplitClause);
  ls = clause_LastSuccedentLitIndex(SplitClause);

  nc = 0; na = 0; ns = 0;

  j = clause_FirstLitIndex();

  for (i = clause_FirstLitIndex(); i <= ls; i++) {
    NextLit = clause_GetLiteral(SplitClause, i);

    NewLit  = clause_LiteralCopy(NextLit);

    if ((lengthNew > 0) && /* To avoid access of Nirvana. */
	list_PointerMember(Literals,NextLit)) {
      /* NewLit is literal for the NewClause. */

      lengthNew--;
      clause_SetLiteral(NewClause, j++, NewLit);
      clause_LiteralSetOwningClause(NewLit, NewClause);
      clause_AddParentClause(NewClause, clause_Number(SplitClause));
      clause_AddParentLiteral(NewClause, i);
      if (i <= lc)
	nc++;
      else if (i <= la)
	na++;
      else 
	ns++;

    } else { /* NewLit is literal for the BlockedClause. */

      clause_SetLiteral(BlockedClause, (i-j), NewLit);
      clause_LiteralSetOwningClause(NewLit, BlockedClause);
      clause_AddParentClause(BlockedClause, clause_Number(SplitClause));
      clause_AddParentLiteral(BlockedClause, i);
    }
  } /* end of 'for all literals'. */

  clause_SetNumOfConsLits(NewClause, nc);
  clause_SetNumOfConsLits(BlockedClause,
			  (clause_NumOfConsLits(SplitClause) - nc));
  clause_SetNumOfAnteLits(NewClause, na);
  clause_SetNumOfAnteLits(BlockedClause,
			  (clause_NumOfAnteLits(SplitClause) - na));
  clause_SetNumOfSuccLits(NewClause, ns);
  clause_SetNumOfSuccLits(BlockedClause,
			  (clause_NumOfSuccLits(SplitClause) - ns));

  clause_ReInit(BlockedClause);
  clause_UpdateSplitDataFromNewSplitting(BlockedClause,
					 SplitClause, prfs_SplitSplitLevel(NewSplit));
  clause_SetFromSplitting(BlockedClause);
  clause_SetParentLiterals(BlockedClause,
    list_NReverse(clause_ParentLiterals(BlockedClause)));

  clause_SetDepth(BlockedClause, clause_Depth(SplitClause)+1);

  prfs_SplitAddBlockedClause(NewSplit, BlockedClause);
  prfs_SplitPutDeletedClauses(NewSplit, list_Nil());

  
  prfs_PushSplitStack(PS, NewSplit);

  clause_ReInit(NewClause);
  clause_UpdateSplitDataFromNewSplitting(NewClause, SplitClause,
					 prfs_SplitSplitLevel(NewSplit));
  clause_SetFromSplitting(NewClause);

  clause_SetParentLiterals(NewClause,
			   list_NReverse(clause_ParentLiterals(NewClause)));

  clause_SetDepth(NewClause, clause_Depth(SplitClause)+1);
  clause_RemoveFlag(NewClause, WORKEDOFF);

  if (clause_IsGround(NewClause)) {
    /* Keep Clauses made from NewClause for refutation case! */
    CLAUSE UnitClause;
    LIST   AtomList;

    la = clause_LastAntecedentLitIndex(NewClause);
    ls = clause_LastSuccedentLitIndex(NewClause);

    Literals = clause_ParentLiterals(NewClause);

    for (i = clause_FirstLitIndex(); i <= ls; i++) {

      NextLit    = clause_GetLiteral(NewClause, i);
      AtomList   = list_List(term_Copy(clause_LiteralAtom(NextLit)));

      if (i <= la)
	UnitClause = clause_Create(list_Nil(), list_Nil(), AtomList);
      else
	UnitClause = clause_Create(list_Nil(), AtomList, list_Nil());

      clause_SetNumber(UnitClause, -1);
      /* To detect forgotten setting at reinsertion! */
      clause_DecreaseCounter();
      /* Reset internally increased counter! */

      list_Delete(AtomList);

      clause_SetFromSplitting(UnitClause);
      clause_UpdateSplitDataFromNewSplitting(UnitClause, SplitClause,
					     prfs_SplitSplitLevel(NewSplit));
      clause_AddParentClause(UnitClause, clause_Number(NewClause));
      clause_AddParentLiteral(UnitClause, i);
      clause_AddParentClause(UnitClause, clause_Number(SplitClause));
      clause_AddParentLiteral(UnitClause, (int)list_Car(Literals));
      Literals = list_Cdr(Literals);
      prfs_SplitAddBlockedClause(NewSplit, UnitClause);
    }
  }
  /* printf("\n\nSPLITTING DONE!");
     printf("\nAus           : "); clause_Print(SplitClause); fflush(stdout);
     printf("\nDer erste Teil: "); clause_Print(NewClause); fflush(stdout);
     printf("\nDer zweite Teil: ");
     clause_Print(BlockedClause); fflush(stdout);
     printf("\nDaher als BlockedClauses:\n");
     clause_ListPrint(prfs_SplitBlockedClauses(NewSplit)); fflush(stdout);
     */
  return NewClause;
}


LIST prfs_GetSplitLiterals(PROOFSEARCH PS, CLAUSE Clause)
/**************************************************************
  INPUT:   A Clause and a proofsearch object
  RETURNS: A list of literals building the bigger part of a 
           variable-disjunct literal partition if one exists,
	   an empty list, else.
  MEMORY:  Allocates memory for the literal list.
***************************************************************/
{
  LITERAL NextLit;
  int     i,length,OldLength;
  LIST    LitList, VarOcc,NextOcc;
  BOOL    Change;

#if DEBUG_RULES_SPLIT
  if (!clause_IsClause(Clause)) {
    printf("\n\nIllegal input in prfs_GetSplitLiterals\n");
    misc_DumpCore();
  }
#endif 

  /* We aren't prepared to cope with empty clauses !!! */
  LitList = list_Nil();

  if (prfs_SplitCounter(PS) != 0) {

    if (clause_HasSuccLits(Clause)) {
      if (clause_HasGroundSuccLit(Clause)) {

	NextLit = clause_GetGroundSuccLit(Clause);
	LitList = list_Cons(NextLit, LitList);

	for (i = clause_LastAntecedentLitIndex(Clause);i >= clause_FirstLitIndex();i--) {
	  NextLit = clause_GetLiteral(Clause, i);
	  if (term_IsGround(clause_LiteralAtom(NextLit)))
	    LitList = list_Cons(NextLit, LitList);
	}
	return LitList;
      }

      NextLit = clause_GetLiteral(Clause, clause_LastLitIndex(Clause)); /* Is a Succedent Literal */
      VarOcc  = term_VariableSymbols(clause_LiteralAtom(NextLit));
      LitList = list_List(NextLit);
      length  = clause_Length(Clause);
      Change  = TRUE;

      while(Change) {
	Change = FALSE;

	for (i=(clause_LastLitIndex(Clause)-1) ; i>=clause_FirstLitIndex(); i--) {

	  NextLit = clause_GetLiteral(Clause, i);
	
	  if (!list_PointerMember(LitList,NextLit)) {
	    NextOcc = term_VariableSymbols(clause_LiteralAtom(NextLit));
	    if (list_HasIntersection(VarOcc, NextOcc)) { 
	      OldLength = list_Length(VarOcc);
	      VarOcc    = list_NPointerUnion(VarOcc, NextOcc);
	      LitList   = list_Cons(NextLit, LitList);
	      if (OldLength < list_Length(VarOcc))
		Change = TRUE;
	    }
	    else
	      list_Delete(NextOcc);
	  }
	}
      }
      if (list_Length(LitList) == length) {
	list_Delete(LitList);
	LitList = list_Nil();
      }
      Change = TRUE;                  /* Check whether not all succedent literls are used */
      for(i=clause_FirstSuccedentLitIndex(Clause) ; i<length && Change;i++)
	if (!list_PointerMember(LitList,clause_GetLiteral(Clause, i)))
	  Change = FALSE;
      if (Change) {
	list_Delete(LitList);
	LitList = list_Nil();
      }
      list_Delete(VarOcc);
    }
  }
  return LitList;
}



CLAUSE prfs_PerformSplitting(PROOFSEARCH Search, CLAUSE Clause)
/**************************************************************
  INPUT:   A proof search object and a usable clause.
  EFFECT:  If <Clause> can be split it is splitted, the first
           part of the split is returned  and the
	   splitted clause is removed from usable and kept in the split stack.
	   Otherwise <Clause> remains unchanged and NULL is returned.
  RETURNS: NULL if <Clause> is not splittable, the first split part otherwise.
***************************************************************/
{
  CLAUSE Result;

  Result = (CLAUSE)NULL;
  
  if (clause_HasSolvedConstraint(Clause)) {
    LIST LitList;

    LitList = prfs_GetSplitLiterals(Search, Clause);

    if (!list_Empty(LitList)) {
      prfs_ExtractUsable(Search,Clause);
      Result = prfs_DoSplitting(Search, Clause, LitList);
      list_Delete(LitList);
    }
  }

  return Result;
}

