/*------------------------------------------------------------------*/
/*      gen.c
/*
/*      Copyright Newborn Software 1996
/*
/*      functions:
/*         generate_clause()        ......................#1
/*         emit_literal()           ......................#2
/*         generate_literal()       ......................#3
/*         repair_literal()         ......................#4
/*         test_equality()          ......................#5
/*------------------------------------------------------------------*/
 
#include "const.h"
#include "clause.h"

/*------------------------------------------------------------------*/
/* externally defined functions                                     */
/*------------------------------------------------------------------*/
 
extern calc_lit_hc();           /* hash.c..calc literal hash code   */
extern calc_clause_hashcode();  /* hash.c..calc clause hash code    */
 
extern long *var_substitution_tbl[];/* tgtp.c..var substitution tbl */
extern clause *clist[];           /* tgtp.c...pointers to clauses   */
 
/*------------------------------------------------------------------*/
/* externally defined constants                                     */
/*------------------------------------------------------------------*/

extern long ply, nextc, base;
extern int o_maxlit;        /* max # lits per clause (0 = no limit) */
extern int o_maxterm;
extern int in_elim;
extern int o_maxvar;
extern long equal_pred;
extern int phase;

/*------------------------------------------------------------------*/
/* internally defined functions                                     */
/*------------------------------------------------------------------*/

clause *generate_clause();
long *emit_literal();
int generate_literal();
void repair_literal();
void repair_clause();
int test_equality();
 
/*------------------------------------------------------------------*/
/* internally defined constants                                     */
/*------------------------------------------------------------------*/
 
static long *ep;                      /* where to emit clause code  */
long constants_in_clause;

/*------------------------------------------------------------------*/ 
/*      *generate_clause(c1,c2,lit1,lit2,ix1,ix2)    ........#1
/*
/*      generate a clause from one or two parent cls
/*      Apply all changes currently in the 
/*      var_substitution_tbl while we gen the new lits. 
/*      The variables are st'dized apart. Note that we destroy
/*      the var_substitution_tbl when we generate the clause
/*
/*      parameters:     c1,c2: left and right parents
/*                      lit1,lit2: literals to leave out
/*                      ix1, ix2 are indices of resolved literals
/*
/*      A clause will not be gen. for various reasons (too many lits,
/*      or the clause is a tautology)
/*
/*      called from:  binary_resolutions()      infer.c
/*                    binary_factors()          infer.c
/*                    bres_ply_1()              elim.c
/*                    bres_link()               elim.c
/*                    bres_replace()            elim.c
/*                    bres_proof_line()         pass2.c
/*                    factor_proof_line()       pass2.c
/*
/*      calls:        generate_literal()        gen.c
/*                    calc_clause_hashcode()    hash.c
/*                    test_equality()           gen.c
/*
/*      returns:      ptr to new clause or NULL if none generated
/*------------------------------------------------------------------*/
 
clause *generate_clause(c1,c2,lit1,lit2,ix1,ix2)
clause *c1, *c2;
literal *lit1, *lit2;
int ix1, ix2;
{
register clause *c;                           /* new clause         */
literal *curl, *prevl;      /* current literal; previous literal    */
register literal *l1, *l2;
register literal *pl;
long oldnv;
int oldnvar;

c = clist[nextc];                      /* where to put the clauses  */
 
/*-----------------------------*/
/* initialize clause header    */
/*-----------------------------*/              
 
c->c_bits = ACTIVE;                       /* assume active clause   */
c->c_lpar = c1;                         /* parents of this clause   */
c->c_rpar = c2;
c->c_nv   = 0L;                                /* filled in later   */
c->c_nlit = 0;                           /* count the # of literals */
c->c_lit  = 0L;
 
/*--------------------------*/
/* prepare to scan literals */
/*--------------------------*/
 
constants_in_clause = 0L;
prevl = 0L;                               /* prev literal in scan   */
ep = (long *) (c + 1);                  /* lit's go after cl header */
 
l1 = c1->c_lit;             /* ptr to first literal; c1 never NULL  */

if(c2) {l2 = c2->c_lit;}
else {l2 = 0L;}
 
while(l1 != 0L || l2 != 0L) {          /* while there are literals  */ 
 
/*------------------------------------------------------*/
/* if l1 or l2 is one of the literals that we are       */
/* not to include then we simply advance the pointer    */
/*------------------------------------------------------*/
 
   if((l1 != 0L) && ((l1 == lit1) || 
                (l1->l_id & HHITMASK) != 0L)) {
      l1 = l1->l_next; continue;}            
   if((l2 != 0L) && ((l2 == lit2) || 
               (l2->l_id & HHITMASK) != 0L)) {
      l2 = l2->l_next; continue;}
 
/*----------------------------------------------------------*/
/* save this information in case literal is deleted---      */
/* in the case that it is identical to another and must     */
/* be deleted                                               */
/*----------------------------------------------------------*/
 
   curl = (literal *) ep;    /* literal that we're making   */
   oldnv = c->c_nv;
	    
   if(l2 == 0L || l1 != 0L) {
      if(generate_literal(c,l1)) {
         return(0L);}
      l1 = l1->l_next;}
   else {
      if(generate_literal(c,l2)) {
         return(0L);}
      l2 = l2->l_next;}
	       
/*----------------------------------------------------------*/
/* eliminate predicates with Equal(x,x) as tautologies      */
/*----------------------------------------------------------*/
 
   if(curl->l_id == equal_pred && phase > 0) { 
      if(test_equality(curl)) {
         return(0L);}}                   /* equal tautology */
 
/*----------------------------------------------------------*/
/* we scan back in the current group to see if there        */
/* are any matching literal checksums. If so we perform     */
/* a literal comparison and a) remove the literal if        */
/* it is identical to another; or b) cancel generation      */
/* of the clause if the literals are identical but of       */
/* opposite signs (tautology).                              */
/*----------------------------------------------------------*/
		
tautology_check:
   for(pl = c->c_lit; pl != 0L; pl = pl->l_next) {
      if((pl->l_X == curl->l_X && pl->l_Y == curl->l_Y) ||
	         (pl->l_X == -curl->l_X && pl->l_Y ==-curl->l_Y)){ 
         if((curl->l_id ^ pl->l_id) != 0L) {

/*-----------------------------------------------------------*/
/* clause is a tautology                                     */
/*-----------------------------------------------------------*/

            if(in_elim == 1) {
                return(0L);}       /* tautology */
            else break;}
	       
/*-----------------------------------------------------------*/
/* merge clause, delete literal (by not adding it)           */
/*-----------------------------------------------------------*/
 
          c->c_nv    = oldnv;
          c->c_bits |= MERGE;               /* remove lit  */
          ep = (long *) curl;
          goto idn;}}
 
/*------------------------------------------------------*/
/* the literal has survived, we patch the previous lit  */
/* so that it points here                               */
/*------------------------------------------------------*/
 
   if(prevl == 0L)  {
      c->c_lit = curl;}
   else  {
      prevl->l_next = curl;}
	
   prevl = curl;
   c->c_nlit++; 
 
/*--------------------------------------------------------------*/
/* if there are too many literals or variables, forget it ...   */
/*--------------------------------------------------------------*/
		
   if(c->c_nlit > o_maxlit) {
      if(o_maxlit)  {
	       return(0L);}}                  
   if((int) c->c_nv > o_maxvar){
      return(0L);}
 
idn:    ;}

if(c->c_nlit == 1) c->c_bits |= UNIT;             /* a unit clause  */
c->c_end = (clause *) ep;                    /* pointer past clause */
c->c_indices = (ix1<<LEFT_INDEX_SHIFT) + ix2; /* parent lit indices */
c->c_id = base + 1L + ply;                  /* give clause an ID    */

calc_clause_hashcode(c);

return(c);
}
 
/*----------------------------------------------------------------*/ 
/*      emit_literal(c,l,ii)           ......................#2
/*
/*      emit_literal is a recursive fxn that will apply the subs
/*      of the var_substitution_tbl to a given agmnt and write the 
/*      expanded version.
/*
/*      parameters:     c : ^ clause header
/*                      l : ^ arguments
/*                      i : number of arguments
/*
/*      Note: we write out the expanded version at the current
/*      "ep" location.
/*
/*      called from:    generate_literal()      gen.c
/*
/*      calls:          emit_literal()          gen.c
/*
/*      returns:        pointer past arguments
/*                      NULL if fcn nesting too deep.
/*--------------------------------------------------------------*/
 
long  *emit_literal(c,l,ii)
clause *c;
long *l,ii;
{
register long sl;
register int i;
 
for(i = (int) ii; i > 0; i--) {            /* for each argument  */
   sl = *l;
   if(sl & FCNMASK) {                   /* FCN:just copy it over */
      *ep++ = sl;
      l++;
      if(sl & ARGMASK) {
	       if((l = emit_literal(c,l,sl & ARGMASK)) == 0L) {
	          return(0L);}}
      else constants_in_clause++;}
   else {
      if((unsigned long) var_substitution_tbl[sl] < MAX_VARS_X2) {
	       if(var_substitution_tbl[sl] == 0L) {
	          var_substitution_tbl[sl] = (long *) ++c->c_nv;}
 
/*----------------*/
/* write it out   */
/*----------------*/
 
	       *ep++ = (long) var_substitution_tbl[sl] - 1L;}
 
      else if(emit_literal(c,var_substitution_tbl[sl],1L) == 0L) {
	       return(0L);} 
      l++;}}
return(l);
}
 
/*----------------------------------------------------------------*/
/*      generate_literal(c,lit)     .....................#3
/*
/*      add literal lit from clause c to new clause
/*
/*      called from:   generate_clause()          gen.c
/*
/*      calls:         emit_literal()             gen.c
/*                     calc_lit_hc()              hash.c
/*
/*      returns:       0 => literal is ok 
/*                     1 => problem with literal (too long, eg)
/*----------------------------------------------------------------*/
 
int generate_literal(c,lit)
clause *c;
literal *lit;
{
register literal *new_literal;
register int new_lit_terms;
	
new_literal = (literal *) ep;                    /* new literal  */
new_literal->l_next = 0L;
new_literal->l_id = lit->l_id;
 
ep = (long  *) (new_literal + 1L);                /* args go here */
 
if(emit_literal(c,lit + 1L,lit->l_id & ARGMASK) == 0L) {
   return(1);}
 
/*------------------------------------------*/
/* determine a hash code for the literal    */
/* and receive the number of term in lit    */
/*------------------------------------------*/
 
new_lit_terms = calc_lit_hc(new_literal,c);
	
/*------------------------------------------*/
/* if too many terms in lit, kill clause    */
/*------------------------------------------*/
 
if(new_lit_terms > o_maxterm) {
   if(in_elim == 1) {
      return(1);}}
 
return(0);
}
 
/*--------------------------------------------------------------*/
/*      repair_literal(c,index)       ......................#4
/*
/*      replaces a literal with its equivalent hash code when, for
/*      example, the literal is of the form ~Equal(x,x) since this
/*      literal always resolves with Equal(x,x).
/*
/*      called from:    install_clause()            search.c
/*
/*      calls:          calc_clause_hashcode()      hash.c
/*
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void repair_literal(c,index)
clause *c;
int index;
{
register int lit_index;
register literal *l;
 
l = c->c_lit;
for(lit_index = index; lit_index > 1; lit_index--) {
   l = l->l_next;}
 
/*-------------------------------*/
/* remove literal by setting     */
/* the bit l->l_id & HHITMASK    */ 
/* to record literal             */
/* is resolved away              */
/*-------------------------------*/
 
l->l_id |= HHITMASK;
 
/*-------------------------------*/
/* if only one literal remains,  */
/* set c_bits field for unit cls */
/*-------------------------------*/
 
if(--(c->c_nlit) == 1) {
   c->c_bits |= UNIT;}
calc_clause_hashcode(c);
 
return;
}
 
/*--------------------------------------------------------------*/
/*      repair_clause(c,index)       ......................#5
/*
/*      replaces all but one literal (index) with its hash code
/*
/*      called from:    install_clause()            search.c
/*
/*      calls:          calc_clause_hashcode()      hash.c
/*
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void repair_clause(c,index)
clause *c;
int index;
{
register int lit_index;
register literal *l;
 
lit_index = 0;
for(l = c->c_lit; l->l_next != 0L; l = l->l_next) {
   lit_index++;
   if(lit_index != index) {
 
/*-------------------------------*/
/* remove literal by setting     */
/* the bit l->l_id & HHITMASK    */ 
/* to record literal             */
/* is resolved away              */
/*-------------------------------*/
 
      if((l->l_id & HHITMASK) == 0L){
         l->l_id |= HHITMASK;
 
/*-------------------------------*/
/* if only one literal remains,  */
/* set c_bits field for unit cls */
/*-------------------------------*/
 
         if(--(c->c_nlit) == 1) {
            c->c_bits |= UNIT;
            calc_clause_hashcode(c);
            return;}}}}
}
 
/*--------------------------------------------------------------*/
/*      test_equality_(l)       ......................#6
/*
/*      tests to see if equal(string1,string1) is the case
/*
/*      called from:   generate_clause()               gen.c
/*                     install_clause()                search.c
/*
/*      calls:         nobody
/*
/*      returns:    0 => if equal(string1,string1) is not the case
/*                  1 => if equal(string1,string1) is the case
/*--------------------------------------------------------------*/
 
int test_equality(l_equal)
literal *l_equal;
{
long *p, *p1, a;
 
p1 = p = &l_equal->l_id;

for(a = (*p++ & ARGMASK) ; a > 1L ; a--) {
   if(*p & FCNMASK) {
      a += *p++ & ARGMASK;}
   else {
      p++;}}
 
/*--------------------------------------*/
/* p now points to beginning of 2nd     */
/* half of the equality predicate.      */
/* p1 now points to the beginning.      */
/* now a scan is made comparing each    */
/* symbol in the two halfs              */
/*--------------------------------------*/

for(a = (*p1++ & ARGMASK) ; a > 1L ; a--) {
   if(*p != *p1) { 
      return(0);}
   if(*p1 & FCNMASK) {
      a +=*p1++ & ARGMASK;
      p++;}
   else {
      p++; p1++;}}
return(1);
}
