/*----------------------------------------------------------------*/
/*      hash.c
/*
/*      Copyright Newborn Software 1996
/*
/*      functions:
/*
/*      These two functions check whether a clause is in the
/*      clause hash table or whether a new clause can be added
/*
/*      is_clause_in_ht().............................#1
/*      add_clause_to_ht()............................#2
/*
/*      The next four functions deal with the management of the unit
/*      clause hash table.
/*
/*      contradiction()...............................#3
/*      is_unit_in_ht()...............................#4
/*      is_comp_in_ht()...............................#5
/*
/*      The following five functions deal with calculating hashcodes
/*      of literals and clauses.
/*
/*      calc_lit_hc().................................#6
/*      normalize_literal()...........................#7
/*      calc_alt_hashcodes()..........................#8
/*      calc_alt_hash()...............................#9
/*      calc_clause_hashcode()  ......................#10
/*----------------------------------------------------------------*/
 
#include <stdio.h>
#include <math.h>
 
#include "const.h"
#include "clause.h"

/*-----------------------------------------------------------------*/
/* externally defined functions                                    */ 
/*-----------------------------------------------------------------*/

extern fatal();

/*-----------------------------------------------------------------*/
/* externally defined constants                                    */
/*-----------------------------------------------------------------*/

extern hnode *clause_hash_table;   /* ptrs to clause hash tables   */
extern long array_size_m1;
 
extern rnode *rand_array;           /* arrary for random numbers   */
 
extern long s_nod, iter_depth, ply, nextc;
extern int o_hash, o_lit;
extern int o_maxterm;
extern int number_fxns, max_lit;
extern int number_const_hash;
extern int number_constants;
extern int extra_pass, phase;
extern int in_elim;

extern long s_unit_sub;
extern long s_unit_res;
extern long s_htelim;
 
extern int numb_rows, rows_less_nfxns;
 
extern long s_htentry, s_hthit, s_htfail;
extern long s_ltentry, s_lthit, s_ltfail;
 
extern FILE  *outfile;   

/*-----------------------------------------------------------------*/
/* internally defined functions                                    */
/*-----------------------------------------------------------------*/ 

int is_clause_in_ht();
long add_clause_to_ht();
int contradiction();
int is_unit_in_ht();
int is_comp_in_ht();
int calc_lit_hc();
void normalize_literal();
long calc_alt_hashcodes();
void calc_alt_hash();
void calc_clause_hashcode();

/*-----------------------------------------------------------------*/
/* internally defined constants                                    */
/*-----------------------------------------------------------------*/ 

int contradiction_found;
int var_array[64];
int var_array_ptr;
 
long magic_hashx,  magic_hashy;  /* hash code for contradictory lit */
long alt_hx,  alt_hy;          /* alternate hash codes for variants */
 
/*----------------------------------------------------------*/
/* displacements from probe location into hash_table        */
/*----------------------------------------------------------*/
 
static long probe[] = 
{0L,311L,-753L,1726L,-2159L,2420L,-3821L,4522L,
		  -5223L,5724L,-6625L,8326L,-11227L,0L};
      
/*-----------------------------------------------------------------*/
/*      is_clause_in_ht(c)      .............................#1
/*
/*      see if clause c is in the hash table
/*      if so, return 0, otherwise, return -1.
/*
/*      called by:  install_clause()         search.c
/*                  bres_ply_1               elim.c
/*
/*      calls:      nobody
/*
/*      returns:    0 if clause not in TT, depth+1 of node if found
/*-----------------------------------------------------------------*/
 
int is_clause_in_ht(c)
clause *c;
{
register hnode *hn;
register long *p, ht;
	     
p = probe;                                  /* ht probe pointer   */
 
do {
 
/*-------------------------*/
/* get bits to decide the  */
/* table location to probe */
/*-------------------------*/   
 
   ht = (c->c_X + *p) & array_size_m1;
   hn = &clause_hash_table[ht];
 
/*-----------------------------------------*/
/* return a -1 if hash table does not have */
/* any entry at this hash table location   */
/*-----------------------------------------*/
 
   if(hn->h_X == 0L) {
      return(-1);}
 
/*-----------------------------------------*/
/* if same hash code found, check to see   */
/* if clause was a base clause; if so      */
/* return a hit.  also if clause was       */
/* previously found on same iteration      */
/* and no deeper in tree, return a hit.    */
/*-----------------------------------------*/
 
   if(hn->h_X == c->c_X) {
      if(hn->h_Y == c->c_Y) {
         if(hn->h_in == 0) {
            s_htelim++;
            return(0);}
         if(hn->h_in == (int) iter_depth) {
            if(((long) hn->h_ply & PLY_MASK) <= ply) {
               s_htelim++;
 
/*-----------------------------*/
/* return 0 on a hit           */
/*-----------------------------*/
 
               return(0);}}}}   
   } while(*(++p));
 
/*-----------------------------*/
/* clause not in hash table    */
/*-----------------------------*/
 
return(-1);
}
 
/*-----------------------------------------------------------------*/
/*    add_clause_to_ht(c)           ........................#2
/*
/*    called by:    hash_all_base_clauses()              tgtp.c
/*                  prove_by_iterative_deepening()       tgtp.c
/*                  install_clause()                     search.c
/*                  add_link_resolvents()                elim.c
/*                  replace_clause()                     elim.c
/*
/*    calls:        nobody
/*
/*    returns:      nil
/*-----------------------------------------------------------------*/
 
long add_clause_to_ht(hx,hy,lits)
long hx, hy;
int lits;
{
register long *p, ht, best_yet;
register hnode *hn, *hm;
 
if(o_hash == 0) {
   return(0L);}

/*-------------------------------------------------*/
/* if clause has only one literal, see if          */
/* complement has been previously hashed.          */
/*-------------------------------------------------*/
 
if(lits == 1L) {
   contradiction_found = contradiction(hx,hy);
   if(contradiction_found) {
      return(2L);}}

/*--------------------------------------------------*/
/* now set p to point to probe displacement array   */
/*--------------------------------------------------*/
 
best_yet = -1L;
p = probe;
 
do {
   
/*------------------------------------*/
/* get bits to decide hash table      */
/* location to probe                  */
/*------------------------------------*/   
   
   ht = (hx + *p) & array_size_m1;
   hn = &clause_hash_table[ht];
 
/*--------------------------------------------------------------*/
/* enter new clauses if hn->h_X is zero (ie, no entry)          */
/*                                                              */
/* each entry is four words                                     */
/*                                                              */
/* hash code (2 long words)                                     */
/* iteration (1 integer)                                        */
/* ply (1 integer) + 256 * number of literals in clause         */
/*--------------------------------------------------------------*/ 
	
   if(hn->h_X == 0L) {
      hn->h_X     = hx;
      hn->h_Y     = hy;
      hn->h_in    = (int) iter_depth;
      hn->h_ply   = (int) (ply) + 256 * lits;  
      if(lits == 1L) {
         s_ltentry++;}
      else {
         s_htentry++;}
      return(0L);}
        
/*------------------------------*/
/* update an existing entry     */
/*------------------------------*/
        
   if(hn->h_X == hx && hn->h_Y == hy) { 
      if(lits == 1L) {
         s_lthit++;}
      else {
         s_hthit++;}
      if(((long) hn->h_ply & PLY_MASK) > 1L) {
         if((long) hn->h_in == iter_depth) {	
            if(((long) hn->h_ply & PLY_MASK) >= ply) {
               hn->h_ply = (int) (ply + 256 * lits);}
            return(1L);}
         else {               /* if different iteration */ 
            hn->h_ply = (int) (ply + 256 * lits);
            hn->h_in  = (int) iter_depth;
            return(1L);}}
      else hn->h_in = (int) iter_depth; /* entry is at ply 1	*/
      return(1L);}
 
/*---------------------------------------------*/
/* if another entry is already here, update    */
/* best_yet (best_yet is the best previous     */
/* entry to replace.  that entry with the      */
/* most literals is chosen first, and then     */
/* the one deepest in the tree is chosen next. */
/*---------------------------------------------*/
 
   else {
      if(hn->h_ply > (int) best_yet) {
         best_yet = (int) hn->h_ply;
         hm = hn;}}
   } while(*(++p));
 
/*--------------------------------------------------------------*/
/* if no location found for new entry, (after trying 13 probes) */
/* overwrite an existing entry -- the one with most literals    */
/* and deepest in tree -- if it is a worse clause than the      */
/* current one.                                                 */
/*--------------------------------------------------------------*/
/*   
if((int) ply + 256 * lits < hm->h_ply) { */
   hm->h_X 	= hx;
   hm->h_Y 	= hy; 
   hm->h_ply  = (int) ply + 256 * lits;
   hm->h_in   = (int) iter_depth;/*}*/
if(lits == 1L) {
   s_ltfail++;}
else {
   s_htfail++;}
return(0L);                        /* overwrite entry return    */
}
 
/*--------------------------------------------------------------*/
/*      contradiction(l)          ......................#3
/*
/*      test unit to see whether its complement is in the
/*      unit clause hash table.
/*
/*      called by:  is_unit_in_hash()        hash.c
/*
/*      calls:      nobody
/*
/*      returns:    0  if no contradiction found
/*                  1  if found (negative hashcode found
/*---------------------------------------------------------------*/
 
int contradiction(lx,ly)
long lx,ly;
{
register long *p, ht;
register hnode *l_try;
 
p = probe;                               /* probe pointer       */
    
do {   
   ht = (-lx + *p) & array_size_m1;
   l_try = &clause_hash_table[ht];
        
   if(l_try->h_X == 0L) {
      return(0);}
   else 
      if(l_try->h_X == -lx && l_try->h_Y == -ly) {
	       magic_hashx = lx;
	       magic_hashy = ly;
         s_unit_res++;
	       s_ltentry++; 
	       return(1);}                                    
      } while(*(++p));
 
return(0);                                 /* no contradiction */
}
 
/*---------------------------------------------------------------*/
/*      is_unit_in_ht(c)          .................#4
/*
/*      examines clause c to see if one of the literals
/*      in c is in hash table and
/*          (1) was placed in the hash table on the
/*              current iteraltion
/*          (2) the hash table unit was at a ply
/*              less than or equal to the current ply
/*
/*      called by:  calc_inferences()          search.c
/*
/*      calls:      nobody
/*
/*      returns:    1 if literal found
/*                  0 if literal not found
/*---------------------------------------------------------------*/
 
int is_unit_in_ht(c)
clause *c;
{
register long  *p, ht;
register literal *l;
register hnode *l_try;
int index;
 
/*----------------------------------*/
/* test for the same literal        */
/*----------------------------------*/
    
index = 0;
 
for(l = c->c_lit ; l != 0L ; l = l->l_next) {  /* scan literals     */
   index++;
   p = probe;                                  /* probe pointer     */
   if((l->l_id & HHITMASK) == 0L) {   /* if literal not masked out  */
      do {
         ht = (l->l_X + *p) & array_size_m1;
         l_try = &clause_hash_table[ht];
	 
         if(l_try->h_X == 0L) break;  /* literal not in hash table  */
            if(l_try->h_X == l->l_X && l_try->h_Y == l->l_Y) {
 
/*----------------------------------*/
/* unit found in hash table         */
/*----------------------------------*/

               if((long) l_try->h_in == iter_depth) { 
                  if(((long) l_try->h_ply & PLY_MASK) <= ply &&
                           ply >= iter_depth){
                     s_unit_sub++;
                     return(index);}}
               else if(l_try->h_in < 2 && 
                    ((long) l_try->h_ply & PLY_MASK) < 2L) {
                  s_unit_sub++;
		              return(index);}}           /* lit in hash_table */
      } while(*(++p));}}
return(0);                               /* wasn't found in table */
}
 
/*----------------------------------------------------------------*/
/*      is_comp_in_ht(c)       .............................#5
/*
/*      examines clause c to see if the complement of
/*      a unit in c is in the hash table;
/*
/*      returns: the index of the first such literal found.
/*
/*      called by:  install_clause()        search.c
/*
/*      calls:      nobody
/*
/*      returns:    0   if literal not found in clause  c  or
/*                  lit_index  (the index of the literal found)
/*----------------------------------------------------------------*/
 
int is_comp_in_ht(c)
clause *c;
{
register long  *p, ht;
register int lit_index; /* records the index of the complem'y lit */ 
register literal *l;
register hnode *l_try;
 
lit_index = 0;
 
/*--------------------------------------------------*/
/* scan literals of c for a complementary match     */
/*--------------------------------------------------*/
 
for(l = c->c_lit ; l != 0L ; l = l->l_next) {
   p = probe;                                   /* probe pointer     */
   lit_index++;
   if((l->l_id & HHITMASK) == 0L) {
      do {
         ht = (-l->l_X + *p) & array_size_m1;
         l_try = &clause_hash_table[ht];
         if(l_try->h_X == 0L) {break;}
         if(l_try->h_X == -l->l_X) {
            if(l_try->h_Y == -l->l_Y) {
               s_unit_res++;
               return(lit_index);}}} while (*(++p));}}
 
return(0);             /* complementary literal not in table   */
}
 
/*---------------------------------------------------------------*/
/*      calc_lit_hc(l,c)       ..........................#6
/*
/*      calc_lit_hc will compute the hashcode of a lit and
/*      store the result in the l_xsum field
/*
/*      called by  hash_all_base_clauses()       tgtp.c
/*                 generate_literal()            gen.c
/*
/*      calls:     nobody
/*
/*      returns:   lt_terms = number of terms in lit
/*---------------------------------------------------------------*/
 
int calc_lit_hc(l,c)
literal *l;
clause *c;
{
register long a, s, s1, t, *p, q;    
register rnode *r1;
int lt_terms;            /* l_terms = # of terms in lit  */
 
/*-------------------------------------------------------*/
/* p..........a pointer to the first term of the literal */
/* q..........used as an index into the rand_array       */
/* s,t........two words of 64-bit hash code for lit l    */
/* lt_terms...count of the number of terms in lit l      */
/*-------------------------------------------------------*/ 
 
p = &l->l_id;        
s = t = q = 0L;     
lt_terms = 0;
 
for(a = *p++ & ARGMASK; a > 0L; a--) {
 
/*--------------------------------------------------------*/
/* get next term; determine the number of arguments       */
/* increment the count of the number of terms in lit;     */
/* return during Phase 1,2 if too many terms in literal   */
/*--------------------------------------------------------*/
 
   if(lt_terms++ > o_maxterm) {
      if(in_elim == 1) {
         return(lt_terms);}}
      
   if(*p & FCNMASK) {           
 
/*---------------------------------------------------------*/
/* function found; set pointer into rand_array             */
/*---------------------------------------------------------*/
 
      r1 = 
        &rand_array[((*p&IAMASK)>>ARGSHIFT)+rows_less_nfxns+q];
      a += *p++ & ARGMASK;}
   else {                     
 
/*---------------------------------------------------------*/
/* variable found; set pointer into rand_array             */
/*---------------------------------------------------------*/
 
      r1 = &rand_array[(*p & MAX_VARS_M1) + q];
      p++;}
	
/*---------------------------------------------------------*/
/* accumulate hashcode in s and t                          */
/*---------------------------------------------------------*/
 
   s = s ^ r1->r_hash1;        
   t = t ^ r1->r_hash2;
	
/*---------------------------------------------------------*/
/* set q to index of top of next row in rand_array         */
/*---------------------------------------------------------*/
 
   q = q + numb_rows;}

/*---------------------------------------------------------*/
/* add in hash code for predicate                          */
/*---------------------------------------------------------*/
 
s1 = (&rand_array[(l->l_id & IAMASK) >> ARGSHIFT])->r_hash1;
s = s ^ s1;
t = t ^ (-s1);

/*---------------------------------------------------------*/
/* If complemented literal, negate hashcode                */
/* save value in literal header elements, l->l_X,l->l_Y    */
/*---------------------------------------------------------*/
 
if(l->l_id & NEGMASK) { 
   l->l_X = -s;
   l->l_Y = -t;}        
else {
   l->l_X = s;
   l->l_Y = t;}
	
return(lt_terms);
}
 
/*------------------------------------------------------------------*/
/*      normalize_literal(l,c)         ....................#7
/*
/*      called by:      install_clause()    search.c
/*                      calc_inferences()   search.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*------------------------------------------------------------------*/
 
void normalize_literal(l,c)
literal *l;
clause *c;
{
register long a, *p, k; /* l_terms = # of terms in lit  */
 
p = &l->l_id;        
var_array_ptr = 0L;     
 
for(k = 0L ; k < (long) c->c_nv ; k++) {var_array[k] = 0;}
	
for(a = *p++ & ARGMASK; a > 0L; a--) {
   if((*p & FCNMASK) != 0L) {           
      a += *p++ & ARGMASK;}
   else {               
 
/*-----------------------------------------------------------*/
/* see if var_array has det a new var name for this variable */
/*-----------------------------------------------------------*/
 
      if(var_array[*p & MAX_VARS_M1] == 0) {
 
/*---------------------------------------------------------------*/
/* if not, enter new var name                                    */
/*---------------------------------------------------------------*/
 
	     var_array[*p & MAX_VARS_M1] = var_array_ptr++ + 1;}
 
      *p = var_array[*p & MAX_VARS_M1] - 1;
      if(c->c_bits & VID) *p += MAX_VARS;
      p++;}}
 
c->c_nv = var_array_ptr;
return;
}
 
/*--------------------------------------------------------------*/
/*      calc_alt_hashcodes(l,c)   ................#8
/*
/*      calculates var in the formula, and for each constant
/*      in the list of constants for theorem, calculate a hash
/*      code and add it to the clause hash table
/*
/*      called by:      install_clause()               search.c
/*
/*      calls:          calc_alt_hash()                hash.c
/*                      add_clause_to_ht()             hash.c
/*
/*      returns:        0L;2L if contradiction found
/*---------------------------------------------------------------*/
 
long calc_alt_hashcodes(l,c) 
literal *l;
clause *c;
{
register long const1, const2, varbl, varbl2;

/*-----------------------------------------------*/
/* if there is one variable, replace it with     */
/* each constant and calculate the resulting     */
/* hash code.  add the hash code to the clause   */
/* hash table.                                   */
/*-----------------------------------------------*/
 
if(c->c_nv == 1L) {
   for(const1 = 1L ; const1 <= (long) number_constants ; const1++) {
      calc_alt_hash(l,0L,-1L,const1,0L);    
      if(add_clause_to_ht(alt_hx,alt_hy,c->c_nlit)==2L) {
	     return(2L);}}

/*-----------------------------------------------*/
/* calc hashcode with single variable replaced   */
/* with the variable y.  most likely, but not    */
/* for sure, this will produce a new hashcode    */
/*-----------------------------------------------*/
 
   calc_alt_hash(l,0L,-1L,1L - (long) rows_less_nfxns,0L);
   if(add_clause_to_ht(alt_hx,alt_hy,c->c_nlit) == 2L) {
      return(2L);}}
 
/*-----------------------------------------------*/
/* if there are two variables, replace each one  */
/* with a constant and calculate the hashcode    */
/*-----------------------------------------------*/

if(c->c_nv == 2L) {
   for(varbl = 0L; varbl < 2L ; varbl++) {
     for(const1 = 1L; const1 <= (long) number_constants ; const1++) {
        calc_alt_hash(l,(long) (var_array[varbl] - 1),-1L,const1,0L);
           if(add_clause_to_ht(alt_hx,alt_hy,c->c_nlit) == 2L) {
	        return(2L);}}}}
			
/*------------------------------------------------*/
/* replace every pair of variables with every     */
/* pair of constants, calculate the hashcode      */
/* and add the hashcode to the literal hash table */
/*------------------------------------------------*/ 

if(c->c_nv >= 2L) {
   for(varbl = 0L; varbl < c->c_nv - 1L; varbl++) {
      for(varbl2 = varbl + 1L ; varbl2 < c->c_nv; varbl2++) {
         for(const1=1L; const1 <= (long) number_const_hash; const1++) {
	    for(const2=1L;const2<=(long) number_const_hash; const2++) {

	       calc_alt_hash(l,varbl,varbl2,const1,const2);

	       if(add_clause_to_ht(alt_hx,alt_hy,c->c_nlit) == 2L) {
	          return(2L);}}}}}}

return(0L);
}
 
/*----------------------------------------------------------------*/
/*      calc_alt_hash(l,v1,v2,sub1,sub2)    ...............#9
/*
/*      calculate the hash code for lit l with varible v1 replaced
/*      by constant sub1 and variable v2 replaced by constant sub2.
/*
/*      called by:    calc_alternate_hashcodes()   hash.c
/*                    elim_base_clause_hc()        pass2.c
/*                    elim_same_cont_hc()          pass2.c
/*                    alt_on_hash_code_list()      pass2.c
/*
/*      calls:        nobody
/*
/*      returns:      nil
/*----------------------------------------------------------------*/
 
void calc_alt_hash(l,v1,v2,sub1,sub2)
literal *l;
long v1, v2, sub1, sub2;
{
register long a, *p, q, s1;
register rnode *r1;

p = &l->l_id;        
alt_hx = alt_hy = q = 0L;     
 
for(a = *p++ & ARGMASK; a > 0L; a--) {
   if(*p & FCNMASK) {
 
/*---------------------------------------------------------------*/
/* function found, create pointer to element                     */
/* ((*p & IAMASK)>>ARGSHIFT) + (long) rows_less_nfxns + q        */
/* in rand_array.  Advance pointer p to next term.               */
/*---------------------------------------------------------------*/
 
      r1 = &rand_array[
         ((*p & IAMASK)>>ARGSHIFT) + (long) rows_less_nfxns + q];
      a += *p++ & ARGMASK;}
      
   else if((*p & MAX_VARS_M1) == v1) {
 
/*---------------------------------------------------------------*/
/* variable v1 found, replace with constant sub1, create ptr to  */
/* element    sub1 + (long) rows_less_nfxns + q                  */
/* in rand_array.  Then advance pointer p to next term           */
/*---------------------------------------------------------------*/
 
      r1 = &rand_array[sub1 + (long) rows_less_nfxns + q];
      p++;}
      
   else if((*p & MAX_VARS_M1) == v2) {
 
/*---------------------------------------------------------------*/
/* variable v2 found, replace with constant sub2, create ptr to  */
/* element    sub1 + (long) rows_less_nfxns + q                  */
/* in rand_array.  Then advance pointer p to next term           */
/*---------------------------------------------------------------*/
 
      r1 = &rand_array[sub2 + (long) rows_less_nfxns + q];
      p++;}
      
   else  {
 
/*---------------------------------------------------------------*/
/* another variable found -- do not substitute -- create ptr to  */
/* element    (*p & MAX_VARS_M1) + q                             */
/* in rand_array.  Then advance pointer p to next term           */
/*---------------------------------------------------------------*/
 
      r1 = &rand_array[(*p & MAX_VARS_M1) + q];
      p++;}
 
/*---------------------------------------------------------------*/
/* exclusive-or random numbers found in rand_array with current  */
/* partially-calculated hash-code                                */
/*---------------------------------------------------------------*/
 
   alt_hx = alt_hx ^ r1->r_hash1;
   alt_hy = alt_hy ^ r1->r_hash2; 
   q = q + numb_rows;}

/*---------------------------------------------------------------*/
/* Add in hash code for predicate................................*/
/*---------------------------------------------------------------*/

s1 = (&rand_array[(l->l_id & IAMASK) >> ARGSHIFT])->r_hash1;
alt_hx = alt_hx ^ s1;
alt_hy = alt_hy ^ (-s1);
	
/*---------------------------------------------------------------*/
/* If complemented lit, negate hashcode                          */
/*---------------------------------------------------------------*/
 
if(l->l_id & NEGMASK) {  /* If compl lit, negate hc */
   alt_hx = -alt_hx;
   alt_hy = -alt_hy;}     
return;
}
 
/*-------------------------------------------------------------*/
/*      calc_clause_hashcode(c)        ...............#10
/*
/*      The hashcode of a cls is calculated. The hash codes for
/*      each literal are placed on a list and then ordered,
/*      then summed and multiplied to yield a hopefully random
/*      number.
/*
/*      called by:  read_in_clause()        tgtp.c
/*                  generate_clause()       gen.c
/*                  install_proof_line()    pass2.c
/*                  install_clause()        search.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*---------------------------------------------------------------*/
 
void calc_clause_hashcode(c)
clause *c;
{
register literal *l;
register long temp_lit_total, j, c_hash1, c_hash2;
register long  l_hash_tempx, l_hash_tempy;  /* used when reordering */
long l_hashx[MAX_LIT_IN_CLAUSE], l_hashy[MAX_LIT_IN_CLAUSE];
 
/*-------------------------------------------------*/
/* if c = null clause, return special hash code    */
/*-------------------------------------------------*/
 
if(c->c_nlit == 0) {
   c->c_X = 362879L; c->c_Y = 4021L; return;}
		
/*-------------------------------------------------*/
/* scan clause for literal not represented by its  */
/* hash code and put its hash value on the top     */
/* of the list of literal hash codes               */
/*-------------------------------------------------*/
 
for(l = c->c_lit; l != 0L ; l = l->l_next) {                           
   if((l->l_id & (long) HHITMASK) == 0L) {
      break;}} 
 
l_hashx[0L]     = l->l_X ;              
l_hashy[0L]     = l->l_Y ;              
l               = l->l_next;                        
temp_lit_total  = 0L;
 
/*-------------------------------------------------*/           
/* list hash values for literals in                */
/* descending order                                */
/*-------------------------------------------------*/
 
while(l != 0L) {
   if((l->l_id & (long) HHITMASK) == 0L) {
      l_hashx[temp_lit_total + 1L] = l->l_X;
      l_hashy[temp_lit_total + 1L] = l->l_Y;
      temp_lit_total++;
	
      for(j = temp_lit_total; j > 0L ; j--) {
	     if(l_hashx[j] < l_hashx[j - 1L])  {
	        break;}
	     else if(l_hashx[j]  == l_hashx[j - 1L]) {
	        if(l_hashy[j] < l_hashy[j - 1L])  {
	           break;}}
	     l_hash_tempx = l_hashx[j - 1L];
	     l_hash_tempy = l_hashy[j - 1L];
	     l_hashx[j - 1L] = l_hashx[j];
	     l_hashy[j - 1L] = l_hashy[j];
	     l_hashx[j] = l_hash_tempx;
	     l_hashy[j] = l_hash_tempy;}}
   l = l->l_next;}
 
c_hash1 = l_hashx[0L];
c_hash2 = l_hashy[0L];
 
/*-------------------------------------------------*/
/* now form c_hash1, and c_hash2 by adding and     */
/* multiplying the values on the list              */
/*-------------------------------------------------*/
 
for(j = 1L ; j <= temp_lit_total ; j++) {

   if(l_hashx[j] > 0L) {
      c_hash1  *= l_hashx[j] * (l_hashy[j] + 5039L);
      c_hash2  += l_hashx[j] +  l_hashy[j];}
   else {       
      c_hash1  *= l_hashx[j] *  (l_hashy[j] - 97L);
      c_hash2  += l_hashx[j] +  (l_hashy[j] + 5039L);}}

if(c->c_nlit != 1) {
   c_hash1 = (c_hash1 >> 16L) ^ c_hash1;}        

c->c_X = c_hash1;
c->c_Y = c_hash2;
 
return;
}
