
/*       pass2.c
/*
/*      Copyright Newborn Software 1996
/*
/*      functions:
/*
/*          functions involved in saving the proof in the
/*          cl_array
/*
/*      init_cl_array() ......................#1
/*      save_line()     ......................#2
/*      elim_base_clause_hc().................#3
/*      elim_same_cont_hc() ..................#4
/*      test_for_base_match().................#5
/*      test_for_cont_match().................#6
/*      stack_hash_codes()  ..................#7
/*      print_hash_code_list()  ..............#8
/*      print_cl_array()    ..................#9
/*      on_hash_code_list() ..................#10
/*      alt_on_hash_code_list()...............#11
/*      delete_found_hcs()....................#12
/*      move_down_renumber()..................#13
/*
/*          function that transfers the proof from the bottom of
/*          cl_array to the top
/*
/*      rearrange_cl_array()..................#14
/*
/*          functions involved in the reconstruction of the proof
/*          from information in the cl_array
/*
/*      bres_proof_line().....................#15
/*      factor_proof_line()...................#16
/*      install_proof_line()..................#17
/*      resolve_proof_line()..................#18
/*      search_proof_line() ..................#19
/*
/* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
/*
/*      The functions carried out in this file can be broken down into
/*      four categories.  They all deal with placing information in the
/*      cl_array and manipulating it so that the proof resides there
/*      as a set of four-tuples.  Each tuple contains the four
/*      arguments necessary to form a resolvent or a factor
/*
/*          1.  clears cl_array, hash_code_list
/*
/*                  init_cl_array()
/*
/*          2.  place lines of proof in cl array along with hash codes
/*              of missing literals.
/*
/*                 save_line()
/*                    delete_found_hash_codes()
/*                    elim_base_clause_hc()
/*                       test_for_base_match()
/*                    elim_same_cont_hc()
/*                       test_for_cont_match()
/*                    move_down_renumber()
/*                 stack_hash_codes()
/*                 on_hash_code_list()
/*                 alt_on_hash_code_list()
/*                 print_cl_array()
/*                 print_hash_code_list()
/*
/*          3.  when the search is done, reorganize the information
/*              in the cl array by reordering clauses and adding
/*              inferences that yielded the hash codes of the missing
/*              literals.
/*
/*                 rearrange_cl_array()
/*
/*          4.  Reconstruct proof by carrying out the inferences
/*              as indicated by the information in the cl array.
/*              each row of the cl array correponds to one inference.
/*
/*                 search_proof_line()
/*                    resolve_proof_line()
/*                       factor_proof_line()
/*                          clear_var_substitution_tble()
/*                          unify() 
/*                          generate_clause()
/*                          install_proof_line()
/*                       bres_proof_line()
/*                          standardize()               
/*                          clear_var_substitution_tble()
/*                          unify()
/*                          generate_clause()
/*                          install_proof_line()
/*                          print_cl_array()      
/*                          print_clause()       
/*
/*-------------------------------------------------------------*/
 
#include <stdio.h>
 
#include "const.h"
#include "clause.h"

/*-------------------------------------------------------------*/
/* externally defined functions                                */
/*-------------------------------------------------------------*/
 
extern fatal();                            /* tgtp.c           */

extern print_clause();                     /* dump.c           */
extern print0();

extern install_clause();                   /* search.c         */

extern clause *generate_clause();          /* gen.c            */
extern int test_equality();

extern standardize();                      /* infer.c          */
extern clear_var_substitution_tble();

extern unify();                            /* unify.c          */
 
extern calc_clause_hashcode();             /* hash.c           */
extern calc_alt_hash();
extern normalize_literal();

/*-------------------------------------------------------------*/
/* externally define constants                                 */
/*-------------------------------------------------------------*/
 
extern FILE *outfile;
 
extern long *clause_space[];
extern clause *clist[];
extern dnode depth[];
 
extern long ply, base, org_base, nc, nextc;
extern long s_nod, iter_depth;
extern long magic_hashx, magic_hashy;
extern long alt_hx, alt_hy;
extern long equal_identity, equal_pred;
extern int o_screen, o_print;
extern int o_debug, big_bug;
extern int extra_pass, phase;
extern int o_maxterm;
extern int o_noprint;
extern int number_const_hash;
extern int number_constants;
extern int rows_less_nfxns;
extern int var_array[64], var_array_ptr;
 
/*-------------------------------------------------------------*/
/* internally defined functions                                */
/*-------------------------------------------------------------*/

void init_cl_array();
void save_line();
int elim_base_clause_hc();
int elim_same_cont_hc();
int test_for_base_match();
int test_for_cont_match();
void stack_hash_codes();
void print_hash_codes();
void print_cl_array();
int on_hash_code_list();
int alt_on_hash_code_list();
void delete_found_hcs();
void delete_alt_hcs();
void move_down_renumber();
void rearrange_cl_array();
void bres_proof_line();
void factor_proof_line();
void install_proof_line();
void resolve_proof_line();
void search_proof_line();

 
/*------------------------------------------------------*/
/* the hash code list saves  hash values of literals    */
/* for which the complement must be found in pass2.     */
/* hash_code_list_active is a 32-bit vector the i-th bit*/
/* of which is a 0 or 1 to indicate whether the i-th    */
/* element on hash_code_list is still to be found       */
/*------------------------------------------------------*/
 
long hash_code_listx[32], hash_code_listy[32];
long hash_code_list_top, hash_code_list_active; 
long current_proof_top, proof_length;
long killer_hashx, killer_hashy, killer_hash_ptr;
 
/*------------------------------------------------------*/
/* cl[PROOF_ARRAY_SIZE][3] stores info on resolut's and */
/* factors that must be performed to obtain the proof.  */
/* a proof can be at most PROOF_ARRAY_SIZE long.        */
/*------------------------------------------------------*/
 
long cl[PROOF_ARRAY_SIZE][3];
int zz;
 
/*---------------------------------------------------------------*/
/*      init_cl_array()                    ................#1
/*
/*      initialize several variables and clears the cl array,
/*      and hash_code_list.  executed before search begins.
/*
/*      called by:      main()              tgtp.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void init_cl_array()
{       
register long i,j;
	
hash_code_list_top    = 0L; /* counter of lits repby hash vals */
hash_code_list_active = 0L;
proof_length          = 0L;
current_proof_top     = 0L;
	
for(i = 0L; i < PROOF_ARRAY_SIZE ; i++) {
   for(j = 0L; j < 3L ; j++)   {
      cl[i][j] = 0L;}}
for(i = 0L; i < 32L  ; i++) {
   hash_code_listx[i] = hash_code_listy[i] = 0L;}
return;
}
 
/*---------------------------------------------------------------*/
/*      save_line()                       ................#2
/*
/*      this function is called to save the path of inferences
/*      that led to the contradiation and to each unit clause
/*      whose hash code was previously saved on hash_code_list.
/*      Inferences are stored in the cl-array.
/*
/*      called by:  main()                  tgtp.c
/*                  install_clause()        search.c
/*
/*      calls:      print_cl_array()        pass2.c
/*                  print_clause()          tdump.c
/*                  delete_found_hcs()      pass2.c
/*                  elim_base_clause_hc()   pass2.c
/*                  elim_same_cont_hc()     pass2.c
/*                  move_down_renumber()    pass2.c
/*
/*      returns:    nil
/*---------------------------------------------------------------*/
 
void save_line(c_killer)
clause *c_killer;
 
/*----------------------------------------------*/
/* c_killer (killer clause) is 0L in phase 1    */
/* and is a pointer to the unit clause ending   */
/* the path which is being saved in phase 2     */
/*----------------------------------------------*/
 
{
register clause *c;
literal *l;
register long i, j, k, l_index, cid, number_of_inserts;
long top_ptr, line_length;
long clpar, cl2par, clindex, crindex, ply1, killer, cr;
long hxtemp,hytemp;
	
/*---------------------------------------------------------------*/
/*
/*      cl array structure
/*
/*      cl[i][0]......bits 28-32 ptr to element on hash_code_list
/*                    that corresponds to the hashcode of the
/*                    missing literal
/*                    bits 16-26 the index of the clause
/*                    bits 05-15 the index of the left parent clause
/*                    bits 00-04 the index of the left literal
/*      cl[i][1]......bits 05-16 the index of the rght parent clause
/*                    bits 00-04 the index of the right literal
/*           ....or   bits 00-31 the l->l_X hash code if literal
/*                    is missing
/*      cl[i][2]......0 if no literal is missing from clause
/*                    l->l_Y if literal is missing
/*
/*--------------------------------------------------------------*/
				
/*--------------------------------------*/
/* print a "." to indicate a line is    */
/* being saved                          */
/*--------------------------------------*/
 
print0(".");
 
/*--------------------------------------*/
/* print line if debug option is on     */
/*--------------------------------------*/
 
if(o_debug > 0) {
   printf(
   "\n\nSAVE LINE PLY:%ld NODES:%ld ITER:%ld \n",
                                ply,s_nod,iter_depth);
   fprintf(outfile,
   "\n\nSAVE LINE PLY:%ld NODES:%ld ITER:%ld \n",
                                ply,s_nod,iter_depth);
 
   if(proof_length == 0L) {
      for(i = 1L; i <= ply; i++)  {
	       print_clause(clist[depth[i].d_tc]);}}
   else {
      for(i = 1L; i < ply; i++)  {
	       print_clause(clist[depth[i].d_tc]);}
      print_clause(c_killer);}} 
 
/*----------------------------------------*/
/* scan main line and hash_code_list for  */
/* missing clauses whose hash codes       */
/* have already been found                */
/*----------------------------------------*/
 
delete_found_hcs(killer_hashx,killer_hashy);             

/*----------------------------------------*/
/* delete variants of clause              */
/*----------------------------------------*/

if(phase == 2) {
   delete_alt_hcs(c_killer);}

top_ptr = 0L;
number_of_inserts = 0L;
 
ply1 = ply;
if(proof_length != 0L) {
   ply1--;}
 
for(i = 1L ; i <= ply1 ; i++) {
 
/*--------------------------------------*/
/* transform clauses in clist to rows   */      
/* in the cl-array                      */
/*--------------------------------------*/
	
   c = clist[depth[i].d_tc];
   cid = c->c_id + number_of_inserts;
 
   clpar = c->c_indices >> LEFT_INDEX_SHIFT;
   cl[top_ptr][0] = (c->c_id << 16L) + (c->c_lpar->c_id << 4L) + clpar;
   if(c->c_rpar != 0L) {
      cl[top_ptr][1] = 
          (c->c_rpar->c_id << 4L) + (c->c_indices & RIGHT_INDEX_MASK);}
   else {
      cl[top_ptr][1] = c->c_indices & RIGHT_INDEX_MASK;}
      
   cl[top_ptr][2] = 0L;
   top_ptr++;

/*------------------------------------------*/
/* now take care of each missing literal    */
/* add a row to the cl array for each       */
/* literal in clause c that is represented  */
/* by its hash code.                        */
/*------------------------------------------*/
 
   l_index = 0L;
   for(l = c->c_lit; l != 0L ; l = l->l_next) {
      l_index++;
      if((l->l_id & HHITMASK) != 0L) {  
	     number_of_inserts++;
	     cid = (cl[top_ptr - 1L][0] >> 16L);            
			
	     clpar = l_index;               
	     for(j = top_ptr - 1L; cl[j][2] != 0L; j--) {
	        cl2par = (cl[j][0] & INDEX_MASK);
	        if(cl2par < clpar) {
	           clpar--;}}
			
	     cl[top_ptr][0] = ((cid + 1L)<< 16L) + (cid<< 4L) + clpar;

	     cl[top_ptr][1] = hxtemp = -l->l_X;
	     cl[top_ptr][2] = hytemp = -l->l_Y;
	     if(elim_base_clause_hc(top_ptr,l) == 0) {
	        if(elim_same_cont_hc(top_ptr,i) == 0) {
		   normalize_literal(l,c); 
		   calc_lit_hc(l,c);        
	           cl[top_ptr][1] = -l->l_X;
	           cl[top_ptr][2] = -l->l_Y;
	 
	           if(elim_base_clause_hc(top_ptr,l) == 0) {
	              elim_same_cont_hc(top_ptr,i);}}}
             if(cl[top_ptr][2] != -1L) {
                cl[top_ptr][1] = hxtemp;
	        cl[top_ptr][2] = hytemp;}
	     top_ptr++;}}}
 
/*----------------------------------------------*/
/* for secondary lines, add unit clause         */
/*----------------------------------------------*/

if(phase == 2) {
   c = c_killer;
   cid = c->c_id + number_of_inserts;
	
   clpar = c->c_indices >> LEFT_INDEX_SHIFT;
   cl[top_ptr][0] = (c->c_id << 16L) + (c->c_lpar->c_id << 4L) + clpar;
   if(c->c_rpar != 0L) {
      cl[top_ptr][1] = 
         (c->c_rpar->c_id << 4L) + (c->c_indices & RIGHT_INDEX_MASK);}
   else {
      cl[top_ptr][1] = c->c_indices & RIGHT_INDEX_MASK;}
   cl[top_ptr][2] = 0L;
   top_ptr++;
 
/*------------------------------------------*/
/* now take care of missing literals from   */
/* killer clause of secondary line          */
/*------------------------------------------*/
 
   l_index = 0L;
   for(l = c->c_lit; l != 0L ; l = l->l_next) {
      l_index++;
      if((l->l_id & HHITMASK) != 0L) {  
	     number_of_inserts++;
	     cid = (cl[top_ptr - 1L][0] >> 16L);
	     clpar = l_index;
	     if(top_ptr != 0L) {
	        for(j = top_ptr - 1L; cl[j][2] != 0L; j--) {
	           cl2par = (cl[j][0] & INDEX_MASK);
	           if(cl2par < clpar) {
	              clpar--;}}}
	     cl[top_ptr][0] = ((cid + 1L)<< 16L) + (cid<< 4L) + clpar;
 
	     cl[top_ptr][1] = hxtemp = -l->l_X;
	     cl[top_ptr][2] = hytemp = -l->l_Y;
	     if(elim_base_clause_hc(top_ptr,l) == 0) {
	        if(elim_same_cont_hc(top_ptr,i) == 0) {
	           normalize_literal(l,c); 
		   calc_lit_hc(l,c);        
	           cl[top_ptr][1] = -l->l_X;
	           cl[top_ptr][2] = -l->l_Y;
	           if(elim_base_clause_hc(top_ptr,l) == 0) {
	              elim_same_cont_hc(top_ptr,i);}}}

             if(cl[top_ptr][2] != -1L) {
                cl[top_ptr][1] = hxtemp;
	        cl[top_ptr][2] = hytemp;}
	     top_ptr++;}}
   cl[top_ptr - 1L][0] += (killer_hash_ptr << 26L);}
 
/*--------------------------------------------------*/
/* if main line ends when complement found in hash  */
/* table, add extra line for this magic resolvent   */
/*--------------------------------------------------*/

if(proof_length == 0L) {
   c = clist[depth[ply].d_tc];
   if(c->c_nlit == 1) {
      if(proof_length == 0L && magic_hashx != 0L) {
         cl[top_ptr][0] = ((cl[top_ptr - 1L][0] >> 16L) + 1L) <<16L; 
         cl[top_ptr][0]=(i<<26L)+cl[top_ptr][0]+(cl[top_ptr][0]<<4L)+1;
         cl[top_ptr][1] = -magic_hashx;
         cl[top_ptr][2] = -magic_hashy;
         if(elim_base_clause_hc(top_ptr,c->c_lit) == 0) {
	    elim_same_cont_hc(top_ptr,i);}              
         top_ptr++;}}}

line_length = top_ptr;
 
/*---------------------------------------------------*/
/* renumber parents; first right then left           */
/*---------------------------------------------------*/

for(i = 1L ; i < line_length ; i++) {
 
/*----------------------------------*/
/* renumber right parents           */
/*----------------------------------*/
 
   if(cl[i][2] == 0L) {
      cr = (cl[i][1] >> 4L) & PARENT_MASK;
      crindex = cl[i][1] & INDEX_MASK;
      if(cr > base + 1L) {

         k = base + 1L; 
         for(j = base + 1L;  ; j++) {
            if(cl[j - base - 1L][2] == 0L) {
               k++; 
               if(k > cr) { 
                   break;}}}

	     cl[i][1] = (j << 4L) + crindex;}} 

/*----------------------------------*/
/* renumber left parents            */
/*----------------------------------*/
 
   killer = (cl[i][0] >> 26L);
   cid = ((cl[i-1][0]) >> 16L) + 1L;
   clpar = cid - 1L;
   clindex = (cl[i][0] & INDEX_MASK);
   cl[i][0] = (killer<<26L) + (cid<<16L) + (clpar<<4L) + clindex;}
 
/*----------------------------------*/
/* unmark hashes to base clauses    */
/*----------------------------------*/

for(i = 1L ; i < line_length ; i++) {
   if(cl[i][2] == -1L) {
      cl[i][2] = 0L;}}
 
proof_length += line_length;

current_proof_top = PROOF_ARRAY_SIZE_M1 - proof_length;
if(o_debug > 0) { 
   print_cl_array(0L);}
move_down_renumber(line_length);
if(o_debug > 0) {
   print_cl_array(PROOF_ARRAY_SIZE - proof_length);}

return;
}
 
/*---------------------------------------------------------------*/
/*      elim_base_clause_hc()              ................#3
/*
/*      fix up cl entries which have hash codes to base clauses
/*
/*      parameters: top_ptr = row in cl_array with missing literal
/*
/*      called by:  save_line()                     pass2.c
/*
/*      calls:      test_for_base_match()           pass2.c
/*                  calc_alt_hash()                 hash.c
/*
/*      returns:    0 => no match; 1 => match
/*---------------------------------------------------------------*/
 
int elim_base_clause_hc(top_ptr,ll)
long top_ptr;
literal *ll;
{
clause *c;
literal *l;
long i, k1, k2, varbl, varbl2;


if(equal_pred != 0L && (ll->l_id & IAMASK) == equal_pred) {
   if(test_equality(ll)) {
      cl[top_ptr][1] = ((equal_identity + 1L) << 4L) + 1L;
      cl[top_ptr][2] = -1L;
      return(1);}}

for(i = 0L ; i <= base ; i++) {
   c = clist[i];
   if(c->c_nlit == 1) { 
      l = c->c_lit;
      normalize_literal(l,c); 
/*----------------------------------------------*/
/* test hash code for match with base clause    */
/*----------------------------------------------*/
 
      if(test_for_base_match(i,l->l_X,l->l_Y,top_ptr)) {
	     return(1);}
 
      if(c->c_nv == 1L)  {
	     for(k1 = 1L; k1 <= (long) number_constants ; k1++) {
	        calc_alt_hash(l,0L,-1L,k1,0L);
	        if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
	           return(1);}}
 
	     calc_alt_hash(l,0L,-1L,1L-(long)rows_less_nfxns,0L);
	     if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
	        return(1);}}
 
/*----------------------------------------------*/
/* if base clause is not a ground instance      */
/* check alternate hash codes                   */
/*----------------------------------------------*/
 
      if(c->c_nv == 2L) {
	     for(varbl = 0L; varbl < 2L ; varbl++) {
	        for(k1 = 1L; k1 <= (long) number_constants; k1++) {
	           calc_alt_hash(l,(long)(var_array[varbl]-1),-1L,k1,0L);
	           if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
		           return(1);}}}}
      if(c->c_nv >= 2L) {
	     for(varbl = 0L; varbl < c->c_nv - 1L; varbl++) {
	        for(varbl2 = varbl + 1L ; varbl2 < c->c_nv; varbl2++) {
	           for(k1 = 1L; k1 <= (long) number_const_hash; k1++) {
		          for(k2 = 1L; k2 <= (long) number_const_hash; k2++) {
		             calc_alt_hash(l,varbl,varbl2,k1,k2);
		             if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
			            return(1);}}}}}}}}
return(0);
}
 
/*---------------------------------------------------------------*/
/*      elim_same_cont_hc()               ................#4
/*
/*      fix up cl entries which have hash codes to clauses on same
/*      continuation
/*
/*      parameters: top_ptr = row in cl_array with missing literal
/*                  pli     = length of current continuation - 1
/*
/*      called by:  save_line()                     pass2.c
/*
/*      calls:      test_for_cont_match()           pass2.c
/*                  calc_alt_hash()                 hash.c
/*
/*      returns:    0 => not elim; 1 => eliminated 
/*------------------------------------------------------------------*/
 
int elim_same_cont_hc(top_ptr,pli)
long top_ptr, pli;
{
clause *c;
literal *l;
long i, k1, k2, varbl, varbl2;
 
for(i = 1L ; i < pli ; i++) { 
   c = clist[depth[i].d_tc]; 
   if(c->c_nlit == 1) {
      for(l = c->c_lit; (l->l_id & HHITMASK) != 0L; l = l->l_next) { };
			
/*--------------------------------------------------*/
/* test hash code for match with contuation clause  */
/*--------------------------------------------------*/

      normalize_literal(l,c); 
      if(test_for_cont_match(i,l->l_X,l->l_Y,top_ptr)) {
         return(1);}

/*----------------------------------------------*/
/* if base clause is not a ground instance      */
/* check alternate hash codes                   */
/*----------------------------------------------*/
 
      if(c->c_nv == 1L) {
         for(k1=1L; k1 <= (long) number_constants; k1++) {
	           calc_alt_hash(l,0L,-1L,k1,0L);  
	           if(test_for_cont_match(i,alt_hx,alt_hy,top_ptr)) {
	              return(1);}}
 
            calc_alt_hash(l,0L,-1L,1L-(long)rows_less_nfxns,0L);  
            if(test_for_base_match(i,alt_hx,alt_hy,top_ptr)) {
	           return(1);}}
 
         if(c->c_nv == 2L) {
            for(varbl = 0L; varbl < 2L ; varbl++) {
	           for(k1 = 1L; k1 <= (long) number_constants ; k1++) {
	              calc_alt_hash(l,(long)(var_array[varbl]-1),-1L,k1,0L);
	              if(test_for_cont_match(i,alt_hx,alt_hy,top_ptr)) {
		             return(1);}}}}
       
         if(c->c_nv >= 2L) {  
            for(varbl = 0L; varbl < c->c_nv - 1L; varbl++) {
	           for(varbl2 = varbl + 1L ; varbl2 < c->c_nv; varbl2++) {
	              for(k1 = 1L; k1 <= (long) number_const_hash ; k1++) {
	                 for(k2 = 1L; k2 <= (long) number_const_hash; k2++) {
	                    calc_alt_hash(l,varbl,varbl2,k1,k2);
	                    if(test_for_cont_match(i,alt_hx,alt_hy,top_ptr)){
		                   return(1);}}}}}}}}
return(0);
}
 
/*-------------------------------------------------------------*/
/*      test_for_base_match()             ................#5
/*
/*      parameters: c_index = the index of the base clause
/*                              under comparison
/*                  hc1,hc2 = hashcodes of missing literal
/*                  top_ptr = row in cl_array with missing lit
/*
/*      called by:  elim_base_clause_hc()             pass2.c
/*
/*      calls:      nobody
/*
/*      returns:    0 if no match found, 1 if match found
/*-------------------------------------------------------------*/
 
int test_for_base_match(c_index,hc1,hc2,top_ptr)
long c_index, hc1, hc2, top_ptr;
{
if(cl[top_ptr][1] == hc1 && cl[top_ptr][2] == hc2) {
   cl[top_ptr][1] = ((c_index +1L) << 4L) + 1L;
   cl[top_ptr][2] = -1L;
   return(1);}
return(0);
}
 
/*-------------------------------------------------------------*/
/*      test_for_cont_match()              ................#6
/*
/*      called by:  elim_cont_clause_hc()          pass2.c
/*
/*      calls:      nobody
/*
/*      returns:    0 => no match; 1 => match
/*-------------------------------------------------------------*/
 
int test_for_cont_match(c_index,hc1,hc2,top_ptr)
long c_index, hc1, hc2, top_ptr;
{
clause *c;
literal *l;
long additional_rows, j;

if(cl[top_ptr][1] == hc1 && cl[top_ptr][2] == hc2) {
   additional_rows = 0L;
   for(j = 1L; j <= c_index ; j++) {
      c = clist[depth[j].d_tc];
      for(l = c->c_lit ; l != 0L ; l = l->l_next) {
	     if((l->l_id & HHITMASK) != 0L) {
	        additional_rows++;}}}
   c_index = c_index + additional_rows; 
   cl[top_ptr][1] = ((c_index + base + 1L) << 4L) + 1L;
   cl[top_ptr][2] = -1L;
   return(1);}
return(0);
}
 
/*-------------------------------------------------------------*/
/*      stack_hash_codes()            ....................#7
/*
/*      put hash codes of lits that must be found on next pass
/*      on hash_code_list
/*
/*      called by:   main()          tgtp.c
/*
/*      calls:       nobody
/*
/*      returns:     nil
/*-------------------------------------------------------------*/
 
void stack_hash_codes()
{
register long i, j;

/*-------------------------------------------------------------*/
/* if hash code for missing lit in clause, stack it for        */
/* latter discovery. do not stack it if it is already on       */
/* hash_code_list                                              */
/*-------------------------------------------------------------*/
								
for(i = current_proof_top ; i < PROOF_ARRAY_SIZE ; i++) {
   if(cl[i][2] != 0L)  {                
	
/*----------------------------------------------------------*/
/* literal in this clause is represented by its hash code   */
/*----------------------------------------------------------*/
      for(j = 0L; j <= hash_code_list_top ; j++)  { 
 
/*----------------------------------------------------------*/
/* if already on hash_code_list, do not add a second time   */
/*----------------------------------------------------------*/
 
	     if(cl[i][1] == hash_code_listx[j]) {
	        if(cl[i][2] == hash_code_listy[j]) {
		   if((hash_code_list_active & (1L << j)) == 0L) {
		      hash_code_list_active += (1L << j);}
	           break;}}
	     if(hash_code_listx[j] == 0L) {
			
/*----------------------------------------------------------*/
/* add hash code of literal at end of hash_code_list, inc   */
/* the top of hash_code_list and fix up the j-th bit of the */
/* hash_code_list_active word                               */
/*----------------------------------------------------------*/
 
	        hash_code_listx[j] = cl[i][1];
	        hash_code_listy[j] = cl[i][2];
	        hash_code_list_top++; 
	        hash_code_list_active += (1L << j);
	        break;}}}}
return;
}
 
/*-------------------------------------------------------------*/
/*      print_hash_code_list()           ................#8
/*
/*      used as a debugging tool; prints the contents of the
/*      missing literal hash_code_list
/*
/*      called by:      print_cl_array()             pass2.c
/*                      interrupt_options()          tdump.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*-------------------------------------------------------------*/
 
void print_hash_code_list()
{
long i;
 
print0("HASH CODE LIST");
 
for(i = 0L; i < hash_code_list_top ; i++) {
   if((i % 3L) == 0L) { print0("\n");}
   printf("H[%ld]:%ld %ld  ",
	     i,hash_code_listx[i],hash_code_listy[i]);
   fprintf(outfile,"H[%ld]:%ld %ld  ",
	     i,hash_code_listx[i],hash_code_listy[i]);}
 
printf("\nAlive_mask: %ld \n?",hash_code_list_active);
fprintf(outfile,"\nAlive_mask: %ld ",hash_code_list_active);
}                       
		
 
/*-------------------------------------------------------------*/
/*      print_cl_array()                ................#9
/*
/*      called when debugging program
/*
/*      called by:  save_line()                   pass2.c
/*                  fix_inferences_in_cl_array()  pass2.c
/*                  interrupt_options()           tdump.c
/*
/*      calls:      print_hash_code_list()        pass2.c
/*
/*      returns:    nil
/*-------------------------------------------------------------*/
 
void print_cl_array(top_of_proof)
long top_of_proof;
{       
long i;
 
print0("\nCL_ARRAY");
		
for(i = top_of_proof; i < PROOF_ARRAY_SIZE ; i++) {
   if(cl[i][0] == 0L) {break;}
   if(cl[i][2] == 0L) {         
      printf(
      "\n%ld L %ld,%ld R %ld,%ld  HC %ld  ",
      (cl[i][0] >> 16) & CIDMASK2,                 /* block  */
      (cl[i][0] >> 4) & PARENT_MASK,              /* l_par   */
      (cl[i][0] & INDEX_MASK),                    /* index1  */
      ((cl[i][1] >> 4) & PARENT_MASK),
      (cl[i][1] & INDEX_MASK),
      (cl[i][0] >> 26));
      fprintf(outfile,
      "\n%ld L %ld,%ld R %ld,%ld  HC %ld  ",
      (cl[i][0] >> 16) & CIDMASK2,                   /* block */
      (cl[i][0] >> 4) & PARENT_MASK,                 /* l_par */
      (cl[i][0] & INDEX_MASK),                      /* index1 */
      ((cl[i][1] >> 4) & PARENT_MASK),
      (cl[i][1] & INDEX_MASK),
      (cl[i][0] >> 26));}
 
   else if(cl[i][2] != 0L && phase != 3) {              
      printf(
      "\n%ld L %ld,%ld H %ld %ld  HC %ld  ",
      (cl[i][0] >> 16) & CIDMASK2,                   /* block  */
      (cl[i][0] >> 4) & PARENT_MASK,                /* l_par   */
      (cl[i][0] & INDEX_MASK),                      /* index1  */
      cl[i][1],
      cl[i][2],
      (cl[i][0] >> 26));
			
      fprintf(outfile,
      "\n%ld L %ld,%ld H %ld,%ld  HC %ld  ",
      (cl[i][0] >> 16) & CIDMASK2,                   /* block   */
      (cl[i][0] >> 4) & PARENT_MASK,                 /* l_par   */
      (cl[i][0] & INDEX_MASK),                       /* index1  */
      cl[i][1],
      cl[i][2],
      (cl[i][0] >> 26));}
			
   else {
      printf(
      "\n%ld L %ld,%ld R %ld,%ld HC %ld  INDEX: %ld  ",
      (cl[i][0] >> 16) & CIDMASK2,                    /* block   */
      (cl[i][0] >> 4) & PARENT_MASK,                  /* l_par   */
      (cl[i][0] & INDEX_MASK),                        /* index1  */
      ((cl[i][1] >> 4) & PARENT_MASK),
      (cl[i][1] & INDEX_MASK),
      (cl[i][0] >> 26),
      cl[i][2]);
 
      fprintf(outfile,
      "\n%ld L %ld,%ld R %ld,%ld HC %ld  INDEX: %ld  ",
      (cl[i][0] >> 16) & CIDMASK2,                   /* block     */
      (cl[i][0] >> 4) & PARENT_MASK,                 /* l_par     */
      (cl[i][0] & INDEX_MASK),                       /* index1    */
      ((cl[i][1] >> 4) & PARENT_MASK),
      (cl[i][1] & INDEX_MASK),
      (cl[i][0] >> 26),
      cl[i][2]);}}
 
print0("\n");
print_hash_code_list(); 
return;         
}
 
/*--------------------------------------------------------------*/
/*      on_hash_code_list(hcx,hcy)         ................#10
/*
/*      see if complement of hcx,hcy is on hash_code_list
/*
/*      called by:   install_clause()        search.c
/*
/*      calls:       nobody
/*
/*      returns:     0 if hash code not on hash_code_list
/*                   1 if hash code on hash_code_list
/*-------------------------------------------------------------*/
 
int on_hash_code_list(hcx,hcy)
long hcx, hcy;
{
register long i;
 
for( i = 0L ; i <= hash_code_list_top ; i++) {
   if(hcx == hash_code_listx[i] && hcy == hash_code_listy[i]) {
      if((hash_code_list_active & (1L << i)) != 0L) {
			
/*----------------------------------------------------------*/
/*  if found on hash_code_list, delete by setting bit on    */
/*  hash_code_list_active to 0.                             */
/*----------------------------------------------------------*/
		
	     hash_code_list_active = hash_code_list_active ^ (1L << i);
	     killer_hashx = hcx; 
	     killer_hashy = hcy;
	     killer_hash_ptr = i + 1L;
	     return(1);}}
   if(hash_code_list_active == 0L) { 
      return(0);}}
return(0);
}
 
/*---------------------------------------------------------------*/
/*      alt_on_hash_code_list()    .......................#11
/*
/*      called by:   install_clause()            search.c
/*
/*      calls:       on_hash_code_list()         pass2.c
/*                   calc_alt_hash()             hash.c
/*
/*      returns:     0 if hash code not on hash_code_list
/*                   2 if hash code on hash_code_list
/*---------------------------------------------------------------*/
 
int alt_on_hash_code_list(l,c)
literal *l;
clause *c;
{
long k1, k2, varbl, varbl2;
int found;
 
found = 0;

if(c->c_nv == 1L) { 
   for(k1 = 1L ; k1 <= (long) number_constants ; k1++) {
      calc_alt_hash(l,0L,-1L,k1,0L);            
      if(on_hash_code_list(alt_hx,alt_hy)) {
         found = 2;}}
 
   calc_alt_hash(l,0L,-1L,1L-(long) rows_less_nfxns,0L);     
   if(on_hash_code_list(alt_hx,alt_hy)) {
      found = 2;}}


if(c->c_nv == 2L) {
   for(varbl = 0L; varbl < 2L ; varbl++) {
      for(k1 = 1L; k1 <= (long) number_constants ; k1++) {
	     calc_alt_hash(l,(long) (var_array[varbl] - 1),-1L,k1,0L);
	     if(on_hash_code_list(alt_hx,alt_hy)) {
	        found = 2;}}}}

if(c->c_nv  >= 2L) {
   for(varbl = 0L; varbl < c->c_nv - 1L; varbl++) {
      for(varbl2 = varbl + 1L ; varbl2 < c->c_nv; varbl2++) {
	     for(k1 = 1L; k1 <= (long) number_const_hash ; k1++) {
	        for(k2 = 1L; k2 <= (long) number_const_hash; k2++) {
	           calc_alt_hash(l,varbl,varbl2,k1,k2);
	           if(on_hash_code_list(alt_hx,alt_hy)) {
		       found = 2;}}}}}}

return(found);
}
 
/*---------------------------------------------------------------*/
/*      delete_found_hcs()               ................#12
/*
/*      deletes hash codes from the line just found for which the
/*      complement's hash code has been previously found and placed
/*      in the hash_array
/*
/*      called by:   save_line()          pass2.c
/*
/*      calls:       nobody
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void delete_found_hcs(hx,hy)
long hx,hy;
{
long m;

for(m = current_proof_top + 1L ; m < PROOF_ARRAY_SIZE ; m++) {
   if(cl[m][1] == hx && cl[m][2] == hy) {
      cl[m][2] = 0L;
      cl[m][1] = (current_proof_top << 4L) + 1L;}}

return;
}
 
/*---------------------------------------------------------------*/
/*      move_down_renumber()                ................#13
/*
/*      adds the continuation stored in the top rows of cl_array to
/*      those in the bottom where the previously-found parts of
/*      the proof have been stored.
/*
/*      called by:      save_line()                pass2.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void move_down_renumber(path_length)
long path_length;
{
long cid, clpar, clindex, crpar, crindex, j, m;
long t_ptr, b_ptr, diff, hash_pointer, old_bottom_ptr;
 
b_ptr = PROOF_ARRAY_SIZE - proof_length;
old_bottom_ptr = b_ptr + path_length;
 
for(t_ptr = 0L ; t_ptr < path_length ; t_ptr++) {
   if(cl[t_ptr][2] != 0L) {
      for(j = 0L; j <= hash_code_list_top; j++) {
	     if(cl[t_ptr][1] == hash_code_listx[j] && 
			   cl[t_ptr][2] == hash_code_listy[j] &&
			   (hash_code_list_active & (1L << j)) == 0L) {
	        for(m = old_bottom_ptr ; m < PROOF_ARRAY_SIZE ; m++) {
	           hash_pointer = (cl[m][0] >> 26L);
	           if(hash_pointer == j + 1L) {
		          cl[t_ptr][1] = (m << 4L) + 1L;
		          cl[t_ptr][2] = 0L;
              break;}}}}}}

for(t_ptr = 0L ; t_ptr < path_length ; t_ptr++) {
   hash_pointer = cl[t_ptr][0] >> 26L;
   cid = b_ptr;
   clpar = (cl[t_ptr][0] >> 4L) & PARENT_MASK;
   if(clpar > base + 1L) {
      diff = base + 2L + t_ptr - clpar;
      clpar = b_ptr - diff;}
   clindex = (cl[t_ptr][0] & INDEX_MASK);
	
   if(cl[t_ptr][2] == 0L) {
      crpar = (cl[t_ptr][1] >> 4L) & PARENT_MASK;
      if(crpar > base + 1L && crpar<PROOF_ARRAY_SIZE - proof_length) {
	     diff = base + 2L + t_ptr - crpar;
	     crpar = b_ptr - diff;}
      crindex = (cl[t_ptr][1] & INDEX_MASK);
		
      cl[b_ptr][0] = 
	   (hash_pointer << 26L) + (cid << 16L) + (clpar << 4L) + clindex;
      cl[b_ptr][1] = (crpar << 4L) + crindex;
      cl[b_ptr][2] = 0L;}
	
   else {
      cl[b_ptr][0] = 
	    (hash_pointer << 26L) + (cid << 16L) + (clpar << 4L) + clindex;
      cl[b_ptr][1] = cl[t_ptr][1];
      cl[b_ptr][2] = cl[t_ptr][2];}
	
   cl[t_ptr][0] = cl[t_ptr][1] = cl[t_ptr][2] = 0L;
	
   b_ptr++;}
		
for(m = b_ptr; m < PROOF_ARRAY_SIZE ; m++) {
   if(cl[m][2] == -1L) {
      cl[m][1] = ((old_bottom_ptr - 1L) << 4L) + 1L;
      cl[m][2] = 0L;}}
return;
}
 
/*-------------------------------------------------------------*/
/*      rearrange_cl_array()            ................#14
/*
/*      rearranges proof in the top rows of the cl_array when this
/*      function is entered, the proof is in the bottom of cl_array
/*      (rows from "PROOF_ARRAY_SIZE - proof_length" 
/*      to PROOF_ARRAY_SIZE).
/*
/*      called by:     main()           tgtp.c
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*-------------------------------------------------------------*/
 
void rearrange_cl_array()
{
long clause_id, m;
long start_index, bottom_of_block, j, k;
long not_ready, rpar, rpar_gone;
long cid, clpar, clindex, crpar, crindex;
long smallest, smallest_ptr, c0;
 
/*--------------------------------------*/
/* determine order of blocks            */
/*--------------------------------------*/
 
clause_id = 1L;

while(clause_id <= proof_length) {
   start_index = PROOF_ARRAY_SIZE - proof_length;
   get_next_blocks:
   for(m = start_index ; m < PROOF_ARRAY_SIZE ; m++) {
      if((cl[m][0] >> 26L) != 0L || m == PROOF_ARRAY_SIZE_M1) {
         bottom_of_block = m;
         break;}}
   not_ready = 0L;
   for(j = start_index; j <= bottom_of_block; j++) {
      if(cl[j][2] != 0L) {not_ready++;}
      if((cl[j][1]>>4L) > base+1L && ((cl[j][1]>>4L) < start_index 
			|| (cl[j][1]>>4L) > bottom_of_block)) {
         rpar = cl[j][1] >> 4L;
         rpar_gone = cl[rpar][2];
         if(rpar_gone == 0L) {
            not_ready++;}}} 
   if(not_ready == 0L) { 
      for(j = start_index; j <= bottom_of_block; j++) {
	     cl[j][2] = clause_id++;}}
   start_index = m + 1L;
   if(m < PROOF_ARRAY_SIZE) {
      goto get_next_blocks;}}
	
/*----------------------------------*/
/* now move up to top of array      */
/*----------------------------------*/
k = 0L;
for(j = PROOF_ARRAY_SIZE - proof_length ; j < PROOF_ARRAY_SIZE ; j++) {
   cid = cl[j][2] + base + 1L;
   clpar = (cl[j][0] >> 4L) & PARENT_MASK;
   if(clpar > base + 1L) {
      clpar = cid - 1L;}
   clindex = cl[j][0] & INDEX_MASK;
   crpar = (cl[j][1] >> 4L) & PARENT_MASK;
   if(crpar > base + 1L) {
      crpar =  cl[crpar][2] + base + 1L;}
   crindex = cl[j][1] & INDEX_MASK;
	
   cl[j][0] = (cid << 16L) + (clpar << 4L) + clindex;
   cl[j][1] = (crpar << 4L) + crindex;}
for(j=PROOF_ARRAY_SIZE-proof_length ;j < PROOF_ARRAY_SIZE_M1 ;j++ ) { 
   smallest = cl[j][2];
   smallest_ptr = j;
   for(m = j + 1L ; m < PROOF_ARRAY_SIZE ; m++) {
      if(cl[m][2] < smallest) {
	     smallest = cl[m][2];
	     smallest_ptr = m;}}
   c0 = cl[j][0]; cl[j][0] = cl[smallest_ptr][0]; 
   cl[smallest_ptr][0] = c0;
   c0 = cl[j][1]; cl[j][1] = cl[smallest_ptr][1]; 
   cl[smallest_ptr][1] = c0;
   c0 = cl[j][2]; cl[j][2] = cl[smallest_ptr][2]; 
   cl[smallest_ptr][2] = c0;}
	
k = 0L;
for(j = PROOF_ARRAY_SIZE - proof_length ;j < PROOF_ARRAY_SIZE ;j++) {
   cl[k][0] = cl[j][0]; cl[j][0] = 0L;
   cl[k][1] = cl[j][1]; cl[j][1] = 0L;
   cl[k][2] = cl[j][2]; cl[j][2] = 0L;
   k++;}
 
return;
}               
 
/*-------------------------------------------------------------*/
/*      bres_proof_line(c1,c2,l1_int,l2_int)...............#15
/*
/*      parameters:  c1,c2:  pointers to the clause to resolve
/*                   ix1,ix2: first and second indices of factor
/*
/*      called by:   resolve_proof_line()           pass2.c
/*
/*      calls:       standardize()                  infer.c
/*                   clear_var_substitution_tble()  infer.c
/*                   unify()                        unify.c
/*                   generate_clause()              gen.c
/*                   install_proof_line()           pass2.c
/*                   printf_cl_array()              pass2.c
/*                   print_clause()                 tdump.c
/*
/*      returns:     nil (exits from program on error)
/*-------------------------------------------------------------*/
 
void bres_proof_line(ci1,ci2,ix1,ix2)
long ci1, ci2;
int ix1, ix2;
{
register literal *l1, *l2;
register long i;
clause *c, *c1, *c2;
	
c1 = clist[ci1 - 1L];
c2 = clist[ci2 - 1L];
		
l1 = c1->c_lit;
for(i = 1L ; i < (long) ix1; i++) {
   l1 = l1->l_next;}
l2 = c2->c_lit;
for(i = 1L ; i < (long) ix2; i++) {
   l2 = l2->l_next;}
	
standardize(c1,c2);             /* standardize variables apart  */
	
	
clear_var_substitution_tble(c1->c_bits,c1->c_nv);
clear_var_substitution_tble(c2->c_bits,c2->c_nv);
if(unify(&l1->l_id,&l2->l_id)) {
   if((c = generate_clause(c1,c2,l1,l2,ix1,ix2)) != 0L) { 
      install_proof_line(c);
      clist[nextc - 1L]->c_id = nextc;
      return;}}
      
/*--------------------------------------------*/
/* at this point, if the clauses cannot be    */
/* installed, there is a bug, most likely     */
/* a hashing error.                           */
/*--------------------------------------------*/
 
else {
   big_bug = 1;
   printf("\nBIG BUG!!..#nodes=%ld\n",s_nod);
   if(o_noprint == 0) {
      fprintf(outfile,"\nBIG BUG!!..#nodes=%ld\n",s_nod); }  
   print_cl_array(0L);  
 
   for(i = 0L; i <= base; i++) {
      print_clause(clist[i]);  
      if((i == org_base) && (org_base != base)) {
	     print0("\nInferred clauses:\n");}    
      if(i == (org_base - nc))  { 
	     print0("\nNegated conclusion:\n");}} 
	
   print0("\nProof:\n");
	
/*----------------------------------*/
/* write out the proof clauses      */
/*----------------------------------*/
	
   for(i = 1L; i <= ply; i++)  {
      c = clist[depth[i].d_tc];
      if(c->c_lpar->c_id != c->c_id - 1L && i != 1L) {
	     print0("\n");}
      print_clause(c);}
 
   print0("\n");
   print_clause(c1);print_clause(c2);
   zz = getchar(); 
   fatal(" ");}
	
}
 
/*---------------------------------------------------------------*/
/*      factor_proof_line(c1,ix1,ix2)     ................#16
/*
/*      parameters:   c1:  pointer to the clause to factor
/*                    ix1,ix2: first and second indices of factor
/*
/*      called by:    resolve_proof_line()              pass2.c
/*
/*      calls:        clear_var_substitution_tble()     infer.c
/*                    unify()                           unify.c
/*                    generate_clause()                 gen.c
/*                    install_proof_line()              pass2.c
/*
/*      returns:      nil
/*---------------------------------------------------------------*/
 
void factor_proof_line(c1,ix1,ix2)
long c1;                            /* index of clause to factor    */
int ix1, ix2;                       /* index of literals to factor  */
{
register literal *l1, *l2;
register clause *c;                 /* clause to factor             */
register clause *new_clause;        /* facotred clause (new clause) */
register int lit_index;
	
c = clist[c1 - 1L];                 /* clause to factor             */
	
/*----------------------------------------------*/
/* scan for literals corresponding to ix1, ix2  */
/*----------------------------------------------*/
 
l1 = c->c_lit;
for(lit_index = 1 ; lit_index < ix1 ; lit_index++)  {
   l1 = l1->l_next;} 
	
l2 = c->c_lit;
for(lit_index = 1 ; lit_index < ix2 ; lit_index++)  {
   l2 = l2->l_next;}
			
clear_var_substitution_tble(c->c_bits,c->c_nv);
	
if(unify(&l1->l_id,&l2->l_id)) {
   new_clause = generate_clause(c,0L,l1,0L,ix1,ix2);    
   if((c->c_bits & MERGE) != 0L) {
      new_clause->c_bits |= MERGE;}
   install_proof_line(new_clause);
   clist[nextc - 1L]->c_id = nextc;}
 
return;
}
 
/*-----------------------------------------------------------*/
/*      install_proof_line(c)             .............#17
/*
/*      install clause c on the clist
/*
/*      called by:  bres_proof_line()       pass2.c
/*                  fact_proof_line()       pass2.c
/*
/*      calls:      nobody
/*
/*      returns:    nil
/*-----------------------------------------------------------*/
 
void install_proof_line(c)
clause *c;
{
dnode *d;
	
d = &depth[ply];                       /* insert at current ply */
c->c_id = base + ply + 1L ;            /* give clause an ID     */
d->d_tc = nextc++;

clist[nextc] = c->c_end;
 
if((long) c->c_end > (long) d->d_rc) {
   d->d_rc = c->c_end;}
 
clist[d->d_tc] = c;
return;
}
  
/*---------------------------------------------------------------*/
/*      resolve_proof_line()            ................#18
/*
/*      determine next inference (factor or binary resol) as indicated
/*      by the cl_array
/*
/*      called by:  search_proof_line()   pass2.c
/*
/*      calls:      bres_proof_line()     pass2.c
/*                  factor_proof_line()   pass2.c
/*
/*      returns:    nil
/*---------------------------------------------------------------*/
 
void resolve_proof_line()
{
long c1, c2;                      /* index of clauses to resolve  */
long c1_index, c2_index;          /* index of literals to resolve */

/*----------------------*/
/* decode cl info       */
/*----------------------*/
	
c1       = (cl[ply - 1L][0] >> 4L) & PARENT_MASK;
c1_index = (cl[ply - 1L][0] & INDEX_MASK);
c2           = (cl[ply - 1L][1] >> 4L) & PARENT_MASK;
c2_index = (cl[ply - 1L][1] & INDEX_MASK);
 
if(c2 == 0L) {
   factor_proof_line(c1,(int) c1_index, (int) c2_index);
   return;}
bres_proof_line(c1,c2,(int) c1_index, (int) c2_index);
 
return;
}                                                       
 
/*---------------------------------------------------------------*/
/*      search_proof_line()             ................#19
/*
/*      search_proof_line generates a single path of inferences as
/*      guided by the cl[] array.  This is performed after phases
/*      1 and 2 are complete
/*
/*      called by:  main()                  tgtp.c
/*
/*      calls:      resolve_proof_line()    pass2.c
/*
/*      returns:    nil
/*---------------------------------------------------------------*/
 
void search_proof_line()
{
dnode *d;
int o_maxterm_temp;

depth[0].d_tc = base;               /* initial generating clause    */
depth[0].d_ec = base;               /* top clause at level 0 = base */
depth[0].d_bc = 0L;                 /* bottom clause at level 0 = 0 */
	
nextc = base +1L;                      /* start storing proof here  */
 
/*--------------------------------------------*/
/* determine where to begin storing pointers  */
/* to clauses in the proof                    */
/*--------------------------------------------*/ 
 
clist[nextc] = depth[0].d_rc = clist[base]->c_end;
 
/*----------------------------------------------*/
/* do not place a limit on the number of terms  */
/* in a literal in the proof                    */
/*----------------------------------------------*/
 
o_maxterm_temp = o_maxterm;
o_maxterm = 60L;
 
/*----------------------------------------------*/
/* pick off resolutions stored in the cl-array  */
/* carry them out to make sure that a hash err  */
/* has not led to an illegal resolution         */
/*----------------------------------------------*/

for(ply = 1L; ply <= proof_length; ply++) {
   d = &depth[ply];
   nextc  = depth[ply - 1L].d_ec + 1L;  
   d->d_rc = 0L;
   d->d_tc = d->d_bc = nextc;
   resolve_proof_line();
   d->d_ec = depth[ply - 1L].d_ec + 1L;
 
/*----------------------------*/
/* break on null clause       */
/*----------------------------*/
 
   if(clist[d->d_tc]->c_nlit == 0) {
      o_maxterm = o_maxterm_temp;
      return;}}
}
 
/*--------------------------------------------------------------*/
/*      delete_alt_hcs(c)   ................#8
/*
/*      calculates var in the formula, and for each constant
/*      in the list of constants for theorem, calculate a hash
/*      code and add it to the clause hash table
/*
/*      called by:      install_clause()               search.c
/*
/*      calls:          calc_alt_hash()                hash.c
/*                      add_clause_to_ht()             hash.c
/*
/*      returns:        0L;2L if contradiction found
/*---------------------------------------------------------------*/
 
void delete_alt_hcs(c) 
clause *c;
{
register long const1, const2, varbl, varbl2;
literal *l;
 
/*-----------------------------------------------*/
/* if there is one variable, replace it with     */
/* each constant and calculate the resulting     */
/* hash code.  add the hash code to the clause   */
/* hash table.                                   */
/*-----------------------------------------------*/

for(l = c->c_lit; (l->l_id & HHITMASK) != 0L; l = l->l_next) { };

if(c->c_nv == 1L) {
   for(const1 = 1L ; const1 <= (long) number_constants ; const1++) {
      calc_alt_hash(l,0L,-1L,const1,0L);    
      delete_found_hcs(alt_hx,alt_hy);}
 
/*-----------------------------------------------*/
/* calc hashcode with single variable replaced   */
/* with the variable y.  most likely, but not    */
/* for sure, this will produce a new hashcode    */
/*-----------------------------------------------*/
 
   calc_alt_hash(l,0L,-1L,1L - (long) rows_less_nfxns,0L);
   delete_found_hcs(alt_hx,alt_hy);}
 
/*-----------------------------------------------*/
/* if there are two variables, replace each one  */
/* with a constant and calculate the hashcode    */
/*-----------------------------------------------*/
 
if(c->c_nv == 2L) {
   for(varbl = 0L; varbl < 2L ; varbl++) {
     for(const1 = 1L; const1 <= (long) number_constants ; const1++) {
	    calc_alt_hash(l,(long) (var_array[varbl]-1),-1L,const1,0L);
	    delete_found_hcs(alt_hx,alt_hy);}}}
			
/*------------------------------------------------*/
/* replace every pair of variables with every     */
/* pair of constants, calculate the hashcode      */
/* and add the hashcode to the literal hash table */
/*------------------------------------------------*/ 
 
if(c->c_nv >= 2L) {
  for(varbl = 0L; varbl < c->c_nv - 1L; varbl++) {
   for(varbl2 = varbl + 1L ; varbl2 < c->c_nv; varbl2++) {
     for(const1 = 1L; const1 <= (long) number_const_hash; const1++) {
       for(const2 = 1L; const2 <= (long) number_const_hash; const2++) {
          calc_alt_hash(l,varbl,varbl2,const1,const2);
          delete_found_hcs(alt_hx,alt_hy);}}}}}
return;
}
