/*
   td_prove.pl
   David Sharpe
   June 30, 1996

Adjustments for the CASC14 competition Bruce Spencer June 1997 (look for EBS 
with your editor)


   To run td_prove on TPTP problem X, do the following two steps:
   1a) type: eclipse -b gen_ap_db.pl [ENTER]
    b) type: tptp('X'). [ENTER]
    c) type: halt. [ENTER]
   Step 1 generates the db file X.ap_db
   2a) type: eclipse -b td_prove.pl
    b) set all the flags, explained below, to control the procedure,
    c) type: ['X.ap_db'] [ENTER] to compile the db
    d) type: go. [ENTER] to run the procedure.

EBS
With eclipse -b gen_ap_db.pl there is another interface, tptp_sub().
See that file for more info.  The output is then put in 'ap_db'

EBS
There is a CASC mode, invoked by running 'go_casc', that 
takes its input from 'ap_db' and prints either
"A proof exists" or
"No proof exists (clauses are satisfiable)"
If the proof exists, it is printed to a file 'proof'.


   td_prove is a general top-down automated reasoning program with the
   following short list of default inference mechanisms:
   1.Anc Merge
   2.Extension Resolution,

   and the following set of options available by setting flags (see
   D.Sharpe AllPaths thesis for a discussion of these options):
   1.Anc Taut.
   2.Left Merge
   3.Left Taut.
   4.Right Merge
   5.Right Taut.
   6.Early (extension time) or Late (goal selection) Taut. detection
   7.Disequality (merge and taut.) N.B.taut. must be EARLY for DIS
   8.Deterministic merges or not (ALPOC only uses nondeterm. merges)
   9.Nondeterministic Left merges or not
   10. Nondeterministic Right merges or not
   11.OC,
   12.UnitNotVL,
   13.IdentNotVL,
   14.balanced IDA*
   15.inference IDA* (default)

   Furthermore, the procedure can be run with a number of different
   settings to produce different output, as follows:
   1. Batch mode (puts output to a file X.count)
   2. proof (accumulates a Prolog representation of the proof tree
      and prints statistics on the final proof tree).
   3. gather statistics (count)
   4. print out the proof tree (printproof).


   The options are chosen by setting flags.  A series of commented out
   flag settings are provided.  The user should uncomment the flag
   settings of choice.
*/

/*
 * Eclipse Compiler directives, should not be unset without a good
 * reason.
 */
:- set_flag(debug_compile,off).
:- use_module(library(quintus)).
:- ensure_loaded(library(lists)).
:- ['unify_comp.pl']. % sound unify with occurs
:- op(100,fx,(++)).
:- op(100,fx,(--)).
:- dynamic restart/11.  % restart with counts & times (not done yet)
anc_merge_flag.         % default merge.
search_iter(1).         % iterative deepening step size    
search_bound(10000).    % arbitrary cut off

/**********************************************************************/
/* Flag options for td_prove
/* These are included here _ALL_ commented out. I suggest that the
/* flags section be copied to a seperate file and the desired flags
/* uncommented there.  Then a record of which flags are set are stored
/* in a file for future reference.
/**********************************************************************/

/*
 * flags that control how td_prove runs (batch,count,proof,printproof).
 * Each flag should have exactly one of the two options uncommented.
 */
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%batch.
batch:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%quiet. %EBS
quiet:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%count.
count:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
proof.
%proof:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
printproof.                        % requires proof to be on
%printproof:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/*
 * Flags that control the inference mechanisms and search pruning
 * tactics.  Because of the general nature of this procedure (and
 * limited time), not all of the option combinations will be done
 * efficiently.  I attempt to state any combinations disallowed; however
 * there may be some I miss.... And of course I make no claims about
 * the soundness, completeness, or indeed the wiseness, of any of the
 * sets of options.
 * Each flag should have exactly one of the two options uncommented.
 */
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
left_merge_flag.
%left_merge_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
right_merge_flag.
%right_merge_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
anc_taut_flag.
%anc_taut_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
left_taut_flag.
%left_taut_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
right_taut_flag.
%right_taut_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
late_taut_flag.
%late_taut_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
early_taut_flag.
%early_taut_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
taut_dis_flag.                     % requires early_taut_flag to be on.
%taut_dis_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
merge_dis_flag.                    % requires early_taut_flag to be on.
%merge_dis_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
det_merge_flag.
%det_merge_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
non_det_left_flag.
%non_det_left_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
non_det_right_flag.
%non_det_right_flag:fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
oc_flag.
%oc_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
balanced_flag.
%balanced_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
unit_not_vl_flag.
%unit_not_vl_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
ident_lit_not_vl_flag.
%ident_lit_not_vl_flag:-fail.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

/**********************************************************************/
/* The top-down prover (td_prove) proof procedure                     */
/**********************************************************************/
/* td_prove Argument List:
   arg1: list of goal literals, each element is a 5 argument term:
      g(literal,node_num,clause_num,level,flag), where
         literal    = positive (++) or negative (--) prolog predicate,
         node_num   = unique integer assigned to every atom in the tree,
         clause_num = input clause number,
         level      = level in the tree, root atoms are level 1,
         flag       = flag for right hook merge head:
            var(flag) true ==> goal is not a right hook merge head,
            var(flag) false ==> goal is a right hook merge head
   arg2: visible literals information in a 4 argument term:
      v(VR0,VA0,VL0,S0), where VR0, VA0, VL0 are VisLit-info(...)
         VR0 info has 3 args: node_num,level,flag (as in arg1).
         VA0 info has 2 args: clause_num,level
         VL0 info has 2 arg:  nas (nearest ancestor support level),
                right support node number list
         S0  = SA:SR, where SA is ancestor support level list, and
                SR is right node number support list.
   arg3: tree information in a 3 argument term: t(N0,D0,B0)
      N0 = greatest node number in the tree input
      D0 = disequality list, [d(pred1,pred2,type),...]
           where pred1 and pred2 are 2 (unsigned) predicates that form a
           disequality if they ever become exactly equal, and type is
           merge_dis, or one of (right_taut,anc_taut,left_taut).
      B0 = iterative deepening bound input,
   arg4: visible literals information output
   arg5: tree information output
   arg6: proof output term (see proof printing procedures for format).
*/

/*
 * success if no goals left
 */
td_prove([],V,T,V,T,[]).

/*
 * Right Hook Merge Head.
 */
td_prove([g(C,N,ClN,L,F)|Gs],V0,T0,V5,T2,[PfG|PfGs]):-
   F == rhm, %test goal flag as a right hook merge head
   v_get_VL(V0,VL0),
   remove_invis_left(N,[],VL0,Invis,VL1),
   v_add_VL(VL1,V0,V1),
   v_remove_RS(V1,N,V2),
   !,
   td_prove([g(C,N,ClN,L,_)],V2,T0,V3,T1,[PfG]),
   v_get_VL(V3,VL2),
   v_get_S(V3,S),
   restore_vis_left(Invis,S,VL2,VL3),
   v_add_VL(VL3,V3,V4),
   td_prove(Gs,V4,T1,V5,T2,PfGs).

/*
 * late tautology check (failure driven)
 */
td_prove([G|_Gs],V0,_T0,_V3,_T1,_Pf):-
   late_taut_flag,
   g_off_v(G,V0,V1),
   g_get_c(G,C),
   \+ tautology_chk([C-_],V1), % goal passed as a visible list
   !,
   fail.

/*
 * goal node is deterministically merged
 */
td_prove([G|Gs],V0,T0,V3,T1,[PfG|PfGs]):-
   (   \+ first_order   % prop logic always deterministic
   ;   det_merge_flag
   ),
   merge_id(G,V0,MergeType,Info),
   !,
   g_off_v(G,V0,V1),
   my_incval(MergeType),
   proof(G,MergeType,0,_,PfG),
   td_prove(Gs,V1,T0,V2,T1,PfGs),
   v_add_a_sup(MergeType,Info,V2,V3).

/*
 * nondeterministic merge
 */
td_prove([G|Gs],V0,T0,V3,T1,[PfG|PfGs]):-
   first_order,
   merge_oc(G,V0,MergeType,Info),
   t_dis_chk(T0),
   g_off_v(G,V0,V1),
   my_incval(MergeType),
   proof(G,MergeType,0,_,PfG),
   td_prove(Gs,V1,T0,V2,T1,PfGs),
   v_add_a_sup(MergeType,Info,V2,V3).

/*
 * extend the tree with a new input clause
 */
td_prove([G|Gs],V0,t(N0,D0,B0),V6,T2,[PfG|PfGs]):-

   % update G and find an extension
   g_update(G,V0,ClM,NegC,Lplus1,V1),        % change G to internal node
   build_merge_dis(G,V1,D0,D1),              % unifiable merges are DIS
   extension_rule(NegC,B0,B1,ClM,Ext,ExtVR), % extend within IDA* bound
   dis_chk(D0),
   my_incval(ext),

   % early tautology path check and tautology disequality build
   (  early_taut_flag
   -> tautology_chk(ExtVR,V1),
      build_taut_dis(ExtVR,V1,D1,D2)
   ;  true
   ),

   % extension clause enumerated and visible to the right, and empty S
   enum_nodes(Ext,Lplus1,N0,N1),
   v_extend(ExtVR,V1,V2),

   % prove new goals in the extension of the current goal
   td_prove(Ext,V2,t(N1,D2,B1),V3,t(N2,D3,B2),PfExt),

   % bookeeping and IDA* search
   ExtSize is N1 - N0,
   update_vis_for_siblings(G,ExtSize,V3,V4),
   proof(G,ClM,ExtSize,PfExt,PfG),
   (  balanced_flag
   -> B3 is B0
   ;  B3 is B2  % default inference count IDA*
   ),

   % prove the open siblings of G
   td_prove(Gs,V4,t(N2,D3,B3),V5,T2,PfGs),
   v_combine_sup(V0,V5,V6).

/**********************************************************************/
/* Procedures called that use 'v' 'g' 'info' or 't' wrappers          */
/**********************************************************************/
v_get_VL(v(_,_,VL,_),VL).
v_get_S(v(_,_,_,S),S).

v_add_VL(VL,v(VR,VA,_,S),v(VR,VA,VL,S)).

v_remove_RS(v(VR,VA,VL,SA:SR0),N,v(VR,VA,VL,SA:SR1)):-
   delete_all(N,SR0,SR1).

v_extend(ExtVR,v(VR0,VA,VL,_),v(VR1,VA,VL,[]:[])):-
   append(ExtVR,VR0,VR1),
   !.

v_combine_sup(v(_,_,_,SA0:SR0),v(VR,VA,VL,SA1:SR1),v(VR,VA,VL,SA:SR)):-
   union(SA0,SA1,SA),
   union(SR0,SR1,SR),
   !.

v_add_a_sup(left,info(SA0,SR0),v(VR,VA,VL,SA1:SR1),v(VR,VA,VL,SA:SR)):-
   SA = [SA0|SA1],
   union(SR0,SR1,SR),
   !.
v_add_a_sup(anc,info(_,SA0),v(VR,VA,VL,SA1:SR),v(VR,VA,VL,[SA0|SA1]:SR)).
v_add_a_sup(right,info(SR,_,_),v(VR,VA,VL,SA:SR1),v(VR,VA,VL,SA:SR2)):-
   union([SR],SR1,SR2),
   !.

g_off_v(g(_,N,_,_,_),v(VR0,VA,VL,S),v(VR1,VA,VL,S)):-
   take_num_off_VR(N,VR0,VR1).
g_off_VR(g(_,N,_,_,_),VR0,VR1):-
   take_num_off_VR(N,VR0,VR1).

g_get_c(g(C,_,_,_,_),C).

/* update for extension by finding Negated clause, level and level plus
 * one, and moving G from a vis right to a vis ancestor
 */
g_update(g(C,N,_,L,_),v(VR0,VA,VL,S),ClM,
   NegC,Lplus1,v(VR1,[NegC-info(ClM,L)|VA],VL,S)):-
   negate(C,NegC),
   Lplus1 is L + 1,
   take_num_off_VR(N,VR0,VR1),
   !.

/* ordered_clause_chk true if the ordered clause restriction is not
   violated   for a potential ancestor merge path:
   arg1: path tail clause number
   arg2: path head info
*/
info_oc_chk(ClN,info(ClM,_)):-
      oc_flag
   -> ClN >= ClM
   ;  true.

t_dis_chk(t(_,D,_)):-
   dis_chk(D).

/**********************************************************************/
/* Procedures called during proof                                     */
/**********************************************************************/
/* first argument returned negated in second argument */
negate(++X,--X).
negate(--X,++X).

/* term in first arg returned without sign in second argument */
atomize(++A,A).
atomize(--A,A).

/* update proof term */
proof(g(C,_N,_ClN,L,_),Type,NumExt,Reason,p(C,Type,L,NumExt,Reason)):-
   proof,!. 
proof(_,_,_,_,_). %EBS Nonexistence of this line appears to be a bug

/* increment global counter */
my_incval(X):-
   count,!,
   incval(X).
my_incval(_). %EBS  This non-existence of this line appears to be a bug.



/* extension_rule finds the next axiom clause head that matches the goal
   by prolog's regular lexical ordered nondeterministic backtracking
   search.  Only extensions within the iterative deepening bound are
   found.
      arg1: negation of the goal
      arg2: depth bound in
      arg3: depth bound out
      arg4: clause number
      arg5: extension literals
      arg6: list of visible right extension literals
*/
extension_rule(++Pred,B0,B1,ClNum,Ext,ExtVR):-
   functor(Pred,PredSymbol,_Arity),
   !,
   pos_ext(PredSymbol,++Pred,ClNum,Estimate,UnifyList,Ext,ExtVR),
   tally(Estimate,B0,B1),
   call_list_mbrs(UnifyList).
extension_rule(--Pred,B0,B1,ClNum,Ext,ExtVR):-
   functor(Pred,PredSymbol,_Arity),
   !,
   neg_ext(PredSymbol,--Pred,ClNum,Estimate,UnifyList,Ext,ExtVR),
   tally(Estimate,B0,B1),
   call_list_mbrs(UnifyList).

/* iterative deepening tally. Modified from O'Keefe,
 * "The Craft of Prolog" p.66
 */
tally(N,B0,B1):-
   B0 >= N,
   B1 is B0 - N.

/* take_num_off_VR removes node number from the visible right list
   arg1: node number
   arg2: input visible right list
   arg3: updated visible right list
*/
take_num_off_VR(N,[_C-info(N,_,_)|T],T):-
   !.
take_num_off_VR(N,[H|T],[H|T1]):-
   take_num_off_VR(N,T,T1).

/*tautology_chk top level clause that fails when a tautology path
  exists between a literal in the extension clause and a literal
  in any of the flaged visible lists.  There are 8 flag combinations.
   arg1: extension clause literal list.
   arg2: visible term
*/
tautology_chk(ExtVR,v(VR,VA,VL,_S)):-
   anc_taut_flag,
   left_taut_flag,
   right_taut_flag,
   !,
   tautology_chk2(ExtVR,anc_taut,VA),
   tautology_chk2(ExtVR,right_taut,VR),
   tautology_chk2(ExtVR,left_taut,VL),
   near_right_tautology_chk2(ExtVR),
   !.
tautology_chk(ExtVR,v(_VR,VA,VL,_S)):-
   anc_taut_flag,
   left_taut_flag,
   \+ right_taut_flag,
   !,
   tautology_chk2(ExtVR,anc_taut,VA),
   tautology_chk2(ExtVR,left_taut,VL),
   !.
tautology_chk(ExtVR,v(VR,VA,_VL,_S)):-
   anc_taut_flag,
   \+ left_taut_flag,
   right_taut_flag,
   !,
   tautology_chk2(ExtVR,anc_taut,VA),
   tautology_chk2(ExtVR,right_taut,VR),
   near_right_tautology_chk2(ExtVR),
   !.
tautology_chk(ExtVR,v(_VR,VA,_VL,_S)):-
   anc_taut_flag,
   \+ left_taut_flag,
   \+ right_taut_flag,
   !,
   tautology_chk2(ExtVR,anc_taut,VA),
   !.
tautology_chk(ExtVR,v(VR,_VA,VL,_S)):-
   \+ anc_taut_flag,
   left_taut_flag,
   right_taut_flag,
   !,
   tautology_chk2(ExtVR,right_taut,VR),
   tautology_chk2(ExtVR,left_taut,VL),
   near_right_tautology_chk2(ExtVR),
   !.
tautology_chk(ExtVR,v(_VR,_VA,VL,_S)):-
   \+ anc_taut_flag,
   left_taut_flag,
   \+ right_taut_flag,
   !,
   tautology_chk2(ExtVR,left_taut,VL),
   !.
tautology_chk(ExtVR,v(VR,_VA,_VL,_S)):-
   \+ anc_taut_flag,
   \+ left_taut_flag,
   right_taut_flag,
   !,
   tautology_chk2(ExtVR,right_taut,VR),
   near_right_tautology_chk2(ExtVR),
   !.
tautology_chk(_ExtVR,_V):-
   \+ anc_taut_flag,
   \+ left_taut_flag,
   \+ right_taut_flag,
   !.

tautology_chk2([],_Type,_V):- !.
tautology_chk2([E-_|_],Type,V):-
   negate(E,NegE),
   tautology_id(NegE,V),
   my_incval(Type),
   !,
   fail.
tautology_chk2([_|Es],Type,V):-
   tautology_chk2(Es,Type,V).

near_right_tautology_chk2([]):-!.
near_right_tautology_chk2([_]):-!.
near_right_tautology_chk2([E-_|Es]):-
   negate(E,NegE),
   tautology_id(NegE,Es),
   my_incval(right_taut),
   !,
   fail.
near_right_tautology_chk2([_|Es]):-
   near_right_tautology_chk2(Es).

/* build_taut_dis constructs the tautology disequalities from the
   extension clause literals into the flaged visible literals,
   calls the second build_taut_dis2 for the real work:
   arg1: extension vis right literals
   arg2: visible lists
   arg3: dis list in
   arg4: dis list out
*/
build_taut_dis(_,_,D,D):-
   \+ first_order,
   !.
build_taut_dis(_,_,D,D):-
   \+ taut_dis_flag,
   !.
build_taut_dis(C,V,D0,D1):-
   build_taut_dis2(C,V,D0,D1).

build_taut_dis2([],_,D,D):-
   !.
build_taut_dis2([C-_|Cs],v(VR,VA,VL,_S),D1,D6):-
   negate(C,NegC),
   (  right_taut_flag
   -> build_dis2(NegC,VR,right_taut,D1,D2),
      build_dis2(NegC,Cs,right_taut,D2,D3)
   ;  D3 is D1
   ),
   (  anc_taut_flag
   -> build_dis2(NegC,VA,anc_taut,D3,D4)
   ;  D4 is D3
   ),
   (  left_taut_flag
   -> build_dis2(NegC,VL,left_taut,D4,D5)
   ;  D5 is D4
   ),
   build_taut_dis2(Cs,v(VR,VA,VL,_S),D5,D6).

/* build_merge_dis constructs the merge disequalities from the
   literal about to be extended into the rest of the visible tree:
   (note that the current goal is not visible to the right as
   is the normal condition)
   arg1: current goal
   arg2: visible lists
   arg3: dis list in
   arg4: dis list out
*/
build_merge_dis(_,_,D,D):-
   \+ first_order,
   !.
build_merge_dis(_,_,D,D):-
   \+ merge_dis_flag,
   !.
build_merge_dis(G,V,D0,D1):-
   build_merge_dis2(G,V,D0,D1).

build_merge_dis2(g(C,_,_,_,_),v(VR,VA,VL,_S),D1,D4):-
   (  anc_merge_flag
   -> build_dis2(C,VA,merge_dis,D1,D2)
   ;  D2 is D1
   ),
   (  left_merge_flag
   -> build_dis2(C,VL,merge_dis,D2,D3)
   ;  D3 is D2
   ),
   (  right_merge_flag
   ->   build_dis2(C,VR,merge_dis,D3,D4)
   ;  D4 is D3
   ),
   !.

/* build_dis2 builds a list of 3 element lists which unify with
   the current literal plus the type:
   arg1: current literal
   arg2: list of visible literals
   arg3: type of disequality being built
   arg4: dis list input
   arg5: dis list output
*/
build_dis2(_,[],_,D,D):-!.
build_dis2(C,[V-_|Vs],Type,D1,D2):-
   \+ unify(C,V),
   !,
   build_dis2(C,Vs,Type,D1,D2).
build_dis2(C,[V-_|Vs],Type,D1,[d(A1,A2,Type)|D2]):-
   atomize(C,A1),
   atomize(V,A2),
   build_dis2(C,Vs,Type,D1,D2).

/* dis_chk: fails if a disequality exists in the dislist:
   (Note dis_chk arg is empty list if dis_flag is not set).
   arg1: disequality list
*/
dis_chk([]):-
   !.
dis_chk([d(A1,A2,Type)|_]):-
   A1 == A2,
   !,
   my_incval(Type),
   my_incval(taut_dis),
   fail.
dis_chk([_|T]):-
   dis_chk(T).

/* enum_nodes assigns a number to each new extension literal */
enum_nodes([],_,N,N):-!.
enum_nodes([g(_,N0,_,L,_)|Gs],L,N0,N2):-
   N1 is N0 + 1,
   enum_nodes(Gs,L,N1,N2).

/* update_vis_for_siblings after a sibling is proved update the visible
   lists and ancestor support if current is an ancestor support:
   arg1: goal just proved by extension
   arg2: extension size: the number of literals in the extension clause
   arg3: input visible term
   arg4: output visible term
*/
update_vis_for_siblings(g(_,_,_,L,_),0,v(VR,VA0,VL,S),v(VR,VA1,VL,S)):-
   unit_not_vl_flag, % nodes with 0 children are not visible to the left
   take_curr_off_VA(L,VA0,VA1),
   !.
update_vis_for_siblings(g(C,_,_,L,_),_,v(VR,VA0,VL0,SA0:SR),
   v(VR,VA1,VL2,SA1:SR)):-
   curr_support(L,SA0,SA1),         %test if C is an ancestor support
   update_vis_left(L,VL0,VL1),      %remove all vis left with nas of L
   calc_nas(SA1,0,NAS),             %new nas initialy set at root (0)
   VL2 = [C-info(NAS,SR)|VL1],      %compose visible left
   take_curr_off_VA(L,VA0,VA1),
   !.
update_vis_for_siblings(g(C,_,_,L,_),_,v(VR,VA0,VL0,SA0:SR),
   v(VR,VA1,VL1,SA0:SR)):-
   take_curr_off_VA(L,VA0,VA1),
   calc_nas(SA0,0,NAS),             %nearest initialy at root (0)
   VL1 = [C-info(NAS,SR)|VL0],      %compose visible left
   !.

/* take_curr_off_VA removes the current literal (at given level) from
   the visible   ancestor list:
   arg1: level
   arg2: input visible ancestor list
   arg3: updated visible ancestor list
*/
take_curr_off_VA(L,[_-info(_,L)|T],T):-
   !.
take_curr_off_VA(L,[H|T],[H|T1]):-
   take_curr_off_VA(L,T,T1).

/* curr_support, fails if the current level is not an ancestor support
   otherwise succeeds with the support list updated so that it does not
   contain the current level:
   arg1: current level
   arg2: input support list
   arg3: updated support list
*/
curr_support(L,[L|T],T1):-
   delete_all(L,T,T1).
curr_support(L,[H|T],[H|T1]):-
   curr_support(L,T,T1).

/* update_vis_left takes literals off of visible left list with ancestor
   support:
   arg1: ancestor support level
   arg2: input visible left list
   arg3: updated visible left list
*/
update_vis_left(_,[],[]):-
   !.
update_vis_left(L,[_-info(L,_)|T],VL):-
   update_vis_left(L,T,VL).
update_vis_left(L,[H|T],[H|VL]):-
   update_vis_left(L,T,VL).

/* calc_nas finds the nearest (at greatest level) of ancestor support
   less than the current level:
   arg1: ancestor support levels
   arg2: input nearest support
   arg3: updated nearest support
*/
calc_nas([],NAS,NAS):-!.
calc_nas([X|Xs],NAS0,NAS1):-
      X > NAS0
   -> % if X is a nearest ancestor support so far
      calc_nas(Xs,X,NAS1)
   ;  % else keep nearest ancestor support so far
      calc_nas(Xs,NAS0,NAS1).

/* remove invisible to the left nodes supported by a right hook merge
   path.
   arg1: right support node number
   arg2: invisible left nodes input
   arg3: visible left nodes input
   arg4: invisible left nodes output
   arg5: visible left nodes output
*/
remove_invis_left(_,Invis,[],Invis,[]):-
   !.
remove_invis_left(N,Invis0,[Lit-info(NAS,RS0)|VL0],
  [Lit-info(NAS,RS1)|Invis1],VL1):-
   det_delete(N,RS0,RS1),
   remove_invis_left(N,Invis0,VL0,Invis1,VL1).
remove_invis_left(N,Invis0,[VL|VL0],Invis1,[VL|VL1]):-
   remove_invis_left(N,Invis0,VL0,Invis1,VL1).

/* restore left nodes supported by a right hook node AFTER the right
   hook merge path is finnished.
   arg1: right support node number
   arg2: invisible left nodes input
   arg3: visible left nodes input
   arg4: invisible left nodes output
   arg5: visible left nodes output
*/
restore_vis_left([],_,VL,VL):-
   !.
restore_vis_left([Lit-info(NAS0,SR0)|Invis0],SA:SR,VL0,
  [Lit-info(NAS1,SR1)|VL1]):-
   ident_lit_not_VL_chk(Lit,VL0),
   !,
   union(SR0,SR,SR1),
   calc_nas(SA,NAS0,NAS1),
   restore_vis_left(Invis0,SA:SR,VL0,VL1).
restore_vis_left([_|InvisT],SA:SR,VL0,VL1):-
   restore_vis_left(InvisT,SA:SR,VL0,VL1).

/* check if there are no identical literals already visible to the
   left (IdentLitNotVL)
*/
ident_lit_not_VL_chk(_,_):-
   \+ ident_lit_not_vl_flag,
   !.
ident_lit_not_VL_chk(_,[]):-
   !.
ident_lit_not_VL_chk(Lit,[VisLit-_|_]):-
   Lit == VisLit,
   !,
   fail.
ident_lit_not_VL_chk(Lit,[_|T]):-
   ident_lit_not_VL_chk(Lit,T).

/* merge current with visible node,with occurs check */
/* merge goal with visible right, ancestor, or left node with occurs*/
/* first try with visible right */
merge_oc(g(C,N,_,_,_),v(VR0,_,_,_),right,Info):-
   right_merge_flag,
   non_det_right_flag,
   take_num_off_VR(N,VR0,VR1),
   merge_oc2(C,VR1,Info),
   Info = info(_,_,rhm). % set info flag
/* next try with visible ancestor, also check if its redundant (oc) */
merge_oc(g(C,_,ClN,_,_),v(_,VA,_,_),anc,Info):-
   anc_merge_flag,
   merge_oc2(C,VA,Info),
   info_oc_chk(ClN,Info).
/* finally try with visible left */
merge_oc(g(C,_,_,_,_),v(_,_,VL,_),left,Info):-
   left_merge_flag,
   non_det_left_flag,
   merge_oc2(C,VL,Info).

/* find unifiable match in visible list */
merge_oc2(C,[H-Info|_T],Info):-
   unify(C,H).
merge_oc2(C,[_|T],Info):-
   merge_oc2(C,T,Info).


/* merge goal with visible right, ancestor, or left node */
/* first try with visible right */
merge_id(g(C,N,_,_,_),v(VR0,_,_,_),right,Info):-
   right_merge_flag,
   take_num_off_VR(N,VR0,VR1),
   merge_id2(C,VR1,Info),
   Info = info(_,_,rhm),
   !.
/* next try with visible ancestor, also check if it redundant (oc) */
merge_id(g(C,_,ClN,_,_),v(_,VA,_,_),anc,Info):-
   anc_merge_flag,
   merge_id2(C,VA,Info),
   !,
   info_oc_chk(ClN,Info).
/* finally try with visible left */
merge_id(g(C,_,_,_,_),v(_,_,VL,_),left,Info):-
   left_merge_flag,
   merge_id2(C,VL,Info),
   !.

/* find identical match in visible list */
merge_id2(C,[H-Info|_T],Info):-
   C == H,
   !.
merge_id2(C,[_|T],Info):-
   merge_id2(C,T,Info).

/* tautology_id succeeds when a (identical) match exists between the
   literal in the first argument and a literal in the visible list.
*/
tautology_id(NegLit,[VisLit-_|_]):-
   NegLit == VisLit,
   !.
tautology_id(NegLit,[_|T]):-
   tautology_id(NegLit,T).

/**********************************************************************/
/* Utilities and top level for allpaths procedure                     */
/**********************************************************************/
/* top level to build a tptp database in an external file, and have it
   compiled in the current session and then run the allpaths procedure
*/
ap_tptp(Prob):-
   consult('gen_ap_db.pl'),
   name(Prob,ProbList),
   name('.ap_db',FileExtension),
   append(ProbList,FileExtension,DBFileList),
   name(DBFile,DBFileList),
   consult(DBFile),
   tptp(Prob),
   go.

/* top level predicate to call the theorem prover */
go_casc_printproof :- %EBS
	compile(ap_db),
	abolish(batch/0), assert((batch:-fail)), % take control of the flags
	abolish(proof/0), assert(proof), 
	abolish(printproof/0), assert(printproof), 
	abolish(quiet/0), assert((quiet:-fail)), 
	abolish(count/0), assert((count:-fail)), 
	statistics(runtime, [Time0,_]),
	(go -> write('A proof exists') ; write('No proof exists')), 
	statistics(runtime, [Time1, _]),
	nl, Time is Time1-Time0,
	write('User CPU Time is '), write(Time), 
	write(' milliseconds for proof.'), nl.


go_casc:- %EBS
	compile(ap_db),
	abolish(batch/0), assert((batch:-fail)), % take control of the flags
	abolish(printproof/0), assert((printproof:-fail)), 
	abolish(proof/0), assert((proof:-fail)), 
	abolish(quiet/0), assert(quiet), 
	abolish(count/0), assert((count:-fail)), 
	statistics(runtime, [Time0,_]),
	(go -> write('A proof exists') ; write('No proof exists')), 
	statistics(runtime, [Time1, _]),
	nl, Time is Time1-Time0,
	write('User CPU Time is '), write(Time), 
	write(' milliseconds for proof.'), nl.

go:-
   problem(Prob),
   query(Query,QueryVR,Min,N),
   search_bound(Max),
   search_iter(Inc),
   (batch ->
      % output to file, else output to standard out
      name(Prob,FileName),
      (count ->
         name('.count',FileExtension)
      ;
         name('.time',FileExtension)
      ),
      append(FileName,FileExtension,FileList),
      name(File,FileList),
      open(File,append,output),
      Stream = output
   ;
      Stream = user
   ),

   build_taut_dis(QueryVR,v([],[],[],[]:[]),[],D0),
   V = v(QueryVR,[],[],[]:[]),
   (  quiet -> true ; %EBS
      put_table_heading(Stream)
   ),% EBS

   (  restart(L,T,Ext,RM,AM,LM,RT,AT,LT,MD,TD)
   -> % reset counters and set level
      setval(ext,Ext),setval(left,LM),setval(anc,AM),setval(right,RM),
      setval(left_taut,LT),setval(anc_taut,AT),setval(right_taut,RT),
      setval(merge_dis,MD),setval(taut_dis,TD),
      Level = L
   ;  % otherwise zero counters and set time and level
      setval(ext,0),setval(left,0),setval(anc,0),setval(right,0),
      setval(left_taut,0),setval(anc_taut,0),setval(right_taut,0),
      setval(merge_dis,0),setval(taut_dis,0),
      Level = Min, T = 0
   ),

   !,

   % run the procedure with iterative deepening wrapper
   statistics(runtime,[Time0|_]),
   iterate(Query,V,N,D0,Level,Max,Inc,Time0,Stream,Proof),
   statistics(runtime,[Time1|_]),

   Time2 is Time1 - Time0 + T,
(quiet -> true ;  %EBS
   nl(Stream), write(Stream, 'Problem '),write(Stream, Prob),
   write(Stream, ' proved '), write(Stream, Time2),
   write(Stream, ' msec.'),nl(Stream),
   (  proof
   -> setval(fin_ext,0),setval(fin_uext,0),setval(fin_left,0),
      setval(fin_anc,0),setval(fin_right,0),
      nl(Stream),
      write(Stream, 'Final Proof used: #Ext, #UnitExt, #RM, #AM, #LM:'),
      nl(Stream), write(Stream, 'Fin, '),
      fin_proof(Proof),
      getval(fin_ext,FE), write(Stream, FE), write(Stream, ', '),
      getval(fin_uext,FUE), write(Stream, FUE), write(Stream, ', '),
      getval(fin_right,FR), write(Stream, FR), write(Stream, ', '),
      getval(fin_anc,FA), write(Stream, FA), write(Stream, ', '),
      getval(fin_left,FL), write(Stream, FL), nl(Stream), nl(Stream)
	;
	  true %EBS Nonexistence of this line appears to be a bug
   ),
   (  printproof
   -> put_proof_head(Stream),
      put_proof(Stream,1,_,Proof),nl(Stream),
      write(Stream, 'Proof Done'),nl(Stream),nl(Stream)
   ;
      true
   )
), %EBS
   (batch -> close(output) ; true).

/* iterate drives the iterative deepening search process */
iterate(Query,V,N,D0,Min,Max,_Inc,Time,Stream,Proof):-
   Min =< Max,
   td_prove(Query,V,t(N,D0,Min),_,_,Proof),
   statistics(runtime,[Time1|_]),
   Time2 is Time1 - Time,
   (quiet -> true; write(Stream, 'y, '),        % mark final line of counts
    put_table(Min,Time2,Stream)). %EBS
iterate(Query,V,N,D0,Min,Max,Inc,Time,Stream,Proof):-
   first_order,
   statistics(runtime,[Time1|_]),
   Time2 is Time1 - Time,
   (batch -> Stream = output ; Stream = user),
   (quiet -> true ; %EBS
	write(Stream, 'n, '),             % mark line as not done
    put_table(Min,Time2,Stream)
   ),%EBS
   IncMin is Min + Inc,
   IncMin =< Max,
   iterate(Query,V,N,D0,IncMin,Max,Inc,Time,Stream,Proof).

/*********************************************************/
/* inference count procedures (for swi and quintus onl(Stream)y) */
/*********************************************************/
/* comment out for eclipse, which has built in destructive assignment */
/*
setval(C,0):- abolish(C,1),P =.. [C,0],assert(P).
my_incval(C):-
   P =.. [C,X],
   retract(P),
   X1 is X + 1,
   P1 =.. [C,X1],
   assert(P1),
   !.
getval(C,X):-
   P =.. [C,X],
   retract(P).
*/

/******************************/
/* proof printing procedures  */
/******************************/
put_table_heading(Stream):-
   nl(Stream),nl(Stream),
   problem(Problem),
   write(Stream, Problem),write(Stream,' Proof Information'),nl(Stream),
   write(Stream, 'Done,Depth,Cum Time'),
   (count ->
      write(Stream, ',#Res,#Right,#Anc,#Left,#Right_taut,#Anc_taut,'),
      write(Stream, ',#Left_taut,#Taut_Dis,#Merge_Dis')
   ;
      true
   ),
   nl(Stream).

put_table(Depth,CumTime,Stream):-
   write(Stream, Depth),write(Stream, ', '),put(Stream,9), %9 =ascii tab
   write(Stream, CumTime),write(Stream, ', '),put(Stream,9),
   (count ->
      getval(ext,C1),getval(right,C2),
      getval(anc,C3),getval(left,C4),
      getval(right_taut,C5),getval(anc_taut,C6),
      getval(left_taut,C7),getval(merge_dis,C8),
      getval(taut_dis,C9),
      write(Stream, C1),write(Stream, ', '),put(Stream,9),
      write(Stream, C2),write(Stream, ', '),put(Stream,9),
      write(Stream, C3),write(Stream, ', '),put(Stream,9),
      write(Stream, C4),write(Stream, ', '),put(Stream,9),
      write(Stream, C5),write(Stream, ', '),put(Stream,9),
      write(Stream, C6),write(Stream, ', '),put(Stream,9),
      write(Stream, C7),write(Stream, ', '),put(Stream,9),
      write(Stream, C8),write(Stream, ', '),put(Stream,9),
      write(Stream, C9)
   ;
      true
   ),
   nl(Stream),
   flush(Stream).

/* count the types of inferences taken in the final proof */
fin_proof([]):-!.
fin_proof([p(_,Type,_,0,_)|PfsRest]):-
   % closed leaf
   (  integer(Type)
   -> incval(fin_uext)
   ;  name(Type,TypeList),
	  %name(fin,  [102, 105, 110, 95]), %EBS
      append([102, 105, 110, 95],TypeList,FinTypeList),
      name(FinType,FinTypeList),
      incval(FinType)
   ),
   fin_proof(PfsRest).
fin_proof([p(_,_,_,_,PfsExt)|PfsRest]):-
   incval(fin_ext),
   fin_proof(PfsExt),
   fin_proof(PfsRest).

/* put proof as Stickel like proof, EXCEPT, goal number listed in the
   extension is simply the number of extensions in round brackets.
   It is an easy fix to get the goal numbers (future)
*/
put_proof_head(Stream):-
   nl(Stream),
   problem(Problem),
   write(Stream, 'Proof of '),write(Stream, Problem),nl(Stream),
   write(Stream, 'Goal#  Wff#  Wff Instance'),nl(Stream),
   write(Stream, '-----  ----  ------------'),nl(Stream).

put_proof(_,N,N,[]).
put_proof(Stream,N0,N2,[p(C,Type,L,0,_)|PfsRest]):-

   % put the closed leaf (0 extension)
   name(N0,N0List),
   length(N0List,N0Len),
   tab(Stream, 3 - N0Len),
   write(Stream, '['),write(Stream, N0),write(Stream, ']'),
   name(Type,TypeList),
   length(TypeList,TLen),
   tab(Stream, 5 - TLen),write(Stream, Type),
   tab(Stream, 3 + L * 3),write(Stream, C),
   nl(Stream),

   % put the rest of the proof
   N1 is N0 + 1,
   put_proof(Stream,N1,N2,PfsRest).
put_proof(Stream,N0,N3,[p(C,Type,L,Ext,PfsExt)|PfsRest]):-

   % put the current node
   name(N0,N0List),
   length(N0List,N0Len),
   tab(Stream, 3 - N0Len),
   write(Stream, '['),write(Stream, N0),write(Stream, ']'),
   name(Type,TypeList),length(TypeList,TLen),
   tab(Stream, 5 - TLen),write(Stream, Type),
   tab(Stream, 3 + L * 3),write(Stream, C),
   (Ext > 0 ->
      write(Stream, ' :- ('),write(Stream, Ext),write(Stream, ').')
   ; %else
      write(Stream, '.')
   ),
   nl(Stream),

   % put the rest of the proof
   N1 is N0 + 1,
   put_proof(Stream,N1,N2,PfsExt),
   put_proof(Stream,N2,N3,PfsRest).

/********************************/
/* General Prolog Utilities     */
/********************************/
/* delete(a1,a2 a3),delete next member of a2 that matches a1 into a3:
   arg1: element to test for
   arg2: input list
   arg3: list with first match of arg1 in arg2 off the list
*/
delete(A1,[A1|A3],A3).
delete(A1,[E|A2],[E|A3]):-
   delete(A1,A2,A3).

/* det_delete(a1,a2 a3), deterministicly delete first member of a2 that
   matches a1 into a3:
   arg1: element to test for
   arg2: input list
   arg3: list with first match of arg1 in arg2 off the list
*/
det_delete(A1,[A1|A3],A3):-!.
det_delete(A1,[E|A2],[E|A3]):-
   det_delete(A1,A2,A3).

/* delete_all(a1, a2, a3) deterministicly deletes all occurances of a1
   in a2 into a2:
   arg1: element to test for
   arg2: input list
   arg3: list with all occurances of a1 off of arg2
*/
delete_all(_,[],[]):-
   !.
delete_all(A,[A|T],T1):-
   delete_all(A,T,T1).
delete_all(A,[X|T],[X|T1]):-
   delete_all(A,T,T1).

/* union of the 1st and 2nd arguments returned in the 3rd argument;
   intersection elements are chosen from the 2nd arg,and if the 1st or
   2nd argument either contain duplicates then the 3rd element may as
   well.
*/
union([],L,L):- !.
union([H|T],L,R):-
   member(H,L),!,
   union(T,L,R).
union([H|T],L,[H|R]):-
   union(T,L,R).

/* call each member of the list */
call_list_mbrs([]):-!.
call_list_mbrs([X|Xs]):-
   call(X),
   call_list_mbrs(Xs).
